// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>
#include <OB/Properties.h>
#include <OB/NTService.h>

#include <Hello_impl.h>
#include <HelloWorldService.h>

#include <HelloMsg.h>

#ifdef HAVE_FSTREAM
#include <fstream>
#else
#include <fstream.h>
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace OB;

// ----------------------------------------------------------------------
// HelloWorldService constructor
// ----------------------------------------------------------------------

HelloWorldService::HelloWorldService()
    : NTService("HelloWorld", "Hello World Service", HELLO_MSG),
      stopped_(false)
{
}

// ----------------------------------------------------------------------
// HelloWorldService public member implementation
// ----------------------------------------------------------------------

void
HelloWorldService::start(int argc, char** argv)
{
    try
    {
	//
	// Create ORB
	//
	orb_ = CORBA::ORB_init(argc, argv);

	Properties::instance() -> setProperty("ooc.boa.port", "12345");
	Properties::instance() -> setProperty("ooc.boa.conc_model", 
					      "threaded");

	CORBA::Object_var poaObj = 
	    orb_ -> resolve_initial_references("RootPOA");
	PortableServer::POA_var rootPoa = PortableServer::POA::_narrow(poaObj);
	PortableServer::POAManager_var manager = rootPoa -> the_POAManager();

        //
	// Create implementation object
	//

	Hello_impl helloImpl;
	Hello_var hello = helloImpl._this();

	CORBA::String_var ior = orb_ -> object_to_string(hello);

	const char* refFile = "c:\\Hello.ref";
	ofstream out(refFile);
	if(out.fail())
	{
	    CORBA::String_var msg = 
		CORBA::string_dup("HelloWorld: can't open '");
	    msg += refFile;
	    msg += "': ";
	    msg += strerror(errno);
	    MessageViewer::instance() -> error(msg);
	    return;
	}

	out << ior << endl;
	out.close();

        //
        // Inform the status manager that the service is running.
        //
	statusUpdate(SERVICE_RUNNING);
        
	//
	// Run implementation
	//
        if(!stopped_)
	{
	    manager -> activate();
	    orb_ -> run();
	}

	PortableServer::ObjectId_var objId = 
	    rootPoa -> servant_to_id(&helloImpl);
	rootPoa -> deactivate_object(objId);
    }
    catch(CORBA::SystemException& ex)
    {
	CORBA::String_var err = ExceptionToString(ex);
	MessageViewer::instance() -> error(err);
    }
}

void
HelloWorldService::stop()
{
    //
    // The service manager calls this in a seperate thread. So first
    // we have inform JTC to adopt this thread.
    //
    JTCAdoptCurrentThread adopt;
    stopped_ = true;
    if(!CORBA::is_nil(orb_))
	orb_ -> shutdown(false);
}

