// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>
#include <OB/Reactor.h>

#include <Receiver_impl.h>

#include <ctype.h>

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

// ----------------------------------------------------------------------
// Receiver constructor and desctructor
// ----------------------------------------------------------------------

Receiver_impl::Receiver_impl(CORBA::ORB_ptr orb, Broadcaster_ptr server)
    : ReceiverBase_impl(orb, server), input_(CORBA::string_dup(""))
{
    //
    // Register with the server
    //
    _cxx_register();

    //
    // Register with the reactor
    //
    OB::Reactor* reactor = OB::Reactor::instance();
    reactor -> registerHandler(this, OB::EventRead, 0);
}

Receiver_impl::~Receiver_impl()
{
    //
    // Unregister with the reactor
    //
    OB::Reactor* reactor = OB::Reactor::instance();
    reactor -> unregisterHandler(this);
}

// ----------------------------------------------------------------------
// Print message
// ----------------------------------------------------------------------

void
Receiver_impl::print(const char* text)
    throw(CORBA::SystemException)
{
    //
    // Remove active input line temporarily
    //
    for(CORBA::ULong i = 0 ; i < strlen(input_) ; i++)
	cout << "\b \b";

    //
    // Print the text
    //
    ReceiverBase_impl::print(text);

    //
    // Restore active input line
    //
    cout << input_ << flush;
}

// ----------------------------------------------------------------------
// ORBacus specific event handling
// ----------------------------------------------------------------------
 
void
Receiver_impl::handleStop()
{
    // Do nothing here
}

void
Receiver_impl::handleEvent(CORBA::ULong mask)
{
    assert(mask == OB::EventRead);

    char c;
    cin.read(&c, 1);

    switch(c)
    {
    case '\b':
    {
	if(strlen(input_) > 0)
	{
	    input_[strlen(input_) - 1] = '\0';
	    cout << "\b \b" << flush;
	}
	break;
    }

    case '\n':
    {
	cout << c << flush;
	CORBA::String_var s = input_;
	input_[0] = '\0';
	parse(s);
	break;
    }

    default:
    {
	if(isprint(c))
	{
	    cout << c << flush;
	    input_ += c;
	}
	break;
    }
    }
}

