// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>
#include <OB/OCI.h>
#include <OB/OCI_IIOP.h>
#include <Broadcaster_impl.h>

#include <stdlib.h>
#include <time.h>

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

int
main(int argc, char* argv[], char*[])
{
    try
    {
	//
	// Create ORB
	//
	CORBA::ORB_var orb = CORBA::ORB_init(argc, argv);

        //
        // Resolve Root POA
        //
        CORBA::Object_var poaObj =
	    orb -> resolve_initial_references("RootPOA");
        PortableServer::POA_var rootPoa = PortableServer::POA::_narrow(poaObj);
 
        //
        // Create persistent POA
        //
        PortableServer::POAManager_var manager = rootPoa -> the_POAManager();
 
        CORBA::PolicyList policies(2);
        policies.length(2);
        policies[0] =
            rootPoa -> create_lifespan_policy(PortableServer::PERSISTENT);
        policies[1] =
            rootPoa -> create_id_assignment_policy(PortableServer::USER_ID);
        PortableServer::POA_var persistentPOA =
            rootPoa -> create_POA("persistent", manager, policies);
	
	//
	// Create implementation object
	//
	Broadcaster_impl broadcasterImpl(orb);

        PortableServer::ObjectId_var id =
            PortableServer::string_to_ObjectId("broadcaster");
        persistentPOA -> activate_object_with_id(id, &broadcasterImpl);
 
        CORBA::Object_var broadcasterObj =
            persistentPOA -> create_reference_with_id(id,
						      "IDL:Broadcaster:1.0");
	Broadcaster_var p = Broadcaster::_narrow(broadcasterObj);
	
	//
	// Print "server started" message
	//
	time_t t;
	char* d;
	t = time(0);
	d = asctime(localtime(&t));
	assert(d && strlen(d) > 0);
	d[strlen(d) - 1] = '\0';
	cout << "*** Server started at " << d << endl;
	
	/*	//
	// Display the host and port
	//
	OCI_AccRegistry_var registry = boa -> get_acc_registry();
	OCI_AcceptorSeq_var acceptors = registry -> get_acceptors();
	
	for(CORBA::ULong i = 0 ; i < acceptors -> length() ; i++)
	{
	    OCI_AcceptorInfo_var info = acceptors[i] -> get_info();
	    OCI_IIOP_AcceptorInfo_var iiopInfo =
		OCI_IIOP_AcceptorInfo::_narrow(info);
	    
	    if(!CORBA::is_nil(iiopInfo))
	    {
		CORBA::String_var host = iiopInfo -> host();
		CORBA::UShort port = iiopInfo -> port();
		
		cout << "*** Running on host `" << host
		     << "', port number " << port << endl;
	    }
	}
	*/
	//
	// Run implementation
	//
        manager -> activate();
        orb -> run();
    }
    catch(CORBA::SystemException& ex)
    {
	OB::PrintException(ex);
	return 1;
    }

    return 0;
}
