// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>

#include <DefaultFactory_impl.h>
#include <LocatorFactory_impl.h>

#include <stdlib.h>
#include <errno.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

int
main(int argc, char* argv[], char*[])
{
    try
    {
        CORBA::Object_var obj;
        PortableServer::ObjectId_var oid;
        CORBA::String_var ref;

        //
        // Create ORB
        //
        CORBA::ORB_var orb = CORBA::ORB_init(argc, argv);

        //
        // Because OB::Timer is used, the Reactive concurrency model
        // is required
        //
        orb -> server_conc_model(CORBA::ORB::ServerConcModelReactive);

        //
        // Get Root POA
        //
        obj = orb -> resolve_initial_references("RootPOA");
        PortableServer::POA_var root = PortableServer::POA::_narrow(obj);

        //
        // Create persistent POA
        //
        PortableServer::POAManager_var poaManager = root -> the_POAManager();
        CORBA::PolicyList policies;
        policies.length(2);
        policies[0] =
            root -> create_id_assignment_policy(PortableServer::USER_ID);
        policies[1] =
            root -> create_lifespan_policy(PortableServer::PERSISTENT);
        PortableServer::POA_var poa =
            root -> create_POA("persistent", poaManager, policies);

        //
        // Create default servant factory implementation object
        //
        DefaultFactory_impl defFactory(orb, poa);
        oid = PortableServer::string_to_ObjectId("def_factory");
        poa -> activate_object_with_id(oid, &defFactory);
        obj = poa -> id_to_reference(oid);
        ref = orb -> object_to_string(obj);

        //
        // Save reference in file
        //
        {
            const char* myRef = "dfactory.ref";
            ofstream out(myRef);
            if(out.fail())
            {
                cerr << argv[0] << ": can't open `" << myRef << "': "
                     << strerror(errno) << endl;
                return EXIT_FAILURE;
            }

            out << ref;
            out.close();
        }

        //
        // Create servant locator factory implementation object
        //
        LocatorFactory_impl locFactory(orb, poa);
        oid = PortableServer::string_to_ObjectId("loc_factory");
        poa -> activate_object_with_id(oid, &locFactory);
        obj = poa -> id_to_reference(oid);
        ref = orb -> object_to_string(obj);

        //
        // Save reference in file
        //
        {
            const char* myRef = "lfactory.ref";
            ofstream out(myRef);
            if(out.fail())
            {
                cerr << argv[0] << ": can't open `" << myRef << "': "
                     << strerror(errno) << endl;
                return EXIT_FAILURE;
            }

            out << ref;
            out.close();
        }

        //
        // Run implementation
        //
        poaManager -> activate();
        orb -> run();
    }
    catch(CORBA::SystemException& ex)
    {
        OB::PrintException(ex);
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}
