// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>

#include <Hello_skel_tie.h>
#include <Hello_impl_tie.h>

#include <stdlib.h>
#include <errno.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

#ifndef HAVE_NO_EXPLICIT_TEMPLATES
template class POA_Hello_tie< Hello_impl_tie >;
#else
#ifdef HAVE_PRAGMA_DEFINE
#pragma define(POA_Hello_tie< Hello_impl_tie >)
#endif
#endif

int
main(int argc, char* argv[], char*[])
{
    try
    {
	//
	// Create ORB
	//
	CORBA::ORB_var orb = CORBA::ORB_init(argc, argv);

	//
	// Resolve Root POA
	//
	CORBA::Object_var poaObj =
            orb -> resolve_initial_references("RootPOA");
	PortableServer::POA_var rootPOA = PortableServer::POA::_narrow(poaObj);

        //
        // Get a reference to the POA manager
        //
	PortableServer::POAManager_var manager = rootPOA -> the_POAManager();

	//
	// Create implementation object
	//
	Hello_impl_tie helloTie;
	POA_Hello_tie< Hello_impl_tie > helloImpl(helloTie, rootPOA);
	Hello_var hello = helloImpl._this();

	//
	// Save reference
	//
	CORBA::String_var s = orb -> object_to_string(hello);

	const char* refFile = "Hello.ref";
	ofstream out(refFile);
	if(out.fail())
	{
	    cerr << argv[0] << ": can't open `" << refFile << "': "
		 << strerror(errno) << endl;
	    return 1;
	}

	out << s << endl;
	out.close();

	//
	// Run implementation
	//
	manager -> activate();
	orb -> run();
    }
    catch(CORBA::SystemException& ex)
    {
	OB::PrintException(ex);
	return 1;
    }

    return 0;
}
