// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>
#include <Peer_impl.h>

#include <stdlib.h>
#include <errno.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

int
main(int argc, char* argv[], char*[])
{
    try
    {
	CORBA::Object_var obj;

	//
	// Create ORB
	//
	CORBA::ORB_var orb = CORBA::ORB_init(argc, argv);

	//
	// Resolve the Root POA
	//
	obj = orb -> resolve_initial_references("RootPOA");
	PortableServer::POA_var root = PortableServer::POA::_narrow(obj);
	PortableServer::POAManager_var poaManager = root -> the_POAManager();

	//
	// Set the concurrency model
	//
#ifdef HAVE_JTC
	if(orb -> conc_model() != CORBA::ORB::ConcModelThreaded ||
	   (orb -> server_conc_model() != CORBA::ORB::ServerConcModelThreadPerRequest &&
	    orb -> server_conc_model() != CORBA::ORB::ServerConcModelThreadPool))
	{
	    orb -> conc_model(CORBA::ORB::ConcModelReactive);
	    orb -> server_conc_model(CORBA::ORB::ServerConcModelReactive);
	}
#else
	orb -> conc_model(CORBA::ORB::ConcModelReactive);
	orb -> server_conc_model(CORBA::ORB::ServerConcModelReactive);
#endif

	//
	// Create implementation object
	//
	Peer_impl peerImpl;
	Peer_var p1 = peerImpl._this();
    
	//
	// Save reference
	//
	CORBA::String_var s = orb -> object_to_string(p1);

	const char* refFile = "Peer.ref";
	ofstream out(refFile);
	if(out.fail())
	{
	    cerr << argv[0] << ": can't open `" << refFile << "': "
		 << strerror(errno) << endl;
	    return 1;
	}

	out << s << endl;
	out.close();

	//
	// Run implementation
	//
	poaManager -> activate();
	orb -> run();
    }
    catch(CORBA::SystemException& ex)
    {
	OB::PrintException(ex);
	return 1;
    }

    return 0;
}
