// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>
#include <OB/Timer.h>

#include <Peer_impl.h>

#include <time.h>
#include <stdlib.h>
#include <errno.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

int
main(int argc, char* argv[], char*[])
{
    try
    {
	CORBA::Object_var obj;

	//
	// Create ORB
	//
	CORBA::ORB_var orb = CORBA::ORB_init(argc, argv);

	//
	// Resolve the Root POA
	//
	obj = orb -> resolve_initial_references("RootPOA");
	PortableServer::POA_var root = PortableServer::POA::_narrow(obj);
	PortableServer::POAManager_var poaManager = root -> the_POAManager();

	//
	// Set the concurrency model
	//
#ifdef HAVE_JTC
	if(orb -> conc_model() != CORBA::ORB::ConcModelThreaded ||
	   (orb -> server_conc_model() != CORBA::ORB::ServerConcModelThreadPerRequest &&
	    orb -> server_conc_model() != CORBA::ORB::ServerConcModelThreadPool))
	{
	    orb -> conc_model(CORBA::ORB::ConcModelReactive);
	    orb -> server_conc_model(CORBA::ORB::ServerConcModelReactive);
	}
#else
	orb -> conc_model(CORBA::ORB::ConcModelReactive);
	orb -> server_conc_model(CORBA::ORB::ServerConcModelReactive);
#endif

	//
	// Create implementation object
	//
	Peer_impl peerImpl;
	Peer_var p2 = peerImpl._this();

	//
	// Get other object
	//
	const char* refFile = "Peer.ref";
	ifstream in; // Must use open(name), not ifstream in(name) (VC++ bug)
	in.open(refFile);
	if(in.fail())
	{
	    cerr << argv[0] << ": can't open `" << refFile << "': "
		 << strerror(errno) << endl;
	    return 1;
	}

	char s[1000];
	in >> s;

	obj = orb -> string_to_object(s);
	assert(!CORBA::is_nil(obj));

	Peer_var p1 = Peer::_narrow(obj);
	assert(!CORBA::is_nil(p1));

	//
	// Initialize the servers so that incoming requests are
	// handled from this point on
	//
	orb -> init_servers();
	poaManager -> activate();

	while(true)
	{
	    //
	    // Start recursive decrement
	    //
	    cout << "Enter a number (0 to exit): ";
	    CORBA::Long value;
	    cin >> value;
	    
	    if(value <= 0)
		break;
	    
	    //
	    // Run demo
	    //
	    timeval tv = OB::TimerList::timeNow();
	    p1 -> decrement(value, p2);
	    tv = OB::TimerList::timeSub(OB::TimerList::timeNow(), tv);
	    cout << "Time needed: "
		 << ((double)tv.tv_sec + (double)tv.tv_usec / 1000000.0)
		 << endl;
	}

	//
	// Deactivate implementation
	//
	orb -> shutdown(CORBA::TRUE);
    }
    catch(CORBA::SystemException& ex)
    {
	OB::PrintException(ex);
	return 1;
    }

    return 0;
}
