// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/OCI_IIOP.h>
#include <OB/Util.h>

#include <Hello_impl.h>

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

class CloseCB_impl : public OCI::CloseCB
{
public:

    virtual void close_cb(OCI::TransportInfo_ptr transport_info)
    {
        OCI::IIOP::TransportInfo_var iiopInfo =
            OCI::IIOP::TransportInfo::_narrow(transport_info);

        if(!CORBA::is_nil(iiopInfo))
        {
            OCI::IIOP::InetAddr_var remoteAddr = iiopInfo -> remote_addr();
            CORBA::UShort remotePort = iiopInfo -> remote_port();

            cout << "Closing connection from: "
                 << (int)remoteAddr[0] << '.' << (int)remoteAddr[1] << '.'
                 << (int)remoteAddr[2] << '.' << (int)remoteAddr[3]
                 << ":" << remotePort << endl;
        }
    }
};

class AcceptCB_impl : public OCI::AcceptCB
{
public:
    AcceptCB_impl() { cout << "AcceptCB_impl instantiated" << endl; }
    ~AcceptCB_impl() { cout << "AcceptCB_impl destroyed" << endl; }

    virtual void accept_cb(OCI::TransportInfo_ptr transport_info)
    {
        OCI::IIOP::TransportInfo_var iiopInfo =
            OCI::IIOP::TransportInfo::_narrow(transport_info);

        if(!CORBA::is_nil(iiopInfo))
        {
            OCI::IIOP::InetAddr_var remoteAddr = iiopInfo -> remote_addr();
            CORBA::UShort remotePort = iiopInfo -> remote_port();

            cout << "Connection from: "
                 << (int)remoteAddr[0] << '.' << (int)remoteAddr[1] << '.'
                 << (int)remoteAddr[2] << '.' << (int)remoteAddr[3]
                 << ":" << remotePort << endl;

            char c;
            do
            {
                cout << "Refuse connection (y/n)? ";
                cin >> c;
            }
            while(c != 'y' && c != 'n');

            if(c == 'y')
                throw CORBA::NO_PERMISSION();

            //
            // Add the close callback object
            //
            OCI::CloseCB_var closeCB = new CloseCB_impl();
            iiopInfo -> add_close_cb(closeCB);
        }
    }
};

int
main(int argc, char* argv[], char*[])
{
    try
    {
        CORBA::Object_var obj;

        //
        // Create ORB
        //
        CORBA::ORB_var orb = CORBA::ORB_init(argc, argv);

        //
        // Resolve the Root POA
        //
        obj = orb -> resolve_initial_references("RootPOA");
        PortableServer::POA_var root = PortableServer::POA::_narrow(obj);
        PortableServer::POAManager_var poaManager = root -> the_POAManager();

        //
        // Get the OCI Current object
        //
        obj = orb -> resolve_initial_references("OCICurrent");
        OCI::Current_var ociCurrent = OCI::Current::_narrow(obj);

        //
        // Create implementation object
        //
        Hello_impl helloImpl(ociCurrent);
        Hello_var p = helloImpl._this();

        //
        // Save reference
        //
        CORBA::String_var s = orb -> object_to_string(p);

        const char* refFile = "Hello.ref";
        ofstream out(refFile);
        if(out.fail())
        {
            cerr << argv[0] << ": can't open `" << refFile << "': "
                 << strerror(errno) << endl;
            return 1;
        }

        out << s << endl;
        out.close();

        //
        // Display the host and port
        //
        OCI::AccRegistry_var registry = orb -> get_acc_registry();
        OCI::AcceptorSeq_var acceptors = registry -> get_acceptors();

        for(CORBA::ULong i = 0 ; i < acceptors -> length() ; i++)
        {
            OCI::AcceptorInfo_var info = acceptors[i] -> get_info();
            OCI::IIOP::AcceptorInfo_var iiopInfo =
                OCI::IIOP::AcceptorInfo::_narrow(info);

            if(!CORBA::is_nil(iiopInfo))
            {
                CORBA::String_var host = iiopInfo -> host();
                CORBA::UShort port = iiopInfo -> port();

                cout << "*** Running on host `" << host
                     << "', port number " << port << endl;

                //
                // Add the accept callback object
                //
                OCI::AcceptCB_var acceptorCB = new AcceptCB_impl();
                iiopInfo -> add_accept_cb(acceptorCB);
            }
        }

        //
        // Run implementation
        //
        poaManager -> activate();
        orb -> run();
    }
    catch(CORBA::SystemException& ex)
    {
        OB::PrintException(ex);
        return 1;
    }

    return 0;
}
