// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>
#include <OB/Timer.h>

#include <Nested_impl.h>

#include <stdlib.h>
#include <errno.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;

int
main(int argc, char* argv[], char*[])
{
    try
    {
	Object_var obj;

	//
	// Create ORB
	//
	ORB_var orb = ORB_init(argc, argv);

	//
	// Resolve the Root POA
	//
	obj = orb -> resolve_initial_references("RootPOA");
	PortableServer::POA_var root = PortableServer::POA::_narrow(obj);
	PortableServer::POAManager_var poaManager = root -> the_POAManager();

	//
	// Get nesting level
	//
	if(argc < 2)
	{
	    cerr << "Usage: " << argv[0] << " nesting-level" << endl;
	    return 1;
	}
	int level = atoi(argv[1]);
	if(level <= 0 || (level % 2) == 1)
	{
	    cerr << argv[0]
		 << ": nesting level must be positive and even"
		 << endl;
	    return 1;
	}
	
	//
	// Get "I" object
	//
	const char* refFile = "I.ref";
	ifstream in; // Must use open(name), not ifstream in(name) (VC++ bug)
	in.open(refFile);
	if(in.fail())
	{
	    cerr << argv[0] << ": can't open `" << refFile << "': "
		 << strerror(errno) << endl;
	    return 1;
	}
	
	char s[1000];
	in >> s;
	
	obj = orb -> string_to_object(s);
	assert(!is_nil(obj));
	
	I_var i = I::_narrow(obj);
	assert(!is_nil(i));
	
	//
	// Create implementation object
	//
	// We divide level by two since we want this to be the
	// nesting level of the servers.
	//
	I_impl impl(i, level / 2);
	I_var p = impl._this();

	//
	// Initialize the servers so that incoming requests are
	// handled from this point on
	//
	orb -> init_servers();
	poaManager -> activate();

	//
	// Tell i to call back into the implementation object p.
	//
	// This also has the side effect of forcing the thread pool to
	// be initialized (if the server is, in fact, using a thread
	// pool)
        //
	timeval tv = OB::TimerList::timeNow();
	i -> connect(p);
	tv = OB::TimerList::timeSub(OB::TimerList::timeNow(), tv);
	cout << "Time needed for initialization: "
	     << ((double)tv.tv_sec + (double)tv.tv_usec / 1000000.0) * 1000.0
 	     << " ms" << endl;

	//
	// Run the performance test
	//
	tv = OB::TimerList::timeNow();
	i -> f();
	tv = OB::TimerList::timeSub(OB::TimerList::timeNow(), tv);
	cout << "Total time needed: "
	     << ((double)tv.tv_sec + (double)tv.tv_usec / 1000000.0) * 1000.0
             << " ms" << endl;
	cout << "Time per method call: "
	     << (((double)tv.tv_sec + (double)tv.tv_usec / 1000000.0)
		/ (double)level) * 1000.0 << " ms" << endl;
    }
    catch(SystemException& ex)
    {
	OB::PrintException(ex);
	return 1;
    }

    return 0;
}
