// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>
#include <Finance.h>
#include <EmployeeFactory_impl.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

int main(int argc, char* argv[])
{
    try
    {
        CORBA::ORB_var orb = CORBA::ORB_init(argc, argv);

        //
        // Instantiate value factory
        //
        // NOTE: There's no need to register this factory with the ORB
        //       unless we will be receiving a Employee value.
        //
        EmployeeFactory_impl* factory = new EmployeeFactory_impl();
        CORBA::ValueFactoryBase_var factoryVar = factory;

        //
        // Obtain the server's object reference from a file
        //
        const char* refFile = "Payroll.ref";
        ifstream in(refFile);
        if(in.fail())
        {
            cerr << argv[0] << ": can't open `" << refFile << "'" << endl;
            return 1;
        }

        char buff[1024];
        in >> buff;
        CORBA::Object_var obj = orb -> string_to_object(buff);
        in.close();

        Payroll_var payroll = Payroll::_narrow(obj);
        assert(!CORBA::is_nil(payroll));

        //
        // Test with null value
        //
        payroll -> pay_employee((Employee*)0);

        //
        // Test with real value
        //
        Employee_var emp1 = factory -> create((CORBA::Float)1000.0);
        emp1 -> emp_id(999);
        emp1 -> name((const char*)"John Smith");
        Employee::Address addr;
        addr.street = CORBA::string_dup("1234 Maiden Lane");
        addr.city = CORBA::string_dup("Somewhere");
        addr.state = CORBA::string_dup("CA");
        addr.zip = CORBA::string_dup("95000");
        emp1 -> addr(addr);
        emp1 -> ssn((const char*)"123-45-6789");
        cout << "Paying " << emp1 -> name() << endl;
        payroll -> pay_employee(emp1);
    }
    catch(const CORBA::SystemException& ex)
    {
        OB::PrintException(ex);
        return 1;
    }

    return 0;
}
