// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>
#include <Finance_skel.h>
#include <EmployeeFactory_impl.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

//
// Implementation of Payroll interface
//
class Payroll_impl : public virtual POA_Payroll
{
public:
    virtual void pay_employee(Employee* emp)
        throw(CORBA::SystemException)
    {
        if(emp != 0)
            emp -> print_check();
    }
};


int main(int argc, char* argv[])
{
    try
    {
        CORBA::ORB_var orb = CORBA::ORB_init(argc, argv);

        //
        // Create & install value factory - if we forget to do this,
        // the client will receive a MARSHAL exception when attempting
        // to transmit a Employee value.
        //
        CORBA::ValueFactoryBase_var factory = new EmployeeFactory_impl();
        CORBA::ValueFactoryBase_var oldFactory =
            orb -> register_value_factory("IDL:Employee:1.0", factory);

        CORBA::Object_var obj;
        obj = orb -> resolve_initial_references("RootPOA");
        PortableServer::POA_var root = PortableServer::POA::_narrow(obj);
        PortableServer::POAManager_var manager = root -> the_POAManager();

        //
        // Instantiate and activate the Payroll servant
        //
        Payroll_impl impl;
        root -> activate_object(&impl);
        obj = root -> servant_to_reference(&impl);

        //
        // Export the object reference to a file
        //
        const char* refFile = "Payroll.ref";
        ofstream out(refFile);
        if(out.fail())
        {
            cerr << argv[0] << ": can't open `" << refFile << "'" << endl;
            return 1;
        }

        CORBA::String_var ref = orb -> object_to_string(obj);
        out << ref << endl;
        out.close();

        manager -> activate();
        orb -> run();
    }
    catch(const CORBA::SystemException& ex)
    {
        OB::PrintException(ex);
        return 1;
    }

    return 0;
}
