// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_CODE_CONVERTER_H
#define OB_CODE_CONVERTER_H

namespace OB
{

class CodeConverterBase
{
    //
    // Source and destination code set
    //
    const CodeSetDatabase::CodeSetInfo* srcSet_;
    const CodeSetDatabase::CodeSetInfo* dstSet_;

public:

    CodeConverterBase(const CodeSetDatabase::CodeSetInfo* = 0,
		      const CodeSetDatabase::CodeSetInfo* = 0);
    virtual ~CodeConverterBase() { };

    //
    // Get source/destination code set
    //
    const CodeSetDatabase::CodeSetInfo* getSource() const
	{ return srcSet_; }
    const CodeSetDatabase::CodeSetInfo* getDest() const
	{ return dstSet_; }

    //
    // Read/write character
    //
    virtual CORBA::Char read_char(const CORBA::Octet*&)
	throw(CORBA::DATA_CONVERSION) = 0;
    virtual void write_char(CORBA::Octet*&, CORBA::Char)
	throw(CORBA::DATA_CONVERSION) = 0;

    //
    // Read/write/count wide character
    //
    virtual CORBA::WChar read_wchar(const CORBA::Octet*&)
	throw(CORBA::DATA_CONVERSION) = 0;
    virtual void write_wchar(CORBA::Octet*&, CORBA::WChar)
	throw(CORBA::DATA_CONVERSION) = 0;
    virtual CORBA::ULong count_wchar(CORBA::WChar) = 0;
};


class CodeConverter : public CodeConverterBase
{
    //
    // Unicode character mapping tables
    //
    const CodeSetDatabase::CharMapInfo* srcMap_;
    const CodeSetDatabase::CharMapInfo* dstMap_;

public:

    CodeConverter(const CodeSetDatabase::CodeSetInfo*,
		  const CodeSetDatabase::CodeSetInfo*);
    virtual ~CodeConverter() { };

    //
    // Read/write character
    //
    CORBA::Char read_char(const CORBA::Octet*&)
	throw(CORBA::DATA_CONVERSION);
    void write_char(CORBA::Octet*&, CORBA::Char)
	throw(CORBA::DATA_CONVERSION);

    //
    // Read/write/count wide character
    //
    CORBA::WChar read_wchar(const CORBA::Octet*&)
	throw(CORBA::DATA_CONVERSION);
    void write_wchar(CORBA::Octet*&, CORBA::WChar)
	throw(CORBA::DATA_CONVERSION);
    CORBA::ULong count_wchar(CORBA::WChar);

    //
    // Convert via Unicode
    //
    void convert(CORBA::WChar&);
};


class LocalCodeConverter : public CodeConverterBase
{
    //
    // The LocalCodeConverter singleton
    //
    static OB_IMPORT LocalCodeConverter* instance_;

public:

    //
    // Get singleton instance
    //
    static LocalCodeConverter* instance();

    //
    // Read/write character
    //
    CORBA::Char read_char(const CORBA::Octet*& in)
	throw(CORBA::DATA_CONVERSION) { return *in++; }
    void write_char(CORBA::Octet*& out, CORBA::Char value)
	throw(CORBA::DATA_CONVERSION) { *out++ = value; }

    //
    // Read/write/count wide character
    //
    CORBA::WChar read_wchar(const CORBA::Octet*&)
	throw(CORBA::DATA_CONVERSION);
    void write_wchar(CORBA::Octet*&, CORBA::WChar)
	throw(CORBA::DATA_CONVERSION);
    CORBA::ULong count_wchar(CORBA::WChar) { return sizeof(CORBA::WChar) + 1; }
};

} // End of namespace OB

#endif
