// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OCI_BUFFER_H
#define OCI_BUFFER_H

namespace OCI
{

class Buffer;
typedef Buffer* Buffer_ptr;
typedef Buffer_ptr BufferRef;

} // End of namespace OCI

void OBDuplicate(OCI::Buffer_ptr);
void OBRelease(OCI::Buffer_ptr);

namespace OCI
{

//
// The Buffer class
//
typedef OB::ObjVar< Buffer > Buffer_var;

class Buffer : public OB::SimpleRefCount
{
    CORBA::ULong pos_;
    OB::Buffer buf_;

public:

    Buffer() : pos_(0) { }
    Buffer(CORBA::ULong n) : pos_(0), buf_(n) { }
    virtual ~Buffer() { }
    
    //
    // Standard IDL to C++ Mapping
    //
    static inline Buffer_ptr _duplicate(Buffer_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline Buffer_ptr _nil()
    { return 0; }
    
    CORBA::Octet* data() { return buf_.data; }
    CORBA::Octet* rest() { return buf_.data + pos_; }
    
    CORBA::ULong length() { return buf_.len; }
    CORBA::ULong rest_length() { return buf_.len - pos_; }
    
    CORBA::ULong pos() { return pos_; }
    void pos(CORBA::ULong pos) { pos_ = pos; }
    void advance(CORBA::ULong delta) { pos_ += delta; }

    CORBA::Boolean is_full() { return pos_ >= buf_.len; }

    //
    // Additional ORBacus specific functions
    //
    OB::Buffer& buf() { return buf_; }
    void alloc(CORBA::ULong n) { buf_.alloc(n); }
    void realloc(CORBA::ULong n) { buf_.realloc(n); }
};

typedef OB::ObjSeq< Buffer, int > BufferSeq;
typedef OB::SeqVar< OB::ObjSeq< Buffer, int > > BufferSeq_var;

} // End of namespace OCI

namespace CORBA
{

inline void
release(OCI::Buffer_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OCI::Buffer_ptr p)
{
    return p == 0;
}

} // End of namespace CORBA

#endif
