// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_SERVER_H
#define OB_SERVER_H

namespace OB
{

class Server;
typedef Server* Server_ptr;
typedef Server_ptr ServerRef;

} // End of namespace OB

void OBDuplicate(OB::Server_ptr);
void OBRelease(OB::Server_ptr);

namespace OB
{

//
// The OCI object key type
//
// TODO: Why is this needed here?
//typedef OB_OctetSeq OCI_ObjectKey;

//
// The Server class
//
typedef ObjVar< Server > Server_var;
typedef ObjSeq< Server, int > ServerSeq;
typedef SeqVar< ObjSeq< Server, int > > ServerSeq_var;

class Server : public RefCount
{
    //
    // Hide copy-constructor and assignment operator
    //
    Server(const Server&);
    void operator=(const Server&);

protected:

    Server() { }
    virtual ~Server() { }

public:

    static inline Server_ptr _duplicate(Server_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline Server_ptr _nil()
    { return 0; }

    //
    // Destroy the server
    //
    virtual void destroy() = 0;

    //
    // Get the OCI acceptor
    //
    virtual OCI::Acceptor_ptr acceptor() = 0;

    //
    // Run the server. This may block or not block, depending on the
    // server kind, The return value is a "done" flag, meaning that
    // the server blocked to do it's work and can now be released.
    //
    virtual bool run() = 0;

    //
    // Disable the server
    //
    virtual void disable() = 0;
};

} // End of namespace OB

namespace CORBA
{
// XXX Are these needed in CORBA?

inline void
release(OB::Server_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::Server_ptr p)
{
    return p == 0;
}

} // End of namespace CORBA

#endif
