// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_VALUE_FACTORY_H
#define OB_VALUE_FACTORY_H

namespace CORBA
{

class ValueBase;
class ValueFactoryBase;

typedef ValueFactoryBase* ValueFactory;

class ValueFactoryBase : public virtual OB::RefCount
{
    virtual ValueBase* create_for_unmarshal() = 0;
    friend class OB::InputStream;

protected:
    ValueFactoryBase();

public:
    virtual ~ValueFactoryBase();
    virtual void _add_ref();
    virtual void _remove_ref();

    static ValueFactory _downcast(ValueFactory);

    virtual void* _OB_downcastHelp(const char*) const;
    virtual const char* _OB_typeId() const;

    static ValueBase* _OB_copyValue(ValueBase*, const char*);

    struct BooleanValue
    {
        BooleanValue(Boolean b) : value(b) {}
        Boolean value;
    };

    struct CharValue
    {
        CharValue(Char c) : value(c) {}
        Char value;
    };

    struct OctetValue
    {
        OctetValue(Octet o) : value(o) {}
        Octet value;
    };

    struct WCharValue
    {
        WCharValue(WChar wc) : value(wc) {}
        WChar value;
    };
};

} // End of namespace CORBA

void OBRelease(CORBA::ValueFactoryBase*);
void OBDuplicate(CORBA::ValueFactoryBase*);

namespace CORBA
{

typedef OB::ObjVar< ValueFactoryBase > ValueFactoryBase_var;

} // End of namespace CORBA

#endif
