// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/Any.h>
#include <OB/OBV.h>
#include <OB/IFR1.h>
#include <OB/IFR4.h>
#include <OB/AbstractBase.h>
#include <OB/POAServantBase.h>
#include <OB/ValueBase.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Template instantiations
// ----------------------------------------------------------------------

#ifndef HAVE_NO_EXPLICIT_TEMPLATES
template class OB::ObjVar< AbstractBase >;
#else
#ifdef HAVE_PRAGMA_DEFINE
#pragma define(OB::ObjVar< AbstractBase >)
#endif
#endif

// ----------------------------------------------------------------------
// AbstractBase constructor and destructor
// ----------------------------------------------------------------------

AbstractBase::AbstractBase()
{
}

AbstractBase::AbstractBase(const AbstractBase&)
{
}
    
AbstractBase::~AbstractBase()
{
}

// ----------------------------------------------------------------------
// AbstractBase public member functions
// ----------------------------------------------------------------------

AbstractBase_ptr
AbstractBase::_duplicate(AbstractBase_ptr p)
{
    if(p)
    {
        Object_ptr obj = p -> _OB_toObject();
        if(!is_nil(obj))
        {
            Object::_duplicate(obj);
        }
        else
        {
            ValueBase* vb = p -> _OB_toValue();
            assert(vb != 0); // XXX exception?
            add_ref(vb);
        }
    }

    return p;
}

AbstractBase_ptr
AbstractBase::_narrow(AbstractBase_ptr p)
{
    return _duplicate(p);
}

AbstractBase_ptr
AbstractBase::_nil()
{
    return 0;
}

Object_ptr
AbstractBase::_OB_toObject()
{
    return Object::_nil();
}

ValueBase*
AbstractBase::_OB_toValue()
{
    return 0;
}

Boolean
CORBA::is_nil(AbstractBase_ptr p)
{
    return (p == 0);
}

void
CORBA::release(AbstractBase_ptr p)
{
    if(p)
    {
        Object_ptr obj = p -> _OB_toObject();
        if(!is_nil(obj))
        {
            release(obj);
        }
        else
        {
            ValueBase* vb = p -> _OB_toValue();
            assert(vb != 0); // XXX exception?
            remove_ref(vb);
        }
    }
}

void
OBRelease(CORBA::AbstractBase_ptr p)
{
    CORBA::release(p);
}

void
OBDuplicate(CORBA::AbstractBase_ptr p)
{
    CORBA::AbstractBase::_duplicate(p);
}
