// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_COLLOCATED_CLIENT_H
#define OB_COLLOCATED_CLIENT_H

namespace OB
{

class CollocatedDowncall;
typedef CollocatedDowncall* CollocatedDowncall_ptr;

class CollocatedClient;
typedef CollocatedClient* CollocatedClient_ptr;

} // End of namespace OB

void OBDuplicate(OB::CollocatedDowncall_ptr);
void OBRelease(OB::CollocatedDowncall_ptr);
void OBDuplicate(OB::CollocatedClient_ptr);
void OBRelease(OB::CollocatedClient_ptr);

namespace OB
{

//
// Forward declarations
//
template <class K, class T> class Hashtable;

typedef OB::ObjVar< CollocatedDowncall > CollocatedDowncall_var;

//
// The CollocatedClient class
//
typedef OB::ObjVar< CollocatedClient > CollocatedClient_var;

class CollocatedClient : public OB::Client
{
    //
    // Hide copy-constructor and assignment operator
    //
    CollocatedClient(const CollocatedClient&);
    void operator=(const CollocatedClient&);

private:
    
#ifdef HAVE_JTC
    JTCMutex idMutex_;
#endif
    CORBA::ULong id_;

protected:
    
    //
    // Return the next available request id
    //
    CORBA::ULong nextId();

public:
    
    //
    // This struct is shared between requests and replies. Some
    // compilers need this to be public.
    //
    struct Request
    {
        StubImplBase_var stub;
        CORBA::String_var op;
        Buffer buf;
        CORBA::ULong reqId;
        long status;
    };
    
    typedef Hashtable<CORBA::ULong, Request*> RequestMap;
    
    friend class CollocatedDowncall;

protected:
    
    CORBA::ULong consumeReply(Request*, Buffer&, bool&, bool&);
    Request* findReply(CORBA::ULong);

    CollocatedDowncall_var downcall_;

    RequestMap* requestMap_;
    
public:

    CollocatedClient();
    virtual ~CollocatedClient();

    static inline CollocatedClient_ptr _duplicate(CollocatedClient_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline CollocatedClient_ptr _nil()
    { return 0; }

    //
    // Destroy the client
    //
    virtual void destroy();

    //
    // Get the OCI Connector
    //
    virtual OCI::Connector_ptr connector();

    //
    // Get the OCI Transport
    //
    virtual OCI::Transport_ptr transport();
    
    //
    // Calculate offset
    //
    virtual CORBA::ULong offset(StubImplBase*, const char*);

    //
    // Send request and receive response
    // With explicit timeout (throws exception on timeout)
    // With optional retry on communication failure
    //
    virtual CORBA::ULong request(StubImplBase*, const char*, Buffer&,
                                 bool&, bool&, bool&, CORBA::Long, bool);

    //
    // Send request oneway
    // With implicit zero timeout (does not throw exception on timeout)
    // With optional retry on communication failure
    //
    virtual void oneway(StubImplBase*, const char*, Buffer&, bool);

    //
    // Send request deferred
    // With implicit zero timeout (does not throw exception on timeout)
    // With optional retry on communication failure
    //
    virtual void deferred(StubImplBase*, const char*, Buffer&,
                          CORBA::ULong&, bool);

    //
    // Get repsonse
    // With explicit timeout (throws exception on timeout)
    //
    virtual CORBA::ULong response(CORBA::ULong, Buffer&, bool&, bool&, bool&,
				  CORBA::Long);
    
    //
    // Poll response
    // With implicit zero timeout (returns zero on timeout)
    //
    virtual CORBA::ULong poll(CORBA::ULong, Buffer&, bool&, bool&, bool&);
};

} // End of namespace OB

namespace CORBA
{
// XXX Are these necessary in CORBA?

inline void
release(OB::CollocatedClient_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::CollocatedClient_ptr p)
{
    return p == 0;
}

} // End of namespace CORBA

#endif
