// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/TemplateI.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/Any.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/IFR4.h>
#include <OB/OBV.h>
#include <OB/Hashtable.h>
#include <OB/HashtableI.h>
#include <OB/POAServantBase.h>
#include <OB/ValueBase.h>
#include <OB/Stream.h>
#include <OB/Custom.h>
#include <OB/Custom_impl.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Implementation of OB::DataOutputStream_impl
// ----------------------------------------------------------------------

DataOutputStream_impl::DataOutputStream_impl(OutputStream& out) :
    out_(out)
{
}

DataOutputStream_impl::~DataOutputStream_impl()
{
}

ValueBase*
DataOutputStream_impl::_copy_value()
{
    return 0;
}

void
DataOutputStream_impl::_OB_marshalValue(OutputStream&)
{
    throw MARSHAL();
}

void
DataOutputStream_impl::_OB_marshalCountValue(CountStream&)
{
    throw MARSHAL();
}

void
DataOutputStream_impl::_OB_unmarshalValue(InputStream&)
{
    throw MARSHAL();
}

void
DataOutputStream_impl::write_any(const Any& value)
{
    CountStream count;
    count.count_any(value);
    out_._OB_addCapacity(count.count(), 4);
    out_.write_any(value);
}

void
DataOutputStream_impl::write_boolean(Boolean value)
{
    out_._OB_addCapacity(1);
    out_.write_boolean(value);
}

void
DataOutputStream_impl::write_char(Char value)
{
    out_._OB_addCapacity(1);
    out_.write_boolean(value);
}

void
DataOutputStream_impl::write_wchar(WChar value)
{
    CountStream count;
    count.count_wchar(value);
    out_._OB_addCapacity(count.count());
    out_.write_wchar(value);
}

void
DataOutputStream_impl::write_octet(Octet value)
{
    out_._OB_addCapacity(1);
    out_.write_octet(value);
}

void
DataOutputStream_impl::write_short(Short value)
{
    out_._OB_addCapacity(2, 2);
    out_.write_short(value);
}

void
DataOutputStream_impl::write_ushort(UShort value)
{
    out_._OB_addCapacity(2, 2);
    out_.write_ushort(value);
}

void
DataOutputStream_impl::write_long(Long value)
{
    out_._OB_addCapacity(4, 4);
    out_.write_long(value);
}

void
DataOutputStream_impl::write_ulong(ULong value)
{
    out_._OB_addCapacity(4, 4);
    out_.write_ulong(value);
}

void
DataOutputStream_impl::write_longlong(LongLong value)
{
    out_._OB_addCapacity(8, 8);
    out_.write_longlong(value);
}

void
DataOutputStream_impl::write_ulonglong(ULongLong value)
{
    out_._OB_addCapacity(8, 8);
    out_.write_ulonglong(value);
}

void
DataOutputStream_impl::write_float(Float value)
{
    out_._OB_addCapacity(4, 4);
    out_.write_float(value);
}

void
DataOutputStream_impl::write_double(Double value)
{
    out_._OB_addCapacity(8, 8);
    out_.write_double(value);
}

void
DataOutputStream_impl::write_longdouble(LongDouble value)
{
    out_._OB_addCapacity(16, 8);
    out_.write_longdouble(value);
}

void
DataOutputStream_impl::write_string(const char* value)
{
    out_._OB_addCapacity(strlen(value) + 1, 4);
    out_.write_string(value);
}

void
DataOutputStream_impl::write_wstring(const wchar_t* value)
{
    CountStream count;
    count.count_wstring(value);
    out_._OB_addCapacity(count.count(), 4);
    out_.write_wstring(value);
}

void
DataOutputStream_impl::write_Object(Object_ptr value)
{
    CountStream count;
    count.count_Object(value);
    out_._OB_addCapacity(count.count(), 4);
    out_.write_Object(value);
}

void
DataOutputStream_impl::write_Value(ValueBase* value)
{
    CountStream count;
    count.count_Value(value);
    out_._OB_addCapacity(count.count(), 4);
    out_.write_Value(value);
}

void
DataOutputStream_impl::write_TypeCode(TypeCode_ptr value)
{
    CountStream count;
    count.count_TypeCode(value);
    out_._OB_addCapacity(count.count(), 4);
    out_.write_TypeCode(value);
}

void
DataOutputStream_impl::write_any_array(const CORBA::AnySeq& seq,
                                       ULong offset,
                                       ULong length)
{
    for(ULong i = offset ; i < offset + length ; i++)
        write_any(seq[i]);
}

void
DataOutputStream_impl::write_boolean_array(const CORBA::BooleanSeq& seq,
                                           ULong offset,
                                           ULong length)
{
    assert_nca(offset + length <= seq.length(), NCASeqRange);
    out_._OB_addCapacity(length);
    out_.write_boolean_array(seq.get_buffer() + offset, length);
}

void
DataOutputStream_impl::write_char_array(const CORBA::CharSeq& seq,
                                        ULong offset,
                                        ULong length)
{
    assert_nca(offset + length <= seq.length(), NCASeqRange);
    out_._OB_addCapacity(length);
    out_.write_char_array(seq.get_buffer() + offset, length);
}

void
DataOutputStream_impl::write_wchar_array(const CORBA::WCharSeq& seq,
                                         ULong offset,
                                         ULong length)
{
    for(ULong i = offset ; i < offset + length ; i++)
        write_wchar(seq[i]);
}

void
DataOutputStream_impl::write_octet_array(const CORBA::OctetSeq& seq,
                                         ULong offset,
                                         ULong length)
{
    assert_nca(offset + length <= seq.length(), NCASeqRange);
    out_._OB_addCapacity(length);
    out_.write_octet_array(seq.get_buffer() + offset, length);
}

void
DataOutputStream_impl::write_short_array(const CORBA::ShortSeq& seq,
                                         ULong offset,
                                         ULong length)
{
    assert_nca(offset + length <= seq.length(), NCASeqRange);
    out_._OB_addCapacity(length * 2, 2);
    out_.write_short_array(seq.get_buffer() + offset * 2, length);
}

void
DataOutputStream_impl::write_ushort_array(const CORBA::UShortSeq& seq,
                                          ULong offset,
                                          ULong length)
{
    assert_nca(offset + length <= seq.length(), NCASeqRange);
    out_._OB_addCapacity(length * 2, 2);
    out_.write_ushort_array(seq.get_buffer() + offset * 2, length);
}

void
DataOutputStream_impl::write_long_array(const CORBA::LongSeq& seq,
                                        ULong offset,
                                        ULong length)
{
    assert_nca(offset + length <= seq.length(), NCASeqRange);
    out_._OB_addCapacity(length * 4, 4);
    out_.write_long_array(seq.get_buffer() + offset * 4, length);
}

void
DataOutputStream_impl::write_ulong_array(const CORBA::ULongSeq& seq,
                                         ULong offset,
                                         ULong length)
{
    assert_nca(offset + length <= seq.length(), NCASeqRange);
    out_._OB_addCapacity(length * 4, 4);
    out_.write_ulong_array(seq.get_buffer() + offset * 4, length);
}

void
DataOutputStream_impl::write_ulonglong_array(const CORBA::ULongLongSeq& seq,
                                             ULong offset,
                                             ULong length)
{
    assert_nca(offset + length <= seq.length(), NCASeqRange);
    out_._OB_addCapacity(length * 8, 8);
    out_.write_ulonglong_array(seq.get_buffer() + offset * 8, length);
}

void
DataOutputStream_impl::write_longlong_array(const CORBA::LongLongSeq& seq,
                                            ULong offset,
                                            ULong length)
{
    assert_nca(offset + length <= seq.length(), NCASeqRange);
    out_._OB_addCapacity(length * 8, 8);
    out_.write_longlong_array(seq.get_buffer() + offset * 8, length);
}

void
DataOutputStream_impl::write_float_array(const CORBA::FloatSeq& seq,
                                         ULong offset,
                                         ULong length)
{
    assert_nca(offset + length <= seq.length(), NCASeqRange);
    out_._OB_addCapacity(length * 4, 4);
    out_.write_float_array(seq.get_buffer() + offset * 4, length);
}

void
DataOutputStream_impl::write_double_array(const CORBA::DoubleSeq& seq,
                                          ULong offset,
                                          ULong length)
{
    assert_nca(offset + length <= seq.length(), NCASeqRange);
    out_._OB_addCapacity(length * 8, 8);
    out_.write_double_array(seq.get_buffer() + offset * 8, length);
}

// ----------------------------------------------------------------------
// Implementation of CORBA::DataInputStream_impl
// ----------------------------------------------------------------------

DataInputStream_impl::DataInputStream_impl(InputStream& in) :
    in_(in)
{
}

DataInputStream_impl::~DataInputStream_impl()
{
}

ValueBase*
DataInputStream_impl::_copy_value()
{
    return 0;
}

void
DataInputStream_impl::_OB_marshalValue(OutputStream&)
{
    throw MARSHAL();
}

void
DataInputStream_impl::_OB_marshalCountValue(CountStream&)
{
    throw MARSHAL();
}

void
DataInputStream_impl::_OB_unmarshalValue(InputStream&)
{
    throw MARSHAL();
}

Any*
DataInputStream_impl::read_any()
{
    return in_.read_any();
}

Boolean
DataInputStream_impl::read_boolean()
{
    return in_.read_boolean();
}

Char
DataInputStream_impl::read_char()
{
    return in_.read_char();
}

WChar
DataInputStream_impl::read_wchar()
{
    return in_.read_wchar();
}

Octet
DataInputStream_impl::read_octet()
{
    return in_.read_octet();
}

Short
DataInputStream_impl::read_short()
{
    return in_.read_short();
}

UShort
DataInputStream_impl::read_ushort()
{
    return in_.read_ushort();
}

Long
DataInputStream_impl::read_long()
{
    return in_.read_long();
}

ULong
DataInputStream_impl::read_ulong()
{
    return in_.read_ulong();
}

LongLong
DataInputStream_impl::read_longlong()
{
    return in_.read_longlong();
}

ULongLong
DataInputStream_impl::read_ulonglong()
{
    return in_.read_ulonglong();
}

Float
DataInputStream_impl::read_float()
{
    return in_.read_float();
}

Double
DataInputStream_impl::read_double()
{
    return in_.read_double();
}

LongDouble
DataInputStream_impl::read_longdouble()
{
    return in_.read_longdouble();
}

char*
DataInputStream_impl::read_string()
{
    return in_.read_string();
}

wchar_t*
DataInputStream_impl::read_wstring()
{
    return in_.read_wstring();
}

Object_ptr
DataInputStream_impl::read_Object()
{
    return in_.read_Object();
}

ValueBase*
DataInputStream_impl::read_Value()
{
    return in_.read_Value();
}

TypeCode_ptr
DataInputStream_impl::read_TypeCode()
{
    return in_.read_TypeCode();
}

void
DataInputStream_impl::read_any_array(CORBA::AnySeq& seq,
                                     ULong offset,
                                     ULong length)
{
    for(ULong i = offset ; i < offset + length ; i++)
        in_.read_any(seq[i]);
}

void
DataInputStream_impl::read_boolean_array(CORBA::BooleanSeq& seq,
                                         ULong offset,
                                         ULong length)
{
    assert_nca(offset + length <= seq.length(), NCASeqRange);
    in_.read_boolean_array(seq.get_buffer() + offset, length);
}

void
DataInputStream_impl::read_char_array(CORBA::CharSeq& seq,
                                      ULong offset,
                                      ULong length)
{
    assert_nca(offset + length <= seq.length(), NCASeqRange);
    in_.read_char_array(seq.get_buffer() + offset, length);
}

void
DataInputStream_impl::read_wchar_array(CORBA::WCharSeq& seq,
                                       ULong offset,
                                       ULong length)
{
    for(ULong i = offset ; i < offset + length ; i++)
        seq[i] = in_.read_wchar();
}

void
DataInputStream_impl::read_octet_array(CORBA::OctetSeq& seq,
                                       ULong offset,
                                       ULong length)
{
    assert_nca(offset + length <= seq.length(), NCASeqRange);
    in_.read_octet_array(seq.get_buffer() + offset, length);
}

void
DataInputStream_impl::read_short_array(CORBA::ShortSeq& seq,
                                       ULong offset,
                                       ULong length)
{
    assert_nca(offset + length <= seq.length(), NCASeqRange);
    in_.read_short_array(seq.get_buffer() + offset * 2, length);
}

void
DataInputStream_impl::read_ushort_array(CORBA::UShortSeq& seq,
                                        ULong offset,
                                        ULong length)
{
    assert_nca(offset + length <= seq.length(), NCASeqRange);
    in_.read_ushort_array(seq.get_buffer() + offset * 2, length);
}

void
DataInputStream_impl::read_long_array(CORBA::LongSeq& seq,
                                      ULong offset,
                                      ULong length)
{
    assert_nca(offset + length <= seq.length(), NCASeqRange);
    in_.read_long_array(seq.get_buffer() + offset * 4, length);
}

void
DataInputStream_impl::read_ulong_array(CORBA::ULongSeq& seq,
                                       ULong offset,
                                       ULong length)
{
    assert_nca(offset + length <= seq.length(), NCASeqRange);
    in_.read_ulong_array(seq.get_buffer() + offset * 4, length);
}

void
DataInputStream_impl::read_ulonglong_array(CORBA::ULongLongSeq& seq,
                                           ULong offset,
                                           ULong length)
{
    assert_nca(offset + length <= seq.length(), NCASeqRange);
    in_.read_ulonglong_array(seq.get_buffer() + offset * 8, length);
}

void
DataInputStream_impl::read_longlong_array(CORBA::LongLongSeq& seq,
                                          ULong offset,
                                          ULong length)
{
    assert_nca(offset + length <= seq.length(), NCASeqRange);
    in_.read_longlong_array(seq.get_buffer() + offset * 8, length);
}

void
DataInputStream_impl::read_float_array(CORBA::FloatSeq& seq,
                                       ULong offset,
                                       ULong length)
{
    assert_nca(offset + length <= seq.length(), NCASeqRange);
    in_.read_float_array(seq.get_buffer() + offset * 4, length);
}

void
DataInputStream_impl::read_double_array(CORBA::DoubleSeq& seq,
                                        ULong offset,
                                        ULong length)
{
    assert_nca(offset + length <= seq.length(), NCASeqRange);
    in_.read_double_array(seq.get_buffer() + offset * 8, length);
}
