// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/TemplateI.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/Any.h>
#include <OB/OBV.h>
#include <OB/TCKind.h>
#include <OB/TypeCode.h>
#include <OB/Environment.h>
#include <OB/Context.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/IFR2.h>
#include <OB/Policy.h>
#include <OB/ORB.h>
#include <OB/NamedValue.h>
#include <OB/DII.h>
#include <OB/Stream.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(CORBA::ExceptionList_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(CORBA::ExceptionList_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

void
OBDuplicate(CORBA::ContextList_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(CORBA::ContextList_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

void
OBDuplicate(CORBA::Request_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(CORBA::Request_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// Template instantiations
// ----------------------------------------------------------------------

#ifndef HAVE_NO_EXPLICIT_TEMPLATES
template class OB::ObjVar< ExceptionList >;
template class OB::ObjOut< ExceptionList >;
//template class OB::ObjForSeq< ExceptionList >;
//template class OB::ObjSeq< ExceptionList >;
//template class OB::SeqVar< OB::ObjSeq< ExceptionList > >;

template class OB::ObjVar< ContextList >;
template class OB::ObjOut< ContextList >;
//template class OB::ObjForSeq< ContextList >;
//template class OB::ObjSeq< ContextList >;
//template class OB::SeqVar< OB::ObjSeq< ContextList > >;

template class OB::ObjVar< Request >;
template class OB::ObjOut< Request >;
template class OB::ObjForSeq< Request >;
template class OB::ObjSeq< Request, int >;
template class OB::SeqVar< OB::ObjSeq< Request, int > >;
//template class OB::SeqOut< OB::ObjSeq< Request, int > >;
#else
#ifdef HAVE_PRAGMA_DEFINE
#pragma define(OB::ObjVar< ExceptionList >)
#pragma define(OB::ObjOut< ExceptionList >)
//#pragma define(OB::ObjForSeq< ExceptionList >)
//#pragma define(OB::ObjSeq< ExceptionList >)
//#pragma define(OB::SeqVar< OB::ObjSeq< ExceptionList > >)

#pragma define(OB::ObjVar< ContextList >)
#pragma define(OB::ObjOut< ContextList >)
//#pragma define(OB::ObjForSeq< ContextList >)
//#pragma define(OB::ObjSeq< ContextList >)
//#pragma define(OB::SeqVar< OB::ObjSeq< ContextList > >)

#pragma define(OB::ObjVar< Request >)
#pragma define(OB::ObjOut< Request >)
#pragma define(OB::ObjForSeq< Request >)
#pragma define(OB::ObjSeq< Request, int >)
#pragma define(OB::SeqVar< OB::ObjSeq< Request, int > >)
//#pragma define(OB::SeqOut< OB::ObjSeq< Request, int > >)
#endif
#endif

// ----------------------------------------------------------------------
// ExceptionList public member implementation
// ----------------------------------------------------------------------

ULong
ExceptionList::count() const
{
    return typeCodeSeq_.length();
}

void
ExceptionList::add(TypeCode_ptr tc)
{
    typeCodeSeq_.append(TypeCode::_duplicate(tc));
}

void
ExceptionList::add_consume(TypeCode_ptr tc)
{
    typeCodeSeq_.append(tc);
}

TypeCode_ptr
ExceptionList::item(ULong n)
{
    if(n >= typeCodeSeq_.length())
        throw Bounds();
    
    return typeCodeSeq_[n];
}

void
ExceptionList::remove(ULong n)
{
    if(n >= typeCodeSeq_.length())
        throw Bounds();

    typeCodeSeq_.remove(n);
}

// ----------------------------------------------------------------------
// ContextList public member implementation
// ----------------------------------------------------------------------

ULong
ContextList::count() const
{
    return stringSeq_.length();
}

void
ContextList::add(const char* ctxt)
{
    assert_nca(ctxt, NCANullString);
    stringSeq_.append(ctxt);
}

void
ContextList::add_consume(char* ctxt)
{
    assert_nca(ctxt, NCANullString);
    stringSeq_.append(ctxt);
}

const char*
ContextList::item(ULong n)
{
    if(n >= stringSeq_.length())
        throw Bounds();

    return stringSeq_[n];
}

void
ContextList::remove(ULong n)
{
    if(n >= stringSeq_.length())
        throw Bounds();

    stringSeq_.remove(n);
}

// ----------------------------------------------------------------------
// Request private member functions
// ----------------------------------------------------------------------

Client_ptr
Request::getClient()
{
    StubImplBase_var stub = target_ -> _OB_getImplBase();
    assert(stub.in());
    return stub -> _OB_client();
}

void
Request::marshal(Buffer& buf)
{
    Client_var clt = getClient();
    if(is_nil(clt))
        throw NO_IMPLEMENT();

    StubImplBase_var stub = target_ -> _OB_getImplBase();
    ULong offset = clt -> offset(stub, operation_);
    CountStream count(offset);
    ULong i;

    for(i = 0 ; i < arguments_ -> count() ; i++)
    {
	NamedValue_ptr nv = arguments_ -> item(i);
	if(nv -> flags() != ARG_OUT)
	{
	    Any* any = nv -> value();
	    count.count_any_NoTypeCode(*any);
	}
    }

    StringSeq ctxSeq;
    if(!is_nil(ctx_))
    {
	for(i = 0 ; i < contexts_ -> count() ; i++)
	{
	    const char* item = contexts_ -> item(i);
	    ctx_ -> _OB_getValues("", 0, item, ctxSeq);
	}

        count.count_ulong(); // ctxSeq.length()
        for(i = 0 ; i < ctxSeq.length() ; i++)
            count.count_string(ctxSeq[i]);
    }

    buf.alloc(count.count());
    OutputStream out(buf, offset);

    for(i = 0 ; i < arguments_ -> count() ; i++)
    {
	NamedValue_ptr nv = arguments_ -> item(i);
	if(nv -> flags() != ARG_OUT)
	{
	    Any* any = nv -> value();
	    out.write_any_NoTypeCode(*any);
	}
    }

    if(!is_nil(ctx_))
    {
        out.write_ulong(ctxSeq.length());
        for(i = 0 ; i < ctxSeq.length() ; i++)
            out.write_string(ctxSeq[i]);
    }
}

void
Request::unmarshal(const Buffer& cbuf, ULong offset, bool swap, bool exc)
{
    InputStream in(cbuf, offset, swap);

    if(exc)
    {
	const Octet* pos = in._OB_pos();
	String_var id = in.read_string();
	in._OB_pos(pos);
	
	for(ULong i = 0 ; i < exceptions_ -> count() ; i++)
	{
	    TypeCode_ptr tc = exceptions_ -> item(i);
	    if(strcmp(tc -> id(), id) == 0)
	    {
		UnknownUserException* ex = new UnknownUserException;
		Any& any = ex -> exception();
		in.read_any_NoTypeCode(any, tc);
		environment_ -> exception(ex);
		return;
	    }
	}
	
	environment_ -> exception(new UNKNOWN);
	return;
    }

    TypeCode_var tc;
    Any* any;
    
    any = result_ -> value();
    tc = any -> type();
    in.read_any_NoTypeCode(*any, tc);
    
    for(ULong i = 0 ; i < arguments_ -> count() ; i++)
    {
	NamedValue_ptr nv = arguments_ -> item(i);
	if(nv -> flags() != ARG_IN)
	{
	    any = nv -> value();
	    tc = any -> type();
	    in.read_any_NoTypeCode(*any, tc);
	}
    }
}

// ----------------------------------------------------------------------
// Request constructor and destructor
// ----------------------------------------------------------------------

Request::Request(Object_ptr target,
		 const char* operation,
		 NVList_ptr arguments,
		 NamedValue_ptr result,
		 ExceptionList_ptr exceptions,
		 ContextList_ptr contexts)
    : target_(Object::_duplicate(target)),
      operation_(operation),
      arguments_(NVList::_duplicate(arguments)),
      result_(NamedValue::_duplicate(result)),
      environment_(new Environment),
      exceptions_(ExceptionList::_duplicate(exceptions)),
      contexts_(ContextList::_duplicate(contexts)),
      reqId_(0)
{
}

Request::Request(Object_ptr target,
		 const char* operation,
		 NVList_ptr arguments,
		 NamedValue_ptr result)
    : target_(Object::_duplicate(target)),
      operation_(operation),
      arguments_(NVList::_duplicate(arguments)),
      result_(NamedValue::_duplicate(result)),
      environment_(new Environment),
      exceptions_(new ExceptionList),
      contexts_(new ContextList),
      reqId_(0)
{
}

Request::Request(Object_ptr target, const char* operation)
    : target_(Object::_duplicate(target)),
      operation_(operation),
      arguments_(new NVList),
      result_(new NamedValue(string_dup(""), new Any, ARG_OUT)),
      environment_(new Environment),
      exceptions_(new ExceptionList),
      contexts_(new ContextList),
      reqId_(0)
{
}

// ----------------------------------------------------------------------
// Request public member implementation
// ----------------------------------------------------------------------

Object_ptr
Request::target() const
{
    return target_;
}

const char*
Request::operation() const
{
    return operation_;
}

NVList_ptr
Request::arguments()
{
    return arguments_;
}

NamedValue_ptr
Request::result()
{
    return result_;
}

Environment_ptr
Request::env()
{
    return environment_;
}

ExceptionList_ptr
Request::exceptions()
{
    return exceptions_;
}

ContextList_ptr
Request::contexts()
{
    return contexts_;
}

void
Request::ctx(Context_ptr ctx)
{
    ctx_ = Context::_duplicate(ctx);
}

Context_ptr
Request::ctx() const
{
    return ctx_;
}

Any&
Request::add_in_arg()
{
    NamedValue_ptr p = arguments_ -> add(ARG_IN);
    return *(p -> value());
}

Any&
Request::add_in_arg(const char* name)
{
    assert_nca(name, NCANullString);
    NamedValue_ptr p = arguments_ -> add_item(name, ARG_IN);
    return *(p -> value());
}

Any&
Request::add_inout_arg()
{
    NamedValue_ptr p = arguments_ -> add(ARG_INOUT);
    return *(p -> value());
}

Any&
Request::add_inout_arg(const char* name)
{
    assert_nca(name, NCANullString);
    NamedValue_ptr p = arguments_ -> add_item(name, ARG_INOUT);
    return *(p -> value());
}

Any&
Request::add_out_arg()
{
    NamedValue_ptr p = arguments_ -> add(ARG_OUT);
    return *(p -> value());
}

Any&
Request::add_out_arg(const char* name)
{
    assert_nca(name, NCANullString);
    NamedValue_ptr p = arguments_ -> add_item(name, ARG_OUT);
    return *(p -> value());
}

void
Request::set_return_type(TypeCode_ptr tc)
{
    result_ -> value() -> replace(tc, 0);
}

Any&
Request::return_value()
{
    return *(result_ -> value());
}

void
Request::invoke()
{
    try
    {
	Buffer buf;
	marshal(buf);
	
	Client_var clt = getClient();
	if(is_nil(clt))
	    throw NO_IMPLEMENT();

	bool swap, exc, fwd;
        StubImplBase_var base = target_ -> _OB_getImplBase();
	ULong offset = clt -> request(base, operation_, buf, swap, exc, fwd,
				      base -> _OB_timeout(),
				      base -> _OB_retry());
	
	assert(offset > 0);

	if(fwd)
	{
            InputStream in(buf, offset, swap);
            IOP::IOR ior;
            IOP::IOR::_OB_unmarshal(ior, in);
            target_ -> _OB_forward(ior);
	    invoke();
	    return;
	}

	unmarshal(buf, offset, swap, exc);
    }
    catch(const SystemException& ex)
    {
	environment_ -> exception(ex._OB_clone());
    }
}

void
Request::send_oneway()
{
    try
    {
	Buffer buf;
	marshal(buf);

	Client_var clt = getClient();
	if(is_nil(clt))
	    throw NO_IMPLEMENT();

        StubImplBase_var base = target_ -> _OB_getImplBase();
	clt -> oneway(base, operation_, buf, base -> _OB_retry());
    }
    catch(const SystemException& ex)
    {
	environment_ -> exception(ex._OB_clone());
    }
}

void
Request::send_deferred()
{
    try
    {
	Buffer buf;
	marshal(buf);
	
	Client_var clt = getClient();
	if(is_nil(clt))
	    throw NO_IMPLEMENT();

        StubImplBase_var base = target_ -> _OB_getImplBase();
	clt -> deferred(base, operation_, buf, reqId_, base -> _OB_retry());
    }
    catch(const SystemException& ex)
    {
	environment_ -> exception(ex._OB_clone());
    }
}

void
Request::get_response()
{
    try
    {
	Buffer buf;

	Client_var clt = getClient();
	if(is_nil(clt))
	    throw NO_IMPLEMENT();

	bool swap, exc, fwd;
        StubImplBase_var base = target_ -> _OB_getImplBase();
	ULong offset = clt -> response(reqId_, buf, swap, exc, fwd,
				       base -> _OB_timeout());

	assert(offset > 0);

	if(fwd)
	{
            InputStream in(buf, offset, swap);
            IOP::IOR ior;
            IOP::IOR::_OB_unmarshal(ior, in);
            target_ -> _OB_forward(ior);
	    send_deferred();
	    get_response();
	    return;
	}

	unmarshal(buf, offset, swap, exc);
    }
    catch(const SystemException& ex)
    {
	environment_ -> exception(ex._OB_clone());
    }
}

Boolean
Request::poll_response()
{
    try
    {
	Buffer buf;
	
	Client_var clt = getClient();
	if(is_nil(clt))
	    throw NO_IMPLEMENT();

	bool swap, exc, fwd;
	ULong offset = clt -> poll(reqId_, buf, swap, exc, fwd);
	
	if(offset == 0)
	    return false;

	if(fwd)
	{
            InputStream in(buf, offset, swap);
            IOP::IOR ior;
            IOP::IOR::_OB_unmarshal(ior, in);
            target_ -> _OB_forward(ior);
	    send_deferred();
	    return false;
	}

	unmarshal(buf, offset, swap, exc);
	return true;
    }
    catch(const SystemException& ex)
    {
	environment_ -> exception(ex._OB_clone());
	return true;
    }
}

// ----------------------------------------------------------------------
// Object public member implementations that are related to the DII
// ----------------------------------------------------------------------

void
Object::_create_request(Context_ptr ctx,
			const char* operation,
			NVList_ptr arg_list,
			NamedValue_ptr result,
			Request_out request,
			Flags /*req_flags*/)
{
    assert_nca(operation, NCANullString);
    request = new Request(this, operation, arg_list, result);
    request -> ctx(ctx);
}

void
Object::_create_request(Context_ptr ctx,
			const char* operation,
			NVList_ptr arg_list,
			NamedValue_ptr result,
			ExceptionList_ptr exceptions,
			ContextList_ptr contexts,
			Request_out request,
			Flags /*req_flags*/)
{
    assert_nca(operation, NCANullString);
    request = new Request(this, operation, arg_list, result,
				exceptions, contexts);
    request -> ctx(ctx);
}

Request_ptr
Object::_request(const char* operation)
{
    assert_nca(operation, NCANullString);
    return new Request(this, operation);
}

// ----------------------------------------------------------------------
// ORB public member implementations that are related to the DII
// ----------------------------------------------------------------------

void
ORB::create_list(Long n, NVList_out list)
{
    if(n < 0)
	n = 0;

    list = new NVList((ULong)n);
}

void
ORB::create_operation_list(OperationDef_ptr opDef, NVList_out list)
{
    //
    // Get operation description
    //
    Contained::Description_var d = opDef -> describe();
    OperationDescription* desc;
    // XXX EGCS 1.1.1 complains about this - Why?
    //Boolean b = d -> value >>= desc;
    Boolean b = ::operator>>=(d -> value, desc);
    assert(b);

    //
    // Create list
    //
    list = new NVList;
    for(ULong i = 0 ; i < desc -> parameters.length() ; i++)
    {
	ParameterDescription& par = desc -> parameters[i];

	char* name = string_dup(par.name);

	Any* any = new Any;
	any -> replace(par.type, 0);

	Flags flags;
	switch(par.mode)
	{
	case PARAM_IN:
	    flags = ARG_IN;
	    break;
	    
	case PARAM_OUT:
	    flags = ARG_OUT;
	    break;
	    
	case PARAM_INOUT:
	    flags = ARG_INOUT;
	    break;
	}

	list -> add_value_consume(name, any, flags);
    }
}

void
ORB::create_named_value(NamedValue_out nv)
{
    nv = new NamedValue;
}

void
ORB::create_exception_list(ExceptionList_out list)
{
    list = new ExceptionList;
}

void
ORB::create_context_list(ContextList_out list)
{
    list = new ContextList;
}

void
ORB::send_multiple_requests_oneway(const RequestSeq& requests)
{
    for(ULong i = 0 ; i < requests.length() ; i++)
	requests[i] -> send_oneway();
}

void
ORB::send_multiple_requests_deferred(const RequestSeq& requests)
{
    deferredRequests_ = requests;
    deferredRequestCompletions_.length(deferredRequests_.length());

    for(ULong i = 0 ; i < deferredRequests_.length() ; i++)
    {
	deferredRequests_[i] -> send_deferred();
	deferredRequestCompletions_[i] = FALSE;
    }
}

Boolean
ORB::poll_next_response()
{
    ULong i;

    for(i = 0 ; i < deferredRequests_.length() ; i++)
	if(!deferredRequestCompletions_[i])
	    if(deferredRequests_[i] -> poll_response())
		deferredRequestCompletions_[i] = TRUE;

    for(i = 0 ; i < deferredRequests_.length() ; i++)
	if(deferredRequestCompletions_[i] &&
	   !is_nil(deferredRequests_[i]))
	    return TRUE;

    return FALSE;
}

void
ORB::get_next_response(Request_out request)
{
    for(ULong i = 0 ; i < deferredRequests_.length() ; i++)
	if(deferredRequestCompletions_[i] &&
	   !is_nil(deferredRequests_[i]))
	{
	    request = Request::_duplicate(deferredRequests_[i]);
	    deferredRequests_[i] = Request::_nil();
	    return;
	}

    request = Request::_nil();
}

// ----------------------------------------------------------------------
// UnkownUserException public member implementation
// ----------------------------------------------------------------------

/* UnkownUserException defined here, it belongs logically to the DII */

UnknownUserException::
UnknownUserException(const UnknownUserException& val)
    : UserException(val),
      any_(val.any_)
{
}

UnknownUserException::UnknownUserException(const Any& any)
    : any_(any)
{
}

UnknownUserException&
UnknownUserException::operator=(const UnknownUserException& val)
{
    if(this != &val)
        any_ = val.any_;

    return *this;
}
