// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/TemplateI.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/Any.h>
#include <OB/OBV.h>
#include <OB/TCKind.h>
#include <OB/TypeCode.h>
#include <OB/Context.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/IFR2.h>
#include <OB/Policy.h>
#include <OB/ORB.h>
#include <OB/Server.h>
#include <OB/NamedValue.h>
#include <OB/POAServantBase.h>
#include <OB/Stream.h>
#include <OB/DSI.h>
#include <OB/NativeTypes.h>
#include <OB/POAServantLocator.h> // For PortableServer_ServantLocator::Cookie
#include <OB/POACurrent.h>
#include <OB/POAInterface.h>

#include <Upcall.h>
#include <POA_impl.h>
#include <POACurrent_impl.h>

using namespace CORBA;
using namespace PortableServer;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(CORBA::ServerRequest_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(CORBA::ServerRequest_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// Template instantiations
// ----------------------------------------------------------------------

#ifndef HAVE_NO_EXPLICIT_TEMPLATES
template class OB::ObjVar< ServerRequest >;
template class OB::ObjForSeq< ServerRequest >;
//template class OB::ObjSeq< ServerRequest >;
//template class OB::SeqVar< OB::ObjSeq< ServerRequest > >;
#else
#ifdef HAVE_PRAGMA_DEFINE
#pragma define(OB::ObjVar< ServerRequest >)
#pragma define(OB::ObjForSeq< ServerRequest >)
//#pragma define(OB::ObjSeq< ServerRequest >)
//#pragma define(OB::SeqVar< OB::ObjSeq< ServerRequest > >)
#endif
#endif

// ----------------------------------------------------------------------
// ServerRequest private member functions
// ----------------------------------------------------------------------

void
ServerRequest::unmarshal()
{
    if(!is_nil(params_))
    {
	TypeCode_var tc;
	Any* any;

	for(ULong i = 0 ; i < params_ -> count() ; i++)
	{
	    NamedValue_ptr nv = params_ -> item(i);
	    if(nv -> flags() != ARG_OUT)
	    {
		any = nv -> value();
		tc = any -> type();
		in_.read_any_NoTypeCode(*any, tc);
	    }
	}
    }
}

void
ServerRequest::unmarshalCtx()
{
    StringSeq* ctxSeq = new StringSeq;
    ULong len = in_.read_ulong();
    ctxSeq -> length(len);
    for(ULong i = 0 ; i < len ; i++)
        (*ctxSeq)[i] = in_.read_string();
    ctx_ = Context::_OB_create(ctxSeq);
}

void
ServerRequest::marshal()
{
    CountStream count(offOut_);
    ULong i;
    Any* any;
    
    any = (Any*)exception_;
    if(any)
    {
	count.count_any_NoTypeCode(*any);
    }
    else
    {
	any = (Any*)result_;
	if(any)
	{
	    count.count_any_NoTypeCode(*any);
	}
	
	if(!is_nil(params_))
	{
	    for(i = 0 ; i < params_ -> count() ; i++)
	    {
		NamedValue_ptr nv = params_ -> item(i);
		if(nv -> flags() != ARG_IN)
		{
		    any = nv -> value();
		    count.count_any_NoTypeCode(*any);
		}
	    }
	}
    }

    buf_.alloc(count.count());
    OutputStream out(buf_, offOut_);

    any = (Any*)exception_;
    if(any)
    {
	out.write_any_NoTypeCode(*any);
    }
    else
    {
	any = (Any*)result_;
	if(any)
	{
	    out.write_any_NoTypeCode(*any);
	}

	if(!is_nil(params_))
	{
	    for(i = 0 ; i < params_ -> count() ; i++)
	    {
		NamedValue_ptr nv = params_ -> item(i);
		if(nv -> flags() != ARG_IN)
		{
		    any = nv -> value();
		    out.write_any_NoTypeCode(*any);
		}
	    }
	}
    }
}

// ----------------------------------------------------------------------
// ServerRequest constructor and destructor
// ----------------------------------------------------------------------

ServerRequest::ServerRequest(Object_ptr object,
			     const char* name,
			     Buffer& buf,
			     bool swap,
			     ULong offIn,
			     ULong offOut)
    : object_(Object::_duplicate(object)),
      name_(name), buf_(buf), offOut_(offOut), in_(buf, offIn, swap)
{
}

// ----------------------------------------------------------------------
// ServerRequest public member implementation
// ----------------------------------------------------------------------

OperationDef_ptr
ServerRequest::op_def()
{
    InterfaceDef_var interf = object_ -> _get_interface();
    if(is_nil(interf))
        return OperationDef::_nil();

    Contained_var contained = interf -> lookup(name_);
    if(is_nil(contained))
        return OperationDef::_nil();

    return OperationDef::_narrow(contained);
}

const char*
ServerRequest::op_name()
{
    return name_;
}

Context_ptr
ServerRequest::ctx()
{
    unmarshalCtx();
    return ctx_;
}

void
ServerRequest::params(NVList_ptr parameters)
{
    params_ = parameters;
    unmarshal();
}

void
ServerRequest::result(Any* value)
{
    result_ = value;
}

void
ServerRequest::exception(Any* value)
{
    exception_ = value;
}

// ----------------------------------------------------------------------
// The PortableServer::DynamicImplementation public member implementation
// ----------------------------------------------------------------------

Object_ptr
DynamicImplementation::_this()
{
    //
    // If invoked outside of a request we raise a WrongPolicy
    // exception
    //
    POA_var thePOA = _default_POA();
    POA_impl* poaImpl = dynamic_cast<POA_impl*>(thePOA.in());
    Current_impl* current = poaImpl -> _OB_POACurrent();

    if(!current -> _OB_inUpcall() || current -> _OB_getServant() != this)
    {
        throw POA::WrongPolicy();
    }

    ObjectId_var oid = current -> get_object_id();
    String_var intf = _primary_interface(oid, thePOA);

    return poaImpl -> _OB_createReference(current -> _OB_getKey(), intf);
}

DispatchStatus
DynamicImplementation::_OB_dispatch(const char* _ob_op,
				    Buffer& _ob_buf,
				    bool _ob_sw,
				    ULong _ob_offIn,
				    ULong _ob_offOut)
{
    DispatchStatus status =
	ServantBase::_OB_dispatch(_ob_op, _ob_buf, _ob_sw,
                                  _ob_offIn, _ob_offOut);

    if(status == DispatchStatusNoOperation)
    {
        Object_var self = _this();
	ServerRequest_var serverRequest =
	    new ServerRequest(self, _ob_op, _ob_buf, _ob_sw,
			      _ob_offIn, _ob_offOut);

	invoke(serverRequest);
	serverRequest -> marshal();
	
	Any* any;
	any = (Any*)(serverRequest -> exception_);

	if(any)
	{
	    TypeCode_var tc = any -> type();
	    if(tc -> _OB_isSystemException())
	    {
		InputStream in(_ob_buf, _ob_offOut, false); // No swap here!
		SystemException::_OB_unmarshalThrow(in);
	    }

	    status = DispatchStatusExcept;
	}
	else
	    status = DispatchStatusOK;
    }

    return status;
}

bool
DynamicImplementation::_OB_isDynamic()
{
    return true;
}

StubImplBase*
DynamicImplementation::_OB_createStubImpl(POA_ptr,
                                          const ObjectId&,
                                          const OctetSeq&)
{
    //XXX - is this correct error? - This should never be called!
    throw INTERNAL();
    return 0;
}
