// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_DOWNCALL_H
#define OB_DOWNCALL_H

namespace OB
{

class Downcall;
typedef Downcall* Downcall_ptr;

} // End of namespace OB

void OBDuplicate(OB::Downcall_ptr);
void OBRelease(OB::Downcall_ptr);

namespace OB
{

//
// Forward declarations
//
typedef ObjVar< Downcall > Downcall_var;

//
// The Downcall class.  This abstract class represents an interface
// that is used to down-call. That is from server to client.
//
class Downcall : public RefCount
{
public:
    
    virtual ~Downcall() { }

    static inline Downcall_ptr _duplicate(Downcall_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline Downcall_ptr _nil()
    { return 0; }
    
    //
    // Regular response
    //
    virtual void response(CORBA::ULong, Buffer&) = 0;

    //
    // Response is an exception
    //
    virtual void systemException(CORBA::ULong, Buffer&,
                                 const CORBA::SystemException&) = 0;

    //
    // Response is a userException
    //
    virtual void userException(CORBA::ULong, Buffer&) = 0;

    //
    // Response is a LOCATION_FORWARD
    //
    virtual void forward(CORBA::ULong, Buffer&, CORBA::Object_ptr) = 0;
};

} // End of namespace OB

namespace CORBA
{
// XXX Are these necessary in CORBA?

inline void
release(OB::Downcall_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::Downcall_ptr p)
{
    return p == 0;
}

} // End of namespace CORBA

#endif
