// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

//
// Author: Philippe Merle
//
// From the Laboratoire d'Informatique Fondamentale de Lille in France
// (it is a computer science laboratory of the Lille french university)
//

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/Any.h>
#include <OB/OBV.h>
#include <OB/TCKind.h>
#include <OB/TypeCode.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/IFR4.h>
#include <OB/Policy.h>
#include <OB/ORB.h>
#include <OB/DynAny.h>
#include <OB/POAServantBase.h>
#include <OB/ValueBase.h>
#include <OB/Stream.h>

#include <DynAny_impl.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Macro to read an Any containing an enum value
// ----------------------------------------------------------------------

#define READ_ENUM_VALUE(ANY) (* ( (ULong*) (ANY).value () ) )

// ----------------------------------------------------------------------
// Narrow DynAny to implementation classes
// ----------------------------------------------------------------------

// ----------------------------------------------------------------------
// ----------------------------------------------------------------------
//
// DynAny_impl class
//
// ----------------------------------------------------------------------
// ----------------------------------------------------------------------

// ----------------------------------------------------------------------
// DynAny_impl constructor and destructor
// ----------------------------------------------------------------------

DynAny_impl::DynAny_impl (TypeCode_ptr type)
    : type_(TypeCode::_duplicate(type))
{
}

DynAny_impl::~DynAny_impl()
{
}

// ----------------------------------------------------------------------
// operation to create a DynAny_impl from a TypeCode
// ----------------------------------------------------------------------

DynAny_ptr
DynAny_impl::create (TypeCode_ptr typecode)
{
    TypeCode_var tc = GetOrigType(typecode);

    switch(tc -> kind())
    {
    case tk_null:
    case tk_void:
    case tk_short:
    case tk_long:
    case tk_longlong:
    case tk_ushort:
    case tk_ulong:
    case tk_ulonglong:
    case tk_float:
    case tk_double:
    case tk_longdouble:
    case tk_boolean:
    case tk_char:
    case tk_wchar:
    case tk_octet:
    case tk_any:
    case tk_TypeCode:
    case tk_Principal:
    case tk_objref:
    case tk_string:
    case tk_wstring:
	return new DynBasic_impl(typecode);

    case tk_struct:
	return new DynStruct_impl(typecode);

    case tk_except:
	return new DynException_impl(typecode);

    case tk_union:
	return new DynUnion_impl(typecode);

    case tk_fixed:
	return new DynFixed_impl(typecode);

    case tk_enum:
	return new DynEnum_impl(typecode);

    case tk_sequence:
	return new DynSequence_impl(typecode);

    case tk_array:
	return new DynArray_impl(typecode);

/*
    case tk_value:
    case tk_value_box:
	return new DynValue_impl(typecode);
*/

    case tk_alias:
	assert(false);
	break;
    }

    assert(false);
    return DynAny::_nil();
}

// ----------------------------------------------------------------------
// DynAny_impl public IDL member implementation
// from DynAny interface
// ----------------------------------------------------------------------

TypeCode_ptr
DynAny_impl::type()
    throw(SystemException)
{
    return TypeCode::_duplicate(type_);
}

// ----------------------------------------------------------------------

#define CONTROL_INSERT(ANY,TC)                        \
    if ( ANY == 0 )                                   \
        throw DynAny::InvalidValue ();          \
    TypeCode_var type = ANY->type();            \
    if ( ! type->equivalent(TC) )                     \
        throw DynAny::InvalidValue ();

void
DynAny_impl::insert_boolean(Boolean value)
    throw(DynAny::InvalidValue,
          SystemException)
{
    Any* any = current_any ();

    CONTROL_INSERT (any,_tc_boolean);

    // this allows to reuse the current TypeCode of the any.
    any->replace (type, new Boolean(value), TRUE);

    next ();
}

void
DynAny_impl::insert_octet(Octet value)
    throw(DynAny::InvalidValue,
          SystemException)
{
    Any* any = current_any ();

    CONTROL_INSERT (any,_tc_octet);

    // this allows to reuse the current TypeCode of the any.
    any->replace (type, new Octet(value), TRUE);

    next ();
}

void
DynAny_impl::insert_char(Char value)
    throw(DynAny::InvalidValue,
          SystemException)
{
    Any* any = current_any ();

    CONTROL_INSERT (any,_tc_char);

    // this allows to reuse the current TypeCode of the any.
    any->replace (type, new Char(value), TRUE);

    next ();
}

void
DynAny_impl::insert_wchar(WChar value)
    throw(DynAny::InvalidValue,
          SystemException)
{
    Any* any = current_any ();

    CONTROL_INSERT (any,_tc_wchar);

    // this allows to reuse the current TypeCode of the any.
    any->replace (type, new WChar(value), TRUE);

    next ();
}

void
DynAny_impl::insert_short(Short value)
    throw(DynAny::InvalidValue,
          SystemException)
{
    Any* any = current_any ();

    CONTROL_INSERT (any,_tc_short);

    // this allows to reuse the current TypeCode of the any.
    any->replace (type, new Short(value), TRUE);

    next ();
}

void
DynAny_impl::insert_ushort(UShort value)
    throw(DynAny::InvalidValue,
              SystemException)
{
    Any* any = current_any ();

    CONTROL_INSERT (any,_tc_ushort);

    // this allows to reuse the current TypeCode of the any.
    any->replace (type, new UShort(value), TRUE);

    next ();
}

void
DynAny_impl::insert_long(Long value)
    throw(DynAny::InvalidValue,
          SystemException)
{
    Any* any = current_any ();

    CONTROL_INSERT (any,_tc_long);

    // this allows to reuse the current TypeCode of the any.
    any->replace (type, new Long(value), TRUE);

    next ();
}

void
DynAny_impl::insert_ulong(ULong value)
    throw(DynAny::InvalidValue,
          SystemException)
{
    Any* any = current_any ();

    CONTROL_INSERT (any,_tc_ulong);

    // this allows to reuse the current TypeCode of the any.
    any->replace (type, new ULong(value), TRUE);

    next ();
}

void
DynAny_impl::insert_longlong(LongLong value)
    throw(DynAny::InvalidValue,
          SystemException)
{
    Any* any = current_any ();

    CONTROL_INSERT (any,_tc_longlong);

    // this allows to reuse the current TypeCode of the any.
    any->replace (type, new LongLong(value), TRUE);

    next ();
}

void
DynAny_impl::insert_ulonglong(ULongLong value)
    throw(DynAny::InvalidValue,
          SystemException)
{
    Any* any = current_any ();

    CONTROL_INSERT (any,_tc_ulonglong);

    // this allows to reuse the current TypeCode of the any.
    any->replace (type, new ULongLong(value), TRUE);

    next ();
}

void
DynAny_impl::insert_float(Float value)
    throw(DynAny::InvalidValue,
          SystemException)
{
    Any* any = current_any ();

    CONTROL_INSERT (any,_tc_float);

    // this allows to reuse the current TypeCode of the any.
    any->replace (type, new Float(value), TRUE);

    next ();
}

void
DynAny_impl::insert_double(Double value)
    throw(DynAny::InvalidValue,
          SystemException)
{
    Any* any = current_any ();

    CONTROL_INSERT (any,_tc_double);

    // this allows to reuse the current TypeCode of the any.
    any->replace (type, new Double(value), TRUE);

    next ();
}

void
DynAny_impl::insert_longdouble(LongDouble value)
    throw(DynAny::InvalidValue,
          SystemException)
{
#if SIZEOF_DOUBLE != 0
    Any* any = current_any ();

    CONTROL_INSERT (any,_tc_longdouble);

    // this allows to reuse the current TypeCode of the any.
    any->replace (type, new LongDouble(value), TRUE);

    next ();
#else
    throw InvalidValidValue();
#endif
}

void
DynAny_impl::insert_string(const char* value)
    throw(DynAny::InvalidValue,
          SystemException)
{
    Any* any = current_any ();

    // Can't use CONTROL_INSERT, as it then wouldn't be possible to
    // insert bounded strings.
    //CONTROL_INSERT (any,_tc_string);
    if(any == 0)
        throw InvalidValue ();
    TypeCode_var type = any -> type();
    TypeCode_var origType = OB::GetOrigType(type);
    if(origType -> kind() != tk_string)
        throw InvalidValue ();

    // this allows to reuse the current TypeCode of the any.
    any->replace (type, string_dup(value), TRUE);

    next ();
}

void
DynAny_impl::insert_wstring(const wchar_t* value)
    throw(DynAny::InvalidValue,
          SystemException)
{

    Any* any = current_any ();

    // Can't use CONTROL_INSERT, as it then wouldn't be possible to
    // insert bounded wstrings.
    //CONTROL_INSERT (any,_tc_wstring);
    if(any == 0)
        throw InvalidValue ();
    TypeCode_var type = any -> type();
    TypeCode_var origType = OB::GetOrigType(type);
    if(origType -> kind() != tk_wstring)
        throw InvalidValue ();

    // this allows to reuse the current TypeCode of the any.
    any->replace (type, wstring_dup(value), TRUE);

    next ();
}

void
DynAny_impl::insert_reference(Object_ptr value)
    throw(DynAny::InvalidValue,
          SystemException)
{
    Any* any = current_any ();

    if ( any == 0 )
        throw DynAny::InvalidValue ();

    TypeCode_var type = any->type();

    TypeCode_var tc = GetOrigType(type);

    if ( tc->kind() != tk_objref )
        throw DynAny::InvalidValue ();

    //
    // Note of Philippe Merle
    //
    // I don't want to this test because it could send an IIOP request.
    // I think the DynAny API must only execute local tasks.
    // This test must be done in applications which use the DynAny API.
    // For instance, in CorbaScript, this test is done by the interpreter.
    //
    // Moreover, OMG don't currently define the semantic of the operation.
    // Then, I think we don't add any network overheads.
    //
    //  if ( !value -> _is_a(tc->id()) )
    //        throw InvalidValue ();
    //

    any->replace (type, Object::_duplicate(value), TRUE);

    next ();
}

void
DynAny_impl::insert_typecode(TypeCode_ptr value)
    throw(DynAny::InvalidValue,
          SystemException)
{
    Any* any = current_any ();

    CONTROL_INSERT (any,_tc_TypeCode);

    // this allows to reuse the current TypeCode of the any.
    any->replace (type, TypeCode::_duplicate(value), TRUE);

    next ();
}

void
DynAny_impl::insert_any(const Any& value)
    throw(DynAny::InvalidValue,
          SystemException)
{
    Any* any = current_any ();

    CONTROL_INSERT (any,_tc_any);

    // this allows to reuse the current TypeCode of the any.
    any->replace (type, new Any(value), TRUE);

    next ();
}

void
DynAny_impl::insert_valuetype(ValueBase* value)
    throw(DynAny::InvalidValue,
          SystemException)
{
    Any* any = current_any ();

    if ( any == 0 )
        throw DynAny::InvalidValue ();

    TypeCode_var type = any->type();

    TypeCode_var tc = GetOrigType(type);

    if ( tc->kind() != tk_value && tc->kind() != tk_value_box )
        throw DynAny::InvalidValue ();

    add_ref(value);
    any->replace (type, value, TRUE);

    next ();
}

// ----------------------------------------------------------------------

#define GET_CHECK_ANY(ANY) ( ( ANY != 0 ) && ( ANY->value() != 0 ) )

Boolean
DynAny_impl::get_boolean()
    throw(DynAny::TypeMismatch,
          SystemException)
{
    Any* any = current_any ();

    Boolean result;
    Any::to_boolean val (result);

    if ( ! GET_CHECK_ANY(any) || ! ( *any >>= val ) )
        throw DynAny::TypeMismatch ();
 
    return result;
}

Octet
DynAny_impl::get_octet()
    throw(DynAny::TypeMismatch,
          SystemException)
{
    Any* any = current_any ();

    Octet result;
    Any::to_octet val (result);

    if ( ! GET_CHECK_ANY(any) || ! ( *any >>= val ) )
        throw DynAny::TypeMismatch ();
 
    return result;
}

Char
DynAny_impl::get_char()
    throw(DynAny::TypeMismatch,
          SystemException)
{
    Any* any = current_any ();

    Char result;
    Any::to_char val (result);

    if ( ! GET_CHECK_ANY (any) || ! ( *any >>= val ) )
        throw DynAny::TypeMismatch ();
 
    return result;
}

WChar
DynAny_impl::get_wchar()
    throw(DynAny::TypeMismatch,
          SystemException)
{
    Any* any = current_any ();

    WChar result;
    Any::to_wchar val (result);

    if ( ! GET_CHECK_ANY (any) || ! ( *any >>= val ) )
        throw DynAny::TypeMismatch ();
 
    return result;
}

Short
DynAny_impl::get_short()
    throw(DynAny::TypeMismatch,
          SystemException)
{
    Any* any = current_any ();

    Short result;

    if ( ! GET_CHECK_ANY(any) || ! ( *any >>= result ) )
        throw DynAny::TypeMismatch ();
 
    return result;
}

UShort
DynAny_impl::get_ushort()
    throw(DynAny::TypeMismatch,
          SystemException)
{
    Any* any = current_any ();

    UShort result;

    if ( ! GET_CHECK_ANY(any) || ! ( *any >>= result ) )
        throw DynAny::TypeMismatch ();
 
    return result;
}

Long
DynAny_impl::get_long()
    throw(DynAny::TypeMismatch,
          SystemException)
{
    Any* any = current_any ();

    Long result;

    if ( ! GET_CHECK_ANY (any) || ! ( *any >>= result ) )
        throw DynAny::TypeMismatch ();
 
    return result;
}

ULong
DynAny_impl::get_ulong()
    throw(DynAny::TypeMismatch,
          SystemException)
{
    Any* any = current_any ();

    ULong result;

    if ( ! GET_CHECK_ANY(any) || ! ( *any >>= result ) )
        throw DynAny::TypeMismatch ();
 
    return result;
}

LongLong
DynAny_impl::get_longlong()
    throw(DynAny::TypeMismatch,
          SystemException)
{
    Any* any = current_any ();

    LongLong result;

    if ( ! GET_CHECK_ANY (any) || ! ( *any >>= result ) )
        throw DynAny::TypeMismatch ();
 
    return result;
}

ULongLong
DynAny_impl::get_ulonglong()
    throw(DynAny::TypeMismatch,
          SystemException)
{
    Any* any = current_any ();

    ULongLong result;

    if ( ! GET_CHECK_ANY(any) || ! ( *any >>= result ) )
        throw DynAny::TypeMismatch ();
 
    return result;
}

Float
DynAny_impl::get_float()
    throw(DynAny::TypeMismatch,
          SystemException)
{
    Any* any = current_any ();

    Float result;

    if ( ! GET_CHECK_ANY(any) || ! ( *any >>= result ) )
        throw DynAny::TypeMismatch ();
 
    return result;
}

Double
DynAny_impl::get_double()
    throw(DynAny::TypeMismatch,
          SystemException)
{
    Any* any = current_any ();

    Double result;

    if ( ! GET_CHECK_ANY(any) || ! ( *any >>= result ) )
        throw DynAny::TypeMismatch ();
 
    return result;
}

LongDouble
DynAny_impl::get_longdouble()
    throw(DynAny::TypeMismatch,
          SystemException)
{
    Any* any = current_any ();

    LongDouble result;

    if ( ! GET_CHECK_ANY(any) || ! ( *any >>= result ) )
        throw DynAny::TypeMismatch ();
 
    return result;
}

char*
DynAny_impl::get_string()
    throw(DynAny::TypeMismatch,
          SystemException)
{
    Any* any = current_any ();

    if ( ! GET_CHECK_ANY(any) )
        throw DynAny::TypeMismatch ();

    TypeCode_var type = any->type();

    TypeCode_var tc = GetOrigType (type);

    if(tc -> kind() != tk_string )
        throw DynAny::TypeMismatch ();

    char* result = (char*)any->value();

    return string_dup(result);
}

wchar_t*
DynAny_impl::get_wstring()
    throw(DynAny::TypeMismatch,
          SystemException)
{
    Any* any = current_any ();

    if ( ! GET_CHECK_ANY(any) )
        throw DynAny::TypeMismatch ();

    TypeCode_var type = any->type();

    TypeCode_var tc = GetOrigType (type);

    if(tc -> kind() != tk_wstring )
        throw DynAny::TypeMismatch ();

    wchar_t* result = (wchar_t*)any->value();

    return wstring_dup(result);
}

Object_ptr
DynAny_impl::get_reference()
    throw(DynAny::TypeMismatch,
          SystemException)
{
    Any* any = current_any ();

    Object_ptr result;
    Any::to_object v(result);

    if ( ! GET_CHECK_ANY(any) || ! ( *any >>= v ) )
        throw DynAny::TypeMismatch ();

    return Object::_duplicate(result);
}

TypeCode_ptr
DynAny_impl::get_typecode()
    throw(DynAny::TypeMismatch,
          SystemException)
{
    Any* any = current_any ();

    TypeCode_ptr result;

    if ( ! GET_CHECK_ANY(any) || ! ( *any >>= result ) )
        throw DynAny::TypeMismatch ();
 
    return TypeCode::_duplicate(result);
}

Any*
DynAny_impl::get_any()
    throw(DynAny::TypeMismatch,
          SystemException)
{
    Any* any = current_any ();

    Any* result;

    if ( ! GET_CHECK_ANY(any) || ! ( *any >>= result ) )
        throw DynAny::TypeMismatch ();
 
    return new Any(*result);
}

ValueBase*
DynAny_impl::get_valuetype()
    throw(DynAny::TypeMismatch,
          SystemException)
{
    Any* any = current_any ();

    if ( ! GET_CHECK_ANY(any) )
        throw DynAny::TypeMismatch ();

    TypeCode_var type = any->type();
    TypeCode_var tc = GetOrigType(type);

    if ( tc->kind() != tk_value && tc->kind() != tk_value_box )
        throw DynAny::TypeMismatch ();

    ValueBase* result = (ValueBase*)any->value();
    add_ref(result);
    return result;
}

void
DynAny_impl::destroy()
    throw(SystemException)
{
    // nothing to do
}

// ----------------------------------------------------------------------
// ----------------------------------------------------------------------
//
// DynBasic_impl implementation
//
// ----------------------------------------------------------------------
// ----------------------------------------------------------------------

// ----------------------------------------------------------------------
// DynBasic_impl constructor and destructor
// ----------------------------------------------------------------------

DynBasic_impl::DynBasic_impl(TypeCode_ptr type)
    : DynAny_impl(type)
    , any_ (new Any (type, 0, FALSE))
{
}

DynBasic_impl::~DynBasic_impl()
{
}

// ----------------------------------------------------------------------
// DynBasic_impl operations to access components
// ----------------------------------------------------------------------

Any*
DynBasic_impl::current_any ()
{
    return any_;
}

Any*
DynBasic_impl::current_any_value ()
{
    return any_;
}

// ----------------------------------------------------------------------
// DynBasic_impl operations
// to marshal and unmarshal components into a complex any value
// ----------------------------------------------------------------------

void
DynBasic_impl::MarshalCount(CountStream& count)
{
    count.count_any_NoTypeCode(any_);
}

void
DynBasic_impl::Marshal(OutputStream& out)
{
    out.write_any_NoTypeCode(any_);
}

void
DynBasic_impl::Unmarshal(InputStream& in)
{
    in.read_any_NoTypeCode(any_.inout(), type_);
}

// ----------------------------------------------------------------------
// DynBasic_impl public IDL member implementation
// from DynAny interface
// ----------------------------------------------------------------------

void
DynBasic_impl::assign(DynAny_ptr dyn_any)
    throw(DynAny::Invalid, SystemException)   
{
   TypeCode_var tc = dyn_any->type();

    if ( ! tc->equivalent(type_) )
        throw DynAny::Invalid ();

    any_ = dyn_any->to_any ();
}

void
DynBasic_impl::from_any(const Any& value)
    throw(DynAny::Invalid, SystemException)
{
    TypeCode_var tc = value.type();

    if ( ! tc->equivalent(type_) )
        throw DynAny::Invalid ();
    
    any_ = new Any (value);
    any_ -> type(type_);
}

Any*
DynBasic_impl::to_any()
    throw(DynAny::Invalid, SystemException)
{
    return new Any (any_);
}

DynAny_ptr
DynBasic_impl::copy()
    throw(SystemException)
{
    TypeCode_var tc = any_->type();

    DynAny_ptr result = new DynBasic_impl (tc);
    result->from_any(any_);
    return result;
}

DynAny_ptr
DynBasic_impl::current_component()
    throw(SystemException)
{
    return DynAny::_duplicate(this);
}

Boolean
DynBasic_impl::next()
    throw(SystemException)
{
    return FALSE;
}

Boolean
DynBasic_impl::seek(Long index)
    throw(SystemException)
{
    if ( index == 0 )
        return TRUE;
    return FALSE;
}

void
DynBasic_impl::rewind()
    throw(SystemException)
{
    // nothing to do
}

// ----------------------------------------------------------------------
// ----------------------------------------------------------------------
//
// DynFixed_impl implementation
//
// ----------------------------------------------------------------------
// ----------------------------------------------------------------------

// ----------------------------------------------------------------------
// DynFixed_impl constructor and destructor
// ----------------------------------------------------------------------

DynFixed_impl::DynFixed_impl(TypeCode_ptr type,
					 Fixed value)
    : DynAny_impl(type)
    , value_ (value)
{
}

DynFixed_impl::~DynFixed_impl()
{
}

// ----------------------------------------------------------------------
// DynFixed_impl  operations to access components
// ----------------------------------------------------------------------

Any*
DynFixed_impl::current_any ()
{
    return 0;
}

Any*
DynFixed_impl::current_any_value ()
{
    return 0;
}

// ----------------------------------------------------------------------
// DynFixed_impl operations
// to marshal and unmarshal components into a complex any value
// ----------------------------------------------------------------------

void
DynFixed_impl::MarshalCount(CountStream& count)
{
    count.count_fixed(value_);
}

void
DynFixed_impl::Marshal(OutputStream& out)
{
    out.write_fixed(value_);
}

void
DynFixed_impl::Unmarshal(InputStream& in)
{
    value_ = in.read_fixed();
}

// ----------------------------------------------------------------------
// DynFixed_impl public IDL member implementation
// from DynAny interface
// ----------------------------------------------------------------------

void
DynFixed_impl::assign(DynAny_ptr dyn_any)
    throw(DynAny::Invalid, SystemException)
{
    DynFixed_var dyn_fixed = DynFixed::_narrow (dyn_any);
    if ( is_nil(dyn_fixed) )
        throw DynAny::Invalid ();

    TypeCode_var dyn_fixed_type = dyn_fixed->type();
    if ( ! dyn_fixed_type->equivalent(type_) )
        throw DynAny::Invalid ();

    DynAny::OctetSeq_var v = dyn_fixed->get_value();
    set_value(v);
}

void
DynFixed_impl::from_any(const Any& value)
    throw(DynAny::Invalid, SystemException)
{
    TypeCode_var tc = value.type();

    if ( ! tc->equivalent(type_) )
        throw DynAny::Invalid ();
//XXX
//    value >>= value_;
}

Any*
DynFixed_impl::to_any()
    throw(DynAny::Invalid, SystemException)
{
    return new Any (type_, new Fixed(value_), TRUE);
}

DynAny_ptr
DynFixed_impl::copy()
    throw(SystemException)
{
    return new DynFixed_impl (type_, value_);
}

DynAny_ptr
DynFixed_impl::current_component()
    throw(SystemException)
{
    return DynFixed::_duplicate(this);
}

Boolean
DynFixed_impl::next()
    throw(SystemException)
{
    return FALSE;
}

Boolean
DynFixed_impl::seek(Long index)
    throw(SystemException)
{
    if ( index == 0 )
        return TRUE;
    return FALSE;
}

void
DynFixed_impl::rewind()
    throw(SystemException)
{
    // nothing to do
}

// ----------------------------------------------------------------------
// DynFixed_impl public IDL member implementation
// from DynFixed interface
// ----------------------------------------------------------------------

DynAny::OctetSeq*
DynFixed_impl::get_value()
    throw(SystemException)
{
    ULong len = type_ -> fixed_digits() + 1;
    Octet* buf = new Octet[len];
    OutputStream out(buf, 0, len);
    out.write_fixed(value_);

    return new DynAny::OctetSeq(*(DynAny::OctetSeq*)buf);
}

void
DynFixed_impl::set_value(const DynAny::OctetSeq& val)
    throw(DynAny::InvalidValue, SystemException)
{
    //
    // Check number of digits
    //
    ULong digits = (val[0] & 0xf0) != 0 ? 1 : 0;
    digits += (val.length() - 1) * 2;
    if(digits != type_ -> fixed_digits())
	throw DynAny::InvalidValue();

    InputStream in(val.get_buffer(), 0, val.length(), false);
    value_ = in.read_fixed();
}

// ----------------------------------------------------------------------
// ----------------------------------------------------------------------
//
// DynEnum_impl implementation
//
// ----------------------------------------------------------------------
// ----------------------------------------------------------------------

// ----------------------------------------------------------------------
// DynEnum_impl constructor and destructor
// ----------------------------------------------------------------------

DynEnum_impl::DynEnum_impl(TypeCode_ptr type,
                                       ULong value)
    : DynAny_impl(type)
    , value_ (value)
{
}

DynEnum_impl::~DynEnum_impl()
{
}

// ----------------------------------------------------------------------
// DynEnum_impl  operations to access components
// ----------------------------------------------------------------------

Any*
DynEnum_impl::current_any ()
{
    return 0;
}

Any*
DynEnum_impl::current_any_value ()
{
    return 0;
}

// ----------------------------------------------------------------------
// DynEnum_impl operations
// to marshal and unmarshal components into a complex any value
// ----------------------------------------------------------------------

void
DynEnum_impl::MarshalCount(CountStream& count)
{
    count.count_ulong();
}

void
DynEnum_impl::Marshal(OutputStream& out)
{
    out.write_ulong(value_);
}

void
DynEnum_impl::Unmarshal(InputStream& in)
{
    value_ = in.read_ulong();
}

// ----------------------------------------------------------------------
// DynEnum_impl public IDL member implementation
// from DynAny interface
// ----------------------------------------------------------------------

void
DynEnum_impl::assign(DynAny_ptr dyn_any)
    throw(DynAny::Invalid, SystemException)
{
    DynEnum_var dyn_enum = DynEnum::_narrow (dyn_any);
    if ( is_nil(dyn_enum) )
        throw DynAny::Invalid ();

    TypeCode_var dyn_enum_type = dyn_enum->type();
    if ( ! dyn_enum_type->equivalent(type_) )
        throw DynAny::Invalid ();

    value_ = dyn_enum->value_as_ulong ();
}

void
DynEnum_impl::from_any(const Any& value)
    throw(DynAny::Invalid, SystemException)
{
    TypeCode_var tc = value.type();

    if ( ! tc->equivalent(type_) )
        throw DynAny::Invalid ();

    if ( value.value () == 0 )
        throw DynAny::Invalid ();

    value_ = READ_ENUM_VALUE (value);
}

Any*
DynEnum_impl::to_any()
    throw(DynAny::Invalid, SystemException)
{
    return new Any (type_, new ULong(value_), TRUE);
}

DynAny_ptr
DynEnum_impl::copy()
    throw(SystemException)
{
    return new DynEnum_impl (type_, value_);
}

DynAny_ptr
DynEnum_impl::current_component()
    throw(SystemException)
{
    return DynEnum::_duplicate(this);
}

Boolean
DynEnum_impl::next()
    throw(SystemException)
{
    return FALSE;
}

Boolean
DynEnum_impl::seek(Long index)
    throw(SystemException)
{
    if ( index == 0 )
        return TRUE;
    return FALSE;
}

void
DynEnum_impl::rewind()
    throw(SystemException)
{
    // nothing to do
}

// ----------------------------------------------------------------------
// DynEnum_impl public IDL member implementation
// from DynEnum interface
// ----------------------------------------------------------------------

char*
DynEnum_impl::value_as_string()
    throw(SystemException)
{
    TypeCode_var tc = GetOrigType(type_);

    return string_dup ( tc->member_name (value_) );
}

void
DynEnum_impl::value_as_string(const char* name)
    throw(SystemException)
{
    TypeCode_var tc = GetOrigType(type_);

    ULong i=0;
    ULong mi = tc->member_count ();

    for (i=0; i<mi; i++)
        if ( strcmp(tc->member_name(i), name) == 0 ) {
            value_ = i;
            break;
        }

    if ( i == mi )
     ; // throw exception ??? this is not specified into the CORBA spec
   
}

ULong
DynEnum_impl::value_as_ulong()
    throw(SystemException)
{
    return value_;
}

void
DynEnum_impl::value_as_ulong(ULong value)
    throw(SystemException)
{
    TypeCode_var tc = GetOrigType(type_);

    if ( value < tc->member_count () )
        value_ = value;
    else
       ; // throw exception ??? this is not specified into the CORBA spec
}

// ----------------------------------------------------------------------
// ----------------------------------------------------------------------
//
// DynComplex_impl implementation
//
// ----------------------------------------------------------------------
// ----------------------------------------------------------------------

// ----------------------------------------------------------------------
// DynComplex_impl constructor and destructor
// ----------------------------------------------------------------------

DynComplex_impl::DynComplex_impl(TypeCode_ptr type)
    : DynAny_impl(type)
    , components_ ()
    , index_ (0)
{
}

DynComplex_impl::~DynComplex_impl()
{
}

// ----------------------------------------------------------------------
// DynComplex_impl operation to init components
// ----------------------------------------------------------------------

void
DynComplex_impl::init_components (ULong count)
{
    components_.length (count);
}

// ----------------------------------------------------------------------
// DynComplex_impl operations to access components
// ----------------------------------------------------------------------

Any*
DynComplex_impl::current_any ()
{
    if ( ( index_ < components_.length() )
      && !is_nil(components_[index_]) )
        return dynamic_cast<DynAny_impl*>
            (components_[index_].in()) -> current_any_value();
    return 0;
}

Any*
DynComplex_impl::current_any_value ()
{
    return 0;
}

// ----------------------------------------------------------------------
// DynComplex_impl operations
// to marshal and unmarshal components into a complex any value
// ----------------------------------------------------------------------

void
DynComplex_impl::MarshalCount(CountStream& count)
{
    for (ULong i=0; i<components_.length(); i++) {
        if ( is_nil(components_[i]) )
            throw DynAny::Invalid();

        dynamic_cast<DynAny_impl*>
            (components_[i].in()) -> MarshalCount (count);
    }
}

void
DynComplex_impl::Marshal(OutputStream& out)
{
    for (ULong i=0; i<components_.length(); i++) {
        if ( is_nil(components_[i]) )
            throw DynAny::Invalid();

        dynamic_cast<DynAny_impl*>
            (components_[i].in()) -> Marshal (out);
    }
}

void
DynComplex_impl::Unmarshal(InputStream& in)
{
    for (ULong i=0; i<components_.length(); i++) {
        if ( is_nil(components_[i]) )
            throw DynAny::Invalid();

        dynamic_cast<DynAny_impl*>
            (components_[i].in()) -> Unmarshal (in);
    }
}

// ----------------------------------------------------------------------
// DynComplex_impl public IDL member implementation
// from DynAny interface
// ----------------------------------------------------------------------

void
DynComplex_impl::from_any(const Any& value)
    throw(DynAny::Invalid, SystemException)
{
    TypeCode_var tc = value.type();

    if ( ! tc->equivalent(type_) )
        throw DynAny::Invalid ();

    const ConstructedInfo* info = value.info();

    if ( !info ) {

	const Buffer* bufp = (const Buffer*)value.value();
	InputStream in(*bufp, 0, false);
	Unmarshal(in);

    } else {

        CountStream count;
        info -> marshalCount(value.value(), count);

        Buffer buff(count.count());
        OutputStream out(buff, 0);

        info->marshal(value.value(), out);

        InputStream in(buff, 0, false);
        Unmarshal(in);
    } 
}

Any*
DynComplex_impl::to_any()
    throw(DynAny::Invalid, SystemException)
{
    CountStream count;
    MarshalCount(count);

    Buffer* bufp = new Buffer(count.count());
    OutputStream out(*bufp, 0);

    try
    {
	OB_DEC_TRY_FIX
        Marshal(out);
    }
    catch (DynAny::Invalid& ex)
    {
        delete bufp;
        throw ex;
    }

    return new Any(type_, bufp, TRUE);
}

void
DynComplex_impl::assign(DynAny_ptr dyn_any)
    throw(DynAny::Invalid, SystemException)
{
    TypeCode_var tc = dyn_any->type();

    if ( ! tc->equivalent(type_) )
        throw DynAny::Invalid ();

    DynComplex_impl* dyn_complex =
        dynamic_cast<DynComplex_impl*>(dyn_any);

    if ( dyn_complex == 0)
        throw DynAny::Invalid ();

    for (ULong i=0; i<components_.length(); i++) {
        DynAny_ptr component = dyn_complex->components_[i];
        if ( is_nil(component) )
            components_[i] = DynAny::_nil();
        else
            components_[i] = component -> copy ();
    }
}
  
DynAny_ptr
DynComplex_impl::current_component()
    throw(SystemException)
{
    if ( index_ < components_.length() ) {
        return DynAny::_duplicate(components_[index_]);
	}

   return DynAny::_nil();
}

Boolean
DynComplex_impl::next()
    throw(SystemException)
{
    index_ ++;
    if ( index_ < components_.length() )
        return TRUE;
    return FALSE;
}

Boolean
DynComplex_impl::seek(Long index)
    throw(SystemException)
{
    if ( ( index >= 0 ) && ( index < Long(components_.length()) ) ) {
        index_ = index;
        return TRUE;
    }
    return FALSE;
}

void
DynComplex_impl::rewind()
    throw(SystemException)
{
    index_ = 0;
}

// ----------------------------------------------------------------------
// ----------------------------------------------------------------------
//
// DynStruct_impl implementation
//
// ----------------------------------------------------------------------
// ----------------------------------------------------------------------

// ----------------------------------------------------------------------
// DynStruct_impl constructor and destructor
// ----------------------------------------------------------------------

DynStruct_impl::DynStruct_impl(TypeCode_ptr type,
                                           Boolean init)
    : DynComplex_impl(type)
    , DynAny_impl (type)
{
    TypeCode_var tc = GetOrigType(type);

    ULong member_count = tc->member_count();
    DynComplex_impl::init_components(member_count);

    if ( init )
       for (ULong i=0; i<member_count; i++)
       {
           TypeCode_var member_type = tc -> member_type(i);
           components_[i] = DynAny_impl::create (member_type);

       }
}

DynStruct_impl::~DynStruct_impl()
{
}

// ----------------------------------------------------------------------
// DynStruct_impl public IDL member implementation
// from DynAny interface
// ----------------------------------------------------------------------

DynAny_ptr
DynStruct_impl::copy()
    throw(SystemException)
{
    DynStruct_impl* result =
                      new DynStruct_impl (type_, FALSE);

    for (ULong i=0; i<components_.length(); i++) {
        result->components_[i] = components_[i] -> copy ();
    }

    return result;
}

// ----------------------------------------------------------------------
// DynStruct_impl public IDL member implementation
// from DynStruct interface
// ----------------------------------------------------------------------

FieldName
DynStruct_impl::current_member_name()
    throw(SystemException)
{
    TypeCode_var tc = GetOrigType(type_);

    return string_dup ( tc->member_name(index_) );
}

TCKind
DynStruct_impl::current_member_kind()
    throw(SystemException)
{
    TypeCode_var tc = GetOrigType(type_);

    TypeCode_var member_type = tc->member_type (index_);
    return member_type -> kind ();
}

NameValuePairSeq*
DynStruct_impl::get_members()
    throw(SystemException)
{
    TypeCode_var tc = GetOrigType(type_);

    NameValuePairSeq* result = new NameValuePairSeq;
    result->length (components_.length());

    for (ULong i=0; i<components_.length(); i++) {
        (*result)[i].id = string_dup (tc->member_name (i));
        Any_var any = components_[i]->to_any ();
        (*result)[i].value = any;
    }
   
    return result;
}

void
DynStruct_impl::set_members(const NameValuePairSeq& value)
    throw(DynAny::InvalidSeq, SystemException)
{
    if ( value.length() != components_.length() )
	throw DynAny::InvalidSeq ();
    
    TypeCode_var tc = GetOrigType(type_);

    ULong i;
    
    for (i=0; i<components_.length(); i++) {
	if ( strcmp (value[i].id, tc->member_name(i)) != 0 )
	    throw DynAny::InvalidSeq ();
	
	TypeCode_var value_type = value[i].value.type();
	TypeCode_var member_type = tc->member_type(i);
	if ( ! value_type->equivalent(member_type) )
	    throw DynAny::InvalidSeq ();
    }
    
    for (i=0; i<components_.length(); i++) {
	components_[i]->from_any (value[i].value);
    }
}

// ----------------------------------------------------------------------
// ----------------------------------------------------------------------
//
// DynException_impl implementation
//
// ----------------------------------------------------------------------
// ----------------------------------------------------------------------

// ----------------------------------------------------------------------
// DynException_impl constructor and destructor
// ----------------------------------------------------------------------

DynException_impl::DynException_impl(TypeCode_ptr type,
                                                 Boolean init)
    : DynStruct_impl(type,init)
    , DynComplex_impl(type)
    , DynAny_impl (type)
    , id_( type->id() )
{
}

DynException_impl::~DynException_impl()
{
}

// ----------------------------------------------------------------------
// DynException_impl operations
// to marshal and unmarshal components into a complex any value
// ----------------------------------------------------------------------

void
DynException_impl::MarshalCount(CountStream& count)
{
    count.count_string(id_);
    DynComplex_impl::MarshalCount(count);
}

void
DynException_impl::Marshal(OutputStream& out)
{
    out.write_string(id_);
    DynComplex_impl::Marshal(out);
}

void
DynException_impl::Unmarshal(InputStream& in)
{
    String_var id = in.read_string();
    assert(strcmp(id, id_) == 0);
    DynComplex_impl::Unmarshal(in);
}

// ----------------------------------------------------------------------
// DynException_impl public IDL member implementation
// from DynAny interface
// ----------------------------------------------------------------------

DynAny_ptr
DynException_impl::copy()
    throw(SystemException)
{
    DynException_impl* result =
                      new DynException_impl (type_, FALSE);

    for (ULong i=0; i<components_.length(); i++) {
        result->components_[i] = components_[i] -> copy ();
    }

    return result;
}

// ----------------------------------------------------------------------
// ----------------------------------------------------------------------
//
// DynUnion_impl
//
// ----------------------------------------------------------------------
// ----------------------------------------------------------------------

// ----------------------------------------------------------------------
// DynUnion_impl constructor and destructor
// ----------------------------------------------------------------------

DynUnion_impl::DynUnion_impl(TypeCode_ptr type)
    : DynComplex_impl(type)
    , DynAny_impl (type)
    , default_index_ (-1)
    , current_discr_index_ (-1)
{
    DynComplex_impl::init_components(2);

    TypeCode_var tc = GetOrigType(type);

    ULong count = tc->member_count();

    member_labels_.length(count);
    member_types_.length(count);

    for(ULong i=0; i<count; i++) {
        Any_var label = tc->member_label(i);
        member_labels_[i] = compute_any(label);
        member_types_[i] = tc->member_type(i);
    }

    TypeCode_var discr_type = tc->discriminator_type();

    default_index_ = tc->default_index();

    if ( default_index_ != -1 ) {
        member_labels_[default_index_] = compute_default_value (discr_type);
    }

    components_[0] = DynAny_impl::create(discr_type);
    components_[1] = DynAny::_nil();
}

DynUnion_impl::~DynUnion_impl()
{
}

// ----------------------------------------------------------------------
// private operations
// ----------------------------------------------------------------------

void
DynUnion_impl::check_discriminator_and_set_member ()
{
    Long discr_index = default_index_;

    Any_var any_discr = components_[0]->to_any ();

    if ( any_discr->value() != 0 ) {
        LongLong label = compute_any (any_discr);
        for (ULong i=0; i<member_labels_.length(); i++) {
            if ( member_labels_[i] == label ) {
                discr_index = i;
                break;
            }
        }
    }

    if ( discr_index != current_discr_index_ ) {
        current_discr_index_ = discr_index;
        if ( discr_index != -1 )
            components_[1] =
                DynAny_impl::create (member_types_[discr_index]);
         else
 	    components_[1] = DynAny::_nil();
    }
}

void
DynUnion_impl::set_dyn_any (DynAny_ptr dyn_any,
                                  LongLong value)
{
    TypeCode_var type = dyn_any->type();
    TypeCode_var tc = GetOrigType(type);

    switch (tc->kind())
    {
    case tk_short:
        dyn_any->insert_short((Short)value);
        break;

    case tk_ushort:
        dyn_any->insert_ushort((UShort)value);
        break;

    case tk_long:
        dyn_any->insert_long((Long)value);
        break;

    case tk_ulong:
        dyn_any->insert_ulong((ULong)value);
        break;

    case tk_longlong:
        dyn_any->insert_longlong((LongLong)value);
        break;

    case tk_ulonglong:
        dyn_any->insert_ulonglong((ULongLong)value);
        break;

    case tk_char:
        dyn_any->insert_char((Char)value);
        break;

    case tk_wchar:
        dyn_any->insert_wchar((WChar)value);
        break;

    case tk_boolean:
        dyn_any->insert_boolean((Boolean)value);
        break;

    case tk_octet:
        dyn_any->insert_octet((Octet)value);
        break;
    
    case tk_enum: {
        DynEnum_var dyn_enum = DynEnum::_narrow (dyn_any);
        dyn_enum->value_as_ulong ((ULong)value);
      } break;

    default:
       assert(false);
       break;
    }
}

LongLong
DynUnion_impl::compute_any (Any* any)
{
    LongLong result = 0;

    TypeCode_var type = any->type();
    TypeCode_var tc = GetOrigType(type);

    switch (tc->kind())
    {
    case tk_short: {
        Short v_short;
        *any >>= v_short;
        result = Long(v_short);
      } break;

    case tk_ushort: {
        UShort v_ushort;
        *any >>= v_ushort;
        result = Long(v_ushort);
      } break;

    case tk_long: {
        Long v_long;
        *any >>= v_long;
        result = Long(v_long);
      } break;

    case tk_ulong: {
        ULong v_ulong;
        *any >>= v_ulong;
        result = Long(v_ulong);
      } break;

    case tk_longlong: {
        LongLong v_longlong;
        *any >>= v_longlong;
        result = LongLong(v_longlong);
      } break;

    case tk_ulonglong: {
        ULongLong v_ulonglong;
        *any >>= v_ulonglong;
        result = LongLong(v_ulonglong);
      } break;

    case tk_char: {
        Char v_char;
        Any::to_char to_char (v_char);
        *any >>= to_char;
        result = Long(v_char);
      } break;

    case tk_boolean: {
        Boolean v_boolean;
        Any::to_boolean to_boolean (v_boolean);
        *any >>= to_boolean;
        result = Long(v_boolean);
      } break;

    case tk_octet: {
        Octet v_octet;
        Any::to_octet to_octet (v_octet);
        *any >>= to_octet;
        result = Long(v_octet);
      } break;
    
    case tk_enum: {
        ULong v_enum = READ_ENUM_VALUE (*any);
        result = Long(v_enum);
      } break;

    default:
       assert(false);
       break;
    }
    return result;
}

//
// WARNING
//
// This is a simplified version
// It doesn't work for big ULong discriminator (>2147483647)
//

LongLong
DynUnion_impl::compute_default_value (TypeCode_ptr discr_type)
{
    LongLong result = 0;
    LongLong min, max;
 
    TypeCode_var tc = GetOrigType(discr_type);
   
    switch(tc -> kind())
    {
    case tk_short:
	min = -32768;
	max = 32767;
	break;
	
    case tk_ushort:
	min = 0;
	max = 65535;
	break;
	
    case tk_long:
	min = -2147483647 - 1;
	max = 2147483647;
	break;
	
    case tk_ulong:
	min = 0;
	max = 2147483647;
//	max = 4294967294U;
	break;

    case tk_longlong:
	min = -2147483647 - 1;
//	min = -9223372036854775807L - 1;
	max = 2147483647;
//	max = 9223372036854775807U;
	break;
	
    case tk_ulonglong:
	min = 0;
	max = 2147483647;
//	max = 9223372036854775807U;
	break;
	
    case tk_enum:
	min = 0;
	max = tc -> member_count() - 1;
	break;
	
    case tk_char:
	min = 0;
	max = 255;
	break;
	
    case tk_boolean:
	min = 0;
	max = 1;
	break;
	
    default:
	assert(false);
	break;
    }

    for(LongLong i = max ; i >= min ; i--) {
	bool ok = true;
	
	for(ULong j = 0 ; ok && j < member_labels_.length() ; j++)
	    if(member_labels_[j] == i)
		    ok = false;
	
	if(ok) {
	    result = i;
	    break;
	}
    }
    return result;
}

// ----------------------------------------------------------------------
// DynUnion_impl operations
// to marshal and unmarshal components into a complex any value
// ----------------------------------------------------------------------

void
DynUnion_impl::Unmarshal(InputStream& in)
{
    if ( is_nil(components_[0]) )
        throw DynAny::Invalid();
    dynamic_cast<DynAny_impl*>
        (components_[0].in()) -> Unmarshal(in);
    
    check_discriminator_and_set_member ();
    
    if ( is_nil(components_[1]) )
        throw DynAny::Invalid();
    dynamic_cast<DynAny_impl*>
        (components_[1].in()) -> Unmarshal(in);
}

// ----------------------------------------------------------------------
// DynUnion_impl public IDL member implementation
// from DynAny interface
// ----------------------------------------------------------------------

void
DynUnion_impl::assign(DynAny_ptr dyn_any)
    throw(DynAny::Invalid, SystemException)
{
    TypeCode_var tc = dyn_any->type();

    if ( ! tc->equivalent(type_) )
        throw DynAny::Invalid ();

    DynUnion_impl* dyn_union =
        dynamic_cast<DynUnion_impl*>(dyn_any);

    if (dyn_union == 0)
        throw DynAny::Invalid ();

    DynAny_var component = dyn_union->components_[0];

    if ( is_nil(component) )
        throw DynAny::Invalid ();

    components_[0] = component -> copy ();

    check_discriminator_and_set_member ();

    component = dyn_union->components_[1];

    if ( is_nil(component) )
        components_[1] = DynAny::_nil();
    else
        components_[1] = component -> copy ();
}
  
DynAny_ptr
DynUnion_impl::copy()
    throw(SystemException)
{
    DynUnion_impl* result = new DynUnion_impl (type_);

    result->current_discr_index_ = current_discr_index_;
 
    for (ULong i=0; i<2; i++) {
        DynAny_ptr component = components_[i];
        if ( !is_nil(component) )
            result->components_[i] = components_[i] -> copy ();
        else
            result->components_[i] = DynAny::_nil();
    }

    return result;
}

Boolean
DynUnion_impl::next()
    throw(SystemException)
{
    Boolean boolean = DynComplex_impl::next();

    if ( boolean && ( index_ == 1 ) ) {
        check_discriminator_and_set_member ();
    }

    return boolean;
}

// ----------------------------------------------------------------------
// DynUnion_impl public IDL member implementation
// from DynUnion interface
// ----------------------------------------------------------------------

//
// WARNING
//
// I not sure than I have understand
// the semantic of the set_as_default attribute
//

Boolean
DynUnion_impl::set_as_default()
    throw(SystemException)
{
    return ( default_index_ != -1 )
        && ( current_discr_index_ == default_index_);
}

void
DynUnion_impl::set_as_default(Boolean b)
    throw(SystemException)
{
    if ( ! b )
        return;

    if ( default_index_ == -1 )
        return;

    set_dyn_any (components_[0], member_labels_[default_index_]);
next();
    check_discriminator_and_set_member ();
    index_ = 1;
}

DynAny_ptr
DynUnion_impl::discriminator()
    throw(SystemException)
{
    return DynAny::_duplicate(components_[0]);
}

TCKind
DynUnion_impl::discriminator_kind()
    throw(SystemException)
{
    TypeCode_var discriminator_type = components_[0]->type();
    return discriminator_type->kind();
}

DynAny_ptr
DynUnion_impl::member()
    throw(SystemException)
{
    check_discriminator_and_set_member ();
    return DynAny::_duplicate(components_[1]);
}

FieldName
DynUnion_impl::member_name()
    throw(SystemException)
{
    if ( current_discr_index_ == -1 )
        return string_dup("");

    TypeCode_var tc = GetOrigType(type_);

    return string_dup (
        tc -> member_name ( ULong(current_discr_index_) ) );
}

void
DynUnion_impl::member_name(const char* name)
    throw(SystemException)
{
    TypeCode_var tc = GetOrigType(type_);

    for (ULong i=0; i<member_labels_.length(); i++)
        if ( strcmp(name, tc->member_name(i)) == 0 ) {
            set_dyn_any (components_[0], member_labels_[i]);
            check_discriminator_and_set_member ();
            index_ = 1;
            return;
        }

    // Must throw an exception?
}

TCKind
DynUnion_impl::member_kind()
    throw(SystemException)
{
    if ( ! is_nil (components_[1]) ) {
        TypeCode_var member_type = components_[1]->type();
        return member_type->kind();
    }

    // Must throw an exception?

    return tk_null;
}

// ----------------------------------------------------------------------
// ----------------------------------------------------------------------
//
// DynSequence implementation
//
// ----------------------------------------------------------------------
// ----------------------------------------------------------------------

// ----------------------------------------------------------------------
// DynSequence_impl constructor and destructor
// ----------------------------------------------------------------------

DynSequence_impl::DynSequence_impl(TypeCode_ptr type)
    : DynComplex_impl(type)
    , DynAny_impl (type)
{
}

DynSequence_impl::~DynSequence_impl()
{
}

// ----------------------------------------------------------------------
// internal operation to init components
// ----------------------------------------------------------------------

void
DynSequence_impl::init_components (ULong value)
{
    TypeCode_var tc = GetOrigType(type_);

    TypeCode_var content_type = tc->content_type();

    DynComplex_impl::init_components (value);

    for (ULong i=0; i<value; i++) {
        if ( is_nil(components_[i]) )
	    components_[i] = DynAny_impl::create (content_type);
    }
}

// ----------------------------------------------------------------------
// DynSequence_impl operations
// to marshal and unmarshal components into a complex any value
// ----------------------------------------------------------------------

void
DynSequence_impl::MarshalCount(CountStream& count)
{
    // Use specific OB function
    count.count_ulong(); // components_.length()
    DynComplex_impl::MarshalCount(count);
}

void
DynSequence_impl::Marshal(OutputStream& out)
{
    // Use specific OB function
    out.write_ulong(components_.length());
    DynComplex_impl::Marshal(out);
}

void
DynSequence_impl::Unmarshal(InputStream& in)
{
    ULong count = in.read_ulong();
    
    init_components(count);
    DynComplex_impl::Unmarshal(in);
}

// ----------------------------------------------------------------------
// DynSequence_impl public IDL member implementation
// from DynAny interface
// ----------------------------------------------------------------------

void
DynSequence_impl::assign(DynAny_ptr dyn_any)
    throw(DynAny::Invalid, SystemException)
{
    TypeCode_var tc = dyn_any->type();

    if ( ! tc->equivalent(type_) )
        throw DynAny::Invalid ();

    DynSequence_impl* dyn_sequence =
        dynamic_cast<DynSequence_impl*>(dyn_any);

    if (dyn_sequence == 0)
        throw DynAny::Invalid ();

    components_.length(dyn_sequence->components_.length());

    for (ULong i=0; i<components_.length(); i++) {
        DynAny_ptr component = dyn_sequence->components_[i];
        if ( is_nil(component) )
            components_[i] = DynAny::_nil();
        else
            components_[i] = component -> copy ();
    }
}
  
DynAny_ptr
DynSequence_impl::copy()
    throw(SystemException)
{
    DynSequence_impl* result =
                      new DynSequence_impl (type_);

    result->components_.length(components_.length());

    for (ULong i=0; i<components_.length(); i++) {
        result->components_[i] = components_[i] -> copy ();
    }

    return result;
}

// ----------------------------------------------------------------------
// DynSequence_impl public IDL member implementation
// from DynSequence interface
// ----------------------------------------------------------------------

ULong
DynSequence_impl::length()
    throw(SystemException)
{
    return components_.length();
}

void
DynSequence_impl::length(ULong value)
    throw(SystemException)
{
    init_components (value);
}

CORBA::AnySeq*
DynSequence_impl::get_elements()
    throw(SystemException)
{
    CORBA::AnySeq* result = new CORBA::AnySeq;
    result->length(components_.length());

    for (ULong i=0; i<components_.length(); i++) {
        Any_var any = components_[i]->to_any();
        (*result)[i] = any; 
    }
    return result;
}

void
DynSequence_impl::set_elements(const CORBA::AnySeq& value)
    throw(DynAny::InvalidSeq, SystemException)
{
    if (value.length() != components_.length())
        throw DynAny::InvalidSeq ();

    TypeCode_var tc = GetOrigType(type_);
    TypeCode_var content_type = tc->content_type();

    ULong i;
    for (i=0; i<components_.length(); i++) {
        TypeCode_var value_type = value[i].type();
        if ( ! value_type->equivalent(content_type) )
             throw DynAny::InvalidSeq ();
    }

    for (i=0; i<components_.length(); i++)
        components_[i]->from_any (value[i]);
}

// ----------------------------------------------------------------------
// ----------------------------------------------------------------------
//
// DynArray_impl implementation
//
// ----------------------------------------------------------------------
// ----------------------------------------------------------------------

// ----------------------------------------------------------------------
// DynArray_impl constructor and destructor
// ----------------------------------------------------------------------

DynArray_impl::DynArray_impl(TypeCode_ptr type,
                                         Boolean init)
    : DynComplex_impl(type)
    , DynAny_impl (type)
{
    TypeCode_var tc = GetOrigType(type);

    DynComplex_impl::init_components(tc->length());

    if (init) {
        TypeCode_var content_type = tc->content_type();
        for (ULong i=0; i<components_.length(); i++) {
	    components_[i] = DynAny_impl::create (content_type);
        }
    }
}

DynArray_impl::~DynArray_impl()
{
}

// ----------------------------------------------------------------------
// DynArray_impl public IDL member implementation
// from DynAny interface
// ----------------------------------------------------------------------

DynAny_ptr
DynArray_impl::copy()
    throw(SystemException)
{
    DynArray_impl* result =
                      new DynArray_impl (type_, FALSE);

    for (ULong i=0; i<components_.length(); i++) {
        result->components_[i] = components_[i] -> copy ();
    }

    return result;
}

// ----------------------------------------------------------------------
// DynArray_impl public IDL member implementation
// from DynArray interface
// ----------------------------------------------------------------------

CORBA::AnySeq*
DynArray_impl::get_elements()
    throw(SystemException)
{
    CORBA::AnySeq* result = new CORBA::AnySeq;
    result->length(components_.length());

    for (ULong i=0; i<components_.length(); i++) {
        Any_var any = components_[i]->to_any();
        (*result)[i] = any; 
    }
    return result;
}

void
DynArray_impl::set_elements(const CORBA::AnySeq& value)
    throw(DynAny::InvalidSeq, SystemException)
{
    if (value.length() != components_.length())
        throw DynAny::InvalidSeq ();

    TypeCode_var tc = GetOrigType(type_);

    TypeCode_var content_type = tc->content_type();

    ULong i;

    for (i=0; i<components_.length(); i++) {
        TypeCode_var value_type = value[i].type();
        if ( ! value_type->equivalent(content_type) )
             throw DynAny::InvalidSeq ();
    }

    for (i=0; i<components_.length(); i++)
        components_[i]->from_any (value[i]);
}

// ----------------------------------------------------------------------
// ----------------------------------------------------------------------
//
// ORB public member implementations that are related to DynAny
//
// ----------------------------------------------------------------------
// ----------------------------------------------------------------------

DynAny_ptr
ORB::create_dyn_any(const Any& value)
{
    TypeCode_var tc = value.type();
    DynAny_ptr result = DynAny_impl::create (tc);
    result->from_any (value);
    return result;
}

DynAny_ptr
ORB::create_basic_dyn_any(TypeCode_ptr type)
{
    TypeCode_var tc = GetOrigType(type);

    switch(tc -> kind())
    {
    case tk_null:
    case tk_void:
    case tk_short:
    case tk_long:
    case tk_longlong:
    case tk_ushort:
    case tk_ulong:
    case tk_ulonglong:
    case tk_float:
    case tk_double:
    case tk_longdouble:
    case tk_boolean:
    case tk_char:
    case tk_wchar:
    case tk_octet:
    case tk_any:
    case tk_TypeCode:
    case tk_Principal:
    case tk_objref:
    case tk_string:
    case tk_wstring:
	break;

    case tk_struct:
    case tk_union:
    case tk_fixed:
    case tk_enum:
    case tk_sequence:
    case tk_array:
    case tk_except:
    case tk_value:
    case tk_value_box:
    case tk_native:
	throw InconsistentTypeCode();

    case tk_alias:
	assert(false);
	break;
    }

    return new DynBasic_impl(type);
}

DynStruct_ptr
ORB::create_dyn_struct(TypeCode_ptr type)
{
    TypeCode_var tc = GetOrigType(type);

    if(tc -> kind() == tk_struct)
        return new DynStruct_impl(type);

    if(tc -> kind() == tk_except)
        return new DynException_impl(type);

    throw InconsistentTypeCode();
    return DynStruct::_nil(); // Some compilers need this
}

DynSequence_ptr
ORB::create_dyn_sequence(TypeCode_ptr type)
{
    TypeCode_var tc = GetOrigType(type);

    if(tc -> kind() != tk_sequence)
	throw InconsistentTypeCode();

    return new DynSequence_impl(type);
}

DynArray_ptr
ORB::create_dyn_array(TypeCode_ptr type)
{
    TypeCode_var tc = GetOrigType(type);

    if(tc -> kind() != tk_array)
	throw InconsistentTypeCode();

    return new DynArray_impl(type);
}

DynUnion_ptr
ORB::create_dyn_union(TypeCode_ptr type)
{
    TypeCode_var tc = GetOrigType(type);

    if(tc -> kind() != tk_union)
	throw InconsistentTypeCode();

    return new DynUnion_impl(type);
}

DynEnum_ptr
ORB::create_dyn_enum(TypeCode_ptr type)
{
    TypeCode_var tc = GetOrigType(type);

    if(tc -> kind() != tk_enum)
	throw InconsistentTypeCode();

    return new DynEnum_impl(type);
}

DynFixed_ptr
ORB::create_dyn_fixed(TypeCode_ptr type)
{
    TypeCode_var tc = GetOrigType(type);

    if(tc -> kind() != tk_fixed)
	throw InconsistentTypeCode();

    assert_nca(false, NCANotImpl);
    return new DynFixed_impl(type);
}

// ----------------------------------------------------------------------
// end of file
// ----------------------------------------------------------------------
