// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/Any.h>
#include <OB/OBV.h>
#include <OB/TCKind.h>
#include <OB/TypeCode.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/POAServantBase.h>
#include <OB/IFR1.h>
#include <OB/IFR4.h>
#include <OB/ValueBase.h>
#include <OB/Stream.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// SystemException public member implementation
// ----------------------------------------------------------------------

SystemException::SystemException(const char* reason)
    : reason_(reason), minor_(0), status_(COMPLETED_NO)
{
}

SystemException::SystemException(ULong minor, CompletionStatus status)
    : reason_(string_dup("")), minor_(minor), status_(status)
{
}

SystemException::SystemException(const char* reason,
				 ULong minor,
				 CompletionStatus status)
    : reason_(reason), minor_(minor), status_(status)
{
}

SystemException::SystemException(const SystemException& se)
    : reason_(se.reason_), minor_(se.minor_), status_(se.status_)
{
}

SystemException&
SystemException::operator=(const SystemException& se)
{
    if(this != &se)
    {
	reason_ = se.reason_;
	minor_ = se.minor_;
	status_ = se.status_;
    }
    return *this;
}

// ----------------------------------------------------------------------
// UserException public member implementation
// ----------------------------------------------------------------------

/* None yet */

// ----------------------------------------------------------------------
// All specific system exception public member implementation
// ----------------------------------------------------------------------

#define OB_MAKE_STANDARD_EXCEPTION_I(T) \
\
T::T(const char* r) \
    : SystemException(r) \
{ \
} \
T::T(ULong m, CompletionStatus s) \
    : SystemException(m, s) \
{ \
} \
T::T(const char* r, ULong m, CompletionStatus s) \
    : SystemException(r, m, s) \
{ \
} \
T::T(const T& ex) \
    : SystemException(ex) \
{ \
} \
\
Exception* \
T::_OB_clone() const \
{ \
    return new T(*this); \
} \
const char* \
T::_OB_typeId() const \
{ \
    return "IDL:omg.org/CORBA/" #T ":1.0"; \
}

/* Note: The _OB_insertInAny() operations for all the system
   exceptions are defined in Any.cpp */

OB_MAKE_STANDARD_EXCEPTION_I(UNKNOWN)
OB_MAKE_STANDARD_EXCEPTION_I(BAD_PARAM)
OB_MAKE_STANDARD_EXCEPTION_I(NO_MEMORY)
OB_MAKE_STANDARD_EXCEPTION_I(IMP_LIMIT)
OB_MAKE_STANDARD_EXCEPTION_I(COMM_FAILURE)
OB_MAKE_STANDARD_EXCEPTION_I(INV_OBJREF)
OB_MAKE_STANDARD_EXCEPTION_I(NO_PERMISSION)
OB_MAKE_STANDARD_EXCEPTION_I(INTERNAL)
OB_MAKE_STANDARD_EXCEPTION_I(MARSHAL)
OB_MAKE_STANDARD_EXCEPTION_I(INITIALIZE)
OB_MAKE_STANDARD_EXCEPTION_I(NO_IMPLEMENT)
OB_MAKE_STANDARD_EXCEPTION_I(BAD_TYPECODE)
OB_MAKE_STANDARD_EXCEPTION_I(BAD_OPERATION)
OB_MAKE_STANDARD_EXCEPTION_I(NO_RESOURCES)
OB_MAKE_STANDARD_EXCEPTION_I(NO_RESPONSE)
OB_MAKE_STANDARD_EXCEPTION_I(PERSIST_STORE)
OB_MAKE_STANDARD_EXCEPTION_I(BAD_INV_ORDER)
OB_MAKE_STANDARD_EXCEPTION_I(TRANSIENT)
OB_MAKE_STANDARD_EXCEPTION_I(FREE_MEM)
OB_MAKE_STANDARD_EXCEPTION_I(INV_IDENT)
OB_MAKE_STANDARD_EXCEPTION_I(INV_FLAG)
OB_MAKE_STANDARD_EXCEPTION_I(INTF_REPOS)
OB_MAKE_STANDARD_EXCEPTION_I(BAD_CONTEXT)
OB_MAKE_STANDARD_EXCEPTION_I(OBJ_ADAPTER)
OB_MAKE_STANDARD_EXCEPTION_I(DATA_CONVERSION)
OB_MAKE_STANDARD_EXCEPTION_I(OBJECT_NOT_EXIST)
OB_MAKE_STANDARD_EXCEPTION_I(TRANSACTION_REQUIRED)
OB_MAKE_STANDARD_EXCEPTION_I(TRANSACTION_ROLLEDBACK)
OB_MAKE_STANDARD_EXCEPTION_I(INVALID_TRANSACTION)
OB_MAKE_STANDARD_EXCEPTION_I(INV_POLICY)

// ----------------------------------------------------------------------
// CDR functions
// ----------------------------------------------------------------------

void
SystemException::_OB_marshal(const SystemException& val, OutputStream& out)
{
    out.write_string(val._OB_typeId());
    out.write_ulong(val.minor());
    out.write_ulong((ULong)val.completed());
}

void
SystemException::_OB_marshalCount(const SystemException& val,
                                  CountStream& count)
{
    count.count_string(val._OB_typeId());
    count.count_ulong(); // val.minor()
    count.count_ulong(); // (ULong)val.completed()
}

SystemException*
SystemException::_OB_unmarshal(InputStream& in)
{
    String_var id = in.read_string();
    ULong minor = in.read_ulong();
    CompletionStatus status = (CompletionStatus)in.read_ulong();

    if(strcmp(id, "IDL:omg.org/CORBA/BAD_PARAM:1.0") == 0)
    {
	return new BAD_PARAM(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/NO_MEMORY:1.0") == 0)
    {
	return new NO_MEMORY(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/IMP_LIMIT:1.0") == 0)
    {
	return new IMP_LIMIT(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/COMM_FAILURE:1.0") == 0)
    {
	return new COMM_FAILURE(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/INV_OBJREF:1.0") == 0)
    {
	return new INV_OBJREF(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/NO_PERMISSION:1.0") == 0)
    {
	return new NO_PERMISSION(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/INTERNAL:1.0") == 0)
    {
	return new INTERNAL(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/MARSHAL:1.0") == 0)
    {
	return new MARSHAL(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/INITIALIZE:1.0") == 0)
    {
	return new INITIALIZE(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/NO_IMPLEMENT:1.0") == 0)
    {
	return new NO_IMPLEMENT(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/BAD_TYPECODE:1.0") == 0)
    {
	return new BAD_TYPECODE(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/BAD_OPERATION:1.0") == 0)
    {
	return new BAD_OPERATION(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/NO_RESOURCES:1.0") == 0)
    {
	return new NO_RESOURCES(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/NO_RESPONSE:1.0") == 0)
    {
	return new NO_RESPONSE(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/PERSIST_STORE:1.0") == 0)
    {
	return new PERSIST_STORE(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/BAD_INV_ORDER:1.0") == 0)
    {
	return new BAD_INV_ORDER(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/TRANSIENT:1.0") == 0)
    {
	return new TRANSIENT(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/FREE_MEM:1.0") == 0)
    {
	return new FREE_MEM(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/INV_IDENT:1.0") == 0)
    {
	return new INV_IDENT(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/INV_FLAG:1.0") == 0)
    {
	return new INV_FLAG(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/INTF_REPOS:1.0") == 0)
    {
	return new INTF_REPOS(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/BAD_CONTEXT:1.0") == 0)
    {
	return new BAD_CONTEXT(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/OBJ_ADAPTER:1.0") == 0)
    {
	return new OBJ_ADAPTER(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/DATA_CONVERSION:1.0") == 0)
    {
	return new DATA_CONVERSION(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/OBJECT_NOT_EXIST:1.0") == 0)
    {
	return new OBJECT_NOT_EXIST(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/TRANSACTION_REQUIRED:1.0") == 0)
    {
	return new TRANSACTION_REQUIRED(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/TRANSACTION_ROLLEDBACK:1.0") == 0)
    {
	return new TRANSACTION_ROLLEDBACK(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/INVALID_TRANSACTION:1.0") == 0)
    {
	return new INVALID_TRANSACTION(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/INV_POLICY:1.0") == 0)
    {
	return new INV_POLICY(minor, status);
    }

    //
    // Unknown exception
    //
    return new UNKNOWN(minor, status);
}

void
SystemException::_OB_unmarshalThrow(InputStream& in)
{
    String_var id = in.read_string();
    ULong minor = in.read_ulong();
    CompletionStatus status = (CompletionStatus)in.read_ulong();

    if(strcmp(id, "IDL:omg.org/CORBA/BAD_PARAM:1.0") == 0)
    {
	throw BAD_PARAM(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/NO_MEMORY:1.0") == 0)
    {
	throw NO_MEMORY(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/IMP_LIMIT:1.0") == 0)
    {
	throw IMP_LIMIT(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/COMM_FAILURE:1.0") == 0)
    {
	throw COMM_FAILURE(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/INV_OBJREF:1.0") == 0)
    {
	throw INV_OBJREF(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/NO_PERMISSION:1.0") == 0)
    {
	throw NO_PERMISSION(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/INTERNAL:1.0") == 0)
    {
	throw INTERNAL(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/MARSHAL:1.0") == 0)
    {
	throw MARSHAL(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/INITIALIZE:1.0") == 0)
    {
	throw INITIALIZE(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/NO_IMPLEMENT:1.0") == 0)
    {
	throw NO_IMPLEMENT(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/BAD_TYPECODE:1.0") == 0)
    {
	throw BAD_TYPECODE(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/BAD_OPERATION:1.0") == 0)
    {
	throw BAD_OPERATION(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/NO_RESOURCES:1.0") == 0)
    {
	throw NO_RESOURCES(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/NO_RESPONSE:1.0") == 0)
    {
	throw NO_RESPONSE(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/PERSIST_STORE:1.0") == 0)
    {
	throw PERSIST_STORE(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/BAD_INV_ORDER:1.0") == 0)
    {
	throw BAD_INV_ORDER(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/TRANSIENT:1.0") == 0)
    {
	throw TRANSIENT(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/FREE_MEM:1.0") == 0)
    {
	throw FREE_MEM(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/INV_IDENT:1.0") == 0)
    {
	throw INV_IDENT(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/INV_FLAG:1.0") == 0)
    {
	throw INV_FLAG(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/INTF_REPOS:1.0") == 0)
    {
	throw INTF_REPOS(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/BAD_CONTEXT:1.0") == 0)
    {
	throw BAD_CONTEXT(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/OBJ_ADAPTER:1.0") == 0)
    {
	throw OBJ_ADAPTER(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/DATA_CONVERSION:1.0") == 0)
    {
	throw DATA_CONVERSION(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/OBJECT_NOT_EXIST:1.0") == 0)
    {
	throw OBJECT_NOT_EXIST(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/TRANSACTION_REQUIRED:1.0") == 0)
    {
	throw TRANSACTION_REQUIRED(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/TRANSACTION_ROLLEDBACK:1.0") == 0)
    {
	throw TRANSACTION_ROLLEDBACK(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/INVALID_TRANSACTION:1.0") == 0)
    {
	throw INVALID_TRANSACTION(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/INV_POLICY:1.0") == 0)
    {
	throw INV_POLICY(minor, status);
    }

    //
    // Unknown exception
    //
    throw UNKNOWN(minor, status);
}
