// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_GIOP_CLIENT_H
#define OB_GIOP_CLIENT_H

namespace OB
{

class GIOPClientWorker;
typedef GIOPClientWorker* GIOPClientWorker_ptr;

class GIOPClient;
typedef GIOPClient* GIOPClient_ptr;

} // End of namespace OB

void OBDuplicate(OB::GIOPClientWorker_ptr);
void OBRelease(OB::GIOPClientWorker_ptr);
void OBDuplicate(OB::GIOPClient_ptr);
void OBRelease(OB::GIOPClient_ptr);

namespace OB
{

//
// Declare GIOPClientWorker as a class
//
typedef ObjVar< GIOPClientWorker > GIOPClientWorker_var;

//
// The GIOPClient class
//
typedef ObjVar< GIOPClient > GIOPClient_var;

class GIOPClient : public Client
{
    //
    // Hide copy-constructor and assignment operator
    //
    GIOPClient(const GIOPClient&);
    void operator=(const GIOPClient&);

protected:

    //
    // The connector
    //
    OCI::Connector_var connector_;
    
    //
    // The worker
    //
    GIOPClientWorker_var worker_;
#ifdef HAVE_JTC
    JTCMutex workerMutex_;
#endif

    //
    // Add message header and request header
    //
    void addHeaders(StubImplBase*, const char*, CORBA::ULong, bool, Buffer&);

    //
    // Get the worker
    //
    GIOPClientWorker_ptr getWorker();
    
public:

    GIOPClient(OCI::Connector_ptr);
    virtual ~GIOPClient();

    static inline GIOPClient_ptr _duplicate(GIOPClient_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline GIOPClient_ptr _nil()
    { return 0; }

    //
    // Destroy the client
    //
    virtual void destroy();

    //
    // Destroy the worker
    //
    void destroyWorker(GIOPClientWorker_ptr);

    //
    // Get the OCI Connector
    //
    virtual OCI::Connector_ptr connector();

    //
    // Get the OCI Transport
    //
    virtual OCI::Transport_ptr transport();
    
    //
    // Calculate offset
    //
    virtual CORBA::ULong offset(StubImplBase*, const char*);

    //
    // Send request and receive response
    // With explicit timeout (throws exception on timeout)
    // With optional retry on communication failure
    //
    virtual CORBA::ULong request(StubImplBase*, const char*, Buffer&,
                                 bool&, bool&, bool&, CORBA::Long, bool);

    //
    // Send request oneway
    // With implicit zero timeout (does not throw exception on timeout)
    // With optional retry on communication failure
    //
    virtual void oneway(StubImplBase*, const char*, Buffer&, bool);

    //
    // Send request deferred
    // With implicit zero timeout (does not throw exception on timeout)
    // With optional retry on communication failure
    //
    virtual void deferred(StubImplBase*, const char*, Buffer&,
                          CORBA::ULong&, bool);

    //
    // Get repsonse
    // With explicit timeout (throws exception on timeout)
    //
    virtual CORBA::ULong response(CORBA::ULong, Buffer&, bool&, bool&, bool&,
				  CORBA::Long);
    
    //
    // Poll response
    // With implicit zero timeout (returns zero on timeout)
    //
    virtual CORBA::ULong poll(CORBA::ULong, Buffer&, bool&, bool&, bool&);
};

} // End of namespace OB


namespace CORBA
{
// XXX Are these necessary in CORBA?

inline void
release(OB::GIOPClient_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::GIOPClient_ptr p)
{
    return p == 0;
}

} // End of namespace CORBA

#endif
