// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/TemplateI.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/Any.h>
#include <OB/OBV.h>
#include <OB/TCKind.h>
#include <OB/TypeCode.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/POAServantBase.h>
#include <OB/IFR1.h>
#include <OB/IFR4.h>
#include <OB/ValueBase.h>
#include <OB/GIOP.h>
#include <OB/OCI.h>
#include <OB/Reactor.h>
#include <OB/Stream.h>

#include <GIOPClient.h>
#include <GIOPClientWorker.h>

#ifdef WIN32
#   include <sys/timeb.h>
#endif

using namespace CORBA;
using namespace OB;

//
// Next request ID
//
ULong GIOPClientWorker::nextRequestId_ = 0;

//
// Discarded request IDs
//
ULongSeq GIOPClientWorker::discReqIdSeq_;

#ifdef HAVE_JTC
//
// Mutex for request ID handling
//
JTCMutex GIOPClientWorker::requestIdMutex_;
#endif

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::GIOPClientWorker_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPClientWorker_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

void
OBDuplicate(OB::GIOPClientWorkerBlocking_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPClientWorkerBlocking_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

void
OBDuplicate(OB::GIOPClientWorkerReactive_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPClientWorkerReactive_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

#ifdef HAVE_JTC
void
OBDuplicate(OB::GIOPClientWorkerThreaded_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPClientWorkerThreaded_ptr p)
{
    if(p)
	p -> _OB_decRef();
}
#endif

// ----------------------------------------------------------------------
// Template instantiations
// ----------------------------------------------------------------------

#ifndef HAVE_NO_EXPLICIT_TEMPLATES
template class OB::ObjVar< GIOPClientWorker >;
template class OB::ObjForSeq< GIOPClientWorker >;
//template class OB::ObjVar< GIOPClientWorkerBlocking >;
//template class OB::ObjForSeq< GIOPClientWorkerBlocking >;
template class OB::ObjVar< GIOPClientWorkerReactive >;
template class OB::ObjForSeq< GIOPClientWorkerReactive >;
#ifdef HAVE_JTC
//template class OB::ObjVar< GIOPClientWorkerThreaded >;
//template class OB::ObjForSeq< GIOPClientWorkerThreaded >;
#endif
#else
#ifdef HAVE_PRAGMA_DEFINE
#pragma define(OB::ObjVar< GIOPClientWorker >)
#pragma define(OB::ObjForSeq< GIOPClientWorker >)
//#pragma define(OB::ObjVar< GIOPClientWorkerBlocking >)
//#pragma define(OB::ObjForSeq< GIOPClientWorkerBlocking >)
#pragma define(OB::ObjVar< GIOPClientWorkerReactive >)
#pragma define(OB::ObjForSeq< GIOPClientWorkerReactive >)
#ifdef HAVE_JTC
//#pragma define(OB::ObjVar< GIOPClientWorkerThreaded >)
//#pragma define(OB::ObjForSeq< GIOPClientWorkerThreaded >)
#endif
#endif
#endif

// ----------------------------------------------------------------------
// GIOPClientWorker private and protected member implementations
// ----------------------------------------------------------------------

void
GIOPClientWorker::addMessageError()
{
    OCI::Buffer_var buf = new OCI::Buffer(12);
    OutputStream out(buf -> buf(), 0);

    GIOP::MessageHeader_1_0 hdr =
    {
	{ 'G', 'I', 'O', 'P' }, { 1, 0 }, Endian, GIOP::MessageError, 0
    };
    GIOP::MessageHeader_1_0::_OB_marshal(hdr, out);
    add(buf);
}

void
GIOPClientWorker::setupBuffer()
{
    //
    // Create new buffer for incoming message
    //
    buf_ = new OCI::Buffer(12);
}

void
GIOPClientWorker::extractHeader()
{
    const Octet* coct = buf_ -> data();
    
    if(coct[0] != 'G' || coct[1] != 'I' || coct[2] != 'O' || coct[3] != 'P')
    {
	addMessageError();
	throw COMM_FAILURE("", MinorNoGIOP, COMPLETED_MAYBE);
    }
    
    InputStream in(buf_ -> buf(), 0, false);
    GIOP::MessageHeader_1_0::_OB_unmarshal(hdr_, in);
    assert(in._OB_pos() - buf_ -> data() == 12);
    bool swap = hdr_.byte_order != Endian;
    // TODO: Is this really portable?  What about alignment?
    // TODO2: If swap is false, the following isn't necessary
    const Octet* coctSz = (const Octet*)&hdr_.message_size;
    InputStream inSz(coctSz, 0, 4, swap);
    ULong sz = inSz.read_ulong();
    hdr_.message_size = sz;
    buf_ -> realloc(12 + sz);
}

void
GIOPClientWorker::add(OCI::Buffer_ptr buf)
{
    //
    // Add new message to the message buffers
    //
    outSeq_.length(outSeq_.length() + 1);
    outSeq_[outSeq_.length() - 1] = OCI::Buffer::_duplicate(buf);
}

void
GIOPClientWorker::execute()
{
    //
    // Save buffer and all relevant information from the header
    //
    bool swap = hdr_.byte_order != Endian;
    GIOP::MsgType_1_0 type = (GIOP::MsgType_1_0)hdr_.message_type;
    Buffer buf;
    buf.consume(buf_ -> buf());
    buf_ = OCI::Buffer::_nil(); // Reset buf_ to allow for reading
                               // the next message
    
    //
    // Check message type
    //
    if(type > GIOP::MessageError)
    {
	addMessageError();
	throw COMM_FAILURE("", MinorUnknownMessage, COMPLETED_MAYBE);
    }
    
    switch(type)
    {
    case GIOP::Reply:
    {
        InputStream in(buf, 12, swap);
	GIOP::ReplyHeader repH;
	GIOP::ReplyHeader::_OB_unmarshal(repH, in);

	//
	// Check if request has been discarded
	//
	bool discard = false;
	{
#ifdef HAVE_JTC
	    JTCSynchronized synchronized(requestIdMutex_);
#endif
	    for(ULong i = 0 ; i < discReqIdSeq_.length() ; i++)
		if(discReqIdSeq_[i] == repH.request_id)
		{
		    discard = true;
		    discReqIdSeq_.remove(i);
		    break;
		}
	}
	
	if(!discard)
	{
	    //
	    // Save the message
	    //
	    Reply* rep = new Reply;
	    rep -> reqId = repH.request_id;
	    rep -> buf.consume(buf);
	    rep -> swap = swap;
	    rep -> status = repH.reply_status;
	    rep -> offset = in._OB_pos() - rep -> buf.data;
	    rep -> next = head_;
	    head_ = rep;
	}
    }
    
    break;
    
    case GIOP::Request:
    case GIOP::LocateRequest:
    case GIOP::CancelRequest:
	
	//
	// These messages may not be sent to a client
	//
	addMessageError();
	throw COMM_FAILURE("", MinorWrongMessage, COMPLETED_MAYBE);
	
    case GIOP::LocateReply:
	
	//
	// Since I am not sending LocateRequest messages,
	// LocateReply may not be sent to this client
	//
	addMessageError();
	throw COMM_FAILURE("", MinorWrongMessage, COMPLETED_MAYBE);
	
    case GIOP::CloseConnection:
	
	//
	// Notify connection closing by throwing a TRANSIENT exception
	//
	throw TRANSIENT();
	
    case GIOP::MessageError:
	
	//
	// How are message error messages handled? I just send an
	// error message back.
	//
	addMessageError();
	throw COMM_FAILURE("", MinorMessageError, COMPLETED_MAYBE);
    }
}

ULong
GIOPClientWorker::findReply(ULong reqId, Buffer& buf, bool& swap,
			    GIOP::ReplyStatusType& status)
{
    //
    // Check buffered reply messages
    //
    Reply** p = &head_;
    while(*p)
    {
	Reply* rep = *p;

	if(rep -> reqId == reqId)
	{
	    //
	    // Reply found
	    //
	    buf.consume(rep -> buf);
	    swap = rep -> swap;
	    status = rep -> status;
	    ULong offset = rep -> offset;

	    //
	    // Remove reply from buffer
	    //
	    *p = rep -> next;
	    delete rep;
	    
	    return offset;
	}
	else
	{
	    p = &(rep -> next);
	}
    }
    
    //
    // Nothing found in buffer
    //
    return 0;
}

Boolean
GIOPClientWorker::checkReply(ULong reqId)
{
    //
    // Check buffered reply messages
    //
    Reply* rep = head_;
    while(rep)
    {
	if(rep -> reqId == reqId)
	    return true;

	rep = rep -> next;
    }
    
    //
    // Nothing found in buffer
    //
    return false;
}

// ----------------------------------------------------------------------
// GIOPClientWorker constructor and destructor
// ----------------------------------------------------------------------

GIOPClientWorker::GIOPClientWorker(GIOPClient_ptr client,
				   OCI::Transport_ptr transport)
    : client_(GIOPClient::_duplicate(client)),
      transport_(OCI::Transport::_duplicate(transport)), head_(0)
{
    //cout << "GIOPClientWorker" << endl;
}

GIOPClientWorker::~GIOPClientWorker()
{
    //cout << "~GIOPClientWorker" << endl;

    //
    // Remove all buffered reply messages
    //
    while(head_)
    {
	Reply* next = head_ -> next;
	delete head_;
	head_ = next;
    }
}

// ----------------------------------------------------------------------
// GIOPClientWorker public member implementations
// ----------------------------------------------------------------------

void
GIOPClientWorker::destroy()
{
    //
    // Try to send all unsent outgoing messages
    //
    for(ULong i = 0 ; i < outSeq_.length() ; i++)
    {
	while(!outSeq_[i] -> is_full())
	{
	    ULong oldPos = outSeq_[i] -> pos();
	    
	    //
	    // Send buffer, non-blocking
	    //
	    try
	    {
          	OB_DEC_TRY_FIX
		transport_ -> send(outSeq_[i], false);
	    }
	    catch(const SystemException&)
	    {
	    }

	    //
            // Don't allow any delay if messages cannot be send
	    //
	    if(outSeq_[i] -> pos() == oldPos)
	    {
		i = outSeq_.length();
		break;
	    }
	}
    }

    //
    // Close the transport
    //
    transport_ -> close();
}

OCI::Transport_ptr
GIOPClientWorker::transport()
{
    return OCI::Transport::_duplicate(transport_);
}

// ----------------------------------------------------------------------
// GIOPClientWorkerBlocking constructor and destructor
// ----------------------------------------------------------------------

GIOPClientWorkerBlocking::GIOPClientWorkerBlocking(GIOPClient_ptr client,
						   OCI::Transport_ptr
						   transport)
    : GIOPClientWorker(client, transport)
{
}

GIOPClientWorkerBlocking::~GIOPClientWorkerBlocking()
{
}

// ----------------------------------------------------------------------
// GIOPClientWorkerBlocking public member implementations
// ----------------------------------------------------------------------

ULong
GIOPClientWorkerBlocking::sendReceiveTimeout(ULong reqId,
					     Buffer& buf, bool& swap,
					     GIOP::ReplyStatusType& status,
					     Long t)
{
    OB_SYNCHRONIZED

    //
    // Convert simple buffer to OCI buffer
    // Add the OCI buffer
    //
    OCI::Buffer_var b = new OCI::Buffer;
    b -> buf().consume(buf);
    add(b);

    //
    // Try to send all messages
    //
    try
    {
	if(t < 0)
	{
	    for(ULong i = 0 ; i < outSeq_.length() ; i++)
	    {
		//
		// Send buffer, blocking
		//
		OCI::Buffer_ptr buf = outSeq_[i];
		transport_ -> send(buf, true);
		assert(buf -> is_full());
	    }

	    //
	    // Remove all messages
	    //
	    outSeq_.length(0);
	}
	else
	{
	    //
	    // Get current time
	    //
#ifdef WIN32
	    struct _timeb tb1;
	    _ftime(&tb1);
#else
	    timeval tv1;
	    gettimeofday(&tv1, 0);
#endif
	    Long millis = 0;

	    while(outSeq_.length() > 0)
	    {
		//
		// Check if time limit is exceeded
		//
		if(millis > t)
		    return 0; // Timeout

		//
		// Send buffer, with timeout
		//
		OCI::Buffer_ptr buf = outSeq_[0];
		ULong oldPos = buf -> pos();
		transport_ -> send_timeout(buf, (ULong)t);
		if(oldPos == buf -> pos())
		    return 0; // Timeout

		if(buf -> is_full())
		{
		    //
		    // Message was sent completely, remove it
		    //
		    for(ULong i = 1 ; i < outSeq_.length() ; i++)
			outSeq_[i - 1] = outSeq_[i];
		    outSeq_.length(outSeq_.length() - 1);
		}

		//
		// Calculate time needed
		//
#ifdef WIN32
		struct _timeb tb2;
		_ftime(&tb2);
		millis = (tb2.time - tb1.time) * 1000;
		millis += tb2.millitm;
		millis -= tb1.millitm;
#else
		timeval tv2;
		gettimeofday(&tv2, 0);
		millis = (tv2.tv_sec - tv1.tv_sec) * 1000;
		millis += tv2.tv_usec / 1000;
		millis -= tv1.tv_usec / 1000;
#endif
	    }

	    //
	    // Check if time limit is exceeded
	    //
	    if(millis > t)
		return 0; // Timeout

	    //
	    // Subtract time needed before receiving reply
	    //
	    t -= millis;
	}
    }
    catch(const SystemException&)
    {
	//
	// Restore the buffer
	//
	buf.consume(b -> buf());

	//
	// Destroy the worker and rethrow
	//
	client_ -> destroyWorker(this);
	throw;
    }
    
    //
    // Return the reply
    //
    try
    {
	return receiveTimeout(reqId, buf, swap, status, t);
    }
    catch(const SystemException&)
    {
	//
	// Restore the buffer and rethrow
	//
	buf.consume(b -> buf());
	throw;
    }
}

ULong
GIOPClientWorkerBlocking::receiveTimeout(ULong reqId, Buffer& buf, bool& swap,
					 GIOP::ReplyStatusType& status,
					 Long t)
{
    OB_SYNCHRONIZED

    ULong offset = 0;

    //
    // Try to receive the reply
    //
    try
    {
	if(t < 0)
	{
	    while((offset = findReply(reqId, buf, swap, status)) == 0)
	    {
		//
		// Setup the incoming message buffer
		//
		assert(is_nil(buf_));
		setupBuffer();

		//
		// Receive header, blocking
		//
		transport_ -> receive(buf_, true);
		assert(buf_ -> is_full());

		//
		// Header is complete
		//
		extractHeader();

		if(!buf_ -> is_full())
		{
		    //
		    // Receive body, blocking
		    //
		    transport_ -> receive(buf_, true);
		    assert(buf_ -> is_full());
		}

		//
		// Execute the message
		//
		execute();
	    }
	}
	else
	{
	    //
	    // Get current time
	    //
#ifdef WIN32
	    struct _timeb tb1;
	    _ftime(&tb1);
#else
	    timeval tv1;
	    gettimeofday(&tv1, 0);
#endif
	    Long millis = 0;

	    while((offset = findReply(reqId, buf, swap, status)) == 0)
	    {
		//
		// Check if time limit is exceeded
		//
		if(millis > t)
		    return 0; // Timeout

		//
		// Setup the incoming message buffer if not already done
		//
		if(is_nil(buf_))
		    setupBuffer();

		if(buf_ -> pos() < 12)
		{
		    //
		    // Receive header, with timeout
		    //
		    ULong oldPos = buf_ -> pos();
		    transport_ -> receive_timeout(buf_, t);
		    if(oldPos == buf_ -> pos())
			return 0; // Timeout

		    //
		    // Header complete?
		    //
		    if(buf_ -> is_full())
			extractHeader();
		}

		if(buf_ -> pos() >= 12)
		{
		    if(!buf_ -> is_full())
		    {
			//
			// Receive body, with timeout
			//
			ULong oldPos = buf_ -> pos();
			transport_ -> receive_timeout(buf_, t);
			if(oldPos == buf_ -> pos())
			    return 0; // Timeout
		    }

		    //
		    // Execute message if buffer is full
		    //
		    if(buf_ -> is_full())
			execute();
		}

		//
		// Calculate time needed
		//
#ifdef WIN32
		struct _timeb tb2;
		_ftime(&tb2);
		millis = (tb2.time - tb1.time) * 1000;
		millis += tb2.millitm;
		millis -= tb1.millitm;
#else
		timeval tv2;
		gettimeofday(&tv2, 0);
		millis = (tv2.tv_sec - tv1.tv_sec) * 1000;
		millis += tv2.tv_usec / 1000;
		millis -= tv1.tv_usec / 1000;
#endif
	    }
	}
    }
    catch(const SystemException&)
    {
	//
	// Destroy the worker and rethrow
	//
	client_ -> destroyWorker(this);
	throw;
    }

    assert(offset);
    return offset;
}

void
GIOPClientWorkerBlocking::sendNonBlocking(Buffer& buf)
{
    OB_SYNCHRONIZED

    //
    // Convert simple buffer to OCI buffer
    // Add the OCI buffer
    //
    OCI::Buffer_var b = new OCI::Buffer;
    b -> buf().consume(buf);
    add(b);

    //
    // Try to send all messages
    //
    try
    {
	while(outSeq_.length() > 0)
	{
	    //
	    // Send buffer, with timeout
	    //
	    OCI::Buffer_ptr buf = outSeq_[0];
	    ULong oldPos = buf -> pos();
	    transport_ -> send(buf, false);
	    if(oldPos == buf -> pos())
		break;

	    if(buf -> is_full())
	    {
		//
		// Message was sent completely, remove it
		//
		for(ULong i = 1 ; i < outSeq_.length() ; i++)
		    outSeq_[i - 1] = outSeq_[i];
		outSeq_.length(outSeq_.length() - 1);
	    }
	}
    }
    catch(const SystemException&)
    {
	//
	// Restore the buffer
	//
	buf.consume(b -> buf());

	//
	// Destroy the worker and rethrow
	//
	client_ -> destroyWorker(this);
	throw;
    }
}

// ---------------------------------------------------------------------
// GIOPClientWorkerReactive private and protected member implementations
// ----------------------------------------------------------------------

void
GIOPClientWorkerReactive::add(OCI::Buffer_ptr buf)
{
    //OB_SYNCHRONIZED
    // Synchronization can be left out for performance reasons, since
    // this operation is only called from operations which are already
    // synchronized.

    if(!destroy_) // Don't add if already destroyed
    {
	GIOPClientWorker::add(buf);
	
	//
	// Since there is now data to write, register as read and write
	// event handler. This is only necessary if this is the first
	// outgoing message, since if there is more then one message in
	// the buffers, the event handler is already registered.
	//
	if(outSeq_.length() == 1)
	{
	    OCI::Handle handle = transport_ -> handle();
	    assert(handle >= 0);
	    Reactor::instance() ->
		registerHandler(this, EventRead | EventWrite, handle);
	}
    }
}

// ----------------------------------------------------------------------
// GIOPClientWorkerReactive constructor and destructor
// ----------------------------------------------------------------------

GIOPClientWorkerReactive::GIOPClientWorkerReactive(GIOPClient_ptr client,
						   OCI::Transport_ptr
						   transport)
    : GIOPClientWorker(client, transport), destroy_(false), exception_(0)
{
    //
    // Register as read event handler
    //
    OCI::Handle handle = transport_ -> handle();
    assert(handle >= 0);
    Reactor::instance() -> registerHandler(this, EventRead, handle);
}

GIOPClientWorkerReactive::~GIOPClientWorkerReactive()
{
    assert(destroy_);
    delete exception_;
}

// ----------------------------------------------------------------------
// GIOPClientWorkerReactive public member implementations
// ----------------------------------------------------------------------

void
GIOPClientWorkerReactive::destroy()
{
    //
    // WARNING: The reactive worker is not thread safe!
    //

    //
    // Don't destroy twice
    //
    if(destroy_)
	return;
    
    //
    // Set the destroy flag
    //
    destroy_ = true;
    
    //
    // Unregister event handler
    //
    Reactor::instance() -> unregisterHandler(this);

    //
    // Call destroy on the super class
    //
    GIOPClientWorker::destroy();
}

ULong
GIOPClientWorkerReactive::sendReceiveTimeout(ULong reqId,
					     Buffer& buf, bool& swap,
					     GIOP::ReplyStatusType& status,
					     Long t)
{
    OB_SYNCHRONIZED

    //
    // Convert simple buffer to OCI buffer
    // Add the OCI buffer
    //
    OCI::Buffer_var b = new OCI::Buffer;
    b -> buf().consume(buf);
    add(b);

    //
    // Return the reply
    //
    try
    {
	return receiveTimeout(reqId, buf, swap, status, t);
    }
    catch(const SystemException&)
    {
	//
	// Restore the buffer and rethrow
	//
	buf.consume(b -> buf());
	throw;
    }
}

ULong
GIOPClientWorkerReactive::receiveTimeout(ULong reqId, Buffer& buf, bool& swap,
					 GIOP::ReplyStatusType& status,
					 Long t)
{
    OB_SYNCHRONIZED

    ULong offset = 0;

    if(t < 0)
    {
	//
	// Wait for reply or exception
	//
	while(!exception_ &&
	      (offset = findReply(reqId, buf, swap, status)) == 0)
	{
	    //
	    // Dispatch one event, using the reactor
	    //
	    if(!Reactor::instance() -> dispatchOneEvent(t))
		assert(false);
	}	
    }
    else
    {
	//
	// Get current time
	//
#ifdef WIN32
	struct _timeb tb1;
	_ftime(&tb1);
#else
	timeval tv1;
	gettimeofday(&tv1, 0);
#endif
	Long millis = 0;
	    
	//
	// Wait for reply or exception
	//
	while(!exception_ &&
	      (offset = findReply(reqId, buf, swap, status)) == 0)
	{
	    //
	    // Check if time limit is exceeded
	    //
	    if(millis > t)
		return 0; // Timeout

	    //
	    // Dispatch one event, using the reactor
	    //
	    if(!Reactor::instance() -> dispatchOneEvent(t))
		return 0; // Timeout

	    //
	    // Calculate time needed
	    //
#ifdef WIN32
	    struct _timeb tb2;
	    _ftime(&tb2);
	    millis = (tb2.time - tb1.time) * 1000;
	    millis += tb2.millitm;
	    millis -= tb1.millitm;
#else
	    timeval tv2;
	    gettimeofday(&tv2, 0);
	    millis = (tv2.tv_sec - tv1.tv_sec) * 1000;
	    millis += tv2.tv_usec / 1000;
	    millis -= tv1.tv_usec / 1000;
#endif
	}
    }
    
    //
    // Check whether there is an exception
    //
    if(exception_)
    {
	//
	// Destroy the worker and raise the exception
	//
	client_ -> destroyWorker(this);
	exception_ -> _raise();
    }

    assert(offset);
    return offset;
}

void
GIOPClientWorkerReactive::sendNonBlocking(Buffer& buf)
{
    OB_SYNCHRONIZED

    //
    // Convert simple buffer to OCI buffer
    // Add the OCI buffer
    //
    OCI::Buffer_var b = new OCI::Buffer;
    b -> buf().consume(buf);
    add(b);

    //
    // Dispatch one event, using the reactor, giving short message a
    // chance to be deliviered immediately
    //
    Reactor::instance() -> dispatchOneEvent(0);

    //
    // Check whether there is an exception
    //
    if(exception_)
    {
	//
	// Restore the buffer
	//
	buf.consume(b -> buf());

	//
	// Destroy the worker and raise the exception
	//
	client_ -> destroyWorker(this);
	exception_ -> _raise();
    }

    //
    // Nothing else to do here
    //
}

void
GIOPClientWorkerReactive::handleEvent(Mask mask)
{
    //
    // Make sure that this worker cannot be released as long as this
    // operation is executed. Note: This must go before the
    // synchronization!
    //
    GIOPClientWorkerReactive_var worker =
	GIOPClientWorkerReactive::_duplicate(this);

    OB_SYNCHRONIZED

    //
    // Handle read events
    //
    if(mask & EventRead)
    {
	try
	{
	    //
	    // Setup the incoming message buffer if not already done
	    //
	    if(is_nil(buf_))
		setupBuffer();
	    
	    if(buf_ -> pos() < 12)
	    {
		//
		// Receive header, non-blocking
		//
		transport_ -> receive(buf_, false);
		
		//
		// Header complete?
		//
		if(buf_ -> is_full())
		    extractHeader();
	    }
	    
	    if(buf_ -> pos() >= 12)
	    {
		if(!buf_ -> is_full())
		{
		    //
		    // Receive body, non-blocking
		    //
		    transport_ -> receive(buf_, false);
		}
		
		//
		// Execute message if buffer is full
		//
		if(buf_ -> is_full())
		    execute();
	    }
	}
	catch(const SystemException& ex)
	{
	    //
	    // Buffer the exception if none is set
	    //
	    if(!exception_)
	    {
		exception_ = SystemException::_downcast(ex._OB_clone());
		assert(exception_);
	    }

	    //
	    // The worker may *not* be removed from the client,
	    // otherwise the client will not detect the communication
	    // failure.
	    //
	    //client_ -> destroyWorker(this);
	    destroy();

	    return;
	}
    }

    //
    // Handle write events
    //
    if((mask & EventWrite) && outSeq_.length() > 0)
    {
	//
	// Send this message
	//
	try
	{
	    //
	    // Send buffer, non-blocking
	    //
	    OCI::Buffer_ptr buf = outSeq_[0];
	    transport_ -> send(buf, false);
	    
	    if(buf -> is_full())
	    {
		//
		// Message was sent completely, remove it
		//
		for(ULong i = 1 ; i < outSeq_.length() ; i++)
		    outSeq_[i - 1] = outSeq_[i];
		outSeq_.length(outSeq_.length() - 1);
		
		if(outSeq_.length() == 0)
		{
		    //
		    // Register as read event handler only, since
		    // there is nothing to write anymore
		    //
		    OCI::Handle handle = transport_ -> handle();
		    assert(handle >= 0);
		    Reactor::instance()
			-> registerHandler(this, EventRead, handle);
		}
	    }
	}
	catch(const SystemException& ex)
	{
	    //
	    // Buffer the exception if none is set
	    //
	    if(!exception_)
	    {
		exception_ = SystemException::_downcast(ex._OB_clone());
		assert(exception_);
	    }

	    //
	    // The worker may *not* be removed from the client,
	    // otherwise the client will not detect the communication
	    // failure.
	    //
	    //client_ -> destroyWorker(this);
	    destroy();

	    return;
	}
    }
}

#ifdef HAVE_JTC

// ----------------------------------------------------------------------
// GIOPClientWorkerThreaded private and protected member implementations
// ----------------------------------------------------------------------

void
GIOPClientWorkerThreaded::add(OCI::Buffer_ptr buf)
{
    //JTCSynchronized synchronized(*this);
    // Synchronization can be left out for performance reasons, since
    // this operation is only called from operations which are already
    // synchronized.

    if(!destroy_) // Don't add if already destroyed
    {
	GIOPClientWorker::add(buf);
	
	//
	// Notify the sender thread. This is only necessary if this is
	// the first outgoing message, since if there is more then one
	// message in the buffers, threads are already notified.
	//
	if(outSeq_.length() == 1)
	    notifyAll();
    }
}

OCI::Buffer_ptr
GIOPClientWorkerThreaded::remove()
{
    JTCSynchronized synchronized(*this);

    //
    // Wait for a message
    //
    while(!destroy_ && outSeq_.length() == 0)
    {
	try
	{
	    wait();
	}
	catch(const JTCInterruptedException&)
	{
	}
    }

    //
    // Return nil buffer if destroy_ was set
    //
    if(destroy_)
	return OCI::Buffer::_nil();

    //
    // Remove and return first message from message buffers
    //
    assert(outSeq_.length() > 0);
    OCI::Buffer_var buf = outSeq_[0];
    
    for(ULong i = 1 ; i < outSeq_.length() ; i++)
	outSeq_[i - 1] = outSeq_[i];
    outSeq_.length(outSeq_.length() - 1);
    
    return buf._retn();
}

// ----------------------------------------------------------------------
// GIOPClientWorkerThreaded constructor and destructor
// ----------------------------------------------------------------------

GIOPClientWorkerThreaded::GIOPClientWorkerThreaded(GIOPClient_ptr client,
						   OCI::Transport_ptr
						   transport)
    : GIOPClientWorker(client, transport), destroy_(false), exception_(0)
{
    //
    // Start sender and receiver thread
    //
    senderThread_ = new SenderThread(this);
    senderThread_ -> start();
    receiverThread_ = new ReceiverThread(this);
    receiverThread_ -> start();
}

GIOPClientWorkerThreaded::~GIOPClientWorkerThreaded()
{
    assert(destroy_);
    delete exception_;
}

// ----------------------------------------------------------------------
// GIOPClientWorkerThreaded public member implementations
// ----------------------------------------------------------------------

void
GIOPClientWorkerThreaded::destroy()
{
    JTCSynchronized synchronized(*this);

    //
    // Don't destroy twice
    //
    if(destroy_)
	return;
    
    //
    // Set the destroy flag
    //
    destroy_ = true;
    
    //
    // Notify the sender and the receiver thread
    //
    notifyAll();
    
    //
    // Shutdown the transport
    // TODO: Try to send all unsent outgoing messages first
    //
    transport_ -> shutdown();
    
    //
    // Wait for the receiver thread to terminate
    //
    if(receiverThread_ &&
       receiverThread_ -> getId() != JTCThread::currentThread() -> getId())
    {
	while(receiverThread_)
	{
	    try
	    {
		wait();
	    }
	    catch(const JTCInterruptedException&)
	    {
	    }
	}
    }
    
    //
    // Wait for the sender thread to terminate
    //
    if(senderThread_ &&
       senderThread_ -> getId() != JTCThread::currentThread() -> getId())
    {
	while(senderThread_)
	{
	    try
	    {
		wait();
	    }
	    catch(const JTCInterruptedException&)
	    {
	    }
	}
    }

    //
    // Call destroy on the super class
    //
    GIOPClientWorker::destroy();
}

ULong
GIOPClientWorkerThreaded::sendReceiveTimeout(ULong reqId,
					     Buffer& buf, bool& swap,
					     GIOP::ReplyStatusType& status,
					     Long t)
{
    JTCSynchronized synchronized(*this);

    //
    // Convert simple buffer to OCI buffer
    // Add the OCI buffer
    //
    OCI::Buffer_var b = new OCI::Buffer;
    b -> buf().consume(buf);
    add(b);

    //
    // Return the reply
    //
    try
    {
	return receiveTimeout(reqId, buf, swap, status, t);
    }
    catch(const SystemException&)
    {
	//
	// Restore the buffer and rethrow
	//
	buf.consume(b -> buf());
	throw;
    }
}

ULong
GIOPClientWorkerThreaded::receiveTimeout(ULong reqId, Buffer& buf,
					 bool& swap,
					 GIOP::ReplyStatusType& status,
					 Long t)
{
    //
    // Yield if zero timeout
    //
    if(t == 0)
	JTCThread::yield();

    JTCSynchronized synchronized(*this);
    {
	ULong offset = 0;

	if(t < 0)
	{
	    //
	    // Wait for reply or exception, without timeout
	    //
	    while(!exception_ &&
		  (offset = findReply(reqId, buf, swap, status)) == 0)
	    {
		try
		{
		    wait();
		}
		catch(const JTCInterruptedException&)
		{
		}
	    }
	}
	else if(t == 0)
	{
	    //
	    // Zero timeout, don't wait for reply
	    //
	    if(!exception_)
		offset = findReply(reqId, buf, swap, status);

	    if(offset == 0)
		return 0; // Timeout
	}
	else
	{
	    //
	    // Get current time
	    //
#ifdef WIN32
	    struct _timeb tb1;
	    _ftime(&tb1);
#else
	    timeval tv1;
	    gettimeofday(&tv1, 0);
#endif
	    Long millis = 0;

	    //
	    // Wait for reply or exception, with timeout
	    //
	    while(!exception_ &&
		  (offset = findReply(reqId, buf, swap, status)) == 0)
	    {
		//
		// Check if time limit is exceeded
		//
		if(millis > t)
		    return 0; // Timeout

		try
		{
		    wait(t);
		}
		catch(const JTCInterruptedException&)
		{
		}

		//
		// Calculate time needed
		//
#ifdef WIN32
		struct _timeb tb2;
		_ftime(&tb2);
		millis = (tb2.time - tb1.time) * 1000;
		millis += tb2.millitm;
		millis -= tb1.millitm;
#else
		timeval tv2;
		gettimeofday(&tv2, 0);
		millis = (tv2.tv_sec - tv1.tv_sec) * 1000;
		millis += tv2.tv_usec / 1000;
		millis -= tv1.tv_usec / 1000;
#endif
	    }
	}

	//
	// Check whether there is an exception
	//
	if(exception_)
	{
	    //
	    // Destroy the worker and raise the exception
	    //
	    client_ -> destroyWorker(this);
	    exception_ -> _raise();
	}

	assert(offset);
	return offset;
    }
}

void
GIOPClientWorkerThreaded::sendNonBlocking(Buffer& buf)
{
    JTCSynchronized synchronized(*this);

    //
    // Convert simple buffer to OCI buffer
    // Add the OCI buffer
    //
    OCI::Buffer_var b = new OCI::Buffer;
    b -> buf().consume(buf);
    add(b);

    //
    // Check whether there is an exception
    //
    if(exception_)
    {
	//
	// Restore the buffer
	//
	buf.consume(b -> buf());

	//
	// Destroy the worker and raise the exception
	//
	client_ -> destroyWorker(this);
	exception_ -> _raise();
    }

    //
    // Nothing else to do here
    //
}

void
GIOPClientWorkerThreaded::senderRun()
{
    while(!destroy_)
    {
	try
	{
	    //
	    // Remove outgoing message from the outgoing message buffers
	    //
	    OCI::Buffer_var buf = remove();

	    //
	    // Check if thread should be stopped and destroyed
	    //
	    if(is_nil(buf))
		return;

	    //
	    // Send buffer, blocking
	    //
	    transport_ -> send(buf, true);
	    assert(buf -> is_full());
	}
	catch(const SystemException& ex)
	{
	    {
		JTCSynchronized synchronized(*this);

		//
		// It is possible that the sender thread has already set
		// an exception
		//
		if(!exception_)
		{
		    exception_ =
			SystemException::_downcast(ex._OB_clone());
		    assert(exception_);
		}
		
		notifyAll();
		
		//
		// The worker may *not* be removed from the client,
		// otherwise the client will not detect the communication
		// failure.
		//
		//client_ -> destroyWorker(this);
		destroy();
	    }

	    return;
	}
    }
}

void
GIOPClientWorkerThreaded::receiverRun()
{
    while(!destroy_)
    {
	try
	{
	    //
	    // Setup the incoming message buffer
	    //
	    assert(is_nil(buf_));
	    setupBuffer();
	    
	    //
	    // Receive header, blocking
	    //
	    transport_ -> receive(buf_, true);
	    assert(buf_ -> is_full());
	    
	    //
	    // Header is complete
	    //
	    extractHeader();
	    
	    if(!buf_ -> is_full())
	    {
		//
		// Receive body, blocking
		//
		transport_ -> receive(buf_, true);
		assert(buf_ -> is_full());
	    }
	    
	    //
	    // Execute the message
	    //
	    {
		JTCSynchronized synchronized(*this);

		if(!destroy_)
		{
		    execute();
		    notifyAll();
		}
	    }
	}
	catch(const SystemException& ex)
	{
	    {
		JTCSynchronized synchronized(*this);

		//
		// It is possible that the sender thread has already set
		// an exception
		//
		if(!exception_)
		{
		    exception_ =
			SystemException::_downcast(ex._OB_clone());
		    assert(exception_);
		}
		
		notifyAll();

		//
		// The worker may *not* be removed from the client,
		// otherwise the client will not detect the communication
		// failure.
		//
		//client_ -> destroyWorker(this);
		destroy();
	    }
	    	    
	    return;
	}
    }
}

// ----------------------------------------------------------------------
// GIOPClientWorkerThreaded::SenderThread public member implementations
// ----------------------------------------------------------------------

void
GIOPClientWorkerThreaded::SenderThread::run()
{
    worker_ -> senderRun();
    
    {
	JTCSynchronized sync(*(worker_.in()));
	worker_ -> senderThread_ = JTCThreadHandle();
	worker_ -> notifyAll();
    }
    
    // Break cyclic object dependency
    worker_ = GIOPClientWorkerThreaded::_nil();
}

// ----------------------------------------------------------------------
// GIOPClientWorkerThreaded::ReceiverThread public member implementations
// ----------------------------------------------------------------------

void
GIOPClientWorkerThreaded::ReceiverThread::run()
{
    worker_ -> receiverRun();
    
    {
	JTCSynchronized sync(*(worker_.in()));
	worker_ -> receiverThread_ = JTCThreadHandle();
	worker_ -> notifyAll();
    }
    
    // Break cyclic object dependency
    worker_ = GIOPClientWorkerThreaded::_nil();
}

#endif
