// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/TemplateI.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/Any.h>
#include <OB/OBV.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/IFR1.h>
#include <OB/Policy.h>
#include <OB/ORB.h>
#include <OB/Server.h>
#include <OB/OCI.h>
#include <OB/GIOP.h>
#include <OB/Reactor.h>

#include <GIOPServer.h>
#include <GIOPServerStarter.h>
#include <Downcall.h>
#include <GIOPServerWorker.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// GIOPServer constructor and destructor
// ----------------------------------------------------------------------

GIOPServer::GIOPServer(OCI::Acceptor_ptr acceptor, Upcall* upcall)
    : acceptor_(OCI::Acceptor::_duplicate(acceptor)),
      starter_(GIOPServerStarter::_nil()),
      enabled_(true),
      upcall_(upcall)
{
    //cout << "GIOPServer" << endl;
}

GIOPServer::~GIOPServer()
{
    //cout << "~GIOPServer" << endl;

    //
    // If starter_ is not nil, destroy() was not called
    //
    assert(is_nil(starter_));
}

// ----------------------------------------------------------------------
// GIOPServer public member implementation
// ----------------------------------------------------------------------

void
GIOPServer::destroy()
{
    if(!is_nil(starter_))
    {
	starter_ -> destroy();
	starter_ = GIOPServerStarter::_nil();
    }
}

OCI::Acceptor_ptr
GIOPServer::acceptor()
{
    return OCI::Acceptor::_duplicate(acceptor_);
}

bool
GIOPServer::run()
{
    bool done = false;

    //XXX OCI ORB CONC
    switch(ORB::server_conc_model())
    {
    case ORB::ServerConcModelBlocking:
    {
	if(!starter_.in())
	    starter_ = new GIOPServerStarterBlocking(acceptor_, enabled_,
                                                     upcall_);
	break;
    }
	
    case ORB::ServerConcModelReactive:
    {
	if(!starter_.in())
	    starter_ = new GIOPServerStarterReactive(acceptor_, enabled_,
                                                     upcall_);
	break;
    }
	
#ifdef HAVE_JTC
    case ORB::ServerConcModelThreaded:
    case ORB::ServerConcModelThreadPerClient:
    case ORB::ServerConcModelThreadPerRequest:
    case ORB::ServerConcModelThreadPool:
    {
	if(!starter_.in())
	    starter_ = new GIOPServerStarterThreaded(acceptor_, enabled_,
                                                     upcall_);
	break;
    }
#endif
    }

    if(enabled_)
	done = starter_ -> startWorker();

    return done;
}

void
GIOPServer::disable()
{
    assert(is_nil(starter_));
    enabled_ = false;
}
