// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_GIOP_SERVER_STARTER_H
#define OB_GIOP_SERVER_STARTER_H

namespace OB
{

class GIOPServerWorker;
typedef GIOPServerWorker* GIOPServerWorker_ptr;

class GIOPServerStarter;
typedef GIOPServerStarter* GIOPServerStarter_ptr;

class Upcall;

class GIOPServerStarterBlocking;
typedef GIOPServerStarterBlocking* GIOPServerStarterBlocking_ptr;

//
// The GIOPServerStarterReactive class
//
class GIOPServerStarterReactive;
typedef GIOPServerStarterReactive* GIOPServerStarterReactive_ptr;

#ifdef HAVE_JTC

//
// The GIOPServerStarterThreaded class
//
class GIOPServerStarterThreaded;
typedef GIOPServerStarterThreaded* GIOPServerStarterThreaded_ptr;

#endif

} // End of namespace OB

void OBDuplicate(OB::GIOPServerWorker_ptr);
void OBRelease(OB::GIOPServerWorker_ptr);
void OBDuplicate(OB::GIOPServerStarter_ptr);
void OBRelease(OB::GIOPServerStarter_ptr);
void OBDuplicate(OB::GIOPServerStarterBlocking_ptr);
void OBRelease(OB::GIOPServerStarterBlocking_ptr);
void OBDuplicate(OB::GIOPServerStarterReactive_ptr);
void OBRelease(OB::GIOPServerStarterReactive_ptr);

#ifdef HAVE_JTC

void OBDuplicate(OB::GIOPServerStarterThreaded_ptr);
void OBRelease(OB::GIOPServerStarterThreaded_ptr);

#endif

namespace OB
{

//
// Declare GIOPServerWorker as a class
//
typedef ObjVar< GIOPServerWorker > GIOPServerWorker_var;
typedef ObjSeq< GIOPServerWorker, int > GIOPServerWorkerSeq;

//
// The GIOPServerStarter class
//
typedef ObjVar< GIOPServerStarter > GIOPServerStarter_var;

class GIOPServerStarter : public RefCount
{
    //
    // Hide copy-constructor and assignment operator
    //
    GIOPServerStarter(const GIOPServerStarter&);
    void operator=(const GIOPServerStarter&);

protected:

    //
    // The acceptor
    //
    OCI::Acceptor_var acceptor_;

    //
    // The workers
    //
    GIOPServerWorkerSeq workers_;
#ifdef HAVE_JTC
    JTCMutex workersMutex_;
#endif

    //
    // The upcall point
    //
    Upcall* upcall_;

public:
    
    GIOPServerStarter(OCI::Acceptor_ptr, bool, Upcall*);
    virtual ~GIOPServerStarter();

    static inline GIOPServerStarter_ptr _duplicate(GIOPServerStarter_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline GIOPServerStarter_ptr _nil()
    { return 0; }

    //
    // Destroy the server starter
    //
    virtual void destroy();

    //
    // Destroy a worker
    //
    virtual void destroyWorker(GIOPServerWorker_ptr);

    //
    // Start the worker
    //
    virtual bool startWorker();
};

//
// The GIOPServerStarterBlocking class
//
typedef ObjVar< GIOPServerStarterBlocking > GIOPServerStarterBlocking_var;

class GIOPServerStarterBlocking : public GIOPServerStarter
{
    //
    // Hide copy-constructor and assignment operator
    //
    GIOPServerStarterBlocking(const GIOPServerStarterBlocking&);
    void operator=(const GIOPServerStarterBlocking&);

public:
    
    GIOPServerStarterBlocking(OCI::Acceptor_ptr, bool, Upcall*);
    virtual ~GIOPServerStarterBlocking();

    static inline GIOPServerStarterBlocking_ptr
    _duplicate(GIOPServerStarterBlocking_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline GIOPServerStarterBlocking_ptr _nil()
    { return 0; }

    //
    // Start the worker
    //
    virtual bool startWorker();
};

//
// The GIOPServerStarterReactive class
//
typedef ObjVar< GIOPServerStarterReactive > GIOPServerStarterReactive_var;

class GIOPServerStarterReactive : public GIOPServerStarter,
				  public EventHandler
{
    //
    // Hide copy-constructor and assignment operator
    //
    GIOPServerStarterReactive(const GIOPServerStarterReactive&);
    void operator=(const GIOPServerStarterReactive&);

protected:

    bool destroy_; // True if destroy() was called

public:
    
    GIOPServerStarterReactive(OCI::Acceptor_ptr, bool, Upcall*);
    virtual ~GIOPServerStarterReactive();

    static inline GIOPServerStarterReactive_ptr
    _duplicate(GIOPServerStarterReactive_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline GIOPServerStarterReactive_ptr _nil()
    { return 0; }

    //
    // Destroy the worker
    //
    virtual void destroy();

    //
    // Handle event
    //
    virtual void handleEvent(Mask);
};

#ifdef HAVE_JTC

//
// The GIOPServerStarterThreaded class
//
typedef ObjVar< GIOPServerStarterThreaded > GIOPServerStarterThreaded_var;

class GIOPServerStarterThreaded : public GIOPServerStarter
{
    //
    // Hide copy-constructor and assignment operator
    //
    GIOPServerStarterThreaded(const GIOPServerStarterThreaded&);
    void operator=(const GIOPServerStarterThreaded&);

protected:

    bool destroy_; // True if destroy() was called

    //
    // The starter thread
    //
    class StarterThread : public JTCThread
    {
	GIOPServerStarterThreaded_var starter_;

    public:
	
	StarterThread(GIOPServerStarterThreaded_ptr starter)
	    : starter_(GIOPServerStarterThreaded::_duplicate(starter)) { }
	virtual ~StarterThread() { }
	virtual void run();
    };
    friend class StarterThread;
    JTCThreadHandle starterThread_;

public:
    
    GIOPServerStarterThreaded(OCI::Acceptor_ptr, bool, Upcall*);
    virtual ~GIOPServerStarterThreaded();

    static inline GIOPServerStarterThreaded_ptr
    _duplicate(GIOPServerStarterThreaded_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline GIOPServerStarterThreaded_ptr _nil()
    { return 0; }

    //
    // Destroy the server starter
    //
    virtual void destroy();

    //
    // Run method for starter thread
    //
    void starterRun();
};

#endif

} // End of namespace OB


namespace CORBA
{
// XXX Are these necessary in CORBA?

inline void
release(OB::GIOPServerStarter_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::GIOPServerStarter_ptr p)
{
    return p == 0;
}

inline void
release(OB::GIOPServerStarterBlocking_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::GIOPServerStarterBlocking_ptr p)
{
    return p == 0;
}

inline void
release(OB::GIOPServerStarterReactive_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::GIOPServerStarterReactive_ptr p)
{
    return p == 0;
}

#ifdef HAVE_JTC

inline void
release(OB::GIOPServerStarterThreaded_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::GIOPServerStarterThreaded_ptr p)
{
    return p == 0;
}

#endif

} // End of namespace CORBA

#endif
