// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/TemplateI.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/Any.h>
#include <OB/OBV.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/IFR1.h>
#include <OB/IFR4.h>
#include <OB/Policy.h>
#include <OB/ORB.h>
#include <OB/GIOP.h>
#include <OB/Reactor.h>
#include <OB/Util.h> // For the message viewer

#include <OB/POAServantBase.h>
#include <OB/ValueBase.h>
#include <OB/Stream.h>
#include <OB/OCI.h>

#include <OCI_impl.h> // For OCI_Current_impl::_OB_...()

#include <GIOPServerStarter.h>
#include <Downcall.h>
#include <GIOPServerWorker.h>

#include <Upcall.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::Downcall_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::Downcall_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

void
OBDuplicate(OB::GIOPServerWorker_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPServerWorker_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

void
OBDuplicate(OB::GIOPServerWorkerBlocking_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPServerWorkerBlocking_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

void
OBDuplicate(OB::GIOPServerWorkerReactive_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPServerWorkerReactive_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

#ifdef HAVE_JTC
void
OBDuplicate(OB::GIOPServerWorkerThreaded_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPServerWorkerThreaded_ptr p)
{
    if(p)
	p -> _OB_decRef();
}
#endif

// ----------------------------------------------------------------------
// Template instantiations
// ----------------------------------------------------------------------

#ifndef HAVE_NO_EXPLICIT_TEMPLATES
template class OB::ObjVar< Downcall >;

template class OB::ObjVar< GIOPServerWorker >;
template class OB::ObjForSeq< GIOPServerWorker >;
template class OB::ObjSeq< GIOPServerWorker, int >;
//template class OB::ObjVar< GIOPServerWorkerBlocking >;
//template class OB::ObjForSeq< GIOPServerWorkerBlocking >;
template class OB::ObjVar< GIOPServerWorkerReactive >;
template class OB::ObjForSeq< GIOPServerWorkerReactive >;
#ifdef HAVE_JTC
//template class OB::ObjVar< GIOPServerWorkerThreaded >;
//template class OB::ObjForSeq< GIOPServerWorkerThreaded >;
#endif
#else
#ifdef HAVE_PRAGMA_DEFINE
#pragma define(OB::ObjVar< Downcall >)
#pragma define(OB::ObjVar< GIOPServerWorker >)
#pragma define(OB::ObjForSeq< GIOPServerWorker >)
#pragma define(OB::ObjSeq< GIOPServerWorker, int >)
//#pragma define(OB::ObjVar< GIOPServerWorkerBlocking >)
//#pragma define(OB::ObjForSeq< GIOPServerWorkerBlocking >)
#pragma define(OB::ObjVar< GIOPServerWorkerReactive >)
#pragma define(OB::ObjForSeq< GIOPServerWorkerReactive >)
#ifdef HAVE_JTC
//#pragma define(OB::ObjVar< GIOPServerWorkerThreaded >)
//#pragma define(OB::ObjForSeq< GIOPServerWorkerThreaded >)
#endif
#endif
#endif

// ----------------------------------------------------------------------
// GIOPServerWorker private and protected member implementations
// ----------------------------------------------------------------------

void
GIOPServerWorker::addMessageError()
{
    OCI::Buffer_var buf = new OCI::Buffer(12);
    OutputStream out(buf -> buf(), 0);

    GIOP::MessageHeader_1_0 hdr =
    {
	{ 'G', 'I', 'O', 'P' }, { 1, 0 },
        Endian, GIOP::MessageError, 0
    };
    GIOP::MessageHeader_1_0::_OB_marshal(hdr, out);
    add(buf);
}

void
GIOPServerWorker::addCloseConnection()
{
    OCI::Buffer_var buf = new OCI::Buffer(12);
    OutputStream out(buf -> buf(), 0);

    GIOP::MessageHeader_1_0 hdr =
    {
	{ 'G', 'I', 'O', 'P' }, { 1, 0 },
	Endian, GIOP::CloseConnection, 0
    };
    GIOP::MessageHeader_1_0::_OB_marshal(hdr, out);
    add(buf);
}

void
GIOPServerWorker::setupBuffer()
{
    //
    // Create new buffer for incoming message
    //
    buf_ = new OCI::Buffer(12);
}

void
GIOPServerWorker::extractHeader()
{
    const Octet* coct = buf_ -> data();
    
    if(coct[0] != 'G' || coct[1] != 'I' || coct[2] != 'O' || coct[3] != 'P')
    {
	addMessageError();
	throw COMM_FAILURE("", MinorNoGIOP, COMPLETED_NO);
    }
    
    InputStream in(buf_ -> buf(), 0, false);
    GIOP::MessageHeader_1_0::_OB_unmarshal(hdr_, in);
    assert(in._OB_pos() - buf_ -> data() == 12);
    bool swap = hdr_.byte_order != Endian;
    // TODO: Is this really portable?  What about alignment?
    // TODO2: If swap is false, the following isn't necessary
    const Octet* coctSz = (const Octet*)&hdr_.message_size;
    InputStream inSz(coctSz, 0, 4, swap);
    ULong sz = inSz.read_ulong();
    hdr_.message_size = sz;
    buf_ -> realloc(12 + sz);
}

void
GIOPServerWorker::add(OCI::Buffer_ptr buf)
{
    //
    // Add new message to the message buffers
    //
    outSeq_.length(outSeq_.length() + 1);
    outSeq_[outSeq_.length() - 1] = OCI::Buffer::_duplicate(buf);
}

void
GIOPServerWorker::execute()
{
    //
    // Save buffer and all relevant information from the header
    //
    bool swap = hdr_.byte_order != Endian;
    GIOP::MsgType_1_0 type = (GIOP::MsgType_1_0)hdr_.message_type;
    Buffer buf;
    buf.consume(buf_ -> buf());
    buf_ = OCI::Buffer::_nil(); // Reset buf_ to allow for reading
                                // the next message
    
    //
    // Check message type
    //
    if(type > GIOP::MessageError)
    {
	addMessageError();
	throw COMM_FAILURE("", MinorUnknownMessage, COMPLETED_NO);
    }
    
    switch(type)
    {
    case GIOP::Request:

	dispatchRequest(buf, swap, transport_);
	break;
    
    case GIOP::CancelRequest:
	
	//
	// Do nothing here (doing nothing is GIOP compliant)
	//
	break;
	
    case GIOP::LocateRequest:
    {
	//
	// Get the locate request header
	//
	GIOP::LocateRequestHeader locReqH;
        InputStream in(buf, 12, swap);
	GIOP::LocateRequestHeader::_OB_unmarshal(locReqH, in);
	
	//
	// Start creating locate reply message
	//
	GIOP::LocateReplyHeader locRepH;
	locRepH.request_id = locReqH.request_id;
        CountStream offsetOut(12);
	GIOP::LocateReplyHeader::_OB_marshalCount(locRepH, offsetOut);
	
	if(upcall_ -> findByKey(locReqH.object_key))
	{
	    ULong count = offsetOut.count();
	    /* No body to marshal here ... */
	    
	    buf.alloc(count);
	    locRepH.locate_status = GIOP::OBJECT_HERE;
	}
	else
	{
	    ULong count = offsetOut.count();
	    /* No body to marshal here ... */
	    
	    buf.alloc(count);
	    locRepH.locate_status = GIOP::UNKNOWN_OBJECT;
	}
	
	//
	// Create locate reply message
	//
	GIOP::MessageHeader_1_0 hdr =
	{
	    { 'G', 'I', 'O', 'P' }, { 1, 0 },
	    Endian, GIOP::LocateReply, buf.len - 12
	};
	
        OutputStream out(buf, 0);
	GIOP::MessageHeader_1_0::_OB_marshal(hdr, out);
	GIOP::LocateReplyHeader::_OB_marshal(locRepH, out);
	
	//
	// Convert simple buffer to OCI buffer
	// Add the OCI buffer
	//
	OCI::Buffer_var b = new OCI::Buffer;
	b -> buf().consume(buf);
	add(b);
	break;
    }
    
    case GIOP::Reply:
    case GIOP::LocateReply:
    case GIOP::CloseConnection:
	
	//
	// These messages may not be sent to a server
	//
	addMessageError();
	throw COMM_FAILURE("", MinorWrongMessage, COMPLETED_NO);
	
    case GIOP::MessageError:
	
	//
	// How are message error messages handled? I just send an
	// error message back.
	//
	addMessageError();
	throw COMM_FAILURE("", MinorMessageError, COMPLETED_NO);
    }
}
	
void
GIOPServerWorker::dispatchRequest(Buffer& buf, bool swap,
				  OCI::Transport_ptr transport)
{
    //
    // Get the request header
    //
    GIOP::RequestHeader_1_1 reqH;
    InputStream in(buf, 12, swap);
    GIOP::RequestHeader_1_1::_OB_unmarshal(reqH, in);
    ULong offsetIn = in._OB_pos() - buf.data;
    
    //
    // Start creating reply message
    //
    CountStream offsetOut(12);
    if(reqH.response_expected)
    {
        //
        // XXX - as an optimization, we can do this once
        //
        GIOP::ReplyHeader repH;
        GIOP::ReplyHeader::_OB_marshalCount(repH, offsetOut);
    }
    
    //cout << "Invoking: " << reqH.request_id << endl;

    OCI::TransportInfo_var info = transport -> get_info();
    upcall_ -> invoke((reqH.response_expected) ? this : Downcall::_nil(),
                      reqH.request_id,
                      reqH.object_key, reqH.operation, buf, swap,
                      offsetIn, offsetOut.count(), info);
}

void
GIOPServerWorker::composeReply(GIOP::ReplyHeader& repH, Buffer& buf)
{
    GIOP::MessageHeader_1_0 hdr =
    {
        { 'G', 'I', 'O', 'P' }, { 1, 0 },
        Endian, GIOP::Reply, buf.len - 12
    };
    
    OutputStream out(buf, 0);
    GIOP::MessageHeader_1_0::_OB_marshal(hdr, out);
    GIOP::ReplyHeader::_OB_marshal(repH, out);
    
    //
    // Convert simple buffer to OCI buffer
    // Add the OCI buffer
    //
    OCI::Buffer_var b = new OCI::Buffer;
    b -> buf().consume(buf);
    add(b);
}

// ----------------------------------------------------------------------
// GIOPServerWorker constructor and destructor
// ----------------------------------------------------------------------

GIOPServerWorker::GIOPServerWorker(GIOPServerStarter_ptr starter,
				   OCI::Transport_ptr transport,
				   Upcall* upcall)
    : starter_(GIOPServerStarter::_duplicate(starter)),
      transport_(OCI::Transport::_duplicate(transport)),
      upcall_(upcall)
{
    //cout << "OBGIOPServerWorker" << endl;
}

GIOPServerWorker::~GIOPServerWorker()
{
    //cout << "~OBGIOPServerWorker" << endl;
}

// ----------------------------------------------------------------------
// GIOPServerWorker public member implementations
// ----------------------------------------------------------------------

void
GIOPServerWorker::destroy()
{
    //
    // Add a "close connection" message
    //
    addCloseConnection();

    //
    // Try to send all unsent outgoing messages
    //
    for(ULong i = 0 ; i < outSeq_.length() ; i++)
    {
	while(!outSeq_[i] -> is_full())
	{
	    ULong oldPos = outSeq_[i] -> pos();
	    
	    //
	    // Send buffer, non-blocking
	    //
	    try
	    {
          	OB_DEC_TRY_FIX
		transport_ -> send(outSeq_[i], false);
	    }
	    catch(const SystemException&)
	    {
	    }

	    //
            // Don't allow any delay if messages cannot be send
	    //
	    if(outSeq_[i] -> pos() == oldPos)
	    {
		i = outSeq_.length();
		break;
	    }
	}
    }

    //
    // Close the transport
    //
    transport_ -> close();
}

//
// Regular response
//
void
GIOPServerWorker::response(ULong requestId, Buffer& buf)
{
    //cout << "Response: " << requestId << endl;
    
    GIOP::ReplyHeader repH;
    repH.request_id = requestId;
    repH.reply_status = GIOP::NO_EXCEPTION;

    composeReply(repH, buf);
}

//
// Response is an system exception
//
void
GIOPServerWorker::systemException(ULong requestId, Buffer& buf,
                                  const SystemException& ex)
{
    
    GIOP::ReplyHeader repH;
    repH.request_id = requestId;
    repH.reply_status = GIOP::SYSTEM_EXCEPTION;

    CountStream offsetOut(12);
    GIOP::ReplyHeader::_OB_marshalCount(repH, offsetOut);
    CountStream count(offsetOut.count());
    SystemException::_OB_marshalCount(ex, count);
    
    buf.alloc(count.count());
    OutputStream out(buf, offsetOut.count());
    SystemException::_OB_marshal(ex, out);

    composeReply(repH, buf);
}


//
// Response is a userException
//
void
GIOPServerWorker::userException(ULong requestId, Buffer& buf)
{
    GIOP::ReplyHeader repH;
    repH.request_id = requestId;
    repH.reply_status = GIOP::USER_EXCEPTION;

    composeReply(repH, buf);
}

//
// Response is a location forward
//
void
GIOPServerWorker::forward(ULong requestId, Buffer& buf, Object_ptr obj)
{
    GIOP::ReplyHeader repH;
    repH.request_id = requestId;
    repH.reply_status = GIOP::LOCATION_FORWARD;

    CountStream offsetOut(12);
    GIOP::ReplyHeader::_OB_marshalCount(repH, offsetOut);
    CountStream count(offsetOut.count());
    count.count_Object(obj);
    
    buf.alloc(count.count());
    OutputStream out(buf, offsetOut.count());
    out.write_Object(obj);

    composeReply(repH, buf);
}

// ----------------------------------------------------------------------
// GIOPServerWorkerBlocking constructor and destructor
// ----------------------------------------------------------------------

GIOPServerWorkerBlocking::GIOPServerWorkerBlocking(
    GIOPServerStarterBlocking_ptr starter, OCI::Transport_ptr transport,
    Upcall* upcall)
    : GIOPServerWorker(starter, transport, upcall)
{
}

GIOPServerWorkerBlocking::~GIOPServerWorkerBlocking()
{
}

// ----------------------------------------------------------------------
// GIOPServerWorkerBlocking public member implementations
// ----------------------------------------------------------------------

void
GIOPServerWorkerBlocking::getRequest()
{
    try
    {
	//
	// Setup the incoming message buffer
	//
	assert(is_nil(buf_));
	setupBuffer();

	//
	// Receive header, blocking, detect connection loss
	//
	if(!transport_ -> receive_detect(buf_, true))
	{
	    starter_ -> destroyWorker(this);
	    return; // Orderly shutdown
	}
	assert(buf_ -> is_full());

	//
	// Header is complete
	//
	extractHeader();

	if(!buf_ -> is_full())
	{
	    //
	    // Receive body, blocking, detect connection loss
	    //
	    if(!transport_ -> receive_detect(buf_, true))
	    {
		starter_ -> destroyWorker(this);
		return; // Orderly shutdown
	    }
	    assert(buf_ -> is_full());
	}

	//
	// Execute the message
	//
	execute();
    }
    catch(const SystemException&)
    {
	//
	// Destroy the worker and rethrow
	//
	starter_ -> destroyWorker(this);
	throw;
    }
}

void
GIOPServerWorkerBlocking::sendReply()
{
    try
    {
	//
	// Try to send all messages
	//
	assert(outSeq_.length() <= 1);
	for(ULong i = 0 ; i < outSeq_.length() ; i++)
	{
	    //
	    // Send buffer, blocking, detect connection loss
	    //
	    OCI::Buffer_ptr buf = outSeq_[i];
	    if(!transport_ -> send_detect(buf, true))
	    {
		starter_ -> destroyWorker(this);
		return; // Orderly shutdown
	    }
	    assert(buf -> is_full());
	}

	//
	// Remove all messages
	//
	outSeq_.length(0);
    }
    catch(const SystemException&)
    {
	//
	// Destroy the worker and rethrow
	//
	starter_ -> destroyWorker(this);
	throw;
    }
}

// ----------------------------------------------------------------------
// GIOPServerWorkerReactive private and protected member implementations
// ----------------------------------------------------------------------

void
GIOPServerWorkerReactive::add(OCI::Buffer_ptr buf)
{
    if(!destroy_) // Don't add if already destroyed
    {
	GIOPServerWorker::add(buf);
	
	//
	// Since there is now data to write, register as read and
	// write event handler. This is only necessary if the message
	// added is the first outgoing message in the outgoing message
	// buffer (since if there is more then one message in the
	// buffers, the event handler is already registered).
	//
	if(outSeq_.length() == 1)
	{
	    OCI::Handle handle = transport_ -> handle();
	    if(handle >= 0) // Don't register if already closed
		Reactor::instance() ->
		    registerHandler(this, EventRead | EventWrite, handle);
	}
    }
}

// ----------------------------------------------------------------------
// GIOPServerWorkerReactive constructor and destructor
// ----------------------------------------------------------------------

GIOPServerWorkerReactive::GIOPServerWorkerReactive(
    GIOPServerStarterReactive_ptr starter, OCI::Transport_ptr transport,
    Upcall* upcall)
    : GIOPServerWorker(starter, transport, upcall), destroy_(false)
{
    //
    // Register as read event handler
    //
    OCI::Handle handle = transport_ -> handle();
    assert(handle >= 0);
    Reactor::instance() -> registerHandler(this, EventRead, handle);
}

GIOPServerWorkerReactive::~GIOPServerWorkerReactive()
{
    assert(destroy_);
}

// ----------------------------------------------------------------------
// GIOPServerWorkerReactive public member implementations
// ----------------------------------------------------------------------

void
GIOPServerWorkerReactive::destroy()
{
    //
    // WARNING: The reactive worker is not thread safe!
    //

    //
    // Don't destroy twice
    //
    if(destroy_)
	return;
    
    //
    // Set the destroy flag
    //
    destroy_ = true;
    
    //
    // Unregister event handler
    //
    Reactor::instance() -> unregisterHandler(this);

    //
    // Call destroy on the super class
    //
    GIOPServerWorker::destroy();
}

void
GIOPServerWorkerReactive::handleEvent(Mask mask)
{
    //
    // Make sure that this worker cannot be released as long as this
    // operation is executed
    //
    GIOPServerWorkerReactive_var worker =
	GIOPServerWorkerReactive::_duplicate(this);

    //
    // Handle read events
    //
    if(mask & EventRead)
    {
	try
	{
	    //
	    // Setup the incoming message buffer if not already done
	    //
	    if(is_nil(buf_))
		setupBuffer();
	    
	    if(buf_ -> pos() < 12)
	    {
		//
		// Receive header, non-blocking, detect connection loss
		//
		if(!transport_ -> receive_detect(buf_, false))
		{
		    starter_ -> destroyWorker(this);
		    return; // Orderly shutdown
		}
		
		//
		// Header complete?
		//
		if(buf_ -> is_full())
		    extractHeader();
	    }
	    
	    if(buf_ -> pos() >= 12)
	    {
		if(!buf_ -> is_full())
		{
		    //
		    // Receive body, non-blocking, detect connection loss
		    //
		    if(!transport_ -> receive_detect(buf_, false))
		    {
			starter_ -> destroyWorker(this);
			return; // Orderly shutdown
		    }
		}
	    }
	}
	catch(const SystemException&)
	{
	    //
	    // Destroy the worker and rethrow
	    //
	    starter_ -> destroyWorker(this);
	    throw;
	}
	
	try
	{
	    //
	    // Execute message if buffer is full
	    //
	    if(buf_ -> is_full())
		execute();
	}
	catch(const SystemException&)
	{
	    //
	    // Destroy the worker and rethrow
	    //
	    starter_ -> destroyWorker(this);
	    throw;
	}
    }

    //
    // Handle write events
    //
    if(mask & EventWrite)
    {
	assert(outSeq_.length() > 0);
	
	//
	// Send this message
	//
	try
	{
	    //
	    // Send buffer, non-blocking, detect connection loss
	    //
	    OCI::Buffer_ptr buf = outSeq_[0];
	    if(!transport_ -> send_detect(buf, false))
	    {
		starter_ -> destroyWorker(this);
		return; // Orderly shutdown
	    }
	    
	    if(buf -> is_full())
	    {
		//
		// Message was sent completely, remove it
		//
		for(ULong i = 1 ; i < outSeq_.length() ; i++)
		    outSeq_[i - 1] = outSeq_[i];
		outSeq_.length(outSeq_.length() - 1);
		
		if(outSeq_.length() == 0)
		{
		    //
		    // Register as read event handler only, since
		    // there is nothing to write anymore
		    //
		    OCI::Handle handle = transport_ -> handle();
		    if(handle >= 0) // Don't register if already closed
			Reactor::instance()
			    -> registerHandler(this, EventRead, handle);
		}
	    }
	}
	catch(const SystemException&)
	{
	    //
	    // Destroy the worker and rethrow
	    //
	    starter_ -> destroyWorker(this);
	    throw;
	}
    }
}

#ifdef HAVE_JTC

void
GIOPServerWorkerThreaded::add(OCI::Buffer_ptr buf)
{
    JTCSynchronized synchronized(*this);

    if(!destroy_) // Don't add if already destroyed
    {
	GIOPServerWorker::add(buf);
	
	//
	// Notify the sender thread. This is only necessary if this is
	// the first outgoing message, since if there is more then one
	// message in the buffers, threads are already notified.
	//
	if(outSeq_.length() == 1)
	    notifyAll();
    }
}

OCI::Buffer_ptr
GIOPServerWorkerThreaded::remove()
{
    JTCSynchronized synchronized(*this);

    //
    // Wait for a message
    //
    while(!destroy_ && outSeq_.length() == 0)
    {
	try
	{
	    wait();
	}
	catch(const JTCInterruptedException&)
	{
	}
    }

    //
    // Return nil buffer if destroy_ was set
    //
    if(destroy_)
	return OCI::Buffer::_nil();

    //
    // Remove and return first message from message buffers
    //
    assert(outSeq_.length() > 0);
    OCI::Buffer_var buf = outSeq_[0];
    
    for(ULong i = 1 ; i < outSeq_.length() ; i++)
	outSeq_[i - 1] = outSeq_[i];
    outSeq_.length(outSeq_.length() - 1);
    
    return buf._retn();
}

void
GIOPServerWorkerThreaded::dispatchRequest(Buffer& buf, bool swap,
					  OCI::Transport_ptr transport)
{
    //XXX OCI ORB CONC
    switch(ORB::server_conc_model())
    {
    case ORB::ServerConcModelBlocking:
    case ORB::ServerConcModelReactive:
	assert(false);
	break;

    case ORB::ServerConcModelThreaded:
	//
	// The synchronization comes later
	//

    case ORB::ServerConcModelThreadPerClient:
	GIOPServerWorker::dispatchRequest(buf, swap, transport);
	break;
	
    case ORB::ServerConcModelThreadPerRequest:
    {
	(new Dispatcher(this, buf, swap, transport)) -> start();
	break;
    }

    case ORB::ServerConcModelThreadPool:
    {
	ThreadPool* pool = ThreadPool::instance();
	pool -> add(this, buf, swap, transport);
	break;
    }
    }
}

// ----------------------------------------------------------------------
// GIOPServerWorkerThreaded constructor and destructor
// ----------------------------------------------------------------------

GIOPServerWorkerThreaded::GIOPServerWorkerThreaded(
    GIOPServerStarterThreaded_ptr starter, OCI::Transport_ptr transport,
    Upcall* upcall)
    : GIOPServerWorker(starter, transport, upcall), destroy_(false)
{
    //
    // Start sender and receiver thread
    //
    senderThread_ = new SenderThread(this);
    senderThread_ -> start();
    receiverThread_ = new ReceiverThread(this);
    receiverThread_ -> start();
}

GIOPServerWorkerThreaded::~GIOPServerWorkerThreaded()
{
    assert(destroy_);
}

// ----------------------------------------------------------------------
// GIOPServerWorkerThreaded public member implementations
// ----------------------------------------------------------------------

void
GIOPServerWorkerThreaded::destroy()
{
    JTCSynchronized synchronized(*this);

    //
    // Set the destroy flag
    //
    assert(!destroy_);
    destroy_ = true;
    
    //
    // Notify the sender and the receiver thread
    //
    notifyAll();
    
    //
    // Shutdown the transport
    // TODO: Add a "close connection" message and Try to send all
    // unsent outgoing messages first
    //
    transport_ -> shutdown();
    
    //
    // Wait for the receiver thread to terminate
    //
    if(receiverThread_ &&
       receiverThread_ -> getId() != JTCThread::currentThread() -> getId())
    {
	while(receiverThread_)
	{
	    try
	    {
		wait();
	    }
	    catch(const JTCInterruptedException&)
	    {
	    }
	}
    }
    
    //
    // Wait for the sender thread to terminate
    //
    if(senderThread_ &&
       senderThread_ -> getId() != JTCThread::currentThread() -> getId())
    {
	while(senderThread_)
	{
	    try
	    {
		wait();
	    }
	    catch(const JTCInterruptedException&)
	    {
	    }
	}
    }

    //
    // Call destroy on the super class
    //
    GIOPServerWorker::destroy();
}

void
GIOPServerWorkerThreaded::senderRun()
{
    while(!destroy_)
    {
	try
	{
	    //
	    // Remove outgoing message from the outgoing message buffers
	    //
	    OCI::Buffer_var buf = remove();

	    //
	    // Check if thread should be stopped and destroyed
	    //
	    if(is_nil(buf))
		return;

	    //
	    // Send buffer, blocking, detect connection loss
	    //
	    if(!transport_ -> send_detect(buf, true))
	    {
		//
		// Destroy the worker
		//
	        starter_ -> destroyWorker(this);

		return; // Orderly shutdown
	    }
	    assert(buf -> is_full());
	}
	catch(const SystemException&)
	{
	    //
	    // Destroy the worker
	    //
	    starter_ -> destroyWorker(this);

	    return;
	}
    }
}

void
GIOPServerWorkerThreaded::receiverRun()
{
    while(!destroy_)
    {
	try
	{
	    //
	    // Setup the incoming message buffer
	    //
	    assert(is_nil(buf_));
	    setupBuffer();
	    
	    //
	    // Receive header, blocking, detect connection loss
	    //
	    if(!transport_ -> receive_detect(buf_, true))
	    {
		//
		// Destroy the worker
		//
		starter_ -> destroyWorker(this);

		return; // Orderly shutdown
	    }

            assert(buf_ -> is_full());

	    //
	    // Check if thread should be stopped and destroyed
            //
	    if(destroy_)
		return;

	    //
	    // Header is complete
	    //
	    extractHeader();
	    
	    if(!buf_ -> is_full())
	    {
		//
		// Receive body, blocking, detect connection loss
		//
		if(!transport_ -> receive_detect(buf_, true))
		{
		    //
		    // Destroy the worker
		    //
		    starter_ -> destroyWorker(this);

		    return; // Orderly shutdown
		}
		assert(buf_ -> is_full());
	    }
	    
	    //
	    // Execute the message
	    //
	    execute();
	}
	catch(const SystemException&)
	{
	    //
	    // Destroy the worker
	    //
	    starter_ -> destroyWorker(this);

	    return;
	}
    }
}

// ----------------------------------------------------------------------
// GIOPServerWorkerThreaded::SenderThread public member implementations
// ----------------------------------------------------------------------

void
GIOPServerWorkerThreaded::SenderThread::run()
{
    worker_ -> senderRun();
    
    {
	JTCSynchronized sync(*(worker_.in()));
	worker_ -> senderThread_ = JTCThreadHandle();
	worker_ -> notifyAll();
    }
    
    // Break cyclic object dependency
    worker_ = GIOPServerWorkerThreaded::_nil();
}

// ----------------------------------------------------------------------
// GIOPServerWorkerThreaded::ReceiverThread public member implementations
// ----------------------------------------------------------------------

void
GIOPServerWorkerThreaded::ReceiverThread::run()
{
    worker_ -> receiverRun();
    
    {
	JTCSynchronized sync(*(worker_.in()));
	worker_ -> receiverThread_ = JTCThreadHandle();
	worker_ -> notifyAll();
    }
    
    // Break cyclic object dependency
    worker_ = GIOPServerWorkerThreaded::_nil();
}

// ----------------------------------------------------------------------
// GIOPServerWorkerThreaded::ThreadPool private and protected member
// implementations
// ----------------------------------------------------------------------

void
GIOPServerWorkerThreaded::ThreadPool::get(
    GIOPServerWorkerThreaded::ThreadPool::Request& req)
{
    JTCSynchronized synchronized(*this);
    
    while(!head_)
    {
	try
	{
	    wait();
	}
	catch(const JTCInterruptedException&)
	{
	}
    }
    
    req.worker = head_ -> worker;
    req.buf.consume(head_ -> buf);
    req.swap = head_ -> swap;
    req.transport = OCI::Transport::_duplicate(head_ -> transport);
    req.next = 0;
    
    Request* next = head_ -> next;
    delete head_;
    head_ = next;
}

// ----------------------------------------------------------------------
// GIOPServerWorkerThreaded::ThreadPool constructor and destructor
// ----------------------------------------------------------------------

GIOPServerWorkerThreaded::ThreadPool::ThreadPool()
    : head_(0)
{
    //XXX OCI ORB CONC
    ULong num = ORB::conc_model_thread_pool();
    for(ULong i = 0 ; i < num ; i++)
	(new Dispatcher(this)) -> start();
    //XXX
}

GIOPServerWorkerThreaded::ThreadPool::~ThreadPool()
{
    while(head_)
    {
	Request* next = head_ -> next;
	delete head_;
	head_ = next;
    }
}

// ----------------------------------------------------------------------
// GIOPServerWorkerThreaded::ThreadPool public
// ----------------------------------------------------------------------

GIOPServerWorkerThreaded::ThreadPool*
GIOPServerWorkerThreaded::ThreadPool::instance()
{
    static GIOPServerWorkerThreaded::ThreadPool* instance = 0;
    static JTCMutex mutex;

    if(!instance) // Double-checked locking
    {
        JTCSynchronized synchronized(mutex);
	
        if(!instance)
	    instance = new ThreadPool();
    }
    
    return instance;
}

void
GIOPServerWorkerThreaded::ThreadPool::add(
    GIOPServerWorkerThreaded_ptr w, Buffer& b, bool s,
    OCI::Transport_ptr t)
{
    JTCSynchronized synchronized(*this);
    
    Request** p = &head_;
    while(*p)
	p = &((*p) -> next);
    
    *p = new Request;
    (*p) -> worker = GIOPServerWorkerThreaded::_duplicate(w);
    (*p) -> buf.consume(b);
    (*p) -> swap = s;
    (*p) -> transport = OCI::Transport::_duplicate(t);
    (*p) -> next = 0;
    
    notify();
}

#endif
