// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/Any.h>
#include <OB/OBV.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/IFR1.h>
#include <OB/Policy.h>
#include <OB/ORB.h>
#include <OB/Server.h>
#include <OB/GIOP.h>
#include <OB/IIOP.h>
#include <OB/Util.h>
#include <OB/Net.h>
#include <OB/Properties.h>
#include <OB/NativeTypes.h>
#include <OB/POAInterface.h>
#include <OB/POAServantBase.h>
#include <OB/OCI_IIOP.h>
#include <OB/CONV_FRAME.h>
#include <OB/CodeSetDatabase.h>

#include <Upcall.h>
#include <OARoot.h>

#include <OCI_IIOP_impl.h> // For new OCI_IIOP_..._impl

using namespace CORBA;
using namespace OCI;
using namespace OB;

// ----------------------------------------------------------------------
// ORB Initialization
// ----------------------------------------------------------------------

ORB_ptr
CORBA::ORB_init(int& ac, char** av, const char* orb_identifier)
{
    assert_nca(orb_identifier, NCANullString);

    MessageViewer* viewer = MessageViewer::instance();

    ORB_var orb = ORB::_OB_instance();
    if(!is_nil(orb))
    {
	viewer -> warning("ORB_init: ORB already initialized");
	return orb._retn();
    }

    //
    // Initialize JThreads/C++, if necessary
    //
#ifdef HAVE_JTC
    JTCInitialize* dummyForJTC = 0;
    
    //
    // If JTC has not been initialized then create a JTCInitialize
    // object to initialize the JThreads/C++ library
    //
    if(!JTCInitialize::initialized())
	dummyForJTC = new JTCInitialize(ac, av);
#endif
    
    //
    // Initialize the properties
    //
    Properties* properties = Properties::init(ac, av);
    
    //
    // Process each argument. Turn each argument into an appropriate
    // property.
    //
    int i = 0;
    while(i < ac)
    {
        const char* option = av[i];
        if(strcmp(option, "-ORBid") == 0)
        {
            if(i + 1 >= ac)
            {
                viewer -> error("ORB_init: argument expected for "
                                "-ORBid");
                throw INITIALIZE();
            }

            properties -> setProperty("ooc.orb.id", av[i + 1]);
            
            for(int j = i ; j + 2 < ac ; j++)
                av[j] = av[j + 2];
            
            ac -= 2;
        }
        else if(strcmp(option, "-ORBservice") == 0)
        {
            if(i + 2 >= ac)
            {
                viewer -> error("ORB_init: arguments expected for "
                                "-ORBservice");
                throw INITIALIZE();
            }

            String_var prop = string_dup("ooc.service.");
            prop += av[i + 1];
            properties -> setProperty(prop, av[i + 2]);
                
            for(int j = i ; j + 3 < ac ; j++)
                av[j] = av[j + 3];
            
            ac -= 3;
        }
        else if(strcmp(option, "-ORBrepository") == 0)
        {
            if(i + 1 >= ac)
            {
                viewer -> error("ORB_init: argument expected for "
                                "-ORBrepository");
                throw INITIALIZE();
            }

            properties -> setProperty("ooc.service.InterfaceRepository",
				      av[i + 1]);

            for(int j = i ; j + 2 < ac ; j++)
                av[j] = av[j + 2];
            
            ac -= 2;
        }
        else if(strcmp(option, "-ORBnaming") == 0)
        {
            if(i + 1 >= ac)
            {
                viewer -> error("ORB_init: argument expected for "
                                "-ORBnaming");
                throw INITIALIZE();
            }

            properties -> setProperty("ooc.service.NameService", av[i + 1]);
                
            for(int j = i ; j + 2 < ac ; j++)
                av[j] = av[j + 2];
            
            ac -= 2;
        }
        else if(strcmp(option, "-ORBtrace_level") == 0)
        {
            if(i + 1 >= ac)
            {
                viewer -> error("ORB::ORB_init: argument expected "
                                "for -ORBtrace_level");
                throw INITIALIZE();
            }

            properties -> setProperty("ooc.orb.trace_level", av[i + 1]);
            
            for(int j = i ; j + 2 < ac ; j++)
                av[j] = av[j + 2];
            
            ac -= 2;
        }        
        else if(strcmp(option, "-ORBblocking") == 0)
        {
            properties -> setProperty("ooc.orb.conc_model", "blocking");
            
            for(int j = i ; j + 1 < ac ; j++)
                av[j] = av[j + 1];
            
            ac -= 1;
        }
        else if(strcmp(option, "-ORBreactive") == 0)
        {
            properties -> setProperty("ooc.orb.conc_model", "reactive");
            
            for(int j = i ; j + 1 < ac ; j++)
                av[j] = av[j + 1];
            
            ac -= 1;
        }
#ifdef HAVE_JTC
        else if(strcmp(option, "-ORBthreaded") == 0)
        {
            properties -> setProperty("ooc.orb.conc_model", "threaded");
            
            for(int j = i ; j + 1 < ac ; j++)
                av[j] = av[j + 1];
            
            ac -= 1;
        }
#endif
        else if(strcmp(option, "-ORBversion") == 0)
        {
            viewer -> message(Version);
            
            for(int j = i ; j + 1 < ac ; j++)
                av[j] = av[j + 1];
            
            ac -= 1;
        }
        else if(strcmp(option, "-ORBlicense") == 0)
        {
            viewer -> message(License);
            
            for(int j = i ; j + 1 < ac ; j++)
                av[j] = av[j + 1];
            
            ac -= 1;
        }
	else if(strcmp(option, "-ORBnative_cs") == 0)
	{
            if(i + 1 >= ac)
            {
                viewer -> error("ORB_init: argument expected for "
                                "-ORBNCS");
                throw INITIALIZE();
            }

            properties -> setProperty("ooc.orb.native_cs", av[i + 1]);
                
            for(int j = i ; j + 2 < ac ; j++)
                av[j] = av[j + 2];
            
            ac -= 2;
	}
	else if(strcmp(option, "-ORBnative_wcs") == 0)
	{
            if(i + 1 >= ac)
            {
                viewer -> error("ORB_init: argument expected for "
                                "-ORBnative_wcs");
                throw INITIALIZE();
            }

            properties -> setProperty("ooc.orb.native_wcs", av[i + 1]);
                
            for(int j = i ; j + 2 < ac ; j++)
                av[j] = av[j + 2];
            
            ac -= 2;
	}
        else if(strncmp(option, "-ORB", 4) == 0)
        {
            String_var err =
                string_dup("ORB_init: unknown option: ");
            err += option;
            viewer -> warning(err);
	    
            for(int j = i ; j + 1 < ac ; j++)
                av[j] = av[j + 1];
	    
            ac -= 1;
        }
        else
            i++;
    }

    //
    // Process each property
    //
    String_var id = orb_identifier;
    ORB::ConcModel concModel = ORB::conc_model();
//    ULong native_cs = o -> native_cs();
//    ULong native_wcs = o -> native_wcs();
// XXX Get these from ORB
    ULong native_cs = CodeSetDatabase::UTF8;
    ULong native_wcs = CodeSetDatabase::UTF16;
    bool addIIOPConnector = true;
    
    Properties::KeySeq keys = properties -> getKeys("ooc.orb.");
    
    ULong j;
    for(j = 0; j < keys.length(); ++j)
    {
        const char* key = keys[j];
        const char* value = properties -> getProperty(key);
        assert(value != 0);

        if(strcmp(key, "ooc.orb.conc_model") == 0)
        {
            if(strcmp(value, "blocking") == 0)
                concModel = ORB::ConcModelBlocking;
            else if(strcmp(value, "reactive") == 0)
                concModel = ORB::ConcModelReactive;
#ifdef HAVE_JTC
            else if(strcmp(value, "threaded") == 0)
                concModel = ORB::ConcModelThreaded;
#endif
            else
            {
                String_var err =
                string_dup("ORB_init:: unknown value for "
                           "ooc.orb.conc_model: ");
                err += value;
                viewer -> warning(err);
            }
        }
        else if(strcmp(key, "ooc.orb.trace_level") == 0)
        {
            MessageViewer::setTraceLevel(atoi(value));
        }
        else if(strcmp(key, "ooc.orb.id") == 0)
        {
            id = string_dup(value);
        }
        else if(strcmp(key, "ooc.orb.add_iiop_connector") == 0)
        {
            if(strcmp(value, "true") == 0)
                addIIOPConnector = true;
            else
                addIIOPConnector = false;
        }
	else if(strcmp(key, "ooc.orb.native_cs") == 0)
	{
	    ULong id = CodeSetDatabase::instance() -> nameToId(value);
	    if(id)
		native_cs = id;
	    else
	    {
		String_var err =
		    string_dup("ORB_init:: unknown value for "
			       "ooc.orb.native_cs: ");
		err += value;
		viewer -> warning(err);
	    }
	}
	else if(strcmp(key, "ooc.orb.native_wcs") == 0)
	{
	    ULong id = CodeSetDatabase::instance() -> nameToId(value);
	    if(id)
		native_wcs = id;
	    else
	    {
		String_var err =
		    string_dup("ORB_init:: unknown value for "
			       "ooc.orb.native_wcs: ");
		err += value;
		viewer -> warning(err);
	    }
	}
        else
        {
            String_var err =
            string_dup("ORB_init: unknown property: ");
            err += key;
            viewer -> warning(err);
        }
    }

    //
    // Verify the ORB id
    //
    if(strlen(id) > 0 && strcmp(id, "OB_ORB") != 0)
    {
        String_var err =
        string_dup("ORB_init: unknown ORB `");
        err += id;
        err += "'";
        viewer -> error(err);
        throw INITIALIZE();
    }

    //
    // Create the ORB
    //
    ORB_var p = new ORB();

#ifdef HAVE_JTC
    //
    // Set the JTCInitialized argument
    //
    p -> dummyForJTC_ = dummyForJTC;
#endif

    //
    // Set the concurrency model
    //
    p -> conc_model(concModel);

    //
    // Add an IIOP connector if necessary
    //
    if(addIIOPConnector)
    {
        ConFactoryRegistry_var registry = p -> get_con_factory_registry();
        assert(!is_nil(registry));
        
        OCI::ConFactory_var factoryForIIOP =
            new OCI::IIOP::ConFactory_impl;
        registry -> add_factory(factoryForIIOP);
    }

    //
    // Parse the POA params. We must do this *before* adding the
    // configured services.
    //
    p -> parsePOAParams(ac, av);
        
    //
    // Add those services configured in the "ooc.service" property.
    // XXX - what if the service is collocated? Is this going to work
    // correctly? (ie. the user gets a fast-call proxy?)
    //
    Properties::KeySeq serviceKeys = properties -> getKeys("ooc.service.");
    for(j = 0 ; j < serviceKeys.length() ; j++)
    {
        const char* key = serviceKeys[j];
        const char* value = properties -> getProperty(key);
        assert(value != 0);
        assert(strncmp(key, "ooc.service.", strlen("ooc.service.")) == 0);
        key += strlen("ooc.service.");
        p -> _OB_addInitialService(key, value);
    }

    return p._retn();
}

// ----------------------------------------------------------------------
// POA Initialization
// ----------------------------------------------------------------------

void
ORB::parsePOAParams(int& ac, char** av)
{
    MessageViewer* viewer = MessageViewer::instance();

    //
    // Get the set of properties
    //
    Properties* properties = Properties::instance();

    //
    // Process each argument. Turn each argument into an appropriate
    // property.
    //
    int i = 0;
    while(i < ac)
    {
        const char* option = av[i];
        assert(option != 0);
        if(strcmp(option, "-OAdisable_iiop_acceptor") == 0)
        {
            properties -> setProperty("ooc.oa.disable_iiop_acceptor", "true");
            
            for(int j = i ; j + 1 < ac ; j++)
                av[j] = av[j + 1];
            
            ac -= 1;
        }
        else if(strcmp(option, "-OAnumeric") == 0)
	{
            properties -> setProperty("ooc.oa.numeric", "true");

            for(int j = i ; j + 1 < ac ; j++)
		av[j] = av[j + 1];

	    ac -= 1;
	}
	else if(strcmp(option, "-OAhost") == 0)
	{
	    if(i + 1 >= ac)
	    {
		viewer -> error("ORB::ORB_init: argument expected "
				"for -OAhost");
		throw INITIALIZE();
	    }

            properties -> setProperty("ooc.oa.host", av[i + 1]);

	    for(int j = i ; j + 2 < ac ; j++)
		av[j] = av[j + 2];

	    ac -= 2;
	}
	else if(strcmp(option, "-OAport") == 0)
	{
	    if(i + 1 >= ac)
	    {
                viewer -> error("ORB::ORB_init: argument expected "
				"for -OAport");
		throw INITIALIZE();
            }

            properties -> setProperty("ooc.oa.port", av[i + 1]);

	    for(int j = i ; j + 2 < ac ; j++)
		av[j] = av[j + 2];

	    ac -= 2;
	}
        else if(strcmp(option, "-OAblocking") == 0)
        {
            properties -> setProperty("ooc.oa.conc_model", "blocking");

            for(int j = i ; j + 1 < ac ; j++)
                av[j] = av[j + 1];
            
            ac -= 1;
        }
        else if(strcmp(option, "-OAreactive") == 0)
        {
            properties -> setProperty("ooc.oa.conc_model", "reactive");
            
            for(int j = i ; j + 1 < ac ; j++)
                av[j] = av[j + 1];
            
            ac -= 1;
        }
#ifdef HAVE_JTC
        else if(strcmp(option, "-OAthreaded") == 0)
        {
            properties -> setProperty("ooc.oa.conc_model", "threaded");
            
            for(int j = i ; j + 1 < ac ; j++)
                av[j] = av[j + 1];
            
            ac -= 1;
        }
        else if(strcmp(option, "-OAthread_per_client") == 0)
        {
            properties -> setProperty("ooc.oa.conc_model",
                                      "thread_per_client");
            
            for(int j = i ; j + 1 < ac ; j++)
                av[j] = av[j + 1];
            
            ac -= 1;
        }
        else if(strcmp(option, "-OAthread_per_request") == 0)
        {
            properties -> setProperty("ooc.oa.conc_model",
                                      "thread_per_request");
            
            for(int j = i ; j + 1 < ac ; j++)
                av[j] = av[j + 1];
            
            ac -= 1;
        }
        else if(strcmp(option, "-OAthread_pool") == 0)
        {
            if(i + 1 >= ac)
            {
                viewer -> error("ORB::ORB_init: argument expected "
                                "for -OAthread_pool");
                throw INITIALIZE();
            }

            properties -> setProperty("ooc.oa.conc_model", "thread_pool");
            properties -> setProperty("ooc.oa.thread_pool", av[i + 1]);
            
            for(int j = i ; j + 2 < ac ; j++)
                av[j] = av[j + 2];
            
            ac -= 2;
        }
#endif
        else if(strncmp(option, "-OA", 3) == 0)
        {
            String_var err =
            string_dup("ORB::ORB_init: unknown option: ");
            err += option;
            viewer -> warning(err);
	    
            for(int j = i ; j + 1 < ac ; j++)
                av[j] = av[j + 1];
	    
            ac -= 1;
        }
        else
            i++;
    }
}

void
ORB::_OB_initializeRootPOA(PortableServer::POA_ptr poa)
{
    //
    // Process each property
    //
    ORB::ServerConcModel concModel = ORB::server_conc_model();
    ULong concModelThreadPool = ORB::conc_model_thread_pool();
    bool iiopAcceptorDisabled = false;
    String_var host = string_dup("");
    UShort port = 0;
    bool numeric = false;
    bool addIIOPAcceptor = true;

    //
    // Get the viewer
    //
    MessageViewer* viewer = MessageViewer::instance();
    
    //
    // Get the set of properties
    //
    Properties* properties = Properties::instance();
    
    Properties::KeySeq keys = properties -> getKeys("ooc.oa.");
    ULong j;
    for(j = 0; j < keys.length(); ++j)
    {
        const char* key = keys[j];
        const char* value = properties -> getProperty(key);
        assert(value != 0);

        if(strcmp(key, "ooc.oa.disable_iiop_acceptor") == 0)
        {
            if(strcmp(value, "true") == 0)
                iiopAcceptorDisabled = true;
            else
                iiopAcceptorDisabled = false;
        }
        else if(strcmp(key, "ooc.oa.add_iiop_acceptor") == 0)
        {
            if(strcmp(value, "true") == 0)
                addIIOPAcceptor = true;
            else
                addIIOPAcceptor = false;
        }
        else if(strcmp(key, "ooc.oa.port") == 0)
        {
            int po = atoi(value);
            if(po < 0 || po > 65535)
            {
                viewer -> error("ORB::ORB_init: port must be between "
                                "0 and 65535");
                throw INITIALIZE();
            }
            port = (UShort)po;
        }
        else if(strcmp(key, "ooc.oa.host") == 0)
        {
            host = string_dup(value);
        }
        else if(strcmp(key, "ooc.oa.numeric") == 0)
        {
            if(strcmp(value, "true") == 0)
                numeric = true;
            else
                numeric = false;
        }
        else if(strcmp(key, "ooc.oa.conc_model") == 0)
        {
            if(strcmp(value, "blocking") == 0)
                concModel = ORB::ServerConcModelBlocking;
            else if(strcmp(value, "reactive") == 0)
                concModel = ORB::ServerConcModelReactive;
#ifdef HAVE_JTC
            else if(strcmp(value, "threaded") == 0)
                concModel = ORB::ServerConcModelThreaded;
            else if(strcmp(value, "thread_per_client") == 0)
                concModel = ORB::ServerConcModelThreadPerClient;
            else if(strcmp(value, "thread_per_request") == 0)
                concModel = ORB::ServerConcModelThreadPerRequest;
            else if(strcmp(value, "thread_pool") == 0)
                concModel = ORB::ServerConcModelThreadPool;
#endif
            else
            {
                String_var err =
                string_dup("ORB::ORB_init: unknown value for "
                           "ooc.oa.conc_model: ");
                err += value;
                viewer -> warning(err);
            }
        }
#ifdef HAVE_JTC
        else if(strcmp(key, "ooc.oa.thread_pool") == 0)
        {
            concModelThreadPool = atoi(value);
        }
#endif
        else
        {
            String_var err =
            string_dup("ORB::ORB_init: unknown property: ");
            err += key;
            viewer -> warning(err);
        }
    }
    
    //
    // Get the hostname
    //
    if(strlen(host) == 0)
    {
        //
        // We use MAXHOSTNAMELEN + 1 since we're not sure whether the
        // len parameter to gethostname() includes '\0' or not.
        //
	char hostname[MAXHOSTNAMELEN + 1];
	if(gethostname(hostname, MAXHOSTNAMELEN) == -1)
	{
	    throw COMM_FAILURE(LastError(), MinorGethostname, COMPLETED_NO);
	}
	
	struct hostent* ent;
	int retry = OB_GETHOSTBYNAME_RETRY;
	do
	{
	    ent = gethostbyname(hostname);
	}
#ifdef WIN32
	while(!ent && WSAGetLastError() == WSATRY_AGAIN && retry-- > 0);
#else
	while(!ent && h_errno == TRY_AGAIN && retry-- > 0);
#endif
	
	if(!ent)
	{
#ifdef WIN32
	    if(WSAGetLastError() == WSANO_DATA)
#else
	    if(h_errno == NO_DATA)
#endif
	    {
		viewer -> warning("ORB::ORB_init: "
				  "can't resolve hostname\n"
				  "using `localhost' (127.0.0.1) "
				  "instead of hostname");

		if(numeric)
		    host = string_dup("127.0.0.1");
		else
		    host = string_dup("localhost");
	    }
	    else
	    {
		throw COMM_FAILURE(LastHError(), MinorGethostbyname,
				   COMPLETED_NO);
	    }
	}
	else
	{
	    if(numeric || strrchr(ent -> h_name, '.') == 0)
	    {
/*
		if(!numeric)
		{
		    viewer -> warning("ORB::BOA_init: "
				      "hostname is a domainless name\n"
				      "using dotted decimal notation "
				      "instead of hostname");
		}
*/
		
		struct in_addr in;
		memcpy(&in.s_addr, ent -> h_addr, sizeof(in.s_addr));
		host = string_dup(inet_ntoa(in));
	    }
	    else
	    {
		host = string_dup(ent -> h_name);
	    }

	    if(strcmp(host, "127.0.0.1") == 0 ||
	       strcmp(host, "localhost") == 0)
	    {
		viewer -> warning("ORB::ORB_init: "
				  "hostname lookup returned "
				  "`localhost' (127.0.0.1)\n"
				  "Use the -OAhost option to select "
				  "some other hostname");
	    }
	}
    }

    //
    // Add the root POA
    //
    String_var poaName = poa -> the_name();
    _OB_addInitialService(poaName, poa, false);

    server_conc_model(concModel);
    conc_model_thread_pool(concModelThreadPool);
    
    //
    // Add an IIOP acceptor if necessary
    //
    if(addIIOPAcceptor)
    {
        AccRegistry_var registry = get_acc_registry();
        assert(!is_nil(registry));
        assert(strlen(host) > 0);
        OCI::Acceptor_var acceptor =
            new OCI::IIOP::Acceptor_impl(host, port);

        registry -> add_acceptor(acceptor);

        //
        // If the acceptor is disabled then inform the POA to disable it
        //
        //if(iiopAcceptorDisabled)
        //p -> disable_acceptor(acceptor -> tag());
    }
}
