// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/TemplateI.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/Any.h>
#include <OB/NamedValue.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(CORBA::NamedValue_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(CORBA::NamedValue_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

void
OBDuplicate(CORBA::NVList_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(CORBA::NVList_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// Template instantiations
// ----------------------------------------------------------------------

#ifndef HAVE_NO_EXPLICIT_TEMPLATES
template class OB::ObjVar< NamedValue >;
template class OB::ObjOut< NamedValue >;
template class OB::ObjForSeq< NamedValue >;
template class OB::ObjSeq< NamedValue, int >;
//template class OB::SeqVar< OB::ObjSeq< NamedValue > >;
//template class OB::SeqOut< OB::ObjSeq< NamedValue > >;

template class OB::ObjVar< NVList >;
template class OB::ObjOut< NVList >;
//template class OB::ObjForSeq< NVList >;
//template class OB::ObjSeq< NVList >;
//template class OB::SeqVar< OB::ObjSeq< NVList > >;
//template class OB::SeqOut< OB::ObjSeq< NVList > >;
#else
#ifdef HAVE_PRAGMA_DEFINE
#pragma define(OB::ObjVar< NamedValue >)
#pragma define(OB::ObjOut< NamedValue >)
#pragma define(OB::ObjForSeq< NamedValue >)
#pragma define(OB::ObjSeq< NamedValue, int >)
//#pragma define(OB::SeqVar< OB::ObjSeq< NamedValue > >)
//#pragma define(OB::SeqOut< OB::ObjSeq< NamedValue > >)

#pragma define(OB::ObjVar< NVList >)
#pragma define(OB::ObjOut< NVList >)
//#pragma define(OB::ObjForSeq< NVList >)
//#pragma define(OB::ObjSeq< NVList >)
//#pragma define(OB::SeqVar< OB::ObjSeq< NVList > >)
//#pragma define(OB::SeqOut< OB::ObjSeq< NVList > >)
#endif
#endif

// ----------------------------------------------------------------------
// NamedValue constructor and destructor
// ----------------------------------------------------------------------

NamedValue::NamedValue()
    :  name_(string_dup("")), value_(new Any), flags_(0)
{
}

NamedValue::NamedValue(char* n, Any* a, Flags f)
    :  name_(n), value_(a), flags_(f)
{
}

// ----------------------------------------------------------------------
// NamedValue public member implementation
// ----------------------------------------------------------------------

const char*
NamedValue::name() const
{
    return name_;
}

Any*
NamedValue::value() const
{
#ifdef HAVE_NO_MUTABLE
    return ((NamedValue*)this) -> value_;
#else
    return value_;
#endif
}

Flags
NamedValue::flags() const
{
    return flags_;
}

// ----------------------------------------------------------------------
// NVList constructor
// ----------------------------------------------------------------------

NVList::NVList(ULong n)
    : namedValueSeq_(n)
{
    namedValueSeq_.length(n);
    for(ULong i = 0 ; i < n ; i++)
	namedValueSeq_[i] = new NamedValue;
}

// ----------------------------------------------------------------------
// NVList public member implementation
// ----------------------------------------------------------------------

ULong
NVList::count() const
{
    return namedValueSeq_.length();
}

NamedValue_ptr
NVList::add(Flags f)
{
    NamedValue_ptr p =
	new NamedValue(string_dup(""), new Any, f);
    namedValueSeq_.append(p);
    return p;
}

NamedValue_ptr
NVList::add_item(const char* n, Flags f)
{
    assert_nca(n, NCANullString);
    NamedValue_ptr p = new NamedValue(string_dup(n),
						  new Any, f);
    namedValueSeq_.append(p);
    return p;
}

NamedValue_ptr
NVList::add_value(const char* n, const Any& a, Flags f)
{
    assert_nca(n, NCANullString);
    NamedValue_ptr p = new NamedValue(string_dup(n),
						  new Any(a), f);
    namedValueSeq_.append(p);
    return p;
}

NamedValue_ptr
NVList::add_item_consume(char* n, Flags f)
{
    assert_nca(n, NCANullString);
    NamedValue_ptr p = new NamedValue(n, new Any, f);
    namedValueSeq_.append(p);
    return p;
}

NamedValue_ptr
NVList::add_value_consume(char* n, Any* a, Flags f)
{
    assert_nca(n, NCANullString);
    NamedValue_ptr p = new NamedValue(n, a, f);
    namedValueSeq_.append(p);
    return p;
}

NamedValue_ptr
NVList::item(ULong n)
{
    if(n >= namedValueSeq_.length())
        throw Bounds();
    
    return namedValueSeq_[n];
}

void
NVList::remove(ULong n)
{
    if(n >= namedValueSeq_.length())
        throw Bounds();
    
    namedValueSeq_.remove(n);
}
