// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_OA_ROOT_H
#define OB_OA_ROOT_H

namespace PortableServer
{

//
// Forward declarations
//
class POA_impl;
class Current_impl;

}

namespace OB
{

class Server;
typedef Server* Server_ptr;
typedef Server_ptr ServerRef;

} // End of namespace OB

void OBDuplicate(OB::Server_ptr);
void OBRelease(OB::Server_ptr);

namespace OB
{

//
// The OARoot class
//
typedef ObjVar< Server > Server_var;
typedef ObjSeq< Server, int > ServerSeq;
typedef SeqVar< ObjSeq< Server, int > > ServerSeq_var;

class OARoot : public Upcall
{
private:

#ifdef HAVE_JTC
    JTCMonitor serversMonitor_;
#endif
    ServerSeq servers_;
    
    bool serversInitialized_;
    
    OCI::AccRegistry_var registry_;

#ifdef HAVE_JTC
    JTCMutex poaMutex_;
#endif
    PortableServer::POA_impl* rootPOA_;
    
    static CORBA::ORB::ServerConcModel serverConcModel_;
    
    static CORBA::ULong serverConcModelThreadPool_;

    //
    // The object adapater singleton
    //
    static OARoot* instance_;
    
    friend class CORBA::ORB;
    OARoot();

public:
    
    ~OARoot();

    static OARoot* instance();
    
    void createNewServers();

    bool initServers();

    void destroyServers();
    
    void run();
   
    void shutdown(CORBA::Boolean);
    
    bool workPending();
    
    void performWork();

    void invoke(Downcall_ptr,
                CORBA::ULong,
                const OctetSeq& key,
                const char* op,
                Buffer& buf,
                bool sw,
                CORBA::ULong offIn,
                CORBA::ULong offOut,
                OCI::TransportInfo_ptr);
    
    bool findByKey(const OctetSeq&);

    //
    // Accessors
    //
    static void serverConcModel(CORBA::ORB::ServerConcModel);
    static CORBA::ORB::ServerConcModel serverConcModel();

    static void serverThreadPool(CORBA::ULong);
    static CORBA::ULong serverThreadPool();

    OCI::AccRegistry_ptr getAccRegistry();

    PortableServer::POA_ptr getRootPOA();
    bool hasRootPOA() const;
};

} // End of namespace OB

#endif
