// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/POAServantBase.h>
#include <OB/OBPolicies.h>

#include <OBPolicies_impl.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// ProtocolPolicy_impl constructor and destructor
// ----------------------------------------------------------------------

ProtocolPolicy_impl::ProtocolPolicy_impl(IOP::ProfileId value)
    : value_(value)
{
}

ProtocolPolicy_impl::~ProtocolPolicy_impl()
{
}

// ----------------------------------------------------------------------
// ProtocolPolicy_impl public member implementation
// ----------------------------------------------------------------------

IOP::ProfileId
ProtocolPolicy_impl::value()
    throw(SystemException)
{
    return value_;
}

PolicyType
ProtocolPolicy_impl::policy_type()
    throw(SystemException)
{
    return PROTOCOL_POLICY;
}

Policy_ptr
ProtocolPolicy_impl::copy()
    throw(SystemException)
{
    return Policy::_duplicate(this);
}

void
ProtocolPolicy_impl::destroy()
    throw(SystemException)
{
}

// ----------------------------------------------------------------------
// ConnectionReusePolicy_impl constructor and destructor
// ----------------------------------------------------------------------

ConnectionReusePolicy_impl::ConnectionReusePolicy_impl(Boolean r)
    : value_(r)
{
}

ConnectionReusePolicy_impl::~ConnectionReusePolicy_impl()
{
}

// ----------------------------------------------------------------------
// ConnectionReusePolicy_impl public member implementation
// ----------------------------------------------------------------------

Boolean
ConnectionReusePolicy_impl::value()
    throw(SystemException)
{
    return value_;
}

PolicyType
ConnectionReusePolicy_impl::policy_type()
    throw(SystemException)
{
    return CONNECTION_REUSE_POLICY;
}

Policy_ptr
ConnectionReusePolicy_impl::copy()
    throw(SystemException)
{
    return Policy::_duplicate(this);
}

void
ConnectionReusePolicy_impl::destroy()
    throw(SystemException)
{
}

// ----------------------------------------------------------------------
// ReconnectPolicy_impl constructor and destructor
// ----------------------------------------------------------------------

ReconnectPolicy_impl::ReconnectPolicy_impl(Boolean r)
    : value_(r)
{
}

ReconnectPolicy_impl::~ReconnectPolicy_impl()
{
}

// ----------------------------------------------------------------------
// ReconnectPolicy_impl public member implementation
// ----------------------------------------------------------------------

Boolean
ReconnectPolicy_impl::value()
    throw(SystemException)
{
    return value_;
}

PolicyType
ReconnectPolicy_impl::policy_type()
    throw(SystemException)
{
    return RECONNECT_POLICY;
}

Policy_ptr
ReconnectPolicy_impl::copy()
    throw(SystemException)
{
    return Policy::_duplicate(this);
}

void
ReconnectPolicy_impl::destroy()
    throw(SystemException)
{
}

// ----------------------------------------------------------------------
// TimeoutPolicy_impl constructor and destructor
// ----------------------------------------------------------------------

TimeoutPolicy_impl::TimeoutPolicy_impl(ULong t)
{
    if(t > 0x7fffffff)
        throw IMP_LIMIT();
    value_ = t;
}

TimeoutPolicy_impl::~TimeoutPolicy_impl()
{
}

// ----------------------------------------------------------------------
// TimeoutPolicy_impl public member implementation
// ----------------------------------------------------------------------

ULong
TimeoutPolicy_impl::value()
    throw(SystemException)
{
    return value_;
}

PolicyType
TimeoutPolicy_impl::policy_type()
    throw(SystemException)
{
    return TIMEOUT_POLICY;
}

Policy_ptr
TimeoutPolicy_impl::copy()
    throw(SystemException)
{
    return Policy::_duplicate(this);
}

void
TimeoutPolicy_impl::destroy()
    throw(SystemException)
{
}


