// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OCI_IIOP_IMPL_H
#define OCI_IIOP_IMPL_H

namespace OCI
{

namespace IIOP
{

//
// Declare classes
//
class Transport_impl;
class TransportInfo_impl;
class Acceptor_impl;
class AcceptorInfo_impl;
class Connector_impl;
class ConnectorInfo_impl;
class ConFactory_impl;
class ConFactoryInfo_impl;

//
// The Transport_impl class
//
class Transport_impl : public Transport
{
    //
    // Hide copy-constructor and assignment operator
    //
    Transport_impl(const Transport_impl&);
    void operator=(const Transport_impl&);

protected:

    Handle fd_; // The handle
    CORBA::Boolean block_; // Set to true if blocking IO is selected
    TransportInfo_impl* info_; // Transport information
    
    void setBlock(CORBA::Boolean); // Set blocking/non-blocking behavior

    //
    // The corresponding info class is a friend
    //
    friend class TransportInfo_impl;

public:

    Transport_impl(Connector_impl*, Handle, CORBA::Boolean);
    Transport_impl(Acceptor_impl*, Handle, CORBA::Boolean);
    virtual ~Transport_impl();

    //
    // Standard IDL to C++ Mapping
    //
    virtual ProfileId tag()
        throw(CORBA::SystemException)
    {
        return IOP::TAG_INTERNET_IOP;
    }
    
    virtual Handle handle()
        throw(CORBA::SystemException)
    {
        return fd_;
    }
    
    virtual CORBA::ULong fragmentation()
        throw(CORBA::SystemException)
    {
        return 0;
    }
    
    virtual void close()
	throw(CORBA::SystemException);
    
    virtual void shutdown()
        throw(CORBA::SystemException);
    
    virtual void receive(Buffer_ptr, CORBA::Boolean)
        throw(CORBA::SystemException);
    
    virtual CORBA::Boolean receive_detect(Buffer_ptr, CORBA::Boolean)
        throw(CORBA::SystemException);
    
    virtual void receive_timeout(Buffer_ptr, CORBA::ULong)
        throw(CORBA::SystemException);
    
    virtual void send(Buffer_ptr, CORBA::Boolean)
        throw(CORBA::SystemException);
    
    virtual CORBA::Boolean send_detect(Buffer_ptr, CORBA::Boolean)
        throw(CORBA::SystemException);
    
    virtual void send_timeout(Buffer_ptr, CORBA::ULong)
        throw(CORBA::SystemException);
    
    virtual OCI::TransportInfo_ptr get_info()
        throw(CORBA::SystemException);
};

//
// The Connector_impl class
//
class Connector_impl : public Connector
{
    //
    // Hide copy-constructor and assignment operator
    //
    Connector_impl(const Connector_impl&);
    void operator=(const Connector_impl&);

protected:

    CORBA::String_var host_; // The host
    CORBA::UShort port_; // The port
    struct sockaddr_in addr_; // The socket address
    Handle fd_; // The handle
    ConnectorInfo_impl* info_; // Connector information

    void close(); // Close the connector

    //
    // The corresponding info class is a friend
    //
    friend class ConnectorInfo_impl;

public:

    Connector_impl(const char*, CORBA::UShort,
                            const ConnectCBSeq& cb);
    virtual ~Connector_impl();

    //
    // Standard IDL to C++ Mapping
    //
    virtual ProfileId tag()
        throw(CORBA::SystemException)
    {
        return IOP::TAG_INTERNET_IOP;
    }
    
    virtual Handle handle()
        throw(CORBA::SystemException)
    {
        return fd_;
    }
    
    virtual OCI::Transport_ptr connect()
        throw(CORBA::SystemException);

    virtual OCI::Transport_ptr connect_timeout(CORBA::ULong)
        throw(CORBA::SystemException);

    virtual ObjectKey* is_usable(const IOR&)
        throw(CORBA::SystemException);
    
    virtual ObjectKey* is_usable_with_policies(const IOR&,
					       const CORBA::PolicyList&)
        throw(CORBA::SystemException);
    
    virtual OCI::ConnectorInfo_ptr get_info()
        throw(CORBA::SystemException);
};

//
// The Acceptor_impl class
//
class Acceptor_impl : public Acceptor
{
    //
    // Hide copy-constructor and assignment operator
    //
    Acceptor_impl(const Acceptor_impl&);
    void operator=(const Acceptor_impl&);

protected:

    CORBA::String_var host_; // The host
    CORBA::UShort port_; // The port
    struct sockaddr_in addr_; // The socket address
    Handle fd_; // The handle
    AcceptorInfo_impl* info_; // Acceptor info

    //
    // The corresponding info class is a friend
    //
    friend class AcceptorInfo_impl;

public:

    Acceptor_impl(const char*, CORBA::UShort);
    virtual ~Acceptor_impl();

    //
    // Standard IDL to C++ Mapping
    //
    virtual ProfileId tag()
        throw(CORBA::SystemException)
    {
        return IOP::TAG_INTERNET_IOP;
    }
    
    virtual Handle handle()
        throw(CORBA::SystemException)
    {
        return fd_;
    }
    
    virtual void close()
        throw(CORBA::SystemException);
    
    virtual void shutdown()
        throw(CORBA::SystemException);
    
    virtual void listen()
        throw(CORBA::SystemException);
    
    virtual OCI::Transport_ptr accept()
        throw(CORBA::SystemException);
    
    virtual void add_profile(const ObjectKey&, IOR&)
        throw(CORBA::SystemException);
    
    virtual ObjectKey* is_local(const IOR&)
        throw(CORBA::SystemException);
    
    virtual OCI::AcceptorInfo_ptr get_info()
        throw(CORBA::SystemException);
};

//
// The ConFactory_impl class
//
class ConFactory_impl : public ConFactory
{
    //
    // Hide copy-constructor and assignment operator
    //
    ConFactory_impl(const ConFactory_impl&);
    void operator=(const ConFactory_impl&);

    ConFactoryInfo_impl* info_; // ConFactory info

public:

    ConFactory_impl();
    virtual ~ConFactory_impl();

    //
    // Standard IDL to C++ Mapping
    //
    virtual ProfileId tag()
        throw(CORBA::SystemException)
    {
        return IOP::TAG_INTERNET_IOP;
    }
    
    virtual OCI::Connector_ptr create(const IOR&)
        throw(CORBA::SystemException);
    
    virtual OCI::Connector_ptr create_with_policies(const IOR&,
					            const CORBA::PolicyList&)
        throw(CORBA::SystemException);
    
    virtual CORBA::Boolean consider_with_policies(const IOR&,
						  const CORBA::PolicyList&)
        throw(CORBA::SystemException);
    
    virtual CORBA::Boolean equivalent(const IOR&, const IOR&)
        throw(CORBA::SystemException);
    
    virtual CORBA::ULong hash(const IOR&, CORBA::ULong)
        throw(CORBA::SystemException);
    
    virtual OCI::ConFactoryInfo_ptr get_info()
        throw(CORBA::SystemException);
};

//
// The TransportInfo_impl class
//
class TransportInfo_impl
    : public TransportInfo OB_COMMA_MONITOR
{
    //
    // Hide copy-constructor and assignment operator
    //
    TransportInfo_impl(const TransportInfo_impl&);
    void operator=(const TransportInfo_impl&);

    OCI::ConnectorInfo_var connectorInfo_; // The connector info
    OCI::AcceptorInfo_var acceptorInfo_; // The acceptor info
    OCI::CloseCBSeq closeCBSeq_; // All close callback objects
    Transport_impl* transport_; // The associated transport
    struct sockaddr_in addr_; // Local address and port
    struct sockaddr_in remoteAddr_; // Remote address and port

    //
    // Initialize data members
    //
    void init();

public:

    TransportInfo_impl(Transport_impl* transport,
				Connector_impl* connector);
    TransportInfo_impl(Transport_impl* transport,
				Acceptor_impl* acceptor);
    virtual ~TransportInfo_impl();

    //
    // Standard IDL to C++ Mapping
    //
    virtual OCI::ProfileId tag()
        throw(CORBA::SystemException)
    {
        return IOP::TAG_INTERNET_IOP;
    }
    
    virtual OCI::ConnectorInfo_ptr connector_info()
        throw(CORBA::SystemException);
    
    virtual OCI::AcceptorInfo_ptr acceptor_info()
        throw(CORBA::SystemException);
    
    virtual void add_close_cb(OCI::CloseCB_ptr cb)
        throw(CORBA::SystemException);
    
    virtual void remove_close_cb(OCI::CloseCB_ptr cb)
        throw(CORBA::SystemException);
    
    virtual InetAddr_slice* addr()
        throw(CORBA::SystemException);
    
    virtual CORBA::UShort port()
        throw(CORBA::SystemException);
    
    virtual InetAddr_slice* remote_addr()
        throw(CORBA::SystemException);
    
    virtual CORBA::UShort remote_port()
        throw(CORBA::SystemException);

    //
    // ORBacus internal functions
    // Application programs must not use these functions directly
    //
    void _OB_callCloseCB(OCI::TransportInfo_ptr);
    void _OB_destroy();
};

//
// The ConnectorInfo_impl class
//
class ConnectorInfo_impl
    : public ConnectorInfo OB_COMMA_MONITOR
{
    //
    // Hide copy-constructor and assignment operator
    //
    ConnectorInfo_impl(const ConnectorInfo_impl&);
    void operator=(const ConnectorInfo_impl&);

    OCI::ConnectCBSeq connectCBSeq_; // All connect callback objects
    Connector_impl* connector_; // The associated connector

public:

    ConnectorInfo_impl(Connector_impl* connector, const OCI::ConnectCBSeq& cb)
	: connectCBSeq_(cb), connector_(connector) { }
    virtual ~ConnectorInfo_impl() { }

    //
    // Standard IDL to C++ Mapping
    //
    virtual OCI::ProfileId tag()
        throw(CORBA::SystemException)
    {
        return IOP::TAG_INTERNET_IOP;
    }
    
    virtual void add_connect_cb(OCI::ConnectCB_ptr cb)
        throw(CORBA::SystemException);
    
    virtual void remove_connect_cb(OCI::ConnectCB_ptr cb)
        throw(CORBA::SystemException);
    
    virtual InetAddr_slice* remote_addr()
        throw(CORBA::SystemException);
    
    virtual CORBA::UShort remote_port()
        throw(CORBA::SystemException);

    //
    // ORBacus internal functions
    // Application programs must not use these functions directly
    //
    void _OB_callConnectCB(OCI::TransportInfo_ptr);
    void _OB_destroy();
};

//
// The AcceptorInfo_impl class
//
class AcceptorInfo_impl
    : public AcceptorInfo OB_COMMA_MONITOR
{
    //
    // Hide copy-constructor and assignment operator
    //
    AcceptorInfo_impl(const AcceptorInfo_impl&);
    void operator=(const AcceptorInfo_impl&);

    OCI::AcceptCBSeq acceptCBSeq_; // All accept callback objects
    Acceptor_impl* acceptor_; // The associated acceptor

public:
    
    AcceptorInfo_impl(Acceptor_impl* acceptor)
	: acceptor_(acceptor) { }
    virtual ~AcceptorInfo_impl() { }

    //
    // Standard IDL to C++ Mapping
    //
    virtual OCI::ProfileId tag()
        throw(CORBA::SystemException)
    {
        return IOP::TAG_INTERNET_IOP;
    }
    
    virtual void add_accept_cb(OCI::AcceptCB_ptr cb)
        throw(CORBA::SystemException);
    
    virtual void remove_accept_cb(OCI::AcceptCB_ptr cb)
        throw(CORBA::SystemException);
    
    virtual char* host()
        throw(CORBA::SystemException);
    
    virtual InetAddr_slice* addr()
        throw(CORBA::SystemException);
    
    virtual CORBA::UShort port()
        throw(CORBA::SystemException);

    //
    // ORBacus internal functions
    // Application programs must not use these functions directly
    //
    void _OB_callAcceptCB(OCI::TransportInfo_ptr);
    void _OB_destroy();
};

//
// The ConFactoryInfo_impl class
//
class ConFactoryInfo_impl :
    public ConFactoryInfo OB_COMMA_MONITOR
{
    //
    // Hide copy-constructor and assignment operator
    //
    ConFactoryInfo_impl(const ConFactoryInfo_impl&);
    void operator=(const ConFactoryInfo_impl&);

    OCI::ConnectCBSeq connectCBSeq_; // All connect callback objects

public:

    ConFactoryInfo_impl() { }

    //
    // Standard IDL to C++ Mapping
    //
    virtual OCI::ProfileId tag()
        throw(CORBA::SystemException)
    {
        return IOP::TAG_INTERNET_IOP;
    }
    
    virtual void add_connect_cb(OCI::ConnectCB_ptr cb)
        throw(CORBA::SystemException);
    
    virtual void remove_connect_cb(OCI::ConnectCB_ptr cb)
        throw(CORBA::SystemException);

    //
    // ORBacus internal functions
    // Application programs must not use these functions directly
    //
    OCI::ConnectCBSeq* _OB_getConnectCBSeq();
};

} // End of namespace IIOP

} // End of namespace OCI

#endif
