// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OCI_IMPL_H
#define OCI_IMPL_H

namespace OB
{

//
// Forward declarations
//
class OARoot;

} // End of namespace OB


namespace OCI
{

//
// The ConFactoryRegistry_impl class
//
class ConFactoryRegistry_impl : virtual public ConFactoryRegistry
{
    //
    // All transport factories
    //
    ConFactorySeq factories_;

    //
    // Hide copy-constructor and assignment operator
    //
    ConFactoryRegistry_impl(const ConFactoryRegistry_impl&);
    void operator=(const ConFactoryRegistry_impl&);

    ConFactoryRegistry_impl() { }
    virtual ~ConFactoryRegistry_impl() { }
    friend class CORBA::ORB; // ORBs may create ConFactoryRegistries_impls

public:

    //
    // Standard IDL to C++ Mapping
    //
    virtual void add_factory(ConFactory_ptr)
        throw(CORBA::SystemException);
    virtual ConFactory_ptr get_factory(const IOR&)
        throw(CORBA::SystemException);
    virtual ConFactory_ptr get_factory_with_policies(const IOR&,
						     const
						     CORBA::PolicyList&)
        throw(CORBA::SystemException);
    virtual ConFactorySeq* get_factories()
        throw(CORBA::SystemException);
    virtual CORBA::Boolean equivalent(const IOR&, const IOR&)
        throw(CORBA::SystemException);
    virtual CORBA::ULong hash(const IOR&, CORBA::ULong maximum)
        throw(CORBA::SystemException);
};

//
// The AccRegistry_impl class
//
class AccRegistry_impl : virtual public AccRegistry
{
    AcceptorSeq acceptors_; // All profile acceptors

    //
    // Hide copy-constructor and assignment operator
    //
    AccRegistry_impl(const AccRegistry_impl&);
    void operator=(const AccRegistry_impl&);

    AccRegistry_impl() { }
    virtual ~AccRegistry_impl() { }
    friend class ::OB::OARoot; // OAs may create AccRegistries_impls

public:

    //
    // Standard IDL to C++ Mapping
    //
    virtual void add_acceptor(Acceptor_ptr)
        throw(CORBA::SystemException);
    virtual AcceptorSeq* get_acceptors()
        throw(CORBA::SystemException);
    virtual void add_profiles(const ObjectKey&, IOR&)
        throw(CORBA::SystemException);
    virtual ObjectKey* is_local(const IOR&)
        throw(CORBA::SystemException);
};

//
// The Current_impl class
//
class CurrentState;
class Current_impl : virtual public Current
{
    //
    // Hide copy-constructor and assignment operator
    //
    Current_impl(const Current_impl&);
    void operator=(const Current_impl&);

#ifdef HAVE_JTC
    JTCThreadKey stateKey_;
#else
    CurrentState* state_;
#endif

public:

    Current_impl();
    virtual ~Current_impl();
    
    //
    // Standard IDL to C++ Mapping
    //
    virtual TransportInfo_ptr get_oci_transport_info()
        throw(CORBA::SystemException);
    virtual AcceptorInfo_ptr get_oci_acceptor_info()
        throw(CORBA::SystemException);
    
    //
    // ORBacus internal functions
    // Application programs must not use these functions directly
    //
    void _OB_preinvoke(TransportInfo_ptr);
    void _OB_postinvoke();
};

} // End of namespace OCI

#endif
