// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/TemplateI.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/Any.h>
#include <OB/OBV.h>
#include <OB/TCKind.h>
#include <OB/TypeCode.h>
#include <OB/Environment.h>
#include <OB/Context.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/GIOP.h>
#include <OB/IIOP.h>
#include <OB/IFR1.h>
#include <OB/IFR4.h>
#include <OB/Policy.h>
#include <OB/POAServantBase.h>
#include <OB/PolicyManager.h>
#include <OB/OBPolicies.h>
#include <OB/ValueBase.h>
#include <OB/ORB.h>
#include <OB/OCI.h>
#include <OB/OCI_IIOP.h>
#include <OB/Properties.h>
#include <OB/Util.h>
#include <OB/Hashtable.h>
#include <OB/HashtableI.h>
#include <OB/NativeTypes.h>
#include <OB/POAServantBase.h>
#include <OB/POAInterface.h>
#include <OB/POAServantLocator.h>
#include <OB/Net.h> // For windows sockets startup
#include <OB/ValueFactoryBase.h>
#include <OB/Stream.h>
#include <OB/CONV_FRAME.h>
#include <OB/CodeSetDatabase.h>
#include <OB/CodeConverter.h>
#include <OB/CodeSetUtil.h>
#include <OB/Hashers.h> // For hash functions

#include <OCI_impl.h> // For new OCI_ConFactoryRegistry_impl
                      // and OCI_Current_impl

#include <ORBPolicyManager_impl.h> // For new ORBPolicyManager_impl
#include <OBPolicies_impl.h> // For new OB_ConnectionReusePolicy_impl

#include <POA_impl.h>

#include <GIOPClient.h>

#include <Upcall.h>
#include <OARoot.h>
#include <CollocatedClient.h>

#include <IIOPUtil.h>

#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>

using namespace CORBA;
using namespace OB;

//
// The ORB - only one instance is allowed
//
ORB_ptr ORB::orb_ = ORB::_nil();

//
// The concurrency model
//
ORB::ConcModel ORB::concModel_ = ORB::ConcModelBlocking;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(CORBA::ORB_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(CORBA::ORB_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// Template instantiations
// ----------------------------------------------------------------------

#ifndef HAVE_NO_EXPLICIT_TEMPLATES
template class OB::ObjVar< ORB >;
//template class OB::ObjForSeq< ORB >;
// template class OB::ObjSeq< ORB >;
// template class OB::SeqVar< OB::ObjSeq< ORB > >;
// template class OB::SeqOut< OB::ObjSeq< ORB > >;

template class OB::Hashtable< String_var, Object_var >;
template class OB::Hashtable< String_var, ValueFactoryBase_var >;
#else
#ifdef HAVE_PRAGMA_DEFINE
#pragma define(OB::ObjVar< ORB >)
// #pragma define(OB::ObjForSeq< ORB >)
// #pragma define(OB::ObjSeq< ORB >)
// #pragma define(OB::SeqVar< OB::ObjSeq< ORB > >)
// #pragma define(OB::SeqOut< OB::ObjSeq< ORB > >)

#pragma define(OB::Hashtable< String_var, Object_var >)
#pragma define(OB::Hashtable< String_var, ValueFactoryBase_var >)
#endif
#endif

// ----------------------------------------------------------------------
// ORB constructor and destructor
// ----------------------------------------------------------------------

ORB::ORB()
#ifdef HAVE_JTC
    : dummyForJTC_(0)
#endif
{
#ifdef HAVE_JTC       
    //                       
    // If JTC has not been initialized then create a JTCInitialize object
    //                      
    if(!JTCInitialize::initialized())
    {                  
        dummyForJTC_ = new JTCInitialize();
    }              
#endif

#ifdef HAVE_WINSOCK_H
    //
    // Winsock startup
    //
    WORD wVersionRequested = MAKEWORD(1, 1);
    WSADATA wsaData;
    if(WSAStartup(wVersionRequested, &wsaData) != 0)
    {
	throw COMM_FAILURE(LastError(), MinorWSAStartup, COMPLETED_NO);
    }
#endif

#ifndef WIN32
    //
    // Set SIGPIPE signal handler
    //
    struct sigaction actPIPE;
    actPIPE.sa_handler = SIG_IGN;
    sigemptyset(&actPIPE.sa_mask);
    actPIPE.sa_flags = 0;
    sigaction(SIGPIPE, &actPIPE, 0);
#endif

    //
    // Check if there is already an ORB singleton
    //
    assert(is_nil(orb_));

    //
    // Set the ORB singleton to this instance
    //
    orb_ = this;

    //
    // Set the native code set defaults
    //
    native_cs_ = CodeSetDatabase::UTF8;
    native_wcs_ = CodeSetDatabase::UTF16;

    //
    // Create the OARoot
    //
    oaRoot_ = new OARoot();

    //
    // Create the initial services map
    //
    services_ = new InitialServiceMap(63);

    //
    // Create the ValueFactoryBase map
    //
    valueFactoryBaseMap_ = new ValueFactoryBaseMap(1023);
    
    //
    // Add the default policies
    //
    ConnectionReusePolicy_impl* reusePolicy =
        new ConnectionReusePolicy_impl(true);
    policies_.append(reusePolicy);
    
    ReconnectPolicy_impl* retryPolicy = new ReconnectPolicy_impl(false);
    policies_.append(retryPolicy);

    //
    // Create the ORBPolicyManager
    //
    PolicyManager_var pm = new ORBPolicyManager_impl(policies_);
    _OB_addInitialService("ORBPolicyManager", pm);

    //
    // Create the OCI::Current object, and add to the set of initial
    // services
    //
    ociCurrent_ = new OCI::Current_impl;
    _OB_addInitialService("OCICurrent", ociCurrent_);

    //
    // Add a fake `RootPOA' entry
    //
    _OB_addInitialService("RootPOA", Object::_nil());
    
    //
    // Create the registry
    //
    registry_ = new OCI::ConFactoryRegistry_impl;
}

ORB::~ORB()
{
    //
    // Destroy the OARoot
    //
    delete oaRoot_;

    //
    // Destroy all clients
    //
    ULong i;
    for(i = 0 ; i < clients_.length() ; i++)
	clients_[i] -> destroy();

    //
    // Clean out the initial services
    //
    delete services_;
    delete valueFactoryBaseMap_;
    release(ociCurrent_);
    
    //
    // Clean out policies
    //
    for(i = 0 ; i < policies_.length() ; i++)
	policies_[i] -> destroy();
    
    //
    // Set ORB singleton to nil
    //
    orb_ = ORB::_nil();

#ifdef HAVE_WINSOCK_H
    //
    // Winsock shutdown
    //
    // This is a hack to fix problems with OB/MT. The client threads
    // can still be trying to run at this point, and calling winsock
    // functions after WSACleanup() has been called causes access
    // violations.
    //
    // Of course, this isn't a good fix.  A better fix would be to
    // wait for those threads to exit, via a ThreadGroup, or something
    // similar.
    //
#   ifndef HAVE_JTC
    if(WSACleanup() != 0)
    {
	throw COMM_FAILURE(LastError(), MinorWSACleanup, COMPLETED_NO);
    }
#   endif
#endif

#ifndef WIN32
    //
    // Reset SIGPIPE signal handler
    //
    struct sigaction actPIPE;
    actPIPE.sa_handler = SIG_DFL;
    sigemptyset(&actPIPE.sa_mask);
    actPIPE.sa_flags = 0;
    sigaction(SIGPIPE, &actPIPE, 0);
#endif

#ifdef HAVE_JTC
    //
    // Delete the JTCInitialize object if it's been created
    //
    delete dummyForJTC_;
#endif
}

// ----------------------------------------------------------------------
// ORB public member implementation
// ----------------------------------------------------------------------

char*
ORB::object_to_string(Object_ptr p)
{
    if(p -> _OB_isInternalLocal())
        throw BAD_PARAM();
    
    IOP::IOR ior;
    
    if(is_nil(p))
	ior.type_id = string_dup("");
    else
	ior = p -> _OB_ior(); 

    return _OB_IORToString(ior);
}

Object_ptr
ORB::string_to_object(const char* s)
{
    assert_nca(s, NCANullString);

    IOP::IOR_var ior = _OB_stringToIOR(s);
    return _OB_createObject(ior);
}

void
ORB::get_default_context(Context_out ctx)
{
    ctx = new Context("");
}

void
ORB::create_environment(Environment_out env)
{
    env = new Environment;
}

TypeCode_ptr
ORB::create_struct_tc(const char* id,
		      const char* name,
		      const StructMemberSeq& members)
{
    assert_nca(id, NCANullString);
    assert_nca(name, NCANullString);

    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_struct;
    tc -> id_ = id;
    tc -> name_ = name;

    tc -> memberNames_.length(members.length());
    tc -> memberTypes_.length(members.length());

    for(ULong i = 0 ; i < members.length() ; i++)
    {
	tc -> memberNames_[i] = members[i].name;
	tc -> memberTypes_[i] = members[i].type;
    }

    TypeCode::_OB_embedRecTC(tc);
    return tc._retn();
}

TypeCode_ptr
ORB::create_union_tc(const char* id,
		     const char* name,
		     TypeCode_ptr discriminator_type,
		     const UnionMemberSeq& members)
{
    assert_nca(id, NCANullString);
    assert_nca(name, NCANullString);

    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_union;
    tc -> id_ = id;
    tc -> name_ = name;
    tc -> discriminatorType_ = TypeCode::_duplicate(discriminator_type);

#ifndef NDEBUG
    TypeCode_var tc1 = GetOrigType(discriminator_type);
    TCKind k = tc1 -> kind();
    
    assert_nca(k == tk_short || k == tk_ushort ||
 	       k == tk_long || k == tk_ulong ||
 	       k == tk_char || k == tk_boolean ||
 	       k == tk_enum, NCADiscType);
#endif
    
    tc -> labels_.length(members.length());
    tc -> memberNames_.length(members.length());
    tc -> memberTypes_.length(members.length());

    for(ULong i = 0 ; i < members.length() ; i++)
    {
#ifndef NDEBUG
	TypeCode_var tc2 = members[i].label.type();
 	if(tc2 -> kind() != tk_octet)
 	    assert_nca(tc1 -> equivalent(tc2), NCADiscType);
#endif

	tc -> labels_[i] = members[i].label;
	tc -> memberNames_[i] = members[i].name;
	tc -> memberTypes_[i] = members[i].type;
    }

    TypeCode::_OB_embedRecTC(tc);
    return tc._retn();
}

TypeCode_ptr
ORB::create_enum_tc(const char* id,
		    const char* name,
		    const EnumMemberSeq& members)
{
    assert_nca(id, NCANullString);
    assert_nca(name, NCANullString);

    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_enum;
    tc -> id_ = id;
    tc -> name_ = name;

    tc -> memberNames_.length(members.length());

    for(ULong i = 0 ; i < members.length() ; i++)
    {
	tc -> memberNames_[i] = members[i];
    }

    return tc._retn();
}

TypeCode_ptr
ORB::create_alias_tc(const char* id,
		     const char* name,
		     TypeCode_ptr original_type)
{
    assert_nca(id, NCANullString);
    assert_nca(name, NCANullString);

    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_alias;
    tc -> id_ = id;
    tc -> name_ = name;

    tc -> contentType_ = TypeCode::_duplicate(original_type);

    return tc._retn();
}

TypeCode_ptr
ORB::create_exception_tc(const char* id,
			 const char* name,
			 const StructMemberSeq& members)
{
    assert_nca(id, NCANullString);
    assert_nca(name, NCANullString);

    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_except;
    tc -> id_ = id;
    tc -> name_ = name;

    tc -> memberNames_.length(members.length());
    tc -> memberTypes_.length(members.length());

    for(ULong i = 0 ; i < members.length() ; i++)
    {
	tc -> memberNames_[i] = members[i].name;
	tc -> memberTypes_[i] = members[i].type;
    }

    TypeCode::_OB_embedRecTC(tc);
    return tc._retn();
}

TypeCode_ptr
ORB::create_interface_tc(const char* id, const char* name)
{
    assert_nca(id, NCANullString);
    assert_nca(name, NCANullString);

    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_objref;
    tc -> id_ = id;
    tc -> name_ = name;

    return tc._retn();
}

TypeCode_ptr
ORB::create_string_tc(ULong bound)
{
    if(bound == 0)
	return TypeCode::_duplicate(_tc_string);

    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_string;
    tc -> length_ = bound;

    return tc._retn();
}

TypeCode_ptr
ORB::create_wstring_tc(ULong bound)
{
    if(bound == 0)
	return TypeCode::_duplicate(_tc_wstring);

    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_wstring;
    tc -> length_ = bound;

    return tc._retn();
}

TypeCode_ptr
ORB::create_fixed_tc(UShort digits, Short scale)
{
    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_fixed;
    tc -> fixedDigits_ = digits;
    tc -> fixedScale_ = scale;

    return tc._retn();
}

TypeCode_ptr
ORB::create_sequence_tc(ULong bound, TypeCode_ptr element_type)
{
    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_sequence;
    tc -> length_ = bound;
    tc -> contentType_ = TypeCode::_duplicate(element_type);

    return tc._retn();
}

TypeCode_ptr
ORB::create_recursive_sequence_tc(ULong /*bound*/, ULong /*offset*/)
{
    assert_nca(false, NCADeprecated);
    return TypeCode::_nil();
}

TypeCode_ptr
ORB::create_array_tc(ULong length, TypeCode_ptr element_type)
{
    TypeCode_var tc = new TypeCode;
    
    tc -> kind_ = tk_array;
    tc -> length_ = length;
    tc -> contentType_ = TypeCode::_duplicate(element_type);
    
    return tc._retn();
}

TypeCode_ptr
ORB::create_value_tc(const char* id,
		     const char* name,
		     ValueModifier type_modifier,
		     TypeCode_ptr concrete_base,
		     const ValueMemberSeq& members)
{
    assert_nca(id, NCANullString);
    assert_nca(name, NCANullString);

    TypeCode_var tc = new TypeCode;
    
    tc -> kind_ = tk_value;
    tc -> id_ = id;
    tc -> name_ = name;
    tc -> typeModifier_ = type_modifier;
    tc -> concreteBaseType_ = TypeCode::_duplicate(concrete_base);

    tc -> memberNames_.length(members.length());
    tc -> memberTypes_.length(members.length());
    tc -> memberVisibility_.length(members.length());

    for(ULong i = 0 ; i < members.length() ; i++)
    {
	tc -> memberNames_[i] = members[i].name;
	tc -> memberTypes_[i] = members[i].type;
	tc -> memberVisibility_[i] = members[i].access;
    }

    TypeCode::_OB_embedRecTC(tc);
    return tc._retn();
}

TypeCode_ptr
ORB::create_boxed_value_tc(const char* id,
			   const char* name,
			   TypeCode_ptr boxed_type)
{
    assert_nca(id, NCANullString);
    assert_nca(name, NCANullString);

    TypeCode_var tc = new TypeCode;
    
    tc -> kind_ = tk_value_box;
    tc -> id_ = id;
    tc -> name_ = name;
    tc -> contentType_ = TypeCode::_duplicate(boxed_type);
    
    return tc._retn();
}

TypeCode_ptr
ORB::create_native_tc(const char* id, const char* name)
{
    assert_nca(id, NCANullString);
    assert_nca(name, NCANullString);

    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_native;
    tc -> id_ = id;
    tc -> name_ = name;

    return tc._retn();
}

TypeCode_ptr
ORB::create_recursive_tc(const char* id)
{
    assert_nca(id, NCANullString);

    TypeCode_var tc = new TypeCode;

    tc -> recId_ = id;

    return tc._retn();
}

TypeCode_ptr
ORB::create_abstract_interface_tc(const char* id, const char* name)
{
    assert_nca(id, NCANullString);
    assert_nca(name, NCANullString);

    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_abstract_interface;
    tc -> id_ = id;
    tc -> name_ = name;

    return tc._retn();
}

ORB::ObjectIdList*
ORB::list_initial_services()
{
    ObjectIdList* list = new ObjectIdList;

#ifdef HAVE_JTC
    JTCSynchronized sync(*services_);
#endif
    
    InitialServiceMap::Enumerator keys = services_ -> keys();
    for(ULong i = 0 ; i < keys.length() ; i++)
	list -> append(keys[i]);

    return list;
}

Object_ptr
ORB::resolve_initial_references(const char* identifier)
{
    assert_nca(identifier, NCANullString);

    Object_var obj;

    {
#ifdef HAVE_JTC
	JTCSynchronized sync(*services_);
#endif
    
	if(!services_ -> get(identifier, obj))
	    throw InvalidName();
    }
    
    //
    // If the service is the RootPOA and it hasn't yet been
    // initialized create it. We could put in some automatic method
    // here for late binding of objects at some later point.
    //
    if(is_nil(obj) && strcmp(identifier, "RootPOA") == 0)
    {
        //
        // This initializes the root POA, if necessary
        //
        obj = OARoot::instance() -> getRootPOA();
    }

    //
    // If the object is an internal l-c object, return the object
    //
    if(obj -> _OB_isInternalLocal())
        return obj._retn();
    
    //
    // If the object is remote, return a new object with the current
    // set of policies applied.
    //
    return obj -> _set_policy_overrides(policies_, SET_OVERRIDE);
}

Boolean
ORB::get_service_information(ServiceType, ServiceInformation_out)
{
    return false;
}

Boolean
ORB::work_pending()
{
    return OARoot::instance() -> workPending();
}

void
ORB::perform_work()
{
    OARoot::instance() -> performWork();
}

void
ORB::shutdown(Boolean wait_for_completion)
{
    OARoot::instance() -> shutdown(wait_for_completion);
}

void
ORB::run()
{
    OARoot::instance() -> run();
}

Policy_ptr
ORB::create_policy(PolicyType policy_type, const Any& val)
{
    //
    // I wanted a case stmt here, but egcs complains: case label
    // `OB_CONNECTION_REUSE_POLICY' does not reduce to an integer
    // constant
    //
    if(policy_type == OB::CONNECTION_REUSE_POLICY)
    {
	Boolean b;
	if(val >>= Any::to_boolean(b))
        {
	    return new OB::ConnectionReusePolicy_impl(b);
        }
        throw PolicyError(BAD_POLICY_TYPE);
    }
    
    if(policy_type == OB::PROTOCOL_POLICY)
    {
        OCI::ProfileId tag;
        if(val >>= tag)
        {
            return new OB::ProtocolPolicy_impl(tag);
        }
        throw PolicyError(BAD_POLICY_TYPE);
    }
    
    if(policy_type == OB::RECONNECT_POLICY)
    {
        Boolean b;
        if(val >>= Any::to_boolean(b))
        {
            return new OB::ReconnectPolicy_impl(b);
        }
        throw PolicyError(BAD_POLICY_TYPE);
    }
    
    if(policy_type == OB::TIMEOUT_POLICY)
    {
        ULong t;
        if(val >>= t)
        {
            return new OB::TimeoutPolicy_impl(t);
        }
        throw PolicyError(BAD_POLICY_TYPE);
    }
    
    throw PolicyError(BAD_POLICY);
    return Policy::_nil(); // Some compilers need this
}

ValueFactory
ORB::register_value_factory(const char* id, ValueFactory factory)
{
    assert_nca(id, NCANullString);
    assert_nca(factory, NCANullValue);

    ValueFactoryBase_var old;
    valueFactoryBaseMap_ -> get(id, old);

    factory -> _add_ref();
    valueFactoryBaseMap_ -> put(id, factory);

    //
    // Caller assumes ownership of previously registered factory (if any)
    //
    return old._retn();
}

void
ORB::unregister_value_factory(const char* id)
{
    assert_nca(id, NCANullString);
    valueFactoryBaseMap_ -> remove(id);
}

ValueFactory
ORB::lookup_value_factory(const char* id)
{
    assert_nca(id, NCANullString);
    
    //
    // Caller must invoke _remove_ref
    //
    ValueFactoryBase_var factory;
    if(valueFactoryBaseMap_ -> get(id, factory))
        return factory._retn();

    //XXX - this correct?
    return 0;
}

void
ORB::init_servers()
{
    OARoot::instance() -> initServers();
}

OCI::ConFactoryRegistry_ptr
ORB::get_con_factory_registry()
{
    return OCI::ConFactoryRegistry::_duplicate(registry_);
}

Object_ptr
ORB::get_inet_object(const char* host, UShort port,
		     const char* name, const char* id)
{
    IOP::IOR_var ior = IIOPCreateIOR(host, port, name, id);
    return _OB_createObject(ior);
}

ORB::ConcModel
ORB::conc_model()
{
    return concModel_;
}

void
ORB::conc_model(ORB::ConcModel model)
{
    concModel_ = model;
}

ULong
ORB::native_cs() const
{
    return native_cs_;
}

void
ORB::native_cs(ULong id)
{
    native_cs_ = id;
}

ULong
ORB::native_wcs() const
{
    return native_wcs_;
}

void
ORB::native_wcs(ULong id)
{
    native_wcs_ = id;
}

void
ORB::add_initial_reference(const char* name, Object_ptr obj)
{
    _OB_addInitialService(name, obj);
}

//
// Set the IIOP port from the port number defined in the IIOP profile
// in an Object
//
void
ORB::set_iiop_port_from_object(Object_ptr obj)
{
    //
    // If the object is nil, do nothing.
    //
    if(is_nil(obj))
	return;

    //
    // Get the OCI connector info.
    //
    OCI::ConnectorInfo_var info = obj -> _get_oci_connector_info();

    //
    // Narrow to the IIOP connector info.
    //
    OCI::IIOP::ConnectorInfo_var iiopInfo =
	OCI::IIOP::ConnectorInfo::_narrow(info);

    //
    // If this is an IIOP connector info then rip out the remote port
    // number, and set the ooc.boa.port property to this port.
    //
    if(!is_nil(iiopInfo))
    {
	//
	// Allocate 10 bytes so that we don't need to reallocate when
	// we add the port.
	//
	String_var s = string_alloc(10);
	s[0] = '\0';
	s += iiopInfo -> remote_port();

	//
	// Set the property.
	//
	Properties::instance() -> setProperty("ooc.boa.port", s);
    }
}

//XXX OCI ORB CONC
ORB::ServerConcModel
ORB::server_conc_model()
{
    return OARoot::serverConcModel();
}
void
ORB::server_conc_model(ORB::ServerConcModel conc)
{
    OARoot::serverConcModel(conc);
}

ULong
ORB::conc_model_thread_pool()
{
    return OARoot::serverThreadPool();
}

void
ORB::conc_model_thread_pool(ULong p)
{
    OARoot::serverThreadPool(p);
}

OCI::AccRegistry_ptr
ORB::get_acc_registry()
{
    return OARoot::instance() -> getAccRegistry();
}
//XXX OCI ORB CONC

ORB_ptr
ORB::_OB_instance()
{
    return ORB::_duplicate(orb_);
}

ORB_ptr
ORB::_OB_instance_nodup()
{
    return orb_;
}

char*
ORB::_OB_IORToString(const IOP::IOR& ior)
{
    CountStream count;
    count.count_boolean(); // Endian
    IOP::IOR::_OB_marshalCount(ior, count);
    
    Buffer buf(count.count());
    OutputStream out(buf, 0);
    
    out.write_boolean(Endian);
    IOP::IOR::_OB_marshal(ior, out);
    
    String_var s = OctetsToAscii(buf.data, count.count());
    
    static const char* prefix = "IOR:";
    char* result = string_alloc(strlen(prefix) + strlen(s));
    strcpy(result, prefix);
    strcat(result, s);

    return result;
}

IOP::IOR*
ORB::_OB_stringToIOR(const char* s)
{
    assert_nca(s, NCANullString);

    IOP::IOR_var ior = new IOP::IOR;

    if(strncmp(s, "iiop://", 7) == 0 || strncmp(s, "IIOP://", 7) == 0)
    {
	String_var str;
	const char* token;
	UShort port;
	String_var host, name;

	str = string_dup(s + 7);
	token = strtok(str.inout(), ":");
	if(!token)
	    throw INV_OBJREF();
	host = string_dup(token);
	token = strtok(0, "/");
	if(!token)
	    throw INV_OBJREF();
	port = atoi(token);
	token = strtok(0, "\n");
	if(!token)
	    throw INV_OBJREF();
	name = string_dup(token);

	ior = IIOPCreateIOR(host, port, name, "");
    }
    else if(strncmp(s, "IOR:", 4) == 0)
    {
	s += 4;

	if(strlen(s) % 2)
	    throw INV_OBJREF();
	
	Octet* oct = AsciiToOctets(s);
        InputStream in(oct, 0, strlen(s) / 2, false);
        in._OB_readEndian();
	IOP::IOR::_OB_unmarshal(ior.inout(), in);
	
	delete [] oct;
    }
    else
    {
	throw INV_OBJREF();
    }

    return ior._retn();
}

Object_ptr
ORB::_OB_createObject(const IOP::IOR& ior)
{
    assert_nca(ior.type_id.in(), NCANullString);

    //
    // Check for nil object reference
    //
    if(strlen(ior.type_id) == 0 && ior.profiles.length() == 0)
	return Object::_nil();

    Object_var p = new Object();
    p -> _OB_ior(ior);
    p -> _OB_policies(policies_);

    return p._retn();
}

void
ORB::_OB_changeStub(StubImplBase* p, const IOP::IOR& ior)
{
    assert_nca(ior.type_id.in(), NCANullString);

    if(strlen(ior.type_id) == 0 && ior.profiles.length() == 0)
	throw INV_OBJREF(); // I can't change an object to a nil object

#ifdef HAVE_JTC
    JTCSynchronized synchronized(clientsMutex_);
#endif

    Client_var clt;

    OCI::AccRegistry_var registry =
        ORB::_OB_instance_nodup() -> get_acc_registry();
    OCI::ObjectKey_var key = registry -> is_local(ior);
    if(key -> length() > 0) // Points to local servant
    {
        clt = new CollocatedClient();

        //
        // XXX - the POA key, and the _OB_key are always going to be
        // the same thing. We can just trash _OB_POAkey.
        //
        p -> _OB_POAkey(key);
    }
    else
    {
        //
        // Find reuse policy among policies
        //
        const PolicyList& policies = p -> _OB_policies();
        OB::ConnectionReusePolicy_var reusePolicy;
        for(ULong i = 0; i < policies.length(); i++)
            if(policies[i] -> policy_type() == OB::CONNECTION_REUSE_POLICY)
                reusePolicy = OB::ConnectionReusePolicy::_narrow(policies[i]);
        
        if(is_nil(reusePolicy) || reusePolicy -> value())
        {
            //
            // Try to reuse clients
            //
            for(ULong i = 0 ; i < clients_.length() ; i++)
            {
                //
                // Check reuse flag
                //
                if(clients_[i] -> reuseClient())
                {
                    OCI::Connector_var connector = clients_[i] -> connector();
                    key = connector -> is_usable_with_policies(ior, policies);
                    if(key -> length() > 0)
                        clt = Client::_duplicate(clients_[i]);
                }
            }
        }
        
        if(is_nil(clt))
        {
            //
            // Create a new connector
            //
            OCI::ConFactory_var factory =
                registry_ -> get_factory_with_policies(ior, policies);
            if(is_nil(factory))
                throw NO_RESOURCES();
            OCI::Connector_var connector =
                factory -> create_with_policies(ior, policies);
            key = connector -> is_usable_with_policies(ior, policies);
            assert(key -> length() > 0);
            
            //
            // Create a new client
            //
            clt = new GIOPClient(connector);

	    // XXX Preliminary, don't always use local converter
	    CONV_FRAME::CodeSetComponentInfo_var info =
		CodeSetUtil::getCodeSetComponentInfoFromIOR(&ior);
	    CodeConverters codeConverters;
	    codeConverters.charConverter = LocalCodeConverter::instance();
	    //
	    // wchar conversion possible?
	    //
	    if(info -> ForWcharData.native_code_set == 0 &&
	       info -> ForWcharData.conversion_code_sets.length() == 0)
		codeConverters.wcharConverter = 0;
	    else
		codeConverters.wcharConverter = LocalCodeConverter::instance();
	    clt -> setCodeConverters(codeConverters);

            clients_.append(clt);
            if(!is_nil(reusePolicy) && !reusePolicy -> value())
                clt -> reuseClient(false);
        }
    }

    assert(!is_nil(clt));
    
    //
    // Setup the stub
    //
    p -> _OB_client(clt);
    p -> _OB_ior(ior);

    //
    // Must be last, as the other operations invalidate the key
    //
    p -> _OB_key(key.in());
}

StubImplBase_ptr
ORB::_OB_getStubImpl(Object_ptr o)
{
    StubImplBase* base = 0;
    while(true)
    {
        const IOP::IOR& ior = o -> _OB_ior();
        
        //
        // If the root POA doesn't exist, or the IOR is remote then this is a
        // remote object.                                                    
        //
        if(!OARoot::instance() -> hasRootPOA())
        {
            base = o -> _OB_createStubImpl();
        }
        else
        {
            OCI::AccRegistry_var registry =
                ORB::_OB_instance_nodup() -> get_acc_registry();
            OCI::ObjectKey_var key = registry -> is_local(ior);
            
            //
            // Object is local
            //
            if(key -> length() > 0)
            {
                PortableServer::POA_var theRootPOA =
                    OARoot::instance() -> getRootPOA();
                PortableServer::POA_impl* rootPOA =
                    dynamic_cast<PortableServer::POA_impl*>(theRootPOA.in());
                PortableServer::POA_impl* poa =
                    rootPOA -> _OB_locatePOA(key.in());
                if(poa != 0)
                {
                    PortableServer::POA_var dummy = poa;
                    try
                    {
                        base = poa -> _OB_createStubImplBase(key.in(),
                                                             ior.type_id);
                    }
                    catch(const StubImplForward& fwd)
                    {
                        o -> _OB_forward(fwd.ior);
                        continue;
                    }
                }
                else
                {
                    //
                    // XXX In this case there is no POA for a local
                    // servant. This is an CORBA::OBJECT_NOT_EXIST
                    // exception.
                    //
                    throw OBJECT_NOT_EXIST();
                }
            }
            
            //
            // If the object is local and the servant is DSI, or the
            // object is remote then we must ask the object itself to
            // create the stub.
            //
            if(base == 0)
                base = o -> _OB_createStubImpl();
        }
        
        //
        // Change the stub implementation
        //
        _OB_changeStub(base, ior);
        
        //
        // Copy the policies from the object reference
        //
        base -> _OB_policies(o -> _OB_policies());
        break;
    }
        
    return base;
}

void
ORB::_OB_incClientUsage(Client_ptr client)
{
#ifdef HAVE_JTC
    JTCSynchronized synchronized(clientsMutex_);
#endif

    client -> incUsage();
}

void
ORB::_OB_decClientUsage(Client_ptr client)
{
#ifdef HAVE_JTC
    JTCSynchronized synchronized(clientsMutex_);
#endif

    bool inUse = client -> decUsage();

    if(!inUse)
    {
	for(ULong i = 0 ; i < clients_.length() ; i++)
	    if(clients_[i].in() == client)
	    {
		client -> destroy();
		clients_.remove(i);
		return;
	    }
	
	//
	// Collocated clients won't be in the list (why?)
	//
        client -> destroy();
    }
}

Boolean
ORB::_OB_equivalent(const IOP::IOR& ior1, const IOP::IOR& ior2)
{
    return registry_ -> equivalent(ior1, ior2);
}

ULong
ORB::_OB_hash(const IOP::IOR& ior, ULong maximum)
{
    return registry_ -> hash(ior, maximum);
}

void
ORB::_OB_addInitialService(const char* name, const char* iorString,
                           bool displayWarning)
{
    assert_nca(name, NCANullString);
    assert_nca(iorString, NCANullString);

    Object_var obj = string_to_object(iorString);
    _OB_addInitialService(name, obj, displayWarning);
}

void
ORB::_OB_addInitialService(const char* name, Object_ptr p, bool displayWarning)
{
    assert_nca(name, NCANullString);

#ifdef HAVE_JTC
    JTCSynchronized sync(*services_);
#endif
    if(services_ -> contains(name) && displayWarning)
    {
        MessageViewer* viewer = MessageViewer::instance();
        String_var str =
        string_dup("ORB_init: initial service `");
        str += name;
        str += "' was overridden";
        viewer -> warning(str);
        
        return;
    }
    services_ -> put(name, Object::_duplicate(p));
}

OCI::Current_impl*
ORB::_OB_OCICurrent() const
{
    return ociCurrent_;
}

// ----------------------------------------------------------------------
// ORB Initialization
// ----------------------------------------------------------------------

/* See IIOPInit.cpp */
