// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/POAServantBase.h>
#include <OB/PolicyManager.h>

#include <ORBPolicyManager_impl.h>
#include <PolicyValidator.h>

using namespace CORBA;

// ----------------------------------------------------------------------
// ORBPolicyManager_impl constructor and destructor
// ----------------------------------------------------------------------

ORBPolicyManager_impl::ORBPolicyManager_impl(PolicyList& policies)
    : policies_(policies)
{
}

ORBPolicyManager_impl::~ORBPolicyManager_impl()
{
}

// ----------------------------------------------------------------------
// ORBPolicyManager_impl public member implementation
// ----------------------------------------------------------------------

Policy_ptr
ORBPolicyManager_impl::get_policy_override(PolicyType policy_type)
    throw(SystemException)
{
    ULong policiesLen = policies_.length();
    for(ULong i = 0; i < policiesLen; ++i)
	if(policies_[i] -> policy_type() == policy_type)
	    return Policy::_duplicate(policies_[i]);

    return Policy::_nil();
}

PolicyList*
ORBPolicyManager_impl::get_specific_policy_overrides(
    const PolicyTypeSeq& types)
    throw(SystemException)
{
    ULong len = policies_.length();
    ULong typesLen = types.length();

    PolicyList_var policies = new PolicyList(typesLen);

    for(ULong i = 0 ; i < typesLen; ++i)
    {
        PolicyType currType = types[i];
        for(ULong j = 0 ; j < len ; ++j)
            if(policies_[j] -> policy_type() == currType)
                policies -> append(policies_[j]);
    }

    return policies._retn();
}

PolicyList*
ORBPolicyManager_impl::get_all_policy_overrides()
    throw(SystemException)
{
    return new PolicyList(policies_);
}

void
ORBPolicyManager_impl::add_policy_overrides(const PolicyList& np)
    throw(PolicyManager::InvalidPolicies,
          SystemException)
{
    PolicyList newPolicies = policies_;
    
    ULong npLen  = np.length();
    for(ULong i = 0 ; i < npLen ; ++i)
    {
        ULong len = newPolicies.length();
        ULong j;
        
        for(j = 0 ; j < len ; ++j)
            if(newPolicies[j] -> policy_type() == np[i] -> policy_type())
                break;
        
        if(j < len)
            newPolicies[j] = np[i];
        else
            newPolicies.append(np[i]);
    }
    PolicyValidator::validatePolicies(newPolicies);
    
    policies_ = newPolicies;
}

void
ORBPolicyManager_impl::remove_policy_overrides(const PolicyTypeSeq& ts)
    throw(SystemException)
{
    ULong typesLen = ts.length();

    for(ULong i = 0 ; i < typesLen; ++i)
    {
        PolicyType currType = ts[i];
        ULong len = policies_.length();
        for(ULong j = 0 ; j < len ; ++j)
        {
            if(policies_[j] -> policy_type() == currType)
            {
                policies_.remove(j);
                return;
            }
        }
    }
}

void
ORBPolicyManager_impl::replace_all_policy_overrides(const PolicyList& policies)
    throw(PolicyManager::InvalidPolicies,
          SystemException) 
{
    PolicyValidator::validatePolicies(policies);
    policies_ = policies;
}
