// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/TemplateI.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/Any.h>
#include <OB/OBV.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/IFR2.h>
#include <OB/Policy.h>
#include <OB/ORB.h>
#include <OB/OCI.h>
#include <OB/OBPolicies.h>
#include <OB/NativeTypes.h>
#include <OB/POAInterface.h>
#include <OB/POAServantBase.h>
#include <OB/AbstractBase.h>
#include <OB/Stream.h>

#include <PolicyValidator.h>

using namespace CORBA;
using namespace OB;

//
// The _ob_ids_ for Object
//
const char* Object::_ob_ids_[] =
{
    "IDL:omg.org/CORBA/Object:1.0",
    0
};

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(CORBA::Object_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(CORBA::Object_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// Template instantiations
// ----------------------------------------------------------------------

#ifndef HAVE_NO_EXPLICIT_TEMPLATES
template class OB::ObjVar< Object >;
template class OB::ObjForSeq< Object >;
#else
#ifdef HAVE_PRAGMA_DEFINE
#pragma define(OB::ObjVar< Object >)
#pragma define(OB::ObjForSeq< Object >)
#endif
#endif

// ----------------------------------------------------------------------
// Object protected members
// ----------------------------------------------------------------------

StubImplBase_ptr
Object::_OB_getStubImpl()
{
#ifdef HAVE_JTC
    JTCSynchronized sync(_ob_implMutex_);
#endif
    if(is_nil(_ob_stubImpl_))
    {
	ORB_ptr orb = ORB::_OB_instance_nodup();
	if(is_nil(orb))
	    throw INITIALIZE("ORB not initialized");
	_ob_stubImpl_ = orb -> _OB_getStubImpl(this);
	assert(_ob_stubImpl_.in() != 0);
    }
    return StubImplBase::_duplicate(_ob_stubImpl_);
}

void
Object::_OB_clearStubImpl()
{
#ifdef HAVE_JTC
    JTCSynchronized sync(_ob_implMutex_);
#endif

    _ob_stubImpl_ = StubImplBase::_nil();
}

// ----------------------------------------------------------------------
// Object constructor and destructor
// ----------------------------------------------------------------------

Object::Object()
    : _ob_locality_(false),
      _ob_stubImpl_(StubImplBase::_nil())
{
}

Object::~Object()
{
}

// ----------------------------------------------------------------------
// Object public member implementation
// ----------------------------------------------------------------------

InterfaceDef_ptr
Object::_get_interface()
{
    if(_OB_isInternalLocal())
        throw NO_IMPLEMENT();
    
    while(true)
    {
        StubImplBase_var impl = _OB_getStubImpl();

        try
        {
            return impl -> _get_interface();
        }
        catch(StubImplForward& fwd)
        {
            _OB_forward(fwd.ior);
        }
        catch(StubImplChanged&)
        {
            _OB_clearStubImpl();
        }
    }
}

Boolean
Object::_is_a(const char* _ob_id)
{
#ifdef HAVE_JTC
    JTCSynchronized sync(_ob_iorMutex_);
#endif
    assert_nca(_ob_id, NCANullString);

    //
    // Check IDL:omg.org/CORBA/Object:1.0
    //
    if(strcmp(_ob_id, _ob_ids_[0]) == 0)
        return true;
    
    //
    // Check the type_id in the IOR
    //
    if(strcmp(_ob_id, _ob_ior_.type_id) == 0)
        return true;

    //
    // Check all ids of ObjectImpl
    //
    const char** ids = _OB_ids();
    assert(ids != 0);
    for(ULong i = 0; ids[i] != 0; i++)
        if(strcmp(_ob_id, ids[i]) == 0)
            return true;

    //
    // If this is an internal l-c object, return false now
    //
    if(_OB_isInternalLocal())
        return false;
    
    while(true)
    {
        StubImplBase_var impl = _OB_getStubImpl();
        try
        {
            return impl -> _is_a(_ob_id);
        }
        catch(StubImplForward& fwd)
        {
            _OB_forward(fwd.ior);
        }
        catch(StubImplChanged&)
        {
            _OB_clearStubImpl();
        }
    }
}

Boolean
Object::_non_existent()
{
    if(_OB_isInternalLocal())
        throw NO_IMPLEMENT();

    while(true)
    {
        StubImplBase_var impl = _OB_getStubImpl();
        try
        {
            return impl -> _non_existent();
        }
        catch(StubImplForward& fwd)
        {
            _OB_forward(fwd.ior);
        }
        catch(StubImplChanged&)
        {
            _OB_clearStubImpl();
        }
    }
}

Boolean
Object::_is_equivalent(Object_ptr p)
{
    //
    // Check locality constrainted objects
    //
    if(_OB_isInternalLocal() || p -> _OB_isInternalLocal())
    {
        if(_OB_isInternalLocal() && p -> _OB_isInternalLocal())
        {
            return this == p;
        }
        return false;
    }
    
    //
    // Compare the IORs
    //
    ORB_ptr orb = ORB::_OB_instance_nodup();
    if(is_nil(orb))
	throw INITIALIZE("ORB not initialized");
    
    return orb -> _OB_equivalent(_OB_ior(), p -> _OB_ior());
}

ULong
Object::_hash(ULong maximum)
{
    if(_OB_isInternalLocal())
    {
        return ((ULong) this / 256) % (maximum + 1);
    }
    
    //
    // Get hash via the ORB
    //
    ORB_ptr orb = ORB::_OB_instance_nodup();
    if(is_nil(orb))
        throw INITIALIZE("ORB not initialized");
    return orb -> _OB_hash(_OB_ior(), maximum);
}

/*
 *
 * _request() and _create_request() implementations in DII.cpp
 *
 */

Policy_ptr
Object::_get_policy(PolicyType policy_type)
{
    return _get_effective_policy_override(policy_type);
}

Policy_ptr
Object::_get_effective_policy_override(PolicyType policy_type)
{
    Policy_var policy = _OB_getPolicy(policy_type);
    if(is_nil(policy))
        throw INV_POLICY();
    return policy._retn();
}

Object_ptr
Object::_set_policy_overrides(const PolicyList& np, SetOverrideType set_add)
{
    if(_OB_isInternalLocal())
        throw BAD_OPERATION();
    
    //
    // Create the new policy list
    //
    PolicyList newPolicies;
    if(set_add == SET_OVERRIDE)
    {
        newPolicies = np;
    }
    else
    {
        newPolicies = _ob_policies_;

        ULong npLen = np.length();
	for(ULong i = 0 ; i < npLen ; ++i)
	{
	    ULong len = newPolicies.length();
	    ULong j;

	    for(j = 0 ; j < len ; ++j)
		if(newPolicies[j] -> policy_type() == np[i] -> policy_type())
		    break;

	    if(j < len)
		newPolicies[j] = np[i];
	    else
		newPolicies.append(np[i]);
	}
    }
    PolicyValidator::validatePolicies(newPolicies);

    //
    // Create a new object
    //
    Object_var p = new Object;

    //
    // Copy the data from this object reference
    //
    p -> _OB_copyFrom(this);

    //
    // Override the policies
    //
    p -> _OB_policies(newPolicies);

    return p._retn();
}

PolicyList*
Object::_get_policy_overrides(const PolicyTypeSeq& types)
{
    ULong len = _ob_policies_.length();
    ULong typesLen = types.length();

    PolicyList_var policies = new PolicyList(typesLen);

    for(ULong i = 0 ; i < typesLen; ++i)
    {
        PolicyType currType = types[i];
        for(ULong j = 0 ; j < len ; ++j)
            if(_ob_policies_[j] -> policy_type() == currType)
                policies -> append(_ob_policies_[j]);
    }
    return policies._retn();
}

Boolean
Object::_validate_policies(const PolicyList&)
{
    //
    // For now all policy combinations are valid.
    //
    return TRUE;
}

OCI::TransportInfo_ptr
Object::_get_oci_transport_info()
{
    if(_OB_isInternalLocal())
        throw NO_IMPLEMENT();

    while(true)
    {
        StubImplBase_var impl = _OB_getStubImpl();
        try
        {
            return impl -> _get_oci_transport_info();
        }
        catch(StubImplForward& fwd)
        {
            _OB_forward(fwd.ior);
        }
        catch(StubImplChanged&)
        {
            _OB_clearStubImpl();
        }
    }
}

OCI::ConnectorInfo_ptr
Object::_get_oci_connector_info()
{
    if(_OB_isInternalLocal())
        throw NO_IMPLEMENT();

    while(true)
    {
        StubImplBase_var impl = _OB_getStubImpl();
        try
        {
            return impl -> _get_oci_connector_info();
        }
        catch(StubImplForward& fwd)
        {
            _OB_forward(fwd.ior);
        }
        catch(StubImplChanged&)
        {
            _OB_clearStubImpl();
        }
    }
}

const char**
Object::_OB_ids() const
{
    return _ob_ids_;
}

const char**
Object::_OB_staticIds()
{
    return _ob_ids_;
}

const PolicyList&
Object::_OB_policies()
{
    return _ob_policies_;
}

//
// This method sets the policies to the list represented by np. It
// also updates internal state variables from the set of policies.
//
void
Object::_OB_policies(const PolicyList& np)
{
    _ob_policies_ = np;
}

//
//XXX
// Currently this method is only used in one member method. Unless
// this changes it should be removed.
//
Policy_ptr
Object::_OB_getPolicy(PolicyType policy_type)
{
    ULong policiesLen = _ob_policies_.length();
    for(ULong i = 0; i < policiesLen; ++i)
	if(_ob_policies_[i] -> policy_type() == policy_type)
	    return Policy::_duplicate(_ob_policies_[i]);
    return Policy::_nil();
}

const char*
Object::_OB_typeId() const
{
    return _ob_ids_[0];
}

//
// This method does not need to be protected since it's always called
// on initialization of the Object.
//
void
Object::_OB_copyFrom(Object_ptr p)
{
    _ob_ior_ = p -> _ob_ior_;
    _ob_policies_ = p -> _ob_policies_;
    _ob_locality_ = p -> _ob_locality_;
}

void
Object::_OB_forward(const IOP::IOR& ior)
{
#ifdef HAVE_JTC
    JTCSynchronized sync(_ob_iorMutex_);
#endif

    //
    // Can't location forward internal l-c objects
    //
    assert(!_OB_isInternalLocal());
    ORB_ptr orb = ORB::_OB_instance_nodup();
    if(is_nil(orb))
	throw INITIALIZE("ORB not initialized");

    //
    // Change the IOR
    //
    _ob_ior_ = ior;

    //
    // We need to re-get the stub implementation
    //
    _OB_clearStubImpl();
}

void
Object::_OB_unmarshal(Object_ptr& v, InputStream& in)
{
    Object_var old = v;
    v = in.read_Object();
}

StubImplBase*
Object::_OB_createStubImpl()
{
    return new ObjectStubImpl();
}

StubImplBase*
Object::_OB_getImplBase()
{
    return _OB_getStubImpl();
}

//
// Determine if this proxy is to an internal locality constrained
// servant
//
Boolean
Object::_OB_isInternalLocal()
{
    return _ob_locality_;
}

//
// This will be overridden by interfaces that derive from an
// abstract interface
//
AbstractBase_ptr
Object::_OB_toAbstractBase()
{
    return AbstractBase::_nil();
}
