// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/TemplateI.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/Any.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/POAServantBase.h>
#include <OB/POAManager.h>
#include <OB/NativeTypes.h>
#include <OB/POAInterface.h> // POA
#include <OB/POAServantLocator.h>

#include <POAManager_impl.h>
#include <POA_impl.h>

using namespace CORBA;
using namespace PortableServer;
using namespace OB;

// ----------------------------------------------------------------------
// Template instantiations
// ----------------------------------------------------------------------
    
#ifndef HAVE_NO_EXPLICIT_TEMPLATES
template class OB::FixSeq< POA_impl*, int >;
#else
#ifdef HAVE_PRAGMA_DEFINE
//XXX - Fix template instantiation stuff
#endif
#endif

// ----------------------------------------------------------------------
// POAManager_impl private member implementation
// ----------------------------------------------------------------------

void
POAManager_impl::stateChange(POAManager::State state)
{
    state_ = state;
    
#ifdef HAVE_JTC
    {
        JTCSynchronized sync(stateChange_);
        stateChange_.notifyAll();
    }
#endif
    
    for(ULong i = 0; i < poas_.length(); ++i)
        poas_[i] -> _OB_stateChange();
}

// ----------------------------------------------------------------------
// POAManager_impl constructor and destructor
// ----------------------------------------------------------------------

POAManager_impl::POAManager_impl()
    : terminating_(false),
      state_(POAManager::HOLDING)
      
{
}

// ----------------------------------------------------------------------
// POAManager_impl public member implementation
// ----------------------------------------------------------------------

void
POAManager_impl::activate()
    throw(POAManager::AdapterInactive, SystemException)
{
    OB_SYNCHRONIZED

    //
    // If the POA manager is in inactive state then raise the
    // AdapterInactive exception
    //
    if(state_ == POAManager::INACTIVE)
        throw POAManager::AdapterInactive();

    stateChange(POAManager::ACTIVE);
}

void
POAManager_impl::hold_requests(Boolean)
    throw(POAManager::AdapterInactive, SystemException)
{
    OB_SYNCHRONIZED
    //
    // If the POA manager is in inactive state then raise the
    // AdapterInactive exception
    //
    if(state_ == POAManager::INACTIVE)
    {
        throw POAManager::AdapterInactive();
    }
    
    stateChange(POAManager::HOLDING);
}

void
POAManager_impl::discard_requests(Boolean)
    throw(POAManager::AdapterInactive, SystemException)
{
    OB_SYNCHRONIZED
    //
    // If the POA manager is in inactive state then raise the
    // AdapterInactive exception
    //
    if(state_ == POAManager::INACTIVE)
        throw POAManager::AdapterInactive();
    stateChange(POAManager::DISCARDING);
}

void
POAManager_impl::deactivate(Boolean etherealize, Boolean wait)
    throw(POAManager::AdapterInactive, SystemException)
{
    OB_SYNCHRONIZED
    //
    // If the POA manager is in inactive state then raise the
    // AdapterInactive exception
    //
    if(state_ == POAManager::INACTIVE)
        throw POAManager::AdapterInactive();
    
    stateChange(POAManager::INACTIVE);

    terminating_ = true;

    for(ULong i = 0; i < poas_.length(); ++i)
	poas_[i] -> destroy(etherealize, wait);
    poas_.length(0);
}

POAManager::State
POAManager_impl::get_state()
    throw(SystemException)
{
    OB_SYNCHRONIZED
    return state_;
}

//
// Register a POA with this POAManager
//
void
POAManager_impl::_OB_add(POA_impl* poa)
{
    OB_SYNCHRONIZED
    poas_.append(poa);
}

//
// Deregister a POA with this POAManager
//
void
POAManager_impl::_OB_remove(POA_impl* poa)
{
    OB_SYNCHRONIZED
    if(terminating_)
	return;

    for(ULong i = 0; i < poas_.length(); ++i)
    {
        if(poas_[i] == poa)
        {
            poas_.remove(i);
            return;
        }
    }
    assert(false);
}

#ifdef HAVE_JTC
void
POAManager_impl::_OB_waitStateChange()
{
    JTCSynchronized sync(stateChange_);
    stateChange_.wait();
}
#endif
