// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/POAServantBase.h>
#include <OB/POAPolicies.h>

#include <POAPolicies_impl.h>

using namespace CORBA;
using namespace PortableServer;

// ----------------------------------------------------------------------
// ThreadPolicy_impl constructor and destructor
// ----------------------------------------------------------------------

ThreadPolicy_impl::ThreadPolicy_impl(ThreadPolicyValue value)
    : value_(value)
{
}

ThreadPolicy_impl::~ThreadPolicy_impl()
{
}

// ----------------------------------------------------------------------
// ThreadPolicy_impl public member implementation
// ----------------------------------------------------------------------

ThreadPolicyValue
ThreadPolicy_impl::value()
    throw(SystemException)
{
    return value_;
}

PolicyType
ThreadPolicy_impl::policy_type()
    throw(SystemException)
{
    return THREAD_POLICY_ID;
}

Policy_ptr
ThreadPolicy_impl::copy()
    throw(SystemException)
{
    return Policy::_duplicate(this);
}

void
ThreadPolicy_impl::destroy()
    throw(SystemException)
{
}

// ----------------------------------------------------------------------
// LifespanPolicy_impl constructor and destructor
// ----------------------------------------------------------------------

LifespanPolicy_impl::LifespanPolicy_impl(LifespanPolicyValue value)
    : value_(value)
{
}

LifespanPolicy_impl::~LifespanPolicy_impl()
{
}

// ----------------------------------------------------------------------
// LifespanPolicy_impl public member implementation
// ----------------------------------------------------------------------

LifespanPolicyValue
LifespanPolicy_impl::value()
    throw(SystemException)
{
    return value_;
}

PolicyType
LifespanPolicy_impl::policy_type()
    throw(SystemException)
{
    return LIFESPAN_POLICY_ID;
}

Policy_ptr
LifespanPolicy_impl::copy()
    throw(SystemException)
{
    return Policy::_duplicate(this);
}

void
LifespanPolicy_impl::destroy()
    throw(SystemException)
{
}

// ----------------------------------------------------------------------
// IdUniquenessPolicy_impl constructor and destructor
// ----------------------------------------------------------------------

IdUniquenessPolicy_impl::IdUniquenessPolicy_impl(IdUniquenessPolicyValue value)
    : value_(value)
{
}

IdUniquenessPolicy_impl::~IdUniquenessPolicy_impl()
{
}

// ----------------------------------------------------------------------
// IdUniquenessPolicy_impl public member implementation
// ----------------------------------------------------------------------

IdUniquenessPolicyValue
IdUniquenessPolicy_impl::value()
    throw(SystemException)
{
    return value_;
}

PolicyType
IdUniquenessPolicy_impl::policy_type()
    throw(SystemException)
{
    return ID_UNIQUENESS_POLICY_ID;
}

Policy_ptr
IdUniquenessPolicy_impl::copy()
    throw(SystemException)
{
    return Policy::_duplicate(this);
}

void
IdUniquenessPolicy_impl::destroy()
    throw(SystemException)
{
}

// ----------------------------------------------------------------------
// IdAssignmentPolicy_impl constructor and destructor
// ----------------------------------------------------------------------

IdAssignmentPolicy_impl::IdAssignmentPolicy_impl(IdAssignmentPolicyValue value)
    : value_(value)
{
}

IdAssignmentPolicy_impl::~IdAssignmentPolicy_impl()
{
}

// ----------------------------------------------------------------------
// IdAssignmentPolicy_impl public member implementation
// ----------------------------------------------------------------------

IdAssignmentPolicyValue
IdAssignmentPolicy_impl::value()
    throw(SystemException)
{
    return value_;
}

PolicyType
IdAssignmentPolicy_impl::policy_type()
    throw(SystemException)
{
    return ID_ASSIGNMENT_POLICY_ID;
}

Policy_ptr
IdAssignmentPolicy_impl::copy()
    throw(SystemException)
{
    return Policy::_duplicate(this);
}

void
IdAssignmentPolicy_impl::destroy()
    throw(SystemException)
{
}

// ----------------------------------------------------------------------
// ImplicitActivationPolicy_impl constructor and destructor
// ----------------------------------------------------------------------

ImplicitActivationPolicy_impl::
ImplicitActivationPolicy_impl(ImplicitActivationPolicyValue value)
    : value_(value)
{
}

ImplicitActivationPolicy_impl::~ImplicitActivationPolicy_impl()
{
}

// ----------------------------------------------------------------------
// ImplicitActivationPolicy_impl public member implementation
// ----------------------------------------------------------------------

ImplicitActivationPolicyValue
ImplicitActivationPolicy_impl::value()
    throw(SystemException)
{
    return value_;
}

PolicyType
ImplicitActivationPolicy_impl::policy_type()
    throw(SystemException)
{
    return IMPLICIT_ACTIVATION_POLICY_ID;
}

Policy_ptr
ImplicitActivationPolicy_impl::copy()
    throw(SystemException)
{
    return Policy::_duplicate(this);
}

void
ImplicitActivationPolicy_impl::destroy()
    throw(SystemException)
{
}

// ----------------------------------------------------------------------
// ServantRetentionPolicy_impl constructor and destructor
// ----------------------------------------------------------------------

ServantRetentionPolicy_impl::
ServantRetentionPolicy_impl(ServantRetentionPolicyValue value)
    : value_(value)
{
}

ServantRetentionPolicy_impl::~ServantRetentionPolicy_impl()
{
}

// ----------------------------------------------------------------------
// ServantRetentionPolicy_impl public member implementation
// ----------------------------------------------------------------------

ServantRetentionPolicyValue
ServantRetentionPolicy_impl::value()
    throw(SystemException)
{
    return value_;
}

PolicyType
ServantRetentionPolicy_impl::policy_type()
    throw(SystemException)
{
    return SERVANT_RETENTION_POLICY_ID;
}

Policy_ptr
ServantRetentionPolicy_impl::copy()
    throw(SystemException)
{
    return Policy::_duplicate(this);
}

void
ServantRetentionPolicy_impl::destroy()
    throw(SystemException)
{
}

// ----------------------------------------------------------------------
// RequestProcessingPolicy_impl constructor and destructor
// ----------------------------------------------------------------------

RequestProcessingPolicy_impl::
RequestProcessingPolicy_impl(RequestProcessingPolicyValue value)
    : value_(value)
{
}

RequestProcessingPolicy_impl::~RequestProcessingPolicy_impl()
{
}

// ----------------------------------------------------------------------
// RequestProcessingPolicy_impl public member implementation
// ----------------------------------------------------------------------

RequestProcessingPolicyValue
RequestProcessingPolicy_impl::value()
    throw(SystemException)
{
    return value_;
}

PolicyType
RequestProcessingPolicy_impl::policy_type()
    throw(SystemException)
{
    return REQUEST_PROCESSING_POLICY_ID;
}

Policy_ptr
RequestProcessingPolicy_impl::copy()
    throw(SystemException)
{
    return Policy::_duplicate(this);
}

void
RequestProcessingPolicy_impl::destroy()
    throw(SystemException)
{
}

//
// ORBacus extensions
//

// ----------------------------------------------------------------------
// SynchronizationPolicy_impl constructor and destructor
// ----------------------------------------------------------------------

SynchronizationPolicy_impl::SynchronizationPolicy_impl(
    SynchronizationPolicyValue value)
    : value_(value)
{
}

SynchronizationPolicy_impl::~SynchronizationPolicy_impl()
{
}

// ----------------------------------------------------------------------
// SynchronizationPolicy_impl public member implementation
// ----------------------------------------------------------------------

SynchronizationPolicyValue
SynchronizationPolicy_impl::value()
    throw(SystemException)
{
    return value_;
}

PolicyType
SynchronizationPolicy_impl::policy_type()
    throw(SystemException)
{
    return SYNCHRONIZATION_POLICY_ID;
}

Policy_ptr
SynchronizationPolicy_impl::copy()
    throw(SystemException)
{
    return Policy::_duplicate(this);
}

void
SynchronizationPolicy_impl::destroy()
    throw(SystemException)
{
}
