// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/TemplateI.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/Any.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/IFR1.h>
#include <OB/OBV.h>
#include <OB/Policy.h>
#include <OB/ORB.h>
#include <OB/IFR2.h>
#include <OB/IFR3.h>
#include <OB/POAServantBase.h>
#include <OB/NativeTypes.h>
#include <OB/POACurrent.h>
#include <OB/POAServantLocator.h>
#include <OB/POAInterface.h>
#include <OB/Stream.h>
#include <OB/OCI.h>

#include <POA_impl.h>
#include <Upcall.h>
#include <OARoot.h>
#include <POACurrent_impl.h>

using namespace CORBA;
using namespace PortableServer;
using namespace OB;


// ----------------------------------------------------------------------
// Template instantiations
// ----------------------------------------------------------------------

#ifndef HAVE_NO_EXPLICIT_TEMPLATES
template class OB::ObjVar< ServantBase >;
#else
#ifdef HAVE_PRAGMA_DEFINE
//XXX
#endif
#endif

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBRelease(PortableServer::ServantBase* servant)
{
    if(servant)
        servant -> _remove_ref();
}

void
OBDuplicate(PortableServer::ServantBase* servant)
{
    if(servant)
        servant -> _add_ref();
}

// ----------------------------------------------------------------------
// ServantBase protected member implementation
// ----------------------------------------------------------------------

ServantBase::ServantBase(const ServantBase&)
{
}

ServantBase&
ServantBase::operator=(const ServantBase&)
{
    return *this;
}

// ----------------------------------------------------------------------
// ServantBase constructor and destructor
// ----------------------------------------------------------------------

ServantBase::ServantBase()
{
}

ServantBase::~ServantBase()
{
}

// ----------------------------------------------------------------------
// ServantBase public member implementation
// ----------------------------------------------------------------------

POA_ptr
ServantBase::_default_POA()
{
    return OARoot::instance() -> getRootPOA();
}

InterfaceDef_ptr
ServantBase::_get_interface()
    throw(SystemException)
{
    POA_var thePOA = _default_POA();
    POA_impl* poaImpl = dynamic_cast<POA_impl*>(thePOA.in());
    Current_impl* current = poaImpl -> _OB_POACurrent();
    
    Object_var obj;

    try
    {
	ORB_ptr orb = ORB::_OB_instance_nodup();
	if(is_nil(orb))
	    throw INITIALIZE("ORB not initialized");
	obj = orb -> resolve_initial_references("InterfaceRepository");
    }
    catch(const ORB::InvalidName&)
    {
	throw INTF_REPOS(MinorNoIntfRepos, COMPLETED_NO);
    }

    if(is_nil(obj))
    {
	throw INTF_REPOS(MinorNoIntfRepos, COMPLETED_NO);
    }
								       
    Repository_var repository = Repository::_narrow(obj);

    if(is_nil(repository))
    {
	throw INTF_REPOS(MinorNoIntfRepos, COMPLETED_NO);
    }

    ObjectId_var oid = current -> get_object_id();
    String_var intf = _primary_interface(oid, thePOA);
    Contained_var contained = repository -> lookup_id(intf);

    if(is_nil(contained))
	return InterfaceDef::_nil();

    return InterfaceDef::_narrow(contained);
}
    
Boolean
ServantBase::_is_a(const char* logical_type_id)
    throw(SystemException)
{
    return false;
}
    
Boolean
ServantBase::_non_existent()
    throw(SystemException)
{
    return false;
}

void
ServantBase::_add_ref()
{
    // Do nothing
}
    
void
ServantBase::_remove_ref()
{
    // Do nothing
}

DispatchStatus
ServantBase::_OB_dispatch(const char* _ob_op,
			  Buffer& _ob_buf,
			  bool _ob_sw,
			  ULong _ob_offIn,
			  ULong _ob_offOut)
{
    return _OB_dispatchBase(_ob_op, _ob_buf, _ob_sw, _ob_offIn, _ob_offOut);
}

DispatchStatus
ServantBase::_OB_dispatchBase(const char* _ob_op,
			      Buffer& _ob_buf,
			      bool _ob_sw,
			      ULong _ob_offIn,
			      ULong _ob_offOut)
{
    //
    // Optimization. All operations that we dispatch start with an '_'
    // character.
    //
    if(*_ob_op != '_')
        return DispatchStatusNoOperation;
    
    static const char* names[] =
    {
        "_implementation",
        "_interface",
        "_is_a",
        "_non_existent"
    };
    static const ULong num = 4;

    switch(_OB_findOperation(_ob_op, names, num))
    {
    
    case 0: // _implementation
    {
	/*
        ImplementationDef_var _ob_r = _get_implementation();

	ULong _ob_cnt = _ob_offOut;
	OBMarshalCount(_ob_r, _ob_cnt);
	
	_ob_buf.alloc(_ob_cnt);
	Octet* _ob_o = _ob_buf.data + _ob_offOut;
	OBMarshal(_ob_r, _ob_o);
	
	return DispatchStatusOK;
	*/
	assert(false);
	return DispatchStatusNoOperation;
    }

    case 1: // _interface
    {
        InterfaceDef_var _ob_r = _get_interface();

        CountStream _ob_cnt(_ob_offOut);
	_ob_cnt.count_Object(_ob_r);

	_ob_buf.alloc(_ob_cnt.count());
        OutputStream _ob_out(_ob_buf, _ob_offOut);
	_ob_out.write_Object(_ob_r);
	
	return DispatchStatusOK;
    }

    case 2: // _is_a
    {
        InputStream _ob_in(_ob_buf, _ob_offIn, _ob_sw);
        String_var _ob_arg0 = _ob_in.read_string();

        Boolean _ob_r = _is_a(_ob_arg0);

	CountStream _ob_cnt(_ob_offOut);
	_ob_cnt.count_boolean();
	
	_ob_buf.alloc(_ob_cnt.count());
        OutputStream _ob_out(_ob_buf, _ob_offOut);
	_ob_out.write_boolean(_ob_r);
	
	return DispatchStatusOK;
    }
    
    case 3: // _non_existent
    {
        Boolean _ob_r = _non_existent();

	CountStream _ob_cnt(_ob_offOut);
	_ob_cnt.count_boolean();
	
	_ob_buf.alloc(_ob_cnt.count());
        OutputStream _ob_out(_ob_buf, _ob_offOut);
	_ob_out.write_boolean(_ob_r);
	
	return DispatchStatusOK;
    }
    }

    return DispatchStatusNoOperation;
}

Long
ServantBase::_OB_findOperation(const char* op,
			       const char* names[],
			       ULong num)
{
    Long l = 0;
    Long r = (Long)num;
    Long m;
    int res;

    while(l < r)
    {
        m = (l + r) / 2;
        res = strcmp(names[m], op);
        if(res == 0)
            return m;
        if(res > 0)
            r = m;
        else
            l = m + 1;
    }
    return -1;
}

Object_ptr
ServantBase::_OB_createReference()
{
    Object_var obj;
    
    POA_var thePOA = _default_POA();
    POA_impl* poaImpl = dynamic_cast<POA_impl*>(thePOA.in());
    Current_impl* current = poaImpl -> _OB_POACurrent();

    //
    // If we're in an upcall to this servant, then obtain the object
    // reference for the target CORBA Object it is incarnating for
    // this request.
    //
    if(current -> _OB_inUpcall() && current -> _OB_getServant() == this)
    {
        ObjectId_var oid = current -> get_object_id();
        String_var intf = _primary_interface(oid, thePOA);
        obj = poaImpl -> _OB_createReference(current -> _OB_getKey(), intf);
    }
    else
    {
        obj = thePOA -> servant_to_reference(this);
    }
    return obj._retn();
}

bool
ServantBase::_OB_isDynamic()
{
    return false;
}

#ifdef HAVE_JTC
void
ServantBase::_OB_lock()
{
    servantSyncMutex_.lock();
}

void
ServantBase::_OB_unlock()
{
    servantSyncMutex_.unlock();
}
#endif

// ----------------------------------------------------------------------
// RefCountServantBase protected member implementation
// ----------------------------------------------------------------------

RefCountServantBase&
RefCountServantBase::operator=(const RefCountServantBase& rhs)
{
    return *this;
}

// ----------------------------------------------------------------------
// RefCountServantBase constructor and destructor
// ----------------------------------------------------------------------
RefCountServantBase::RefCountServantBase()
{
}

RefCountServantBase::~RefCountServantBase()
{
}

// ----------------------------------------------------------------------
// RefCountServantBase public member implementation
// ----------------------------------------------------------------------

RefCountServantBase::RefCountServantBase(
    const RefCountServantBase&)
{
}

void
RefCountServantBase::_add_ref()
{
    _OB_incRef();
}

void
RefCountServantBase::_remove_ref()
{
    _OB_decRef();
}
