// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef POA_IMPL_H
#define POA_IMPL_H

namespace OB
{

class Downcall;
typedef Downcall* Downcall_ptr;

} // End of namespace OB

void OBDuplicate(OB::Downcall_ptr);
void OBRelease(OB::Downcall_ptr);

namespace OB
{

//
// Forward declarations
//
typedef ObjVar< Downcall > Downcall_var;

class StubImplBase;
class POAPolicies;

template <class K, class T> class Hashtable;

} // End of namespace OB


namespace PortableServer
{

class ServantBase;

class ServantLocator;
typedef ServantLocator* ServantLocator_ptr;

class ServantActivator;
typedef ServantActivator* ServantActivator_ptr;

} // End of namespace PortableServer

void OBDuplicate(PortableServer::ServantBase*);
void OBRelease(PortableServer::ServantBase*);
void OBDuplicate(PortableServer::ServantLocator_ptr);
void OBRelease(PortableServer::ServantLocator_ptr);
void OBDuplicate(PortableServer::ServantActivator_ptr);
void OBRelease(PortableServer::ServantActivator_ptr);

namespace PortableServer
{

//
// Forward declarations
//
typedef OB::ObjVar< ServantBase > ServantBase_var;

class Current_impl;

typedef OB::ObjVar< ServantLocator > ServantLocator_var;

typedef OB::ObjVar< ServantActivator > ServantActivator_var;


//
// The POA_impl class
//
class POA_impl : public POA
{
#ifdef HAVE_JTC
    JTCRecursiveMutex mutex_;
#endif
    
    //
    // The name of this POA
    //
    CORBA::String_var name_;

    //
    // My parent POA
    //
    POA_impl* parent_;

    //
    // My POA manager
    //
    POAManager_var manager_;

    //
    // The default servant
    //
    ServantBase_var defaultServant_;

    //
    // The servant locator
    //
    ServantLocator_var servantLocator_;

    //
    // The servant activator
    //
    ServantActivator_var servantActivator_;

    //
    // The POA ID
    //
    OB::StringSeq poaId_;

    //
    // Some internal types
    //
    typedef OB::Hashtable<ObjectId,
                          ServantBase_var> ActiveObjectMap;
    
    typedef OB::Hashtable<CORBA::String_var, POA_var> POAHashTable;
    
    typedef OB::Hashtable<Servant,
                          ObjectId> ServantIdMap;

    typedef OB::FixSeq<OB::StubImplBase*, int> StubImplBasePtrSeq;
    typedef OB::Hashtable<ObjectId,
                          StubImplBasePtrSeq*> LocalStubImplSeqMap;

    //
    // The associated AdapterActivator
    //
    AdapterActivator_var adapterActivator_;
    
    //
    // Our children.
    //
    POAHashTable* children_;

    //
    // All connected implementations.
    //
    ActiveObjectMap* activeObjectMap_;

    //
    // If using UNIQUE_ID policy then this will map servant to it's id
    //
    ServantIdMap* servantIdMap_;

    //
    // Mapping between oid and LocalStubImpl
    //
    LocalStubImplSeqMap* localStubImplSeqMap_;

    //
    // My policies.
    //
    OB::POAPolicies* policies_;
    
    //
    // Used for generating SYSTEM_ID object ids.
    //
    CORBA::ULong nextSystemID_;
    
    //
    // Used for generated TRANSIENT keys
    //
    CORBA::ULong num_;

    //
    // Persistency value for lifespan
    //
    bool persistent_;

    //
    // The POA create if, if not persistent
    //
    struct timeval poaCreateTime_;

    //
    // This flag is true if the POA has been destroyed.
    //
    bool destroyed_;

#ifdef HAVE_JTC
    //
    // If this POA has the single thread policy set then this mutex is
    // used ensure this policy is met.
    //
    JTCRecursiveMutex poaSyncMutex_;
    static JTCRecursiveMutex orbSyncMutex_;
#endif

    //                                                    
    // The PortableServer::Current implementation. The RootPOA is
    // responsible for the creation and deletion of the one and only
    // PortableSerer::Current.
    //                                               
    Current_impl* poaCurrent_;

    //
    // Create a new system generated object id
    //
    void createObjectId(ObjectId*&);

    //
    // Determine if this ID was system generated
    //
    bool isSystemGeneratedId(const ObjectId&);

public:
    //
    // Some compilers need this to be public.
    //
    typedef OB::OctetSeq ObjectKey;
    typedef OB::SeqVar< OB::OctetSeq > ObjectKey_var;

private:
    //
    // The data contained in an object key
    //
    struct ObjectKeyData
    {
        OB::StringSeq poaID; // The POA to which this key refers
        ObjectId oid; // The object-id to which this key refers
        bool persistent; // Is the POA that created this key persistent?
        time_t createTime; // If transient, what time was the POA created?
    };

    //
    // Given an object id create an object key.
    //
    ObjectKey* createObjectKey(const ObjectId&);

    //
    // Parse an object key, filling in the fields of the ObjectKeyData
    //
    bool parseObjectKey(const ObjectKey&, ObjectKeyData& data);

    //
    // Remove a child POA given it's namep
    //
    void removeChildPOA(const char*);

    //
    // Generate a new object id, and add the servant to the active object map
    //
    ObjectId* generateIdAddServant(ServantBase*);

    //
    // Add a servant to the active object map
    //
    void addServantActiveObjectMap(const ObjectId&, ServantBase*);

    //
    // Internal find_POA operation
    //
    POA_impl* findPOA(const char*, CORBA::Boolean, bool);

    //
    // Determine if this POA matches the provided object key
    //
    bool poaMatches(const ObjectKeyData&);

    //
    // Validate the set of provided POA policies
    //
    bool validatePolicies();

    //
    // Create the stub impl base
    //
    OB::StubImplBase* createStubImplBase(const ObjectKey&,
                                         const ObjectKeyData&,
                                         const char*);
    
    //
    // Setup a method invocation
    //
    void setupInvoke(OB::Downcall_ptr,
                     CORBA::ULong,
                     const ObjectKey&,
                     const ObjectId&,
                     const char*,
                     OB::Buffer&,
                     bool,
                     CORBA::ULong,
                     CORBA::ULong,
                     OCI::TransportInfo_ptr);
public:
    struct InvocationRecord
    {
        OB::Downcall_var downcall;
        CORBA::ULong requestId;
        ObjectKey key;
        ObjectId oid;
        CORBA::String_var op;
        OB::Buffer buf;
        bool sw;
        CORBA::ULong offIn;
        CORBA::ULong offOut;
        OCI::TransportInfo_var info;

        InvocationRecord* next;
    };
    typedef OB::VarSeq< InvocationRecord*, int > InvocationRecordSeq;
private:
    InvocationRecordSeq records_;
    
    //
    // Invoke this request on the appropriate servant.
    //
    void invoke(OB::Downcall_ptr,
                CORBA::ULong,
                const ObjectKey&,
                const ObjectId&,
                const char*,
                OB::Buffer&,
                bool,
                CORBA::ULong,
                CORBA::ULong,
                OCI::TransportInfo_ptr);

    //
    // Common initialization code
    //
    void init(const CORBA::PolicyList&, POAManager_ptr);

    //
    // Private, internal constructor
    //
    POA_impl(Current_impl*,
             POA_impl*,
             POAManager_ptr,
             const char*,
             const CORBA::PolicyList&);

protected:
    //
    // Stop `class has private destructor and no friends' warning
    //
    ~POA_impl();
    
public:
    //
    // Public constructor. This constructor creates the Root POA.
    //
    POA_impl();

    //
    // Standard IDL to C++ Mapping
    //
    
    //
    // IDL:omg.org/PortableServer/POA/create_POA:1.0
    //
    virtual POA_ptr
    create_POA(const char* adapter_name,
               POAManager_ptr a_POAManager,
               const CORBA::PolicyList& policies)
        throw(POA::AdapterAlreadyExists,
              POA::InvalidPolicy,
              CORBA::SystemException);

    //
    // IDL:omg.org/PortableServer/POA/find_POA:1.0
    //
    virtual POA_ptr find_POA(const char* adapter_name,
                             CORBA::Boolean activate_it)
        throw(POA::AdapterNonExistent,
              CORBA::SystemException);

    //
    // IDL:omg.org/PortableServer/POA/destroy:1.0
    //
    virtual void destroy(CORBA::Boolean etherealize_objects,
                         CORBA::Boolean wait_for_completion)
        throw(CORBA::SystemException);

    //
    // IDL:omg.org/PortableServer/POA/create_thread_policy:1.0
    //
    virtual ThreadPolicy_ptr
    create_thread_policy(ThreadPolicyValue value)
        throw(CORBA::SystemException);

    //
    // IDL:omg.org/PortableServer/POA/create_lifespan_policy:1.0
    //
    virtual LifespanPolicy_ptr
    create_lifespan_policy(LifespanPolicyValue value)
        throw(CORBA::SystemException);

    //
    // IDL:omg.org/PortableServer/POA/create_id_uniqueness_policy:1.0
    //
    virtual IdUniquenessPolicy_ptr
    create_id_uniqueness_policy(IdUniquenessPolicyValue value)
        throw(CORBA::SystemException);

    //
    // IDL:omg.org/PortableServer/POA/create_id_assignment_policy:1.0
    //
    virtual IdAssignmentPolicy_ptr
    create_id_assignment_policy(IdAssignmentPolicyValue value)
        throw(CORBA::SystemException);

    //
    // IDL:omg.org/PortableServer/POA/create_implicit_activation_policy:1.0
    //
    virtual ImplicitActivationPolicy_ptr
    create_implicit_activation_policy(
        ImplicitActivationPolicyValue value)
        throw(CORBA::SystemException);

    //
    // IDL:omg.org/PortableServer/POA/create_servant_retention_policy:1.0
    //
    virtual ServantRetentionPolicy_ptr
    create_servant_retention_policy(
        ServantRetentionPolicyValue value)
        throw(CORBA::SystemException);

    //
    // IDL:omg.org/PortableServer/POA/create_request_processing_policy:1.0
    //
    virtual RequestProcessingPolicy_ptr
    create_request_processing_policy(
        RequestProcessingPolicyValue value)
        throw(CORBA::SystemException);

    //
    // IDL:omg.org/PortableServer/POA/create_synchronization_policy:1.0
    //
    virtual SynchronizationPolicy_ptr
    create_synchronization_policy(
        SynchronizationPolicyValue value)
        throw(CORBA::SystemException);

    //
    // IDL:omg.org/PortableServer/POA/the_name:1.0
    //
    virtual char* the_name()
        throw(CORBA::SystemException);

    //
    // IDL:omg.org/PortableServer/POA/the_parent:1.0
    //
    virtual POA_ptr the_parent()
        throw(CORBA::SystemException);

    //
    // IDL:omg.org/PortableServer/POA/the_POAManager:1.0
    //
    virtual POAManager_ptr the_POAManager()
        throw(CORBA::SystemException);

    //
    // IDL:omg.org/PortableServer/POA/the_activator:1.0
    //
    virtual AdapterActivator_ptr the_activator()
        throw(CORBA::SystemException);
    virtual void the_activator(AdapterActivator_ptr)
        throw(CORBA::SystemException);

    //
    // IDL:omg.org/PortableServer/POA/get_servant_manager:1.0
    //
    virtual ServantManager_ptr get_servant_manager()
        throw(POA::WrongPolicy,
              CORBA::SystemException);

    //
    // IDL:omg.org/PortableServer/POA/set_servant_manager:1.0
    //
    virtual void set_servant_manager(ServantManager_ptr imgr)
        throw(POA::WrongPolicy,
              CORBA::SystemException);

    //
    // IDL:omg.org/PortableServer/POA/get_servant:1.0
    //
    virtual Servant get_servant()
        throw(POA::NoServant,
              POA::WrongPolicy,
              CORBA::SystemException);

    //
    // IDL:omg.org/PortableServer/POA/set_servant:1.0
    //
    virtual void set_servant(Servant p_servant)
        throw(POA::WrongPolicy,
              CORBA::SystemException);

    //
    // IDL:omg.org/PortableServer/POA/activate_object:1.0
    //
    virtual ObjectId*
    activate_object(Servant p_servant)
        throw(POA::ServantAlreadyActive,
              POA::WrongPolicy,
              CORBA::SystemException);

    //
    // IDL:omg.org/PortableServer/POA/activate_object_with_id:1.0
    //
    virtual void activate_object_with_id(const ObjectId& id,
                                         Servant p_servant)
        throw(POA::ServantAlreadyActive,
              POA::ObjectAlreadyActive,
              POA::WrongPolicy,
              CORBA::SystemException);

    //
    // IDL:omg.org/PortableServer/POA/deactivate_object:1.0
    //
    virtual void deactivate_object(const ObjectId& oid)
        throw(POA::ObjectNotActive,
              POA::WrongPolicy,
              CORBA::SystemException);

    //
    // IDL:omg.org/PortableServer/POA/create_reference:1.0
    //
    virtual CORBA::Object_ptr create_reference(const char* intf)
        throw(POA::WrongPolicy,
              CORBA::SystemException);

    //
    // IDL:omg.org/PortableServer/POA/create_reference_with_id:1.0
    //
    virtual CORBA::Object_ptr
    create_reference_with_id(const ObjectId& oid,
                             const char* intf)
        throw(POA::WrongPolicy,
              CORBA::SystemException);

    //
    // IDL:omg.org/PortableServer/POA/servant_to_id:1.0
    //
    virtual ObjectId*
    servant_to_id(Servant p_servant)
        throw(POA::ServantNotActive,
              POA::WrongPolicy,
              CORBA::SystemException);

    //
    // IDL:omg.org/PortableServer/POA/servant_to_reference:1.0
    //
    virtual CORBA::Object_ptr
    servant_to_reference(Servant p_servant)
        throw(POA::ServantNotActive,
              POA::WrongPolicy,
              CORBA::SystemException);

    //
    // IDL:omg.org/PortableServer/POA/reference_to_servant:1.0
    //
    virtual Servant
    reference_to_servant(CORBA::Object_ptr reference)
        throw(POA::ObjectNotActive,
              POA::WrongAdapter,
              POA::WrongPolicy,
              CORBA::SystemException);

    //
    // IDL:omg.org/PortableServer/POA/reference_to_id:1.0
    //
    virtual ObjectId*
    reference_to_id(CORBA::Object_ptr reference)
        throw(POA::WrongAdapter,
              POA::WrongPolicy,
              CORBA::SystemException);

    //
    // IDL:omg.org/PortableServer/POA/id_to_servant:1.0
    //
    virtual Servant
    id_to_servant(const ObjectId& oid)
        throw(POA::ObjectNotActive,
              POA::WrongPolicy,
              CORBA::SystemException);

    //
    // IDL:omg.org/PortableServer/POA/id_to_reference:1.0
    //
    virtual CORBA::Object_ptr
    id_to_reference(const ObjectId& oid)
        throw(POA::ObjectNotActive,
              POA::WrongPolicy,
              CORBA::SystemException);

    //
    // ORBacus internal functions
    // Application programs must not use these functions directly
    //

    //
    // Called before a method invocation on a servant
    //
    void _OB_preinvoke(const char*,
		       const ObjectId&, const ObjectKey&,
                       ServantBase*,
                       ServantLocator::Cookie,
                       OCI::TransportInfo_ptr);

    //
    // Called after a method invocation on a servant
    //
    void _OB_postinvoke();
    
    //
    // Invoke a request on the POA and object-id derived from the object-key.
    //
    void _OB_invoke(OB::Downcall_ptr,
                    CORBA::ULong,
                    const ObjectKey&,
                    const char*,
                    OB::Buffer&,
                    bool,
                    CORBA::ULong,
                    CORBA::ULong,
                    OCI::TransportInfo_ptr);
    
    //
    // Locate a POA, from the provided key data
    //
    POA_impl* _OB_locatePOA(const ObjectKeyData&);

    //
    // Find an object-id described by the provided key
    //
    bool _OB_findByKey(const ObjectKey&);

    //
    // Determine if a servant is present
    //
    bool _OB_findServant(const ObjectId&);

    //
    // Locate the POA given the provided object key
    //
    POA_impl* _OB_locatePOA(const ObjectKey&);

    //
    // Create an object reference given an object key and interface name
    //
    CORBA::Object_ptr _OB_createReference(const ObjectKey&, const char*);
        
    //
    // Ask a servant to create the StubImplBase, if possible.
    //
    OB::StubImplBase* _OB_createStubImplBase(const ObjectKey&, const char*);

    //
    // Locate a servant, given it's object id and operation name
    //
    ServantBase*
    _OB_locateServant(const ObjectId&,
                      const char* op,
                      ServantLocator::Cookie&);

    //
    // Called by the POAManager on a state change.
    //
    void _OB_stateChange();

    Current_impl* _OB_POACurrent();

    void _OB_removeStub(const ObjectId&, OB::StubImplBase*);
};

} // End of namespace PortableServer

#endif
