// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_SELECT_REACTOR_H
#define OB_SELECT_REACTOR_H

namespace OB
{

//
// The SelectReactor class
//
class SelectReactor : public Reactor
{
    //
    // Hide copy-constructor and assignment operator
    //
    SelectReactor(const SelectReactor&);
    void operator=(const SelectReactor&);

    //
    // If set to true, event dispatching must be stopped
    //
    bool stop_;

    //
    // Struct for an event handler info
    //
    struct HandlerInfo
    {
	EventHandler* handler;
	Mask mask;
	Handle handle;
	HandlerInfo* next;
    };

    //
    // Head of the list with all event handler infos
    //
    HandlerInfo* handlerInfoList_;

    //
    // Counter which keeps track of how often the list with event
    // handler infos is currently in use
    //
    CORBA::ULong handlerInfoListUsageCount_;

    //
    // Reap the list of event handler infos
    //
    void reapHandlerInfoList();

    //
    // Set file descriptor sets
    //
    int setFdSets(fd_set&, fd_set&) const;

    //
    // Evaluate file descriptor sets
    //
    void evalFdSets(fd_set&, fd_set&);

protected: // Make the constructor and destructor protected to avoid
           // stupid warning messages from some compilers
    SelectReactor();
    virtual ~SelectReactor();

public:

    //
    // Get the reactor singleton
    //
    static Reactor* instance();

    //
    // Register and unregister event handler
    //
    virtual void registerHandler(EventHandler*, Mask, Handle);
    virtual void unregisterHandler(EventHandler*);

    //
    // Dispatch events
    //
    virtual void dispatch();
    virtual bool dispatchOneEvent(CORBA::Long);

    //
    // Test to see if an event is available
    //
    virtual bool eventReady();

    //
    // Stop event dispatching
    //
    virtual void stop() { stop_ = true; }
};

} // End of namespace OB

#endif
