// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/TemplateI.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/Any.h>
#include <OB/OBV.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/Policy.h>
#include <OB/IFR2.h>
#include <OB/ORB.h>
#include <OB/OCI.h>
#include <OB/OBPolicies.h>
#include <OB/NativeTypes.h>
#include <OB/POAServantBase.h>
#include <OB/POAInterface.h>
#include <OB/POAServantLocator.h>
#include <OB/Stream.h>

#include <POA_impl.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::Client_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::Client_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

void
OBDuplicate(OB::StubImplBase_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::StubImplBase_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// Template instantiations
// ----------------------------------------------------------------------

#ifndef HAVE_NO_EXPLICIT_TEMPLATES
template class OB::ObjVar< Client >;
template class OB::ObjForSeq< Client >;
template class OB::ObjSeq< Client, int >;
template class OB::SeqVar< OB::ObjSeq< Client, int > >;
//template class OB::SeqOut< OB::ObjSeq< Client > >;

template class OB::ObjVar< StubImplBase >;
//template class OB::ObjForSeq< StubImplBase >;
//template class OB::ObjSeq< StubImplBase >;
#else
#ifdef HAVE_PRAGMA_DEFINE
#pragma define(OB::ObjVar< Client >)
#pragma define(OB::ObjForSeq< Client >)
#pragma define(OB::ObjSeq< Client, int >)
#pragma define(OB::SeqVar< OB::ObjSeq< Client, int > >)
//#pragma define(OB::SeqOut< OB::ObjSeq< Client > >)

#pragma define(OB::ObjVar< StubImplBase >)
//#pragma define(OB::ObjForSeq< StubImplBase >)
//#pragma define(OB::ObjSeq< StubImplBase >)
#endif
#endif

// ----------------------------------------------------------------------
// Client constructor and destructor
// ----------------------------------------------------------------------

Client::Client()
    : codeConverters_(0), reuseClient_(true), usage_(0)
{
}

Client::~Client()
{
    if(codeConverters_)
	delete codeConverters_;
}

// ----------------------------------------------------------------------
// Client public member implementation
// ----------------------------------------------------------------------

CodeConverters*
Client::getCodeConverters() const
{
    return codeConverters_;
}

void
Client::setCodeConverters(const CodeConverters& codeConverters)
{
    if(!codeConverters_)
	codeConverters_ = new CodeConverters();

    //
    // Copy code set converters
    //
    codeConverters_ -> charConverter = codeConverters.charConverter;
    codeConverters_ -> wcharConverter = codeConverters.wcharConverter;
}

// ----------------------------------------------------------------------
// StubImplBase constructor and destructor
// ----------------------------------------------------------------------

StubImplBase::StubImplBase()
    : deactivated_(false), _ob_servant_(0)
{
}

StubImplBase::~StubImplBase()
{
    //cout << "~StubImplBase" << endl;
    if(!is_nil(_ob_clt_))
    {
	ORB_ptr orb = ORB::_OB_instance_nodup();
	//
	// If the ORB was already destroyed, don't care about decrementing
	// the client usage - the ORB's destructor destroyed all clients
	// anyway.
	//
	if(!is_nil(orb))
	    orb -> _OB_decClientUsage(_ob_clt_);
    }
    if(!is_nil(_ob_poa_))
    {
        PortableServer::POA_impl* poa =
            dynamic_cast<PortableServer::POA_impl*>(_ob_poa_.in());
        poa -> _OB_removeStub(_ob_oid_, this);
    }
}

// ----------------------------------------------------------------------
// StubImplBase public member implementation
// ----------------------------------------------------------------------

InterfaceDef_ptr
StubImplBase::_get_interface()
{
    if(_ob_servant_ != 0)
        return _ob_servant_ -> _get_interface();
    
    if(is_nil(_ob_clt_))
        throw NO_IMPLEMENT();
    
    ULong _ob_off = _ob_clt_ -> offset(this, "_interface");
    
    Buffer _ob_buf(_ob_off);
    
    bool _ob_sw, _ob_ex, _ob_fo;
    _ob_off = _ob_clt_ -> request(this, "_interface", _ob_buf,
				  _ob_sw, _ob_ex, _ob_fo,
				  _ob_tout_, _ob_retry_);
    
    InputStream _ob_in(_ob_buf, _ob_off, _ob_sw);
    if(_ob_fo)
    {
        IOP::IOR ior;
        IOP::IOR::_OB_unmarshal(ior, _ob_in);
        throw StubImplForward(ior);
    }

    if(_ob_ex)
        throw UNKNOWN();

    InterfaceDef_ptr _ob_r = InterfaceDef::_nil();
    InterfaceDef::_OB_unmarshal(_ob_r, _ob_in);
	
    return _ob_r;
}

Boolean
StubImplBase::_non_existent()
{
    if(_ob_servant_ != 0)
        return _ob_servant_ -> _non_existent();
    
    if(is_nil(_ob_clt_))
        throw NO_IMPLEMENT();
    
    ULong _ob_off = _ob_clt_ -> offset(this, "_non_existent");
    
    Buffer _ob_buf(_ob_off);
    
    bool _ob_sw, _ob_ex, _ob_fo;
    _ob_off = _ob_clt_ -> request(this, "_non_existent", _ob_buf,
				  _ob_sw, _ob_ex, _ob_fo,
				  _ob_tout_, _ob_retry_);
    
    InputStream _ob_in(_ob_buf, _ob_off, _ob_sw);
    if(_ob_fo)
    {
        IOP::IOR ior;
        IOP::IOR::_OB_unmarshal(ior, _ob_in);
        throw StubImplForward(ior);
    }

    if(_ob_ex)
        throw UNKNOWN();
    
    Boolean _ob_r = _ob_in.read_boolean();
    return _ob_r;
}

Boolean
StubImplBase::_is_a(const char* id)
{
    if(_ob_servant_ != 0)
        return _ob_servant_ -> _is_a(id);
    
    //
    // Check implementation object
    //
    if(is_nil(_ob_clt_))
        throw NO_IMPLEMENT();
	
    ULong _ob_off = _ob_clt_ -> offset(this, "_is_a");
    CountStream _ob_cnt(_ob_off);
    _ob_cnt.count_string(id);
	
    Buffer _ob_buf(_ob_cnt.count());
    OutputStream _ob_out(_ob_buf, _ob_off);
    _ob_out.write_string(id);
	
    bool _ob_sw, _ob_ex, _ob_fo;
    _ob_off =  _ob_clt_ -> request(this, "_is_a", _ob_buf,
                                   _ob_sw, _ob_ex, _ob_fo,
                                   _ob_tout_, _ob_retry_);
	
    InputStream _ob_in(_ob_buf, _ob_off, _ob_sw);
    if(_ob_fo)
    {
        IOP::IOR ior;
        IOP::IOR::_OB_unmarshal(ior, _ob_in);
        throw StubImplForward(ior);
    }
	
    if(_ob_ex)
        throw UNKNOWN();
	
    Boolean _ob_r = _ob_in.read_boolean();
    return _ob_r;
}

OCI::TransportInfo_ptr
StubImplBase::_get_oci_transport_info()
{
    if(is_nil(_ob_clt_))
        return OCI::TransportInfo::_nil();

    OCI::Transport_var transport = _ob_clt_ -> transport();
    if(is_nil(transport))
	return OCI::TransportInfo::_nil();

    return transport -> get_info();
}

OCI::ConnectorInfo_ptr
StubImplBase::_get_oci_connector_info()
{
    if(is_nil(_ob_clt_))
        return OCI::ConnectorInfo::_nil();

    OCI::Connector_var connector = _ob_clt_ -> connector();
    if(is_nil(connector))
	return OCI::ConnectorInfo::_nil();

    return connector -> get_info();
}

Client_ptr
StubImplBase::_OB_client()
{
    return Client::_duplicate(_ob_clt_);
}

void
StubImplBase::_OB_client(Client_ptr clt)
{
    _ob_key_.length(0); // Invalidate key cache

    ORB_ptr orb = ORB::_OB_instance_nodup();
    if(is_nil(orb))
	throw INITIALIZE("ORB not initialized");

    if(!is_nil(clt))
	orb -> _OB_incClientUsage(clt);

    if(!is_nil(_ob_clt_))
	orb -> _OB_decClientUsage(_ob_clt_);

    _ob_clt_ = Client::_duplicate(clt);
}

bool
StubImplBase::_OB_isLocal()
{
    return false;
}

Policy_ptr
StubImplBase::_OB_getPolicy(PolicyType policy_type)
{
    ULong policiesLen = _ob_policies_.length();
    for(ULong i = 0; i < policiesLen; ++i)
	if(_ob_policies_[i] -> policy_type() == policy_type)
	    return Policy::_duplicate(_ob_policies_[i]);
    return Policy::_nil();
}

void
StubImplBase::_OB_policies(const PolicyList& policies)
{
    _ob_policies_ = policies;
        
    TimeoutPolicy_var toutPolicy;
    Policy_var policy = _OB_getPolicy(TIMEOUT_POLICY);
    if(!is_nil(policy))
    {
        toutPolicy = TimeoutPolicy::_narrow(policy);
        assert(!is_nil(toutPolicy));
    }
    
    ReconnectPolicy_var reconnectPolicy;
    policy = _OB_getPolicy(RECONNECT_POLICY);
    if(!is_nil(policy))
    {
        reconnectPolicy = ReconnectPolicy::_narrow(policy);
        assert(!is_nil(reconnectPolicy));
    }
    
    _ob_tout_ = is_nil(toutPolicy)
	? -1
	: (Long)toutPolicy -> value();

    _ob_retry_ = is_nil(reconnectPolicy)
	? false
	: (reconnectPolicy -> value() ? TRUE : FALSE);
}

Long
StubImplBase::_OB_timeout()
{
    return _ob_tout_;
}

bool
StubImplBase::_OB_retry()
{
    return _ob_retry_;
}

void
StubImplBase::_OB_deactivate()
{
#ifdef HAVE_JTC
    JTCSynchronized sync(deactivatedMutex_);
#endif
    deactivated_ = true;
}

// ----------------------------------------------------------------------
// InvocationHandler constructor and destructor
// ----------------------------------------------------------------------

InvocationHandler::InvocationHandler(PortableServer::ServantBase* servant,
				     PortableServer::POA_ptr poa,
				     const char* op,
				     const PortableServer::ObjectId& oid,
				     const OctetSeq& key)
    : poa_(poa) // No need to duplicate
{
    PortableServer::POA_impl* poaImpl =
	dynamic_cast<PortableServer::POA_impl*>(poa);
    poaImpl -> _OB_preinvoke(op, oid, key, servant, 0,
			     OCI::TransportInfo::_nil());
}

InvocationHandler::~InvocationHandler()
{
    PortableServer::POA_impl* poaImpl =
	dynamic_cast<PortableServer::POA_impl*>(poa_);
    poaImpl -> _OB_postinvoke();
}
