// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/Any.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/GIOP.h>
#include <OB/IIOP.h>
#include <OB/IFR4.h>
#include <OB/POAServantBase.h>
#include <OB/ValueBase.h>
#include <OB/Util.h>
#include <OB/Stream.h>
#include <OB/CONV_FRAME.h>
#include <OB/CodeSetUtil.h>
#include <OB/CodeSetDatabase.h>
#include <OB/OCI_IIOP.h>

#include <ctype.h>

#include <IIOPUtil.h>

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;
using namespace OB;

MessageViewer* MessageViewer::instance_ = 0;
int MessageViewer::traceLevel_ = 0;

namespace OB
{

class MessageViewerDestroyer
{
public:

    ~MessageViewerDestroyer()
    {
	delete MessageViewer::instance_;
	MessageViewer::instance_ = 0;
    }
};

} // End of namespace OB

static MessageViewerDestroyer destroyer;

// ----------------------------------------------------------------------
// MessageViewer public member implementation
// ----------------------------------------------------------------------

MessageViewer*
MessageViewer::instance()
{
    if(!instance_)
	instance_ = new MessageViewer();

    return instance_;
}

int
MessageViewer::getTraceLevel()
{
    return traceLevel_;
}

void
MessageViewer::setTraceLevel(int level)
{
    traceLevel_ = level;
}

void
MessageViewer::message(const char* msg)
{
    cerr << msg << endl;
}

void
MessageViewer::error(const char* msg)
{
    const char* head = "Error: ";
    String_var s = string_alloc(strlen(head) + strlen(msg));
    strcpy(s.inout(), head);
    strcat(s.inout(), msg);
    message(s);
}

void
MessageViewer::warning(const char* msg)
{
    const char* head = "Warning: ";
    String_var s = string_alloc(strlen(head) + strlen(msg));
    strcpy(s.inout(), head);
    strcat(s.inout(), msg);
    message(s);
}

void
MessageViewer::nca(const char* msg)
{
    const char* head = "Non-compliant application error detected:\n";
    String_var s = string_alloc(strlen(head) + strlen(msg));
    strcpy(s.inout(), head);
    strcat(s.inout(), msg);
    message(s);
}

void
MessageViewer::trace(int level, const char* msg)
{
    if(traceLevel_ >= level)
    {
        String_var s = string_alloc(strlen(msg) + 2);
        strcpy(s.inout(), "[");
        strcat(s.inout(), msg);
        strcat(s.inout(), "]");
        message(s);
    }
}

// ----------------------------------------------------------------------
// Print system exception
// ----------------------------------------------------------------------

void
OB::PrintException(const SystemException& ex)
{
    String_var s = ExceptionToString(ex);
    MessageViewer* viewer = MessageViewer::instance();
    viewer -> message(s);
}

// ----------------------------------------------------------------------
// Get stringified system exception
// ----------------------------------------------------------------------

char*
OB::ExceptionToString(const SystemException& ex)
{
    String_var msg = string_dup("System exception ");

    if(UNKNOWN::_downcast((SystemException*)&ex))
	msg += "`UNKNOWN'";
    else if(BAD_PARAM::_downcast((SystemException*)&ex))
	msg += "`BAD_PARAM'";
    else if(NO_MEMORY::_downcast((SystemException*)&ex))
	msg += "`NO_MEMORY'";
    else if(IMP_LIMIT::_downcast((SystemException*)&ex))
	msg += "`IMP_LIMIT'";
    else if(COMM_FAILURE::_downcast((SystemException*)&ex))
	msg += "`COMM_FAILURE'";
    else if(INV_OBJREF::_downcast((SystemException*)&ex))
	msg += "`INV_OBJREF'";
    else if(NO_PERMISSION::_downcast((SystemException*)&ex))
	msg += "`NO_PERMISSION'";
    else if(INTERNAL::_downcast((SystemException*)&ex))
	msg += "`INTERNAL'";
    else if(MARSHAL::_downcast((SystemException*)&ex))
	msg += "`MARSHAL'";
    else if(INITIALIZE::_downcast((SystemException*)&ex))
	msg += "`INITIALIZE'";
    else if(NO_IMPLEMENT::_downcast((SystemException*)&ex))
	msg += "`NO_IMPLEMENT'";
    else if(BAD_TYPECODE::_downcast((SystemException*)&ex))
	msg += "`BAD_TYPECODE'";
    else if(BAD_OPERATION::_downcast((SystemException*)&ex))
	msg += "`BAD_OPERATION'";
    else if(NO_RESOURCES::_downcast((SystemException*)&ex))
	msg += "`NO_RESOURCES'";
    else if(NO_RESPONSE::_downcast((SystemException*)&ex))
	msg += "`NO_RESPONSE'";
    else if(PERSIST_STORE::_downcast((SystemException*)&ex))
	msg += "`PERSIST_STORE'";
    else if(BAD_INV_ORDER::_downcast((SystemException*)&ex))
	msg += "`BAD_INV_ORDER'";
    else if(TRANSIENT::_downcast((SystemException*)&ex))
	msg += "`TRANSIENT'";
    else if(FREE_MEM::_downcast((SystemException*)&ex))
	msg += "`FREE_MEM'";
    else if(INV_IDENT::_downcast((SystemException*)&ex))
	msg += "`INV_IDENT'";
    else if(INV_FLAG::_downcast((SystemException*)&ex))
	msg += "`INV_FLAG'";
    else if(INTF_REPOS::_downcast((SystemException*)&ex))
	msg += "`INTF_REPOS'";
    else if(BAD_CONTEXT::_downcast((SystemException*)&ex))
	msg += "`BAD_CONTEXT'";
    else if(OBJ_ADAPTER::_downcast((SystemException*)&ex))
	msg += "`OBJ_ADAPTER'";
    else if(DATA_CONVERSION::_downcast((SystemException*)&ex))
	msg += "`DATA_CONVERSION'";
    else if(OBJECT_NOT_EXIST::_downcast((SystemException*)&ex))
	msg += "`OBJECT_NOT_EXIST'";
    else if(TRANSACTION_REQUIRED::_downcast((SystemException*)&ex))
	msg += "`TRANSACTION_REQUIRED'";
    else if(TRANSACTION_ROLLEDBACK::_downcast((SystemException*)&ex))
	msg += "`TRANSACTION_ROLLEDBACK'";
    else if(INVALID_TRANSACTION::_downcast((SystemException*)&ex))
	msg += "`INVALID_TRANSACTION'";
    else if(INV_POLICY::_downcast((SystemException*)&ex))
	msg += "`INV_POLICY'";
    else
	assert(false);

    const char* reason = ex.reason();
    assert(reason);
    if(strlen(reason) > 0)
    {
	msg += "\nReason: ";
	msg += reason;
    }

    msg += "\nCompleted: ";

    switch(ex.completed())
    {
    case COMPLETED_YES:
	msg += "yes";
	break;

    case COMPLETED_NO:
	msg += "no";
	break;

    case COMPLETED_MAYBE:
	msg += "maybe";
	break;
    }

    msg += "\nMinor code: ";
    msg += ex.minor();

    const COMM_FAILURE* commFailure =
	COMM_FAILURE::_downcast((SystemException*)&ex);

    if(commFailure)
    {
	switch(commFailure -> minor())
	{
	case MinorRecv:
	    msg += " (recv() failed)";
	    break;

	case MinorRecvZero:
	    msg += " (recv() returned zero)";
	    break;

	case MinorSend:
	    msg += " (send() failed)";
	    break;

	case MinorSendZero:
	    msg += " (send() returned zero)";
	    break;

	case MinorSocket:
	    msg += " (socket() failed)";
	    break;

	case MinorSetsockopt:
	    msg += " (setsockopt() failed)";
	    break;

	case MinorGetsockopt:
	    msg += " (getsockopt() failed)";
	    break;

	case MinorBind:
	    msg += " (bind() failed)";
	    break;

	case MinorListen:
	    msg += " (listen() failed)";
	    break;

	case MinorConnect:
	    msg += " (connect() failed)";
	    break;

	case MinorAccept:
	    msg += " (accept() failed)";
	    break;

	case MinorSelect:
	    msg += " (select() failed)";
	    break;

	case MinorGethostname:
	    msg += " (gethostname() failed)";
	    break;

	case MinorGethostbyname:
	    msg += " (gethostbyname() failed)";
	    break;

	case MinorWSAStartup:
	    msg += " (WSAStartup() failed)";
	    break;

	case MinorWSACleanup:
	    msg += " (WSACleanup() failed)";
	    break;

	case MinorNoGIOP:
	    msg += " (not a GIOP message)";
	    break;

	case MinorUnknownMessage:
	    msg += " (unknown GIOP message)";
	    break;

	case MinorWrongMessage:
	    msg += " (wrong GIOP message)";
	    break;

	case MinorCloseConnection:
	    msg += " (got a `close connection' message)";
	    break;

	case MinorMessageError:
	    msg += " (got a `message error' message)";
	    break;
	}
    }

    const INTF_REPOS* intfRepos =
	INTF_REPOS::_downcast((SystemException*)&ex);

    if(intfRepos)
    {
	switch(intfRepos -> minor())
	{
	case MinorNoIntfRepos:
	    msg += " (interface repository is not available)";
	    break;

	case MinorLookupAmbiguous:
	    msg += " (search name for lookup() is ambiguous)";
	    break;

	case MinorIllegalRecursion:
	    msg += " (illegal recursion)";
	    break;
	}
    }

    const BAD_PARAM* badParam =
	BAD_PARAM::_downcast((SystemException*)&ex);

    if(badParam)
    {
	switch(badParam -> minor())
	{
	case 2:
	    msg += " (repository id already exists)";
	    break;

	case 3:
            msg += " (name already exists)";
	    break;

	case 4:
	    msg += " (object is not a Container)";
	    break;

	case 5:
            msg += " (name collision with inherited definition)";
	    break;

	case MinorDuplicateFormalParam:
            msg += " (duplicate formal parameter)";
	    break;

	case MinorInvalidValueInheritance:
            msg += " (invalid valuetype inheritance)";
	    break;

	case MinorInvalidValueModifier:
            msg += " (invalid valuetype modifier)";
	    break;

	case MinorDuplicateValueInit:
            msg += " (duplicate valuetype initializer)";
	    break;

	case MinorAbstractValueInit:
            msg += " (abstract valuetype cannot have initializers)";
	    break;

	case MinorDuplicateBaseType:
            msg += " (base type appears more than once)";
	    break;
	}
    }

    const BAD_INV_ORDER* badInvOrder =
	BAD_INV_ORDER::_downcast((SystemException*)&ex);

    if(badInvOrder)
    {
	switch(badInvOrder -> minor())
	{
	case 1:
	    msg += " (dependency prevents destruction of object)";
	    break;

	case 2:
            msg += " (destroy invoked on indestructible object)";
	    break;
	}
    }

    const MARSHAL* marshal =
	MARSHAL::_downcast((SystemException*)&ex);

    if(marshal)
    {
	switch(marshal -> minor())
	{
        case MinorReadNoValueFactory:
            msg += " (no valuetype factory)";
	    break;

        case MinorReadOverflow:
            msg += " (input stream buffer overflow)";
	    break;

        case MinorReadBooleanOverflow:
            msg += " (overflow while reading boolean)";
	    break;

        case MinorReadCharOverflow:
            msg += " (overflow while reading char)";
	    break;

        case MinorReadWCharOverflow:
            msg += " (overflow while reading wchar)";
	    break;

        case MinorReadOctetOverflow:
            msg += " (overflow while reading octet)";
	    break;

        case MinorReadShortOverflow:
            msg += " (overflow while reading short)";
	    break;

        case MinorReadUShortOverflow:
            msg += " (overflow while reading ushort)";
	    break;

        case MinorReadLongOverflow:
            msg += " (overflow while reading long)";
	    break;

        case MinorReadULongOverflow:
            msg += " (overflow while reading ulong)";
	    break;

        case MinorReadLongLongOverflow:
            msg += " (overflow while reading longlong)";
	    break;

        case MinorReadULongLongOverflow:
            msg += " (overflow while reading ulonglong)";
	    break;

        case MinorReadFloatOverflow:
            msg += " (overflow while reading float)";
	    break;

        case MinorReadDoubleOverflow:
            msg += " (overflow while reading double)";
	    break;

        case MinorReadLongDoubleOverflow:
            msg += " (overflow while reading longdouble)";
	    break;

        case MinorReadStringOverflow:
            msg += " (overflow while reading string)";
	    break;

        case MinorReadStringZeroLength:
            msg += " (encountered zero-length string)";
	    break;

        case MinorReadStringNullChar:
            msg += " (encountered null char in string)";
	    break;

        case MinorReadWStringOverflow:
            msg += " (overflow while reading string)";
	    break;

        case MinorReadWStringZeroLength:
            msg += " (encountered zero-length wstring)";
	    break;

        case MinorReadWStringNullWChar:
            msg += " (encountered null wchar in wstring)";
	    break;

        case MinorReadFixedOverflow:
            msg += " (overflow while reading wstring)";
	    break;

        case MinorReadFixedInvalid:
            msg += " (invalid encoding for fixed value)";
	    break;

        case MinorReadBooleanArrayOverflow:
            msg += " (overflow while reading boolean)";
	    break;

        case MinorReadCharArrayOverflow:
            msg += " (overflow while reading char)";
	    break;

        case MinorReadWCharArrayOverflow:
            msg += " (overflow while reading wchar)";
	    break;

        case MinorReadOctetArrayOverflow:
            msg += " (overflow while reading octet)";
	    break;

        case MinorReadShortArrayOverflow:
            msg += " (overflow while reading short)";
	    break;

        case MinorReadUShortArrayOverflow:
            msg += " (overflow while reading ushort)";
	    break;

        case MinorReadLongArrayOverflow:
            msg += " (overflow while reading long)";
	    break;

        case MinorReadULongArrayOverflow:
            msg += " (overflow while reading ulong)";
	    break;

        case MinorReadLongLongArrayOverflow:
            msg += " (overflow while reading longlong)";
	    break;

        case MinorReadULongLongArrayOverflow:
            msg += " (overflow while reading ulonglong)";
	    break;

        case MinorReadFloatArrayOverflow:
            msg += " (overflow while reading float)";
	    break;

        case MinorReadDoubleArrayOverflow:
            msg += " (overflow while reading double)";
	    break;

        case MinorReadLongDoubleArrayOverflow:
            msg += " (overflow while reading longdouble)";
	    break;

	case MinorReadInvTypeCodeIndirection:
            msg += " (invalid TypeCode indirection)";
	    break;

        case MinorWriteOverflow:
            msg += " (output stream buffer overflow)";
	    break;

        case MinorWriteBooleanOverflow:
            msg += " (overflow while writing boolean)";
	    break;

        case MinorWriteCharOverflow:
            msg += " (overflow while writing char)";
	    break;

        case MinorWriteWCharOverflow:
            msg += " (overflow while writing wchar)";
	    break;

        case MinorWriteWCharInvalid:
            msg += " (invalid wchar)";
	    break;

        case MinorWriteOctetOverflow:
            msg += " (overflow while writing octet)";
	    break;

        case MinorWriteShortOverflow:
            msg += " (overflow while writing short)";
	    break;

        case MinorWriteUShortOverflow:
            msg += " (overflow while writing ushort)";
	    break;

        case MinorWriteLongOverflow:
            msg += " (overflow while writing long)";
	    break;

        case MinorWriteULongOverflow:
            msg += " (overflow while writing ulong)";
	    break;

        case MinorWriteLongLongOverflow:
            msg += " (overflow while writing longlong)";
	    break;

        case MinorWriteULongLongOverflow:
            msg += " (overflow while writing ulonglong)";
	    break;

        case MinorWriteFloatOverflow:
            msg += " (overflow while writing float)";
	    break;

        case MinorWriteDoubleOverflow:
            msg += " (overflow while writing double)";
	    break;

        case MinorWriteLongDoubleOverflow:
            msg += " (overflow while writing longdouble)";
	    break;

        case MinorWriteStringOverflow:
            msg += " (overflow while writing string)";
	    break;

        case MinorWriteWStringOverflow:
            msg += " (overflow while writing string)";
	    break;

        case MinorWriteFixedOverflow:
            msg += " (overflow while writing wstring)";
	    break;

        case MinorWriteBooleanArrayOverflow:
            msg += " (overflow while writing boolean)";
	    break;

        case MinorWriteCharArrayOverflow:
            msg += " (overflow while writing char)";
	    break;

        case MinorWriteWCharArrayOverflow:
            msg += " (overflow while writing wchar)";
	    break;

        case MinorWriteOctetArrayOverflow:
            msg += " (overflow while writing octet)";
	    break;

        case MinorWriteShortArrayOverflow:
            msg += " (overflow while writing short)";
	    break;

        case MinorWriteUShortArrayOverflow:
            msg += " (overflow while writing ushort)";
	    break;

        case MinorWriteLongArrayOverflow:
            msg += " (overflow while writing long)";
	    break;

        case MinorWriteULongArrayOverflow:
            msg += " (overflow while writing ulong)";
	    break;

        case MinorWriteLongLongArrayOverflow:
            msg += " (overflow while writing longlong)";
	    break;

        case MinorWriteULongLongArrayOverflow:
            msg += " (overflow while writing ulonglong)";
	    break;

        case MinorWriteFloatArrayOverflow:
            msg += " (overflow while writing float)";
	    break;

        case MinorWriteDoubleArrayOverflow:
            msg += " (overflow while writing double)";
	    break;

        case MinorWriteLongDoubleArrayOverflow:
            msg += " (overflow while writing longdouble)";
	    break;

        case MinorWriteObjectLocal:
            msg += " (attempt to marshal a locality-constrained object)";
	    break;

        case MinorCountObjectLocal:
            msg += " (attempt to count a locality-constrained object)";
	    break;

        case MinorCountWCharInvalid:
            msg += " (invalid wchar)";
	    break;

        case MinorLongDoubleNotSupported:
            msg += " (long double is not supported)";
	    break;
	}
    }

    return msg._retn();
}

// ----------------------------------------------------------------------
// Print octets for debug reasons
// ----------------------------------------------------------------------

void
OB::PrintOctets(ostream& out, const Octet* oct, ULong count)
{
    static const ULong inc = 8;
    
    ULong j;

    for(j = 0 ; j < count ; j += inc)
    {
	ULong k;

	for(k = j ; k - j < inc ; k++)
	{
	    if(k < count)
	    {
		out.width(3);
		out << (int)oct[k] << ' ';
	    }
	    else
		out << "    ";
	}

	out << '"';
	
	for(k = j ; k < count && k - j < inc ; k++)
	{
	    if(isprint((char)oct[k]))
		out << (char)oct[k];
	    else
		out << '.';
	}
	
	out << '"' << endl;
    }
}

// ----------------------------------------------------------------------
// Print object reference
// ----------------------------------------------------------------------

void
OB::PrintObjref(ostream& out, Object_ptr p)
{
    const IOP::IOR& ior = p -> _OB_ior();
    PrintObjref(out, ior);
}

void
OB::PrintObjref(ostream& out, const char* s)
{
    assert_nca(s, NCANullString);
    
    if(strncmp(s, "IOR:", 4) != 0)
	throw INV_OBJREF();
    
    s += 4;
    
    Octet* buf = AsciiToOctets(s);
    InputStream in(buf, 0, strlen(s) / 2, false);
    
    Boolean endian = in.read_boolean();
    in._OB_swap(endian != Endian);

    IOP::IOR ior;
    IOP::IOR::_OB_unmarshal(ior, in);
    
    delete [] buf;

    out << "byteorder: " << (endian ? "little" : "big") << " endian\n";

    PrintObjref(out, ior);
}

void
OB::PrintObjref(ostream& out, const IOP::IOR& ior)
{
    ULong countIOP = 1;
    ULong countMulti = 1;

    out << "type_id: " << ior.type_id << '\n';

    for(ULong i = 0 ; i < ior.profiles.length() ; i++)
    {
	if(ior.profiles[i].tag == IOP::TAG_INTERNET_IOP)
	{
	    out << "IIOP profile #" << countIOP++ << ":\n";
	    IIOPPrintProfileBody(out, ior.profiles[i]);
	}
	else if(ior.profiles[i].tag == IOP::TAG_MULTIPLE_COMPONENTS)
	{
	    out << "Multiple component profile #" << countMulti++ << ": ";

	    const Octet* coct = ior.profiles[i].profile_data.get_buffer();
	    ULong len = ior.profiles[i].profile_data.length();
	    InputStream in(coct, 0, len, false);
	    in._OB_readEndian();

	    ULong cnt = in.read_ulong();
	    for(ULong j = 0 ; j < cnt ; j++)
	    {
		IOP::TaggedComponent comp;
		IOP::TaggedComponent::_OB_unmarshal(comp, in);

		if(comp.tag == IOP::TAG_CODE_SETS)
		{
		    out << "IOP::TAG_CODE_SETS" << endl;
		    PrintTagCodeSets(out, comp);
		}
		else
		    out << "Unknown tag: " << comp.tag << endl;
	    }
	}
    }

    out << flush;
}

// ----------------------------------------------------------------------
// Print code set information
// ----------------------------------------------------------------------

void
OB::PrintTagCodeSets(ostream& out, const IOP::TaggedComponent& component)
{
    const Octet* coct = component.component_data.get_buffer();
    ULong len = component.component_data.length();
    InputStream in(coct, 0, len, false);
    in._OB_readEndian();
    CONV_FRAME::CodeSetComponentInfo info;
    CONV_FRAME::CodeSetComponentInfo::_OB_unmarshal(info, in);

    CodeSetDatabase* db = CodeSetDatabase::instance();
    ULong i;

    //
    // Print char code set information
    //
    out << "Native char code set: " << endl;
    out << "  \"" << db -> idToName(info.ForCharData.native_code_set) << "\""
	<< endl;
    for(i = 0 ; i < info.ForCharData.conversion_code_sets.length() ; i++)
    {
	if(!i)
	    out << "Char conversion code sets:" << endl;

	out << "  \""
	    << db -> idToName(info.ForCharData.conversion_code_sets[i])
	    << "\"" << endl;
    }

    //
    // Print wchar code set information
    //
    out << "Native wchar code set: " << endl;
    out << "  \"" << db -> idToName(info.ForWcharData.native_code_set) << "\"" 
	<< endl;
    for(i = 0 ; i < info.ForWcharData.conversion_code_sets.length() ; i++)
    {
	if(!i)
	    out << "Wchar conversion code sets:" << endl;

	out << "  \""
	    << db -> idToName(info.ForWcharData.conversion_code_sets[i])
	    << "\"" << endl;
    }
}

// ----------------------------------------------------------------------
// Add code set information to IOR
// ----------------------------------------------------------------------

void
OB::IOPAddCodeSetInformation(IOP::IOR* ior, ULong native_cs, ULong native_wcs)
{
    ULong i;

    //
    // Ceate MultipleComponentProfile with code set information
    //
    IOP::MultipleComponentProfile multi;
    IOP::TaggedComponent_var info =
	CodeSetUtil::createTaggedComponentForCodeSets(native_cs, native_wcs);
    multi.append(info);

    //
    // Add MultipleComponentProfile to IOR
    //
    ULong index = ior -> profiles.length();
    ior -> profiles.length(index + 1);
    ior -> profiles[index].tag = IOP::TAG_MULTIPLE_COMPONENTS;
    CountStream count;
    count.count_boolean(); // Endian
    count.count_ulong();
    for(i = 0 ; i < multi.length() ; i++)
	IOP::TaggedComponent::_OB_marshalCount(multi[i], count);
    ior -> profiles[index].profile_data.length(count.count());
    Octet* oct = ior -> profiles[index].profile_data.get_buffer();
#ifdef OB_CLEAR_MEM
    memset(oct, 0, count.count());
#endif
    OutputStream out(oct, 0, count.count());
    out.write_boolean(Endian);
    out.write_ulong(multi.length());
    for(i = 0 ; i < multi.length() ; i++)
	IOP::TaggedComponent::_OB_marshal(multi[i], out);
}
