// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Template.h>
#include <OB/OBTypes.h>
#include <OB/Declarations.h>
#include <OB/IOP.h>
#include <OB/StubImplBase.h>
#include <OB/Object.h>
#include <OB/Any.h>
#include <OB/IOP.h>
#include <OB/IFR4.h>
#include <OB/Policy.h>
#include <OB/OBV.h>
#include <OB/ORB.h>
#include <OB/POAServantBase.h>
#include <OB/ValueBase.h>
#include <OB/AbstractBase.h>
#include <OB/Stream.h>

using namespace CORBA;
using namespace PortableServer;
using namespace OB;

// ----------------------------------------------------------------------
// Template instantiations
// ----------------------------------------------------------------------

#ifndef HAVE_NO_EXPLICIT_TEMPLATES
template class OB::ObjVar< ValueBase >;
#else
#ifdef HAVE_PRAGMA_DEFINE
#pragma define(OB::ObjVar< ValueBase >)
#endif
#endif

// ----------------------------------------------------------------------
// Static members
// ----------------------------------------------------------------------

const char* ValueBase::_ob_id_ = "IDL:omg.org/CORBA/ValueBase:1.0";
const char* ValueBase::_ob_truncatable_ids_[] =
{
    0
};

// ----------------------------------------------------------------------
// ValueBase constructor and destructor
// ----------------------------------------------------------------------

ValueBase::ValueBase()
{
}

ValueBase::ValueBase(const ValueBase&)
{
}
    
ValueBase::~ValueBase()
{
}

// ----------------------------------------------------------------------
// ValueBase public member functions
// ----------------------------------------------------------------------

ValueDef_ptr
ValueBase::_get_value_def()
    throw(SystemException)
{
    Object_var obj;

    try
    {
        ORB_ptr orb = ORB::_OB_instance_nodup();
        if(is_nil(orb))
            throw INITIALIZE("ORB not initialized");
        obj = orb -> resolve_initial_references("InterfaceRepository");
    }
    catch(const ORB::InvalidName&)
    {
        throw INTF_REPOS(MinorNoIntfRepos, COMPLETED_NO);
    }

    if(is_nil(obj))
    {
        throw INTF_REPOS(MinorNoIntfRepos, COMPLETED_NO);
    }

    Repository_var repository = Repository::_narrow(obj);

    if(is_nil(repository))
    {
        throw INTF_REPOS(MinorNoIntfRepos, COMPLETED_NO);
    }

    //
    // Obtain our dynamic type & attempt to find the Contained object
    //
    const char* id = _OB_typeId();
    Contained_var contained = repository -> lookup_id(id);

    if(is_nil(contained))
        return ValueDef::_nil();

    return ValueDef::_narrow(contained);
}

ValueBase*
ValueBase::_downcast(ValueBase* v)
{
    return v;
}

Boolean
ValueBase::_OB_isCustom() const
{
    return FALSE;
}

Boolean
ValueBase::_OB_isTruncatable() const
{
    return FALSE;
}

const char*
ValueBase::_OB_typeId() const
{
    return _ob_id_;
}

const char*
ValueBase::_OB_id()
{
    return _ob_id_;
}

const char**
ValueBase::_OB_truncatableIds() const
{
    return _ob_truncatable_ids_;
}

void
ValueBase::_OB_marshalValue(OutputStream&)
{
    // This should not be called
    assert(false);
}

void
ValueBase::_OB_marshalCountValue(CountStream&)
{
    // This should not be called
    assert(false);
}

void
ValueBase::_OB_unmarshalValue(InputStream&)
{
    // This should not be called
    assert(false);
}

void
ValueBase::_OB_marshal(ValueBase* v, OB::OutputStream& _ob_out)
{
    _ob_out.write_Value(v);
}

void
ValueBase::_OB_marshalCount(ValueBase* v, OB::CountStream& _ob_cnt)
{
    _ob_cnt.count_Value(v);
}

void
ValueBase::_OB_unmarshal(ValueBase*& v, OB::InputStream& _ob_in)
{
    v = _ob_in.read_Value();
}

//
// This will be overridden by valuetypes which support an abstract interface
//
AbstractBase_ptr
ValueBase::_OB_toAbstractBase()
{
    return AbstractBase::_nil();
}

void
OBRelease(CORBA::ValueBase* v)
{
    remove_ref(v);
}

void
OBDuplicate(CORBA::ValueBase* v)
{
    add_ref(v);
}

// ----------------------------------------------------------------------
// DefaultValueRefCountBase constructor and destructor
// ----------------------------------------------------------------------

DefaultValueRefCountBase::DefaultValueRefCountBase()
{
}

DefaultValueRefCountBase::DefaultValueRefCountBase(
    const DefaultValueRefCountBase&)
{
}
    
DefaultValueRefCountBase::~DefaultValueRefCountBase()
{
}

// ----------------------------------------------------------------------
// DefaultValueRefCountBase public member functions
// ----------------------------------------------------------------------

void
DefaultValueRefCountBase::_add_ref()
{
    _OB_incRef();
}

void
DefaultValueRefCountBase::_remove_ref()
{
    _OB_decRef();
}

ULong
DefaultValueRefCountBase::_refcount_value()
{
    return _OB_getRef();
}

// ----------------------------------------------------------------------
// ValueRefCountBase constructor and destructor
// ----------------------------------------------------------------------

ValueRefCountBase::ValueRefCountBase()
{
}

ValueRefCountBase::ValueRefCountBase(
    const ValueRefCountBase&)
{
}
    
ValueRefCountBase::~ValueRefCountBase()
{
}

// ----------------------------------------------------------------------
// ValueRefCountBase public member functions
// ----------------------------------------------------------------------

void
ValueRefCountBase::_add_ref()
{
    RefCountServantBase::_add_ref();
}

void
ValueRefCountBase::_remove_ref()
{
    RefCountServantBase::_remove_ref();
}

ULong
ValueRefCountBase::_refcount_value()
{
    return _OB_getRef();
}
