// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_WIN_REACTOR_H
#define OB_WIN_REACTOR_H

namespace OB
{

//
// The WinReactor class
//
class WinReactor : public Reactor
{
    //
    // Hide copy-constructor and assignment operator
    //
    WinReactor(const WinReactor&);
    void operator=(const WinReactor&);

    //
    // If set to true, event dispatching must be stopped
    //
    bool stop_;

    //
    // The window used for message notifications
    //
    HWND hWnd_;

    //
    // Struct for an event handler info
    //
    struct HandlerInfo
    {
	EventHandler* handler;
	Mask mask;
	Handle handle;
	HandlerInfo* next;
    };

    //
    // Head of the list with all event handler infos
    //
    HandlerInfo* handlerInfoList_;

    WinReactor(HINSTANCE);
    virtual ~WinReactor();

public:

    //
    // Get the reactor singleton
    //
    static Reactor* instance(HINSTANCE);

    //
    // Register and unregister event handler
    //
    virtual void registerHandler(EventHandler*, Mask, Handle);
    virtual void unregisterHandler(EventHandler*);

    //
    // Dispatch events
    //
    virtual void dispatch();
    virtual bool dispatchOneEvent(CORBA::Long);

    //
    // Test to see if an event is available
    //
    virtual bool eventReady();

    //
    // Stop event dispatching
    //
    virtual void stop() { stop_ = true; }

    //
    // Windows message processing function
    //
    LRESULT wndProc(HWND, UINT, WPARAM, LPARAM);
};

//
// The windows message viewer class
//
class WinViewer : public MessageViewer
{
    //
    // Hide copy-constructor and assignment operator
    //
    WinViewer(const WinViewer&);
    void operator=(const WinViewer&);

protected:

    WinViewer() { }
    virtual ~WinViewer() { }

public:

    //
    // Get the message viewer singleton
    //
    static MessageViewer* instance();

    //
    // Print messages
    //
    virtual void message(const char*);
};

} // End of namespace OB

#endif
