// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_X11_REACTOR_H
#define OB_X11_REACTOR_H

namespace OB
{

//
// The X11Reactor class
//
class X11Reactor : public Reactor
{
    //
    // Hide copy-constructor and assignment operator
    //
    X11Reactor(const X11Reactor&);
    void operator=(const X11Reactor&);

    //
    // If set to true, event dispatching must be stopped
    //
    bool stop_;

    //
    // The application context
    //
    XtAppContext context_;

    //
    // Struct for an event handler info
    //
    struct HandlerInfo
    {
	EventHandler* handler;
	Mask mask;
	Handle handle;
	XtInputId readId;
	XtInputId writeId;
	HandlerInfo* next;
    };

    //
    // Head of the list with all event handler infos
    //
    HandlerInfo* handlerInfoList_;

protected: // Make the constructor and destructor protected to avoid
           // stupid warning messages from some compilers
    X11Reactor(XtAppContext);
    virtual ~X11Reactor();

public:

    //
    // Get the reactor singleton
    //
    static Reactor* instance(XtAppContext);

    //
    // Register and unregister event handler
    //
    virtual void registerHandler(EventHandler*, Mask, Handle);
    virtual void unregisterHandler(EventHandler*);

    //
    // Dispatch events
    //
    virtual void dispatch();
    virtual bool dispatchOneEvent(CORBA::Long);

    //
    // Test to see if an event is available
    //
    virtual bool eventReady();

    //
    // Stop event dispatching
    //
    virtual void stop() { stop_ = true; }
};


} // End of namespace OB

#endif
