// **********************************************************************
//
// Copyright (c) 1997
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>

#include <stdlib.h>
#include <stdio.h>

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;
using namespace OB;

static void
usage(const char* progName)
{
    cerr << "Usage:\n";
    cerr << progName << " [options] scoped-name ...\n"
	"\n"
	"Options:\n"
	"-h, --help             Show this message.\n"
	"-v, --version          Show ORBacus version.\n"
	;
}

int
run(int argc, char* argv[])
{
    //
    // Create ORB
    //
    ORB_var orb = ORB_init(argc, argv);

    //
    // Set program name
    //
    const char* progName = argv[0];

    //
    // Get options
    //
    Long i;
    for(i = 1 ; i < argc && *argv[i] == '-' ; i++)
    {
	if(strcmp(argv[i], "--help") == 0 ||
	   strcmp(argv[i], "-h") == 0)
	{
	    usage(progName);
	    return 0;
	}
	else if(strcmp(argv[i], "--version") == 0 ||
		strcmp(argv[i], "-v") == 0)
	{
	    cerr << "ORBacus " << Version << endl;
	    return 0;
	}
	else
	{
	    cerr << progName << ": unknown option `"
		 << argv[i] << "'\n" << endl;
	    usage(progName);
	    return 1;
	}
    }

    Object_var obj;
    try
    {
        obj = orb -> resolve_initial_references("InterfaceRepository");
    }
    catch(ORB::InvalidName&)
    {
    }
    if(is_nil(obj))
    {
	cerr << "No `InterfaceRepository' initial reference." << endl;
	return 1;
    }

    Repository_var repository = Repository::_narrow(obj);
    if(is_nil(repository))
    {
	cerr << "`InterfaceRepository' initial reference is not a repository."
             << endl;
	return 1;
    }
    
    if(i == argc)
    {
	cerr << progName << ": no scoped name" << endl;
	usage(progName);
	return 1;
    }

    for( ; i < argc ; i++)
    {
	const char* scopedName = argv[i];

	Contained_var contained = repository -> lookup(scopedName);

	if(is_nil(contained))
	{
	    cerr << progName << ": warning: `" << scopedName
		 << "' does not exist" << endl;
	}
	else
	{
	    contained -> destroy();
	}
    }

    return 0;
}

int
main(int argc, char* argv[], char*[])
{
    int status;

    try
    {
	//
	// Generate code
	//
	status = run(argc, argv);
    }
    catch(SystemException& ex)
    {
	PrintException(ex);
	status = 1;
    }

    return status;
}
