// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/TemplateI.h>
#include <OB/Hashtable.h>
#include <OB/HashtableI.h>
#include <OB/Hashers.h>

#include <Types.h>
#include <Util.h>
#include <GenUtil.h>
#include <GenCPP.h>

#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;
using namespace OB;

#ifndef HAVE_NO_EXPLICIT_TEMPLATES
template class OB::Hashtable< String_var, String_var >;
#else
#ifdef HAVE_PRAGMA_DEFINE
#pramga define(OB::Hashtable( String_var, String_var >)
#endif
#endif

//
// The header
//
const char* IdlCPPGenerator::header_ =
"// **********************************************************************\n"
"//\n"
"// Generated by the ORBacus IDL-to-C++ Translator\n"
"//\n"
"// Copyright (c) 1999\n"
"// Object-Oriented Concepts, Inc.\n"
"// Billerica, MA, USA\n"
"//\n"
"// Copyright (c) 1999\n"
"// Object-Oriented Concepts GmbH\n"
"// Ettlingen, Germany\n"
"//\n"
"// All Rights Reserved\n"
"//\n"
"// **********************************************************************\n";

// ----------------------------------------------------------------------
// C++ code generator constructor and destructor
// ----------------------------------------------------------------------

IdlCPPGenerator::IdlCPPGenerator(const char* prog,
                                 ORB_ptr orb,
				 Repository_ptr rep,
				 bool noTC,
                                 bool local,
				 bool noVirtual,
                                 bool noSkel,
				 const char* thisHeaderDir,
				 const char* otherHeaderDir,
				 const char* dllImport,
				 const char* hSuff,
				 const char* cSuff)
    : prog_(prog),
      orb_(ORB::_duplicate(orb)),
      repository_(Repository::_duplicate(rep)),
      partial_(false),
      noTC_(noTC),
      local_(local),
      noVirtual_(noVirtual),
      noSkel_(noSkel),
      thisHeaderDir_(thisHeaderDir),
      otherHeaderDir_(otherHeaderDir),
      dllImport_(dllImport),
      hSuff_(hSuff),
      cSuff_(cSuff),
      absMap_(new IdlCPPGenerator::StringMap(1023)),
      uniMap_(new IdlCPPGenerator::StringMap(1023))
{
    if(local_)
	noTC_ = true;
}

IdlCPPGenerator::IdlCPPGenerator(const char* prog,
                                 ORB_ptr orb,
				 Repository_ptr rep,
				 const RepositoryIdSeq& idSeq,
				 const IdlStringSeq& includeSeq,
				 bool noTC,
                                 bool local,
				 bool noVirtual,
                                 bool noSkel,
				 const char* thisHeaderDir,
				 const char* otherHeaderDir,
				 const char* dllImport,
				 const char* hSuff,
				 const char* cSuff)
    : prog_(prog),
      orb_(ORB::_duplicate(orb)),
      repository_(Repository::_duplicate(rep)),
      partial_(true),
      idSeq_(idSeq),
      includeSeq_(includeSeq),
      noTC_(noTC),
      local_(local),
      noVirtual_(noVirtual),
      noSkel_(noSkel),
      thisHeaderDir_(thisHeaderDir),
      otherHeaderDir_(otherHeaderDir),
      dllImport_(dllImport),
      hSuff_(hSuff),
      cSuff_(cSuff),
      absMap_(new IdlCPPGenerator::StringMap(1023)),
      uniMap_(new IdlCPPGenerator::StringMap(1023))
{
    if(local_)
	noTC_ = true;
}

IdlCPPGenerator::~IdlCPPGenerator()
{
    delete absMap_;
    delete uniMap_;
}

// ----------------------------------------------------------------------
// Generate C++ code
// ----------------------------------------------------------------------

int
IdlCPPGenerator::gen(const char* dir, const char* name)
{
    //
    // Open output files
    //
    String_var outputFileH = dir;
    outputFileH += name;
    String_var outputFileC = dir;
    outputFileC += name;
    
    outputFileH += hSuff_;
    outputFileC += cSuff_;
    
    ofstream outH(outputFileH);
    if(!outH.good())
    {
	cerr << prog_ << ": can't open \"" << outputFileH
	     << "\": " << strerror(errno) << endl;
	
	return 1;
    }
    
    ofstream outC(outputFileC);
    if(!outC.good())
    {
	outH.close();

	cerr << prog_ << ": can't open \"" << outputFileC
	     << "\": " << strerror(errno) << endl;
	
	return 1;
    }

    //
    // Print headers
    //
    String_var macro = IdlPathToIdent(outputFileH);
    
    outH << header_;
    outH << "\n// Version: " << Version;
    outH << "\n// License: " << License;
    outH << '\n';
    outH << "\n#ifndef __" << macro << "__";
    outH << "\n#define __" << macro << "__";

    for(ULong i = 0 ; i < includeSeq_.length() ; i++)
    {
	if(i == 0)
	    outH << '\n';
	
	outH << "\n#include <";
	if(strlen(otherHeaderDir_) > 0)
	    outH << otherHeaderDir_ << '/';
	outH << includeSeq_[i] << hSuff_ << '>';
    }
    
    outC << header_;
    outC << "\n// Version: " << Version;
    outC << "\n// License: " << License;
    outC << '\n';
    outC << "\n#include <OB/CORBA.h>";
    outC << "\n#include <OB/TemplateI.h>";
    outC << "\n#include <";
    if(strlen(thisHeaderDir_) > 0)
	outC << thisHeaderDir_ << '/';
    outC << outputFileH << '>';

    //
    // Generate code
    //
    IdlPrettyPrint ppH(outH);
    IdlPrettyPrint ppC(outC);

    genCPlusPlusHFwd(repository_, ppH);
    genCPlusPlusHGlobalFwd(repository_, ppH);
    genCPlusPlusHVarFwd(repository_, ppH);
    if(!local_)
    {
        genCPlusPlusHStubImplFwd(repository_, ppH);
        genCPlusPlusHStubImplGlobalFwd(repository_, ppH);
        genCPlusPlusHStubImplVarFwd(repository_, ppH);
    }
    genCPlusPlusH(repository_, ppH);
    if(!local_)
        genCPlusPlusHStubImpl(repository_, ppH);
    genCPlusPlusHOBV(repository_, ppH);
    genCPlusPlusHGlobalNamespace(repository_, ppH);
    genCPlusPlusC(repository_, ppC);
    if(!local_)
        genCPlusPlusCStubImpl(repository_, ppC);
    genCPlusPlusCOBV(repository_, ppC);
    genCPlusPlusCDefs(repository_, ppC);
    
    //
    // That's it
    //
    outH << '\n';
    outH << "\n#endif";
    outH << '\n';
    outC << '\n';

    return 0;
}

int
IdlCPPGenerator::genSkel(const char* dir, const char* name)
{
    if(local_)
        return 0;
    
    assert(!noSkel_);
    
    //
    // Open output files
    //
    String_var outputFileH = dir;
    outputFileH += name;
    String_var outputFileSkelH = dir;
    outputFileSkelH += name;
    String_var outputFileSkelC = dir;
    outputFileSkelC += name;
    
    outputFileSkelH += "_skel";
    outputFileSkelC += "_skel";

    outputFileH += hSuff_;
    outputFileSkelH += hSuff_;
    outputFileSkelC += cSuff_;
   
    ofstream outSkelH(outputFileSkelH);
    if(!outSkelH.good())
    {
	cerr << prog_ << ": can't open \"" << outputFileSkelH
	     << "\": " << strerror(errno) << endl;
	
	return 1;
    }
    
    ofstream outSkelC(outputFileSkelC);
    if(!outSkelC.good())
    {
	outSkelH.close();

	cerr << prog_ << ": can't open \"" << outputFileSkelC
	     << "\": " << strerror(errno) << endl;
	
	return 1;
    }

    //
    // Print headers
    //
    String_var macroSkel = IdlPathToIdent(outputFileSkelH);

    outSkelH << header_;
    outSkelH << "\n// Version: " << Version;
    outSkelH << "\n// License: " << License;
    outSkelH << '\n';
    outSkelH << "\n#ifndef __" << macroSkel << "__";
    outSkelH << "\n#define __" << macroSkel << "__";
    outSkelH << '\n';
    outSkelH << "\n#include <";
    if(strlen(thisHeaderDir_) > 0)
	outSkelH << thisHeaderDir_ << '/';
    outSkelH << outputFileH << '>';
    
    for(ULong i = 0 ; i < includeSeq_.length() ; i++)
    {
	if(i == 0)
	    outSkelH << '\n';
	
	outSkelH << "\n#include <";
	if(strlen(otherHeaderDir_) > 0)
	    outSkelH << otherHeaderDir_ << '/';
	outSkelH << includeSeq_[i] << "_skel" << hSuff_ << '>';
    }

    outSkelC << header_;
    outSkelC << "\n// Version: " << Version;
    outSkelC << "\n// License: " << License;
    outSkelC << '\n';
    outSkelC << "\n#include <OB/CORBA.h>";
    outSkelC << "\n#include <";
    if(strlen(thisHeaderDir_) > 0)
	outSkelC << thisHeaderDir_ << '/';
    outSkelC << outputFileSkelH << '>';
    
    //
    // Generate code
    //
    IdlPrettyPrint ppSkelH(outSkelH);
    IdlPrettyPrint ppSkelC(outSkelC);

    genCPlusPlusSkelH(repository_, ppSkelH);
    genCPlusPlusSkelHStubImpl(repository_, ppSkelH);
    genCPlusPlusSkelC(repository_, ppSkelC);
    genCPlusPlusSkelCStubImpl(repository_, ppSkelC);
    
    //
    // That's it
    //
    outSkelH << '\n';
    outSkelH << "\n#endif";
    outSkelH << '\n';
    outSkelC << '\n';

    return 0;
}

int
IdlCPPGenerator::genTie(const char* dir, const char* name)
{
    if(local_)
	return 0;

    assert(!noSkel_);

    //
    // Open output files
    //
    String_var outputFileSkelH = dir;
    outputFileSkelH += name;
    String_var outputFileTieH = dir;
    outputFileTieH += name;
    
    outputFileSkelH += "_skel";
    outputFileTieH += "_skel_tie";

    outputFileSkelH += hSuff_;
    outputFileTieH += hSuff_;
    
    ofstream outTieH(outputFileTieH);
    if(!outTieH.good())
    {
	cerr << prog_ << ": can't open \"" << outputFileTieH
	     << "\": " << strerror(errno) << endl;
	
	return 1;
    }
    
    //
    // Print headers
    //
    String_var macroTie = IdlPathToIdent(outputFileTieH);

    outTieH << header_;
    outTieH << "\n// Version: " << Version;
    outTieH << "\n// License: " << License;
    outTieH << '\n';
    outTieH << "\n#ifndef __" << macroTie << "__";
    outTieH << "\n#define __" << macroTie << "__";
    outTieH << '\n';
    outTieH << "\n#include <";
    if(strlen(thisHeaderDir_) > 0)
	outTieH << thisHeaderDir_ << '/';
    outTieH << outputFileSkelH << '>';
    
    //
    // Generate code
    //
    IdlPrettyPrint ppTieH(outTieH);

    genCPlusPlusTieH(repository_, ppTieH);
    
    //
    // That's it
    //
    outTieH << '\n';
    outTieH << "\n#endif";
    outTieH << '\n';

    return 0;
}

int
IdlCPPGenerator::genImpl(const char* dir, const char* name, bool all)
{
    assert(!noSkel_);

    //
    // Open output files
    //
    String_var outputFileSkelH = dir;
    outputFileSkelH += name;
    String_var outputFileImplH = dir;
    outputFileImplH += name;
    String_var outputFileImplC = dir;
    outputFileImplC += name;
    
    outputFileSkelH += "_skel";
    outputFileImplH += "_impl";
    outputFileImplC += "_impl";

    outputFileSkelH += hSuff_;
    outputFileImplH += hSuff_;
    outputFileImplC += cSuff_;

    //
    // Do NOT overwrite existing files!
    //
    struct stat st;
    if(stat(outputFileImplH, &st) == 0 ||
       stat(outputFileImplC, &st) == 0)
    {
        cerr << prog_ << ": will not overwrite \"" << outputFileImplH
             << "\" or \"" << outputFileImplC << "\"" << endl;

        return 1;
    }

    ofstream outImplH(outputFileImplH);
    if(!outImplH.good())
    {
	cerr << prog_ << ": can't open \"" << outputFileImplH
	     << "\": " << strerror(errno) << endl;
	
	return 1;
    }
    
    ofstream outImplC(outputFileImplC);
    if(!outImplC.good())
    {
	cerr << prog_ << ": can't open \"" << outputFileImplC
	     << "\": " << strerror(errno) << endl;
	
	return 1;
    }

    //
    // Print headers
    //
    String_var macroImpl = IdlPathToIdent(outputFileImplH);

    outImplH << header_;
    outImplH << "\n// Version: " << Version;
    outImplH << "\n// License: " << License;
    outImplH << '\n';
    outImplH << "\n#ifndef __" << macroImpl << "__";
    outImplH << "\n#define __" << macroImpl << "__";
    outImplH << '\n';
    outImplH << "\n#include <";
    if(strlen(thisHeaderDir_) > 0)
	outImplH << thisHeaderDir_ << '/';
    outImplH << outputFileSkelH << '>';
    
    for(ULong i = 0 ; i < includeSeq_.length() ; i++)
    {
	if(i == 0)
	    outImplH << '\n';
	
	outImplH << "\n#include <";
	if(strlen(otherHeaderDir_) > 0)
	    outImplH << otherHeaderDir_ << '/';
	outImplH << includeSeq_[i] << "_skel" << hSuff_ << '>';
    }

    outImplC << header_;
    outImplC << "\n// Version: " << Version;
    outImplC << "\n// License: " << License;
    outImplC << '\n';
    outImplC << "\n#include <OB/CORBA.h>";
    outImplC << "\n#include <";
    if(strlen(thisHeaderDir_) > 0)
	outImplC << thisHeaderDir_ << '/';
    outImplC << outputFileImplH << '>';

    //
    // Generate code
    //
    IdlPrettyPrint ppImplH(outImplH);
    IdlPrettyPrint ppImplC(outImplC);

    genCPlusPlusImplH(repository_, all, ppImplH);
    genCPlusPlusImplC(repository_, all, ppImplC);
    
    //
    // That's it
    //
    outImplH << '\n';
    outImplH << "\n#endif";
    outImplH << '\n';
    outImplC << '\n';

    return 0;
}
