// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <Types.h>
#include <Util.h>
#include <GenUtil.h>
#include <GenCPP.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Generate code for the source file, general part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusC(Container_ptr container, IdlPrettyPrint& out)
{
    //
    // Get scope
    //
    String_var scope = getScope(container);

    //
    // Get container contents descriptions
    //
    Container::DescriptionSeq_var seq =
        container -> describe_contents(dk_all, true, -1);

    ULong i, j;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	Container_var cont = Container::_narrow(seq[i].contained_object);

	switch(seq[i].kind)
	{
	case dk_Interface:
        {
            //
            // Interface
            //
            InterfaceDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            Identifier_var name = fixKwd(desc -> name);
            ScopedName_var absolute = getAbsolute(desc -> id);

            comment(desc -> id, out);

            //
            // Template instantiations
            //
            out.sep();
            out.direct("\n#ifndef HAVE_NO_EXPLICIT_TEMPLATES");
            out << "\ntemplate class OB::ObjVar< " << absolute << " >;";
            out << "\ntemplate class OB::ObjOut< " << absolute << " >;";
            out << "\ntemplate class OB::ObjForSeq< " << absolute << " >;";
            out.direct("\n#else");
            out.direct("\n#ifdef HAVE_PRAGMA_DEFINE");
            out << "\n#pragma define(OB::ObjVar< " << absolute << " >)";
            out << "\n#pragma define(OB::ObjOut< " << absolute << " >)";
            out << "\n#pragma define(OB::ObjForSeq< " << absolute << " >)";
            out.direct("\n#endif"); // HAVE_PRAGMA_DEFINE
            out.direct("\n#endif"); // HAVE_NO_EXPLICIT_TEMPLATES

            InterfaceDef_var interf =
                InterfaceDef::_narrow(seq[i].contained_object);
            assert(!is_nil(interf));

            IdlBaseInterfaceInfoSeq allBaseInfo;
            IdlGetBaseInterfaces(interf, allBaseInfo, true);
            IdlStringSeq allBaseAbsolute;
            allBaseAbsolute.length(allBaseInfo.length());
            for(j = 0 ; j < allBaseInfo.length() ; j++)
            {
                //
                // Compute the absolute name of each base interface.
                // For an abstract base interface M1::M2::Foo, the
                // absolute name is M1::M2::OBAbstractImpl_Foo.
                //
                if(allBaseInfo[j].is_abstract)
                {
                    String_var baseAbs;
                    String_var baseName = allBaseInfo[j].intf -> name();
                    baseName = fixKwd(baseName);
                    Container_var c = allBaseInfo[j].intf -> defined_in();
                    baseAbs = getScope(c);
                    baseAbs += "OBAbstractImpl_";
                    baseAbs += baseName;
                    allBaseAbsolute[j] = baseAbs;
                }
                else
                    allBaseAbsolute[j] = getAbsolute(allBaseInfo[j].id);
            }

            String_var classAbs;
            if(desc -> is_abstract)
            {
                classAbs = scope;
                classAbs += "OBAbstractImpl_";
                classAbs += name;
            }
            else
                classAbs = absolute;

            //
            // _ob_ids_
            //
            out.sep();
            out << "\nconst char* " << classAbs << "::_ob_ids_[] =";
            out.start();
            out << "\n\"" << desc -> id << "\",";
            for(j = 0 ; j < allBaseInfo.length() ; j++)
            {
                out << "\n\"" << allBaseInfo[j].id << "\",";
            }
            out << "\n0";
            out.end();
            out << ';';
            
            //
            // Constructor and destructor
            //

            /*
             * Nothing to do here, constructor is defined inline,
             * destructor is not needed
             */

            if(!local_)
            {
                //
                // We have to use these mangled names to avoid a bug
                // in MSVC
                //
                String_var flat = getFlattenedName(absolute);
                String_var derivedStub = string_dup("_ob_stubImpl_");
                derivedStub += flat;
                derivedStub += "_";

                //
                // _OB_getStubImpl
                //
                out.sep();
                out << "\nOBStubImpl_" << absolute << "_ptr\n"
		    << classAbs << "::_OB_getStubImpl()";
                out.start();
                unsigned int indent = out.getIndent();
                out.setIndent(0);
                out << "\n#ifdef HAVE_JTC";
                out.setIndent(indent);
                out << "\nJTCSynchronized sync(_ob_implMutex_);";
                out.setIndent(0);
                out << "\n#endif";
                out.setIndent(indent);
                out.sep();
                out << "\nif(CORBA::is_nil(" << derivedStub << "))";
		out.start();
                out << "\nCORBA::ORB_ptr orb = "
                    << "CORBA::ORB::_OB_instance_nodup();";
                out << "\nif(CORBA::is_nil(orb))";
                out.inc();
                out << "\nthrow CORBA::INITIALIZE(\"ORB not initialized\");";
                out.dec();
                out << "\n_ob_stubImpl_ = orb -> _OB_getStubImpl(this);";
                out << '\n' << derivedStub << " = OBStubImpl_" << absolute
                    << "::_duplicate(dynamic_cast<OBStubImpl_"
                    << absolute << "*>(_ob_stubImpl_.in()));";
                out << "\nassert(" << derivedStub << ".in());";

                if(allBaseInfo.length() > 0)
                {
		    for(j = 0 ; j < allBaseInfo.length() ; j++)
		    {
			ScopedName_var baseAbsolute =
                            getAbsolute(allBaseInfo[j].id);
                        //
                        // We have to use these mangled names to avoid a bug
                        // in MSVC
                        //
                        String_var baseFlat = getFlattenedName(baseAbsolute);
                        String_var baseDerivedStub =
                            string_dup("_ob_stubImpl_");
                        baseDerivedStub += baseFlat;
                        baseDerivedStub += "_";
			out << '\n' << baseDerivedStub << " = OBStubImpl_"
			    << baseAbsolute << "::_duplicate(" << derivedStub
                            << ");";
		    }
                }
		out.end();
		out << "\nreturn OBStubImpl_" << absolute
		    << "::_duplicate(" << derivedStub << ");";

                out.end();

                //
                // _OB_clearStubImpl
                //
                out.sep();
                out << "\nvoid\n" << classAbs << "::_OB_clearStubImpl()";
                out.start();

                indent = out.getIndent();
                out.setIndent(0);
                out << "\n#ifdef HAVE_JTC";
                out.setIndent(indent);
                out << "\nJTCSynchronized sync(_ob_implMutex_);";
                out.setIndent(0);
                out << "\n#endif";
                out.setIndent(indent);
                out.sep();
                
                out << "\n_ob_stubImpl_ = OB::StubImplBase::_nil();";
                out << '\n' << derivedStub << " = OBStubImpl_" << absolute
                    << "::_nil();";

                if(allBaseInfo.length() > 0)
                {
		    for(j = 0; j < allBaseInfo.length(); j++)
		    {
			ScopedName_var baseAbsolute =
                            getAbsolute(allBaseInfo[j].id);
                        //
                        // We have to use these mangled names to avoid a bug
                        // in MSVC
                        //
                        String_var baseFlat = getFlattenedName(baseAbsolute);
                        String_var baseDerivedStub =
                            string_dup("_ob_stubImpl_");
                        baseDerivedStub += baseFlat;
                        baseDerivedStub += "_";
			out << '\n' << baseDerivedStub << " = OBStubImpl_"
			    << baseAbsolute << "::_nil();";
		    }
		}

                out.end();
            }
                
            //
            // release, is_nil, _duplicate, _nil
            //

            /*
             * Nothing to do here, all functions are defined inline
             */

            //
            // Duplicate and Release
            //
            out.sep();	    
            out << "\nvoid\nOBDuplicate(" << absolute << "_ptr p)";
            out.start();
            if(desc -> is_abstract)
            {
                out << '\n' << absolute << "::_duplicate(p);";
            }
            else
            {
                out << "\nif(p)";
                out.inc();
                out << "\np -> _OB_incRef();";
                out.dec();
            }
            out.end();

            out.sep();
            out << "\nvoid\nOBRelease(" << absolute << "_ptr p)";
            out.start();
            if(desc -> is_abstract)
            {
                out << "\nCORBA::release((CORBA::AbstractBase_ptr)p);";
            }
            else
            {
                out << "\nif(p)";
                out.inc();
                out << "\np -> _OB_decRef();";
                out.dec();
            }
            out.end();

            /*
            //
            // _is_a
            //
            out.sep();
            out << "\nCORBA::Boolean\n" << absolute
            << "::_is_a(const char* _ob_id)";
            out.start();
            out << "\nassert_nca(_ob_id, OB::NCANullString);";
            out.sep();
            out << "\nif(strcmp(\"" << desc -> id << "\", _ob_id) == 0)";
	    out.inc();
	    out << "\nreturn true;";
	    out.dec();
	    if(desc -> base_interfaces.length() == 0)
	    {
		out << "\nelse";
		out.inc();
		out << "\nreturn CORBA::Object::_is_a(_ob_id);";
		out.dec();
	    }
	    else
	    {
		for(j = 0 ; j < desc -> base_interfaces.length() ; j++)
		{
		    const char* id = desc -> base_interfaces[j];
		    ScopedName_var baseAbsolute = getAbsolute(id);
		    out << "\nelse if(" << baseAbsolute << "::_is_a(_ob_id))";
		    out.inc();
		    out << "\nreturn true;";
		    out.dec();
		}
		out << "\nelse";
		out.inc();
		out << "\nreturn false;";
		out.dec();
	    }
	    out.end();
            */

	    //
	    // _narrow
	    //
	    out.sep();
	    if(desc -> is_abstract)
	    {
	        out << '\n' << absolute << "_ptr\n" << absolute
		    << "::_narrow(CORBA::AbstractBase_ptr p)";
	    }
	    else
	    {
	        out << '\n' << absolute << "_ptr\n" << absolute
		    << "::_narrow(CORBA::Object_ptr p)";
	    }
	    out.start();
	    out << "\nif(!CORBA::is_nil(p))";
	    out.start();
	    out << '\n' << name << "_ptr v = dynamic_cast<"
		<< name << "_ptr>(p);";
	    out << "\nif(v)";
	    out.inc();
	    out << "\nreturn _duplicate(v);";
	    out.dec();
	    // XXX Can is_a be used for abstract interfaces?
	    if(!desc -> is_abstract && !local_)
	    {
                out.sep();
                out << "\nif(p -> _is_a(_ob_ids_[0]))";
                out.start();
                out << '\n' << name << "_ptr val = new " << name << ';';
                out << "\nval -> _OB_copyFrom(p);";
                out << "\nreturn val;";
                out.end();
	    }
	    out.end();
	    out.sep();
	    out << "\nreturn _nil();";
	    out.end();

            //
            // _OB_createStubImpl
            //
            if(!local_)
            {
                out.sep();
                out << "\nOB::StubImplBase*\n" << classAbs
                    << "::_OB_createStubImpl()";
                out.start();
		out << "\nreturn new OBStubImpl_" << absolute << ';';
                out.end();
            }

            //
            // _OB_ids
            //
            out.sep();
            out << "\nconst char**";
            out << '\n' << classAbs << "::_OB_ids() const";
            out.start();
            out << "\nreturn _ob_ids_;";
            out.end();

            //
            // _OB_staticIds
            //
            out.sep();
            out << "\nconst char**";
            out << '\n' << classAbs << "::_OB_staticIds()";
            out.start();
            out << "\nreturn _ob_ids_;";
            out.end();

	    //
	    // CDR functions
	    //
	    out.sep();
	    out << "\nvoid\n" << absolute << "::_OB_unmarshal(" << name
		<< "_ptr& _ob_v, OB::InputStream& _ob_in)";
	    out.start();
	    if(!local_)
	    {
		// Don't release here! Old value must be released on exit
		// of this function
		out << '\n' << name << "_var old = _ob_v;";
                if(desc -> is_abstract)
                {
                    out << "\nif(_ob_in.read_boolean())";
                    out.start();
		    out << "\nCORBA::Object_var p = _ob_in.read_Object();";
		    out.sep();
		    out << "\nif(!CORBA::is_nil(p))";
                    out.start();
		    out << '\n' << classAbs << "* _ob_obj = new "
                        << classAbs << ';';
		    out << "\n_ob_v = _ob_obj;";
		    out << "\n_ob_obj -> _OB_copyFrom(p);";
		    out.end();
		    out << "\nelse";
		    out.inc();
		    out << "\n_ob_v = " << name << "::_nil();";
		    out.dec();
		    out.end();
                    out << "\nelse";
                    out.start();
		    out << "\nCORBA::ValueBase_var p = _ob_in.read_Value();";
		    out.sep();
		    out << "\nif(p.in() != 0)";
                    out.start();
		    out << "\nCORBA::AbstractBase_ptr a = "
                        << "p -> _OB_toAbstractBase();";
                    out << "\nassert(a);";
		    out << "\n_ob_v = " << name << "::_narrow(a);";
		    out.end();
		    out << "\nelse";
		    out.inc();
		    out << "\n_ob_v = " << name << "::_nil();";
		    out.dec();
		    out.end();
                }
                else
                {
		    out << "\nCORBA::Object_var p = _ob_in.read_Object();";
		    out.sep();
		    out << "\nif(!CORBA::is_nil(p))";
		    out.start();
		    out << "\n_ob_v = new " << name << ';';
		    out << "\n_ob_v -> _OB_copyFrom(p);";
		    out.end();
		    out << "\nelse";
		    out.inc();
		    out << "\n_ob_v = " << name << "::_nil();";
		    out.dec();
                }
	    }
	    else
	    {
		out << "\nthrow CORBA::MARSHAL();";
	    }
	    out.end();

            //
            // _OB_toObject
            //
            if(desc -> is_abstract)
            {
                out.sep();
                out << "\nCORBA::Object_ptr" << '\n' << classAbs
                    << "::_OB_toObject()";
                out.start();
                out << "\nreturn this;";
                out.end();
            }

            //
            // If the interface is abstract, or derives from an abstract
            // interface, we need to generate some additional functions
            //
            bool hasAbstractBase = false;
            for(j = 0 ; j < allBaseInfo.length() ; j++)
            {
                if(allBaseInfo[j].is_abstract)
                {
                    hasAbstractBase = true;
                    break;
                }
            }

            if(desc -> is_abstract || hasAbstractBase)
            {
                out.sep();
                out << "\nCORBA::AbstractBase_ptr" << '\n' << classAbs
                    << "::_OB_toAbstractBase()";
                out.start();
                out << "\nreturn this;";
                out.end();

                out.sep();
                out << "\nvoid" << '\n' << classAbs
                    << "::_OB_insertAny(CORBA::Any& _ob_any)";
                out.start();
                if(!noTC_)
                    out << "\n_ob_any.replace(_tc_" << name
                        << ", (CORBA::Object_ptr)this, true);";
                else
                    out << "\nassert(false);";
                out.end();
            }

	    if(!noTC_)
	    {
                //XXX - fix for new stub structure
                
		//
		// Any insertion and extraction
		//
		out.sep();
		out << "\nvoid\noperator<<=(CORBA::Any& any, " << absolute
		    << "_ptr* v)";
		out.start();
                if(desc -> is_abstract)
                {
                    out << "\nif(!CORBA::is_nil(*v))";
                    out.inc();
                    out << "\n(*v) -> _OB_insertAny(any);";
                    out.dec();
                    out << "\nelse";
                    out.inc();
		    out << "\nany.replace(" << scope << "_tc_" << name
			<< ", 0, true);";
                    out.dec();
                }
                else
                {
		    out << "\nany.replace(" << scope << "_tc_" << name
			<< ", (CORBA::Object_ptr)*v, true);";
                }
		out.end();

		out.sep();
		out << "\nvoid\noperator<<=(CORBA::Any& any, " << absolute
		    << "_ptr v)";
		out.start();
		out << '\n' << absolute << "_ptr val = " << absolute
		    << "::_duplicate(v);";
		out << "\nany <<= &val;";
		out.end();

                if(!desc -> is_abstract)
                {
		    out.sep();
		    out << "\nCORBA::Boolean\noperator>>=("
                        << "const CORBA::Any& any, " << absolute << "_ptr& v)";
		    out.start();
		    out << "\nif(any.check_type("
			<< scope << "_tc_" << name << "))";
		    out.start();
		    out << "\nCORBA::Object_ptr val = "
			<< "(CORBA::Object_ptr)any.value();";
		    out.sep();
		    out << "\nif(!CORBA::is_nil(val))";
		    out.start();
		    out << "\nif(!(v = dynamic_cast<" << absolute
			<< "_ptr>(val)))";
		    out.start();
		    out << "\nv = new " << absolute << ';';
		    out << "\nv -> _OB_copyFrom(val);";
		    out << "\n(CORBA::Any&)any <<= &v;";
		    out.end();
		    out.end();
		    out << "\nelse";
		    out.inc();
		    out << "\nv = " << absolute << "::_nil();";
		    out.dec();
		    out.sep();
		    out << "\nreturn true;";
		    out.end();
		    out << "\nelse";
		    out.inc();
		    out << "\nreturn false;";
		    out.dec();
		    out.end();
                }
	    }
	    
	    break;
	}

        case dk_Value:
        {
            //
            // Valuetype
            //
            ValueDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

            ValueDef_var value = ValueDef::_narrow(cont);
            assert(! is_nil(value));

            if(!check(desc -> id))
                continue;

            Identifier_var name = fixKwd(desc -> name);
            ScopedName_var absolute = getAbsolute(desc -> id);

            comment(desc -> id, out);

            //
            // Template instantiations
            //
            out.sep();
            out.direct("\n#ifndef HAVE_NO_EXPLICIT_TEMPLATES");
            out << "\ntemplate class OB::ObjVar< " << absolute << " >;";
            out << "\ntemplate class OB::ObjOut< " << absolute << " >;";
            out << "\ntemplate class OB::ObjForSeq< " << absolute << " >;";
            out.direct("\n#else");
            out.direct("\n#ifdef HAVE_PRAGMA_DEFINE");
            out << "\n#pragma define(OB::ObjVar< " << absolute << " >)";
            out << "\n#pragma define(OB::ObjOut< " << absolute << " >)";
            out << "\n#pragma define(OB::ObjForSeq< " << absolute << " >)";
            out.direct("\n#endif"); // HAVE_PRAGMA_DEFINE
            out.direct("\n#endif"); // HAVE_NO_EXPLICIT_TEMPLATES

            //
            // _ob_id_
            //
            out.sep();
            out << "\nconst char* " << absolute << "::_ob_id_ = \""
                << desc -> id << "\";";

            //
            // _ob_truncatable_ids_ - if there are any truncatable bases,
            // then generate an array
            //
            ValueDefSeq truncBases;
            IdlGetAllTruncatableBaseValues(value, truncBases);
            if(truncBases.length() > 0)
            {
                out.sep();
                out << "\nconst char* " << absolute
                    << "::_ob_truncatable_ids_[] =";
                out.start();
                out << "\n\"" << desc -> id << "\",";
                for(j = 0 ; j < truncBases.length() ; j++)
                {
                    RepositoryId_var id = truncBases[j] -> id();
                    out << "\n\"" << id << "\",";
                }
                out << "\n0";
                out.end();
                out << ';';
            }

            //
            // Duplicate and Release
            //
            out.sep();            
            out << "\nvoid\nOBDuplicate(" << absolute << "* p)";
            out.start();
            out << "\nCORBA::add_ref(p);";
            out.end();

            out.sep();
            out << "\nvoid\nOBRelease(" << absolute << "* p)";
            out.start();
            out << "\nCORBA::remove_ref(p);";
            out.end();

            //
            // _downcast
            //
            out.sep();
            out << '\n' << absolute << "*\n" << absolute
                << "::_downcast(CORBA::ValueBase* p)";
            out.start();
            out << "\nreturn dynamic_cast<" << name << "*>(p);";
            out.end();

            //
            // _OB_isCustom
            //
            if(desc -> is_custom)
            {
                out.sep();
                out << "\nCORBA::Boolean\n" << absolute
                    << "::_OB_isCustom() const";
                out.start();
                out << "\nreturn CORBA::TRUE;";
                out.end();
            }

            //
            // _OB_isTruncatable
            //
            if(desc -> is_truncatable)
            {
                out.sep();
                out << "\nCORBA::Boolean\n" << absolute
                    << "::_OB_isTruncatable() const";
                out.start();
                out << "\nreturn CORBA::TRUE;";
                out.end();
            }

            //
            // _OB_typeId
            //
            out.sep();
            out << "\nconst char*\n" << absolute << "::_OB_typeId() const";
            out.start();
            out << "\nreturn _ob_id_;";
            out.end();

            //
            // _OB_id
            //
            out.sep();
            out << "\nconst char*\n" << absolute << "::_OB_id()";
            out.start();
            out << "\nreturn _ob_id_;";
            out.end();

            //
            // _OB_truncatableIds - if there are any truncatable bases, then
            // generate this function, otherwise we inherit the one from
            // ValueBase
            //
            if(truncBases.length() > 0)
            {
                out.sep();
                out << "\nconst char**";
                out << '\n' << absolute << "::_OB_truncatableIds() const";
                out.start();
                out << "\nreturn _ob_truncatable_ids_;";
                out.end();
            }

            if(!local_ && ! desc -> is_abstract)
            {
                //
                // CDR member functions
                //

                //
                // Determine if this value has a concrete base value
                //
                ValueDef_var base = value -> base_value();
                ScopedName_var baseAbs;
                if(! is_nil(base))
                {
                    RepositoryId_var bid = base -> id();
                    baseAbs = getAbsolute(bid);
                }

                //
                // Collect state members. We also need all inherited base
                // members for MSVC++.
                //
                Container::DescriptionSeq_var members =
                    value -> describe_contents(dk_ValueMember, TRUE, -1);
                ValueMemberSeq baseMembers;
                if(!is_nil(base))
                    IdlGetAllValueMembers(base, baseMembers);

                //
                // _OB_marshalValue
                //
                out.sep();
                out << "\nvoid\n" << absolute
                    << "::_OB_marshalValue(OB::OutputStream& _ob_out)";
                out.start();
                if(desc -> is_custom)
                {
                    //
                    // Custom marshalling
                    //
                    out << "\nCORBA::DataOutputStream_var _ob_dos = "
                        << "new CORBA::DataOutputStream_impl(_ob_out);";
                    out << "\nmarshal(_ob_dos);";
                }
                else
                {
                    ULong iter = 0;

                    //
                    // Marshal state members - if we have a concrete base
                    // value, then marshal its members first
                    //
                    if(! is_nil(base))
                    {
                        //
                        // Due to a bug in MSVC++, we have to generate
                        // the code for base value members in-line
                        //
                        out.direct("\n#ifdef HAVE_VCPLUSPLUS_BUGS");
                        for(j = 0 ; j < baseMembers.length() ; j++)
                        {
                            const ValueMember& vm = baseMembers[j];
                            String_var memberName = fixKwd(vm.name);
                            String_var arg = string_dup(memberName);
                            arg += "()";
                            String_var uniqueType = absolute;
                            uniqueType += "::OBUnique_";
                            uniqueType += vm.name;
                            writeReadWriteCount(scope, vm.type, arg, iter,
                                                RWCWriteIn, uniqueType, out);
                        }
                        out.direct("\n#else");
                        out << '\n' << baseAbs
                            << "::_OB_marshalValue(_ob_out);";
                        out.direct("\n#endif");
                    }

                    for(j = 0 ; j < members -> length() ; j++)
                    {
                        ValueMember* vm;
                        b = members[j].value >>= vm;
                        assert(b);
                        String_var memberName = fixKwd(vm -> name);
                        String_var arg = string_dup(memberName);
                        arg += "()";
                        String_var uniqueType = absolute;
                        uniqueType += "::OBUnique_";
                        uniqueType += vm -> name;
                        writeReadWriteCount(scope, vm -> type, arg, iter,
                                            RWCWriteIn, uniqueType, out);
                    }
                }
                out.end();

                //
                // _OB_marshalCountValue
                //
                out.sep();
                out << "\nvoid\n" << absolute
                    << "::_OB_marshalCountValue(OB::CountStream& _ob_cnt)";
                out.start();
                if(desc -> is_custom)
                {
                    //
                    // Add 8 bytes on an 8-byte alignment.  This allows
                    // some extra space for alignment, because we don't know
                    // in advance how large a custom valuetype will be.
                    //
                    out << "\n_ob_cnt.add(8, 8);";
                }
                else
                {
                    ULong iter = 0;

                    //
                    // Count state members - if we have a concrete base
                    // value, then count its members first
                    //
                    if(!is_nil(base))
                    {
                        //
                        // Due to a bug in MSVC++, we have to generate
                        // the code for base value members in-line
                        //
                        out.direct("\n#ifdef HAVE_VCPLUSPLUS_BUGS");
                        for(j = 0 ; j < baseMembers.length() ; j++)
                        {
                            const ValueMember& vm = baseMembers[j];
                            String_var memberName = fixKwd(vm.name);
                            String_var arg = string_dup(memberName);
                            arg += "()";
                            String_var uniqueType = absolute;
                            uniqueType += "::OBUnique_";
                            uniqueType += vm.name;
                            writeReadWriteCount(scope, vm.type, arg, iter,
                                                RWCCountIn, uniqueType, out);
                        }
                        out.direct("\n#else");
                        out << '\n' << baseAbs
                            << "::_OB_marshalCountValue(_ob_cnt);";
                        out.direct("\n#endif");
                    }

                    for(j = 0 ; j < members -> length() ; j++)
                    {
                        ValueMember* vm;
                        b = members[j].value >>= vm;
                        assert(b);
                        String_var memberName = fixKwd(vm -> name);
                        String_var arg = string_dup(memberName);
                        arg += "()";
                        String_var uniqueType = absolute;
                        uniqueType += "::OBUnique_";
                        uniqueType += vm -> name;
                        writeReadWriteCount(scope, vm -> type, arg, iter,
                                            RWCCountIn, uniqueType, out);
                    }
                }
                out.end();

                //
                // _OB_unmarshalValue
                //
                out.sep();
                out << "\nvoid\n" << absolute
                    << "::_OB_unmarshalValue(OB::InputStream& _ob_in)";
                out.start();
                if(desc -> is_custom)
                {
                    //
                    // Custom marshalling
                    //
                    out << "\nCORBA::DataInputStream_var _ob_dis = "
                        << "new CORBA::DataInputStream_impl(_ob_in);";
                    out << "\nunmarshal(_ob_dis);";
                }
                else
                {
                    ULong iter = 0;
                    ULong argIndex = 0;

                    //
                    // Unmarshal state members - if we have a concrete base
                    // value, then unmarshal its members first
                    //
                    if(!is_nil(base))
                    {
                        //
                        // Due to a bug in MSVC++, we have to generate
                        // the code for base value members in-line
                        //
                        out.direct("\n#ifdef HAVE_VCPLUSPLUS_BUGS");
                        for(j = 0 ; j < baseMembers.length() ; j++, argIndex++)
                        {
                            const ValueMember& vm = baseMembers[j];
                            String_var arg = string_dup("_ob_m");
                            arg += argIndex;
                            writeValueMemberUnmarshal(scope, absolute, vm.name,
                                                      arg, vm.type, iter, out);
                        }
                        out.direct("\n#else");
                        out << '\n' << baseAbs
                            << "::_OB_unmarshalValue(_ob_in);";
                        out.direct("\n#endif");
                    }

                    for(j = 0 ; j < members -> length() ; j++, argIndex++)
                    {
                        ValueMember* vm;
                        b = members[j].value >>= vm;
                        assert(b);
                        String_var arg = string_dup("_ob_m");
                        arg += argIndex;
                        writeValueMemberUnmarshal(scope, absolute, vm -> name,
                                                  arg, vm -> type, iter, out);
                    }
                }
                out.end();
            }

            if(!local_)
            {
                //
                // CDR static functions
                //

                //
                // _OB_marshal
                //
                out.sep();
                out << "\nvoid\n" << absolute << "::_OB_marshal("
                    << absolute << "* v, OB::OutputStream& out)";
                out.start();
                out << "\nout.write_Value(v, _OB_id());";
                out.end();

                //
                // _OB_marshalCount
                //
                out.sep();
                out << "\nvoid\n" << absolute << "::_OB_marshalCount("
                    << absolute << "* v, OB::CountStream& count)";
                out.start();
                out << "\ncount.count_Value(v, _OB_id());";
                out.end();

                //
                // _OB_unmarshal
                //
                out.sep();
                out << "\nvoid\n" << absolute << "::_OB_unmarshal("
                    << absolute << "*& v, OB::InputStream& in)";
                out.start();
                out << "\nCORBA::ValueBase* vb = in.read_Value(_OB_id());";
                out << "\nv = _downcast(vb);";
                out.end();
            }

            if(!noTC_)
            {
                //
                // Any insertion and extraction
                //
                out.sep();
                out << "\nvoid\noperator<<=(CORBA::Any& any, " << absolute
                    << "** v)";
                out.start();
                out << "\nany.replace(" << scope << "_tc_" << name
                    << ", (CORBA::ValueBase*)*v, true);";
                out.end();

                out.sep();
                out << "\nvoid\noperator<<=(CORBA::Any& any, " << absolute
                    << "* v)";
                out.start();
                out << "\nCORBA::add_ref(v);";
                out << "\nany <<= &v;";
                out.end();

                out.sep();
                out << "\nCORBA::Boolean\noperator>>=(const CORBA::Any& any, "
                    << absolute << "*& v)";
                out.start();
                out << "\nif(any.check_type(" << scope << "_tc_"
                    << name << "))";
                out.start();
                out << "\nCORBA::ValueBase* vb = "
                    << "(CORBA::ValueBase*)any.value();";
                out << "\nv = dynamic_cast<" << absolute << "*>(vb);";
                out << "\nreturn true;";
                out.end();
                out << "\nelse";
                out.inc();
                out << "\nreturn false;";
                out.dec();
                out.end();
	    }

            //
            // Generate type-specific value factory class code, if necessary
            //
            genCPlusPlusCValueFactory(value, out);

            break;
        }

        case dk_ValueBox:
        {
            //
            // Value box
            //
            TypeDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            ValueBoxDef_var value =
                ValueBoxDef::_narrow(seq[i].contained_object);
            assert(! is_nil(value));

            IDLType_var origTypeDef = value -> original_type_def();
            TypeCode_var type = origTypeDef -> type();

            Identifier_var name = fixKwd(desc -> name);
            ScopedName_var absolute = getAbsolute(desc -> id);

            comment(desc -> id, out);

            //
            // Template instantiations
            //
            out.sep();
            out.direct("\n#ifndef HAVE_NO_EXPLICIT_TEMPLATES");
            out << "\ntemplate class OB::ObjVar< " << absolute << " >;";
            out << "\ntemplate class OB::ObjOut< " << absolute << " >;";
            out << "\ntemplate class OB::ObjForSeq< " << absolute << " >;";
            out.direct("\n#else");
            out.direct("\n#ifdef HAVE_PRAGMA_DEFINE");
            out << "\n#pragma define(OB::ObjVar< " << absolute << " >)";
            out << "\n#pragma define(OB::ObjOut< " << absolute << " >)";
            out << "\n#pragma define(OB::ObjForSeq< " << absolute << " >)";
            out.direct("\n#endif"); // HAVE_PRAGMA_DEFINE
            out.direct("\n#endif"); // HAVE_NO_EXPLICIT_TEMPLATES

            //
            // If boxed type is an anonymous sequence, then we need to
            // generate additional template instantiations
            //
            if(type -> kind() == tk_sequence)
            {
                String_var uniqueType = absolute;
                uniqueType += "::OBUnique";

                String_var normType = getTypeString(scope, type,
                                                          GetTypeNormal, false,
                                                          0, uniqueType);
                String_var outType = getTypeString(scope, type,
                                                         GetTypeOut, false,
                                                         0, uniqueType);
                String_var varType = getTypeString(scope, type,
                                                         GetTypeVar, false,
                                                         0, uniqueType);

                out.sep();
                out.direct("\n#ifndef HAVE_NO_EXPLICIT_TEMPLATES");
                out << "\ntemplate class " << normType << ';';
                out << "\ntemplate class " << outType << ';';
                out << "\ntemplate class " << varType << ';';
                out.direct("\n#else");
                out.direct("\n#ifdef HAVE_PRAGMA_DEFINE");
                out << "\n#pragma define(" << normType << ')';
                out << "\n#pragma define(" << outType << ')';
                out << "\n#pragma define(" << varType << ')';
                out.direct("\n#endif"); // HAVE_PRAGMA_DEFINE
                out.direct("\n#endif"); // HAVE_NO_EXPLICIT_TEMPLATES
            }

            //
            // _ob_id_
            //
            out.sep();
            out << "\nconst char* " << absolute << "::_ob_id_ = \""
                << desc -> id << "\";";

            //
            // Duplicate and Release
            //
            out.sep();            
            out << "\nvoid\nOBDuplicate(" << absolute << "* p)";
            out.start();
            out << "\nCORBA::add_ref(p);";
            out.end();

            out.sep();
            out << "\nvoid\nOBRelease(" << absolute << "* p)";
            out.start();
            out << "\nCORBA::remove_ref(p);";
            out.end();

            //
            // _copy_value
            //
            out.sep();
            out << "\nCORBA::ValueBase*\n" << absolute << "::_copy_value()";
            out.start();
            out << "\nreturn new " << name << "(_boxed_in());";
            out.end();

            //
            // _downcast
            //
            out.sep();
            out << '\n' << absolute << "*\n" << absolute
                << "::_downcast(CORBA::ValueBase* p)";
            out.start();
            out << "\nreturn dynamic_cast<" << name << "*>(p);";
            out.end();

            //
            // _OB_typeId
            //
            out.sep();
            out << "\nconst char*\n" << absolute << "::_OB_typeId() const";
            out.start();
            out << "\nreturn _ob_id_;";
            out.end();

            //
            // _OB_id
            //
            out.sep();
            out << "\nconst char*\n" << absolute << "::_OB_id()";
            out.start();
            out << "\nreturn _ob_id_;";
            out.end();

            if(!local_)
            {
                //
                // CDR functions
                //
                ULong iter;
                String_var uniqueType = scope;
                uniqueType += "OBUnique_";
                uniqueType += desc -> name;

                String_var inArg = string_dup("_ob_v_");
                String_var inOutArg = string_dup("_ob_v_");

                TypeCode_var origType = GetOrigType(type);
                TCKind kind = origType -> kind();
                if(kind == tk_objref || kind == tk_sequence ||
                   kind == tk_struct || kind == tk_string ||
                   kind == tk_wstring || kind == tk_union ||
                   kind == tk_any)
                {
                    inArg += ".in()";
                    inOutArg += ".inout()";
                }

                //
                // _OB_marshalValue
                //
                out.sep();
                out << "\nvoid\n" << absolute
                    << "::_OB_marshalValue(OB::OutputStream& _ob_out)";
                out.start();
                //
                // Marshal underlying value
                //
                iter = 0;
                writeReadWriteCount(scope, type, inArg, iter,
                                    RWCWriteIn, uniqueType, out);
                out.end();

                //
                // _OB_marshalCountValue
                //
                out.sep();
                out << "\nvoid\n" << absolute
                    << "::_OB_marshalCountValue(OB::CountStream& _ob_cnt)";
                out.start();
                iter = 0;
                writeReadWriteCount(scope, type, inArg, iter,
                                    RWCCountIn, uniqueType, out);
                out.end();

                //
                // _OB_unmarshalValue
                //
                out.sep();
                out << "\nvoid\n" << absolute
                    << "::_OB_unmarshalValue(OB::InputStream& _ob_in)";
                out.start();
                iter = 0;
                writeReadWriteCount(scope, type, inOutArg, iter,
                                    RWCReadStructElem, uniqueType, out);
                out.end();

                //
                // CDR static functions
                //

                //
                // _OB_marshal
                //
                out.sep();
                out << "\nvoid\n" << absolute << "::_OB_marshal("
                    << absolute << "* v, OB::OutputStream& out)";
                out.start();
                out << "\nout.write_Value(v, _OB_id());";
                out.end();

                //
                // _OB_marshalCount
                //
                out.sep();
                out << "\nvoid\n" << absolute << "::_OB_marshalCount("
                    << absolute << "* v, OB::CountStream& count)";
                out.start();
                out << "\ncount.count_Value(v, _OB_id());";
                out.end();

                //
                // _OB_unmarshal
                //
                out.sep();
                out << "\nvoid\n" << absolute << "::_OB_unmarshal("
                    << absolute << "*& v, OB::InputStream& in)";
                out.start();
                out << "\nCORBA::ValueBase* vb = in.read_Value(_OB_id());";
                out << "\nv = _downcast(vb);";
                out.end();
            }

            if(!noTC_)
            {
                //
                // Any insertion and extraction
                //
                out.sep();
                out << "\nvoid\noperator<<=(CORBA::Any& any, " << absolute
                    << "** v)";
                out.start();
                out << "\nany.replace(" << scope << "_tc_" << name
                    << ", (CORBA::ValueBase*)*v, true);";
                out.end();

                out.sep();
                out << "\nvoid\noperator<<=(CORBA::Any& any, " << absolute
                    << "* v)";
                out.start();
                out << "\nCORBA::add_ref(v);";
                out << "\nany <<= &v;";
                out.end();

                out.sep();
                out << "\nCORBA::Boolean\noperator>>=(const CORBA::Any& any, "
                    << absolute << "*& v)";
                out.start();
                out << "\nif(any.check_type(" << scope << "_tc_"
                    << name << "))";
                out.start();
                out << "\nCORBA::ValueBase* vb = "
                    << "(CORBA::ValueBase*)any.value();";
                out << "\nv = dynamic_cast<" << absolute << "*>(vb);";
                out << "\nreturn true;";
                out.end();
                out << "\nelse";
                out.inc();
                out << "\nreturn false;";
                out.dec();
                out.end();
	    }

            break;
        }

	case dk_Alias:
	{
	    //
	    // Alias
	    //
	    TypeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    TypeCode_var contentType = desc -> type -> content_type();

	    if(contentType -> kind() == tk_sequence)
	    {
		comment(desc -> id, out);
		
		Identifier_var name = fixKwd(desc -> name);
		String_var absolute = getAbsolute(desc -> id);

		String_var uniqueType = scope;
		uniqueType += "OBUnique_";
		uniqueType += name;
		
		//
		// Template instantiations
		//
		String_var templ;
		for(ULong j = 0 ; j < 2 ; j++)
		{
		    if(j == 0)
		    {
			out.sep();
			out.direct("\n#ifndef HAVE_NO_EXPLICIT_TEMPLATES");
		    }
		    else
		    {
			out.direct("\n#else");
			out.direct("\n#ifdef HAVE_PRAGMA_DEFINE");
		    }


		    TypeCode_var iterType = contentType;
		    while(iterType -> kind() == tk_sequence)
		    {
			templ = getTypeString("", iterType, GetTypeNormal,
					      false, 0, uniqueType);
			out << "\ntemplate class " << templ << ';';
			iterType = iterType -> content_type();
		    }

		    templ = getTypeString("", contentType, GetTypeVar, false,
					  0, uniqueType);
		    out << "\ntemplate class " << templ << ';';

		    templ = getTypeString("", contentType, GetTypeOut, false,
					  0, uniqueType);
		    out << "\ntemplate class " << templ << ';';

		    if(j == 1)
		    {
			out.direct("\n#endif"); // HAVE_PRAGMA_DEFINE
			out.direct("\n#endif"); // HAVE_NO_EXPLICIT_TEMPLATES
		    }
		}

		if(!noTC_)
		{
		    //
		    // Any insertion and extraction
		    //
		    String_var infoType = scope;
		    infoType += "OBInfo_";
		    infoType += name;
		    ULong iter;
  

		    out.sep();
		    out << "\nvoid\n" << infoType
			<< "::marshal(const void* _ob_v, "
			<< "OB::OutputStream& _ob_out) const";
		    out.start();
                    out << "\nconst " << absolute << "& _ob_seq = *(const "
                        << absolute << "*)_ob_v;";
		    iter = 0;
		    writeReadWriteCount("", contentType, "_ob_seq", iter,
					RWCWriteIn, uniqueType, out);
		    out.end();

		    out.sep();
		    out << "\nvoid\n" << infoType
			<< "::marshalCount(const void* _ob_v, "
			<< "OB::CountStream& _ob_cnt) const";
		    out.start();
		    out << "\nconst " << absolute << "& _ob_seq = *(const "
			<< absolute << "*)_ob_v;";
		    iter = 0;
		    writeReadWriteCount("", contentType, "_ob_seq", iter,
					RWCCountIn, uniqueType, out);
		    out.end();

		    out.sep();
		    out << "\nvoid\n" << infoType
			<< "::unmarshal(void* _ob_v, "
			<< "OB::InputStream& _ob_in) const";
		    out.start();
                    out << '\n' << absolute << "& _ob_seq = *("
                        << absolute << "*)_ob_v;";
		    iter = 0;
		    writeReadWriteCount("", contentType, "_ob_seq", iter,
					RWCReadInOut, uniqueType, out);
		    out.end();

		    out.sep();
		    out << "\nvoid\noperator<<=(CORBA::Any& any, " << absolute
			<< "* v)";
		    out.start();
		    out << "\nstatic const " << scope << "OBInfo_" << name
                        << " info;";
		    out << "\nany.replace(" << scope << "_tc_" << name
			<< ", v, true, &info);";
		    out.end();

		    out.sep();
		    out << "\nvoid\noperator<<=(CORBA::Any& any, const "
			<< absolute << "& v)";
		    out.start();
		    out << "\nany <<= new " << absolute << "(v);";
		    out.end();

		    out.sep();
		    out << "\nCORBA::Boolean";
		    out << "\noperator>>=(const CORBA::Any& any, " << absolute
			<< "*& v)";
		    out.start();
		    out << "\nif(any.check_type(" << scope << "_tc_"
			<< name << "))";
		    out.start();
		    out << "\nif(!any.info())";
		    out.start();
		    out << "\nconst OB::Buffer* cbufp = "
			<< "(const OB::Buffer*)any.value();";
		    out << "\nOB::InputStream _ob_in(*cbufp, 0, false);";
		    out << '\n' << absolute << "* val = new " << absolute
			<< ';';
		    iter = 0;
		    writeReadWriteCount("", contentType, "(*val)", iter,
					RWCReadInOut, uniqueType, out);
		    out << "\n(CORBA::Any&)any <<= val;";
		    out.end();
		    out.sep();
		    out << "\nv = (" << absolute << "*)any.value();";
		    out << "\nreturn true;";
		    out.end();
		    out << "\nelse";
		    out.inc();
		    out << "\nreturn false;";
		    out.dec();
		    out.end();
		}
	    }
	    else if(contentType -> kind() == tk_array)
	    {
		comment(desc -> id, out);
		
		Identifier_var name = fixKwd(desc -> name);
		String_var absolute = getAbsolute(desc -> id);

		String_var uniqueType = scope;
		uniqueType += "OBUnique_";
		uniqueType += name;

		//
		// Template instantiations
		//
		String_var templ;
		out.sep();
		out.direct("\n#ifndef HAVE_NO_EXPLICIT_TEMPLATES");
		templ = getTypeString("", desc -> type, GetTypeVar, true,
				      0, uniqueType);
		out << "\ntemplate class " << templ << ';';
		templ = getTypeString("", desc -> type, GetTypeForAny, true,
				      0, uniqueType);
		out << "\ntemplate class " << templ << ';';
		out.direct("\n#else");
		out.direct("\n#ifdef HAVE_PRAGMA_DEFINE");
		templ = getTypeString("", desc -> type, GetTypeVar, true,
				      0, uniqueType);
		out << "\n#pragma define(" << templ << ");";
		templ = getTypeString("", desc -> type, GetTypeForAny, true,
				      0, uniqueType);
		out << "\n#pragma define(" << templ << ");";
		out.direct("\n#endif"); // HAVE_PRAGMA_DEFINE
		out.direct("\n#endif"); // HAVE_NO_EXPLICIT_TEMPLATES

		if(!noTC_)
		{
		    //
		    // Any insertion and extraction
		    //
		    String_var infoType = scope;
		    infoType += "OBInfo_";
		    infoType += name;

		    ULong iter;

		    out.sep();
		    out << "\nvoid\n" << infoType
			<< "::marshal(const void* _ob_v, "
			<< "OB::OutputStream& _ob_out) const";
		    out.start();
		    out << "\nconst " << absolute << "_slice* _ob_sl = (const "
			<< absolute << "_slice*)_ob_v;";
		    iter = 0;
		    writeReadWriteCount("", desc -> type, "_ob_sl", iter,
					RWCWriteIn, uniqueType, out);
		    out.end();

		    //
		    // Determine if we actually need the array in order
		    // to count its size - avoids compiler warnings about
		    // unused variables
		    //
		    bool needData = needDataForCount(desc -> type);
		    out.sep();
		    if(needData)
			out << "\nvoid\n" << infoType
			    << "::marshalCount(const void* _ob_v, "
			    << "OB::CountStream& _ob_cnt) const";
		    else
			out << "\nvoid\n" << infoType
			    << "::marshalCount(const void*, "
			    << "OB::CountStream& _ob_cnt) const";
		    out.start();
		    if(needData)
			out << "\nconst " << absolute << "_slice* _ob_sl = "
                            << "(const " << absolute << "_slice*)_ob_v;";
		    iter = 0;
		    writeReadWriteCount("", desc -> type, "_ob_sl", iter,
					RWCCountIn, uniqueType, out);
		    out.end();

		    out.sep();
		    out << "\nvoid\n" << infoType
			<< "::unmarshal(void* _ob_v, "
			<< "\nOB::InputStream& _ob_in) const";
		    out.start();
		    out << '\n' << absolute << "_slice* _ob_sl = ("
			<< absolute << "_slice*)_ob_v;";
		    iter = 0;
		    writeReadWriteCount("", desc -> type, "_ob_sl", iter,
					RWCReadInOut, uniqueType, out);
		    out.end();
		    
		    out.sep();
		    out << "\nvoid\noperator<<=(CORBA::Any& any, const "
			<< absolute << "_forany& r)";
		    out.start();
		    out << "\nconst " << absolute << "_slice* s;";
		    out.sep();
		    out << "\nif(r.nocopy())";
		    out.inc();
		    out << "\ns = r.in();";
		    out.dec();
		    out << "\nelse";
		    out.inc();
		    out.direct("\n#ifdef HAVE_VCPLUSPLUS_BUGS");
		    out << "\ns = (const " << absolute << "_slice*)"
			<< absolute << "_dup(r.in());";
		    out.direct("\n#else");
		    out << "\ns = " << absolute << "_dup(r);";
		    out.direct("\n#endif"); // HAVE_VCPLUSPLUS_BUGS
		    out.dec();
		    out.sep();
		    out << "\nstatic const " << infoType << " info;";
		    out << "\nany.replace(" << scope << "_tc_" << name
			<< ", (void*)s, true, &info);";
		    out.end();
		    
		    out.sep();
		    out << "\nCORBA::Boolean\noperator>>=("
			<< "const CORBA::Any& any, " << absolute
			<< "_forany& r)";
		    out.start();
		    out << "\nif(any.check_type(" << scope << "_tc_"
			<< name << "))";
		    out.start();
		    out << "\nif(!any.info())";
		    out.start();
		    out << "\nconst OB::Buffer* cbufp = "
			<< "(const OB::Buffer*)any.value();";
		    // TODO: Inconsistent naming: "_ob_in" vs. "s", "r", etc.
		    out << "\nOB::InputStream _ob_in(*cbufp, 0, false);";
		    out << '\n' << absolute << "_slice* s = " << absolute
			<< "_alloc();";
		    iter = 0;
		    writeReadWriteCount("", desc -> type, "s", iter,
					RWCReadInOut, uniqueType, out);
		    out << "\n(CORBA::Any&)any <<= " << absolute
			<< "_forany(s, true);";
		    out.end();
		    out.sep();
		    out << "\nr = (" << absolute << "_slice*)any.value();";
		    out << "\nreturn true;";
		    out.end();
		    out << "\nelse";
		    out.inc();
		    out << "\nreturn false;";
		    out.dec();
		    out.end();
		}
	    }

	    break;
	}

	case dk_Enum:
	{
	    //
	    // Enum
	    //
	    TypeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);
		
	    if(!check(desc -> id))
		continue;
		
	    Identifier_var name = fixKwd(desc -> name);
	    ScopedName_var absolute = getAbsolute(desc -> id);
		
	    comment(desc -> id, out);

	    if(!noTC_)
	    {
		//
		// Any insertion and extraction
		//
		out.sep();
		out << "\nvoid\noperator<<=(CORBA::Any& any, " << absolute
		    << " v)";
		out.start();
		out << "\nany.replace(" << scope << "_tc_" << name
		    << ", new CORBA::ULong((CORBA::ULong)v), true);";
		out.end();
		
		out.sep();
		out << "\nCORBA::Boolean\noperator>>=(const CORBA::Any& any, "
		    << absolute << "& v)";
		out.start();
		out << "\nif(any.check_type(" << scope << "_tc_" << name
		    << "))";
		out.start();
		out << "\nv = (" << absolute
		    << ")(*(CORBA::ULong*)any.value());";
		out << "\nreturn true;";
		out.end();
		out << "\nelse";
		out.inc();
		out << "\nreturn false;";
		out.dec();
		out.end();
	    }

	    break;
	}

	case dk_Struct:
	case dk_Exception:
	{
	    RepositoryId_var id;
	    Identifier_var name;
	    TypeCode_var type;
	    
	    if(seq[i].kind == dk_Struct)
	    {
		//
		// Struct
		//
		TypeDescription* desc;
		Boolean b = seq[i].value >>= desc;
		assert(b);

		name = fixKwd(desc -> name);
		id = desc -> id;
		type = desc -> type;
	    }
	    else
	    {
		//
		// Exception
		//
		ExceptionDescription* desc;
		Boolean b = seq[i].value >>= desc;
		assert(b);

		name = fixKwd(desc -> name);
		id = desc -> id;
		type = desc -> type;
	    }

	    if(!check(id))
		continue;
	    
	    ScopedName_var absolute = getAbsolute(id);
	    
	    comment(id, out);
	    
	    bool variable = IdlIsVariable(type);
	    
	    //
	    // Template instantiations
	    //
	    if(seq[i].kind == dk_Struct)
	    {
		out.sep();
		out.direct("\n#ifndef HAVE_NO_EXPLICIT_TEMPLATES");
		if(variable)
                {
		    out << "\ntemplate class OB::VarVar< "
                        << absolute << " >;";
                    out << "\ntemplate class OB::VarOut< "
                        << absolute << " >;";
                }
		else
                {
		    out << "\ntemplate class OB::FixVar< "
                        << absolute << " >;";
                }
		out.direct("\n#else");
		out.direct("\n#ifdef HAVE_PRAGMA_DEFINE");
		if(variable)
                {
		    out << "\n#pragma define(OB::VarVar< "
                        << absolute << " >)";
                    out << "\n#pragma define(OB::VarOut< "
                        << absolute << " >)";
                }
		else
                {
		    out << "\n#pragma define(OB::FixVar< "
                        << absolute << " >)";
                }
		out.direct("\n#endif"); // HAVE_PRAGMA_DEFINE
		out.direct("\n#endif"); // HAVE_NO_EXPLICIT_TEMPLATES
	    }
	    
	    if(variable)
	    {
		//
		// Constructor(s) and assignment operator
		//
		bool hasFixed = false;
		for(j = 0 ; j < type -> member_count() ; j++)
		{
		    TypeCode_var tc = type -> member_type(j);
		    if(!IdlIsVariable(tc))
		    {
			hasFixed = true;
			break;
		    }
		}
		
		if(hasFixed)
		{
		    out.sep();
		    out.direct("\n#ifdef OB_CLEAR_MEM");
		    out << '\n' << absolute << "::" << name << "()";
		    out.start();
		    for(j = 0 ; j < type -> member_count() ; j++)
		    {
			TypeCode_var tc = type -> member_type(j);
			if(!IdlIsVariable(tc))
			{
			    Identifier_var memberName =
				type -> member_name(j);
			    Identifier_var fixedMemberName =
				fixKwd(memberName);
			    out << "\nmemset(&" << fixedMemberName
				<< ", 0, sizeof(" << fixedMemberName << "));";
			}
		    }
		    out.end();
		    out.direct("\n#endif"); // OB_CLEAR_MEM
		}
		
		out.sep();
		out << '\n' << absolute << "::" << name << "(const "
		    << name << "& _ob_a)";
		out.inc();
		bool first = true;
		if(seq[i].kind == dk_Exception)
		{
		    out << "\n: CORBA::UserException(_ob_a)";
		    first = false;
		}
		for(j = 0 ; j < type -> member_count() ; j++)
		{
		    TypeCode_var tc = type -> member_type(j);
		    tc = GetOrigType(tc);
		    
		    if(tc -> kind() != tk_array)
		    {
			if(first)
			{
			    first = false;
			    out << "\n: ";
			}
			else
			    out << ",\n  ";
			
			Identifier_var memberName =
			    type -> member_name(j);
			Identifier_var fixedMemberName =
			    fixKwd(memberName);
			out << fixedMemberName << "(_ob_a." << fixedMemberName
			    << ')';
		    }
		}
		out.dec();

		out.start();
		for(j = 0 ; j < type -> member_count() ; j++)
		{
		    TypeCode_var memberType = type -> member_type(j);
		    TypeCode_var origMemberType =
			GetOrigType(memberType);
		    
		    if(origMemberType -> kind() == tk_array)
		    {
			Identifier_var memberName =
			    type -> member_name(j);
                        String_var target = fixKwd(memberName);
                        String_var source = string_dup("_ob_a.");
                        source += target;
                        String_var index = string_dup("_ob_i");
                        index += j;
                        writeArrayCopy(scope, memberType, target, source,
                                       index, out);
		    }
		}
		out.end();

		if(seq[i].kind == dk_Exception &&
		   type -> member_count() > 0)
		{
		    out.sep();
		    out << '\n' << absolute << "::" << name << '(';
		    unsigned int indent = out.getIndent();
		    out.setIndent(out.getPos());
		    for(j = 0 ; j < type -> member_count() ; j++)
		    {
			String_var arg = string_dup("_ob_a");
			arg += j;

			TypeCode_var memberType = type -> member_type(j);
			String_var s;
			if(memberType -> kind() == tk_sequence)
			{
			    String_var memberName = type -> member_name(j);
			    s = string_dup("const _");
			    s += memberName;
			    s += "_seq&";
			    s += ' ';
			    s += arg;
			}
			else
			    s = getTypeString(scope, memberType, GetTypeIn,
					      false, arg);
			
			if(j > 0)
			    out << ",\n";
			
			out << s;
		    }
		    out << ')';
		    out.setIndent(indent);
		    out.inc();
		    bool first = true;
		    for(j = 0 ; j < type -> member_count() ; j++)
		    {
			TypeCode_var memberType = type -> member_type(j);
			TypeCode_var tc = GetOrigType(memberType);
			
			if(tc -> kind() != tk_array)
			{
			    if(first)
			    {
				first = false;
				out << "\n: ";
			    }
			    else
				out << ",\n  ";
			    
			    Identifier_var memberName = type -> member_name(j);
			    Identifier_var fixedMemberName =
				fixKwd(memberName);

			    if(tc -> kind() == tk_objref ||
			       tc -> kind() == tk_Principal ||
			       tc -> kind() == tk_TypeCode)
			    {
				String_var s =
				    getTypeString(scope, memberType,
						  GetTypeNormal);
				out << fixedMemberName << '(' << s
				    << "::_duplicate(_ob_a" << j << "))";
			    }
			    else
				out << fixedMemberName << "(_ob_a" << j << ')';
			}
		    }
		    out.dec();
		    out.start();
		    for(j = 0 ; j < type -> member_count() ; j++)
		    {
			TypeCode_var memberType = type -> member_type(j);
			TypeCode_var origMemberType =
			    GetOrigType(memberType);
			
			if(origMemberType -> kind() == tk_array)
			{
			    Identifier_var memberName =
			        type -> member_name(j);
                            String_var target = fixKwd(memberName);
                            String_var source =
				string_dup("_ob_a");
			    source += j;
                            String_var index = string_dup("_ob_i");
                            index += j;
                            writeArrayCopy(scope, memberType, target, source,
                                           index, out);
			}
		    }
		    out.end();
		}

		out.sep();
		out << '\n' << absolute << "&\n" << absolute
		    << "::operator=(const " << name << '&';
		if(type -> member_count() > 0)
		    out << " _ob_a)";
		else
		    out << ')';
		out.start();
		if(type -> member_count() > 0)
		{
		    out << "\nif(this != &_ob_a)";
		    out.start();
		    
		    for(j = 0 ; j < type -> member_count() ; j++)
		    {
			TypeCode_var memberType = type -> member_type(j);
			TypeCode_var origMemberType = GetOrigType(memberType);
			
			Identifier_var memberName = type -> member_name(j);
			Identifier_var fixedMemberName = fixKwd(memberName);
			    
			if(origMemberType -> kind() == tk_array)
			{
			    Identifier_var memberName = type -> member_name(j);
                            String_var target = fixKwd(memberName);
                            String_var source = string_dup("_ob_a.");
                            source += target;
                            String_var index = string_dup("_ob_i");
                            index += j;
                            writeArrayCopy(scope, memberType, target, source,
                                           index, out);
			}
			else
			    out << '\n' << fixedMemberName << " = _ob_a."
				<< fixedMemberName << ';';
		    }

		    out.end();
		}
		out << "\nreturn *this;";
		out.end();
	    }

	    if(seq[i].kind == dk_Exception)
	    {
		//
		// _downcast
		//
		out.sep();
		out << '\n' << absolute << "*\n" << absolute
		    << "::_downcast(CORBA::Exception* p)";
		out.start();
		out << "\nreturn dynamic_cast<" << name << "*>(p);";
		out.end();

                //
		// _downcast (const version)
		//
		out.sep();
		out << "\nconst " << absolute << "*\n" << absolute
		    << "::_downcast(const CORBA::Exception* p)";
		out.start();
		out << "\nreturn dynamic_cast<const " << name << "*>(p);";
		out.end();
		
		//
		// _OB_clone
		//
		out.sep();
		out << "\nCORBA::Exception*\n" << absolute
		    << "::_OB_clone() const";
		out.start();
		out << "\nreturn new " << name << "(*this);";
		out.end();

                //
                // _OB_insertInAny
                //
		out.sep();
		out << "\nvoid\n" << absolute
		    << "::_OB_insertInAny(CORBA::Any& any)";
		out.start();
                if(!noTC_)
		    out << "\nany <<= this;";
		else
		    out << "\nassert_nca(false, OB::NCANoExceptionTC);";
		out.end();
	    }

	    if(!local_)
	    {
		//
		// CDR functions
		//
		ULong iter;

		out.sep();
		if(type -> member_count() > 0)
		    out << "\nvoid\n" << absolute << "::_OB_marshal(const "
                        << name << "& _ob_v, OB::OutputStream& _ob_out)";
		else
		    out << "\nvoid\n" << absolute << "::_OB_marshal(const "
			<< name << "&, OB::OutputStream& _ob_out)";
		out.start();
		if(seq[i].kind == dk_Exception)
		    out << "\n_ob_out.write_string(\"" << id << "\");";
		iter = 0;
		for(j = 0 ; j < type -> member_count() ; j++)
		{
		    Identifier_var memberName = type -> member_name(j);
		    Identifier_var fixedMemberName = fixKwd(memberName);
		    TypeCode_var memberType = type -> member_type(j);
		    String_var uniqueType = absolute;
		    uniqueType += "::OBUnique_";
		    uniqueType += memberName;
		    String_var arg = string_dup("_ob_v.");
		    arg += fixedMemberName;
		    writeReadWriteCount(scope, memberType, arg, iter,
					RWCWriteIn, uniqueType, out);
		}
		out.end();
		
		out.sep();
		if(type -> member_count() > 0)
		    out << "\nvoid\n" << absolute << "::_OB_marshalCount("
                        << "const " << name
			<< "& _ob_v, OB::CountStream& _ob_cnt)";
		else
		    out << "\nvoid\n" << absolute << "::_OB_marshalCount("
			<< "const " << name
                        << "&, OB::CountStream& _ob_cnt)";
		out.start();
		if(seq[i].kind == dk_Exception)
		    out << "\n_ob_cnt.count_string(\"" << id << "\");";
		iter = 0;
		for(j = 0 ; j < type -> member_count() ; j++)
		{
		    Identifier_var memberName = type -> member_name(j);
		    Identifier_var fixedMemberName = fixKwd(memberName);
		    TypeCode_var memberType = type -> member_type(j);
		    String_var uniqueType = absolute;
		    uniqueType += "::OBUnique_";
		    uniqueType += memberName;
		    String_var arg = string_dup("_ob_v.");
		    arg += fixedMemberName;
		    writeReadWriteCount(scope, memberType, arg, iter,
					RWCCountIn, uniqueType, out);
		}
		out.end();
		
		out.sep();
		if(type -> member_count() > 0)
		    out << "\nvoid\n" << absolute << "::_OB_unmarshal("
		        << name << "& _ob_v, OB::InputStream& _ob_in)";
		else
		    out << "\nvoid\n" << absolute << "::_OB_unmarshal("
		        << name << "&, OB::InputStream& _ob_in)";
		out.start();
		if(seq[i].kind == dk_Exception)
		{
		    out << "\nCORBA::String_var id = _ob_in.read_string();";
		    out << "\nassert(strcmp(id, \"" << id << "\") == 0);";
		}
		iter = 0;
		for(j = 0 ; j < type -> member_count() ; j++)
		{
		    Identifier_var memberName = type -> member_name(j);
		    Identifier_var fixedMemberName = fixKwd(memberName);
		    TypeCode_var memberType = type -> member_type(j);
		    String_var uniqueType = absolute;
		    uniqueType += "::OBUnique_";
		    uniqueType += memberName;
		    String_var arg = string_dup("_ob_v.");
		    arg += fixedMemberName;
		    writeReadWriteCount(scope, memberType, arg, iter,
					RWCReadStructElem, uniqueType, out);
		}
		out.end();
	    }

	    if(!noTC_)
	    {
		//
		// Any insertion and extraction
		//
		out.sep();
		out.direct("\n#ifndef HAVE_NO_EXPLICIT_TEMPLATES");
		out << "\ntemplate class OB::Info< " << absolute << " >;";
		out.direct("\n#else");
		out.direct("\n#ifdef HAVE_PRAGMA_DEFINE");
		out << "\n#pragma define(OB::Info< " << absolute << " >)";
		out.direct("\n#endif"); // HAVE_PRAGMA_DEFINE
		out.direct("\n#endif"); // HAVE_NO_EXPLICIT_TEMPLATES

		out.sep();
		out << "\nvoid\noperator<<=(CORBA::Any& any, " << absolute
		    << "* v)";
		out.start();
		out << "\nstatic const OB::Info< " << absolute << " > info;";
		out << "\nany.replace(" << scope << "_tc_" << name
		    << ", v, true, &info);";
		out.end();
		
		out.sep();
		out << "\nvoid\noperator<<=(CORBA::Any& any, const "
                    << absolute << "& v)";
		out.start();
		out << "\nany <<= new " << absolute << "(v);";
		out.end();
		
		out.sep();
		out << "\nCORBA::Boolean\noperator>>=(const CORBA::Any& any, "
		    << absolute << "*& v)";
		out.start();
		out << "\nif(any.check_type(" << scope << "_tc_" << name
		    << "))";
		out.start();
		out << "\nif(!any.info())";
		out.start();
		out << "\nconst OB::Buffer* cbufp = "
		    << "(const OB::Buffer*)any.value();";
		out << "\nOB::InputStream in(*cbufp, 0, false);";
		out << '\n' << absolute << "* val = new " << absolute << ';';
		out << '\n' << absolute << "::_OB_unmarshal(*val, in);";
		out << "\n(CORBA::Any&)any <<= val;";
		out.end();
		out.sep();
		out << "\nv = (" << absolute << "*)any.value();";
		out << "\nreturn true;";
		out.end();
		out << "\nelse";
		out.inc();
		out << "\nreturn false;";
		out.dec();
		out.end();
	    }

	    //
	    // Anonymous sequences template instantiations
	    //
	    for(j = 0 ; j < type -> member_count() ; j++)
	    {
		TypeCode_var memberType = type -> member_type(j);
		if(memberType -> kind() == tk_sequence)
		{
		    Identifier_var memberName = type -> member_name(j);

		    String_var uniqueType = absolute;
		    uniqueType += "::OBUnique_";
		    uniqueType += memberName;

		    String_var templ;

		    for(ULong k = 0 ; k < 2 ; k++)
		    {
			if(k == 0)
			{
			    out.sep();
			    out.direct("\n#ifndef HAVE_NO_EXPLICIT_TEMPLATES");
			}
			else
			{
			    out.direct("\n#else");
			    out.direct("\n#ifdef HAVE_PRAGMA_DEFINE");
			}
			
			TypeCode_var iterType = memberType;
			while(iterType -> kind() == tk_sequence)
			{
			    templ = getTypeString("", iterType, GetTypeNormal,
						  false, 0, uniqueType);
			    out << "\ntemplate class " << templ << ';';
			    iterType = iterType -> content_type();
			}
			
			if(k == 1)
			{
			    out.direct("\n#endif"); // HAVE_PRAGMA_DEFINE
                            // HAVE_NO_EXPLICIT_TEMPLATES
			    out.direct("\n#endif");
			}
		    }
		}
	    }

	    break;
	}

	case dk_Union:
	{
	    //
	    // Union
	    //
	    TypeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);
	    ScopedName_var absolute = getAbsolute(desc -> id);

	    TypeCode_var discType = desc -> type -> discriminator_type();
	    TypeCode_var origDiscType = GetOrigType(discType);

	    String_var disc = getTypeString(scope, discType, GetTypeNormal);

	    String_var absDisc = getTypeString("", discType, GetTypeNormal);

	    String_var cast;
	    String_var absCast;
	    if(origDiscType -> kind() == tk_enum)
	    {
		cast = string_alloc(1 + strlen(disc) + 1);
		strcpy(cast.inout(), "(");
		strcat(cast.inout(), disc);
		strcat(cast.inout(), ")");

		absCast = string_alloc(1 + strlen(absDisc) + 1);
		strcpy(absCast.inout(), "(");
		strcat(absCast.inout(), absDisc);
		strcat(absCast.inout(), ")");
	    }
	    else
	    {
		cast = string_dup("");
		absCast = string_dup("");
	    }

	    IdlUnionMemberInfoSeq_var infoSeq =	IdlGetUnionInfo(desc -> type);

	    IdlStringSeq fixedNames;
	    fixedNames.length(infoSeq -> length());
	    for(j = 0 ; j < infoSeq -> length() ; j++)
		fixedNames[j] = fixKwd(infoSeq[j].name);

	    bool variable = IdlIsVariable(desc -> type);

	    comment(desc -> id, out);

	    //
	    // Template instantiations
	    //
	    out.sep();
	    out.direct("\n#ifndef HAVE_NO_EXPLICIT_TEMPLATES");
	    if(variable)
            {
		out << "\ntemplate class OB::VarVar< " << absolute << " >;";
                out << "\ntemplate class OB::VarOut< " << absolute << " >;";
            }
	    else
            {
		out << "\ntemplate class OB::FixVar< " << absolute << " >;";
            }
	    out.direct("\n#else");
	    out.direct("\n#ifdef HAVE_PRAGMA_DEFINE");
	    if(variable)
            {
		out << "\n#pragma define(OB::VarVar< " << absolute << " >)";
                out << "\n#pragma define(OB::VarOut< " << absolute << " >)";
            }
	    else
            {
		out << "\n#pragma define(OB::FixVar< " << absolute << " >)";
            }
	    out.direct("\n#endif"); // HAVE_PRAGMA_DEFINE
	    out.direct("\n#endif"); // HAVE_NO_EXPLICIT_TEMPLATES
	    
	    //
	    // The remove function
	    //
	    if(variable)
	    {
		out.sep();
		out << "\nvoid\n" << absolute << "::_OB_remove()";
		out.start();
		out << "\nif(_ob_i_)";
		out.start();
		out << "\nswitch(_ob_d_)";
		out.start();
		for(j = 0; j < infoSeq -> length() ; j++)
		{
		    IdlUnionMemberInfo& info = infoSeq[j];
		    const char* fixedName = fixedNames[j];

		    if(j > 0)
			out.sep();
		    
		    printUnionLabels(info, out, cast);
		    
		    if(!is_nil(info.type))
		    {
			String_var norm;
			if(info.type -> kind() == tk_sequence)
			{
			    String_var anonName = string_dup("_");
			    anonName += info.name;
			    anonName += "_seq";
			    norm = anonName;
			}
			else
			    norm = getTypeString(scope, info.type,
						 GetTypeNormal);
			
			TypeCode_var origMemberType =
			    GetOrigType(info.type);
			
			switch(origMemberType -> kind())
			{
			case tk_null:
			case tk_void:
			case tk_except:
			case tk_native:
			    assert(false);
			    break;
			    
			case tk_short:
			case tk_long:
			case tk_longlong:
			case tk_ushort:
			case tk_ulong:
			case tk_ulonglong:
			case tk_float:
			case tk_double:
			case tk_longdouble:
			case tk_fixed:
			case tk_boolean:
			case tk_char:
			case tk_wchar:
			case tk_octet:
			case tk_enum:
			    break;
			    
			case tk_struct:
			case tk_union:
			case tk_sequence:
			case tk_any:
			    if(IdlIsVariable(origMemberType))
			    {
				out << "\ndelete _ob_v_." << fixedName << ';';
				// The following is not needed since
				// _ob_i is set to false
				//out << "\n_ob_v_." << fixedName << " = 0;";
			    }
			    break;
			    
			case tk_TypeCode:
			case tk_Principal:
			case tk_objref:
			case tk_abstract_interface:
			{
			    out << "\nCORBA::release(_ob_v_." << fixedName
				<< ");";
			    // The following is not needed since _ob_i
			    // is set to false
			    //norm = getTypeString(scope, info.type);
			    //out << "\n_ob_v_." << fixedName << " = " << norm
			    //	<< "::_nil();";
			    break;
			}
			    
			case tk_string:
			    
			    out << "\nCORBA::string_free(_ob_v_." << fixedName
				<< ");";
			    // The following is not needed since _ob_i
			    // is set to false
			    //out << "\n_ob_v_." << fixedName << " = 0;";
			    break;
			    
			case tk_wstring:
			    
			    out << "\nCORBA::wstring_free(_ob_v_." << fixedName
				<< ");";
			    // The following is not needed since _ob_i
			    // is set to false
			    //out << "\n_ob_v_." << fixedName << " = 0;";
			    break;
			    
			case tk_array:
			    if(IdlIsVariable(origMemberType))
			    {
				String_var arg =
				    string_dup("_ob_v_.");
				arg += fixedName;
				writeArrayFree(scope, info.type, arg, out);
				// The following is not needed since _ob_i
				// is set to false
			        //out << "\n_ob_v_." << fixedName << " = 0;";
			    }
			    break;

			case tk_value:
			case tk_value_box:

			    out << "\nCORBA::remove_ref(_ob_v_." << fixedName
				<< ");";
			    // The following is not needed since _ob_i
			    // is set to false
			    //out << "\n_ob_v_." << fixedName << " = 0;";
			    break;
			    
			case tk_alias:
			    assert(false);
			    break;
			}
		    }
		    
		    out << "\nbreak;";
		}
		out.end();
		out.sep();
		out << "\n_ob_i_ = false;";
		out.end();
		out.end();
	    }

	    //
	    // The check function
	    //
	    out.sep();
	    out << "\nbool\n" << absolute << "::_OB_check(" << disc
		<< " _ob_d) const";
	    out.start();
	    if(variable)
	    {
		out << "\nassert_nca(_ob_i_, OB::NCAUnionInit);";
		out.sep();
	    }
	    bool needSwitch = false;
	    for(j = 0 ; j < infoSeq -> length() ; j++)
	    {
		IdlUnionMemberInfo& info = infoSeq[j];
		if(info.nLabels.length() + info.pLabels.length() > 1 ||
		   info.isDefault)
		{
		    needSwitch = true;
		    break;
		}
	    }
	    if(needSwitch)
	    {
		out << '\n' << disc << " d[2];";
		out << "\nd[0] = _ob_d;";
		out << "\nd[1] = _ob_d_;";
		out.sep();
		out << "\nfor(int i = 0 ; i < 2 ; i++)";
		out.start();
		out << "\nswitch(d[i])";
		out.start();
		for(j = 0 ; j < infoSeq -> length() ; j++)
		{
		    IdlUnionMemberInfo& info = infoSeq[j];
		    
		    if(j > 0)
			out.sep();
		    
		    printUnionLabels(info, out, cast);
		    
		    if(info.nLabels.length() + info.pLabels.length() > 1 ||
		       info.isDefault)
		    {
			if(info.pLabels.length() > 0)
			    out << "\nd[i] = " << cast
				<< info.pLabels[0] << ';';
			else
			    out << "\nd[i] = " << cast << '-'
				<< info.nLabels[0] << ';';
		    }
		    
		    out << "\nbreak;";
		}
		out.end();
		out.end();
		out.sep();
		out << "\nreturn d[0] == d[1];";
	    }
	    else
	    {
		out << "\nreturn _ob_d == _ob_d_;";
	    }
	    out.end();

	    if(variable)
	    {
		//
		// Default constructor
		//
		out.sep();
		out.direct("\n#ifdef OB_CLEAR_MEM");
		out << '\n' << absolute << "::" << name << "()";
		out.inc();
		out << "\n: _ob_i_(false)";
		out.dec();
		out.start();
		out << "\nmemset(&_ob_v_, 0, sizeof(_ob_v_));";
		out << "\nmemset(&_ob_d_, 0, sizeof(_ob_d_));";
		out.end();
		out.direct("\n#endif"); // OB_CLEAR_MEM
		
		//
		// Copy constructor
		//
		out.sep();
		out << '\n' << absolute << "::" << name << "(const "
		    << name << "& _ob_a)";
		out.inc();
		out << "\n: _ob_i_(false)";
		out.dec();
		out.start();
		out << "\n*this = _ob_a;";
		out.end();
		
		//
		// Assignment operator
		//
		out.sep();
		out << '\n' << absolute << "&\n" << absolute << "::operator="
		    << "(const " << name << "& _ob_a)";
		out.start();
		out << "\nif(this != &_ob_a)";
		out.start();
		out << "\n_OB_remove();";
		out.sep();
		out << "\nif(_ob_a._ob_i_)";
		out.start();
		out << "\n_ob_i_ = true;";
		out << "\n_ob_d_ = _ob_a._ob_d_;";
		out.sep();
		out << "\nswitch(_ob_d_)";
		out.start();
		for(j = 0 ; j < infoSeq -> length() ; j++)
		{
		    IdlUnionMemberInfo& info = infoSeq[j];
		    const char* fixedName = fixedNames[j];

		    if(j > 0)
			out.sep();
		    
		    printUnionLabels(info, out, cast);
		    
		    if(!is_nil(info.type))
		    {
			String_var norm;
			if(info.type -> kind() == tk_sequence)
			{
			    String_var anonName = string_dup("_");
			    anonName += info.name;
			    anonName += "_seq";
			    norm = anonName;
			}
			else
			    norm = getTypeString(scope, info.type,
						 GetTypeNormal);
			
			TypeCode_var origMemberType = GetOrigType(info.type);
			
			out.dec();
			out.start();
			switch(origMemberType -> kind())
			{
			case tk_null:
			case tk_void:
			case tk_except:
			case tk_native:
			    assert(false);
			    break;
			    
			case tk_short:
			case tk_long:
			case tk_longlong:
			case tk_ushort:
			case tk_ulong:
			case tk_ulonglong:
			case tk_float:
			case tk_double:
			case tk_longdouble:
			case tk_fixed:
			case tk_boolean:
			case tk_char:
			case tk_wchar:
			case tk_octet:
			case tk_enum:
			    
			    out << "\n_ob_v_." << fixedName
				<< " = _ob_a._ob_v_." << fixedName << ';';
			    break;
			    
			case tk_struct:
			case tk_union:
			case tk_sequence:
			case tk_any:
			
			    if(IdlIsVariable(origMemberType))
			    {
				out << "\n_ob_v_." << fixedName
				    << " = new " << norm << "(*_ob_a._ob_v_."
				    << fixedName << ");";
			    }
			    else
			    {
				out << "\n_ob_v_." << fixedName
				    << " = _ob_a._ob_v_." << fixedName << ';';
			    }
			    break;
			    
			case tk_TypeCode:
			case tk_Principal:
			case tk_objref:
			case tk_abstract_interface:
			    
			    out << "\n_ob_v_." << fixedName << " = " << norm
				<< "::_duplicate(_ob_a._ob_v_."
				<< fixedName << ");";
			    break;
			    
			case tk_string:
			    
			    out << "\n_ob_v_." << fixedName
				<< " = CORBA::string_dup(_ob_a._ob_v_."
				<< fixedName << ");";
			    break;
			    
			case tk_wstring:
			    
			    out << "\n_ob_v_." << fixedName
				<< " = CORBA::wstring_dup(_ob_a._ob_v_."
				<< fixedName << ");";
			    break;
			    
			case tk_array:
			{
                            String_var target =
                                string_dup("_ob_v_.");
                            target += fixedName;
                            String_var source =
                                string_dup("_ob_a._ob_v_.");
                            source += fixedName;
			    if(IdlIsVariable(origMemberType))
                                writeArrayDup(scope, info.type, target,
					      source, "_ob_i", out);
			    else
                                writeArrayCopy(scope, info.type, target,
					       source, "_ob_i", out);
			    break;
			}

			case tk_value:
			case tk_value_box:

			    out << "\n_ob_v_." << fixedName << " = "
				<< "_ob_a._ob_v_." << fixedName << ';';
			    out << "\nCORBA::add_ref(_ob_v_." << fixedName
				<< ");";
			    break;
			    
			case tk_alias:
			    assert(false);
			    break;
			}
			out << "\nbreak;";
			out.end();
			out.inc();
		    }
		    else
			out << "\nbreak;";
		}
		out.end();
		out.end();
		out.end();
		out.sep();
		out << "\nreturn *this;";
		out.end();
	    }
	    
	    //
	    // Set and get discriminant value
	    //
	    out.sep();
	    out << "\nvoid\n" << absolute << "::_d(" << disc << " _ob_d)";
	    out.start();
	    out << "\nassert_nca(_OB_check(_ob_d), OB::NCADiscMismatch);";
	    out << "\n_ob_d_ = _ob_d;";
	    out.end();

	    out.sep();
	    out << '\n' << absDisc << '\n' << absolute << "::_d() const";
	    out.start();
	    if(variable)
		out << "\nassert_nca(_ob_i_, OB::NCAUnionInit);";
	    out << "\nreturn _ob_d_;";
	    out.end();

	    //
	    // Set and get functions
	    //
	    for(j = 0 ; j < infoSeq -> length() ; j++)
	    {
		IdlUnionMemberInfo& info = infoSeq[j];
		const char* fixedName = fixedNames[j];

		String_var label = cast;
		if(info.pLabels.length() > 0)
		{
		    label += info.pLabels[0];
		}
		else
		{
		    label += '-';
		    label += info.nLabels[0];
		}

		if(is_nil(info.type))
		{
		    out.sep();
		    out << "\nvoid\n" << absolute << "::_default()";
		    out.start();
		    if(variable)
		    {
			out << "\n_OB_remove();";
			out << "\n_ob_i_ = true;";
		    }
		    out << "\n_ob_d_ = " << label << ';';
		    out.end();
		}
		else
		{
		    String_var norm;
		    String_var absNorm;
		    if(info.type -> kind() == tk_sequence)
		    {
			String_var anonName = string_dup("_");
			anonName += info.name;
			anonName += "_seq";
			norm = anonName;
			absNorm = absolute;
			absNorm += "::";
			absNorm += anonName;
		    }
		    else
		    {
			norm = getTypeString(scope, info.type, GetTypeNormal);
			absNorm = getTypeString("", info.type, GetTypeNormal);
		    }

		    TypeCode_var origMemberType = GetOrigType(info.type);

		    switch(origMemberType -> kind())
		    {
		    case tk_null:
		    case tk_void:
		    case tk_except:
		    case tk_native:
			assert(false);
			break;
			
		    case tk_short:
		    case tk_long:
		    case tk_longlong:
		    case tk_ushort:
		    case tk_ulong:
		    case tk_ulonglong:
		    case tk_float:
		    case tk_double:
		    case tk_longdouble:
		    case tk_fixed:
		    case tk_boolean:
		    case tk_char:
		    case tk_wchar:
		    case tk_octet:
		    case tk_enum:
			
			out.sep();
			out << "\nvoid\n" << absolute << "::" << fixedName
			    << '(' << norm << " _ob_a)";
			out.start();
			if(variable)
			{
			    out << "\n_OB_remove();";
			    out << "\n_ob_i_ = true;";
			}
			out << "\n_ob_d_ = " << label << ';';
			out << "\n_ob_v_." << fixedName << " = _ob_a;";
			out.end();
			
			out.sep();
			out << '\n' << absNorm << '\n' << absolute << "::"
			    << fixedName << "() const";
			out.start();
			out << "\nassert_nca(_OB_check(" << label
			    << "), OB::NCADiscMismatch);";
			out << "\nreturn _ob_v_." << fixedName << ';';
			out.end();
			break;
			
		    case tk_struct:
		    case tk_union:
		    case tk_sequence:
		    case tk_any:
			
			out.sep();
			out << "\nvoid\n" << absolute << "::"
			    << fixedName << "(const " << norm << "& _ob_a)";
			out.start();
			if(variable)
			{
			    out << "\n_OB_remove();";
			    out << "\n_ob_i_ = true;";
			}
			out << "\n_ob_d_ = " << label << ';';
			if(IdlIsVariable(origMemberType))
			    out << "\n_ob_v_." << fixedName << " = new "
				<< norm	<< "(_ob_a);";
			else
			    out << "\n_ob_v_." << fixedName << " = _ob_a;";
			out.end();
			
			out.sep();
			out << "\nconst " << absNorm << "&\n" << absolute
			    << "::" << fixedName << "() const";
			out.start();
			out << "\nassert_nca(_OB_check(" << label
			    << "), OB::NCADiscMismatch);";
			if(IdlIsVariable(origMemberType))
			    out << "\nreturn *_ob_v_." << fixedName << ';';
			else
			    out << "\nreturn _ob_v_." << fixedName << ';';
			out.end();
			
			out.sep();
			out << '\n' << absNorm << "&\n" << absolute << "::"
			    << fixedName << "()";
			out.start();
			out << "\nassert_nca(_OB_check(" << label
			    << "), OB::NCADiscMismatch);";
			if(IdlIsVariable(origMemberType))
			    out << "\nreturn *_ob_v_." << fixedName << ';';
			else
			    out << "\nreturn _ob_v_." << fixedName << ';';
			out.end();
			
			break;
			
		    case tk_TypeCode:
		    case tk_Principal:
		    case tk_objref:
		    case tk_abstract_interface:
			
			out.sep();
			out << "\nvoid\n" << absolute << "::"
			    << fixedName << '(' << norm << "_ptr _ob_a)";
			out.start();
			if(variable)
			{
			    out << "\n_OB_remove();";
			    out << "\n_ob_i_ = true;";
			}
			out << "\n_ob_d_ = " << label << ';';
			out << "\n_ob_v_." << fixedName << " = " << norm
			    << "::_duplicate(_ob_a);";
			out.end();
			
			out.sep();
			out << '\n' << absNorm << "_ptr\n" << absolute << "::"
			    << fixedName << "() const";
			out.start();
			out << "\nassert_nca(_OB_check(" << label
			    << "), OB::NCADiscMismatch);";
			out << "\nreturn _ob_v_." << fixedName << ';';
			out.end();
			
			break;
			
		    case tk_string:
			
			out.sep();
			out << "\nvoid\n" << absolute << "::" << fixedName
			    << "(char* _ob_a)";
			out.start();
			if(variable)
			{
			    out << "\n_OB_remove();";
			    out << "\n_ob_i_ = true;";
			}
			out << "\n_ob_d_ = " << label << ';';
			out << "\n_ob_v_." << fixedName << " = _ob_a;";
			out.end();
			
			out.sep();
			out << "\nvoid\n" << absolute << "::"
			    << fixedName << "(const char* _ob_a)";
			out.start();
			if(variable)
			{
			    out << "\n_OB_remove();";
			    out << "\n_ob_i_ = true;";
			}
			out << "\n_ob_d_ = " << label << ';';
			out << "\n_ob_v_." << fixedName
			    << " = CORBA::string_dup(_ob_a);";
			out.end();
			
			out.sep();
			out << "\nvoid\n" << absolute << "::" << fixedName
			    << "(const CORBA::String_var& _ob_a)";
			out.start();
			if(variable)
			{
			    out << "\n_OB_remove();";
			    out << "\n_ob_i_ = true;";
			}
			out << "\n_ob_d_ = " << label << ';';
			out << "\n_ob_v_." << fixedName
			    << " = CORBA::string_dup(_ob_a);";
			out.end();
			
			out.sep();
			out << "\nconst char*\n" << absolute << "::"
			    << fixedName << "() const";
			out.start();
			out << "\nassert_nca(_OB_check(" << label
			    << "), OB::NCADiscMismatch);";
			out << "\nreturn _ob_v_." << fixedName << ';';
			out.end();
			
			break;
			
		    case tk_wstring:
			
			out.sep();
			out << "\nvoid\n" << absolute << "::" << fixedName
			    << "(wchar_t* _ob_a)";
			out.start();
			if(variable)
			{
			    out << "\n_OB_remove();";
			    out << "\n_ob_i_ = true;";
			}
			out << "\n_ob_d_ = " << label << ';';
			out << "\n_ob_v_." << fixedName << " = _ob_a;";
			out.end();
			
			out.sep();
			out << "\nvoid\n" << absolute << "::"
			    << fixedName << "(const wchar_t* _ob_a)";
			out.start();
			if(variable)
			{
			    out << "\n_OB_remove();";
			    out << "\n_ob_i_ = true;";
			}
			out << "\n_ob_d_ = " << label << ';';
			out << "\n_ob_v_." << fixedName
			    << " = CORBA::wstring_dup(_ob_a);";
			out.end();
			
			out.sep();
			out << "\nvoid\n" << absolute << "::" << fixedName
			    << "(const CORBA::WString_var& _ob_a)";
			out.start();
			if(variable)
			{
			    out << "\n_OB_remove();";
			    out << "\n_ob_i_ = true;";
			}
			out << "\n_ob_d_ = " << label << ';';
			out << "\n_ob_v_." << fixedName
			    << " = CORBA::wstring_dup(_ob_a);";
			out.end();
			
			out.sep();
			out << "\nconst wchar_t*\n" << absolute << "::"
			    << fixedName << "() const";
			out.start();
			out << "\nassert_nca(_OB_check(" << label
			    << "), OB::NCADiscMismatch);";
			out << "\nreturn _ob_v_." << fixedName << ';';
			out.end();
			
			break;
			
		    case tk_array:
		    {
			String_var slice;
			String_var absSlice;
                        String_var in =
                            getTypeString(scope, info.type, GetTypeNormal,
                                          false, "_ob_a");

			if(info.type -> kind() == tk_alias)
			{
			    slice = getTypeString(scope, info.type,
						  GetTypeSlice);

			    absSlice = getTypeString("", info.type,
						     GetTypeSlice);
			}
			else
			{
			    const char* sl = "_slice";
			    slice = string_alloc(1 + strlen(info.name)
						 + strlen(sl));
			    strcpy(slice.inout(), "_");
			    strcat(slice.inout(), fixedName);
			    strcat(slice.inout(), sl);

			    absSlice = string_alloc(strlen(absolute)
                                                    + 2 + strlen(slice));
			    strcpy(absSlice.inout(), absolute);
			    strcat(absSlice.inout(), "::");
			    strcat(absSlice.inout(), slice);
			}
			
			out.sep();
			out << "\nvoid\n" << absolute << "::"
			    << fixedName << '(' << in << ')';
			out.start();
			if(variable)
			{
			    out << "\n_OB_remove();";
			    out << "\n_ob_i_ = true;";
			}
			out << "\n_ob_d_ = " << label << ';';
                        String_var target =
                            string_dup("_ob_v_.");
                        target += fixedName;
                        if(IdlIsVariable(origMemberType))
                            writeArrayDup(scope, info.type, target,
                                          "_ob_a", "_ob_i", out);
                        else
                            writeArrayCopy(scope, info.type, target,
                                           "_ob_a", "_ob_i", out);
			out.end();

			out.sep();
			out << '\n' << absSlice << "*\n" << absolute
			    << "::" << fixedName << "() const";
			out.start();
			out << "\nassert_nca(_OB_check(" << label
			    << "), OB::NCADiscMismatch);";
			out << "\nreturn (" << slice << "*)_ob_v_."
			    << fixedName << ';';
			out.end();
			
			break;
		    }

		    case tk_value:
		    case tk_value_box:

			out.sep();
			out << "\nvoid\n" << absolute << "::"
			    << fixedName << '(' << norm << "* _ob_a)";
			out.start();
			if(variable)
			{
			    out << "\n_OB_remove();";
			    out << "\n_ob_i_ = true;";
			}
			out << "\n_ob_d_ = " << label << ';';
			out << "\n_ob_v_." << fixedName << " = _ob_a;";
			out << "\nCORBA::add_ref(_ob_v_." << fixedName << ");";
			out.end();
			
			out.sep();
			out << '\n' << absNorm << "*\n" << absolute << "::"
			    << fixedName << "() const";
			out.start();
			out << "\nassert_nca(_OB_check(" << label
			    << "), OB::NCADiscMismatch);";
			out << "\nreturn _ob_v_." << fixedName << ';';
			out.end();
			
			break;

		    case tk_alias:
			assert(false);
			break;
		    }
		}
	    }

	    if(!local_)
	    {
		//
		// CDR functions
		//
		ULong iter;

		out.sep();
		out << "\nvoid\n" << absolute << "::_OB_marshal(const "
                    << name << "& _ob_v, OB::OutputStream& _ob_out)";
		out.start();
		if(variable)
		    out << "\nassert_nca(_ob_v._ob_i_, OB::NCAUnionInit);";
		iter = 0;
		writeReadWriteCount(scope, discType, "_ob_v._ob_d_", iter,
				    RWCWriteIn, 0, out);
		out.sep();
		out << "\nswitch(_ob_v._ob_d_)";
		out.start();
		for(j = 0 ; j < infoSeq -> length() ; j++)
		{
		    IdlUnionMemberInfo& info = infoSeq[j];
		    const char* fixedName = fixedNames[j];
		    
		    if(j > 0)
			out.sep();
		    
		    printUnionLabels(info, out, absCast);
		    
		    if(!is_nil(info.type))
		    {
			String_var uniqueType = absolute;
			uniqueType += "::OBUnique_";
			uniqueType += info.name;

			TypeCode_var origMemberType =
			    GetOrigType(info.type);
                        String_var fixedMemberName =
                            string_dup("_ob_v._ob_v_.");
                        fixedMemberName += fixedName;

			out.dec();
			out.start();
			iter = 0;
			if(IdlIsVariable(origMemberType) &&
			   (origMemberType -> kind() == tk_any ||
			    origMemberType -> kind() == tk_struct ||
			    origMemberType -> kind() == tk_union ||
			    origMemberType -> kind() == tk_sequence))
			{
			    String_var deref = string_dup("(*");
			    deref += fixedMemberName;
			    deref += ')';
			    writeReadWriteCount(scope, info.type, deref,
						iter, RWCWriteIn, uniqueType,
						out);
			}
			else
			{
			    writeReadWriteCount(scope, info.type,
                                                fixedMemberName, iter,
                                                RWCWriteIn, uniqueType, out);
			}
			out << "\nbreak;";
			out.end();
			out.inc();
		    }
		    else
			out << "\nbreak;";
		}
		out.end();
		out.end();
		
		out.sep();
		out << "\nvoid\n" << absolute << "::_OB_marshalCount(const "
                    << name << "& _ob_v, OB::CountStream& _ob_cnt)";
		out.start();
		if(variable)
		    out << "\nassert_nca(_ob_v._ob_i_, OB::NCAUnionInit);";
		iter = 0;
		writeReadWriteCount(scope, discType, "_ob_v._ob_d_", iter,
				    RWCCountIn, 0, out);
		out.sep();
		out << "\nswitch(_ob_v._ob_d_)";
		out.start();
		for(j = 0 ; j < infoSeq -> length() ; j++)
		{
		    IdlUnionMemberInfo& info = infoSeq[j];
		    const char* fixedName = fixedNames[j];
		    
		    if(j > 0)
			out.sep();
		    
		    printUnionLabels(info, out, absCast);
		    
		    if(!is_nil(info.type))
		    {
			String_var uniqueType = absolute;
			uniqueType += "::OBUnique_";
			uniqueType += info.name;

			TypeCode_var origMemberType =
			    GetOrigType(info.type);
                        String_var fixedMemberName =
                            string_dup("_ob_v._ob_v_.");
                        fixedMemberName += fixedName;

			out.dec();
			out.start();
			iter = 0;
			if(IdlIsVariable(origMemberType) &&
			   (origMemberType -> kind() == tk_any ||
			    origMemberType -> kind() == tk_struct ||
			    origMemberType -> kind() == tk_union ||
			    origMemberType -> kind() == tk_sequence))
			{
			    String_var deref = string_dup("(*");
			    deref += fixedMemberName;
			    deref += ')';
			    writeReadWriteCount(scope, info.type, deref,
						iter, RWCCountIn, uniqueType,
						out);
			}
			else
			{
			    writeReadWriteCount(scope, info.type,
                                                fixedMemberName, iter,
                                                RWCCountIn, uniqueType, out);
			}
			out << "\nbreak;";
			out.end();
			out.inc();
		    }
		    else
			out << "\nbreak;";
		}
		out.end();
		out.end();
		
		out.sep();
		out << "\nvoid\n" << absolute << "::_OB_unmarshal(" << name
		    << "& _ob_v, OB::InputStream& _ob_in)";
		out.start();
		if(variable)
		{
		    out << "\n_ob_v._OB_remove();";
		    out.sep();
		}
		if(variable)
		    out << "\n_ob_v._ob_i_ = true;";
		iter = 0;
		writeReadWriteCount(scope, discType, "_ob_v._ob_d_", iter,
				    RWCReadInOut, 0, out);
		out.sep();
		out << "\nswitch(_ob_v._ob_d_)";
		out.start();
		for(j = 0 ; j < infoSeq -> length() ; j++)
		{
		    IdlUnionMemberInfo& info = infoSeq[j];
		    const char* fixedName = fixedNames[j];
		    
		    if(j > 0)
			out.sep();
		    
		    printUnionLabels(info, out, absCast);
		    
		    if(!is_nil(info.type))
		    {
			String_var uniqueType = absolute;
			uniqueType += "::OBUnique_";
			uniqueType += info.name;

                        String_var fixedMemberName =
                            string_dup("_ob_v._ob_v_.");
                        fixedMemberName += fixedName;

			out.dec();
			out.start();
			iter = 0;
			writeReadWriteCount(scope, info.type, fixedMemberName,
					    iter, RWCReadOut, uniqueType, out);
			out << "\nbreak;";
			out.end();
			out.inc();
		    }
		    else
			out << "\nbreak;";
		}
		out.end();
		out.end();
	    }
	    
	    if(!noTC_)
	    {
		//
		// Any insertion and extraction
		//
		out.sep();
		out.direct("\n#ifndef HAVE_NO_EXPLICIT_TEMPLATES");
		out << "\ntemplate class OB::Info< " << absolute << " >;";
		out.direct("\n#else");
		out.direct("\n#ifdef HAVE_PRAGMA_DEFINE");
		out << "\n#pragma define(OB::Info< " << absolute << " >)";
		out.direct("\n#endif"); // HAVE_PRAGMA_DEFINE
		out.direct("\n#endif"); // HAVE_NO_EXPLICIT_TEMPLATES

		out.sep();
		out << "\nvoid\noperator<<=(CORBA::Any& any, " << absolute
		    << "* v)";
		out.start();
		out << "\nstatic const OB::Info< " << absolute << " > info;";
		out << "\nany.replace(" << scope << "_tc_" << name
		    << ", v, true, &info);";
		out.end();
		
		out.sep();
		out << "\nvoid\noperator<<=(CORBA::Any& any, const "
                    << absolute << "& v)";
		out.start();
		out << "\nany <<= new " << absolute << "(v);";
		out.end();
		
		out.sep();
		out << "\nCORBA::Boolean\noperator>>=(const CORBA::Any& any, "
		    << absolute << "*& v)";
		out.start();
		out << "\nif(any.check_type(" << scope << "_tc_" << name
		    << "))";
		out.start();
		out << "\nif(!any.info())";
		out.start();
		out << "\nconst OB::Buffer* cbufp = "
		    << "(const OB::Buffer*)any.value();";
		out << "\nOB::InputStream in(*cbufp, 0, false);";
		out << '\n' << absolute << "* val = new " << absolute << ';';
		out << '\n' << absolute << "::_OB_unmarshal(*val, in);";
		out << "\n(CORBA::Any&)any <<= val;";
		out.end();
		out.sep();
		out << "\nv = (" << absolute << "*)any.value();";
		out << "\nreturn true;";
		out.end();
		out << "\nelse";
		out.inc();
		out << "\nreturn false;";
		out.dec();
		out.end();
	    }

	    //
	    // Anonymous sequences template instantiations
	    //
	    for(j = 0 ; j < infoSeq -> length() ; j++)
	    {
		IdlUnionMemberInfo& info = infoSeq[j];
		if(is_nil(info.type))
		    continue;
		
		if(info.type -> kind() == tk_sequence)
		{
		    String_var uniqueType = absolute;
		    uniqueType += "::OBUnique_";
		    uniqueType += info.name;

		    String_var templ;

		    for(ULong k = 0 ; k < 2 ; k++)
		    {
			if(k == 0)
			{
			    out.sep();
			    out.direct("\n#ifndef HAVE_NO_EXPLICIT_TEMPLATES");
			}
			else
			{
			    out.direct("\n#else");
			    out.direct("\n#ifdef HAVE_PRAGMA_DEFINE");
			}
			
			TypeCode_var iterType = info.type;
			while(iterType -> kind() == tk_sequence)
			{
			    templ = getTypeString("", iterType, GetTypeNormal,
						  false, 0, uniqueType);
			    out << "\ntemplate class " << templ << ';';
			    iterType = iterType -> content_type();
			}
			
			if(k == 1)
			{
			    out.direct("\n#endif"); // HAVE_PRAGMA_DEFINE
                            // HAVE_NO_EXPLICIT_TEMPLATES
			    out.direct("\n#endif");
			}
		    }
		}
	    }

	    break;
	}

	case dk_Attribute:
	{
            if(local_)
                continue;
            
            //
            // Don't generate code for valuetype attributes
            //
            if(container -> def_kind() == dk_Value)
                continue;

            //
            // Get the absolute name of the container
            //
            Contained_var contained = Contained::_narrow(container);
            assert(!is_nil(contained));
            ScopedName_var contAbsolute = getAbsolute(contained);

	    //
	    // Attribute
	    //
	    AttributeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);
            
	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);
	    ScopedName_var absolute = getAbsolute(desc -> id);

            //
            // If the container is an abstract interface, then we need
            // to use a special absolute name
            //
            ScopedName_var attribAbs;
            InterfaceDef_var interf = InterfaceDef::_narrow(container);
            if(!is_nil(interf) && interf -> is_abstract())
            {
                String_var interfName = interf -> name();
                interfName = fixKwd(interfName);
                Container_var c = interf -> defined_in();
                attribAbs = getScope(c);
                attribAbs += "OBAbstractImpl_";
                attribAbs += interfName;
                attribAbs += "::";
                attribAbs += name;
            }
            else
                attribAbs = absolute;

	    comment(desc -> id, out);

	    String_var ret = getTypeString("", desc -> type,
					   GetTypeReturn, false);
	    out.sep();
	    out << '\n' << ret << '\n' << attribAbs << "()";
	    out.start();
	    out << "\nwhile(true)";
	    out.start();

	    out << "\nOBStubImpl_" << contAbsolute
		<< "_var impl = _OB_getStubImpl();";
	    out << "\ntry";
	    out.start();
	    out << "\nreturn impl -> " << name << "();";
	    out.end();
	    out << "";
	    
	    out << "\ncatch(OB::StubImplForward& fwd)";
	    out.start();
	    out << "\n_OB_forward(fwd.ior);";
	    out.end();
	    out << "\ncatch(OB::StubImplChanged&)";
	    out.start();
	    out << "\n_OB_clearStubImpl();";
	    out.end();
	    out.end();
	    out.end();
                
	    if(desc -> mode == ATTR_NORMAL)
	    {
		String_var in = getTypeString(scope, desc -> type,
					      GetTypeIn, false);
		out.sep();
		out << "\nvoid\n" << attribAbs << '(' << in
		    << " _ob_a)";
		out.start();
		out << "\nwhile(true)";
		out.start();

		out << "\nOBStubImpl_" << contAbsolute
		    << "_var impl = _OB_getStubImpl();";
		out << "\ntry";
		out.start();
		out << "\nimpl -> " << name << "(_ob_a);";
		out << "\nreturn;";
		out.end();
		out << "";
		
		out << "\ncatch(OB::StubImplForward& fwd)";
		out.start();
		out << "\n_OB_forward(fwd.ior);";
		out.end();
		out << "\ncatch(OB::StubImplChanged&)";
		out.start();
		out << "\n_OB_clearStubImpl();";
		out.end();
		out.end();
		out.end();
	    }
		
	    break;
	}

	case dk_Operation:
	{
            if(local_)
                continue;
            
            //
            // Don't generate code for valuetype operations
            //
            if(container -> def_kind() == dk_Value)
                continue;

            //
            // Get the absolute name of the container
            //
            Contained_var contained = Contained::_narrow(container);
            assert(!is_nil(contained));
            ScopedName_var contAbsolute = getAbsolute(contained);

	    //
	    // Operation
	    //
	    OperationDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);
	    ScopedName_var absolute = getAbsolute(desc -> id);

            //
            // If the container is an abstract interface, then we need
            // to use a special absolute name
            //
            ScopedName_var opAbs;
            InterfaceDef_var interf = InterfaceDef::_narrow(container);
            if(!is_nil(interf) && interf -> is_abstract())
            {
                String_var interfName = interf -> name();
                interfName = fixKwd(interfName);
                Container_var c = interf -> defined_in();
                opAbs = getScope(c);
                opAbs += "OBAbstractImpl_";
                opAbs += interfName;
                opAbs += "::";
                opAbs += name;
            }
            else
                opAbs = absolute;

	    comment(desc -> id, out);

	    String_var ret = getTypeString("", desc -> result, GetTypeReturn);
	    
	    out.sep();
	    out << '\n' << ret << '\n' << opAbs << "(";
	    
	    unsigned int indent = out.getIndent();
	    out.setIndent(out.getPos());

	    for(j = 0 ; j < desc -> parameters.length() ; j++)
	    {
		GetType tt;
		switch(desc -> parameters[j].mode)
		{
		case PARAM_IN:
		    tt = GetTypeIn;
		    break;

		case PARAM_INOUT:
		    tt = GetTypeInOut;
		    break;

		case PARAM_OUT:
		    tt = GetTypeOut;
		    break;
		}

		String_var arg = string_dup("_ob_a");
		arg += j;

		String_var str =
		    getTypeString(scope, desc -> parameters[j].type, tt,
				  false, arg);

		if(j > 0)
		    out << ",\n";

		out << str;
	    }

	    if(desc -> contexts.length() > 0)
	    {
		if(desc -> parameters.length() > 0)
		    out << ",\n";
		
		out << "CORBA::Context_ptr _ob_ctx";
	    }

	    out << ")";

	    out.setIndent(indent);
	    out.start();
	    out << "\nwhile(true)";
	    out.start();

	    out << "\nOBStubImpl_" << contAbsolute
		<< "_var impl = _OB_getStubImpl();";
	    out << "\ntry";
	    out.start();
	    TypeCode_var origRet;
	    if(desc -> mode != OP_ONEWAY)
		origRet = GetOrigType(desc -> result);
	    if(desc -> mode != OP_ONEWAY &&
	       origRet -> kind() != tk_void)
		out << "\nreturn ";
	    else
		out << '\n';

	    out << "impl -> " << name << "(";
	    for(j = 0 ; j < desc -> parameters.length() ; j++)
	    {
		if(j > 0)
		    out << ", ";
		out << "_ob_a" << j;
	    }
	    if(desc -> contexts.length() > 0)
	    {
		if(desc -> parameters.length() > 0)
		    out << ", ";
		out << "_ob_ctx";
	    }
	    out << ");";
	    if(desc -> mode == OP_ONEWAY ||
	       origRet -> kind() == tk_void)
		out << "\nreturn;";
	    out.end();
	    if(desc -> mode != OP_ONEWAY)
	    {
		out << "\ncatch(OB::StubImplForward& fwd)";
		out.start();
		out << "\n_OB_forward(fwd.ior);";
		out.end();
	    }
	    out << "\ncatch(OB::StubImplChanged&)";
	    out.start();
	    out << "\n_OB_clearStubImpl();";
	    out.end();
	    out.end();
	    out.end();

	    break;
	}

	default:
	    break;
	}

	//
	// Check for recursion
	//
	if(!is_nil(cont))
	    genCPlusPlusC(cont, out);
    }
}

// ----------------------------------------------------------------------
// Generate code for the source file, variable definitions
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusCDefs(Container_ptr container,
			           IdlPrettyPrint& out)
{
    //
    // We need to place TypeCode definitions inside namespace scopes
    // to avoid compiler bugs.
    //

    //
    // Get scope
    //
    String_var scope = getScope(container);
    String_var defScope = scope;

    //
    // Determine the enclosing scope.  Since we are placing these
    // definitions inside 'namespace' scope, we still need to scope
    // definitions that reside within interfaces, valuetypes, etc.
    //
    Container_var parent = Container::_duplicate(container);
    while(parent -> def_kind() != dk_Module &&
          parent -> def_kind() != dk_Repository)
    {
        Contained_var cont = Contained::_narrow(parent);
        assert(!is_nil(cont));
        parent = cont -> defined_in();
    }
    String_var parentScope = string_dup("");
    Contained_var parentCont = Contained::_narrow(parent);
    if(!is_nil(parentCont))
    {
        parentScope = getScope(parent);
        IdlRemoveScope(parentScope, defScope.inout());
    }

    //
    // Get container contents descriptions
    //
    Container::DescriptionSeq_var seq =
        container -> describe_contents(dk_all, true, -1);

    ULong i;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	Container_var cont = Container::_narrow(seq[i].contained_object);

	switch(seq[i].kind)
	{
        case dk_Module:
        {
            //
            // Module
            //
            ModuleDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

            // Don't add this here!!!
            // if(!check(desc -> id))
            //     continue;

            Identifier_var name = fixKwd(desc -> name);

            //comment(desc -> id, out);
            out.sep();
            out << "\nnamespace " << name;
            out << "\n{";

            assert(!is_nil(cont));
            genCPlusPlusCDefs(cont, out);

            out.sep();
            out << "\n} // End of namespace " << name;

            break;
        }

	case dk_Constant:
	{
	    //
	    // Constant
	    //
	    ConstantDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);
	    
	    if(!check(desc -> id))
		continue;
	    
	    Identifier_var name = fixKwd(desc -> name);
            String_var defName = defScope;
            defName += name;

	    comment(desc -> id, out);

	    String_var s = getTypeString(parentScope, desc -> type,
                                         GetTypeConst, false, defName);
	    
	    out.sep();
	    out << '\n' << s << " = ";
	    
	    TypeCode_var tc = desc -> value.type();
	    tc = GetOrigType(tc);
	    
	    switch(tc -> kind())
	    {
	    case tk_short:
	    {
		Short v;
		Boolean b = desc -> value >>= v;
		assert(b);
		out << v;
		break;
	    }
	    
	    case tk_ushort:
	    {
		UShort v;
		Boolean b = desc -> value >>= v;
		assert(b);
		out << v;
		break;
	    }
	    
	    case tk_long:
	    {
		Long v;
		Boolean b = desc -> value >>= v;
		assert(b);
		out << v;
		break;
	    }
	    
	    case tk_ulong:
	    {
		ULong v;
		Boolean b = desc -> value >>= v;
		assert(b);
		out << v;
		break;
	    }
	    
	    case tk_longlong:
	    {
		LongLong v;
		Boolean b = desc -> value >>= v;
		assert(b);
		out << v;
		break;
	    }
	    
	    case tk_ulonglong:
	    {
		ULongLong v;
		Boolean b = desc -> value >>= v;
		assert(b);
		out << v;
		break;
	    }
	    
	    case tk_float:
	    {
		Float v;
		Boolean b = desc -> value >>= v;
		assert(b);
		out << v;
		break;
	    }
	    
	    case tk_double:
	    {
		Double v;
		Boolean b = desc -> value >>= v;
		assert(b);
		out << v;
		break;
	    }
	    
	    case tk_longdouble:
	    {
		LongDouble v;
		Boolean b = desc -> value >>= v;
		assert(b);
		out << v;
		break;
	    }
	    
	    case tk_string:
	    {
		char* v;
		ULong l = tc -> length();
		Any::to_string to(v, l);
		Boolean b = desc -> value >>= to;
		assert(b);
		String_var s = string_dup(v);
		IdlAddEscapes(s.inout(), true);
		out << s;
		break;
	    }
	    
	    case tk_wstring:
	    {
		wchar_t* v;
		Boolean b = desc -> value >>= v;
		assert(b);
		WString_var s = wstring_dup(v);
		IdlAddEscapes(s.inout(), true);
		out << s;
		break;
	    }
	    
	    case tk_char:
	    {
		Char v;
		Boolean b = desc -> value >>= Any::to_char(v);
		assert(b);
		String_var s;
		if(v == '\0')
		    s = string_dup("\\0");
		else
		{
		    s = string_alloc(1);
		    s[0] = v;
		    s[1] = '\0';
		    IdlAddEscapes(s.inout(), false);
		}
		out << "(CORBA::Char)'" << s << "'";
		break;
	    }
	    
	    case tk_wchar:
	    {
		WChar v;
		Boolean b = desc -> value >>= Any::to_wchar(v);
		assert(b);
		WString_var s;
		if(v == L'\0')
		    s = wstring_dup(L"\\0");
		else
		{
		    s = wstring_alloc(1);
		    s[0] = v;
		    s[1] = L'\0';
		    IdlAddEscapes(s.inout(), false);
		}
		out << "(CORBA::WChar)'" << s << "'";
		break;
	    }
	    
	    case tk_boolean:
	    {
		Boolean v;
		Boolean b = desc -> value >>= Any::to_boolean(v);
		assert(b);
		out << (v ? "CORBA::TRUE" : "CORBA::FALSE");
		break;
	    }
	    
	    default:
		assert(false);
		break;
	    }
	    
	    out << ';';
	    
	    break;
	}

	case dk_Interface:
        {
            //
            // Interface
            //
            InterfaceDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            //comment(desc -> id, out);

            if(!noTC_)
            {
		Identifier_var name = fixKwd(desc -> name);

		//
		// TypeCode constant
		//
		out.sep();
		out << "\nOB::TypeCodeConst " << defScope << "_tc_"
                    << name << '(';
		
		InterfaceDef_var interf =
		    InterfaceDef::_narrow(seq[i].contained_object);
		TypeCode_var type = interf -> type();
		IdlOutputAsciiTypeCode(type, out);
		out << ");";
	    }

	    genCPlusPlusCDefs(cont, out);

	    break;
	}

        case dk_Value:
        {
            //
            // Valuetype
            //
            ValueDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            //comment(desc -> id, out);

            if(!noTC_)
            {
		Identifier_var name = fixKwd(desc -> name);

		ValueDef_var value =
		    ValueDef::_narrow(seq[i].contained_object);
		assert(! is_nil(value));

		//
		// TypeCode constant
		//
		out.sep();
		out << "\nOB::TypeCodeConst " << defScope << "_tc_"
                    << name << '(';
		TypeCode_var type = value -> type();
		IdlOutputAsciiTypeCode(type, out);
		out << ");";
	    }

	    genCPlusPlusCDefs(cont, out);

            break;
        }

        case dk_ValueBox:
        {
            //
            // Value box
            //
            TypeDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            if(noTC_)
                continue;

            //comment(desc -> id, out);

            Identifier_var name = fixKwd(desc -> name);

	    ValueBoxDef_var box =
		ValueBoxDef::_narrow(seq[i].contained_object);
	    assert(! is_nil(box));

	    //
	    // TypeCode constant
	    //
	    out.sep();
	    out << "\nOB::TypeCodeConst " << defScope << "_tc_" << name << '(';
	    TypeCode_var boxType = box -> type();
	    IdlOutputAsciiTypeCode(boxType, out);
	    out << ");";

            break;
        }

	case dk_Alias:
	{
	    //
	    // Alias
	    //
	    TypeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

            if(noTC_)
                continue;

            //comment(desc -> id, out);

            Identifier_var name = fixKwd(desc -> name);

	    //
	    // TypeCode constant
	    //
	    out.sep();
	    out << "\nOB::TypeCodeConst " << defScope << "_tc_" << name << '(';
	    IdlOutputAsciiTypeCode(desc -> type, out);
	    out << ");";

	    break;
	}

	case dk_Enum:
	{
	    //
	    // Enum
	    //
	    TypeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);
		
	    if(!check(desc -> id))
		continue;
		
            if(noTC_)
                continue;

	    //comment(desc -> id, out);

            Identifier_var name = fixKwd(desc -> name);

	    //
	    // TypeCode constant
	    //
	    out.sep();
	    out << "\nOB::TypeCodeConst " << defScope << "_tc_" << name << '(';
	    IdlOutputAsciiTypeCode(desc -> type, out);
	    out << ");";

	    break;
	}

	case dk_Struct:
	case dk_Exception:
	{
	    RepositoryId_var id;
	    Identifier_var name;
	    TypeCode_var type;
	    
	    if(seq[i].kind == dk_Struct)
	    {
		//
		// Struct
		//
		TypeDescription* desc;
		Boolean b = seq[i].value >>= desc;
		assert(b);

		name = fixKwd(desc -> name);
		id = desc -> id;
		type = desc -> type;
	    }
	    else
	    {
		//
		// Exception
		//
		ExceptionDescription* desc;
		Boolean b = seq[i].value >>= desc;
		assert(b);

		name = fixKwd(desc -> name);
		id = desc -> id;
		type = desc -> type;
	    }

	    if(!check(id))
		continue;
	    
            if(noTC_)
                continue;

	    //comment(id, out);
	    
	    //
	    // TypeCode constant
	    //
	    out.sep();
	    out << "\nOB::TypeCodeConst " << defScope << "_tc_" << name << '(';
	    IdlOutputAsciiTypeCode(type, out);
	    out << ");";

	    break;
	}

	case dk_Union:
	{
	    //
	    // Union
	    //
	    TypeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

            if(noTC_)
                continue;

	    //comment(id, out);
	    
            Identifier_var name = fixKwd(desc -> name);

	    //
	    // TypeCode constant
	    //
	    out.sep();
	    out << "\nOB::TypeCodeConst " << defScope << "_tc_" << name << '(';
	    IdlOutputAsciiTypeCode(desc -> type, out);
	    out << ");";

	    break;
	}

	default:
	    break;
	}
    }
}

// ----------------------------------------------------------------------
// Generate code for the source file, stub implementation
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusCStubImpl(Container_ptr container,
                                       IdlPrettyPrint& out)
{
    //
    // Get the equivalent StubImpl name for this container
    //
    Contained_var contained = Contained::_narrow(container);
    ScopedName_var className;
    if(!is_nil(contained))
    {
        ScopedName_var contAbsolute = getAbsolute(contained);
        className = string_dup("OBStubImpl_");
        className += contAbsolute;
    }
    
    //
    // Get scope
    //
    String_var scope = getScope(container);

    //
    // Get container contents descriptions
    //
    Container::DescriptionSeq_var seq =
    container -> describe_contents(dk_all, true, -1);

    ULong i, j;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	switch(seq[i].kind)
	{

        case dk_Interface:
        {
            //
            // Interface
            //
            InterfaceDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);
            
            if(!check(desc -> id))
                continue;
            
            Identifier_var name = fixKwd(desc -> name);
            ScopedName_var absolute = getAbsolute(desc -> id);

            className = string_dup("OBStubImpl_");
            className += absolute;

            comment(desc -> id, out);

            String_var prefix;
            if(strlen(scope) > 0)
            {
                prefix = string_dup("OBStubImpl_");
                prefix += scope;
            }
            else
                prefix = string_dup("");

            //
            // Template instantiations
            //
	    out.sep();
	    out.direct("\n#ifndef HAVE_NO_EXPLICIT_TEMPLATES");
	    out << "\ntemplate class OB::ObjVar< " << className << " >;";
	    out.direct("\n#else");
	    out.direct("\n#ifdef HAVE_PRAGMA_DEFINE");
	    out << "\n#pragma define(OB::ObjVar< " << className << " >)";
	    out.direct("\n#endif"); // HAVE_PRAGMA_DEFINE
	    out.direct("\n#endif"); // HAVE_NO_EXPLICIT_TEMPLATES
            
            //
            // Duplicate and Release
            //
            out.sep();	    
            out << "\nvoid\nOBDuplicate(" << className << "_ptr p)";
            out.start();
            out << "\nif(p)";
            out.inc();
            out << "\np -> _OB_incRef();";
            out.dec();
            out.end();

            out.sep();
            out << "\nvoid\nOBRelease(" << className << "_ptr p)";
            out.start();
            out << "\nif(p)";
            out.inc();
            out << "\np -> _OB_decRef();";
            out.dec();
            out.end();

            break;
        }

        case dk_Attribute:
	{
	    //
	    // Attribute
	    //
	    AttributeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);
            
            //
            // Don't generate code for valuetype attributes
            //
            if(container -> def_kind() == dk_Value)
                continue;

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);
	    ScopedName_var absolute = getAbsolute(desc -> id);

	    comment(desc -> id, out);

	    String_var ret = getTypeString("", desc -> type,
						 GetTypeReturn, false);
	    out.sep();
	    out << '\n' << ret << '\n' << className << "::" << name << "()";
            ULong iter = 0;
            out.start();
            out << "\nif(CORBA::is_nil(_ob_clt_))";
            out.inc();
            out << "\nthrow CORBA::NO_IMPLEMENT();";
            out.dec();
            out.sep();
            out << "\nCORBA::ULong _ob_off = _ob_clt_ -> "
                << "offset(this, \"_get_" << desc -> name << "\");";
            out.sep();
            out << "\nOB::Buffer _ob_buf(_ob_off);";
            out.sep();
            out << "\nbool _ob_sw, _ob_ex, _ob_fo;";
            out << "\n_ob_off = _ob_clt_ -> request(this, \"_get_"
                << desc -> name
                << "\", _ob_buf, _ob_sw, _ob_ex, _ob_fo, "
                << "_ob_tout_, _ob_retry_);";
            out.sep();
            out << "\nOB::InputStream _ob_in(_ob_clt_ -> getCodeConverters(), _ob_buf, _ob_off, _ob_sw);";
            out.sep();
            out << "\nif(_ob_fo)";
            out.start();
            out << "\nIOP::IOR ior;";
            out << "\nIOP::IOR::_OB_unmarshal(ior, _ob_in);";
            out << "\nthrow OB::StubImplForward(ior);";
            out.end();
            out.sep();
            out << "\nif(_ob_ex)";
            out.inc();
            out << "\nthrow CORBA::UNKNOWN();";
            out.dec();
            out.sep();
            writeReadWriteCount("", desc -> type, "_ob_r", iter,
                                RWCReadReturnDirect, 0, out);
            out.end();
                
	    if(desc -> mode == ATTR_NORMAL)
	    {
		String_var in = getTypeString("", desc -> type,
						    GetTypeIn, false);
		out.sep();
		out << "\nvoid\n" << className << "::" << name << '('
                    << in << " _ob_a)";

                ULong iter = 0;
                out.start();
                out << "\nif(CORBA::is_nil(_ob_clt_))";
                out.inc();
                out << "\nthrow CORBA::NO_IMPLEMENT();";
                out.dec();
                out.sep();
                out << "\nCORBA::ULong _ob_off = _ob_clt_ -> "
                    << "offset(this, \"_set_" << desc -> name << "\");";
                out << "\nOB::CountStream _ob_cnt(_ob_off);";
                writeReadWriteCount("", desc -> type, "_ob_a", iter,
                                    RWCCountIn, 0, out);
                out.sep();
                out << "\nOB::Buffer _ob_buf(_ob_cnt.count());";
                out << "\nOB::OutputStream _ob_out(_ob_clt_ -> getCodeConverters(), _ob_buf, _ob_off);";
                writeReadWriteCount("", desc -> type, "_ob_a", iter,
                                    RWCWriteIn, 0, out);
                out.sep();
                out << "\nbool _ob_sw, _ob_ex, _ob_fo;";
                out << "\n_ob_clt_ -> request(this, \"_set_"
                    << desc -> name
                    << "\", _ob_buf, _ob_sw, _ob_ex, _ob_fo, "
                    << "_ob_tout_, _ob_retry_);";
                out.sep();
                out << "\nif(_ob_fo)";
                out.start();
                out << "\nOB::InputStream _ob_in(_ob_buf, _ob_off, _ob_sw);";
                out << "\nIOP::IOR ior;";
                out << "\nIOP::IOR::_OB_unmarshal(ior, _ob_in);";
                out << "\nthrow OB::StubImplForward(ior);";
                out.end();
                out.sep();
                out << "\nif(_ob_ex)";
                out.inc();
                out << "\nthrow CORBA::UNKNOWN();";
                out.dec();
                out.end();
            }
		
	    break;
	}

	case dk_Operation:
	{
	    //
	    // Operation
	    //
	    OperationDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

            //
            // Don't generate code for valuetype operations
            //
            if(container -> def_kind() == dk_Value)
                continue;

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);
	    ScopedName_var absolute = getAbsolute(desc -> id);

	    comment(desc -> id, out);

	    String_var ret =
		getTypeString("", desc -> result, GetTypeReturn);
	    
	    out.sep();
	    out << '\n' << ret << '\n' << className << "::" << name << "(";
	    
	    unsigned int indent = out.getIndent();
	    out.setIndent(out.getPos());

	    for(j = 0 ; j < desc -> parameters.length() ; j++)
	    {
		GetType tt;
		switch(desc -> parameters[j].mode)
		{
		case PARAM_IN:
		    tt = GetTypeIn;
		    break;

		case PARAM_INOUT:
		    tt = GetTypeInOut;
		    break;

		case PARAM_OUT:
		    tt = GetTypeOut;
		    break;
		}

		String_var arg = string_dup("_ob_a");
		arg += j;

		String_var str =
		    getTypeString("", desc -> parameters[j].type, tt,
				  false, arg);

		if(j > 0)
		    out << ",\n";

		out << str;
	    }

	    if(desc -> contexts.length() > 0)
	    {
		if(desc -> parameters.length() > 0)
		    out << ",\n";
		
		out << "CORBA::Context_ptr _ob_ctx";
	    }

	    out << ")";

	    out.setIndent(indent);

            ULong iter = 0;
            
            out.start();
            out << "\nif(CORBA::is_nil(_ob_clt_))";
            out.inc();
            out << "\nthrow CORBA::NO_IMPLEMENT();";
            out.dec();
            
            out.sep();
            out << "\nCORBA::ULong _ob_off = _ob_clt_ -> offset(this, \""
                << desc -> name << "\");";
            out << "\nOB::CountStream _ob_cnt(_ob_clt_ -> getCodeConverters(), _ob_off);";
            bool haveInParams = false;
            for(j = 0 ; j < desc -> parameters.length() ; j++)
            {
                if(desc -> parameters[j].mode != PARAM_OUT)
                {
                    TypeCode_ptr argType =
                        desc -> parameters[j].type;
                    String_var arg = string_dup("_ob_a");
                    arg += j;
                    if(desc -> parameters[j].mode == PARAM_IN)
                        writeReadWriteCount("", argType, arg, iter,
                                            RWCCountIn, 0, out);
                    else
                        writeReadWriteCount("", argType, arg, iter,
                                            RWCCountInOut, 0, out);
                    haveInParams = true;
                }
            }
            if(desc -> contexts.length() > 0)
            {
                out << "\nOB::StringSeq _ob_ctxSeq;";
                for(j = 0 ; j < desc -> contexts.length() ; j++)
                {
                    out << "\n_ob_ctx -> _OB_getValues(\"\", 0, \""
                        << desc -> contexts[j] << "\", _ob_ctxSeq);";
                }
                //
                // No typecode for OB::StringSeq
                //
                TypeCode_var ctxType =
                    orb_ -> create_sequence_tc(0, _tc_string);
		writeReadWriteCount("", ctxType, "_ob_ctxSeq", iter,
				    RWCCountIn, 0, out);
            }
            
            out.sep();
            out << "\nOB::Buffer _ob_buf(_ob_cnt.count());";
            if(haveInParams || desc -> contexts.length() > 0)
            {
                out << "\nOB::OutputStream _ob_out(_ob_clt_ -> getCodeConverters(), _ob_buf, _ob_off);";
            }
            for(j = 0 ; j < desc -> parameters.length() ; j++)
            {
                if(desc -> parameters[j].mode != PARAM_OUT)
                {
                    TypeCode_ptr argType =
                        desc -> parameters[j].type;
                    String_var arg = string_dup("_ob_a");
                    arg += j;
                    if(desc -> parameters[j].mode == PARAM_IN)
                        writeReadWriteCount("", argType, arg, iter,
                                            RWCWriteIn, 0, out);
                    else
                        writeReadWriteCount("", argType, arg, iter,
                                            RWCWriteInOut, 0, out);
                }
            }
            if(desc -> contexts.length() > 0)
            {
                //
                // No typecode for OB::StringSeq
                //
                TypeCode_var ctxType =
                    orb_ -> create_sequence_tc(0, _tc_string);
		writeReadWriteCount("", ctxType, "_ob_ctxSeq", iter,
				    RWCWriteIn, 0, out);
            }
            
            out.sep();
            if(desc -> mode == OP_ONEWAY)
            {
                out << "\n_ob_clt_ -> oneway(this, \"" << desc -> name
                    << "\", _ob_buf, _ob_retry_);";
            }
            else
            {
                out << "\nbool _ob_sw, _ob_ex, _ob_fo;";
                out << "\n_ob_off = _ob_clt_ -> request(this, \""
                    << desc -> name << "\", _ob_buf, _ob_sw, "
                    << "_ob_ex, _ob_fo, _ob_tout_, _ob_retry_);";
                
                TypeCode_var origRet = GetOrigType(desc -> result);
                bool haveOutParams = false;
                for(j = 0 ; j < desc -> parameters.length() ; j++)
                {
                    if(desc -> parameters[j].mode == PARAM_OUT ||
                       desc -> parameters[j].mode == PARAM_INOUT)
                    {
                        haveOutParams = true;
                        break;
                    }
                }
                
                out.sep();
                out << "\nif(_ob_fo)";
                out.start();
		out << "\nOB::InputStream _ob_in(_ob_clt_ -> getCodeConverters(), _ob_buf, _ob_off, _ob_sw);";
                out << "\nIOP::IOR ior;";
                out << "\nIOP::IOR::_OB_unmarshal(ior, _ob_in);";
                out << "\nthrow OB::StubImplForward(ior);";
                out.end();
                
                if(haveOutParams || origRet -> kind() != tk_void ||
                   desc -> exceptions.length() > 0)
                {
                    out.sep();
                    out << "\nOB::InputStream "
                        << "_ob_in(_ob_buf, _ob_off, _ob_sw);";
                }
                
                out.sep();
                out << "\nif(_ob_ex)";
                if(desc -> exceptions.length() > 0)
                    out.start();
                else
                    out.inc();
                for(j = 0 ; j < desc -> exceptions.length() ; j++)
                {
                    if(j == 0)
                    {
                        out << "\nconst CORBA::Octet* _ob_pos = "
                            << "_ob_in._OB_pos();";
                        out << "\nCORBA::String_var _ob_id = "
                            << "_ob_in.read_string();";
                        out << "\n_ob_in._OB_pos(_ob_pos);";
                        out.sep();
                        out << '\n';
                    }
                    else
                        out << "\nelse ";
                    
                    out << "if(strcmp(_ob_id, \""
                        << desc -> exceptions[j].id << "\") == 0)";
                    out.start();
                    TypeCode_ptr exType = desc -> exceptions[j].type;
                    String_var s = getTypeString("", exType,
                                                       GetTypeNormal);
                    out << '\n' << s << " _ob_except;";
                    out << '\n' << s << "::_OB_unmarshal(_ob_except, _ob_in);";
                    out.sep();
                    out << "\nthrow _ob_except;";
                    out.end();
                }
                if(desc -> exceptions.length() > 0)
                {
                    out << "\nelse";
                    out.inc();
                }
                out << "\nthrow CORBA::UNKNOWN();";
                if(desc -> exceptions.length() > 0)
                {
                    out.dec();
                    out.end();
                }
                else
                    out.dec();
                
                if(haveOutParams || origRet -> kind() != tk_void)
                    out.sep();
                
                if(origRet -> kind() != tk_void)
                {
                    if(haveOutParams)
                        writeReadWriteCount("", desc -> result, "_ob_r",
                                            iter, RWCReadReturnAlloc, 0,
                                            out);
                    else
                        writeReadWriteCount("", desc -> result, "_ob_r",
                                            iter, RWCReadReturnDirect, 0,
                                            out);
                }
                
                if(haveOutParams)
                {
                    for(j = 0 ; j < desc -> parameters.length() ; j++)
                    {
                        TypeCode_ptr argType =
                            desc -> parameters[j].type;
                        String_var arg = string_dup("_ob_a");
                        arg += j;
                        if(desc -> parameters[j].mode == PARAM_OUT)
                            writeReadWriteCount("", argType, arg,
                                                iter, RWCReadOut, 0, out);
                        else
			    if(desc -> parameters[j].mode == PARAM_INOUT)
				writeReadWriteCount("", argType, arg,
						    iter, RWCReadInOut, 0,
						    out);
                    }

                    if(origRet -> kind() != tk_void)
                    {
                        if(IdlIsVariable(origRet) ||
                           origRet -> kind() == tk_array)
                            out << "\nreturn _ob_r._retn();";
                        else
                            out << "\nreturn _ob_r;";
                    }
                }
            }
            
            out.end();
        
	    break;
	}

	default:
	    break;
	}

	//
	// Check for recursion
	//
	Container_var cont =
	    Container::_narrow(seq[i].contained_object);
	if(!is_nil(cont))
	    genCPlusPlusCStubImpl(cont, out);
    }
}

// ----------------------------------------------------------------------
// Generate OBV_* classes
// ----------------------------------------------------------------------

// TODO: Q: Why is this a separate operation?
// TODO: A: For consistency with genCPlusPlusHOBV.  The OBV_* classes
//       are different than the valuetype classes and are generated
//       as a separate step (similar to the local stubs). (mes)
void
IdlCPPGenerator::genCPlusPlusCOBV(Container_ptr container,
                                  IdlPrettyPrint& out)
{
    //
    // Get container contents descriptions
    //
    Container::DescriptionSeq_var seq =
        container -> describe_contents(dk_all, true, -1);

    ULong i, j;

    for(i = 0 ; i < seq -> length() ; i++)
    {
        Container_var cont = Container::_narrow(seq[i].contained_object);

        switch(seq[i].kind)
        {
        case dk_Value:
        {
            //
            // Valuetype
            //
            ValueDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            //
            // Don't generate OBV class for abstract interface
            //
            if(desc -> is_abstract)
                continue;

            String_var name = getPrefixedName("OBV", seq[i].contained_object);


            //
            // Get the name of the value's `OBV' class
            //
            Identifier_var absolute = getAbsolute(desc -> id);
            Identifier_var obvAbs = getPrefixedAbsolute("OBV", desc -> id, "");

            comment(desc -> id, out);

            //
            // Get all value members
            //
            Container::DescriptionSeq_var members =
                cont -> describe_contents(dk_ValueMember, true, -1);

            //
            // Default constructor
            //
            out.sep();
            out << '\n' << obvAbs << "::" << name << "()";
            out.start();

            bool hasFixed = false;
            for(j = 0 ; j < members -> length() ; j++)
            {
                ValueMember* member;
                b = members[j].value >>= member;
                assert(b);

                if(!IdlIsVariable(member -> type))
                {
                    hasFixed = true;
                    break;
                }
            }

            if(hasFixed)
            {
                //
                // Clear all fixed members
                //
                out.direct("\n#ifdef OB_CLEAR_MEM");
                for(j = 0 ; j < members -> length() ; j++)
                {
                    ValueMember* member;
                    b = members[j].value >>= member;
                    assert(b);

                    if(!IdlIsVariable(member -> type))
                    {
                        Identifier_var memberName = string_dup("_ob_");
                        memberName += member -> name;
                        memberName += "_";
                        out << "\nmemset(&" << memberName << ", 0, sizeof("
                            << memberName << "));";
                    }
                }
                out.direct("\n#endif"); // OB_CLEAR_MEM
            }

            out.end();

            //
            // Constructor with initializer for each member
            //
            if(members -> length() > 0)
            {
                out.sep();
                out << '\n' << obvAbs << "::" << name << "(";
                unsigned int indent = out.getIndent();
                out.setIndent(out.getPos());
                ULong arrayCount = 0;
                for(j = 0 ; j < members -> length() ; j++)
                {
                    ValueMember* member;
                    b = members[j].value >>= member;
                    assert(b);

                    TypeCode_var origMemberType = GetOrigType(member -> type);

                    Identifier_var argName = fixKwd(member -> name);

                    if(j > 0)
                        out << ",\n";

                    //
                    // Determine type name for member - check for
                    // anonymous types
                    //
                    if(member -> type -> kind() == tk_sequence)
                    {
                        //
                        // Anonymous sequence
                        //
                        out << "const " << "_" << member -> name << "_seq& "
                            << argName;
                    }
                    else if(origMemberType -> kind() == tk_array)
                    {
                        arrayCount++;
                        String_var s = getTypeString("", member -> type,
                                                     GetTypeNormal, false,
                                                     argName);
                        out << s;
                    }
                    else
                    {
                        String_var s =
                            getTypeString("", member -> type, GetTypeIn);
                        out << s << ' ' << argName;
                    }
                }

                //
                // Generate member initializer list, but we cannot
                // initialize array members this way, so we need to
                // make sure there's at least one non-array member.
                //
                out << ')';
                if(members -> length() > arrayCount)
                {
                    out << " :";
                    out.setIndent(indent);
                    out.inc();
                    out << '\n';
                    ULong initCount = 0;
                    for(j = 0 ; j < members -> length() ; j++)
                    {
                        ValueMember* member;
                        b = members[j].value >>= member;
                        assert(b);

                        Identifier_var argName = fixKwd(member -> name);
                        TypeCode_var origMemberType =
                            GetOrigType(member -> type);

                        //
                        // Initialize array members in constructor body
                        //
                        if(origMemberType -> kind() == tk_array)
                            continue;

                        String_var memberName = string_dup("_ob_");
                        memberName += member -> name;
                        memberName += "_";

                        if(initCount > 0)
                            out << ",\n";

                        out << memberName << '(';

                        if(origMemberType -> kind() == tk_objref ||
                           origMemberType -> kind() == tk_TypeCode ||
                           origMemberType -> kind() == tk_Principal ||
                           origMemberType -> kind() == tk_abstract_interface)
                        {
                            String_var norm = getTypeString("", member -> type,
                                                            GetTypeNormal);
                            out << norm << "::_duplicate(" << argName << ')';
                        }
                        else
                            out << argName;
                        out << ')';
                        initCount++;
                    }
                    out.dec();
                }
                else
                    out.setIndent(indent);

                out.start();

                //
                // Perform additional initialization
                //
                for(j = 0 ; j < members -> length() ; j++)
                {
                    ValueMember* member;
                    b = members[j].value >>= member;
                    assert(b);

                    Identifier_var argName = fixKwd(member -> name);
                    String_var memberName = string_dup("_ob_");
                    memberName += member -> name;
                    memberName += "_";
                    TypeCode_var origMemberType = GetOrigType(member -> type);

                    if(origMemberType -> kind() == tk_value ||
                       origMemberType -> kind() == tk_value_box)
                    {
                        out << "\nCORBA::add_ref(" << memberName << ");";
                    }
                    else if(origMemberType -> kind() == tk_array)
                    {
                        String_var index = string_dup("_ob_i");
                        index += j;
                        writeArrayCopy("", member -> type, memberName,
                                       argName, index, out);
                    }
                }
                out.end();
            }

            //
            // Destructor
            //
            out.sep();
            out << '\n' << obvAbs << "::~" << name << "()";
            out.start();
            out.end();

            //
            // Generate functions for each value member
            //
            for(j = 0 ; j < members -> length() ; j++)
            {
                ValueMember* member;
                b = members[j].value >>= member;
                assert(b);
                genCPlusPlusCValueMember(cont, member, obvAbs, out);
            }

            break;
        }

        default:
            break;
        }

        //
        // Check for recursion
        //
        if(!is_nil(cont))
            genCPlusPlusCOBV(cont, out);
    }
}

// ----------------------------------------------------------------------
// Generate accessor and mutator functions for value member
// ----------------------------------------------------------------------

// TODO: This should probably be renamed to writeValueMember
//       <But there's a corresponding genCPlusPlusHValueMember - mes>
void
IdlCPPGenerator::genCPlusPlusCValueMember(Container_ptr value,
                                          ValueMember* desc,
                                          const char* className,
                                          IdlPrettyPrint& out)
{
    if(!check(desc -> id))
        return;

    ValueDef_var valueDef = ValueDef::_narrow(value);
    assert(! is_nil(valueDef));
    RepositoryId_var id = valueDef -> id();
    ScopedName_var absValue = getAbsolute(id);

    comment(desc -> id, out);

    Identifier_var name = fixKwd(desc -> name);

    String_var function = string_dup(className);
    function += "::";
    function += name;

    String_var memberName = string_dup("_ob_");
    memberName += desc -> name;
    memberName += "_";

    TypeCode_var origMemberType = GetOrigType(desc -> type);

    String_var norm, in, inOut, ret, var;
    if(desc -> type -> kind() == tk_sequence)
    {
        //
        // Anonymous sequence
        //
        norm = string_dup("_");
        norm += desc -> name;
        norm += "_seq";
        in = string_dup("const ");
        in += norm;
        in += "&";
        inOut = norm;
        inOut += "&";
    }
    else
    {
        norm = getTypeString("", desc -> type, GetTypeNormal);
        in = getTypeString("", desc -> type, GetTypeIn);
        inOut = getTypeString("", desc -> type, GetTypeInOut);
        ret = getTypeString("", desc -> type, GetTypeReturn);
        var = getTypeString("", desc -> type, GetTypeVar);
    }

    switch(origMemberType -> kind())
    {
    case tk_null:
    case tk_void:
    case tk_except:
    case tk_native:
        assert(false);
        break;

    case tk_short:
    case tk_long:
    case tk_longlong:
    case tk_ushort:
    case tk_ulong:
    case tk_ulonglong:
    case tk_float:
    case tk_double:
    case tk_longdouble:
    case tk_fixed:
    case tk_boolean:
    case tk_char:
    case tk_wchar:
    case tk_octet:
    case tk_enum:
        out.sep();
        out << "\nvoid\n" << function << '(' << in << " _ob_a)";
        out.start();
        out << '\n' << memberName << " = _ob_a;";
        out.end();
        out.sep();
        out << '\n' << ret << '\n' << function << "() const";
        out.start();
        out << "\nreturn " << memberName << ';';
        out.end();
        break;

    case tk_value:
    case tk_value_box:
        out.sep();
        out << "\nvoid\n" << function << '(' << in << " _ob_a)";
        out.start();
        out << '\n' << memberName << " = _ob_a;";
        out << "\nCORBA::add_ref(" << memberName << ");";
        out.end();
        out.sep();
        out << '\n' << ret << '\n' << function << "() const";
        out.start();
        out << "\nreturn " << memberName << ".in();";
        out.end();
        break;

    case tk_struct:
    case tk_union:
    case tk_sequence:
    case tk_any:
        out.sep();
        out << "\nvoid\n" << function << '(' << in << " _ob_a)";
        out.start();
        out << '\n' << memberName << " = _ob_a;";
        out.end();
        out.sep();
        out << '\n' << in << '\n' << function << "() const";
        out.start();
        out << "\nreturn " << memberName << ';';
        out.end();
        out.sep();
        out << '\n' << inOut << '\n' << function << "()";
        out.start();
        out << "\nreturn " << memberName << ';';
        out.end();
        break;

    case tk_TypeCode:
    case tk_Principal:
    case tk_objref:
    case tk_abstract_interface:
        out.sep();
        out << "\nvoid\n" << function << '(' << in << " _ob_a)";
        out.start();
        out << '\n' << memberName << " = " << norm << "::_duplicate(_ob_a);";
        out.end();
        out.sep();
        out << '\n' << ret << '\n' << function << "() const";
        out.start();
        out << "\nreturn " << memberName << ".in();";
        out.end();
        break;

    case tk_string:
    case tk_wstring:
        out.sep();
        out << "\nvoid\n" << function << '(' << norm << " _ob_a)";
        out.start();
        out << '\n' << memberName << " = _ob_a;";
        out.end();
        out.sep();
        out << "\nvoid\n" << function << "(const " << norm << " _ob_a)";
        out.start();
        out << '\n' << memberName << " = _ob_a;";
        out.end();
        out.sep();
        out << "\nvoid\n" << function << "(const " << var << "& _ob_a)";
        out.start();
        out << '\n' << memberName << " = _ob_a;";
        out.end();
        out.sep();
        out << "\nconst " << norm << '\n' << function << "() const";
        out.start();
        out << "\nreturn " << memberName << ".in();";
        out.end();
        break;

    case tk_array:
    {
        String_var slice;
        String_var in =
            getTypeString("", desc -> type, GetTypeNormal, false, "_ob_a");

        if(desc -> type -> kind() == tk_alias)
        {
            slice = getTypeString("", desc -> type, GetTypeSlice);
        }
        else
        {
            slice = absValue;
            slice += "::_";
            slice += desc -> name;
            slice += "_slice";
        }
        
        out.sep();
        out << "\nvoid\n" << function << '(' << in << ')';
        out.start();
        writeArrayCopy("", desc -> type, memberName, "_ob_a", "_ob_i", out);
        out.end();

        out.sep();
        out << '\n' << slice << "*\n" << function << "() const";
        out.start();
        out << "\nreturn (" << slice << "*)" << memberName << ';';
        out.end();
        break;
    }

    case tk_alias:
        assert(false);
        break;
    }
}

// ----------------------------------------------------------------------
// Generate type-specific value factory implementation
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusCValueFactory(ValueDef_ptr value,
                                           IdlPrettyPrint& out)
{
    ValueDef::FullValueDescription_var desc = value -> describe_value();

    if(desc -> initializers.length() == 0)
        return;

    //
    // Don't generate factory class for abstract interface
    //
    if(desc -> is_abstract)
        return;

    Identifier_var name = fixKwd(desc -> name);

    //
    // Compose the name of the factory's class
    //
    Identifier_var absolute = getAbsolute(desc -> id);
    String_var className = string_dup(absolute);
    className += "_init";

    //
    // _downcast
    //
    out.sep();
    out << '\n' << className << "*\n" << className
        << "::_downcast(CORBA::ValueFactory f)";
    out.start();
    out << "\nreturn dynamic_cast<" << desc -> name << "_init*>(f);";
    out.end();

    //
    // _OB_typeId
    //
    out.sep();
    out << "\nconst char*\n" << className << "::_OB_typeId() const";
    out.start();
    out << "\nreturn " << absolute << "::_OB_id();";
    out.end();
}

// ----------------------------------------------------------------------
// Write value member unmarshalling code
// ----------------------------------------------------------------------

void
IdlCPPGenerator::writeValueMemberUnmarshal(const char* scope,
                                           const char* absolute,
                                           const char* name,
                                           const char* arg,
                                           CORBA::TypeCode_ptr type,
                                           CORBA::ULong& iter,
                                           IdlPrettyPrint& out)
{
    String_var memberName = fixKwd(name);
    String_var uniqueType = absolute;
    uniqueType += "::OBUnique_";
    uniqueType += name;

    TypeCode_var origMemberType = GetOrigType(type);
    TCKind k = origMemberType -> kind();

    //
    // The mapping for valuetype members of these types
    // provides an accessor that returns a reference,
    // which we can use to unmarshal directly into the
    // member, rather than into a temporary which is then
    // copied.
    //
    // XXX Optimize arrays also?
    //
    if(k == tk_sequence || k == tk_union || k == tk_struct)
    {
        String_var str;
        //
        // Get type name for anonymous sequence
        //
        if(type -> kind() == tk_sequence)
        {
            str = string_dup("_");
            str += name;
            str += "_seq";
        }
        else
        {
            str = getTypeString(scope, type, GetTypeStruct);
        }
        out << '\n' << str << "& " << arg << " = " << memberName << "();";
        writeReadWriteCount(scope, type, arg, iter, RWCReadStructElem,
                            uniqueType, out);
    }
    else
    {
        //
        // Declare a temporary, unmarshal into the
        // temporary, then invoke the mutator for the
        // member.
        //
        String_var str;
        str = getTypeString(scope, type, GetTypeStruct, false, arg);
        out << '\n' << str << ';';
        writeReadWriteCount(scope, type, arg, iter, RWCReadStructElem,
                            uniqueType, out);
        out << '\n' << memberName << '(' << arg << ");";
    }
}
