// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <Types.h>
#include <Util.h>
#include <GenUtil.h>
#include <GenCPP.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Generate code for the header file, forward declarations
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHFwd(Container_ptr container,
				  IdlPrettyPrint& out)
{
    //
    // Get scope
    //
    String_var scope = getScope(container);

    //
    // Get container contents descriptions
    //
    Container::DescriptionSeq_var seq =
	container -> describe_contents(dk_all, true, -1);

    // ------------------------------------------------------------
    // Declare interface, valuetype and value box classes
    // ------------------------------------------------------------

    ULong i;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get container
	//
	Container_var cont = Container::_narrow(seq[i].contained_object);

	switch(seq[i].kind)
	{
	case dk_Module:
	{
	    //
	    // Module
	    //
	    ModuleDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);
	    
	    // Don't add this here!!!
	    // if(!check(desc -> id))
	    //     continue;

            Identifier_var name = fixKwd(desc -> name);

	    //comment(desc -> id, out);
            out.sep();
            out << "\nnamespace " << name;
            out << "\n{";

	    assert(!is_nil(cont));
	    genCPlusPlusHFwd(cont, out);

            out.sep();
            out << "\n} // End of namespace " << name;

	    break;
	}

	case dk_Interface:
	{
	    //
	    // Interface
	    //
	    InterfaceDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    Identifier_var name = fixKwd(desc -> name);

	    //comment(desc -> id, out);

	    out.sep();
	    out << "\nclass " << name << ';';
	    out << "\ntypedef " << name << "* " << name << "_ptr;";
	    out << "\ntypedef " << name << "* " << name << "Ref;";

            break;
	}

	case dk_Value:
	{
	    //
	    // Valuetype
	    //
	    ValueDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

            if(strcmp(desc -> id, "IDL:omg.org/CORBA/ValueBase:1.0") == 0)
                continue;

	    Identifier_var name = fixKwd(desc -> name);

	    //comment(desc -> id, out);
	    out.sep();
	    out << "\nclass " << name << ';';

            break;
	}

	case dk_ValueBox:
	{
	    //
	    // Value box
	    //
	    TypeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    Identifier_var name = fixKwd(desc -> name);

	    //comment(desc -> id, out);
	    out.sep();
	    out << "\nclass " << name << ';';

            break;
	}

        default:
            break;
        }
    }
}

// ----------------------------------------------------------------------
// Generate code for the header file, global function forward declarations
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHGlobalFwd(Container_ptr container,
				        IdlPrettyPrint& out)
{
    //
    // Get scope
    //
    String_var scope = getScope(container);

    //
    // Get container contents descriptions
    //
    Container::DescriptionSeq_var seq =
	container -> describe_contents(dk_all, true, -1);

    // ------------------------------------------------------------
    // Declare OBDuplicate and OBRelease global functions
    // ------------------------------------------------------------

    ULong i;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get container
	//
	Container_var cont = Container::_narrow(seq[i].contained_object);

	switch(seq[i].kind)
	{
	case dk_Module:
	{
	    //
	    // Module
	    //
	    
	    // Don't add this here!!!
	    // if(!check(desc -> id))
	    //     continue;

	    genCPlusPlusHGlobalFwd(cont, out);

	    break;
	}

	case dk_Interface:
	{
	    //
	    // Interface
	    //
	    InterfaceDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

            ScopedName_var absolute = getAbsolute(desc -> id);

	    //comment(desc -> id, out);

	    out.sep();
	    out << "\nvoid OBDuplicate(" << absolute << "_ptr);";
	    out << "\nvoid OBRelease(" << absolute << "_ptr);";

            break;
	}

	case dk_Value:
	{
	    //
	    // Valuetype
	    //
	    ValueDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

            if(strcmp(desc -> id, "IDL:omg.org/CORBA/ValueBase:1.0") == 0)
                continue;

            ScopedName_var absolute = getAbsolute(desc -> id);

	    //comment(desc -> id, out);

	    out.sep();
	    out << "\nvoid OBDuplicate(" << absolute << "*);";
	    out << "\nvoid OBRelease(" << absolute << "*);";

            break;
	}

	case dk_ValueBox:
	{
	    //
	    // Value box
	    //
	    TypeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

            ScopedName_var absolute = getAbsolute(desc -> id);

	    //comment(desc -> id, out);

	    out.sep();
	    out << "\nvoid OBDuplicate(" << absolute << "*);";
	    out << "\nvoid OBRelease(" << absolute << "*);";

            break;
	}

        default:
            break;
        }
    }
}

// ----------------------------------------------------------------------
// Generate code for the header file, _var forward declarations
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHVarFwd(Container_ptr container,
				     IdlPrettyPrint& out)
{
    //
    // Get scope
    //
    String_var scope = getScope(container);

    //
    // Get container contents descriptions
    //
    Container::DescriptionSeq_var seq =
	container -> describe_contents(dk_all, true, -1);

    // ------------------------------------------------------------
    // Declare interface, valuetype and value box _var types
    // ------------------------------------------------------------

    ULong i;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get container
	//
	Container_var cont = Container::_narrow(seq[i].contained_object);

	switch(seq[i].kind)
	{
	case dk_Module:
	{
	    //
	    // Module
	    //
	    ModuleDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);
	    
	    // Don't add this here!!!
	    // if(!check(desc -> id))
	    //     continue;

            Identifier_var name = fixKwd(desc -> name);

	    //comment(desc -> id, out);
            out.sep();
            out << "\nnamespace " << name;
            out << "\n{";

	    assert(!is_nil(cont));
	    genCPlusPlusHVarFwd(cont, out);

            out.sep();
            out << "\n} // End of namespace " << name;

	    break;
	}

	case dk_Interface:
	{
	    //
	    // Interface
	    //
	    InterfaceDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    Identifier_var name = fixKwd(desc -> name);

	    //comment(desc -> id, out);

	    out.sep();
	    out << "\ntypedef OB::ObjVar< " << name << " > "
                << name << "_var;";
            out << "\ntypedef OB::ObjOut< " << name << " > "
                << name << "_out;";

            break;
	}

	case dk_Value:
	{
	    //
	    // Valuetype
	    //
	    ValueDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

            if(strcmp(desc -> id, "IDL:omg.org/CORBA/ValueBase:1.0") == 0)
                continue;

	    Identifier_var name = fixKwd(desc -> name);

	    //comment(desc -> id, out);

	    out.sep();
	    out << "\ntypedef OB::ObjVar< " << name << " > "
                << name << "_var;";
            out << "\ntypedef OB::ObjOut< " << name << " > "
                << name << "_out;";

            break;
	}

	case dk_ValueBox:
	{
	    //
	    // Value box
	    //
	    TypeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    Identifier_var name = fixKwd(desc -> name);

	    //comment(desc -> id, out);

	    out.sep();
	    out << "\ntypedef OB::ObjVar< " << name << " > "
                << name << "_var;";
            out << "\ntypedef OB::ObjOut< " << name << " > "
                << name << "_out;";

            break;
	}

        default:
            break;
        }
    }
}

// ----------------------------------------------------------------------
// Generate code for the header file, StubImpl forward declarations
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHStubImplFwd(Container_ptr container,
                                          IdlPrettyPrint& out)
{
    //
    // Get scope
    //
    String_var scope = getScope(container);

    //
    // Get container contents descriptions
    //
    Container::DescriptionSeq_var seq =
	container -> describe_contents(dk_all, true, -1);

    // ------------------------------------------------------------
    // Declare stub classes for interfaces
    // ------------------------------------------------------------

    ULong i;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get container
	//
	Container_var cont = Container::_narrow(seq[i].contained_object);

	switch(seq[i].kind)
	{
	case dk_Module:
	{
	    //
	    // Module
	    //
	    ModuleDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);
	    
	    // Don't add this here!!!
	    // if(!check(desc -> id))
	    //     continue;

            String_var name =
                getPrefixedName("OBStubImpl", seq[i].contained_object);

	    //comment(desc -> id, out);
	    out.sep();
            out << "\nnamespace " << name;
            out << "\n{";

	    assert(!is_nil(cont));
	    genCPlusPlusHStubImplFwd(cont, out);

            out.sep();
            out << "\n} // End of namespace " << name;

	    break;
	}

	case dk_Interface:
	{
	    //
	    // Interface
	    //
	    InterfaceDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

            String_var name =
                getPrefixedName("OBStubImpl", seq[i].contained_object);

	    //comment(desc -> id, out);

	    out.sep();
	    out << "\nclass " << name << ';';
	    out << "\ntypedef " << name << "* " << name << "_ptr;";
	    out << "\ntypedef " << name << "* " << name << "Ref;";

            break;
	}

        default:
            break;
        }
    }
}

// ----------------------------------------------------------------------
// Generate code for the header file, StubImpl global functions
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHStubImplGlobalFwd(Container_ptr container,
                                                IdlPrettyPrint& out)
{
    //
    // Get scope
    //
    String_var scope = getScope(container);

    //
    // Get container contents descriptions
    //
    Container::DescriptionSeq_var seq =
	container -> describe_contents(dk_all, true, -1);

    // ------------------------------------------------------------
    // Declare stub classes for interfaces
    // ------------------------------------------------------------

    ULong i;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get container
	//
	Container_var cont = Container::_narrow(seq[i].contained_object);

	switch(seq[i].kind)
	{
	case dk_Module:
	{
	    //
	    // Module
	    //
	    
	    // Don't add this here!!!
	    // if(!check(desc -> id))
	    //     continue;

	    assert(!is_nil(cont));
	    genCPlusPlusHStubImplGlobalFwd(cont, out);

	    break;
	}

	case dk_Interface:
	{
	    //
	    // Interface
	    //
	    InterfaceDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

            ScopedName_var absolute =
                getPrefixedAbsolute("OBStubImpl", desc -> id, "");

	    //comment(desc -> id, out);

	    out.sep();
	    out << "\nvoid OBDuplicate(" << absolute << "_ptr);";
	    out << "\nvoid OBRelease(" << absolute << "_ptr);";

            break;
	}

        default:
            break;
        }
    }
}

// ----------------------------------------------------------------------
// Generate code for the header file, StubImpl _var declarations
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHStubImplVarFwd(Container_ptr container,
                                             IdlPrettyPrint& out)
{
    //
    // Get scope
    //
    String_var scope = getScope(container);

    //
    // Get container contents descriptions
    //
    Container::DescriptionSeq_var seq =
	container -> describe_contents(dk_all, true, -1);

    // ------------------------------------------------------------
    // Declare stub classes for interfaces
    // ------------------------------------------------------------

    ULong i;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get container
	//
	Container_var cont = Container::_narrow(seq[i].contained_object);

	switch(seq[i].kind)
	{
	case dk_Module:
	{
	    //
	    // Module
	    //
	    ModuleDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);
	    
	    // Don't add this here!!!
	    // if(!check(desc -> id))
	    //     continue;

            String_var name =
                getPrefixedName("OBStubImpl", seq[i].contained_object);

	    //comment(desc -> id, out);
	    out.sep();
            out << "\nnamespace " << name;
            out << "\n{";

	    assert(!is_nil(cont));
	    genCPlusPlusHStubImplVarFwd(cont, out);

            out.sep();
            out << "\n} // End of namespace " << name;

	    break;
	}

	case dk_Interface:
	{
	    //
	    // Interface
	    //
	    InterfaceDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

            String_var name =
                getPrefixedName("OBStubImpl", seq[i].contained_object);

	    //comment(desc -> id, out);

	    out.sep();
	    out << "\ntypedef OB::ObjVar< " << name << " > "
                << name << "_var;";

            break;
	}

        default:
            break;
        }
    }
}

// ----------------------------------------------------------------------
// Generate code for the header file, general part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusH(Container_ptr container, IdlPrettyPrint& out)
{
    //
    // Get scope
    //
    String_var scope = getScope(container);

    //
    // Get container contents descriptions
    //
    Container::DescriptionSeq_var seq =
	container -> describe_contents(dk_all, true, -1);

    ULong i, j;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get container
	//
	Container_var cont = Container::_narrow(seq[i].contained_object);

	switch(seq[i].kind)
	{
	case dk_Constant:
	{
	    //
	    // Constant
	    //
	    ConstantDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);
	    
	    if(!check(desc -> id))
		continue;
	    
	    Identifier_var name = fixKwd(desc -> name);
	    
	    comment(desc -> id, out);
	    
	    out.sep();
	    out << '\n';
	    
	    if(strlen(dllImport_) > 0)
		out << dllImport_ << ' ';

	    if(isTypeCodeExtern(container))
		out << "extern ";
	    else
		out << "static ";
	    
	    String_var s = getTypeString(scope, desc -> type,
					       GetTypeConst, false, name);
	    out << s << ';';
	    
	    break;
	}

	case dk_Module:
	{
	    //
	    // Module
	    //
	    ModuleDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    // Don't add this here!!!
	    // if(!check(desc -> id))
	    //     continue;

            Identifier_var name = fixKwd(desc -> name);

	    comment(desc -> id, out);
            out << "\nnamespace " << name;
            out << "\n{";

	    assert(!is_nil(cont));
	    genCPlusPlusH(cont, out);

            out.sep();
            out << "\n} // End of namespace " << name;
	    
	    break;
	}
	
	case dk_Interface:
	{
	    //
	    // Interface
	    //
	    InterfaceDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);
	    
	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);
	    ScopedName_var absolute = getAbsolute(desc -> id);

	    comment(desc -> id, out);
		
	    //
	    // For an abstract interface, we first generate an abstract
	    // class that derives from CORBA::AbstractBase instead of
	    // CORBA::Object
	    //
	    if(desc -> is_abstract)
            {
                out << "\nclass " << name << " : ";

                unsigned int indent = out.getIndent();
                out.setIndent(out.getPos());

                if(desc -> base_interfaces.length() > 0)
                {
                    for(j = 0 ; j < desc -> base_interfaces.length() ; j++)
                    {
                        const char* id = desc -> base_interfaces[j];
                        //
                        // Using fully-qualified class name to avoid
                        // ambiguities
                        //
                        ScopedName_var baseAbs = getAbsolute(id);

                        if(j > 0)
                            out << ",\n";

                        //
                        // Specification requires use of virtual inheritance
                        // for abstract interfaces (all base interfaces of
                        // an abstract interface must be abstract)
                        //
                        out << "virtual public " << baseAbs;
                    }
                }
                else
                {
                    out << "virtual public CORBA::AbstractBase";
                }

                out.setIndent(indent);

                out.start();

                out << "\nvoid operator=(const " << name << "&);";

                out.sep();
                out.sep();
                out.dec();
                out << "\nprotected:";
                out.inc();

                //
                // Abstract interfaces require a protected default constructor,
                // copy constructor and virtual destructor.
                //
                out.sep();
                out << '\n' << name << "() { }";
                out << '\n' << name << "(const " << name << "&) { }";
                out << "\nvirtual ~" << name << "() { }";

                out.sep();
                out.dec();
                out << "\npublic:";
                out.inc();

                out.sep();
                out << "\ntypedef " << name << "_ptr _ptr_type;";
                out << "\ntypedef " << name << "_var _var_type;";

                out.sep();
                out << "\nstatic " << name
                    << "_ptr _narrow(CORBA::AbstractBase_ptr);";

                out.sep();
                out << "\nstatic " << name << "_ptr _duplicate("
                    << name << "_ptr p)";
		out.start();
		out << "\nCORBA::AbstractBase::_duplicate(p);";
		out << "\nreturn p;";
		out.end();

                out.sep();
                out << "\nstatic inline " << name << "_ptr\n_nil()";
                out.start();
                out << "\nreturn 0;";
                out.end();

		assert(!is_nil(cont));
		genCPlusPlusH(cont, out);

		//
		// CDR functions
		//
		out.sep();
		out << "\nstatic void _OB_unmarshal(" << name
		    << "_ptr&, OB::InputStream&);";

		out.end();
		out << ';';
		out.sep();
            }

            //
            // Generate the interface class.  For an abstract interface Foo,
            // the name of the interface class is OBAbstractImpl_Foo.
            //

            Identifier_var className;
            if(desc -> is_abstract)
            {
                className = string_dup("OBAbstractImpl_");
                className += name;
            }
            else
                className = name;

	    if(local_)
	    {
		out << "\n// Locality constrained";
		out << "\n//";
	    }
	    out << "\nclass " << className << " : ";
	    
	    unsigned int indent = out.getIndent();
	    out.setIndent(out.getPos());

            //
            // If the interface is abstract, this class must also inherit
            // from the abstract base class
            //
            if(desc -> is_abstract)
                out << "virtual public " << absolute << ",\n";

            InterfaceDef_var intf =
                InterfaceDef::_narrow(seq[i].contained_object);
	    assert(!is_nil(intf));
            IdlBaseInterfaceInfoSeq baseInfo;
            IdlGetBaseInterfaces(intf, baseInfo, false);

	    if(baseInfo.length() > 0)
	    {
		for(j = 0 ; j < baseInfo.length() ; j++)
		{
		    if(j > 0)
			out << ",\n";
		    
		    //
		    // Use fully-qualified class name to avoid ambiguities
		    //
		    if(baseInfo[j].is_abstract)
		    {
			out << "virtual public ";
                        String_var baseName = baseInfo[j].intf -> name();
                        baseName = fixKwd(baseName);
                        Container_var c = baseInfo[j].intf -> defined_in();
                        String_var baseScope = getScope(c);
                        out << baseScope << "OBAbstractImpl_" << baseName;
		    }
		    else
		    {
			ScopedName_var baseAbs = getAbsolute(baseInfo[j].id);

			if(!noVirtual_)
			    out << "virtual ";

			out << "public " << baseAbs;
		    }
		}
	    }
	    else
	    {
		if(!noVirtual_)
		    out << "virtual ";
		
		out << "public CORBA::Object";
	    }
	    
	    out.setIndent(indent);
	    
	    out.start();
	    
	    out << '\n' << className << "(const " << className << "&);";
	    out << "\nvoid operator=(const " << className << "&);";
	    out.sep();
	    out << "\nstatic ";
	    if(strlen(dllImport_) > 0)
		out << dllImport_ << ' ';
	    out << "const char* _ob_ids_[];";

	    out.sep();
	    out.dec();
	    if(local_)
	       out << "\npublic:";
	    else
	       out << "\nprotected:";
	    out.inc();
	    out.sep();
	    if(!local_)
	    {
		out << "\nOBStubImpl_" << absolute
		    << "_ptr _OB_getStubImpl();";
		out << "\nvoid _OB_clearStubImpl();";

                //
                // We have to use these mangled names to avoid a bug
                // in MSVC
                //
                String_var flat = getFlattenedName(absolute);
                String_var derivedStub = string_dup("_ob_stubImpl_");
                derivedStub += flat;
                derivedStub += "_";

		out.sep();
		out << "\nOBStubImpl_" << absolute
		    << "_var " << derivedStub << ';';

		// XXX Friends don't work right with egcs 1.1.1, so
		//     we need to make the constructor public.  We
		//     should remove this when it is supported on all
		//     platforms.
		out.sep();
		//out.direct("\n#ifdef __GNUC__");
		out.dec();
		out << "\npublic:";
		out.inc();
		//out.direct("\n#endif");

		out.sep();
		out << '\n' << className << "() : " << derivedStub
                    << "(0) { }";
	    }
	    else
	    {
		out.sep();
		out << '\n' << className << "() { _ob_locality_ = true; }";
	    }

	    if(!local_)
	    {
		out.sep();
		out.dec();
		out << "\npublic:";
		out.inc();
	    }

            //
            // These aren't needed in the concrete class for an abstract
            // interface
            //
            if(!desc -> is_abstract)
            {
		out.sep();
		out << "\ntypedef " << className << "_ptr _ptr_type;";
		out << "\ntypedef " << className << "_var _var_type;";
		out.sep();
		out << "\nstatic inline " << className << "_ptr\n_duplicate("
		    << className << "_ptr p)";
		out.start();
		out << "\nif(p)";
		out.inc();
		out << "\np -> _OB_incRef();";
		out.dec();
		out << "\nreturn p;";
		out.end();
		
		out.sep();
		out << "\nstatic inline " << className << "_ptr\n_nil()";
		out.start();
		out << "\nreturn 0;";
		out.end();
		
		out.sep();
		// out << "\nvirtual CORBA::Boolean _is_a(const char*);";
		out << "\nstatic " << className
		    << "_ptr _narrow(CORBA::Object_ptr);";
            }

	    out.sep();
	    out << "\nvirtual const char** _OB_ids() const;";
	    out << "\nstatic const char** _OB_staticIds();";

            if(desc -> is_abstract)
            {
                //
                // For an abstract interface class, we generate attribute
                // and operation declarations, but no nested types.  In
                // other words, don't do this:
                //
		//genCPlusPlusH(cont, out);
                //
                Container::DescriptionSeq_var contents =
                    intf -> describe_contents(dk_all, true, -1);

                for(j = 0 ; j < contents -> length() ; j++)
                {
                    if(contents[j].kind == dk_Operation)
                        genCPlusPlusHOperation(contents[j], "", false, false,
                                               out);
                    else if(contents[j].kind == dk_Attribute)
                        genCPlusPlusHAttribute(contents[j], "", false, false,
                                               out);
                }

                //
                // Override function inherited from AbstractBase
                //
                out.sep();
                out << "\nvirtual CORBA::Object_ptr _OB_toObject();";

                //
                // The abstract base class provides the unmarshalling function
                //
            }
            else
            {
		assert(!is_nil(cont));
		genCPlusPlusH(cont, out);

		//
		// CDR functions - _OB_marshal and _OB_marshalCount are not
		// necessary; write_Object() and count_Object() are used
		// directly
		//
		out.sep();
		out << "\nstatic void _OB_unmarshal(" << className
		    << "_ptr&, OB::InputStream&);";
		
		if(!noTC_)
		{
		    // XXX Friends don't work right with egcs 1.1.1.
		    //     We should enable this when it is supported
		    //     on all platforms.  Also see default constructor.
		    //out.sep();
		    //out.direct("\n#ifndef __GNUC__");
		    //out << "\nfriend CORBA::Boolean ::operator>>=("
			//<< "const CORBA::Any&, " << className << "_ptr&);";
		    //out.direct("\n#endif");
		}
	    }

	    if(!local_)
	    {
		out.sep();
		out << "\nvirtual OB::StubImplBase* _OB_createStubImpl();";
	    }

            //
            // If the interface is abstract, or derives from an abstract
            // interface, we need to generate some additional functions
            //
            bool hasAbstractBase = false;
            IdlBaseInterfaceInfoSeq allBaseInfo;
            IdlGetBaseInterfaces(intf, allBaseInfo, true);
            for(j = 0 ; j < allBaseInfo.length() ; j++)
            {
                if(allBaseInfo[j].is_abstract)
                {
                    hasAbstractBase = true;
                    break;
                }
            }

            if(desc -> is_abstract || hasAbstractBase)
            {
		out.sep();
		out << "\nvirtual CORBA::AbstractBase_ptr "
                    << "_OB_toAbstractBase();";
		out.sep();
		out << "\nvirtual void _OB_insertAny(CORBA::Any&);";
            }

            out.end();
	    out << ';';
	    
	    if(!noTC_)
	    {
		out.sep();
		
		out << '\n';
		
		if(strlen(dllImport_) > 0)
		    out << dllImport_ << ' ';
		
		if(isTypeCodeExtern(container))
		    out << "extern OB::TypeCodeConst " << "_tc_"
                        << name << ';';
		else
		    out << "static OB::TypeCodeConst " << "_tc_"
                        << name << ';';
	    }
	    
	    break;
	}
	
	case dk_Value:
        {
            //
            // Valuetype
            //
            ValueDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            ValueDef_var value = ValueDef::_narrow(cont);
            assert(! is_nil(value));

            Identifier_var name = fixKwd(desc -> name);
            ScopedName_var absolute = getAbsolute(desc -> id);

            //
            // Need to generate inherited operation and attribute
            // signatures if this valuetype supports an interface
            //
            InterfaceDef_var intf = value -> supported_interface();
            bool supportsInterface = (! is_nil(intf));

            comment(desc -> id, out);

            out.sep();
            out << "\nclass " << name << " : ";

            unsigned int indent = out.getIndent();
            out.setIndent(out.getPos());

            if(desc -> is_custom && strlen(desc -> base_value) == 0)
            {
                //
                // Class must inherit from CustomMarshal
                //
                out << "virtual public CORBA::CustomMarshal";

                if(desc -> abstract_base_values.length() > 0)
                    out << ",\n";

                for(j = 0 ; j < desc -> abstract_base_values.length() ; j++)
                {
                    const char* id = desc -> abstract_base_values[j];
                    //
                    // Using fully-qualified class name to avoid
                    // ambiguities
                    //
                    ScopedName_var baseAbs = getAbsolute(id);

                    if(j > 0)
                        out << ",\n";

                    out << "virtual public " << baseAbs;
                }
            }
            else if(strlen(desc -> base_value) > 0 ||
                    desc -> abstract_base_values.length() > 0)
            {
                if(strlen(desc -> base_value) > 0)
                {
                    //
                    // Using fully-qualified class name to avoid
                    // ambiguities
                    //
                    ScopedName_var baseAbs =
                        getAbsolute(desc -> base_value);

                    out << "virtual public " << baseAbs;

                    if(desc -> abstract_base_values.length() > 0)
                        out << ",\n";
                }

                for(j = 0 ; j < desc -> abstract_base_values.length() ; j++)
                {
                    const char* id = desc -> abstract_base_values[j];
                    //
                    // Using fully-qualified class name to avoid
                    // ambiguities
                    //
                    ScopedName_var baseAbs = getAbsolute(id);

                    if(j > 0)
                        out << ",\n";

                    out << "virtual public " << baseAbs;
                }
            }
            else
            {
                out << "virtual public CORBA::ValueBase";
            }

            out.setIndent(indent);

            out.start();

            out << "\nvoid operator=(const " << name << "&);";

            out.sep();
            out << "\nstatic const char* _ob_id_;";

            //
            // _ob_truncatable_ids_ - if there are any truncatable bases, then
            // declare a static array
            //
            ValueDefSeq truncBases;
            IdlGetAllTruncatableBaseValues(value, truncBases);
            if(truncBases.length() > 0)
            {
                out.sep();
                out << "\nstatic const char* _ob_truncatable_ids_[];";
            }

            out.sep();
            out.dec();
            out << "\nprotected:";
            out.inc();

            out.sep();
            out << '\n' << name << "() { }";
            out << "\nvirtual ~" << name << "() { }";

            out.sep();
            out.dec();
            out << "\npublic:";
            out.inc();

            out.sep();
            out << "\nstatic " << name << "* _downcast("
                << "CORBA::ValueBase*);";

            //
            // _OB_id - Function to provide static type
            //
            out.sep();
            out << "\nstatic const char* _OB_id();";

            //
            // _OB_truncatableIds - if there are any truncatable bases, then
            // declare this function, otherwise we inherit the one from
            // ValueBase
            //
            if(truncBases.length() > 0)
            {
                out.sep();
                out << "\nvirtual const char** _OB_truncatableIds() const;";
            }

            //
            // Generate operations and attributes from supported
            // interface(s)
            //
            if(supportsInterface)
            {
                Container::DescriptionSeq_var ops =
                    intf -> describe_contents(dk_Operation, false, -1);

                for(j = 0 ; j < ops -> length() ; j++)
                    genCPlusPlusHOperation(ops[j], scope, true, true, out);

                Container::DescriptionSeq_var attrs =
                    intf -> describe_contents(dk_Attribute, false, -1);

                for(j = 0 ; j < attrs -> length() ; j++)
                    genCPlusPlusHAttribute(attrs[j], scope, true, true, out);
            }

            //
            // Generate non-ValueMember children
            //
            assert(! is_nil(cont));
            genCPlusPlusH(cont, out);

            //
            // Get value scope
            //
            String_var valScope = getScope(cont);

            //
            // Get all value members
            //
            Container::DescriptionSeq_var members =
                cont -> describe_contents(dk_ValueMember, true, -1);

            //
            // Generate methods for public value members
            //
            bool havePrivate = false;
            for(j = 0 ; j < members -> length() ; j++)
            {
                ValueMember* member;
                b = members[j].value >>= member;
                assert(b);

                if(member -> access == PUBLIC_MEMBER)
                    genCPlusPlusHValueMember(cont, member, valScope,
                                             false, out);
                else
                    havePrivate = true;
            }

            if(havePrivate)
            {
                out.sep();
                out.dec();
                out << "\nprotected:";
                out.inc();

                //
                // Generate private value members
                //
                for(j = 0 ; j < members -> length() ; j++)
                {
                    ValueMember* member;
                    b = members[j].value >>= member;
                    assert(b);

                    if(member -> access == PRIVATE_MEMBER)
                        genCPlusPlusHValueMember(cont, member, valScope,
                                                 false, out);
                }

                out.sep();
                out.dec();
                out << "\npublic:";
                out.inc();
            }

            //
            // Function to provide dynamic type
            //
            out.sep();
            out << "\nvirtual const char* _OB_typeId() const;";

            if(desc -> is_custom)
            {
                out.sep();
                out << "\nvirtual CORBA::Boolean _OB_isCustom() const;";
            }

            if(desc -> is_truncatable)
            {
                out.sep();
                out << "\nvirtual CORBA::Boolean _OB_isTruncatable() const;";
            }

            //
            // Generate marshalling function signatures if the
            // valuetype is not abstract
            //
            if(! desc -> is_abstract)
            {
                out.sep();
                out << "\nvirtual void _OB_marshalValue(OB::OutputStream&);";
                out << "\nvirtual void _OB_marshalCountValue("
                    << "OB::CountStream&);";
                out << "\nvirtual void _OB_unmarshalValue(OB::InputStream&);";
            }

            if(!local_)
            {
                //
                // CDR
                //
		out.sep();
                out << "\nstatic void _OB_marshal(" << name
                    << "*, OB::OutputStream&);";
                out << "\nstatic void _OB_marshalCount(" << name
                    << "*, OB::CountStream&);";
                out << "\nstatic void _OB_unmarshal(" << name
                    << "*&, OB::InputStream&);";
            }

            //
            // End of class
            //
            out.end();
            out << ';';

            if(!noTC_)
            {
                out.sep();

                out << '\n';

                if(strlen(dllImport_) > 0)
                    out << dllImport_ << ' ';

                if(isTypeCodeExtern(container))
                    out << "extern OB::TypeCodeConst " << "_tc_"
                        << name << ';';
                else
                    out << "static OB::TypeCodeConst " << "_tc_"
                        << name << ';';
            }

            //
            // Generate type-specific value factory class, if necessary
            //
            genCPlusPlusHValueFactory(value, out);

            break;
        }

	case dk_ValueBox:
	{
	    //
	    // Value box
	    //
	    TypeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);
	    
	    if(!check(desc -> id))
		continue;
	    
	    ValueBoxDef_var value =
		ValueBoxDef::_narrow(seq[i].contained_object);
	    assert(! is_nil(value));
	    
	    Identifier_var name = fixKwd(desc -> name);
	    ScopedName_var absolute = getAbsolute(desc -> id);
	    
	    comment(desc -> id, out);
	    
	    out.sep();
	    out << "\nclass " << name << " : ";
	    
	    if(!noVirtual_)
		out << "virtual ";
	    
	    out << "public CORBA::DefaultValueRefCountBase";
	    
	    out.start();

            out << "\nstatic const char* _ob_id_;";

            IDLType_var origTypeDef = value -> original_type_def();
            TypeCode_var type = origTypeDef -> type();
            TypeCode_var origType = GetOrigType(type);
	    
	    switch(origType -> kind())
	    {
	    case tk_null:
	    case tk_void:
		assert(false);
		break;
		
	    case tk_native:
	    case tk_Principal:
	    case tk_except:
	    case tk_TypeCode:
	    case tk_value:
	    case tk_value_box:
	    case tk_abstract_interface:
		assert_nca(false, NCANotImpl);
		break;
		
	    case tk_short:
	    case tk_long:
	    case tk_longlong:
	    case tk_ushort:
	    case tk_ulong:
	    case tk_ulonglong:
	    case tk_float:
	    case tk_double:
	    case tk_longdouble:
	    case tk_boolean:
	    case tk_octet:
	    case tk_char:
	    case tk_wchar:
	    case tk_enum:
	    case tk_objref:
		genCPlusPlusHBasicBox(value, scope, absolute, desc, type, out);
		break;
		
	    case tk_struct:
		genCPlusPlusHStructBox(value, scope, absolute, desc, type,
                                       out);
		break;
		
	    case tk_string:
	    case tk_wstring:
		genCPlusPlusHStringBox(value, scope, absolute, desc, type,
                                       out);
		break;
		
	    case tk_fixed:
		genCPlusPlusHFixedBox(value, scope, absolute, desc, type,
                                      out);
		break;
		
	    case tk_union:
		genCPlusPlusHUnionBox(value, scope, absolute, desc, type, out);
		break;
		
	    case tk_sequence:
		genCPlusPlusHSequenceBox(value, scope, absolute, desc, type,
                                         out);
		break;

	    case tk_any:
		genCPlusPlusHAnyBox(value, scope, absolute, desc, type, out);
		break;

	    case tk_array:
		genCPlusPlusHArrayBox(value, scope, absolute, desc, type, out);
		break;

	    case tk_alias:
		assert(false);
		break;
	    }

	    out.sep();
	    out << "\nvirtual CORBA::ValueBase* _copy_value();";
	    out << "\nstatic " << name << "* _downcast(CORBA::ValueBase*);";

            //
            // Function to provide static type
            //
            out.sep();
            out << "\nstatic const char* _OB_id();";

            //
            // Function to provide dynamic type
            //
            out.sep();
	    out << "\nvirtual const char* _OB_typeId() const;";

	    //
	    // Dynamic marshalling functions
	    //
	    out.sep();
	    out << "\nvirtual void _OB_marshalValue(OB::OutputStream&);";
	    out << "\nvirtual void _OB_marshalCountValue(OB::CountStream&);";
	    out << "\nvirtual void _OB_unmarshalValue(OB::InputStream&);";

            if(!local_)
            {
                //
                // CDR
                //
		out.sep();
                out << "\nstatic void _OB_marshal(" << name
                    << "*, OB::OutputStream&);";
                out << "\nstatic void _OB_marshalCount(" << name
                    << "*, OB::CountStream&);";
                out << "\nstatic void _OB_unmarshal(" << name
                    << "*&, OB::InputStream&);";
            }

	    //
	    // End of class
	    //
	    out.end();
	    out << ';';

	    if(!noTC_)
	    {
		out.sep();

		out << '\n';

		if(strlen(dllImport_) > 0)
		    out << dllImport_ << ' ';

		if(isTypeCodeExtern(container))
		    out << "extern OB::TypeCodeConst " << "_tc_"
                        << name << ';';
		else
		    out << "static OB::TypeCodeConst " << "_tc_"
                        << name << ';';
	    }

	    break;
	}

	case dk_Alias:
	{
	    //
	    // Alias
	    //
	    TypeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);

	    comment(desc -> id, out);

	    TypeCode_var contentType = desc -> type -> content_type();
	    static const char* extPtr = "_ptr";
	    static const char* extRef = "Ref";
	    static const char* extVar = "_var";
	    static const char* extOut = "_out";
	    static const char* extSlice = "_slice";
	    static const char* extForAny = "_forany";

	    String_var nameNorm = name;

	    String_var namePtr = name;
	    namePtr += extPtr;

	    String_var nameRef = name;
	    nameRef += extRef;

	    String_var nameVar = name;
	    nameVar += extVar;

	    String_var nameOut = name;
	    nameOut += extOut;

	    String_var nameSlice = name;
	    nameSlice += extSlice;

	    String_var nameForAny = name;
	    nameForAny += extForAny;

	    String_var uniqueType;
	    if(contentType -> kind() == tk_sequence)
	    {
		uniqueType = string_dup("OBUnique_");
		uniqueType += name;
		out << "\nclass " << uniqueType << " { };";
	    }

	    String_var norm = getTypeString(scope, contentType, GetTypeNormal,
					    false, nameNorm, uniqueType);
	    String_var ptr = getTypeString(scope, contentType, GetTypePtr,
					   false, namePtr, uniqueType);
	    String_var ref = getTypeString(scope, contentType, GetTypeRef,
					   false, nameRef, uniqueType);
	    String_var var = getTypeString(scope, contentType, GetTypeVar,
					   false, nameVar, uniqueType);
	    // Can't name this out, because this is already used for
	    // the output stream
	    String_var outStr = getTypeString(scope, contentType, GetTypeOut,
					      false, nameOut, uniqueType);
	    String_var slice = getTypeString(scope, contentType, GetTypeSlice,
					     false, nameSlice, uniqueType);
	    String_var forAny = getTypeString(scope, contentType,
					      GetTypeForAny, false, nameForAny,
					      uniqueType);

	    out.sep();
	    
	    if(strncmp("???", norm, 3) != 0)
		out << "\ntypedef " << norm << ';';
	    
	    if(strncmp("???", ptr, 3) != 0)
		out << "\ntypedef " << ptr << ';';
	    
	    if(strncmp("???", ref, 3) != 0)
		out << "\ntypedef " << ref << ';';
	    
	    if(strncmp("???", slice, 3) != 0)
		out << "\ntypedef " << slice << ';';
	    
	    if(strncmp("???", var, 3) != 0)
		out << "\ntypedef " << var << ';';
	    
	    if(strncmp("???", outStr, 3) != 0)
		out << "\ntypedef " << outStr << ";";
	    
	    if(strncmp("???", forAny, 3) != 0)
		out << "\ntypedef " << forAny << ';';


	    String_var externStr = string_dup("\n");
	    String_var inlineStr = string_dup("\n");

	    if(strlen(dllImport_) > 0)
	    {
		externStr += dllImport_;
		externStr += ' ';
	    }

	    if(isTypeCodeExtern(container))
	    {
		externStr += "extern ";
		inlineStr += "inline ";
	    }
	    else
	    {
		externStr += "static ";
		inlineStr += "static inline ";
	    }

	    if(!noTC_)
		out << externStr << "OB::TypeCodeConst " << "_tc_"
		    << name << ';';

	    TypeCode_var origContentType = GetOrigType(contentType);
	    if(origContentType -> kind() == tk_array)
	    {
		//
		// Create array memory management operations
		//
		TypeCode_var arrayContentType;
		ULong length;
		getArrayType(origContentType, arrayContentType.out(), length);
		String_var cont = getTypeString(scope, arrayContentType,
						      GetTypeArray, true);

		out.sep();
		out << inlineStr << nameSlice << "*\n" << name << "_alloc()";
		out.start();
		out << "\nreturn ";
		if(length != origContentType -> length())
		    out << '(' << nameSlice << "*) ";
		out << "new " << cont << '[' << length << "];";
		out.end();

		out.sep();
		out << inlineStr << nameSlice << "*\n" << name
		    << "_dup(const " << nameSlice << "* s)";
		out.start();
		out << '\n' << nameSlice << "* to = ";
		if(length != origContentType -> length())
		    out << '(' << nameSlice << "*) ";
		out << "new " << cont << '[' << length << "];";
                writeArrayCopy(scope, contentType, "to", "s", "i", out);
		out << "\nreturn to;";
		out.end();

		out.sep();
		out << inlineStr << "void\n" << name
		    << "_copy(const " << nameSlice << "* s, "
		    << nameSlice << "* to)";
		out.start();
                writeArrayCopy(scope, contentType, "to", "s", "i", out);
		out.end();

		out.sep();
		out << inlineStr << "void\n" << name
		    << "_free(" << nameSlice << "* s)";
		out.start();
                writeArrayFree(scope, contentType, "s", out);
		out.end();

		out.sep();
		out.direct("\n#ifdef HAVE_VCPLUSPLUS_BUGS");

		out.sep();
		out << inlineStr << nameSlice << "*\n" << name
		    << "_dup(" << nameSlice << "* s)";
		out.start();
		out << "\nreturn " << name << "_dup((const "
		    << nameSlice << "*) s);";
		out.end();

		out.sep();
		out << inlineStr << "void\n" << name
		    << "_copy(" << nameSlice << "* s, "
		    << nameSlice << "* to)";
		out.start();
		out << '\n' << name << "_copy((const "
		    << nameSlice << "*) s, to);";
		out.end();
		out.sep();
		out.direct("\n#endif");	// HAVE_VCPLUSPLUS_BUGS

                if(!noTC_)
                {
		    //
		    // Create the "info type" for needed for any insertion
		    // and extraction
		    //
		    String_var infoType = string_dup("OBInfo_");
		    infoType += name;

		    out.sep();
		    out << "\nstruct " << infoType
			<< " : public OB::ConstructedInfo";
		    out.start();
		    out << '\n' << infoType << "() { }";
		    out.sep();
		    out << "\nvirtual void* alloc() const";
		    out.start();
		    out << "\nreturn " << name << "_alloc();";
		    out.end();
		    out.sep();
		    out << "\nvirtual void free(void* p) const";
		    out.start();
		    out << '\n' << name << "_free((" << name << "_slice*)p);";
		    out.end();
		    out.sep();
		    out << "\nvirtual void* dup(const void* p) const";
		    out.start();
		    out << "\nreturn " << name << "_dup((const "
			<< name << "_slice*)p);";
		    out.end();
		    out.sep();
		    out << "\nvirtual void marshal(const void*, "
			<< "OB::OutputStream&) const;";
		    out << "\nvirtual void marshalCount(const void*, "
			<< "OB::CountStream&) const;";
		    out << "\nvirtual void unmarshal(void*, "
			<< "OB::InputStream&) const;";
		    out.end();
		    out << ';';
		}
	    }
	    else if(origContentType -> kind() == tk_sequence)
	    {
                if(!noTC_)
                {
		    //
		    // Create the "info type" for needed for any insertion
		    // and extraction
		    //
		    String_var infoType = string_dup("OBInfo_");
		    infoType += name;

		    out.sep();
		    out << "\nstruct " << infoType
			<< " : public OB::ConstructedInfo";
		    out.start();
		    out << '\n' << infoType << "() { }";
		    out.sep();
		    out << "\nvirtual void* alloc() const";
		    out.start();
		    out << "\nreturn new " << name << ';';
		    out.end();
		    out.sep();
		    out << "\nvirtual void free(void* p) const";
		    out.start();
		    out << "\ndelete (" << name << "*)p;";
		    out.end();
		    out.sep();
		    out << "\nvirtual void* dup(const void* p) const";
		    out.start();
		    out << "\nreturn new " << name << "(*(const "
			<< name << "*)p);";
		    out.end();
		    out.sep();
		    out << "\nvirtual void marshal(const void*, "
			<< "OB::OutputStream&) const;";
		    out << "\nvirtual void marshalCount(const void*, "
			<< "OB::CountStream&) const;";
		    out << "\nvirtual void unmarshal(void*, "
			<< "OB::InputStream&) const;";
		    out.end();
		    out << ';';
		}
	    }

	    break;
	}

	case dk_Enum:
        {
            //
            // Enum
            //
            TypeDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            Identifier_var name = fixKwd(desc -> name);

            comment(desc -> id, out);

            out.sep();
            out << "\nenum " << name;
            out.start();

            for(j = 0 ; j < desc -> type -> member_count() ; j++)
            {
                Identifier_var memberName =
                    desc -> type -> member_name(j);
                memberName = fixKwd(memberName);
                out << '\n' << memberName;
                if(j < desc -> type -> member_count() - 1)
                    out << ',';
            }

            out.end();
            out << ';';

            out.sep();
            out << "\ntypedef " << name << "& " << name << "_out;";

            if(!noTC_)
            {
                out.sep();

                out << '\n';

                if(strlen(dllImport_) > 0)
                    out << dllImport_ << ' ';

                if(isTypeCodeExtern(container))
                    out << "extern OB::TypeCodeConst " << "_tc_"
                        << name << ';';
                else
                    out << "static OB::TypeCodeConst " << "_tc_"
                        << name << ';';
            }

            break;
        }

	case dk_Struct:
	case dk_Exception:
	{
	    RepositoryId_var id;
	    Identifier_var name;
	    TypeCode_var type;
	    
	    if(seq[i].kind == dk_Struct)
	    {
		//
		// Struct
		//
		TypeDescription* desc;
		Boolean b = seq[i].value >>= desc;
		assert(b);
		
		name = fixKwd(desc -> name);
		id = desc -> id;
		type = desc -> type;
	    }
	    else
	    {
		//
		// Exception
		//
		ExceptionDescription* desc;
		Boolean b = seq[i].value >>= desc;
		assert(b);
		
		name = fixKwd(desc -> name);
		id = desc -> id;
		type = desc -> type;
	    }
	    
	    if(!check(id))
		continue;
	    
	    bool variable = IdlIsVariable(type);
	    
	    comment(id, out);
	    if(seq[i].kind == dk_Struct)
	    {
		out.sep();
		out << "\nstruct " << name << ';';
		if(variable)
		{
		    out << "\ntypedef OB::VarVar< " << name << " > "
			<< name << "_var;";
		    out << "\ntypedef OB::VarOut< " << name << " > "
			<< name << "_out;";
		}
		else
		{
		    out << "\ntypedef OB::FixVar< " << name << " > "
			<< name << "_var;";
		    out << "\ntypedef " << name << "& " << name << "_out;";
		}
	    }
	    
	    out.sep();
	    out << "\nstruct " << name;
	    if(seq[i].kind == dk_Exception)
		out << " : public CORBA::UserException";
	    
	    out.start();
	    out.ignoreNextSep(true);
	    
	    if(variable) // Note: Exceptions are always variable
	    {
		bool hasFixed = false;
		for(j = 0 ; j < type -> member_count() ; j++)
		{
		    TypeCode_var tc = type -> member_type(j);
		    if(!IdlIsVariable(tc))
		    {
			hasFixed = true;
			break;
		    }
		}

		if(hasFixed)
		{
		    out.sep();
		    out.direct("\n#ifdef OB_CLEAR_MEM");
		    out << '\n' << name << "();";
		    out.direct("\n#else");
		    out << '\n' << name << "() { }";
		    out.direct("\n#endif"); // OB_CLEAR_MEM
		}
		else
		{
		    out.sep();
		    out << '\n' << name << "() { }";
		}
		
		out << '\n' << name << "(const " << name << "&);";
		
		out << '\n' << name << "& operator=(const " << name << "&);";
		
		if(seq[i].kind == dk_Exception)
		{
		    out.sep();
		    out << "\nstatic " << name
			<< "* _downcast(CORBA::Exception*);";
		    out << "\nstatic const " << name
			<< "* _downcast(const CORBA::Exception*);";
		    out << "\nvirtual void _raise() const { throw *this; }";
		    out.sep();
		    out << "\nvirtual CORBA::Exception* _OB_clone() const;";
		    out << "\nvirtual void _OB_insertInAny(CORBA::Any&);";
		}
	    }
	    
	    if(seq[i].kind == dk_Struct)
	    {
		out.sep();
		out << "\ntypedef " << name << "_var _var_type;";
	    }

	    for(j = 0 ; j < type -> member_count() ; j++)
	    {
		if(j == 0)
		    out.sep();

		TypeCode_var memberType = type -> member_type(j);
		Identifier_var memberName = type -> member_name(j);
		Identifier_var fixedName = fixKwd(memberName);
		
		if(memberType -> kind() == tk_sequence)
		{
		    String_var anonName = string_dup("_");
		    anonName += memberName;
		    anonName += "_seq";

		    String_var uniqueType =
			string_dup("OBUnique_");
		    uniqueType += memberName;
		    out << "\nclass " << uniqueType << " { };";

		    String_var norm = getTypeString("", memberType,
							  GetTypeNormal,
							  false, anonName,
							  uniqueType);

		    out << "\ntypedef " << norm << ';';
		    out << '\n' << anonName << ' ' << fixedName << ';';
		}
		else
		{
		    String_var str =
			getTypeString(scope, memberType, GetTypeStruct, false,
				      fixedName);
		    out << '\n' << str << ';';
		}
	    }
	    
	    //
	    // Generate this constructor after the anon sequence types
	    // to avoid pre-declaration worries.
	    //
	    if(seq[i].kind == dk_Exception &&
	       type -> member_count() > 0)
	    {
		out.sep();
		out << '\n' << name << '(';
		unsigned int indent = out.getIndent();
		out.setIndent(out.getPos());
		for(j = 0 ; j < type -> member_count() ; j++)
		{
		    TypeCode_var memberType = type -> member_type(j);
		    
		    String_var s;
		    if(memberType -> kind() == tk_sequence)
		    {
			String_var memberName = type -> member_name(j);
			s = string_dup("const _");
			s += memberName;
			s += "_seq&";
		    }
		    else
			s = getTypeString(scope, memberType, GetTypeIn);
		    
		    if(j > 0)
			out << ",\n";
		    
		    out << s;
		}
		out << ");";
		out.setIndent(indent);
	    }

            if(!local_)
            {
                //
                // CDR
                //
		out.sep();
                out << "\nstatic void _OB_marshal(const " << name
                    << "&, OB::OutputStream&);";
                out << "\nstatic void _OB_marshalCount(const " << name
                    << "&, OB::CountStream&);";
                out << "\nstatic void _OB_unmarshal(" << name
                    << "&, OB::InputStream&);";
            }
	    
	    out.end();
	    out << ';';
	    
	    if(!noTC_)
	    {
		out.sep();
		
		out << '\n';
		
		if(strlen(dllImport_) > 0)
		    out << dllImport_ << ' ';
		
		if(isTypeCodeExtern(container))
		    out << "extern OB::TypeCodeConst " << "_tc_"
                        << name << ';';
		else
		    out << "static OB::TypeCodeConst " << "_tc_"
                        << name << ';';
	    }
	    
	    break;
	}

	case dk_Union:
	{
	    //
	    // Union
	    //
	    TypeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);
	    
	    if(!check(desc -> id))
		continue;
	    
	    Identifier_var name = fixKwd(desc -> name);
	    
	    TypeCode_var discType = desc -> type -> discriminator_type();
	    String_var disc = getTypeString(scope, discType,
						  GetTypeNormal);
	    
	    IdlUnionMemberInfoSeq_var infoSeq = IdlGetUnionInfo(desc -> type);

	    IdlStringSeq fixedNames;
	    fixedNames.length(infoSeq -> length());
	    for(j = 0 ; j < infoSeq -> length() ; j++)
		fixedNames[j] = fixKwd(infoSeq[j].name);

	    bool variable = IdlIsVariable(desc -> type);
	    
	    comment(desc -> id, out);

	    out.sep();
	    out << "\nclass " << name << ';';
	    if(variable)
	    {
		out << "\ntypedef OB::VarVar< " << name
		    << " > " << name << "_var;";
		out << "\ntypedef OB::VarOut< " << name
		    << " > " << name << "_out;";
	    }
	    else
	    {
		out << "\ntypedef OB::FixVar< " << name
		    << " > " << name << "_var;";
		out << "\ntypedef " << name
		    << "& " << name << "_out;";
	    }
	    
	    //
	    // TODO: Check whether this is really needed
	    //
	    // This is to deal with MSVC bugs. It cannot deal with
	    // code like:
	    // class foo
	    // {
	    // public:
	    //     struct bar { };
	    //     void blah(foo::bar x);
	    // };
	    //
	    String_var realScope = string_dup(scope);
	    realScope += name;
	    realScope += "::";

	    out.sep();
	    out << "\nclass " << name;
	    out.start();
	    bool haveAnonSeq = false;

	    for(j = 0 ; j < infoSeq -> length() ; j++)
	    {
		IdlUnionMemberInfo& info = infoSeq[j];
		if(is_nil(info.type))
		    continue;

		if(info.type -> kind() == tk_sequence)
		{
		    if(!haveAnonSeq)
		    {
			out.dec();
			out << "\npublic:";
			out.inc();
			out.sep();
			haveAnonSeq = true;
		    }

		    String_var anonName = string_dup("_");
		    anonName += info.name;
		    anonName += "_seq";
		    
		    String_var uniqueType =
			string_dup("OBUnique_");
		    uniqueType += info.name;
		    out << "\nclass " << uniqueType << " { };";

		    String_var norm = getTypeString("", info.type,
							  GetTypeNormal,
							  false, anonName,
							  uniqueType);
		    
		    out << "\ntypedef " << norm << ';';
		    out.ignoreNextSep(false);
		}
	    }
	    
	    if(haveAnonSeq)
	    {
		out.sep();
		out.dec();
		out << "\nprivate:";
		out.inc();
		out.sep();
	    }
	    
	    out << "\nunion";
	    out.start();
	    for(j = 0 ; j < infoSeq -> length() ; j++)
	    {
		IdlUnionMemberInfo& info = infoSeq[j];
		if(is_nil(info.type))
		    continue;

		const char* fixedName = fixedNames[j];

		String_var norm;
		if(info.type -> kind() == tk_sequence)
		{
		    String_var anonName = string_dup("_");
		    anonName += info.name;
		    anonName += "_seq";
		    norm = anonName;
		}
		else
		    norm = getTypeString(realScope, info.type, GetTypeNormal);
		
		TypeCode_var origMemberType = GetOrigType(info.type);
		
		switch(origMemberType -> kind())
		{
		case tk_null:
		case tk_void:
		case tk_except:
		case tk_native:
		    assert(false);
		    break;
		    
		case tk_short:
		case tk_long:
		case tk_longlong:
		case tk_ushort:
		case tk_ulong:
		case tk_ulonglong:
		case tk_float:
		case tk_double:
		case tk_longdouble:
		case tk_fixed:
		case tk_boolean:
		case tk_char:
		case tk_wchar:
		case tk_octet:
		case tk_enum:
		case tk_struct:
		case tk_union:
		case tk_sequence:
		case tk_any:
		    if(IdlIsVariable(origMemberType))
			out << '\n' << norm << "* " << fixedName << ';';
		    else
			out << '\n' << norm << ' ' << fixedName << ';';
		    break;
		    
		case tk_TypeCode:
		case tk_Principal:
		case tk_objref:
		case tk_abstract_interface:
		    out << '\n' << norm << "_ptr " << fixedName << ';';
		    break;
		    
		case tk_string:
		    out << "\nchar* " << fixedName << ';';
		    break;
		    
		case tk_wstring:
		    out << "\nwchar_t* " << fixedName << ';';
		    break;
		    
		case tk_array:
		{
		    String_var s;
		    if(IdlIsVariable(origMemberType))
			s = getTypeString(realScope, info.type, GetTypeReturn,
					  false, fixedName);
		    else
			s = getTypeString(realScope, info.type, GetTypeNormal,
					  false, fixedName);
		    
		    out << '\n' << s << ';';
		    break;
		}
		
		case tk_value:
		case tk_value_box:
		    out << '\n' << norm << "* " << fixedName << ';';
		    break;
		    
		case tk_alias:
		    assert(false);
		    break;
		}
	    }
	    out.end();
	    out << " _ob_v_;";
	    
	    out.sep();
	    if(variable)
		out << "\nbool _ob_i_;";
	    out << '\n' << disc << " _ob_d_;";
	    
	    out.sep();
	    if(variable)
		out << "\nvoid _OB_remove();";
	    out << "\nbool _OB_check(" << disc << ") const;";
	    
	    out.sep();
	    out.dec();
	    out << "\npublic:";
	    out.inc();
	    
	    out.sep();
	    out << "\ntypedef " << name << "_var _var_type;";
	    
	    if(variable)
	    {
		out.sep();
		out.direct("\n#ifdef OB_CLEAR_MEM");
		out << '\n' << name << "();";
		out.direct("\n#else");
		out << '\n' << name << "() : _ob_i_(false) { }";
		out.direct("\n#endif"); // OB_CLEAR_MEM
		out << '\n' << name << "(const " << name << "&);";
		out << "\n~" << name << "() { _OB_remove(); }";
		
		out.sep();
		out << '\n' << name << "& operator=(const " << name << "&);";
	    }
	    
	    out.sep();
	    out << "\nvoid _d(" << disc << ");";
	    out << '\n' << disc << " _d() const;";
	    
	    for(j = 0 ; j < infoSeq -> length() ; j++)
	    {
		IdlUnionMemberInfo& info = infoSeq[j];
		const char* fixedName = fixedNames[j];
		
		if(is_nil(info.type))
		{
		    out.sep();
		    out << "\nvoid _default();";
		}
		else
		{
		    String_var norm;
		    if(info.type -> kind() == tk_sequence)
		    {
			String_var anonName = string_dup("_");
			anonName += info.name;
			anonName += "_seq";
			norm = anonName;
		    }
		    else
			norm = getTypeString(realScope, info.type,
					     GetTypeNormal);

		    TypeCode_var origMemberType =
			GetOrigType(info.type);
		    
		    switch(origMemberType -> kind())
		    {
		    case tk_null:
		    case tk_void:
		    case tk_except:
		    case tk_native:
			assert(false);
			break;
			
		    case tk_short:
		    case tk_long:
		    case tk_longlong:
		    case tk_ushort:
		    case tk_ulong:
		    case tk_ulonglong:
		    case tk_float:
		    case tk_double:
		    case tk_longdouble:
		    case tk_fixed:
		    case tk_boolean:
		    case tk_char:
		    case tk_wchar:
		    case tk_octet:
		    case tk_enum:
			out.sep();
			out << "\nvoid " << fixedName << '(' << norm << ");";
			out << '\n' << norm << ' ' << fixedName << "() const;";
			break;
			
		    case tk_struct:
		    case tk_union:
		    case tk_sequence:
		    case tk_any:
			out.sep();
			out << "\nvoid " << fixedName << "(const " << norm
			    << "&);";
			out << "\nconst " << norm << "& " << fixedName
			    << "() const;";
			out << '\n' << norm << "& " << fixedName << "();";
			break;
			
		    case tk_TypeCode:
		    case tk_Principal:
		    case tk_objref:
		    case tk_abstract_interface:
			out.sep();
			out << "\nvoid " << fixedName << '(' << norm
			    << "_ptr);";
			out << '\n' << norm << "_ptr " << fixedName
			    << "() const;";
			break;
			
		    case tk_string:
			out.sep();
			out << "\nvoid " << fixedName << "(char*);";
			out << "\nvoid " << fixedName << "(const char*);";
			out << "\nvoid " << fixedName
			    << "(const CORBA::String_var&);";
			out << "\nconst char* " << fixedName << "() const;";
			break;
			
		    case tk_wstring:
			out.sep();
			out << "\nvoid " << fixedName << "(wchar_t*);";
			out << "\nvoid " << fixedName << "(const wchar_t*);";
			out << "\nvoid " << fixedName
			    << "(const CORBA::WString_var&);";
			out << "\nconst wchar_t* " << fixedName << "() const;";
			break;
			
		    case tk_array:
		    {
			String_var in =
			    getTypeString(realScope, info.type, GetTypeNormal);
			
			String_var slice;
			
			if(info.type -> kind() == tk_alias)
			{
			    slice = getTypeString(realScope, info.type,
						  GetTypeSlice);
			}
			else
			{
			    const char* sl = "_slice";
			    slice = string_alloc(1 + strlen(info.name)
						       + strlen(sl));
			    strcpy(slice.inout(), "_");
			    strcat(slice.inout(), info.name);
			    strcat(slice.inout(), sl);
			    
			    String_var sliceDef =
				getTypeString(realScope, info.type,
					      GetTypeSlice, false, slice);
			    
			    out.sep();
			    out << "\ntypedef " << sliceDef << ';';
			}
			
			out.sep();
			out << "\nvoid " << fixedName << '(' << in << ");";
			out << '\n' << slice << "* " << fixedName
			    << "() const;";
			break;
		    }
		    
		    case tk_value:
		    case tk_value_box:
			out.sep();
			out << "\nvoid " << fixedName << '(' << norm << "*);";
			out << '\n' << norm << "* " << fixedName
			    << "() const;";
			break;
			
		    case tk_alias:
			assert(false);
			break;
		    }
		}
	    }
	    
            if(!local_)
            {
                //
                // CDR
                //
		out.sep();
                out << "\nstatic void _OB_marshal(const " << name
                    << "&, OB::OutputStream&);";
                out << "\nstatic void _OB_marshalCount(const " << name
                    << "&, OB::CountStream&);";
                out << "\nstatic void _OB_unmarshal(" << name
                    << "&, OB::InputStream&);";
            }
	    
	    out.end();
	    out << ';';
	    
	    if(!noTC_)
	    {
		out.sep();
		
		out << '\n';
		
		if(strlen(dllImport_) > 0)
		    out << dllImport_ << ' ';
		
		if(isTypeCodeExtern(container))
		    out << "extern OB::TypeCodeConst " << "_tc_"
			<< name << ';';
		else
		    out << "static OB::TypeCodeConst " << "_tc_"
			<< name << ';';
	    }
	    
	    break;
	}

	case dk_Attribute:
        {
            //
            // Attribute
            //

            DefinitionKind kind = container -> def_kind();

            bool pure;
            if(kind == dk_Interface)
            {
                InterfaceDef_var intf =
                    InterfaceDef::_narrow(container);
                assert(!is_nil(intf));
                pure = intf -> is_abstract() ? true : false;
            }
            else
                pure = (kind == dk_Value);
	    pure |= local_;

            genCPlusPlusHAttribute(seq[i], scope, false, pure, out);

            break;
        }

	case dk_Operation:
        {
            //
            // Operation
            //

            DefinitionKind kind = container -> def_kind();

            bool pure;
            if(kind == dk_Interface)
            {
                InterfaceDef_var intf =
                    InterfaceDef::_narrow(container);
                assert(!is_nil(intf));
                pure = intf -> is_abstract() ? true : false;
            }
            else
                pure = (kind == dk_Value);

	    pure |= local_;

            genCPlusPlusHOperation(seq[i], scope, false, pure, out);

            break;
        }

	case dk_Native:
        {
            //
            // Native
            //
            TypeDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            Identifier_var name = fixKwd(desc -> name);
            ScopedName_var absolute = getAbsolute(desc -> id);

            comment(desc -> id, out);

            out.sep();
            out << "\ntypedef OBNative_" << absolute << ' ' << name << ';';

            break;
        }

	default:
	    break;
	}
    }
}

// ----------------------------------------------------------------------
// Generate code for the header file, stub part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHStubImpl(Container_ptr container,
                                       IdlPrettyPrint& out)
{
    //
    // Get scope
    //
    String_var scope = getScope(container);

    //
    // Get container contents descriptions
    //
    Container::DescriptionSeq_var seq =
	container -> describe_contents(dk_all, true, -1);

    ULong i, j;
    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get container
	//
	Container_var cont =
	    Container::_narrow(seq[i].contained_object);

	switch(seq[i].kind)
	{

        case dk_Module:
        {
	    //
	    // Module
	    //
	    ModuleDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);
	    
	    // Don't add this here!!!
	    // if(!check(desc -> id))
	    //     continue;

            String_var name =
                getPrefixedName("OBStubImpl", seq[i].contained_object);

	    comment(desc -> id, out);
            out << "\nnamespace " << name;
            out << "\n{";

	    assert(!is_nil(cont));
	    genCPlusPlusHStubImpl(cont, out);

            out.sep();
            out << "\n} // End of namespace " << name;

            break;
        }

	case dk_Interface:
	{
	    //
	    // Interface
	    //
	    InterfaceDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

            String_var name =
                getPrefixedName("OBStubImpl", seq[i].contained_object);
            String_var stubScope = string_dup("OBStubImpl_");
            stubScope += scope;

            ScopedName_var absolute = getAbsolute(desc -> id);

	    comment(desc -> id, out);

            //XXX
            //out.sep();
            //out << "\nclass " << pre << name << ";";
            //out << "\ntypedef " << pre << name << "* "
            //   << pre << name << "_ptr;";
            
	    out.sep();
	    out << "\nclass " << name << " : ";
            unsigned int indent = out.getIndent();
	    out.setIndent(out.getPos());

            if(desc -> base_interfaces.length() > 0)
	    {
		for(j = 0 ; j < desc -> base_interfaces.length() ; j++)
		{
		    const char* id = desc -> base_interfaces[j];
                    //
                    // Using fully-qualified class name to avoid
                    // ambiguities
                    //
                    String_var baseAbs =
                        getPrefixedAbsolute("OBStubImpl", id, "");

                    if(j > 0)
                        out << ",\n";

		    if(!noVirtual_)
			out << "virtual ";

		    out << "public " << baseAbs;
		}
	    }
            else
            {
                if(!noVirtual_)
                    out << "virtual ";
                out << "public OB::StubImplBase";
            }
              
            out.setIndent(indent);

	    out.start();

	    out << '\n' << name << "(const " << name << "&);";
	    out << "\nvoid operator=(const " << name << "&);";

            out.dec();
            out.sep();
            out << "\npublic:";
            out.inc();

            out.sep();
            out << '\n' << name << "() { }";

            out.sep();
            out << "\nstatic inline " << name << "_ptr\n_duplicate("
                << name << "_ptr p)";
	    out.start();
	    out << "\nif(p)";
	    out.inc();
	    out << "\np -> _OB_incRef();";
	    out.dec();
	    out << "\nreturn p;";
	    out.end();
            out.sep();
            out << "\nstatic inline " << name << "_ptr\n_nil()";
	    out.start();
            out << "\nreturn 0;";
	    out.end();

	    assert(!is_nil(cont));
	    genCPlusPlusHStubImpl(cont, out);

	    out.end();
	    out << ';';

	    break;
	}

	case dk_Attribute:
	{
	    //
	    // Attribute
	    //
	    AttributeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);

	    comment(desc -> id, out);

	    String_var ret = getTypeString("", desc -> type, GetTypeReturn);

	    out.sep();
            out << "\nvirtual " << ret << ' ' << name << "();";
	    if(desc -> mode == ATTR_NORMAL)
	    {
		String_var in = getTypeString("", desc -> type, GetTypeIn);
                out << "\nvirtual void " << name << '(' << in << ");";
	    }

	    break;
	}

	case dk_Operation:
	{
	    //
	    // Operation
	    //
	    OperationDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);

	    comment(desc -> id, out);

	    String_var ret = getTypeString("", desc -> result, GetTypeReturn);

	    out.sep();
	    out << "\nvirtual " << ret << ' ' << name << "(";

	    unsigned int indent = out.getIndent();
	    out.setIndent(out.getPos());

	    for(j = 0 ; j < desc -> parameters.length() ; j++)
	    {
		GetType tt;
		switch(desc -> parameters[j].mode)
		{
		case PARAM_IN:
		    tt = GetTypeIn;
		    break;

		case PARAM_INOUT:
		    tt = GetTypeInOut;
		    break;

		case PARAM_OUT:
		    tt = GetTypeOut;
		    break;
		}

		String_var argName = fixKwd(desc -> parameters[j].name);

		String_var s =
		    getTypeString("", desc -> parameters[j].type, tt,
				  false, argName);

		if(j > 0)
		    out << ",\n";

		out << s;
	    }

	    if(desc -> contexts.length() > 0)
	    {
		if(desc -> parameters.length() > 0)
		    out << ",\n";

		out << "CORBA::Context_ptr";
	    }

	    out << ");";

	    out.setIndent(indent);

	    break;
	}

	default:
	    break;
	}
    }
}

// ----------------------------------------------------------------------
// Generate code for the header file, global namespace part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHGlobalNamespace(Container_ptr container,
					      IdlPrettyPrint& out)
{
    //
    // Get scope
    //
    String_var scope = getScope(container);

    //
    // Get container contents descriptions
    //
    Container::DescriptionSeq_var seq =
	container -> describe_contents(dk_all, true, -1);

    ULong i;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	switch(seq[i].kind)
	{
	case dk_Interface:
	{
	    //
	    // Interface
	    //
	    InterfaceDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);
	    ScopedName_var absolute = getAbsolute(desc -> id);

	    comment(desc -> id, out);

	    //
	    // release and is_nil
	    //
	    out.sep();
            out << "\nnamespace CORBA";
            out << "\n{";
            out.sep();
    	    out << "\ninline void\nrelease(" << absolute << "_ptr p)";
	    out.start();
            if(desc -> is_abstract)
            {
                out << "\nrelease((AbstractBase_ptr)p);";
            }
            else
            {
	        out << "\nif(p)";
	        out.inc();
	        out << "\np -> _OB_decRef();";
	        out.dec();
	    }
	    out.end();

	    out.sep();
    	    out << "\ninline Boolean\nis_nil(" << absolute
		<< "_ptr p)";
	    out.start();
	    out << "\nreturn p == 0;";
	    out.end();

            if(!local_)
            {
		out.sep();
		out << "\ninline void\nrelease(OBStubImpl_" << absolute
		    << "_ptr p)";
		out.start();
		out << "\nif(p)";
		out.inc();
		out << "\np -> _OB_decRef();";
		out.dec();
		out.end();

		out.sep();
		out << "\ninline Boolean\nis_nil(OBStubImpl_" << absolute
		    << "_ptr p)";
		out.start();
		out << "\nreturn p == 0;";
		out.end();
            }

	    out.sep();
            out << "\n} // End of namespace CORBA";

	    if(!noTC_)
	    {
		//
		// Any insertion and extraction operators.  No extraction
		// operators are necessary for abstract interfaces.
		//
		out.sep();
		out << "\nvoid operator<<=(CORBA::Any&, "
		    << absolute << "_ptr*);";
		out << "\nvoid operator<<=(CORBA::Any&, "
		    << absolute << "_ptr);";

                if(!desc -> is_abstract)
		    out << "\nCORBA::Boolean operator>>=(const CORBA::Any&, "
		        << absolute << "_ptr&);";

		out.sep();
		out << "\ninline void\noperator<<=(CORBA::Any_var& any, "
		    << absolute << "_ptr* val)";
		out.start();
		out << "\nany.inout() <<= val;";
		out.end();

		out.sep();
		out << "\ninline void\noperator<<=(CORBA::Any_var& any, "
		    << absolute << "_ptr val)";
		out.start();
		out << "\nany.inout() <<= val;";
		out.end();

                if(!desc -> is_abstract)
                {
		    out.sep();
		    out << "\ninline CORBA::Boolean";
		    out << "\noperator>>=(const CORBA::Any_var& any, "
		        << absolute << "_ptr& val)";
		    out.start();
		    out << "\nreturn any.in() >>= val;";
		    out.end();
                }
	    }

	    break;
	}

	case dk_Alias:
	{
	    //
	    // Alias
	    //
	    TypeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    TypeCode_var contentType = desc -> type -> content_type();

	    if(contentType -> kind() == tk_sequence)
	    {
		if(!local_ || !noTC_)
		{
		    comment(desc -> id, out);
		    
		    Identifier_var name = fixKwd(desc -> name);
		    String_var absolute = getAbsolute(desc -> id);
		    
		    if(!noTC_)
		    {
			//
			// Any insertion and extraction
			//
			out.sep();
			out << "\nvoid operator<<=(CORBA::Any&, " << absolute
			    << "*);";
			out << "\nvoid operator<<=(CORBA::Any&, const "
			    << absolute << "&);";
			out << "\nCORBA::Boolean operator>>=("
                            << "const CORBA::Any&, " << absolute << "*&);";
			
			out.sep();
			out << "\ninline void\noperator<<=(CORBA::Any_var& "
			    << "any, " << absolute << "* val)";
			out.start();
			out << "\nany.inout() <<= val;";
			out.end();
			
			out.sep();
			out << "\ninline void\noperator<<=("
                            << "CORBA::Any_var& any, "
			    << "const " << absolute << "& val)";
			out.start();
			out << "\nany.inout() <<= val;";
			out.end();
			
			out.sep();
			out << "\ninline CORBA::Boolean";
			out << "\noperator>>=(const CORBA::Any_var& any, "
			    << absolute << "*& val)";
			out.start();
			out << "\nreturn any.in() >>= val;";
			out.end();
		    }
		}
	    }
	    else if(contentType -> kind() == tk_array)
	    {
		if(!noTC_)
		{
		    comment(desc -> id, out);
		    
		    //
		    // Any insertion and extraction
		    //
		    String_var absolute = getAbsolute(desc -> id);
		    String_var absForAny = absolute;
		    absForAny += "_forany";
		    
		    out.sep();
		    out << "\nvoid operator<<=(CORBA::Any&, const "
                        << absForAny << "&);";
		    out << "\nCORBA::Boolean operator>>=(const CORBA::Any&, "
			<< absForAny << "&);";
		    
		    out.sep();
		    out << "\ninline void\noperator<<=(CORBA::Any_var& any, "
			<< "const " << absForAny << "& val)";
		    out.start();
		    out << "\nany.inout() <<= val;";
		    out.end();
		    
		    out.sep();
		    out << "\ninline CORBA::Boolean";
		    out << "\noperator>>=(const CORBA::Any_var& any, "
			<< absForAny << "& val)";
		    out.start();
		    out << "\nreturn any.in() >>= val;";
		    out.end();
		}
	    }

	    break;
	}

	case dk_Enum:
	{
	    //
	    // Enum
	    //
	    TypeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);
	    ScopedName_var absolute = getAbsolute(desc -> id);

	    comment(desc -> id, out);

	    if(!noTC_)
	    {
		//
		// Any insertion and extraction
		//
		out.sep();
		out << "\nvoid operator<<=(CORBA::Any&, " << absolute << ");";
		out << "\nCORBA::Boolean operator>>=(const CORBA::Any&, "
		    << absolute << "&);";
	    }

	    break;
	}

	case dk_Struct:
	case dk_Exception:
	{
	    RepositoryId_var id;
	    Identifier_var name;
	    TypeCode_var type;

	    if(seq[i].kind == dk_Struct)
	    {
		//
		// Struct
		//
		TypeDescription* desc;
		Boolean b = seq[i].value >>= desc;
		assert(b);

		name = fixKwd(desc -> name);
		id = desc -> id;
		type = desc -> type;
	    }
	    else
	    {
		//
		// Exception
		//
		ExceptionDescription* desc;
		Boolean b = seq[i].value >>= desc;
		assert(b);

		name = fixKwd(desc -> name);
		id = desc -> id;
		type = desc -> type;
	    }

	    if(!check(id))
		continue;

	    ScopedName_var absolute = getAbsolute(id);

	    comment(id, out);

	    if(!noTC_)
	    {
		//
		// Any insertion and extraction
		//
		out.sep();
		out << "\nvoid operator<<=(CORBA::Any&, " << absolute << "*);";
		out << "\nvoid operator<<=(CORBA::Any&, const "
		    << absolute << "&);";
		out << "\nCORBA::Boolean operator>>=(const CORBA::Any&, "
		    << absolute << "*&);";

		out.sep();
		out << "\ninline void\noperator<<=(CORBA::Any_var& any, "
		    << absolute << "* val)";
		out.start();
		out << "\nany.inout() <<= val;";
		out.end();

		out.sep();
		out << "\ninline void\noperator<<=(CORBA::Any_var& any, const "
		    << absolute << "& val)";
		out.start();
		out << "\nany.inout() <<= val;";
		out.end();

		out.sep();
		out << "\ninline CORBA::Boolean";
		out << "\noperator>>=(const CORBA::Any_var& any, "
		    << absolute << "*& val)";
		out.start();
		out << "\nreturn any.in() >>= val;";
		out.end();
	    }

	    break;
	}

	case dk_Union:
	{
	    //
	    // Union
	    //
	    TypeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);
	    ScopedName_var absolute = getAbsolute(desc -> id);

	    IdlUnionMemberInfoSeq_var infoSeq = IdlGetUnionInfo(desc -> type);

	    comment(desc -> id, out);

	    if(!noTC_)
	    {
		//
		// Any insertion and extraction
		//
		out.sep();
		out << "\nvoid operator<<=(CORBA::Any&, " << absolute << "*);";
		out << "\nvoid operator<<=(CORBA::Any&, const " << absolute
		    << "&);";
		out << "\nCORBA::Boolean operator>>=(const CORBA::Any&, "
		    << absolute << "*&);";

		out.sep();
		out << "\ninline void\noperator<<=(CORBA::Any_var& any, "
		    << absolute << "* val)";
		out.start();
		out << "\nany.inout() <<= val;";
		out.end();

		out.sep();
		out << "\ninline void\noperator<<=(CORBA::Any_var& any, const "
		    << absolute << "& val)";
		out.start();
		out << "\nany.inout() <<= val;";
		out.end();

		out.sep();
		out << "\ninline CORBA::Boolean";
		out << "\noperator>>=(const CORBA::Any_var& any, "
		    << absolute << "*& val)";
		out.start();
		out << "\nreturn any.in() >>= val;";
		out.end();
	    }
	    
	    break;
	}

        case dk_Value:
        case dk_ValueBox:
        {
            RepositoryId_var id;
            Identifier_var name;

            if(seq[i].kind == dk_Value)
            {
                //
                // Valuetype
                //
                ValueDescription* desc;
                Boolean b = seq[i].value >>= desc;
                assert(b);

                name = fixKwd(desc -> name);
                id = desc -> id;
            }
            else
            {
                //
                // Value box
                //
                TypeDescription* desc;
                Boolean b = seq[i].value >>= desc;
                assert(b);

                name = fixKwd(desc -> name);
                id = desc -> id;
            }

            if(!check(id))
                continue;

            ScopedName_var absolute = getAbsolute(id);

            comment(id, out);

            if(!noTC_)
            {
                //
                // Any insertion and extraction
                //
                out.sep();
                out << "\nvoid operator<<=(CORBA::Any&, "
                    << absolute << "**);";
                out << "\nvoid operator<<=(CORBA::Any&, "
                    << absolute << "*);";
                out << "\nCORBA::Boolean operator>>=(const CORBA::Any&, "
                    << absolute << "*&);";

                out.sep();
                out << "\ninline void\noperator<<=(CORBA::Any_var& any, "
                    << absolute << "** val)";
                out.start();
                out << "\nany.inout() <<= val;";
                out.end();

                out.sep();
                out << "\ninline void\noperator<<=(CORBA::Any_var& any, "
                    << absolute << "* val)";
                out.start();
                out << "\nany.inout() <<= val;";
                out.end();

                out.sep();
                out << "\ninline CORBA::Boolean";
                out << "\noperator>>=(const CORBA::Any_var& any, "
                    << absolute << "*& val)";
                out.start();
                out << "\nreturn any.in() >>= val;";
                out.end();
            }

            break;
        }

	default:
	    break;
	}

	//
	// Check for recursion
	//
	Container_var cont =
	    Container::_narrow(seq[i].contained_object);
	if(!is_nil(cont))
	    genCPlusPlusHGlobalNamespace(cont, out);
    }
}

// ----------------------------------------------------------------------
// Generate OBV_* classes for the header file, general part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHOBV(Container_ptr container, IdlPrettyPrint& out)
{
    //
    // Get scope
    //
    String_var scope = getScope(container);

    //
    // Get container contents descriptions
    //
    Container::DescriptionSeq_var seq =
        container -> describe_contents(dk_all, true, -1);

    ULong i, j;

    for(i = 0 ; i < seq -> length() ; i++)
    {
        //
        // Get container
        //
        Container_var cont =
            Container::_narrow(seq[i].contained_object);

        switch(seq[i].kind)
        {
        case dk_Module:
        {
	    ModuleDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);
	    
	    // Don't add this here!!!
	    // if(!check(desc -> id))
	    //     continue;

            String_var name =
                getPrefixedName("OBV", seq[i].contained_object);

	    comment(desc -> id, out);
            out << "\nnamespace " << name;
            out << "\n{";

	    assert(!is_nil(cont));
	    genCPlusPlusHOBV(cont, out);

            out.sep();
            out << "\n} // End of namespace " << name;

            break;
        }

        case dk_Value:
        {
            //
            // Valuetype
            //
            ValueDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            //
            // Don't generate OBV class for abstract valuetype
            //
            if(desc -> is_abstract)
                continue;

            String_var name = getPrefixedName("OBV", seq[i].contained_object);
            ScopedName_var absolute = getAbsolute(desc -> id);

            comment(desc -> id, out);

            out.sep();
            out << "\nclass " << name << " : ";

            unsigned int indent = out.getIndent();
            out.setIndent(out.getPos());

            out << "virtual public " << absolute;

            //
            // Also inherit from OBV_ class of concrete base value
            //
            if(strlen(desc -> base_value) > 0)
            {
                out << ",\n";

                //
                // Using fully-qualified class name to avoid
                // ambiguities
                //
                String_var baseAbs = getPrefixedAbsolute("OBV",
                    desc -> base_value, "");

                out << "virtual public " << baseAbs;
            }

            out.setIndent(indent);

            out.start();

            out << "\nvoid operator=(const " << name << "&);";

            //
            // Get all value members
            //
            Container::DescriptionSeq_var members =
                cont -> describe_contents(dk_ValueMember, true, -1);

            if(members -> length() > 0)
                out.sep();

            //
            // Create class member for each value member
            //
            for(j = 0 ; j < members -> length() ; j++)
            {
                ValueMember* member;
                b = members[j].value >>= member;
                assert(b);

                Identifier_var memberName = string_dup("_ob_");
                memberName += member -> name;
                memberName += "_";

                if(member -> type -> kind() == tk_sequence)
                {
                    //
                    // Compose anonymous sequence type name
                    //
                    String_var seqName = string_dup("_");
                    seqName += member -> name;
                    seqName += "_seq";
                    out << '\n' << seqName << ' ' << memberName << ';';
                }
                else
                {
                    String_var str = getTypeString("", member -> type,
                                                         GetTypeStruct,
                                                         false, memberName);
                    out << '\n' << str << ';';
                }
            }

            out.sep();
            out.dec();
            out << "\nprotected:";
            out.inc();
            out.sep();
            out << '\n' << name << "();";

            //
            // Constructor with initializer for each member
            //
            if(members -> length() > 0)
            {
                out << '\n' << name << "(";
                indent = out.getIndent();
                out.setIndent(out.getPos());
                for(j = 0 ; j < members -> length() ; j++)
                {
                    ValueMember* member;
                    b = members[j].value >>= member;
                    assert(b);

                    TypeCode_var origMemberType =
                        GetOrigType(member -> type);

                    //
                    // Determine type name for member - check for
                    // anonymous types
                    //
                    String_var typeStr;
                    if(member -> type -> kind() == tk_sequence)
                    {
                        //
                        // Anonymous sequence
                        //
                        typeStr = string_dup("const _");
                        typeStr += member -> name;
                        typeStr += "_seq&";
                    }
                    else if(origMemberType -> kind() == tk_array)
                    {
                        typeStr = getTypeString("", member -> type,
                                                GetTypeNormal);
                    }
                    else
                    {
                        typeStr = getTypeString("", member -> type, GetTypeIn);
                    }

                    if(j > 0)
                        out << ",\n";

                    out << typeStr;
                }
                out << ");";
                out.setIndent(indent);
            }

            out << "\nvirtual ~" << name << "();";

            //
            // Private state member functions
            //
            bool havePublic = false;
            for(j = 0 ; j < members -> length() ; j++)
            {
                ValueMember* member;
                b = members[j].value >>= member;
                assert(b);

                if(member -> access == PRIVATE_MEMBER)
                    genCPlusPlusHValueMember(cont, member, "", true, out);
                else
                    havePublic = true;
            }

            if(havePublic)
            {
                out.sep();
                out.dec();
                out << "\npublic:";
                out.inc();

                //
                // Public state member functions
                //
                for(j = 0 ; j < members -> length() ; j++)
                {
                    ValueMember* member;
                    b = members[j].value >>= member;
                    assert(b);

                    if(member -> access == PUBLIC_MEMBER)
                        genCPlusPlusHValueMember(cont, member, "", true, out);
                }
            }

            out.end();
            out << ';';

            break;
        }

        default:
            break;
        }
    }
}

// ----------------------------------------------------------------------
// Generate code for a value member
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHValueMember(Container_ptr value,
                                          ValueMember* desc,
                                          const char* scope,
                                          bool skel,
                                          IdlPrettyPrint& out)
{
    ValueDef_var valueDef = ValueDef::_narrow(value);
    assert(! is_nil(valueDef));
    Container_var container = valueDef -> defined_in();

    String_var pure;
    if(skel)
        pure = string_dup("");
    else
        pure = string_dup(" = 0");

    if(!check(desc -> id))
        return;

    Identifier_var name = fixKwd(desc -> name);

    comment(desc -> id, out);

    String_var norm, in, inOut, ret, var;
    if(desc -> type -> kind() == tk_sequence)
    {
        //
        // Anonymous sequence
        //
        norm = string_dup("_");
        norm += desc -> name;
        norm += "_seq";
        in = string_dup("const ");
        in += norm;
        in += "&";
        inOut = norm;
        inOut += "&";
    }
    else
    {
        norm = getTypeString(scope, desc -> type, GetTypeNormal);
        in = getTypeString(scope, desc -> type, GetTypeIn);
        inOut = getTypeString(scope, desc -> type, GetTypeInOut);
        ret = getTypeString(scope, desc -> type, GetTypeReturn);
        var = getTypeString(scope, desc -> type, GetTypeVar);
    }

    TypeCode_var origMemberType = GetOrigType(desc -> type);

    switch(origMemberType -> kind())
    {
    case tk_null:
    case tk_void:
    case tk_except:
    case tk_native:
        assert(false);
        break;

    case tk_short:
    case tk_long:
    case tk_longlong:
    case tk_ushort:
    case tk_ulong:
    case tk_ulonglong:
    case tk_float:
    case tk_double:
    case tk_longdouble:
    case tk_fixed:
    case tk_boolean:
    case tk_char:
    case tk_wchar:
    case tk_octet:
    case tk_enum:
    case tk_value:
    case tk_value_box:
        out.sep();
        out << "\nvirtual void " << name << '(' << in << ")" << pure << ';';
        out << "\nvirtual " << ret << ' ' << name << "() const" << pure << ';';
        break;

    case tk_struct:
    case tk_union:
    case tk_sequence:
    case tk_any:
        out.sep();
        out << "\nvirtual void " << name << '(' << in << ')' << pure << ';';
        out << "\nvirtual " << in << ' ' << name << "() const" << pure << ';';
        out << "\nvirtual " << inOut << ' ' << name << "()" << pure << ';';
        break;

    case tk_TypeCode:
    case tk_Principal:
    case tk_objref:
    case tk_abstract_interface:
        out.sep();
        out << "\nvirtual void " << name << '(' << in << ')' << pure << ';';
        out << "\nvirtual " << ret << ' ' << name << "() const" << pure << ';';
        break;

    case tk_string:
    case tk_wstring:
        out.sep();
        out << "\nvirtual void " << name << '(' << norm << ')' << pure << ';';
        out << "\nvirtual void " << name << "(const " << norm << ')'
            << pure << ';';
        out << "\nvirtual void " << name << "(const " << var << "&)"
            << pure << ';';
        out << "\nvirtual const " << norm << ' ' << name << "() const"
            << pure << ';';
        break;

    case tk_array:
    {
        String_var slice;

        if(desc -> type -> kind() == tk_alias)
        {
            slice = getTypeString(scope, desc -> type, GetTypeSlice);
        }
        else
        {
            slice = string_dup("_");
            slice += desc -> name;
            slice += "_slice";

            if(! skel)
            {
                String_var sliceDef =
                    getTypeString(scope, desc -> type, GetTypeSlice, false,
                                  slice);
                out.sep();
                out << "\ntypedef " << sliceDef << ';';
            }
        }

        out.sep();
        out << "\nvirtual void " << name << '(' << norm << ")" << pure << ';';
        out << "\nvirtual " << slice << "* " << name << "() const"
            << pure << ';';
        break;
    }

    case tk_alias:
        assert(false);
        break;
    }
}

// ----------------------------------------------------------------------
// Generate type-specific value factory class
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHValueFactory(ValueDef_ptr value,
                                           IdlPrettyPrint& out)
{
    ValueDef::FullValueDescription_var desc = value -> describe_value();

    if(desc -> initializers.length() == 0)
        return;

    //
    // Don't generate factory class for abstract interface
    //
    if(desc -> is_abstract)
        return;

    //
    // Get scope of container
    //
    Container_var container = value -> defined_in();
    String_var scope;
    if(!is_nil(container))
        scope = getScope(container);
    else
        scope = string_dup("");

    Identifier_var name = fixKwd(desc -> name);

    //
    // Compose the name of the factory's class
    //
    String_var className = string_dup(desc -> name);
    className += "_init";

    out.sep();
    out << "\nclass " << className << " : ";

    if(!noVirtual_)
        out << "virtual ";

    out << "public CORBA::ValueFactoryBase";

    out.start();

    out.dec();
    out << "\nprotected:";
    out.inc();
    out << '\n' << className << "()";
    out.start();
    out.end();

    out.sep();
    out.dec();
    out << "\npublic:";
    out.inc();

    out << "\nvirtual ~" << className << "()";
    out.start();
    out.end();

    out.sep();
    out << "\nstatic " << className << "* _downcast(CORBA::ValueFactory);";

    //
    // Create function for each initializer
    //
    for(ULong i = 0 ; i < desc -> initializers.length() ; i++)
    {
        StructMemberSeq members = desc -> initializers[i].members;
        out.sep();
        out << "\nvirtual " << name << "* create(";

        unsigned int indent = out.getIndent();
        out.setIndent(out.getPos());

        for(ULong j = 0 ; j < members.length() ; j++)
        {
            TypeCode_var origMemberType = GetOrigType(members[j].type);
            String_var argName = fixKwd(members[j].name);
            String_var type;

            if(j > 0)
                out << ",\n";

            if(origMemberType -> kind() == tk_array)
            {
                type = getTypeString("", members[j].type, GetTypeNormal);
                out << type << ' ' << argName;
            }
            //
            // Fix ambiguous types
            //
            else if(origMemberType -> kind() == tk_boolean)
            {
                out << "CORBA::ValueFactoryBase::BooleanValue " << argName;
            }
            else if(origMemberType -> kind() == tk_char)
            {
                out << "CORBA::ValueFactoryBase::CharValue " << argName;
            }
            else if(origMemberType -> kind() == tk_wchar)
            {
                out << "CORBA::ValueFactoryBase::WCharValue " << argName;
            }
            else if(origMemberType -> kind() == tk_octet)
            {
                out << "CORBA::ValueFactoryBase::OctetValue " << argName;
            }
            else
            {
                type = getTypeString(scope, members[j].type, GetTypeIn);
                out << type << ' ' << argName;
            }
        }

        out << ") = 0;";
        out.setIndent(indent);
    }

    out.sep();
    out << "\nvirtual const char* _OB_typeId() const;";

    out.end();
    out << ';';
}

// ----------------------------------------------------------------------
// Generate operation
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHOperation(Container::Description& d,
                                        const char* scope, bool ex,
                                        bool pure, IdlPrettyPrint& out)
{
    ULong i;

    OperationDescription* desc;
    Boolean b = d.value >>= desc;
    assert(b);

    if(!check(desc -> id))
        return;

    Identifier_var name = fixKwd(desc -> name);

    comment(desc -> id, out);

    String_var ret = getTypeString(scope, desc -> result, GetTypeReturn);

    out.sep();
    out << "\nvirtual " << ret << ' ' << name << "(";

    unsigned int indent = out.getIndent();
    out.setIndent(out.getPos());

    for(i = 0 ; i < desc -> parameters.length() ; i++)
    {
        GetType tt;
        switch(desc -> parameters[i].mode)
        {
        case PARAM_IN:
            tt = GetTypeIn;
            break;

        case PARAM_INOUT:
            tt = GetTypeInOut;
            break;

        case PARAM_OUT:
            tt = GetTypeOut;
            break;
        }

        String_var argName =
            fixKwd(desc -> parameters[i].name);

        String_var s = getTypeString(scope, desc -> parameters[i].type,
                                           tt, false, argName);

        if(i > 0)
            out << ",\n";

        out << s;
    }

    if(desc -> contexts.length() > 0)
    {
        if(desc -> parameters.length() > 0)
            out << ",\n";

        out << "CORBA::Context_ptr";
    }

    out << ')';

    out.setIndent(indent);

    //
    // Exception specifications
    //
    if(ex)
    {
        String_var sysEx = string_dup("CORBA::SystemException");

        out.inc();
        out << "\nthrow(";

        indent = out.getIndent();
        out.setIndent(out.getPos());

        for(i = 0 ; i < desc -> exceptions.length() ; i++)
        {
            //
            // Using fully-qualified class name to avoid ambiguities
            //
            Identifier_var exName =
                getAbsolute(desc -> exceptions[i].id);
            out << exName << ",\n";
        }

        out << sysEx << ')';

        out.setIndent(indent);
        out.dec();
    }

    if(pure)
        out << " = 0";

    out << ';';
}

// ----------------------------------------------------------------------
// Generate attribute
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHAttribute(Container::Description& d,
                                        const char* scope, bool ex,
                                        bool pure, IdlPrettyPrint& out)
{
    AttributeDescription* desc;
    Boolean b = d.value >>= desc;
    assert(b);

    if(!check(desc -> id))
        return;

    Identifier_var name = fixKwd(desc -> name);
    String_var sysEx = string_dup("CORBA::SystemException");

    comment(desc -> id, out);

    String_var ret = getTypeString(scope, desc -> type, GetTypeReturn);

    out.sep();
    out << "\nvirtual " << ret << ' ' << name << "()";
    if(ex)
    {
        out.inc();
        out << "\nthrow(" << sysEx << ')';
        out.dec();
    }
    if(pure)
        out << " = 0";
    out << ';';

    if(desc -> mode == ATTR_NORMAL)
    {
        String_var in = getTypeString(scope, desc -> type, GetTypeIn);
        out << "\nvirtual void " << name << '(' << in << ')';
        if(ex)
        {
            out.inc();
            out << "\nthrow(" << sysEx << ')';
            out.dec();
        }
        if(pure)
            out << " = 0";
        out << ';';
    }
}

// ----------------------------------------------------------------------
// Generate basic value box
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHBasicBox(ValueBoxDef_ptr /*value*/,
                                       const char* scope,
                                       const char* absolute,
                                       TypeDescription* desc,
                                       TypeCode_ptr type,
                                       IdlPrettyPrint& out)
{
    TypeCode_var origType = GetOrigType(type);
    String_var name = fixKwd(desc -> name);

    bool isObjRef = (origType -> kind() == tk_objref);

    String_var normType;
    if(isObjRef)
        normType = getTypeString(scope, type, GetTypePtr);
    else
        normType = getTypeString(scope, type, GetTypeNormal);

    String_var inType = getTypeString(scope, type, GetTypeIn);
    String_var inOutType = getTypeString(scope, type, GetTypeInOut);
    String_var varType = getTypeString(scope, type, GetTypeVar);
    String_var retType = getTypeString(scope, type, GetTypeReturn);

    if(isObjRef)
        out << '\n' << varType << " _ob_v_;";
    else
        out << '\n' << normType << " _ob_v_;";

    out.sep();
    out << '\n' << name << "& operator=(const " << name << "&);";

    out.sep();
    out.dec();
    out << "\nprotected:";
    out.inc();

    out << "\nvirtual ~" << name << "() { }";

    out.sep();
    out.dec();
    out << "\npublic:";
    out.inc();

    //
    // Constructors
    //

    out << '\n' << name << "()";
    out.start();
    if(! isObjRef)
    {
        out.direct("\n#ifdef OB_CLEAR_MEM");
        out << "\nmemset(&_ob_v_, 0, sizeof(_ob_v_));";
        out.direct("\n#endif"); // OB_CLEAR_MEM
    }
    out.end();
    out.sep();

    out << '\n' << name << "(const " << name << "& _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a._ob_v_) { }";
    out.dec();
    out.sep();

    out << '\n' << name << '(' << inType << " _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a) { }";
    out.dec();

    out.sep();

    //
    // Assignment operators
    //
    out << '\n' << name << "& operator=(" << inType << " _ob_a)";
    out.start();
    out << "\n_ob_v_ = _ob_a;";
    out << "\nreturn *this;";
    out.end();

    out.sep();

    //
    // Accessor
    //
    out << '\n' << retType << " _value() const";
    out.start();
    if(isObjRef)
        out << "\nreturn _ob_v_.in();";
    else
        out << "\nreturn _ob_v_;";
    out.end();

    out.sep();

    //
    // Modifiers
    //
    out << "\nvoid _value(" << inType << " _ob_a) { _ob_v_ = _ob_a; }";

    out.sep();

    //
    // Argument-passing conversions
    //
    if(isObjRef)
    {
        out << '\n' << inType << " _boxed_in() const "
            << "{ return _ob_v_.in(); }";
        out << '\n' << inOutType << " _boxed_inout() "
            << "{ return _ob_v_.inout(); }";
        out << '\n' << inOutType << " _boxed_out() "
            << "{ return _ob_v_.out(); }";
    }
    else
    {
        out << '\n' << inType << " _boxed_in() const { return _ob_v_; }";
        out << '\n' << inOutType << " _boxed_inout() { return _ob_v_; }";
        out << '\n' << inOutType << " _boxed_out() { return _ob_v_; }";
    }
}

// ----------------------------------------------------------------------
// Generate struct value box
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHStructBox(ValueBoxDef_ptr value,
                                        const char* scope,
                                        const char* absolute,
                                        TypeDescription* desc,
                                        TypeCode_ptr type,
                                        IdlPrettyPrint& out)
{
    TypeCode_var origType = GetOrigType(type);
    String_var name = fixKwd(desc -> name);

    String_var normType = getTypeString(scope, type, GetTypeNormal);
    String_var inType = getTypeString(scope, type, GetTypeIn);
    String_var inOutType = getTypeString(scope, type, GetTypeInOut);
    String_var outType = getTypeString(scope, type, GetTypeOut);
    String_var varType = getTypeString(scope, type, GetTypeVar);
    String_var retType = getTypeString(scope, type, GetTypeReturn);

    out << '\n' << varType << " _ob_v_;";

    out.sep();
    out << '\n' << name << "& operator=(const " << name << "&);";

    out.sep();
    out.dec();
    out << "\nprotected:";
    out.inc();

    out << "\nvirtual ~" << name << "() { }";

    out.sep();
    out.dec();
    out << "\npublic:";
    out.inc();

    //
    // Constructors
    //

    out << '\n' << name << "()";
    out.inc();
    out << "\n: _ob_v_(new " << normType << ") { }";
    out.dec();
    out.sep();

    out << '\n' << name << "(const " << name << "& _ob_a)";
    out.inc();
    out << "\n: _ob_v_(new " << normType << "(_ob_a._ob_v_.in())) { }";
    out.dec();
    out.sep();

    out << '\n' << name << '(' << inType << " _ob_a)";
    out.inc();
    out << "\n: _ob_v_(new " << normType << "(_ob_a)) { }";
    out.dec();

    out.sep();

    //
    // Assignment operator
    //
    out << '\n' << name << "& operator=(" << inType << " _ob_a)";
    out.start();
    out << "\n_ob_v_ = new " << normType << "(_ob_a);";
    out << "\nreturn *this;";
    out.end();

    out.sep();

    //
    // Accessors
    //
    out << '\n' << inType << " _value() const";
    out.start();
    out << "\nreturn _ob_v_.in();";
    out.end();
    out.sep();
    out << '\n' << inOutType << " _value()";
    out.start();
    out << "\nreturn _ob_v_.inout();";
    out.end();

    out.sep();

    //
    // Modifier
    //
    out << "\nvoid _value(" << inType << " _ob_a) "
        << "{ _ob_v_ = new " << normType << "(_ob_a); }";

    out.sep();

    //
    // Argument-passing conversions
    //
    out << '\n' << inType << " _boxed_in() const "
        << "{ return _ob_v_.in(); }";
    out << '\n' << inOutType << " _boxed_inout() "
        << "{ return _ob_v_.inout(); }";
    out << '\n' << retType << "& _boxed_out() "
        << "{ return _ob_v_.out(); }";

    //
    // Struct member functions
    //
    IDLType_var idlType = value -> original_type_def();
    StructDef_var structDef = StructDef::_narrow(idlType);
    assert(! is_nil(structDef));
    StructMemberSeq_var members = structDef -> members();
    RepositoryId_var structId = structDef -> id();
    for(ULong i = 0 ; i < members -> length() ; i++)
    {
        genCPlusPlusHStructBoxMember(structId, members[i], scope, out);
    }
}

// ----------------------------------------------------------------------
// Generate code for a struct value box member
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHStructBoxMember(const char* structId,
                                              const StructMember& desc,
                                              const char* scope,
                                              IdlPrettyPrint& out)
{
    Identifier_var name = fixKwd(desc.name);

    String_var norm = getTypeString(scope, desc.type, GetTypeNormal);

    TypeCode_var origMemberType = GetOrigType(desc.type);

    switch(origMemberType -> kind())
    {
    case tk_null:
    case tk_void:
    case tk_except:
    case tk_native:
        assert(false);
        break;

    case tk_short:
    case tk_long:
    case tk_longlong:
    case tk_ushort:
    case tk_ulong:
    case tk_ulonglong:
    case tk_float:
    case tk_double:
    case tk_longdouble:
    case tk_fixed:
    case tk_boolean:
    case tk_char:
    case tk_wchar:
    case tk_octet:
    case tk_enum:
        out.sep();
        out << "\nvirtual void " << name << '(' << norm << " _ob_a) "
            << "{ _ob_v_ -> " << name << " = _ob_a; }";
        out << "\nvirtual " << norm << ' ' << name << "() const "
            << "{ return _value()." << name << "; }";
        break;

    case tk_value:
    case tk_value_box:
        out.sep();
        out << "\nvirtual void " << name << '(' << norm << "* _ob_a) "
            << "{ _ob_v_ -> " << name << " = _ob_a; }";
        out << "\nvirtual " << norm << "* " << name << "() const "
            << "{ return _value()." << name << "; }";
        break;

    case tk_struct:
    case tk_union:
    case tk_sequence:
    case tk_any:
        //
        // Check for anonymous sequence
        //
        if(desc.type -> kind() == tk_sequence)
        {
            norm = getAbsolute(structId, scope);
            norm += "::_";
            norm += desc.name;
            norm += "_seq";
        }

        out.sep();
        out << "\nvirtual void " << name << "(const " << norm << "& _ob_a) "
            << "{ _ob_v_ -> " << name << " = _ob_a; }";
        out << "\nvirtual const " << norm << "& " << name << "() const "
            << "{ return _value()." << name << "; }";
        out << "\nvirtual " << norm << "& " << name << "() "
            << "{ return _value()." << name << "; }";
        break;

    case tk_TypeCode:
    case tk_Principal:
    case tk_objref:
    case tk_abstract_interface:
        out.sep();
        out << "\nvirtual void " << name << '(' << norm << "_ptr _ob_a) "
            << "{ _ob_v_ -> " << name << " = _ob_a; }";
        out << "\nvirtual " << norm << "_ptr " << name << "() const "
            << "{ return _value()." << name << "; }";
        break;

    case tk_string:
        out.sep();
        out << "\nvirtual void " << name << "(char* _ob_a) "
            << "{ _ob_v_ -> " << name << " = _ob_a; }";
        out << "\nvirtual void " << name << "(const char* _ob_a) "
            << "{ _ob_v_ -> " << name << " = _ob_a; }";
        out << "\nvirtual void " << name << "(const CORBA::String_var& _ob_a) "
            << "{ _ob_v_ -> " << name << " = _ob_a; }";
        out << "\nvirtual const char* " << name << "() const "
            << "{ return _value()." << name << "; }";
        break;

    case tk_wstring:
        out.sep();
        out << "\nvirtual void " << name << "(wchar_t* _ob_a) "
            << "{ _ob_v_ -> " << name << " = _ob_a; }";
        out << "\nvirtual void " << name << "(const wchar_t* _ob_a) "
            << "{ _ob_v_ -> " << name << " = _ob_a; }";
        out << "\nvirtual void " << name
            << "(const CORBA::WString_var& _ob_a) " << "{ _ob_v_ -> "
            << name << " = _ob_a; }";
        out << "\nvirtual const wchar_t* " << name << "() const "
            << "{ return _value()." << name << "; }";
        break;

    case tk_array:
    {
        String_var in =
            getTypeString(scope, desc.type, GetTypeNormal, false, "_ob_a");
        ScopedName_var arrayAbsolute =
            getTypeString(scope, desc.type, GetTypeNormal, false);

        String_var slice;

        bool isAnon = !(desc.type -> kind() == tk_alias);

        if(!isAnon)
        {
            slice = getTypeString(scope, desc.type, GetTypeSlice);
        }
        else
        {
            slice = string_dup("_");
            slice += desc.name;
            slice += "_slice";

            //
            // Emit a typedef for the slice, since the struct doesn't
            // generate one.
            //
            String_var sliceDef =
                getTypeString("", desc.type, GetTypeSlice, true, slice);
            out.sep();
            out << "\ntypedef " << sliceDef << ';';
        }

        out.sep();
        out << "\nvirtual void " << name << '(' << in << ")";
        out.start();
        String_var target = string_dup("_ob_v_ -> ");
        target += name;
        writeArrayCopy(scope, desc.type, target, "_ob_a", "_ob_i", out);
        out.end();

        out.sep();
        out << "\nvirtual " << slice << "* " << name << "() const";
        out.start();
        out << "\nreturn (" << slice << "*)(_value()." << name << ");";
        out.end();
        break;
    }

    case tk_alias:
        assert(false);
        break;
    }
}

// ----------------------------------------------------------------------
// Generate string/wstring value box
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHStringBox(ValueBoxDef_ptr /*value*/,
                                        const char* scope,
                                        const char* absolute,
                                        TypeDescription* desc,
                                        TypeCode_ptr type,
                                        IdlPrettyPrint& out)
{
    TypeCode_var origType = GetOrigType(type);
    String_var name = fixKwd(desc -> name);
    const char* charType;
    if(origType -> kind() == tk_string)
        charType = "char";
    else
        charType = "wchar_t";

    String_var normType = getTypeString(scope, type, GetTypeNormal);
    String_var inType = getTypeString(scope, type, GetTypeIn);
    String_var inOutType = getTypeString(scope, type, GetTypeInOut);
    String_var varType = getTypeString(scope, type, GetTypeVar);

    out << '\n' << varType << " _ob_v_;";
    out.sep();
    out << '\n' << name << "& operator=(const " << name << "&);";

    out.sep();
    out.dec();
    out << "\nprotected:";
    out.inc();

    out << "\nvirtual ~" << name << "() { }";

    out.sep();
    out.dec();
    out << "\npublic:";
    out.inc();

    //
    // Constructors
    //
    out << '\n' << name << "()";
    out.inc();
    if(origType -> kind() == tk_string)
        out << "\n: _ob_v_((const char*)\"\") { }";
    else
        out << "\n: _ob_v_((const wchar_t*)L\"\") { }";
    out.dec();
    out.sep();
    out << '\n' << name << "(const " << name << "& _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a._ob_v_) { }";
    out.dec();
    out.sep();
    out << '\n' << name << '(' << normType << " _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a) { }";
    out.dec();
    out.sep();
    out << '\n' << name << '(' << inType << " _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a) { }";
    out.dec();
    out.sep();
    out << '\n' << name << "(const " << varType << "& _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a) { }";
    out.dec();

    out.sep();

    //
    // Assignment operators
    //
    out << '\n' << name << "& operator=(" << normType << " _ob_a)";
    out.start();
    out << "\n_ob_v_ = _ob_a;";
    out << "\nreturn *this;";
    out.end();
    out.sep();
    out << '\n' << name << "& operator=(" << inType << " _ob_a)";
    out.start();
    out << "\n_ob_v_ = _ob_a;";
    out << "\nreturn *this;";
    out.end();
    out.sep();
    out << '\n' << name << "& operator=(const " << varType << "& _ob_a)";
    out.start();
    out << "\n_ob_v_ = _ob_a;";
    out << "\nreturn *this;";
    out.end();

    out.sep();

    //
    // Accessor
    //
    out << '\n' << inType << " _value() const { return _ob_v_.in(); }";

    out.sep();

    //
    // Modifiers
    //
    out << "\nvoid _value(" << normType << " _ob_a) { _ob_v_ = _ob_a; }";
    out << "\nvoid _value(" << inType << " _ob_a) { _ob_v_ = _ob_a; }";
    out << "\nvoid _value(const " << varType << "& _ob_a) "
        << "{ _ob_v_ = _ob_a; }";

    out.sep();

    //
    // Argument-passing conversions
    //
    out << '\n' << inType << " _boxed_in() const { return _ob_v_.in(); }";
    out << '\n' << inOutType << " _boxed_inout() "
        << "{ return _ob_v_.inout(); }";
    out << '\n' << inOutType << " _boxed_out() { return _ob_v_.out(); }";

    out.sep();

    //
    // String_var functions
    //
    out << '\n' << charType << "& operator[](short idx) "
        << "{ return _ob_v_[idx]; }";
    out << '\n' << charType << " operator[](short idx) const "
        << "{ return _ob_v_[idx]; }";
    out << '\n' << charType << "& operator[](unsigned short idx) "
        << "{ return _ob_v_[idx]; }";
    out << '\n' << charType << " operator[](unsigned short idx) const "
        << "{ return _ob_v_[idx]; }";
    out << '\n' << charType << "& operator[](int idx) "
        << "{ return _ob_v_[idx]; }";
    out << '\n' << charType << " operator[](int idx) const "
        << "{ return _ob_v_[idx]; }";
    out << '\n' << charType << "& operator[](unsigned int idx) "
        << "{ return _ob_v_[idx]; }";
    out << '\n' << charType << " operator[](unsigned int idx) const "
        << "{ return _ob_v_[idx]; }";
    out << '\n' << charType << "& operator[](long idx) "
        << "{ return _ob_v_[idx]; }";
    out << '\n' << charType << " operator[](long idx) const "
        << "{ return _ob_v_[idx]; }";
    out << '\n' << charType << "& operator[](unsigned long idx) "
        << "{ return _ob_v_[idx]; }";
    out << '\n' << charType << " operator[](unsigned long idx) const "
        << "{ return _ob_v_[idx]; }";

    out.sep();

    //
    // ORBacus-specific functions
    //
    if(origType -> kind() == tk_string)
    {
        out << '\n' << name << "& operator+=(const char* _ob_a)";
        out.start();
        out << "\n_ob_v_ += _ob_a;\nreturn *this;";
        out.end();
        out.sep();
        out << '\n' << name << "& operator+=(char _ob_a)";
        out.start();
        out << "\n_ob_v_ += _ob_a;\nreturn *this;";
        out.end();
        out.sep();
        out << '\n' << name << "& operator+=(unsigned char _ob_a)";
        out.start();
        out << "\n_ob_v_ += _ob_a;\nreturn *this;";
        out.end();
        out.sep();
        out << '\n' << name << "& operator+=(long _ob_a)";
        out.start();
        out << "\n_ob_v_ += _ob_a;\nreturn *this;";
        out.end();
        out.sep();
        out << '\n' << name << "& operator+=(unsigned long _ob_a)";
        out.start();
        out << "\n_ob_v_ += _ob_a;\nreturn *this;";
        out.end();
        out.sep();
        out << '\n' << name << "& operator+=(short _ob_a)";
        out.start();
        out << "\n_ob_v_ += _ob_a;\nreturn *this;";
        out.end();
        out.sep();
        out << '\n' << name << "& operator+=(unsigned short _ob_a)";
        out.start();
        out << "\n_ob_v_ += _ob_a;\nreturn *this;";
        out.end();
        out.sep();
        out << '\n' << name << "& operator+=(int _ob_a)";
        out.start();
        out << "\n_ob_v_ += _ob_a;\nreturn *this;";
        out.end();
        out.sep();
        out << '\n' << name << "& operator+=(unsigned int _ob_a)";
        out.start();
        out << "\n_ob_v_ += _ob_a;\nreturn *this;";
        out.end();
    }
    else
    {
        out << '\n' << name << "& operator+=(const wchar_t* _ob_a)";
        out.start();
        out << "\n_ob_v_ += _ob_a;\nreturn *this;";
        out.end();
        out.sep();
        out << '\n' << name << "& operator+=(wchar_t _ob_a)";
        out.start();
        out << "\n_ob_v_ += _ob_a;\nreturn *this;";
        out.end();
    }
}

// ----------------------------------------------------------------------
// Generate union value box
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHUnionBox(ValueBoxDef_ptr value,
                                       const char* scope,
                                       const char* absolute,
                                       TypeDescription* desc,
                                       TypeCode_ptr type,
                                       IdlPrettyPrint& out)
{
    IDLType_var idlType = value -> original_type_def();
    UnionDef_var unionDef = UnionDef::_narrow(idlType);
    assert(! is_nil(unionDef));

    TypeCode_var origType = GetOrigType(type);
    String_var name = fixKwd(desc -> name);

    String_var normType = getTypeString(scope, type, GetTypeNormal);
    String_var inType = getTypeString(scope, type, GetTypeIn);
    String_var inOutType = getTypeString(scope, type, GetTypeInOut);
    String_var outType = getTypeString(scope, type, GetTypeOut);
    String_var varType = getTypeString(scope, type, GetTypeVar);
    String_var retType = getTypeString(scope, type, GetTypeReturn);

    out << '\n' << varType << " _ob_v_;";

    out.sep();
    out << '\n' << name << "& operator=(const " << name << "&);";

    out.sep();
    out.dec();
    out << "\nprotected:";
    out.inc();

    out << "\nvirtual ~" << name << "() { }";

    out.sep();
    out.dec();
    out << "\npublic:";
    out.inc();

    //
    // Constructors
    //

    out << '\n' << name << "()";
    out.inc();
    out << "\n: _ob_v_(new " << normType << ") { }";
    out.dec();
    out.sep();

    out << '\n' << name << "(const " << name << "& _ob_a)";
    out.inc();
    out << "\n: _ob_v_(new " << normType << "(_ob_a._ob_v_.in())) { }";
    out.dec();
    out.sep();

    out << '\n' << name << '(' << inType << " _ob_a)";
    out.inc();
    out << "\n: _ob_v_(new " << normType << "(_ob_a)) { }";
    out.dec();

    out.sep();

    //
    // Assignment operator
    //
    out << '\n' << name << "& operator=(" << inType << " _ob_a)";
    out.start();
    out << "\n_ob_v_ = new " << normType << "(_ob_a);";
    out << "\nreturn *this;";
    out.end();

    out.sep();

    //
    // Accessors
    //
    out << '\n' << inType << " _value() const";
    out.start();
    out << "\nreturn _ob_v_.in();";
    out.end();
    out.sep();
    out << '\n' << inOutType << " _value()";
    out.start();
    out << "\nreturn _ob_v_.inout();";
    out.end();

    out.sep();

    //
    // Modifier
    //
    out << "\nvoid _value(" << inType << " _ob_a) "
        << "{ _ob_v_ = new " << normType << "(_ob_a); }";

    out.sep();

    //
    // Argument-passing conversions
    //
    out << '\n' << inType << " _boxed_in() const "
        << "{ return _ob_v_.in(); }";
    out << '\n' << inOutType << " _boxed_inout() "
        << "{ return _ob_v_.inout(); }";
    out << '\n' << retType << "& _boxed_out() "
        << "{ return _ob_v_.out(); }";

    //
    // Union member functions
    //
    IdlUnionMemberInfoSeq_var infoSeq = IdlGetUnionInfo(type);
    RepositoryId_var unionId = unionDef -> id();

    for(ULong i = 0 ; i < infoSeq -> length() ; i++)
    {
        if(is_nil(infoSeq[i].type))
            continue;
        genCPlusPlusHUnionBoxMember(unionId, infoSeq[i], scope, out);
    }
}

// ----------------------------------------------------------------------
// Generate code for a union value box member
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHUnionBoxMember(const char* unionId,
                                             const IdlUnionMemberInfo& desc,
                                             const char* scope,
                                             IdlPrettyPrint& out)
{
    Identifier_var name = fixKwd(desc.name);

    String_var norm = getTypeString(scope, desc.type, GetTypeNormal);

    TypeCode_var origMemberType = GetOrigType(desc.type);

    switch(origMemberType -> kind())
    {
    case tk_null:
    case tk_void:
    case tk_except:
    case tk_native:
        assert(false);
        break;

    case tk_short:
    case tk_long:
    case tk_longlong:
    case tk_ushort:
    case tk_ulong:
    case tk_ulonglong:
    case tk_float:
    case tk_double:
    case tk_longdouble:
    case tk_fixed:
    case tk_boolean:
    case tk_char:
    case tk_wchar:
    case tk_octet:
    case tk_enum:
        out.sep();
        out << "\nvirtual void " << name << '(' << norm << " _ob_a) "
            << "{ _ob_v_ -> " << name << "(_ob_a); }";
        out << "\nvirtual " << norm << ' ' << name << "() const "
            << "{ return _value()." << name << "(); }";
        break;

    case tk_value:
    case tk_value_box:
        out.sep();
        out << "\nvirtual void " << name << '(' << norm << "* _ob_a) "
            << "{ _ob_v_ -> " << name << "(_ob_a); }";
        out << "\nvirtual " << norm << "* " << name << "() const "
            << "{ return _value()." << name << "(); }";
        break;

    case tk_struct:
    case tk_union:
    case tk_sequence:
    case tk_any:
        //
        // Check for anonymous sequence
        //
        if(desc.type -> kind() == tk_sequence)
        {
            norm = getAbsolute(unionId, scope);
            norm += "::_";
            norm += desc.name;
            norm += "_seq";
        }

        out.sep();
        out << "\nvirtual void " << name << "(const " << norm << "& _ob_a) "
            << "{ _ob_v_ -> " << name << "(_ob_a); }";
        out << "\nvirtual const " << norm << "& " << name << "() const "
            << "{ return _value()." << name << "(); }";
        out << "\nvirtual " << norm << "& " << name << "() "
            << "{ return _ob_v_ -> " << name << "(); }";
        break;

    case tk_TypeCode:
    case tk_Principal:
    case tk_objref:
    case tk_abstract_interface:
        out.sep();
        out << "\nvirtual void " << name << '(' << norm << "_ptr _ob_a) "
            << "{ _ob_v_ -> " << name << "(_ob_a); }";
        out << "\nvirtual " << norm << "_ptr " << name << "() const "
            << "{ return _value()." << name << "(); }";
        break;

    case tk_string:
        out.sep();
        out << "\nvirtual void " << name << "(char* _ob_a) "
            << "{ _ob_v_ -> " << name << "(_ob_a); }";
        out << "\nvirtual void " << name << "(const char* _ob_a) "
            << "{ _ob_v_ -> " << name << "(_ob_a); }";
        out << "\nvirtual void " << name << "(const CORBA::String_var& _ob_a) "
            << "{ _ob_v_ -> " << name << "(_ob_a); }";
        out << "\nvirtual const char* " << name << "() const "
            << "{ return _value()." << name << "(); }";
        break;

    case tk_wstring:
        out.sep();
        out << "\nvirtual void " << name << "(wchar_t* _ob_a) "
            << "{ _ob_v_ -> " << name << "(_ob_a); }";
        out << "\nvirtual void " << name << "(const wchar_t* _ob_a) "
            << "{ _ob_v_ -> " << name << "(_ob_a); }";
        out << "\nvirtual void " << name
            << "(const CORBA::WString_var& _ob_a) " << "{ _ob_v_ -> "
            << name << "(_ob_a); }";
        out << "\nvirtual const wchar_t* " << name << "() const "
            << "{ return _value()." << name << "(); }";
        break;

    case tk_array:
    {
        String_var slice;
        String_var in =
            getTypeString(scope, desc.type, GetTypeNormal, false, "_ob_a");

        bool isAnon = !(desc.type -> kind() == tk_alias);

        if(! isAnon)
        {
            slice = getTypeString(scope, desc.type, GetTypeSlice);
        }
        else
        {
            slice = getAbsolute(unionId, scope);
            slice += "::_";
            slice += desc.name;
            slice += "_slice";
        }

        out.sep();
        out << "\nvirtual void " << name << '(' << in << ") "
            << "{ _ob_v_ -> " << name << "(_ob_a); }";

        out << "\nvirtual " << slice << "* " << name << "() const "
            << "{ return _value()." << name << "(); }";
        break;
    }

    case tk_alias:
        assert(false);
        break;
    }
}

// ----------------------------------------------------------------------
// Generate sequence value box
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHSequenceBox(ValueBoxDef_ptr value,
                                          const char* scope,
                                          const char* absolute,
                                          TypeDescription* desc,
                                          TypeCode_ptr type,
                                          IdlPrettyPrint& out)
{
    String_var name = fixKwd(desc -> name);
    TypeCode_var origType = GetOrigType(type);
    TypeCode_var content = origType -> content_type();
    TypeCode_var origContentType = GetOrigType(content);
    bool bounded = (origType -> length() > 0);
    String_var uniqueType = string_dup("OBUnique");

    String_var normContent =
        getTypeString(scope, content, GetTypeNormal, false, 0, uniqueType);
    String_var inContent =
        getTypeString(scope, content, GetTypeIn, false, 0, uniqueType);

    //
    // If the contents is an anon-type then we have to munge the
    // sequence type string.
    //
    if(origContentType -> kind() == tk_array)
    {
        normContent += "_slice*";
    }
    else if(origContentType -> kind() == tk_objref)
    {
        normContent += "_ptr";
    }
    String_var contentPtr = string_dup(normContent);
    contentPtr += "*";

    String_var subContent;
    String_var constBuff;
    if(origContentType -> kind() == tk_string)
    {
        subContent = string_dup("OB::StrForSeq");
        constBuff = string_dup("const char* const *");
    }
    else if(origContentType -> kind() == tk_wstring)
    {
        subContent = string_dup("OB::WStrForSeq");
        constBuff = string_dup("const wchar_t* const *");
    }
    else
    {
        subContent = string_dup(normContent);
        subContent += "&";
        constBuff = string_dup("const ");
        constBuff += contentPtr;
    }

    String_var normType, inType, inOutType, outType, varType, retType;

    //
    // If type is anonymous, emit a unique class and some typedefs
    //
    // XXX Mapping is unclear on what to do for a value box of an
    //     anonymous type. Without typedefs, code becomes vendor-specific.
    // XXX Also need a typedef for content?  What about something like
    //     valuetype Foo sequence< sequence<long> >; ??
    //
    if(type -> kind() == tk_sequence)
    {
        normType =
            getTypeString(scope, type, GetTypeNormal, false, 0, uniqueType);
        outType = getTypeString(scope, type, GetTypeOut, false, 0, uniqueType);
        varType = getTypeString(scope, type, GetTypeVar, false, 0, uniqueType);
        out << "\nclass " << uniqueType << " { };";
        out << "\ntypedef " << normType << " _seq_type;";
        out << "\ntypedef " << varType << " _var_type;";
        out << "\ntypedef " << outType << " _out_type;";
        normType = string_dup("_seq_type");
        inType = string_dup("const _seq_type&");
        inOutType = string_dup("_seq_type&");
        outType = string_dup("_out_type");
        varType = string_dup("_var_type");
        retType = string_dup("_seq_type*");
    }
    else
    {
        normType = getTypeString(scope, type, GetTypeNormal);
        inType = getTypeString(scope, type, GetTypeIn);
        inOutType = getTypeString(scope, type, GetTypeInOut);
        outType = getTypeString(scope, type, GetTypeOut);
        varType = getTypeString(scope, type, GetTypeVar);
        retType = getTypeString(scope, type, GetTypeReturn);
    }

    out << '\n' << varType << " _ob_v_;";

    out.sep();
    out << '\n' << name << "& operator=(const " << name << "&);";

    out.sep();
    out.dec();
    out << "\nprotected:";
    out.inc();

    out << "\nvirtual ~" << name << "() { }";

    out.sep();
    out.dec();
    out << "\npublic:";
    out.inc();

    //
    // Constructors
    //

    out << '\n' << name << "()";
    out.inc();
    out << "\n: _ob_v_(new " << normType << ") { }";
    out.dec();
    out.sep();

    if(!bounded)
    {
        out << '\n' << name << "(CORBA::ULong max)";
        out.inc();
        out << "\n: _ob_v_(new " << normType << "(max)) { }";
        out.dec();

        //
        // MSVC cannot handle default parameter values that refer to
        // namespace-scoped values, so we have to provide two versions
        // of this function which are functionally equivalent.
        //
        // This is what we really want:
        //
        //out.sep();
        //out << '\n' << name << "(CORBA::ULong max, CORBA::ULong l, "
        //    << contentPtr << " v, CORBA::Boolean rel = CORBA::FALSE)";
        //out.inc();
        //out << "\n: _ob_v_(new " << normType << "(max, l, v, rel)) { }";
        //out.dec();

        out.sep();
        out << '\n' << name << "(CORBA::ULong max, CORBA::ULong l, "
            << contentPtr << " v, CORBA::Boolean rel)";
        out.inc();
        out << "\n: _ob_v_(new " << normType << "(max, l, v, rel)) { }";
        out.dec();
        out.sep();
        out << '\n' << name << "(CORBA::ULong max, CORBA::ULong l, "
            << contentPtr << " v)";
        out.inc();
        out << "\n: _ob_v_(new " << normType
            << "(max, l, v, CORBA::FALSE)) { }";
        out.dec();
    }
    else
    {
        //
        // MSVC cannot handle default parameter values that refer to
        // namespace-scoped values, so we have to provide two versions
        // of this function which are functionally equivalent.
        //
        // This is what we really want:
        //
        //out << '\n' << name << "(CORBA::ULong l, "
        //    << contentPtr << " v, CORBA::Boolean rel = CORBA::FALSE)";
        //out.inc();
        //out << "\n: _ob_v_(new " << normType << "(l, v, rel)) { }";
        //out.dec();

        out << '\n' << name << "(CORBA::ULong l, "
            << contentPtr << " v, CORBA::Boolean rel)";
        out.inc();
        out << "\n: _ob_v_(new " << normType << "(l, v, rel)) { }";
        out.dec();
        out.sep();
        out << '\n' << name << "(CORBA::ULong l, "
            << contentPtr << " v)";
        out.inc();
        out << "\n: _ob_v_(new " << normType << "(l, v, CORBA::FALSE)) { }";
        out.dec();
    }

    out.sep();

    out << '\n' << name << "(const " << name << "& _ob_a)";
    out.inc();
    out << "\n: _ob_v_(new " << normType << "(_ob_a._ob_v_.in())) { }";
    out.dec();
    out.sep();

    out << '\n' << name << '(' << inType << " _ob_a)";
    out.inc();
    out << "\n: _ob_v_(new " << normType << "(_ob_a)) { }";
    out.dec();

    out.sep();

    //
    // Assignment operator
    //
    out << '\n' << name << "& operator=(" << inType << " _ob_a)";
    out.start();
    out << "\n_ob_v_ = new " << normType << "(_ob_a);";
    out << "\nreturn *this;";
    out.end();

    out.sep();

    //
    // Accessors
    //
    out << '\n' << inType << " _value() const";
    out.start();
    out << "\nreturn _ob_v_.in();";
    out.end();
    out.sep();
    out << '\n' << inOutType << " _value()";
    out.start();
    out << "\nreturn _ob_v_.inout();";
    out.end();

    out.sep();

    //
    // Modifier
    //
    out << "\nvoid _value(" << inType << " _ob_a)";
    out.start();
    out << "\n_ob_v_ = new " << normType << "(_ob_a);";
    out.end();

    out.sep();

    //
    // Argument-passing conversions
    //
    out << '\n' << inType << " _boxed_in() const "
        << "{ return _ob_v_.in(); }";
    out << '\n' << inOutType << " _boxed_inout() "
        << "{ return _ob_v_.inout(); }";
    out << '\n' << retType << "& _boxed_out() "
        << "{ return _ob_v_.out(); }";

    out.sep();

    //
    // Sequence functions
    //
    out << "\nCORBA::ULong maximum() const { return _ob_v_.in().maximum(); }";
    out.sep();
    out << "\nCORBA::ULong length() const { return _ob_v_.in().length(); }";
    out << "\nvoid length(CORBA::ULong l) { _ob_v_ -> length(l); }";
    out.sep();
    out << '\n' << subContent << " operator[](CORBA::ULong idx) "
        << "{ return _ob_v_[idx]; }";
    out << '\n' << inContent << " operator[](CORBA::ULong idx) const "
        << "{ return _ob_v_[idx]; }";
    out.sep();
    out << "\nCORBA::Boolean release() const "
        << "{ return _ob_v_.in().release(); }";
    out.sep();

    if(bounded)
    {
        //
        // MSVC cannot handle default parameter values that refer to
        // namespace-scoped values, so we have to provide two versions
        // of this function which are functionally equivalent.
        //
        // This is what we really want:
        //
        //out << "\nvoid replace(CORBA::ULong l, " << contentPtr
        //    << " d, CORBA::Boolean rel = CORBA::FALSE) "
        //    << "{ _ob_v_ -> replace(l, d, rel); }";

        out << "\nvoid replace(CORBA::ULong l, " << contentPtr
            << " d, CORBA::Boolean rel) "
            << "{ _ob_v_ -> replace(l, d, rel); }";
        out << "\nvoid replace(CORBA::ULong l, " << contentPtr << " d) "
            << "{ _ob_v_ -> replace(l, d, CORBA::FALSE); }";
    }
    else
    {
        //
        // MSVC cannot handle default parameter values that refer to
        // namespace-scoped values, so we have to provide two versions
        // of this function which are functionally equivalent.
        //
        // This is what we really want:
        //
        //out << "\nvoid replace(CORBA::ULong m, CORBA::ULong l, " << contentPtr
        //    << " d, CORBA::Boolean rel = CORBA::FALSE) "
        //    << "{ _ob_v_ -> replace(m, l, d, rel); }";

        out << "\nvoid replace(CORBA::ULong m, CORBA::ULong l, " << contentPtr
            << " d, CORBA::Boolean rel) "
            << "{ _ob_v_ -> replace(m, l, d, rel); }";
        out << "\nvoid replace(CORBA::ULong m, CORBA::ULong l, " << contentPtr
            << " d) { _ob_v_ -> replace(m, l, d, CORBA::FALSE); }";
    }
    out.sep();

    //
    // MSVC cannot handle default parameter values that refer to
    // namespace-scoped values, so we have to provide two versions
    // of this function which are functionally equivalent.
    //
    // This is what we really want:
    //
    //out << '\n' << contentPtr
    //    << " get_buffer(CORBA::Boolean orphan = CORBA::FALSE) "
    //    << "{ return _ob_v_ -> get_buffer(orphan); }";

    out << '\n' << contentPtr
        << " get_buffer(CORBA::Boolean orphan) "
        << "{ return _ob_v_ -> get_buffer(orphan); }";
    out << '\n' << contentPtr
        << " get_buffer() "
        << "{ return _ob_v_ -> get_buffer(CORBA::FALSE); }";

    out << '\n' << constBuff << " get_buffer() const "
        << "{ return _ob_v_.in().get_buffer(); }";

    //
    // ORBacus-specific functions
    //
    if(! bounded)
    {
        out.sep();

        if(origContentType -> kind() == tk_string)
        {
            out << "\nvoid insert(char* _ob_a) "
                << "{ _ob_v_ -> insert(_ob_a); }";
            out << "\nvoid insert(const char* _ob_a) "
                << "{ _ob_v_ -> insert(_ob_a); }";
            out << "\nvoid insert(const CORBA::String_var& _ob_a) "
                << "{ _ob_v_ -> insert(_ob_a); }";
            out << "\nvoid insert(const OB::StrForSeq& _ob_a) "
                << "{ _ob_v_ -> insert(_ob_a); }";
            out.sep();
            out << "\nvoid append(char* _ob_a) "
                << "{ _ob_v_ -> append(_ob_a); }";
            out << "\nvoid append(const char* _ob_a) "
                << "{ _ob_v_ -> append(_ob_a); }";
            out << "\nvoid append(const CORBA::String_var& _ob_a) "
                << "{ _ob_v_ -> append(_ob_a); }";
            out << "\nvoid append(const OB::StrForSeq& _ob_a) "
                << "{ _ob_v_ -> append(_ob_a); }";
            out.sep();
        }
        else if(origContentType -> kind() == tk_wstring)
        {
            out << "\nvoid insert(wchar_t* _ob_a) "
                << "{ _ob_v_ -> insert(_ob_a); }";
            out << "\nvoid insert(const wchar_t* _ob_a) "
                << "{ _ob_v_ -> insert(_ob_a); }";
            out << "\nvoid insert(const CORBA::WString_var& _ob_a) "
                << "{ _ob_v_ -> insert(_ob_a); }";
            out << "\nvoid insert(const OB::WStrForSeq& _ob_a) "
                << "{ _ob_v_ -> insert(_ob_a); }";
            out.sep();
            out << "\nvoid append(wchar_t* _ob_a) "
                << "{ _ob_v_ -> append(_ob_a); }";
            out << "\nvoid append(const wchar_t* _ob_a) "
                << "{ _ob_v_ -> append(_ob_a); }";
            out << "\nvoid append(const CORBA::WString_var& _ob_a) "
                << "{ _ob_v_ -> append(_ob_a); }";
            out << "\nvoid append(const OB::WStrForSeq& _ob_a) "
                << "{ _ob_v_ -> append(_ob_a); }";
            out.sep();
        }
        else
        {
            out << "\nvoid insert(" << inContent << " _ob_a) "
                << "{ _ob_v_ -> insert(_ob_a); }";
            out << "\nvoid append(" << inContent << " _ob_a) "
                << "{ _ob_v_ -> append(_ob_a); }";
        }

        out << "\nvoid remove(CORBA::ULong idx) { _ob_v_ -> remove(idx); }";

        // XXX consume
    }
}

// ----------------------------------------------------------------------
// Generate any value box
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHAnyBox(ValueBoxDef_ptr value,
                                     const char* scope,
                                     const char* absolute,
                                     TypeDescription* desc,
                                     TypeCode_ptr type,
                                     IdlPrettyPrint& out)
{
    TypeCode_var origType = GetOrigType(type);
    String_var name = fixKwd(desc -> name);

    String_var normType = getTypeString(scope, type, GetTypeNormal);
    String_var inType = getTypeString(scope, type, GetTypeIn);
    String_var inOutType = getTypeString(scope, type, GetTypeInOut);
    String_var varType = getTypeString(scope, type, GetTypeVar);
    String_var retType = getTypeString(scope, type, GetTypeReturn);

    out << '\n' << varType << " _ob_v_;";

    out.sep();
    out << '\n' << name << "& operator=(const " << name << "&);";

    out.sep();
    out.dec();
    out << "\nprotected:";
    out.inc();

    out << "\nvirtual ~" << name << "() { }";

    out.sep();
    out.dec();
    out << "\npublic:";
    out.inc();

    //
    // Constructors
    //

    out << '\n' << name << "()";
    out.inc();
    out << "\n: _ob_v_(new " << normType << ") { }";
    out.dec();
    out.sep();

    //
    // MSVC cannot handle default parameter values that refer to
    // namespace-scoped values, so we have to provide two versions
    // of this function which are functionally equivalent.
    //
    // This is what we really want:
    //
    //out << '\n' << name << "(CORBA::TypeCode_ptr tc, void* d, "
    //    << "CORBA::Boolean rel = CORBA::FALSE)";
    //out.inc();
    //out << "\n: _ob_v_(new " << normType << "(tc, d, rel)) { }";
    //out.dec();

    out << '\n' << name << "(CORBA::TypeCode_ptr tc, void* d, "
        << "CORBA::Boolean rel)";
    out.inc();
    out << "\n: _ob_v_(new " << normType << "(tc, d, rel)) { }";
    out.dec();
    out << '\n' << name << "(CORBA::TypeCode_ptr tc, void* d)";
    out.inc();
    out << "\n: _ob_v_(new " << normType << "(tc, d, CORBA::FALSE)) { }";
    out.dec();

    out.sep();

    out << '\n' << name << "(const " << name << "& _ob_a)";
    out.inc();
    out << "\n: _ob_v_(new " << normType << "(_ob_a._ob_v_.in())) { }";
    out.dec();
    out.sep();

    out << '\n' << name << '(' << inType << " _ob_a)";
    out.inc();
    out << "\n: _ob_v_(new " << normType << "(_ob_a)) { }";
    out.dec();

    out.sep();

    //
    // Assignment operator
    //
    out << '\n' << name << "& operator=(" << inType << " _ob_a)";
    out.start();
    out << "\n_ob_v_ = new " << normType << "(_ob_a);";
    out << "\nreturn *this;";
    out.end();

    out.sep();

    //
    // Accessors
    //
    out << '\n' << inType << " _value() const";
    out.start();
    out << "\nreturn _ob_v_.in();";
    out.end();
    out.sep();
    out << '\n' << inOutType << " _value()";
    out.start();
    out << "\nreturn _ob_v_.inout();";
    out.end();

    out.sep();

    //
    // Modifier
    //
    out << "\nvoid _value(" << inType << " _ob_a) "
        << "{ _ob_v_ = new " << normType << "(_ob_a); }";

    out.sep();

    //
    // Argument-passing conversions
    //
    out << '\n' << inType << " _boxed_in() const "
        << "{ return _ob_v_.in(); }";
    out << '\n' << inOutType << " _boxed_inout() "
        << "{ return _ob_v_.inout(); }";
    out << '\n' << retType << "& _boxed_out() "
        << "{ return _ob_v_.out(); }";

    out.sep();

    //
    // Public interface
    //

    //
    // MSVC cannot handle default parameter values that refer to
    // namespace-scoped values, so we have to provide two versions
    // of this function which are functionally equivalent.
    //
    // This is what we really want:
    //
    //out << "\nvoid replace(CORBA::TypeCode_ptr tc, void* d, "
    //    << "CORBA::Boolean rel = CORBA::FALSE) "
    //    << "{ _ob_v_ -> replace(tc, d, rel); }";

    out << "\nvoid replace(CORBA::TypeCode_ptr tc, void* d, "
        << "CORBA::Boolean rel) { _ob_v_ -> replace(tc, d, rel); }";
    out << "\nvoid replace(CORBA::TypeCode_ptr tc, void* d) "
        << "{ _ob_v_ -> replace(tc, d, CORBA::FALSE); }";

    out.sep();
    out << "\nCORBA::TypeCode_ptr type() const { return _ob_v_.in().type(); }";
    out << "\nconst void* value() const { return _ob_v_.in().value(); }";

    out.sep();

    // XXX include operators?

    //
    // ORBacus-specific functions
    //
    out << "\nvoid replace(CORBA::TypeCode_ptr tc, void* d, "
        << "CORBA::Boolean rel, OB::ConstructedInfo* info) "
        << "{ _ob_v_ -> replace(tc, d, rel, info); }";
    out << "\nCORBA::Boolean check_type(CORBA::TypeCode_ptr tc) const "
        << "{ return _ob_v_.in().check_type(tc); }";
    out << "\nconst OB::ConstructedInfo* info() const "
        << "{ return _ob_v_.in().info(); }";
}

// ----------------------------------------------------------------------
// Generate fixed value box
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHFixedBox(ValueBoxDef_ptr value,
                                       const char* scope,
                                       const char* absolute,
                                       TypeDescription* desc,
                                       TypeCode_ptr type,
                                       IdlPrettyPrint& out)
{
    TypeCode_var origType = GetOrigType(type);
    String_var name = fixKwd(desc -> name);

    String_var normType = getTypeString(scope, type, GetTypeNormal);
    String_var inType = getTypeString(scope, type, GetTypeIn);
    String_var inOutType = getTypeString(scope, type, GetTypeInOut);
    String_var retType = getTypeString(scope, type, GetTypeReturn);

    out << '\n' << normType << " _ob_v_;";

    out.sep();
    out << '\n' << name << "& operator=(const " << name << "&);";

    out.sep();
    out.dec();
    out << "\nprotected:";
    out.inc();

    out << "\nvirtual ~" << name << "() { }";

    out.sep();
    out.dec();
    out << "\npublic:";
    out.inc();

    //
    // Constructors
    //

    out << '\n' << name << "() { }";
    out.sep();

    out << '\n' << name << "(int _ob_a = 0)";
    out.inc();
    out << "\n: _ob_v_(_ob_a) { }";
    out.dec();
    out << '\n' << name << "(unsigned int _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a) { }";
    out.dec();
    out << '\n' << name << "(CORBA::Long _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a) { }";
    out.dec();
    out << '\n' << name << "(CORBA::ULong _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a) { }";
    out.dec();
    out << '\n' << name << "(CORBA::LongLong _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a) { }";
    out.dec();
    out << '\n' << name << "(CORBA::ULongLong _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a) { }";
    out.dec();
    out << '\n' << name << "(CORBA::Double _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a) { }";
    out.dec();
    out << '\n' << name << "(CORBA::LongDouble _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a) { }";
    out.dec();
    out << '\n' << name << '(' << inType << " _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a) { }";
    out.dec();
    out << '\n' << name << "(const char* _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a) { }";
    out.dec();

    out.sep();

    out << '\n' << name << "(const " << name << "& _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a._ob_v_) { }";
    out.dec();

    out.sep();

    //
    // Assignment operator
    //
    out << '\n' << name << "& operator=(" << inType << " _ob_a)";
    out.start();
    out << "\n_ob_v_ = _ob_a;";
    out << "\nreturn *this;";
    out.end();

    out.sep();

    //
    // Accessors
    //
    out << '\n' << inType << " _value() const";
    out.start();
    out << "\nreturn _ob_v_;";
    out.end();
    out.sep();
    out << '\n' << inOutType << " _value()";
    out.start();
    out << "\nreturn _ob_v_;";
    out.end();

    out.sep();

    //
    // Modifier
    //
    out << "\nvoid _value(" << inType << " _ob_a) { _ob_v_ = _ob_a; }";

    out.sep();

    //
    // Argument-passing conversions
    //
    out << '\n' << inType << " _boxed_in() const { return _ob_v_; }";
    out << '\n' << inOutType << " _boxed_inout() { return _ob_v_; }";
    out << '\n' << retType << "& _boxed_out() { return _ob_v_; }";

    out.sep();

    //
    // Public interface
    //
    out << '\n' << normType << " round(CORBA::UShort _ob_a) const "
        << "{ return _ob_v_.round(_ob_a); }";
    out << '\n' << normType << " truncate(CORBA::UShort _ob_a) const "
        << "{ return _ob_v_.truncate(_ob_a); }";

    out << '\n' << normType << " abs() const "
        << "{ return _ob_v_.abs(); }";
    out << "\nCORBA::Short signum() const "
        << "{ return _ob_v_.signum(); }";

    out.sep();

    out << "\nCORBA::Short scale() const "
        << "{ return _ob_v_.scale(); }";
    out << '\n' << normType << " setScale(CORBA::Short _ob_a) const "
        << "{ return _ob_v_.setScale(_ob_a); }";

    out.sep();

    out << '\n' << normType << " movePointLeft(CORBA::Short _ob_a) const "
        << "{ return _ob_v_.movePointLeft(_ob_a); }";
    out << '\n' << normType << " movePointRight(CORBA::Short _ob_a) const "
        << "{ return _ob_v_.movePointRight(_ob_a); }";

    out.sep();

    out << "\nchar* to_string() const "
        << "{ return _ob_v_.to_string(); }";

    // XXX include operators?
}

// ----------------------------------------------------------------------
// Generate array value box
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHArrayBox(ValueBoxDef_ptr value,
                                       const char* scope,
                                       const char* absolute,
                                       TypeDescription* desc,
                                       TypeCode_ptr type,
                                       IdlPrettyPrint& out)
{
    String_var name = fixKwd(desc -> name);
    TypeCode_var origType = GetOrigType(type);
    TypeCode_var content = origType -> content_type();
    TypeCode_var origContentType = GetOrigType(content);

    //
    // Grammar doesn't allow valueboxes for anonymous arrays
    //

    String_var normType =
        getTypeString(scope, type, GetTypeNormal, false, "_ob_v_");
    String_var sliceType = getTypeString(scope, type, GetTypeSlice);
    String_var inType = getTypeString(scope, type, GetTypeIn);
    String_var inOutType = getTypeString(scope, type, GetTypeInOut);
    String_var outType = getTypeString(scope, type, GetTypeOut);
    String_var varType = getTypeString(scope, type, GetTypeVar);
    String_var retType = getTypeString(scope, type, GetTypeReturn);

    out << '\n' << normType << ';';

    out.sep();
    out << '\n' << name << "& operator=(const " << name << "&);";

    out.sep();
    out.dec();
    out << "\nprotected:";
    out.inc();

    out << "\nvirtual ~" << name << "() { }";

    out.sep();
    out.dec();
    out << "\npublic:";
    out.inc();

    //
    // Constructors
    //

    out << '\n' << name << "()";
    out.start();
    if(! IdlIsVariable(type))
    {
        out.direct("\n#ifdef OB_CLEAR_MEM");
        out << "\nmemset(&_ob_v_, 0, sizeof(_ob_v_));";
        out.direct("\n#endif"); // OB_CLEAR_MEM
    }
    out.end();

    out.sep();
    out << '\n' << name << "(const " << name << "& _ob_a)";
    out.start();
    writeArrayCopy(scope, type, "_ob_v_", "_ob_a._ob_v_", "i", out);
    out.end();

    out.sep();
    out << '\n' << name << '(' << inType << " _ob_a)";
    out.start();
    writeArrayCopy(scope, type, "_ob_v_", "_ob_a", "i", out);
    out.end();

    out.sep();

    //
    // Assignment operator
    //
    out << '\n' << name << "& operator=(" << inType << " _ob_a)";
    out.start();
    writeArrayCopy(scope, type, "_ob_v_", "_ob_a", "i", out);
    out << "\nreturn *this;";
    out.end();

    out.sep();

    //
    // Accessors
    //
    out << '\n' << sliceType << "* _value() const";
    out.start();
    out << "\nreturn (" << sliceType << "*)_ob_v_;";
    out.end();
    out.sep();
    out << '\n' << sliceType << "* _value()";
    out.start();
    out << "\nreturn (" << sliceType << "*)_ob_v_;";
    out.end();

    out.sep();

    //
    // Modifier
    //
    out << "\nvoid _value(" << inType << " _ob_a)";
    out.start();
    writeArrayCopy(scope, type, "_ob_v_", "_ob_a", "i", out);
    out.end();

    out.sep();

    //
    // Argument-passing conversions
    //
    out << '\n' << sliceType << "* _boxed_in() const "
        << "{ return (" << sliceType << "*)_ob_v_; }";
    out << '\n' << sliceType << "* _boxed_inout() "
        << "{ return (" << sliceType << "*)_ob_v_; }";
    out << '\n' << sliceType << "* _boxed_out() "
        << "{ return (" << sliceType << "*)_ob_v_; }";

    out.sep();

    //
    // Array _var functions
    //
    out << '\n' << sliceType << "& operator[](short idx)";
    out << "\n{ return _ob_v_[idx]; }";
    out << "\nconst " << sliceType << "& operator[](short idx) const";
    out << "\n{ return ((const " << sliceType << "*)_ob_v_)[idx]; }";
    out << '\n' << sliceType << "& operator[](unsigned short idx)";
    out << "\n{ return _ob_v_[idx]; }";
    out << "\nconst " << sliceType << "& operator[](unsigned short idx) const";
    out << "\n{ return ((const " << sliceType << "*)_ob_v_)[idx]; }";
    out << '\n' << sliceType << "& operator[](int idx)";
    out << "\n{ return _ob_v_[idx]; }";
    out << "\nconst " << sliceType << "& operator[](int idx) const";
    out << "\n{ return ((const " << sliceType << "*)_ob_v_)[idx]; }";
    out << '\n' << sliceType << "& operator[](unsigned int idx)";
    out << "\n{ return _ob_v_[idx]; }";
    out << "\nconst " << sliceType << "& operator[](unsigned int idx) const";
    out << "\n{ return ((const " << sliceType << "*)_ob_v_)[idx]; }";
    out << '\n' << sliceType << "& operator[](long idx)";
    out << "\n{ return _ob_v_[idx]; }";
    out << "\nconst " << sliceType << "& operator[](long idx) const";
    out << "\n{ return ((const " << sliceType << "*)_ob_v_)[idx]; }";
    out << '\n' << sliceType << "& operator[](unsigned long idx)";
    out << "\n{ return _ob_v_[idx]; }";
    out << "\nconst " << sliceType << "& operator[](unsigned long idx) const";
    out << "\n{ return ((const " << sliceType << "*)_ob_v_)[idx]; }";
}
