// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <Types.h>
#include <GenUtil.h>
#include <GenCPP.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Generate code for the example source file
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusImplC(Container_ptr container, bool all,
                                   IdlPrettyPrint& out)
{
    //
    // Get scope
    //
    String_var scope = getScope(container);

    //
    // Get container contents descriptions
    //
    Container::DescriptionSeq_var seq =
        container -> describe_contents(dk_all, true, -1);

    ULong i, j;

    for(i = 0 ; i < seq -> length() ; i++)
    {
        //
        // Get container
        //
        Container_var cont =
            Container::_narrow(seq[i].contained_object);

        switch(seq[i].kind)
        {
        case dk_Module:
        {
            //
            // Module
            //
            ModuleDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

            // Don't add this here!!!
            // if(!check(desc -> id))
            //     continue;

            comment(desc -> id, out);

            out.sep();

            assert(!is_nil(cont));
            genCPlusPlusImplC(cont, all, out);

            break;
        }

        case dk_Interface:
        {
            //
            // Interface
            //
            InterfaceDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            if(desc -> is_abstract)
                continue;

            Identifier_var name = fixKwd(desc -> name);
            ScopedName_var absolute = getAbsolute(desc -> id);

            comment(desc -> id, out);

            out.sep();
            out << '\n' << absolute << "_impl::" << name << "_impl()";
            out.start();
            out.end();
            out.sep();
            out << '\n' << absolute << "_impl::~" << name << "_impl()";
            out.start();
            out.end();

            //
            // Generate attributes and operations for any base abstract
            // interfaces
            //
            if(!all)
            {
                InterfaceDef_var intf = InterfaceDef::_narrow(cont);
                assert(!is_nil(intf));
                IdlBaseInterfaceInfoSeq baseInfo;
                IdlGetBaseInterfaces(intf, baseInfo, false);

                ULong k, l;
                IdlStringSeq ids;

                for(j = 0 ; j < baseInfo.length() ; j++)
                {
                    if(!baseInfo[j].is_abstract)
                        continue;

                    Container::DescriptionSeq_var contents =
                        baseInfo[j].intf -> describe_contents(dk_all,
                                                              false, -1);

                    for(k = 0 ; k < contents -> length() ; k++)
                    {
                        if(contents[k].kind == dk_Operation)
                        {
                            OperationDescription* opDesc;
                            Boolean b = (contents[k].value) >>= opDesc;
                            assert(b);
                            for(l = 0 ; l < ids.length() ; l++)
                                if(strcmp(ids[l], opDesc -> id) == 0)
                                    break;
                            if(l >= ids.length())
                            {
                                genCPlusPlusImplCOperation(absolute,
                                                           opDesc, out);
                                ids.append(opDesc -> id);
                            }
                        }
                        else if(contents[k].kind == dk_Attribute)
                        {
                            AttributeDescription* atDesc;
                            Boolean b = (contents[k].value) >>= atDesc;
                            assert(b);
                            for(l = 0 ; l < ids.length() ; l++)
                                if(strcmp(ids[l], atDesc -> id) == 0)
                                    break;
                            if(l >= ids.length())
                            {
                                genCPlusPlusImplCAttribute(absolute,
                                                           atDesc, out);
                                ids.append(atDesc -> id);
                            }
                        }
                    }
                }
            }

            //
            // Attributes and operations
            //

            Container::DescriptionSeq_var contents =
                cont -> describe_contents(dk_all, !all, -1);

            for(j = 0 ; j < contents -> length() ; j++)
            {
                if(contents[j].kind == dk_Attribute)
                {
                    AttributeDescription* atDesc;
                    Boolean b = (contents[j].value >>= atDesc);
                    assert(b);
                    genCPlusPlusImplCAttribute(absolute, atDesc, out);
                }
                else
                {
                    OperationDescription* opDesc;
                    Boolean b = (contents[j].value >>= opDesc);
                    assert(b);
                    genCPlusPlusImplCOperation(absolute, opDesc, out);
                }
            }

            break;
        }

        case dk_Value:
        {
            //
            // Valuetype
            //
            ValueDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            ValueDef_var value = ValueDef::_narrow(cont);
            assert(! is_nil(value));

            Identifier_var name = fixKwd(desc -> name);
            ScopedName_var absolute = getAbsolute(desc -> id);
            InterfaceDef_var supportedInterface =
                value -> supported_interface();

            comment(desc -> id, out);

            out.sep();
            out << '\n' << absolute << "_impl::" << name << "_impl()";
            out.start();
            out.end();
            out.sep();
            out << '\n' << absolute << "_impl::~" << name << "_impl()";
            out.start();
            out.end();

            //
            // _copy_value
            //
            out.sep();
            out << "\nCORBA::ValueBase*\n" << absolute
                << "_impl::_copy_value()";
            out.start();
            out << "\n// TODO: Implementation";
            out.end();

            //
            // Custom valuetype
            //
            if(desc -> is_custom)
            {
                out.sep();
                out << "\nvoid\n" << absolute
                    << "_impl::marshal(CORBA::DataOutputStream* os)";
                out.start();
                out << "\n// TODO: Implementation";
                out.end();
                out.sep();
                out << "\nvoid\n" << absolute
                    << "_impl::unmarshal(CORBA::DataInputStream* is)";
                out.start();
                out << "\n// TODO: Implementation";
                out.end();
            }

            //
            // Valuetype attributes and operations
            //
            ContainedSeq_var attributes =
                value -> contents(dk_Attribute, !all);

            for(j = 0 ; j < attributes -> length() ; j++)
            {
                Contained::Description_var contDesc =
                    attributes[j] -> describe();
                AttributeDescription* atDesc;
                Boolean b = (contDesc -> value) >>= atDesc;
                assert(b);
                genCPlusPlusImplCAttribute(absolute, atDesc, out);
            }

            ContainedSeq_var operations =
                value -> contents(dk_Operation, !all);

            for(j = 0 ; j < operations -> length() ; j++)
            {
                Contained::Description_var contDesc =
                    operations[j] -> describe();
                OperationDescription* opDesc;
                Boolean b = (contDesc -> value) >>= opDesc;
                assert(b);
                genCPlusPlusImplCOperation(absolute, opDesc, out);
            }

            //
            // Supported interface attributes and operations
            //
            if(!all && !is_nil(supportedInterface))
            {
                attributes =
                    supportedInterface -> contents(dk_Attribute, !all);

                for(j = 0 ; j < attributes -> length() ; j++)
                {
                    Contained::Description_var contDesc =
                        attributes[j] -> describe();
                    AttributeDescription* atDesc;
                    Boolean b = (contDesc -> value) >>= atDesc;
                    assert(b);
                    genCPlusPlusImplCAttribute(absolute, atDesc, out);
                }

                operations =
                    supportedInterface -> contents(dk_Operation, !all);

                for(j = 0 ; j < operations -> length() ; j++)
                {
                    Contained::Description_var contDesc =
                        operations[j] -> describe();
                    OperationDescription* opDesc;
                    Boolean b = (contDesc -> value) >>= opDesc;
                    assert(b);
                    genCPlusPlusImplCOperation(absolute, opDesc, out);
                }
            }

            break;
        }

        default:
            break;
        }
    }
}

// ----------------------------------------------------------------------
// Generate code for an example attribute implementation
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusImplCAttribute(const char* absolute,
                                            AttributeDescription* desc,
                                            IdlPrettyPrint& out)
{
    Identifier_var name = fixKwd(desc -> name);

    comment(desc -> id, out);
    out.sep();

    TypeCode_var origType = GetOrigType(desc -> type);
    String_var ret = getTypeString("", desc -> type, GetTypeReturn);
    out << '\n' << ret << '\n' << absolute << "_impl::" << name << "()";
    out.inc();
    out << "\nthrow(CORBA::SystemException)";
    out.dec();
    out.start();
    out << "\n// TODO: Implementation";
    genCPlusPlusImplCReturn(desc -> type, "_r", out);
    out << "\nreturn _r;";
    out.end();

    if(desc -> mode == ATTR_NORMAL)
    {
        out.sep();

        String_var in = getTypeString("", desc -> type, GetTypeIn);

        out << "\nvoid\n" << absolute << "_impl::" << name << '(' << in
            << " a)";
        out.inc();
        out << "\nthrow(CORBA::SystemException)";
        out.dec();
        out.start();
        out << "\n// TODO: Implementation";
        out.end();
    }
}

// ----------------------------------------------------------------------
// Generate code for an example operation implementation
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusImplCOperation(const char* absolute,
                                            OperationDescription* desc,
                                            IdlPrettyPrint& out)
{
    Identifier_var name = fixKwd(desc -> name);

    comment(desc -> id, out);
    out.sep();

    String_var ret = getTypeString("", desc -> result, GetTypeReturn);

    out << '\n' << ret << '\n' << absolute << "_impl::" << name << '(';

    unsigned int indent = out.getIndent();
    out.setIndent(out.getPos());

    ULong i;
    for(i = 0 ; i < desc -> parameters.length() ; i++)
    {
        GetType tt;
        switch(desc -> parameters[i].mode)
        {
        case PARAM_IN:
            tt = GetTypeIn;
            break;

        case PARAM_INOUT:
            tt = GetTypeInOut;
            break;

        case PARAM_OUT:
            tt = GetTypeOut;
            break;
        }

        String_var argName = fixKwd(desc -> parameters[i].name);

        String_var s = getTypeString("", desc -> parameters[i].type,
                                           tt, false, argName);

        if(i > 0)
            out << ",\n";

        out << s;
    }

    if(desc -> contexts.length() > 0)
    {
        if(desc -> parameters.length() > 0)
            out << ",\n";
    
        out << "CORBA::Context_ptr context";
    }

    out << ')';

    out.setIndent(indent);

    out.inc();
    out << "\nthrow(";

    indent = out.getIndent();
    out.setIndent(out.getPos());

    for(i = 0 ; i < desc -> exceptions.length() ; i++)
    {
        Identifier_var exName = getAbsolute(desc -> exceptions[i].id);
        out << exName << ",\n";
    }

    out << "CORBA::SystemException)";
    out.setIndent(indent);
    out.dec();

    out.start();
    out << "\n// TODO: Implementation";

    if(desc -> result -> kind() != tk_void)
        genCPlusPlusImplCReturn(desc -> result, "_r", out);

    //
    // Show example code for those out and inout parameters requiring
    // allocation
    //
    for(i = 0 ; i < desc -> parameters.length() ; i++)
    {
        if(desc -> parameters[i].mode == PARAM_IN)
            continue;

        String_var argName = fixKwd(desc -> parameters[i].name);

        //
        // Get original type
        //
        TypeCode_var origType =
            GetOrigType(desc -> parameters[i].type);

        //
        // Check if original type is fixed-sized or variable-sized
        //
        bool variable = IdlIsVariable(origType);

        String_var norm =
            getTypeString("", desc -> parameters[i].type, GetTypeNormal);
        String_var ret =
            getTypeString("", desc -> parameters[i].type, GetTypeReturn);

        TCKind kind = origType -> kind();
        switch(kind)
        {
        case tk_void:
        case tk_null:
        case tk_alias:
        case tk_except:
            assert(false);
            break;

        case tk_boolean:
        case tk_short:
        case tk_long:
        case tk_longlong:
        case tk_ushort:
        case tk_ulong:
        case tk_ulonglong:
        case tk_octet:
        case tk_float:
        case tk_double:
        case tk_longdouble:
        case tk_char:
        case tk_wchar:
        case tk_enum:
        case tk_fixed:
            // nothing to do
            break;

        case tk_string:
            if(desc -> parameters[i].mode == PARAM_INOUT)
                out << "\nCORBA::string_free(" << argName << ");";
            out << '\n' << argName << " = CORBA::string_dup(\"\");";
            break;

        case tk_wstring:
            if(desc -> parameters[i].mode == PARAM_INOUT)
                out << "\nCORBA::wstring_free(" << argName << ");";
            out << '\n' << argName << " = CORBA::wstring_dup(L\"\");";
            break;

        case tk_any:
            if(desc -> parameters[i].mode == PARAM_OUT)
                out << '\n' << argName << " = new " << norm << ';';
            break;

        case tk_TypeCode:
        case tk_Principal:
        case tk_objref:
        case tk_abstract_interface:
            if(desc -> parameters[i].mode == PARAM_INOUT)
                out << "\nCORBA::release(" << argName << ");";
            out << '\n' << argName << " = " << norm << "::_nil();";
            break;

        case tk_struct:
        case tk_union:
            if(desc -> parameters[i].mode == PARAM_OUT && variable)
                out << '\n' << argName << " = new " << norm << ';';
            break;

        case tk_sequence:
            if(desc -> parameters[i].mode == PARAM_OUT)
                out << '\n' << argName << " = new " << norm << ';';
            break;

        case tk_array:
            if(desc -> parameters[i].mode == PARAM_OUT && variable)
            {
                String_var arr =
                    getTypeString("", desc -> parameters[i].type,
                                  GetTypeArray);
                out << '\n' << argName << " = " << arr << "_alloc();";
            }
            break;

        case tk_native:
            out << '\n' << argName << " = (" << norm << ")0;";
            break;

        case tk_value:
        case tk_value_box:
            out << '\n' << argName << " = 0;";
            break;
        }
    }

    if(desc -> result -> kind() != tk_void)
        out << "\nreturn _r;";

    out.end();
}

// ----------------------------------------------------------------------
// Generate example return statement
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusImplCReturn(TypeCode_ptr type,
                                         const char* name,
                                         IdlPrettyPrint& out)
{
    //
    // Get original type
    //
    TypeCode_var origType = GetOrigType(type);

    //
    // Check if original type is fixed-sized or variable-sized
    //
    bool variable = IdlIsVariable(origType);

    String_var norm = getTypeString("", type, GetTypeNormal);
    String_var ret = getTypeString("", type, GetTypeReturn);

    //
    // Get result
    //
    TCKind kind = origType -> kind();
    switch(kind)
    {
    case tk_void:
    case tk_null:
    case tk_alias:
    case tk_except:
        assert(false);
        break;

    case tk_boolean:
        out << '\n' << ret << ' ' << name << " = CORBA::FALSE;";
        break;

    case tk_short:
    case tk_long:
    case tk_longlong:
    case tk_ushort:
    case tk_ulong:
    case tk_ulonglong:
    case tk_octet:
        out << '\n' << ret << ' ' << name << " = 0;";
        break;

    case tk_float:
    case tk_double:
    case tk_longdouble:
        out << '\n' << ret << ' ' << name << " = 0.0;";
        break;

    case tk_char:
        out << '\n' << ret << ' ' << name << " = '\\0';";
        break;

    case tk_wchar:
        out << '\n' << ret << ' ' << name << " = L'\\0';";
        break;

    case tk_enum:
        out << '\n' << ret << ' ' << name << " = (" << ret << ")0;";
        break;

    case tk_fixed:
        out << '\n' << ret << ' ' << name << "(\"0\");";
        break;

    case tk_string:
        out << '\n' << ret << ' ' << name << " = CORBA::string_dup(\"\");";
        break;

    case tk_wstring:
        out << '\n' << ret << ' ' << name << " = CORBA::wstring_dup(L\"\");";
        break;

    case tk_any:
        out << '\n' << ret << ' ' << name << " = new " << norm << ';';
        break;

    case tk_TypeCode:
    case tk_Principal:
    case tk_objref:
    case tk_abstract_interface:
        out << '\n' << ret << ' ' << name << " = " << norm << "::_nil();";
        break;

    case tk_struct:
    case tk_union:
        if(variable)
            out << '\n' << ret << ' ' << name << " = new " << norm << ';';
        else
            out << '\n' << ret << ' ' << name << ';';
        break;

    case tk_sequence:
        out << '\n' << ret << ' ' << name << " = new " << norm << ';';
        break;

    case tk_array:
    {
        String_var arr = getTypeString("", type, GetTypeArray);
        out << '\n' << ret << ' ' << name << " = " << arr << "_alloc();";
        break;
    }

    case tk_native:
        out << '\n' << ret << ' ' << name << " = (" << norm << ")0;";
        break;

    case tk_value:
    case tk_value_box:
        out << '\n' << ret << ' ' << name << " = 0;";
        break;
    }
}
