// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <Types.h>
#include <GenUtil.h>
#include <GenCPP.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Generate code for the example header file
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusImplH(Container_ptr container, bool all,
                                   IdlPrettyPrint& out)
{
    //
    // Get scope
    //
    String_var scope = getScope(container);

    //
    // Get container contents descriptions
    //
    Container::DescriptionSeq_var seq =
        container -> describe_contents(dk_all, true, -1);

    ULong i, j;

    for(i = 0 ; i < seq -> length() ; i++)
    {
        //
        // Get container
        //
        Container_var cont = Container::_narrow(seq[i].contained_object);

        switch(seq[i].kind)
        {
        case dk_Module:
        {
            //
            // Module
            //
            ModuleDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

            // Don't add this here!!!
            // if(!check(desc -> id))
            //     continue;

            Identifier_var name = fixKwd(desc -> name);

            comment(desc -> id, out);
            out << "\nnamespace " << name;
            out << "\n{";

            assert(!is_nil(cont));
            genCPlusPlusImplH(cont, all, out);

            out.sep();
            out << "\n} // End of namespace " << name;

            break;
        }

        case dk_Interface:
        {
            //
            // Interface
            //
            InterfaceDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            if(desc -> is_abstract)
                continue;

            Identifier_var name = fixKwd(desc -> name);
            ScopedName_var absolute = getAbsolute(desc -> id);

            comment(desc -> id, out);

            out.sep();
            out << "\nclass " << name << "_impl : ";

            unsigned int indent = out.getIndent();
            out.setIndent(out.getPos());

            // XXX is virtual inheritance mandatory if there are base
            // XXX interfaces and all == false?
            if(!noVirtual_)
                out << "virtual ";
            out << "public " << "POA_" << absolute;

            InterfaceDef_var intf = InterfaceDef::_narrow(cont);
            assert(!is_nil(intf));
            IdlBaseInterfaceInfoSeq baseInfo;
            IdlGetBaseInterfaces(intf, baseInfo, false);

            if(!all)
            {
                for(j = 0 ; j < baseInfo.length() ; j++)
                {
                    //
                    // Impl classes are not generated for abstract interfaces
                    //
                    if(!baseInfo[j].is_abstract)
                    {
                        ScopedName_var baseAbsolute =
                            getAbsolute(baseInfo[j].id, "");

                        out << ",\n";

                        if(!noVirtual_)
                            out << "virtual ";

                        out << "public " << baseAbsolute << "_impl";
                    }
                }
            }

            out.setIndent(indent);

            out.start();

            out << '\n' << name << "_impl(const " << name
                << "_impl&);";
            out << "\nvoid operator=(const " << name << "_impl&);";

            out.sep();
            out.dec();
            out << "\npublic:";
            out.inc();
            out.sep();
            out << '\n' << name << "_impl();";
            out << "\n~" << name << "_impl();";

            //
            // Generate attributes and operations for any base abstract
            // interfaces
            //
            if(!all)
            {
                ULong k, l;
                IdlStringSeq ids;

                for(j = 0 ; j < baseInfo.length() ; j++)
                {
                    if(!baseInfo[j].is_abstract)
                        continue;

                    Container::DescriptionSeq_var contents =
                        baseInfo[j].intf -> describe_contents(dk_all,
                                                              false, -1);

                    for(k = 0 ; k < contents -> length() ; k++)
                    {
                        if(contents[k].kind == dk_Operation)
                        {
                            OperationDescription* opDesc;
                            Boolean b = (contents[k].value) >>= opDesc;
                            assert(b);
                            for(l = 0 ; l < ids.length() ; l++)
                                if(strcmp(ids[l], opDesc -> id) == 0)
                                    break;
                            if(l >= ids.length())
                            {
                                genCPlusPlusImplHOperation(absolute,
                                                           opDesc, out);
                                ids.append(opDesc -> id);
                            }
                        }
                        else if(contents[k].kind == dk_Attribute)
                        {
                            AttributeDescription* atDesc;
                            Boolean b = (contents[k].value) >>= atDesc;
                            assert(b);
                            for(l = 0 ; l < ids.length() ; l++)
                                if(strcmp(ids[l], atDesc -> id) == 0)
                                    break;
                            if(l >= ids.length())
                            {
                                genCPlusPlusImplHAttribute(absolute,
                                                           atDesc, out);
                                ids.append(atDesc -> id);
                            }
                        }
                    }
                }
            }

            //
            // Attributes and operations
            //

            Container::DescriptionSeq_var contents =
                cont -> describe_contents(dk_all, !all, -1);

            for(j = 0 ; j < contents -> length() ; j++)
            {
                if(contents[j].kind == dk_Attribute)
                {
                    AttributeDescription* atDesc;
                    Boolean b = (contents[j].value >>= atDesc);
                    assert(b);
                    genCPlusPlusImplHAttribute(absolute, atDesc, out);
                }
                else
                {
                    OperationDescription* opDesc;
                    Boolean b = (contents[j].value >>= opDesc);
                    assert(b);
                    genCPlusPlusImplHOperation(absolute, opDesc, out);
                }
            }

            out.end();
            out << ';';

            break;
        }

        case dk_Value:
        {
            //
            // Valuetype
            //
            ValueDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            ValueDef_var value = ValueDef::_narrow(cont);
            assert(! is_nil(value));

            Identifier_var name = fixKwd(desc -> name);
            ScopedName_var absolute = getAbsolute(desc -> id);
            ValueDef_var baseValue = value -> base_value();
            InterfaceDef_var supportedInterface =
                value -> supported_interface();
            ScopedName_var baseAbsolute;
            InterfaceDef_var baseSupportedInterface;
            if(!is_nil(baseValue))
            {
                baseAbsolute = getAbsolute(desc -> base_value);
                baseSupportedInterface = baseValue -> supported_interface();
            }

            comment(desc -> id, out);

            out.sep();
            out << "\nclass " << absolute << "_impl : ";

            unsigned int indent = out.getIndent();
            out.setIndent(out.getPos());

            if(!is_nil(supportedInterface))
            {
                out << "virtual public " << "POA_" << absolute << ",";
                if(!is_nil(baseValue) && !all)
                    out << "\nvirtual public " << baseAbsolute << "_impl";
                else
                    out << "\nvirtual public "
                        << "PortableServer::ValueRefCountBase";
            }
            else
            {
                out << "virtual public " << "OBV_" << absolute << ",";
                if(!is_nil(baseValue) && !all)
                    out << "\nvirtual public " << baseAbsolute << "_impl";
                else
                    out << "\nvirtual public CORBA::DefaultValueRefCountBase";
            }

            out.setIndent(indent);

            out.start();

            out << '\n' << name << "_impl(const " << name << "_impl&);";
            out << "\nvoid operator=(const " << name << "_impl&);";

            out.sep();
            out.dec();
            out << "\npublic:";
            out.inc();
            out.sep();
            out << '\n' << name << "_impl();";
            out << "\n~" << name << "_impl();";

            //
            // _copy_value
            //
            out.sep();
            out << "\nvirtual CORBA::ValueBase* _copy_value();";

            //
            // Custom valuetype
            //
            if(desc -> is_custom)
            {
                out.sep();
                out << "\nvirtual void marshal(CORBA::DataOutputStream* os);";
                out << "\nvirtual void unmarshal(CORBA::DataInputStream* is);";
            }

            //
            // Valuetype attributes and operations
            //
            ContainedSeq_var attributes =
                value -> contents(dk_Attribute, !all);

            for(j = 0 ; j < attributes -> length() ; j++)
            {
                Contained::Description_var contDesc =
                    attributes[j] -> describe();
                AttributeDescription* atDesc;
                Boolean b = (contDesc -> value) >>= atDesc;
                assert(b);
                genCPlusPlusImplHAttribute(absolute, atDesc, out);
            }

            ContainedSeq_var operations =
                value -> contents(dk_Operation, !all);

            for(j = 0 ; j < operations -> length() ; j++)
            {
                Contained::Description_var contDesc =
                    operations[j] -> describe();
                OperationDescription* opDesc;
                Boolean b = (contDesc -> value) >>= opDesc;
                assert(b);
                genCPlusPlusImplHOperation(absolute, opDesc, out);
            }

            //
            // Supported interface attributes and operations
            //
            if(!all && !is_nil(supportedInterface))
            {
                attributes =
                    supportedInterface -> contents(dk_Attribute, !all);

                for(j = 0 ; j < attributes -> length() ; j++)
                {
                    Contained::Description_var contDesc =
                        attributes[j] -> describe();
                    AttributeDescription* atDesc;
                    Boolean b = (contDesc -> value) >>= atDesc;
                    assert(b);
                    genCPlusPlusImplHAttribute(absolute, atDesc, out);
                }

                operations =
                    supportedInterface -> contents(dk_Operation, !all);

                for(j = 0 ; j < operations -> length() ; j++)
                {
                    Contained::Description_var contDesc =
                        operations[j] -> describe();
                    OperationDescription* opDesc;
                    Boolean b = (contDesc -> value) >>= opDesc;
                    assert(b);
                    genCPlusPlusImplHOperation(absolute, opDesc, out);
                }
            }

            out.end();
            out << ';';

            break;
        }

        default:
            break;
        }
    }
}

// ----------------------------------------------------------------------
// Generate header code for an example attribute
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusImplHAttribute(const char* absolute,
                                            AttributeDescription* desc,
                                            IdlPrettyPrint& out)
{
    if(!check(desc -> id))
        return;

    Identifier_var name = fixKwd(desc -> name);

    comment(desc -> id, out);

    String_var ret = getTypeString("", desc -> type, GetTypeReturn);

    out.sep();
    out << "\nvirtual " << ret << ' ' << name << "()";
    out.inc();
    out << "\nthrow(CORBA::SystemException);";
    out.dec();

    if(desc -> mode == ATTR_NORMAL)
    {
        String_var in = getTypeString("", desc -> type, GetTypeIn);
        out << "\nvirtual void " << name << '(' << in << ')';
        out.inc();
        out << "\nthrow(CORBA::SystemException);";
        out.dec();
    }
}

// ----------------------------------------------------------------------
// Generate header code for an example operation
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusImplHOperation(const char* absolute,
                                            OperationDescription* desc,
                                            IdlPrettyPrint& out)
{
    if(!check(desc -> id))
        return;

    Identifier_var name = fixKwd(desc -> name);

    comment(desc -> id, out);

    String_var ret = getTypeString("", desc -> result, GetTypeReturn);

    out.sep();
    out << "\nvirtual " << ret << ' ' << name << "(";

    unsigned int indent = out.getIndent();
    out.setIndent(out.getPos());

    ULong i;
    for(i = 0 ; i < desc -> parameters.length() ; i++)
    {
        GetType tt;
        switch(desc -> parameters[i].mode)
        {
        case PARAM_IN:
            tt = GetTypeIn;
            break;

        case PARAM_INOUT:
            tt = GetTypeInOut;
            break;

        case PARAM_OUT:
            tt = GetTypeOut;
            break;
        }

        String_var argName = fixKwd(desc -> parameters[i].name);

        String_var s =
            getTypeString("", desc -> parameters[i].type, tt, false, argName);

        if(i > 0)
            out << ",\n";

        out << s;
    }

    if(desc -> contexts.length() > 0)
    {
        if(desc -> parameters.length() > 0)
            out << ",\n";

        out << "CORBA::Context_ptr";
    }

    out << ")";

    out.setIndent(indent);

    //
    // Servants must declare throw() clauses
    //

    out.inc();
    out << "\nthrow(";

    indent = out.getIndent();
    out.setIndent(out.getPos());

    for(i = 0 ; i < desc -> exceptions.length() ; i++)
    {
        Identifier_var exName = getAbsolute(desc -> exceptions[i].id);
        out << exName << ",\n";
    }

    out << "CORBA::SystemException);";
    out.setIndent(indent);
    out.dec();
}
