// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <Types.h>
#include <GenUtil.h>
#include <GenCPP.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Generate code for the skeleton source file, general part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusSkelC(Container_ptr container,
				   IdlPrettyPrint& out)
{
    //
    // Get scope
    //
    String_var scope = getScope(container);

    //
    // Get container contents descriptions
    //
    Container::DescriptionSeq_var seq =
	container -> describe_contents(dk_all, true, -1);

    ULong i, j;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get container
	//
	Container_var cont = Container::_narrow(seq[i].contained_object);

	switch(seq[i].kind)
	{
	case dk_Module:
	{
	    //
	    // Module
	    //
	    ModuleDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    // Don't add this here!!!
	    // if(!check(desc -> id))
	    //     continue;

 	    comment(desc -> id, out);

	    out.sep();

	    assert(!is_nil(cont));
	    genCPlusPlusSkelC(cont, out);

	    break;
	}

	case dk_Interface:
	{
	    assert(!is_nil(cont));
	    genCPlusPlusSkelCProtected(cont, out);

	    //
	    // Interface
	    //
	    InterfaceDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

	    if(!check(desc -> id))
		continue;

	    InterfaceDef_var interf =
		InterfaceDef::_narrow(seq[i].contained_object);
	    assert(!is_nil(interf));

            ScopedName_var absolute = getAbsolute(desc -> id);
            Identifier_var name = fixKwd(desc -> name);

            ScopedName_var objAbs, classAbs;
            if(desc -> is_abstract)
            {
                objAbs = scope;
                objAbs += "OBAbstractImpl_";
                objAbs += name;
                classAbs = getPrefixedAbsolute("POA", interf, "",
                                               "OBAbstract");
            }
            else
            {
                objAbs = absolute;
                classAbs = getPrefixedAbsolute("POA", desc -> id, "");
            }


	    comment(desc -> id, out);

            if(!desc -> is_abstract)
            {
		//
		// _OB_dispatch()
		//
		unsigned int indent;
		indent = out.getIndent();
		out.sep();
		out << "\nOB::DispatchStatus\n" << classAbs
		    << "::_OB_dispatch(";
		out.setIndent(out.getPos());
		out << "const char* _ob_op,";
		out << "\nOB::Buffer& _ob_buf,";
		out << "\nbool _ob_sw,";
		out << "\nCORBA::ULong _ob_offIn,";
		out << "\nCORBA::ULong _ob_offOut)";
		out.setIndent(indent);
		out.start();

		//
		// Method name table for operations
		//
		ContainedSeq_var operations =
		    interf -> contents(dk_Operation, false);
		ContainedSeq_var attributes =
		    interf -> contents(dk_Attribute, false);
		IdlStringSeq_var names =
		    IdlGetMethodNames(attributes, operations);
		
		if(names -> length() > 0)
		{
		    //
		    // Emit the operation table
		    //
		    out << "\nstatic const char* _ob_names[] =";
		    out.start();
		    for(j = 0; j < names -> length(); j++)
		    {
			if(j != 0)
			    out << ",";
			out << "\n\"" << names[j] << "\"";
		    }
		    out.end();
		    out << ';';
		    out << "\nstatic const CORBA::ULong _ob_numNames = "
			<< names -> length() << ";";

		    //
		    // Emit the giant switch statement
		    //
		    out.sep();
		    out << "\nswitch(_OB_findOperation(_ob_op, _ob_names, "
			<< "_ob_numNames))";
		    out.start();
		    out.dec();
		    
		    for(j = 0 ; j < names -> length(); j++)
		    {
			if(j > 0)
			    out.sep();
			out << "\ncase " << j << ": // " << names[j];
			out.inc();
			int len = strlen(names[j]);
			out << "\nreturn _OB";
			if(!(len > 5 && (strncmp(names[j], "_get_", 5) == 0 ||
					 strncmp(names[j], "_set_", 5) == 0)))
			    out << "_op_";
			out << names[j]
			    << "(_ob_buf, _ob_sw, _ob_offIn, _ob_offOut);";
			out.dec();
		    }

		    out.inc();
		    out.end();
		}
		
		if(names -> length() > 0)
		    out.sep();

		out << "\nreturn _OB_dispatchBase("
		    << "_ob_op, _ob_buf, _ob_sw, _ob_offIn, _ob_offOut);";
		out.end();

		//
		// _this()
		//
		out.sep();
		out << '\n' << absolute << "_ptr\n" << classAbs << "::_this()";
		out.start();
		out << "\nCORBA::Object_var obj = _OB_createReference();";
		out << '\n' << absolute << "_var result = " << absolute
		    << "::_narrow(obj);";
		out << "\nreturn result._retn();";
		out.end();

		//
		// _primary_interface
		//
		indent = out.getIndent();
		out.sep();
		out << "\nCORBA::RepositoryId" << '\n' << classAbs
		    << "::_primary_interface(";
		out.setIndent(out.getPos());
		out << "const PortableServer::ObjectId&,";
		out << "\nPortableServer::POA_ptr)";
		out.setIndent(indent);
		out.start();
		out << "\nreturn CORBA::string_dup(" << absolute
                    << "::_OB_staticIds()[0]);";
		out.end();

		//
		// _OB_createStubImpl
		//
		out.sep();
		out << "\nOB::StubImplBase*" << '\n' << classAbs
		    << "::_OB_createStubImpl(";
		indent = out.getIndent();
		out.setIndent(out.getPos());
		out << "PortableServer::POA_ptr poa,";
		out << "\nconst PortableServer::ObjectId& oid,";
		out << "\nconst OB::OctetSeq& key)";
		out.setIndent(indent);
		out.start();
		out << "\nreturn new OBLocalStubImpl_" << absolute <<
		    "(poa, oid, key, this);";
		out.end();
	    }
            
	    //
	    // _is_a
	    //
	    out.sep();
	    out << "\nCORBA::Boolean" << '\n' << classAbs
		<< "::_is_a(const char* type)";
            out.inc();
            out << "\nthrow(CORBA::SystemException)";
            out.dec();
	    out.start();
            out << "\nconst char** _ob_ids = " << objAbs
                << "::_OB_staticIds();";
            out << "\nfor(CORBA::ULong _ob_i = 0 ; _ob_ids[_ob_i] != 0 ; "
                << "_ob_i++)";
            out.inc();
	    out << "\nif(strcmp(type, _ob_ids[_ob_i]) == 0)";
	    out.inc();
	    out << "\nreturn true;";
	    out.dec();
	    out.dec();
	    out << "\nreturn false;";
	    out.end();
	    break;
	}

	default:
	    break;
	}
    }
}

// ----------------------------------------------------------------------
// Generate code for the skeleton source file, general part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusSkelCStubImpl(Container_ptr container,
                                           IdlPrettyPrint& out)
{
    //
    // Get the equivalent StubImpl name for this container
    //
    Contained_var contained = Contained::_narrow(container);
    ScopedName_var className;
    if(!is_nil(contained))
    {
        ScopedName_var contAbsolute = getAbsolute(contained);
        className = string_dup("OBLocalStubImpl_");
        className += string_dup(contAbsolute);
    }
    
    //
    // Get scope
    //
    String_var scope = getScope(container);

    //
    // Get container contents descriptions
    //
    Container::DescriptionSeq_var seq =
	container -> describe_contents(dk_all, true, -1);

    ULong i, j;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get container
	//
	Container_var cont = Container::_narrow(seq[i].contained_object);

	switch(seq[i].kind)
	{
	case dk_Module:
	{
	    //
	    // Module
	    //
	    ModuleDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    // Don't add this here!!!
	    // if(!check(desc -> id))
	    //     continue;

	    assert(!is_nil(cont));
	    genCPlusPlusSkelCStubImpl(cont, out);

	    break;
	}

        case dk_Interface:
        {
            //
            // Interface
            //
            InterfaceDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);
            
            if(!check(desc -> id))
                continue;

            Identifier_var name = fixKwd(desc -> name);
            ScopedName_var absolute = getAbsolute(desc -> id);

            //
            // We have to use these mangled names to avoid a bug
            // in MSVC
            //
            String_var flat = getFlattenedName(absolute);
            String_var servant = string_dup("_ob_servant_");
            servant += flat;
            servant += "_";

            className = string_dup("OBLocalStubImpl_");
            className += string_dup(absolute);

	    out.sep();
	    out << "\n//";
	    out << "\n// Prepare the servant for a method invocation";
	    out << "\n//";
	    out << "\nPortableServer::ServantBase*";
	    out << '\n' << className << "::_OB_setupServant()";
	    out.start();
	    unsigned int indent = out.getIndent();
	    out.setIndent(0);
	    out << "\n#ifdef HAVE_JTC";
	    out.setIndent(indent);
	    out << "\nJTCSynchronized sync(deactivatedMutex_);";
	    out.setIndent(0);
	    out << "\n#endif";
	    out.setIndent(indent);
	    out << "\nif(deactivated_)";
	    out.inc();
	    out << "\nthrow OB::StubImplChanged();";
	    out.dec();
	    out << '\n' << servant << " -> _add_ref();";
	    out << "\nreturn " << servant << ";";
	    out.end();

            if(desc -> is_abstract)
	    {
		assert(!is_nil(cont));
		genCPlusPlusSkelCStubImpl(cont, out);
                continue;
	    }

            InterfaceDef_var intf =
                InterfaceDef::_narrow(seq[i].contained_object);
            String_var constrName = getPrefixedName("OBLocalStubImpl", intf);

            comment(desc -> id, out);

	    out << '\n' << className << "::" << constrName << "(";
	    out.inc();
	    out << "\nPortableServer::POA_ptr poa,";
	    out << "\nconst PortableServer::ObjectId& oid,";
	    out << "\nconst OB::OctetSeq& key,";
	    out << "\nPOA_" << absolute << "* servant)";
	    out.dec();
	    out.start();
	    out << "\n_ob_poa_ = PortableServer::POA::_duplicate(poa);";
	    out << "\n_ob_oid_ = oid;";
	    out << "\n_ob_poa_key_ = key;";
	    out.sep();
	    out << '\n' << servant << " = servant;";

	    InterfaceDefSeq allBaseInterfaces;
	    IdlGetAllBaseInterfaces(intf, allBaseInterfaces);
	    
	    out << "\n_ob_servant_ = servant;";

	    for(j = 0; j < allBaseInterfaces.length(); j++)
	    {
		RepositoryId_var id = allBaseInterfaces[j] -> id();
		ScopedName_var baseAbsolute = getAbsolute(id);
                //
                // We have to use these mangled names to avoid a bug
                // in MSVC
                //
                String_var baseFlat = getFlattenedName(baseAbsolute);
                String_var baseServant = string_dup("_ob_servant_");
                baseServant += baseFlat;
                baseServant += "_";
		out << '\n' << baseServant << " = servant;";
	    }

	    out.end();

	    assert(!is_nil(cont));
	    genCPlusPlusSkelCStubImpl(cont, out);

            break;
        }

        case dk_Attribute:
	{
	    //
	    // Attribute
	    //
	    AttributeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);
            
	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);
	    ScopedName_var absolute = getAbsolute(desc -> id);

            InterfaceDef_var intf = InterfaceDef::_narrow(container);
            assert(!is_nil(intf));
            RepositoryId_var id = intf -> id();
            ScopedName_var intfAbs = getAbsolute(id);
            //
            // We have to use these mangled names to avoid a bug
            // in MSVC
            //
            String_var flat = getFlattenedName(intfAbs);
            String_var servant = string_dup("_ob_servant_");
            servant += flat;
            servant += "_";

	    comment(desc -> id, out);

	    String_var ret = getTypeString("", desc -> type,
					   GetTypeReturn, false);
	    out.sep();
	    out << '\n' << ret << '\n' << className << "::" << name << "()";
            out.start();

	    out << "\nPortableServer::ServantBase_var _ob_dummy_ = "
		<< "_OB_setupServant();";
	    out << "\nOB::InvocationHandler _ob_handler_(";
	    out.inc();
	    out << '\n' << servant << ", _ob_poa_, "
		<< '"' << name << '"' << ", _ob_oid_, _ob_poa_key_);";
            out << "\nreturn " << servant << " -> " << name << "();";
	    out.dec();

            out.end();
                
	    if(desc -> mode == ATTR_NORMAL)
	    {
                String_var in = getTypeString("", desc -> type,
					      GetTypeIn, false);
                out.sep();
                out << "\nvoid\n" << className << "::" << name << '('
                    << in << " _ob_a)";
                out.start();

		out << "\nPortableServer::ServantBase_var _ob_dummy_ = "
		    << "_OB_setupServant();";
		out << "\nOB::InvocationHandler _ob_handler_(";
		out.inc();
		out << '\n' << servant << ", _ob_poa_, "
		    << '"' << name << '"' << ", _ob_oid_, _ob_poa_key_);";
		out.dec();

                TypeCode_var tc = GetOrigType(desc -> type);

                //
                // Valuetype arguments to collocated attributes and operations
                // must be copied to preserve location transparency
                //
                // XXX More logic needed for abstract interfaces
                //
                String_var arg;
		if(tc -> kind() == tk_value || tc -> kind() == tk_value_box)
                {
                    arg = string_dup("_ob_v");
                    String_var varType =
		        getTypeString("", desc -> type, GetTypeVar);
                    const char* valueId = desc -> type -> id();
                    String_var valueAbs = getTypeString("", desc -> type);
                    out << '\n' << varType << " _ob_v;";
                    out << "\nif(_ob_a != 0)";
                    out.start();
                    out << "\nCORBA::ValueBase_var _ob_vb = "
                        << "CORBA::ValueFactoryBase::_OB_copyValue(_ob_a, "
                        << valueAbs << "::_OB_id());";
                    if(strcmp(valueId, "IDL:omg.org/CORBA/ValueBase:1.0") != 0)
                    {
                        out << "\n_ob_v = " << valueAbs
                            << "::_downcast(_ob_vb._retn());";
                    }
                    else
                        out << "\n_ob_v = _ob_vb._retn();";
                    out.end();
                }
                else
                {
                    arg = string_dup("_ob_a");
                }

                out << '\n' << servant << " -> " << name << '(' << arg << ");";
		out.end();
	    }
		
	    break;
        }

	case dk_Operation:
	{
	    //
	    // Operation
	    //
	    OperationDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);
	    ScopedName_var absolute = getAbsolute(desc -> id);

            InterfaceDef_var intf = InterfaceDef::_narrow(container);
            assert(!is_nil(intf));
            RepositoryId_var id = intf -> id();
            ScopedName_var intfAbs = getAbsolute(id);
            //
            // We have to use these mangled names to avoid a bug
            // in MSVC
            //
            String_var flat = getFlattenedName(intfAbs);
            String_var servant = string_dup("_ob_servant_");
            servant += flat;
            servant += "_";

	    comment(desc -> id, out);

	    String_var ret =
		getTypeString("", desc -> result, GetTypeReturn);
	    
	    out.sep();
	    out << '\n' << ret << '\n' << className << "::" << name << "(";
	    
	    unsigned int indent = out.getIndent();
	    out.setIndent(out.getPos());

	    for(j = 0 ; j < desc -> parameters.length() ; j++)
	    {
		GetType tt;
		switch(desc -> parameters[j].mode)
		{
		case PARAM_IN:
		    tt = GetTypeIn;
		    break;

		case PARAM_INOUT:
		    tt = GetTypeInOut;
		    break;

		case PARAM_OUT:
		    tt = GetTypeOut;
		    break;
		}

		String_var arg = string_dup("_ob_a");
		arg += j;

		String_var str =
		    getTypeString("", desc -> parameters[j].type, tt,
				  false, arg);

		if(j > 0)
		    out << ",\n";

		out << str;
	    }

	    if(desc -> contexts.length() > 0)
	    {
		if(desc -> parameters.length() > 0)
		    out << ",\n";
		
		out << "CORBA::Context_ptr _ob_ctx";
	    }

	    out << ")";

	    out.setIndent(indent);
            
            out.start();

	    out << "\nPortableServer::ServantBase_var _ob_dummy_ = "
		<< "_OB_setupServant();";
	    out << "\nOB::InvocationHandler _ob_handler_(";
	    out.inc();
	    out << '\n' << servant << ", _ob_poa_, "
		<< '"' << name << '"' << ", _ob_oid_, _ob_poa_key_);";
	    out.dec();

            bool emitReturn = false;
            if(desc -> mode != OP_ONEWAY)
            {
                TypeCode_var origRet = GetOrigType(desc -> result);
                if(origRet -> kind() != tk_void)
                    emitReturn = true;
            }
            //
            // 'in' valuetype parameters must be copied to preserve
            // location transparency
            //
            OB::StringSeq argNames;
            argNames.length(desc -> parameters.length());

	    for(j = 0 ; j < desc -> parameters.length() ; j++)
	    {
                TypeCode_var tc =
                    GetOrigType(desc -> parameters[j].type);

                //
                // Valuetype arguments to collocated attributes and operations
                // must be copied to preserve location transparency
                //
                // XXX More logic needed for abstract interfaces
                //
		if(desc -> parameters[j].mode == PARAM_IN &&
                   (tc -> kind() == tk_value || tc -> kind() == tk_value_box))
                {
		    String_var arg = string_dup("_ob_v");
		    arg += j;
                    argNames[j] = arg;
                    String_var varType =
		        getTypeString("", desc -> parameters[j].type,
                                      GetTypeVar);
                    const char* valueId = desc -> parameters[j].type -> id();
                    String_var valueAbs =
                        getTypeString("", desc -> parameters[j].type);
                    out << '\n' << varType << ' ' << arg << ';';
                    out << "\nif(_ob_a" << j << " != 0)";
                    out.start();
                    out << "\nCORBA::ValueBase_var _ob_v = "
                        << "CORBA::ValueFactoryBase::_OB_copyValue(_ob_a" << j
                        << ", " << valueAbs << "::_OB_id());";
                    if(strcmp(valueId, "IDL:omg.org/CORBA/ValueBase:1.0") != 0)
                    {
                        ScopedName_var valueAbs = getAbsolute(valueId);
                        out << "\n_ob_v" << j << " = " << valueAbs
                            << "::_downcast(_ob_v._retn());";
                    }
                    else
                        out << "\n_ob_v" << j << " = _ob_v._retn();";
                    out.end();
                }
                else
                {
		    String_var arg = string_dup("_ob_a");
		    arg += j;
                    argNames[j] = arg;
                }
	    }

            out << '\n';
            if(emitReturn)
                out << "return ";
            out << servant << " -> " << name << "(";
            for(j = 0 ; j < desc -> parameters.length() ; j++)
            {
                if(j > 0)
                    out << ", ";
                out << argNames[j];
            }
            if(desc -> contexts.length() > 0)
            {
                if(desc -> parameters.length() > 0)
                    out << ", ";
                out << "_ob_ctx";
            }
            out << ");";
            out.end();
	    break;
        }

        default:
	    break;
	}
    }
}

// ----------------------------------------------------------------------
// Generate code for the skeleton source file, protected part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusSkelCProtected(Container_ptr container,
					    IdlPrettyPrint& out)
{
    //
    // Get scope
    //
    String_var scope = getScope(container);

    InterfaceDef_var interf = InterfaceDef::_narrow(container);
    assert(!is_nil(interf));

    ScopedName_var classAbs;
    if(interf -> is_abstract())
        classAbs = getPrefixedAbsolute("POA", interf, "", "OBAbstract");
    else
        classAbs = getPrefixedAbsolute("POA", interf, "");

    //
    // Get container contents descriptions
    //
    Container::DescriptionSeq_var seq =
	container -> describe_contents(dk_all, true, -1);

    ULong i, j;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	switch(seq[i].kind)
	{
	case dk_Attribute:
	{
	    //
	    // Attribute
	    //
	    AttributeDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);

	    comment(desc -> id, out);

	    //
	    // Don't use name here instead of desc -> name. We need
	    // the name "unfixed".
	    //
	    ULong iter = 0;
	    out.sep();
            unsigned int indent = out.getIndent();
	    out << "\nOB::DispatchStatus\n" << classAbs << "::_OB_get_"
		<< desc -> name << '(';
            out.setIndent(out.getPos());
            out << "OB::Buffer& _ob_buf,";
            out << "\nbool _ob_sw,";
            out << "\nCORBA::ULong _ob_offIn,";
            out << "\nCORBA::ULong _ob_offOut)";
            out.setIndent(indent);
            out.start();
	    TypeCode_var origType = GetOrigType(desc -> type);
	    bool useVar = IdlIsVariable(origType) ||
		origType -> kind() == tk_array;
	    String_var s;
	    if(useVar)
		s = getTypeString("", desc -> type, GetTypeVar);
	    else
		s = getTypeString("", desc -> type, GetTypeNormal);
	    out << '\n' << s << " _ob_r = " << name << "();";
	    out.sep();
	    out << "\nOB::CountStream _ob_cnt(_ob_offOut);";
	    if(useVar)
		writeReadWriteCount("", desc -> type, "_ob_r.in()", iter,
				    RWCCountIn, 0, out);
	    else
		writeReadWriteCount("", desc -> type, "_ob_r", iter,
				    RWCCountIn, 0, out);
	    out.sep();
	    out << "\n_ob_buf.alloc(_ob_cnt.count());";
	    out << "\nOB::OutputStream _ob_out(_ob_buf, _ob_offOut);";
	    if(useVar)
		writeReadWriteCount("", desc -> type, "_ob_r.in()", iter,
				    RWCWriteIn, 0, out);
	    else
		writeReadWriteCount("", desc -> type, "_ob_r", iter,
				    RWCWriteIn, 0, out);
	    out.sep();
	    out << "\nreturn OB::DispatchStatusOK;";
	    out.end();

	    if(desc -> mode == ATTR_NORMAL)
	    {
		//
		// Don't use name here instead of desc -> name. We need
		// the name "unfixed".
		//
		/*ULong*/ iter = 0;
		out.sep();
		unsigned int indent = out.getIndent();
		out << "\nOB::DispatchStatus\n" << classAbs << "::_OB_set_"
		    << desc -> name << '(';
		out.setIndent(out.getPos());
		out << "OB::Buffer& _ob_buf,";
		out << "\nbool _ob_sw,";
		out << "\nCORBA::ULong _ob_offIn,";
		out << "\nCORBA::ULong _ob_offOut)";
		out.setIndent(indent);
		out.start();
		out << "\nOB::InputStream _ob_in(_ob_buf, _ob_offIn, _ob_sw);";
		// TODO: This shouldn't just use GetTypeStruct
		String_var s = getTypeString("", desc -> type,
						   GetTypeStruct);
		out << '\n' << s << " _ob_a;";
		writeReadWriteCount("", desc -> type, "_ob_a", iter,
				    RWCReadStructElem, 0, out);
		out.sep();
		if(origType -> kind() == tk_array)
		{
		    out.direct("\n#ifdef HAVE_VCPLUSPLUS_BUGS");
		    String_var conv =
			getTypeString("", desc -> type, GetTypeSlice);
		    out << '\n' << name << "((const " << conv << "*)_ob_a);";
		    out.direct("\n#else");
		}
// 		if(origType -> kind() == tk_string ||
// 		   origType -> kind() == tk_wstring ||
// 		   origType -> kind() == tk_objref ||
// 		   origType -> kind() == tk_TypeCode ||
// 		   origType -> kind() == tk_Principal ||
// 		   origType -> kind() == tk_abstract_interface)
// 		    out << '\n' << name << "(_ob_a.in());";
// 		else
		    out << '\n' << name << "(_ob_a);";
		if(origType -> kind() == tk_array)
		    out.direct("\n#endif"); // HAVE_VCPLUSPLUS_BUGS
		out.sep();
		out << "\n_ob_buf.alloc(_ob_offOut);";
		out.sep();
		out << "\nreturn OB::DispatchStatusOK;";
		out.end();
	    }
            
            break;
        }
        
	case dk_Operation:
	{
	    //
	    // Operation
	    //
	    OperationDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);

	    comment(desc -> id, out);

	    //
	    // Don't use name here instead of desc -> name. We need
	    // the name "unfixed".
	    //
	    ULong iter = 0;
	    out.sep();
            unsigned int indent = out.getIndent();
	    out << "\nOB::DispatchStatus\n" << classAbs << "::_OB_op_"
		<< desc -> name << '(';
            out.setIndent(out.getPos());
            out << "OB::Buffer& _ob_buf,";
            out << "\nbool _ob_sw,";
            out << "\nCORBA::ULong _ob_offIn,";
            out << "\nCORBA::ULong _ob_offOut)";
            out.setIndent(indent);
            out.start();
	    
	    bool needUnmarshal = false;
	    for(j = 0 ; j < desc -> parameters.length() ; j++)
	    {
		if(desc -> parameters[j].mode != PARAM_OUT)
		{
		    needUnmarshal = true;
		    break;
		}
	    }
	    if(desc -> contexts.length() > 0)
		needUnmarshal = true;
	    
	    if(needUnmarshal)
		out << "\nOB::InputStream _ob_in(_ob_buf, _ob_offIn, _ob_sw);";
	    
	    for(j = 0 ; j < desc -> parameters.length() ; j++)
	    {
		TypeCode_ptr parType = desc -> parameters[j].type;

		if(desc -> parameters[j].mode != PARAM_OUT)
		{
		    // TODO: This shouldn't just use GetTypeStruct
		    String_var s =
			getTypeString("", parType, GetTypeStruct);
		    out << '\n' << s << " _ob_a" << j << ';';
		    String_var arg = string_dup("_ob_a");
		    arg += j;
		    writeReadWriteCount("", parType, arg, iter,
					RWCReadStructElem, 0, out);
		}
		else
		{
		    String_var s;
		    if(IdlIsVariable(parType))
			s = getTypeString("", parType, GetTypeVar);
		    else
			s = getTypeString("", parType, GetTypeNormal);
		    out << '\n' << s << " _ob_a" << j << ';';
		}
	    }
	    if(desc -> contexts.length() > 0)
	    {
		out << "\nOB::StringSeq_var _ob_ctxSeq = new OB::StringSeq;";
                //
                // No typecode for OB::StringSeq
                //
                TypeCode_var ctxType =
                    orb_ -> create_sequence_tc(0, _tc_string);
                writeReadWriteCount("", ctxType, "_ob_ctxSeq.inout()", iter,
                                    RWCReadStructElem, 0, out);
		out << "\nCORBA::Context_var _ob_ctx = "
		    << "CORBA::Context::_OB_create(_ob_ctxSeq._retn());";
	    }	    

	    if(desc -> parameters.length() > 0 ||
	       desc -> contexts.length() > 0)
		out.sep();
	    
	    if(desc -> exceptions.length() > 0)
	    {
		out << "\ntry";
		out.start();
	    }
	    
	    TypeCode_var origRet = GetOrigType(desc -> result);
	    
	    int extra = 0;
	    for(j = 0 ; j < desc -> parameters.length() ; j++)
	    {
		if(desc -> parameters[j].mode == PARAM_IN)
		{
		    TypeCode_var origPar = desc -> parameters[j].type;
		    origPar = GetOrigType(origPar);
		    if(origPar -> kind() == tk_array)
			extra = 2;
		}
	    }
	    
	    do
	    {
		if(extra == 2)
		    out.direct("\n#ifdef HAVE_VCPLUSPLUS_BUGS");
		if(origRet -> kind() != tk_void)
		{
		    String_var s;
		    if(IdlIsVariable(origRet) ||
		       origRet -> kind() == tk_array)		       
			s = getTypeString("", desc -> result, GetTypeVar);
		    else
			s = getTypeString("", desc -> result, GetTypeNormal);
		    out << '\n' << s << " _ob_r = ";
		}
		else
		    out << '\n';
		out << name << '(';
		for(j = 0 ; j < desc -> parameters.length() ; j++)
		{
		    if(j != 0)
			out << ", ";
		    
		    TypeCode_var origParType =
			desc -> parameters[j].type;
		    origParType = GetOrigType(origParType);
		    
		    if(extra == 2 &&
		       desc -> parameters[j].mode == PARAM_IN)
		    {
			if(origParType -> kind() == tk_array)
			{
			    String_var conv =
				getTypeString("", desc -> parameters[j].type,
					      GetTypeSlice);
			    out << "(const " << conv << "*)";
			}
		    }
		    
		    switch(desc -> parameters[j].mode)
		    {
		    case PARAM_IN:
// 			if(origParType -> kind() == tk_string ||
// 			   origParType -> kind() == tk_wstring ||
// 			   origParType -> kind() == tk_objref ||
// 			   origParType -> kind() == tk_TypeCode||
// 			   origParType -> kind() == tk_Principal)
// 			   origParType -> kind() == tk_abstract_interface)
// 			    out << "_ob_a" << j << ".in()";
// 			else
			    out << "_ob_a" << j;
			break;
			
		    case PARAM_INOUT:
			if(origParType -> kind() == tk_string ||
			   origParType -> kind() == tk_wstring ||
			   origParType -> kind() == tk_objref ||
			   origParType -> kind() == tk_TypeCode ||
			   origParType -> kind() == tk_Principal ||
			   origParType -> kind() == tk_abstract_interface)
			    out << "_ob_a" << j << ".inout()";
			else
			    out << "_ob_a" << j;
			break;
			
		    case PARAM_OUT:
			if(IdlIsVariable(origParType))
			    out << "_ob_a" << j << ".out()";
			else
			    out << "_ob_a" << j;
			break;
		    }
		}
		if(desc -> contexts.length() > 0)
		{
		    if(desc -> parameters.length() > 0)
			out << ", ";
		    
		    out << "_ob_ctx.in()";
		}
		out << ");";
		if(extra == 2)
		    out.direct("\n#else");
		extra--;
	    }
	    while(extra > 0);
	    
	    if(extra == 0)
	    {
		out.direct("\n#endif");
		// HAVE_VCPLUSPLUS_BUGS
	    }
                        
	    out.sep();
	    
	    // TODO: This should not be emitted in case there is nothing to
	    // unmarshal
	    out << "\nOB::CountStream _ob_cnt(_ob_offOut);";
	    
	    bool needMarshal = false;
	    if(origRet -> kind() != tk_void)
	    {
		if(IdlIsVariable(origRet) ||
		   origRet -> kind() == tk_array)
		    writeReadWriteCount("", desc -> result, "_ob_r.in()", iter,
					RWCCountIn, 0, out);
		else
		    writeReadWriteCount("", desc -> result, "_ob_r", iter,
					RWCCountIn, 0, out);
		needMarshal = true;
	    }
	    
	    for(j = 0 ; j < desc -> parameters.length() ; j++)
		if(desc -> parameters[j].mode != PARAM_IN)
		{
		    TypeCode_ptr parType = desc -> parameters[j].type;
		    String_var arg = string_dup("_ob_a");
		    arg += j;
		    if(desc -> parameters[j].mode != PARAM_OUT)
		    {
			writeReadWriteCount("", parType, arg, iter, RWCCountIn,
					    0, out);
		    }
		    else
		    {
			if(IdlIsVariable(parType))
			    arg += ".in()";
			writeReadWriteCount("", parType, arg, iter, RWCCountIn,
					    0, out);
		    }
		    
		    needMarshal = true;
		}
	    
	    // TODO: This should not be emitted in case there is nothing to
	    // unmarshal
	    out.sep();
	    out << "\n_ob_buf.alloc(_ob_cnt.count());";
	    
	    if(needMarshal)
	    {
		out << "\nOB::OutputStream _ob_out(_ob_buf, _ob_offOut);";
		
		if(origRet -> kind() != tk_void)
		{
		    if(IdlIsVariable(origRet) ||
		       origRet -> kind() == tk_array)
			writeReadWriteCount("", desc -> result, "_ob_r.in()",
					    iter, RWCWriteIn, 0, out);
		    else
			writeReadWriteCount("", desc -> result, "_ob_r", iter,
					    RWCWriteIn, 0, out);
		}
		
		for(j = 0 ; j < desc -> parameters.length() ; j++)
		    if(desc -> parameters[j].mode != PARAM_IN)
		    {
			TypeCode_ptr parType =
			    desc -> parameters[j].type;
			String_var arg = string_dup("_ob_a");
			arg += j;
			if(desc -> parameters[j].mode != PARAM_OUT)
			{
			    writeReadWriteCount("", parType, arg, iter,
						RWCWriteIn, 0, out);
			}
			else
			{
			    if(IdlIsVariable(parType))
				arg += ".in()";
			    writeReadWriteCount("", parType, arg, iter,
						RWCWriteIn, 0, out);
			}
		    }
	    }
	    
	    out.sep();
	    out << "\nreturn OB::DispatchStatusOK;";
	    
	    if(desc -> exceptions.length() > 0)
	    {
		out.end();
		for(j = 0 ; j < desc -> exceptions.length() ; j++)
		{
		    TypeCode_ptr exType = desc -> exceptions[j].type;
		    String_var s = getTypeString("", exType, GetTypeNormal);
		    out << "\ncatch(const " << s << "& _ob_except)";
		    out.start();
		    out << "\nOB::CountStream _ob_cnt(_ob_offOut);";
		    out << '\n' << s
                        << "::_OB_marshalCount(_ob_except, _ob_cnt);";
		    out.sep();
		    out << "\n_ob_buf.alloc(_ob_cnt.count());";
		    out << "\nOB::OutputStream _ob_out(_ob_buf, _ob_offOut);";
		    out << '\n' << s
		        << "::_OB_marshal(_ob_except, _ob_out);";
		    out.sep();
		    out << "\nreturn OB::DispatchStatusExcept;";
		    out.end();
		}
	    }
	    
	    out.end();
	}

        default:
            break;
        }
    }
}
