// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <Types.h>
#include <GenUtil.h>
#include <GenCPP.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Generate code for the tie header file
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusTieH(Container_ptr container,
				  IdlPrettyPrint& out)
{
    //
    // Get scope
    //
    String_var scope = getScope(container);

    //
    // Get container contents descriptions
    //
    Container::DescriptionSeq_var seq =
	container -> describe_contents(dk_all, true, -1);

    ULong i, j, k;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get container
	//
	Container_var cont = Container::_narrow(seq[i].contained_object);

	switch(seq[i].kind)
	{
	case dk_Module:
	{
	    //
	    // Module
	    //
	    ModuleDescription* desc;
	    Boolean b = seq[i].value >>= desc;
	    assert(b);

	    // Don't add this here!!!
	    // if(!check(desc -> id))
	    //     continue;

            String_var name = getPrefixedName("POA", seq[i].contained_object);

            comment(desc -> id, out);
            out << "\nnamespace " << name;
            out << "\n{";

            assert(!is_nil(cont));
	    genCPlusPlusTieH(cont, out);

            out.sep();
            out << "\n} // End of namespace " << name;

	    break;
	}

	case dk_Interface:
	{
	    //
	    // Interface
	    //
	    InterfaceDef_var interf =
		InterfaceDef::_narrow(seq[i].contained_object);
	    assert(!is_nil(interf));
	    InterfaceDef::FullInterfaceDescription_var desc =
		interf -> describe_interface();

	    if(!check(desc -> id))
		continue;

            String_var name =
                getPrefixedName("POA", seq[i].contained_object);

	    //
	    // Get the name of the interface's skeleton class
	    //
            ScopedName_var absolute = getAbsolute(seq[i].contained_object);

	    comment(desc -> id, out);

	    //
	    // The tie template
	    //
	    out.sep();
	    out << "\ntemplate<class T>"
		<< "\nclass " << name << "_tie : ";

	    out << "virtual public " << name;

	    out.start();

	    //
	    // Private members
	    //
	    out << "\nT* ptr_;";
	    out << "\nPortableServer::POA_ptr poa_;";
	    out << "\nCORBA::Boolean rel_;";
	    out.sep();
	    out << '\n' << name << "_tie(const " << name << "_tie<T>&) { }";
	    out << "\nvoid operator=(const " << name << "_tie<T>&) { }";

	    //
	    // Public members
	    //
	    out.sep();
	    out.direct("\npublic:");

	    out.sep();
	    out << '\n' << name << "_tie(T& t)";
	    out.inc();
	    out << "\n: ptr_(&t), poa_(PortableServer::POA::_nil()), "
                << "rel_(CORBA::FALSE)";
	    out.dec();
	    out.start();
	    out.end();

	    out.sep();
	    out << '\n' << name << "_tie(T& t, PortableServer::POA_ptr poa)";
	    out.inc();
	    out << "\n: ptr_(&t), poa_(PortableServer::POA::_duplicate(poa)), "
		<< "rel_(CORBA::FALSE)";
	    out.dec();
	    out.start();
	    out.end();

            // XXX
            // egcs 1.1.1 has a bug that causes failure when a default
            // function parameter refers to a namespace.  Until it's
            // fixed, we must use two functions.
            //

#if 0 // Enable this when egcs 1.1.1 is fixed
	    out.sep();
	    out << '\n' << name << "_tie(T* p, "
		<< "CORBA::Boolean release = CORBA::TRUE)";
	    out.inc();
	    out << "\n: ptr_(p), poa_(PortableServer::POA::_nil()), "
		<< "rel_(release)";
	    out.dec();
	    out.start();
	    out.end();
#endif

            // XXX Remove this when egcs 1.1.1 is fixed
	    out.sep();
	    out << '\n' << name << "_tie(T* p)";
	    out.inc();
	    out << "\n: ptr_(p), poa_(PortableServer::POA::_nil()), "
		<< "rel_(CORBA::TRUE)";
	    out.dec();
	    out.start();
	    out.end();

            // XXX Remove this when egcs 1.1.1 is fixed
	    out.sep();
	    out << '\n' << name << "_tie(T* p, CORBA::Boolean release)";
	    out.inc();
	    out << "\n: ptr_(p), poa_(PortableServer::POA::_nil()), "
		<< "rel_(release)";
	    out.dec();
	    out.start();
	    out.end();

#if 0 // Enable this when egcs 1.1.1 is fixed
	    out.sep();
	    out << '\n' << name << "_tie(T* p, PortableServer::POA_ptr poa, "
                << "CORBA::Boolean release = CORBA::TRUE)";
	    out.inc();
	    out << "\n: ptr_(p), poa_(PortableServer::POA::_duplicate(poa)), "
		<< "rel_(release)";
	    out.dec();
	    out.start();
	    out.end();
#endif

            // XXX Remove this when egcs 1.1.1 is fixed
	    out.sep();
	    out << '\n' << name << "_tie(T* p, PortableServer::POA_ptr poa)";
	    out.inc();
	    out << "\n: ptr_(p), poa_(PortableServer::POA::_duplicate(poa)), "
		<< "rel_(CORBA::TRUE)";
	    out.dec();
	    out.start();
	    out.end();

            // XXX Remove this when egcs 1.1.1 is fixed
	    out.sep();
	    out << '\n' << name << "_tie(T* p, PortableServer::POA_ptr poa, "
                << "CORBA::Boolean release)";
	    out.inc();
	    out << "\n: ptr_(p), poa_(PortableServer::POA::_duplicate(poa)), "
		<< "rel_(release)";
	    out.dec();
	    out.start();
	    out.end();

	    out.sep();
	    out << "\nvirtual\n~" << name << "_tie()";
	    out.start();
	    out << "\nif(rel_)";
	    out.inc();
	    out << "\ndelete ptr_;";
	    out.dec();
 	    out.sep();
 	    out << "\nCORBA::release(poa_);";
	    out.end();

	    //
	    // Tie-specific functions
	    //
	    out.sep();
	    out << "\nT*\n_tied_object()";
	    out.start();
	    out << "\nreturn ptr_;";
	    out.end();

	    out.sep();
	    out << "\nvoid\n_tied_object(T& obj)";
	    out.start();
	    out << "\nif(rel_)";
	    out.inc();
	    out << "\ndelete ptr_;";
	    out.dec();
	    out.sep();
	    out << "\nptr_ = &obj;";
	    out << "\nrel_ = CORBA::FALSE;";
	    out.end();

#if 0 // Enable this when egcs 1.1.1 is fixed
	    out.sep();
	    out << "\nvoid\n_tied_object(T* obj, "
		<< "CORBA::Boolean release = CORBA::TRUE)";
	    out.start();
	    out << "\nif(rel_)";
	    out.inc();
	    out << "\ndelete ptr_;";
	    out.dec();
	    out.sep();
	    out << "\nptr_ = obj;";
	    out << "\nrel_ = release;";
	    out.end();
#endif

            // XXX Remove this when egcs 1.1.1 is fixed
	    out.sep();
	    out << "\nvoid\n_tied_object(T* obj)";
	    out.start();
	    out << "\n_tied_object(obj, CORBA::TRUE);";
	    out.end();

            // XXX Remove this when egcs 1.1.1 is fixed
	    out.sep();
	    out << "\nvoid\n_tied_object(T* obj, CORBA::Boolean release)";
	    out.start();
	    out << "\nif(rel_)";
	    out.inc();
	    out << "\ndelete ptr_;";
	    out.dec();
	    out.sep();
	    out << "\nptr_ = obj;";
	    out << "\nrel_ = release;";
	    out.end();

	    out.sep();
	    out << "\nCORBA::Boolean\n_is_owner()";
	    out.start();
	    out << "\nreturn rel_;";
	    out.end();

	    out.sep();
	    out << "\nvoid\n_is_owner(CORBA::Boolean b)";
	    out.start();
	    out << "\nrel_ = b;";
	    out.end();

	    //
	    // Attributes and operations
	    //
	    ContainedSeq_var attributes =
		interf -> contents(dk_Attribute, false);

	    ContainedSeq_var operations =
		interf -> contents(dk_Operation, false);

	    String_var sysEx =
                string_dup("CORBA::SystemException");

	    for(j = 0 ; j < attributes -> length() ; j++)
	    {
		Contained::Description_var contDesc =
		    attributes[j] -> describe();
		AttributeDescription* atDesc;
		Boolean b = (contDesc -> value) >>= atDesc;
		assert(b);

		Identifier_var atName = fixKwd(atDesc -> name);

		out.sep();

		TypeCode_var origType = GetOrigType(atDesc -> type);
		String_var ret =
		    getTypeString("", atDesc -> type, GetTypeReturn);
		out << "\nvirtual " << ret << '\n' << atName << "()";
		out.inc();
		out << "\nthrow(" << sysEx << ")";
		out.dec();
		out.start();
		out << "\nreturn ptr_ -> " << atName << "();";
		out.end();

		if(atDesc -> mode == ATTR_NORMAL)
		{
		    out.sep();

		    String_var in = getTypeString("", atDesc -> type,
							GetTypeIn);

		    out << "\nvirtual void\n" << atName << '(' << in << " a)";
		    out.inc();
		    out << "\nthrow(" << sysEx << ")";
		    out.dec();
		    out.start();
		    out << "\nptr_ -> " << atName << "(a);";
		    out.end();
		}
	    }

	    for(j = 0 ; j < operations -> length() ; j++)
	    {
		Contained::Description_var contDesc =
		    operations[j] -> describe();
		OperationDescription* opDesc;
		Boolean b = (contDesc -> value) >>= opDesc;
		assert(b);
		
		Identifier_var opName = fixKwd(opDesc -> name);

		out.sep();

		String_var ret =
		    getTypeString("", opDesc -> result, GetTypeReturn);

		out << "\nvirtual " << ret << '\n' << opName << '(';

		unsigned int indent = out.getIndent();
		out.setIndent(out.getPos());

		for(k = 0 ; k < opDesc -> parameters.length() ; k++)
		{
		    GetType tt;
		    switch(opDesc -> parameters[k].mode)
		    {
		    case PARAM_IN:
			tt = GetTypeIn;
			break;

		    case PARAM_INOUT:
			tt = GetTypeInOut;
			break;

		    case PARAM_OUT:
			tt = GetTypeOut;
			break;
		    }

		    String_var argName =
			fixKwd(opDesc -> parameters[k].name);

		    String_var s =
			getTypeString("", opDesc -> parameters[k].type, tt,
				      false, argName);

		    if(k > 0)
			out << ",\n";

		    out << s;
		}

		if(opDesc -> contexts.length() > 0)
		{
		    if(opDesc -> parameters.length() > 0)
			out << ",\n";
		
		    out << "CORBA::Context_ptr context";
		}

		out << ')';

		out.setIndent(indent);

		out.inc();
		out << "\nthrow(";

		indent = out.getIndent();
		out.setIndent(out.getPos());

		for(k = 0 ; k < opDesc -> exceptions.length() ; k++)
		{
		    Identifier_var exName =
			getAbsolute(opDesc -> exceptions[k].id);
		    out << exName << ",\n";
		}

		out << sysEx << ")";
		out.setIndent(indent);
		out.dec();

		out.start();
		out << '\n';
		if(opDesc -> result -> kind() != tk_void)
		    out << "return ";
		out << "ptr_ -> " << opName << '(';

		for(k = 0 ; k < opDesc -> parameters.length() ; k++)
		{
		    if(k > 0)
			out << ", ";

		    String_var argName = fixKwd(opDesc -> parameters[k].name);

		    out << argName;
		}

		if(opDesc -> contexts.length() > 0)
		{
		    if(opDesc -> parameters.length() > 0)
			out << ", ";
		
		    out << "context";
		}

		out << ");";

		out.end();
	    }

	    //
	    // Override ServantBase operations
	    //
	    out.sep();
	    out << "\nPortableServer::POA_ptr _default_POA()";
	    out.start();
	    out << "\nif(!CORBA::is_nil(poa_))";
	    out.inc();
	    out << "\nreturn PortableServer::POA::_duplicate(poa_);";
	    out.dec();
	    out << "\nelse";
	    out.inc();
	    out << "\nreturn " << name << "::_default_POA();";
	    out.dec();
	    out.end();

	    out.end();
	    out << ';';

	    break;
	}

	default:
	    break;
	}
    }
}
