// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef GEN_HTML_H
#define GEN_HTML_H

//
// The HTML code generator class
//
class IdlHTMLGenerator : public GenTextBase
{
    //
    // Special HTML "Pretty-Print" class
    //
    class PrettyPrint : public IdlPrettyPrint
    {
	CORBA::String_var path_;
	OB_STD(ofstream) out_;

    public:

	PrettyPrint(const char*, const char*, const char*, unsigned int = 4);
	virtual ~PrettyPrint();

	const char* path() const { return path_; }
	int good() { return out_.good(); }
	virtual void printString(const char*);
    };

    //
    // The top-level file name (without ".html")
    //
    static const char* top_;

    //
    // The program name
    //
    const char* prog_;

    //
    // Comments from IDL file
    //
    IdlCommentSeq commentSeq_;

    //
    // If true, don't sort symbols
    //
    bool noSort_;

    //
    // If true, ignore case when sorting
    //
    bool ignoreCase_;

    //
    // If set, write .html files to this path
    //
    const char* outDir_;

    //
    // Cache for links
    //
    IdlStringSeq linkSeq_;

    //
    // Print comment
    //
    void comment(const char*, IdlPrettyPrint&);

    //
    // Print intro
    //
    void intro(const char*, const char*, IdlPrettyPrint&);

    char* getTypeString(const char*, CORBA::TypeCode_ptr,
			GetType = GetTypeNormal,
			const char* = 0);

    //
    // Get link
    //
    char* getLink(const char*);

    //
    // Generate code for the header file, general part
    //
    int genHTML(CORBA::Container_ptr, IdlPrettyPrint&);

public:

    IdlHTMLGenerator(const char*, CORBA::Repository_ptr,
		     const IdlCommentSeq&, const char*, bool, bool);

    //
    // Generate HTML code
    //
    int gen();
};

#endif
