// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <Types.h>
#include <Util.h>
#include <GenUtil.h>
#include <GenTextBase.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#include <GenHTML.h>

#include <stdlib.h>
#include <ctype.h>
#include <errno.h>

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Special HTML "Pretty-Print" class member implementation
// ----------------------------------------------------------------------

IdlHTMLGenerator::PrettyPrint::PrettyPrint(const char* prog,
					   const char* fullName,
					   const char* dir,
					   unsigned int num)
    : IdlPrettyPrint(out_, num)
{
    //
    // Create complete file name path
    //
    path_ = string_dup(dir);
    if(strlen(path_) > 0)
	path_ += '/';
    path_ += fullName;
    path_ += ".html";

    //
    // Create the file
    //
    out_.open(path_);

    //
    // Write header
    //
    if(good())
    {
	*this << "<html>\n<head>\n\n";

	*this << "<!-- Generated by the ORBacus IDL-to-HTML translator "
	      << "-->\n"
	      << '\n'
	      << "<!-- Copyright (c) 1998 -->\n"
	      << "<!-- Object-Oriented Concepts, Inc. -->\n"
	      << "<!-- Billerica, MA, USA -->\n"
	      << '\n'
	      << "<!-- Copyright (c) 1998 -->\n"
	      << "<!-- Object-Oriented Concepts GmbH -->\n"
	      << "<!-- Ettlingen, Germany -->\n"
	      << '\n'
	      << "<!-- All Rights Reserved -->\n"
	      << '\n'
	      << "<!-- Version: " << Version << " -->\n"
	      << "<!-- License: " << License << " -->\n";

	*this << "\n<title>IDL Documentation for &quot;"
	      << fullName << "&quot;</title>";

	*this << "\n</head>\n<body>\n";
    }
    else
    {
	cerr << prog << ": can't open \"" << path_
	     << "\": " << strerror(errno) << endl;
    }
}

IdlHTMLGenerator::PrettyPrint::~PrettyPrint()
{
    *this << "\n<hr></hr>";

    *this << "\n<p><em>";

    *this << "\nGenerated by the "
	  << "<a href=http://www.ooc.com/ob/>"
	  << "ORBacus</a> IDL-to-HTML translator ("
	  << License << ')';

    *this << "\n</em></p>";

    *this << "\n</body>\n</html>\n";
}

void
IdlHTMLGenerator::PrettyPrint::printString(const char* s)
{
    if(!s)
	return;

    IdlPrettyPrint::printString(s);

    const char* linkStart = strstr(s, "<a ");
    const char* linkEnd = strchr(s, '>');

    //
    // Don't count characters belonging to an HTML link
    //
    if(linkStart && linkEnd && linkEnd > linkStart)
	setPos(getPos() - (linkEnd - linkStart) - 5);
}

// ----------------------------------------------------------------------
// Print header comment
// ----------------------------------------------------------------------

void
IdlHTMLGenerator::comment(const char* id, IdlPrettyPrint& out)
{
    for(ULong i = 0 ; i < commentSeq_.length() ; i++)
    {
	if(strcmp(id, commentSeq_[i].id) == 0)
	{
	    IdlStringSeq memberSeq;
	    IdlStringSeq exceptionSeq;
	    IdlStringSeq paramSeq;
	    IdlStringSeq returnSeq;
	    IdlStringSeq seeSeq;
	    IdlStringSeq sinceSeq;
	    IdlStringSeq authorSeq;
	    IdlStringSeq versionSeq;

	    String_var comment = commentSeq_[i].comment;

	    const char* tok = tokenize(comment.inout());

	    while(tok)
	    {
		if(strncmp(tok, keyException_, strlen(keyException_)) == 0)
		{
		    exceptionSeq.append(tok + strlen(keyException_));
		}
		else if(strncmp(tok, keyMember_, strlen(keyMember_)) == 0)
		{
		    memberSeq.append(tok + strlen(keyMember_));
		}
		else if(strncmp(tok, keyParam_, strlen(keyParam_)) == 0)
		{
		    paramSeq.append(tok + strlen(keyParam_));
		}
		else if(strncmp(tok, keyReturn_, strlen(keyReturn_)) == 0)
		{
		    returnSeq.append(tok + strlen(keyReturn_));
		}
		else if(strncmp(tok, keySee_, strlen(keySee_)) == 0)
		{
		    seeSeq.append(tok + strlen(keySee_));
		}
		else if(strncmp(tok, keySince_, strlen(keySince_)) == 0)
		{
		    sinceSeq.append(tok + strlen(keySince_));
		}
		else if(strncmp(tok, keyAuthor_, strlen(keyAuthor_)) == 0)
		{
		    authorSeq.append(tok + strlen(keyAuthor_));
		}
		else if(strncmp(tok, keyVersion_, strlen(keyVersion_)) == 0)
		{
		    versionSeq.append(tok + strlen(keyVersion_));
		}

		tok = tokenize(0);
	    }

	    char* end = strchr(comment.inout(), '@');
	    if(end)
		*end = '\0';

	    out << "\n<p>" << comment << "\n</p>";

	    if(memberSeq.length() > 0)
	    {
		out << "\n<dl>\n<dt><b>Members:</b>";

		for(ULong j = 0 ; j < memberSeq.length() ; j++)
		{
		    String_var s = memberSeq[j];
		    const char* tok = strtok(s.inout(), " \t");
		    if(tok)
		    {
// 			out << "\n<dd><code><strong>" << tok
// 			    << "</strong></code>"
// 			    << " - " << tok + strlen(tok) + 1 << "</dd>";
			out << "\n<dd><code>" << tok
			    << "</code>"
			    << " - " << tok + strlen(tok) + 1 << "</dd>";
		    }
		}

		out << "\n</dl>";
	    }

	    if(paramSeq.length() > 0)
	    {
		out << "\n<dl>\n<dt><b>Parameters:</b>";

		for(ULong j = 0 ; j < paramSeq.length() ; j++)
		{
		    String_var s = paramSeq[j];
		    const char* tok = strtok(s.inout(), " \t");
		    if(tok)
		    {
// 			out << "\n<dd><code><strong>" << tok
// 			    << "</strong></code>"
// 			    << " - " << tok + strlen(tok) + 1 << "</dd>";
			out << "\n<dd><code>" << tok
			    << "</code>"
			    << " - " << tok + strlen(tok) + 1 << "</dd>";
		    }
		}

		out << "\n</dl>";
	    }

	    if(returnSeq.length() > 0)
	    {
		out << "\n<dl>\n<dt><b>Returns:</b>";

		for(ULong j = 0 ; j < returnSeq.length() ; j++)
		{
		    out << "\n<dd>" << returnSeq[j] << "</dd>";
		}

		out << "\n</dl>";
	    }

	    if(exceptionSeq.length() > 0)
	    {
		out << "\n<dl>\n<dt><b>Raises:</b>";

		for(ULong j = 0 ; j < exceptionSeq.length() ; j++)
		{
		    String_var s = exceptionSeq[j];
		    const char* tok = strtok(s.inout(), " \t");
		    if(tok)
		    {
// 			out << "\n<dd><code><strong>" << tok
// 			    << "</strong></code>"
// 			    << " - " << tok + strlen(tok) + 1 << "</dd>";
			out << "\n<dd><code>" << tok
			    << "</code>"
			    << " - " << tok + strlen(tok) + 1 << "</dd>";
		    }
		}
		 
		out << "\n</dl>";
	    }

	    if(seeSeq.length() > 0)
	    {
		out << "\n<dl>\n<dt><b>See Also:</b>";

		for(ULong j = 0 ; j < seeSeq.length() ; j++)
		{
		    out << "\n<dd>" << seeSeq[j] << "</dd>";
		}

		out << "\n</dl>";
	    }

	    if(sinceSeq.length() > 0)
	    {
		out << "\n<dl>\n<dt><b>Available Since:</b>";

		for(ULong j = 0 ; j < sinceSeq.length() ; j++)
		{
		    out << "\n<dd>" << sinceSeq[j] << "</dd>";
		}

		out << "\n</dl>";
	    }

	    if(authorSeq.length() > 0)
	    {
		out << "\n<dl>\n<dt><b>Author:</b>";

		for(ULong j = 0 ; j < authorSeq.length() ; j++)
		{
		    out << "\n<dd>" << authorSeq[j] << "</dd>";
		}

		out << "\n</dl>";
	    }

	    if(versionSeq.length() > 0)
	    {
		out << "\n<dl>\n<dt><b>Version:</b>";

		for(ULong j = 0 ; j < versionSeq.length() ; j++)
		{
		    out << "\n<dd>" << versionSeq[j] << "</dd>";
		}

		out << "\n</dl>";
	    }

	    out << "<br></br>";

	    break;
	}
    }
}

// ----------------------------------------------------------------------
// Print intro
// ----------------------------------------------------------------------

void
IdlHTMLGenerator::intro(const char* id, const char* name, IdlPrettyPrint& out)
{
    String_var link = getLink(id);

    out << "\n<dt><a href=\"" << link << "\">" << name
	<< "</a>\n<br></br>\n<dd>";

    for(ULong i = 0 ; i < commentSeq_.length() ; i++)
    {
	if(strcmp(id, commentSeq_[i].id) == 0)
	{
	    String_var comment = commentSeq_[i].comment;

	    char* s = comment.inout();
	    char* end1;

	    while(true)
	    {
		end1 = strchr(s, '.');

		if(end1 == 0)
		    break;

		end1++;
		if(*end1 == ' ' || *end1 == '\t' || *end1 == '\n')
		    break;

		s = end1;
	    }

	    char* end2 = strchr(comment.inout(), '@');

	    char* end;
	    if(end1 && end2)
		end = end1 < end2 ? end1 : end2;
	    else if(end1)
		end = end1;
	    else
		end = end2;

	    if(end)
		*end = '\0';

	    out << comment;

	    break;
	}
    }
}

// ----------------------------------------------------------------------
// Create type string
// ----------------------------------------------------------------------

char*
IdlHTMLGenerator::getTypeString(const char* scope, TypeCode_ptr type,
				GetType gt, const char* ident)
{
    //
    // Get result
    //
    String_var result = string_dup("???");
    switch(type -> kind())
    {
    case tk_void:
    case tk_short:
    case tk_long:
    case tk_longlong:
    case tk_ushort:
    case tk_ulong:
    case tk_ulonglong:
    case tk_float:
    case tk_double:
    case tk_longdouble:
    case tk_fixed:
    case tk_boolean:
    case tk_char:
    case tk_wchar:
    case tk_octet:
    case tk_any:
    case tk_TypeCode:
    case tk_Principal:
    {
	// **************************************************
	// Basic types (without strings)
	// **************************************************

	static const char* names[] =
	{
	    "void",
	    "short", "long",
	    "unsigned short", "unsigned long",
	    "float", "double",
	    "boolean", "char", "octet",
	    "any",
	    "TypeCode",
	    "Principal"
	};
	
	result = names[(int)(type -> kind()) - (int)tk_void];

	break;
    }

    case tk_objref:
    {
	// **************************************************
	// Object references (including pseudo object refereces)
        // **************************************************

	if(strcmp(type -> id(), "IDL:omg.org/CORBA/Object:1.0") == 0)
	{
	    result = string_dup("Object");
	}
	else
	{
	    result = getAbsolute(type -> id());
	    IdlRemoveScope(scope, result.inout());
	}

	break;
    }
    
    case tk_struct:
    case tk_union:
    case tk_except:
    case tk_enum:
    case tk_alias:
    {
	// **************************************************
	// Structs, unions, exceptions, enums
	// **************************************************

	result = getAbsolute(type -> id());
	IdlRemoveScope(scope, result.inout());

	break;
    }

    case tk_string:
    {
	// **************************************************
	// Strings
	// **************************************************

	result = string_dup("string");

	if(type -> length())
	{
	    result += '<';
	    result += type -> length();
	    result += '>';
	}

	break;
    }

    case tk_wstring:
    {
	// **************************************************
	// Wide strings
	// **************************************************

	result = string_dup("wstring");

	if(type -> length())
	{
	    result += '<';
	    result += type -> length();
	    result += '>';
	}

	break;
    }

    case tk_sequence:
    {
	// **************************************************
	// Sequences
	// **************************************************

	TypeCode_var contentType = type -> content_type();
	String_var contentTypeString = getTypeString(scope, contentType);
	
	switch(contentType -> kind())
	{
	case tk_null:
	case tk_void:
	    assert(false);
	    break;
	    
	default:

	    result = string_dup("sequence");
	    result += "&lt;";
	    result += contentTypeString;
	    
	    if(type -> length())
	    {
		result += ", ";
		result += type -> length();
	    }
	    
	    result += "&gt;";
	    
	    break;
	}

	break;
    }

    case tk_array:
    {
	// **************************************************
	// Arrays
	// **************************************************

	OB::ULongSeq sizes;
	
	TypeCode_var contentType = TypeCode::_duplicate(type);
	
	while(contentType -> kind() == tk_array)
	{
	    sizes.append(contentType -> length());
	    contentType = contentType -> content_type();
	}
	
	String_var array = string_dup("");
	
	for(ULong i = 0 ; i < sizes.length() ; i++)
	{
	    array += '[';
	    array += sizes[i];
	    array += ']';
	}
	
	result = getTypeString(scope, contentType);
	result += ' ';
	result += ident;
	result += array;

	break;
    }

    case tk_null:
	assert(false);
	break;
    }

    //
    // Add link
    //
    switch(type -> kind())
    {
    case tk_null:
    case tk_void:
    case tk_short:
    case tk_long:
    case tk_longlong:
    case tk_ushort:
    case tk_ulong:
    case tk_ulonglong:
    case tk_float:
    case tk_double:
    case tk_longdouble:
    case tk_fixed:
    case tk_boolean:
    case tk_char:
    case tk_wchar:
    case tk_octet:
    case tk_any:
    case tk_TypeCode:
    case tk_Principal:
    case tk_sequence:
    case tk_array:
    case tk_string:
    case tk_wstring:
	break;
	
    case tk_objref:
    case tk_struct:
    case tk_union:
    case tk_except:
    case tk_enum:
    case tk_alias:
    {
	if(!(type -> kind() == tk_objref &&
	     strcmp(type -> id(), "IDL:omg.org/CORBA/Object:1.0") == 0))
	{
	    String_var tmp = result._retn();
	    
	    String_var link = getLink(type -> id());
	    
	    result = string_dup("<a href=\"");
	    result += link;
	    result += "\">";
	    result += tmp;
	    result += "</a>";
	}

	break;
    }
    }
    
    //
    // Add in, inout, or out
    //
    switch(gt)
    {
    case GetTypeNormal:
	break;

    case GetTypeIn:
    {
	String_var tmp = result._retn();
	result = string_dup("in ");
	result += tmp;
	break;
    }
    
    case GetTypeInOut:
    {
	String_var tmp = result._retn();
	result = string_dup("inout ");
	result += tmp;
	break;
    }
    
    case GetTypeOut:
    {
	String_var tmp = result._retn();
	result = string_dup("out ");
	result += tmp;
	break;
    }
    }

    //
    // Add identifier
    //
    if(type -> kind() != tk_array)
    {
	if(ident && strlen(result))
	{
	    result += ' ';
	    result += ident;
	}
    }

    return result._retn();
}

// ----------------------------------------------------------------------
// Get link
// ----------------------------------------------------------------------

char*
IdlHTMLGenerator::getLink(const char* id)
{
    for(ULong i = 0 ; i < linkSeq_.length() ; i += 2)
    {
	if(strcmp(linkSeq_[i], id) == 0)
	    return string_dup(linkSeq_[i + 1]);
    }

    String_var absolute = getAbsolute(id);
    absolute = fixAbsolute(absolute);

    String_var link;

    Contained_var contained = repository_ -> lookup_id(id);
    assert(!is_nil(contained));
    
    DefinitionKind defKind = contained -> def_kind();
    if(defKind == dk_Module || defKind == dk_Interface)
    {
	link = absolute._retn();
	link += ".html";
    }
    else
    {
	char* name = strrchr(absolute.inout(), '.');
	if(name)
	{
	    name++;
	    assert(*name != '\0');
	    *(name - 1) = '\0';
	    link = string_dup(absolute);
	    link += ".html#";
	    link += name;
	}
	else
	{
	    link = string_dup(top_);
	    link += ".html#";
	    link += absolute;
	}
    }

    ULong len = linkSeq_.length();
    linkSeq_.length(len + 2);
    linkSeq_[len] = id;
    linkSeq_[len + 1] = link;
    return link._retn();
}
