// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef GEN_RTF_H
#define GEN_RTF_H

//
// The RTF code generator class
//
class IdlRTFGenerator : public GenTextBase
{
    //
    // Special RTF "Pretty-Print" class
    //
    class PrettyPrint : public IdlPrettyPrint
    {
	CORBA::String_var path_;
	OB_STD(ofstream) out_;

    public:

	PrettyPrint(const char*, const char*, const char*, const char*,
		    const char*, const char*, const char*, unsigned int = 4);
	virtual ~PrettyPrint();

	const char* path() const { return path_; }
	int good() { return out_.good(); }
	virtual void nl();
	virtual void start();
	virtual void end();
	virtual void printString(const char*);
    };

    //
    // RTF style constants
    //
    static const char* boldon_;  // <b>
    static const char* boldoff_; // </b>

    //
    // The top-level file name (without ".rtf")
    //
    static const char* top_;

    //
    // The program name
    //
    const char* prog_;

    //
    // Comments from IDL file
    //
    IdlCommentSeq commentSeq_;

    //
    // If true, don't sort symbols
    //
    bool noSort_;

    //
    // If true, ignore case when sorting
    //
    bool ignoreCase_;

    //
    // If true, create index
    //
    bool withIndex_;

    //
    // If set, write .rtf files to this path
    //
    const char* outDir_;

    //
    // If set, write .rtf code to this file
    //
    const char* singleFile_;

    //
    // Font names
    //
    const char* bodyFont_;
    const char* literalFont_;
    const char* titleFont_;
    const char* headingFont_;

    //
    // RTF font settings
    //
    CORBA::String_var body_;
    CORBA::String_var literal_;
    CORBA::String_var title_;
    CORBA::String_var heading_;

    //
    // Cache for links
    //
    IdlStringSeq linkSeq_;

    //
    // Print comment
    //
    void comment(const char*, IdlPrettyPrint&);

    //
    // Print paragraph
    //
    void paragraph(const char*, IdlPrettyPrint&);

    //
    // Find and replace substrings, case insensitive
    //
    char* replace(const char*, const char*, const char*);

    //
    // Print intro
    //
    void intro(const char*, const char*, IdlPrettyPrint&);

    char* getTypeString(const char*, CORBA::TypeCode_ptr,
			GetType = GetTypeNormal,
			const char* = 0);

    //
    // Generate code for the header file, general part
    //
    int genRTF(CORBA::Container_ptr, IdlPrettyPrint&);

public:

    IdlRTFGenerator(const char*, CORBA::Repository_ptr,
		    const IdlCommentSeq&, const char*, const char*, bool, bool,
		    bool, const char*, const char*, const char*, const char*,
		    unsigned int, unsigned int, unsigned int, unsigned int);

    //
    // Generate RTF code
    //
    int gen();
};

#endif
