// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef GEN_TEXT_BASE_H
#define GEN_TEXT_BASE_H

//
// The RTF/HTML code generator base class
//
class GenTextBase
{
    //
    // Cache for absolute names
    //
    IdlStringSeq absSeq_;

    //
    // Keyword list
    //
    IdlStringSeq keywordSeq_;

protected:

    GenTextBase(CORBA::Repository_ptr);
    virtual ~GenTextBase() { };

    //
    // Generate code
    //
    int gen();

    //
    // The repository
    //
    CORBA::Repository_var repository_;

    //
    // The keywords
    //
    static const char* keyException_;
    static const char* keyMember_;
    static const char* keyParam_;
    static const char* keyReturn_;
    static const char* keySee_;
    static const char* keySince_;
    static const char* keyAuthor_;
    static const char* keyVersion_;

    //
    // Create type string
    //
    enum GetType
    {
	GetTypeNormal,  // Normal type string
	GetTypeIn,      // Type string for use as in parameter
	GetTypeOut,     // Type string for use as out parameter
	GetTypeInOut    // Type string for use as inout parameter
    };
    
    //
    // Get type string
    //
    virtual char* getTypeString(const char*, CORBA::TypeCode_ptr, GetType,
				const char*) = 0;

    //
    // Get absolute name
    //
    char* getAbsolute(CORBA::Contained_ptr);
    char* getAbsolute(const char*);
   
    //
    // Fix absolute path name, i.e. replace "::" by "."
    //
    char* fixAbsolute(const char*);

    //
    // Get scope
    //
    void getScope(CORBA::Container_ptr, char*&);

    //
    // Print union labels
    //
    void printUnionLabels(const IdlUnionMemberInfo&, IdlPrettyPrint&,
			  CORBA::TypeCode_ptr);

    //
    // Tokenize
    //
    char* tokenize(char*);

    //
    // Check for keyword
    //
    bool isKeyword(const char*);
};

#endif

