%{

// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <Types.h>
#include <Util.h>
#include <GramUtil.h>

#include <stdlib.h>
#include <ctype.h>

#ifdef WIN32
#   include <malloc.h>
#ifndef BORLAND
#   define alloca _alloca
#endif
#endif

#ifdef __DECCXX
#   include <alloca.h>
#endif

using namespace CORBA;
using namespace OB;

void yyerror(const char* s);
int yylex();

static void TranslateAny(const Any&,
			 ULong&, Boolean&,
			 Long&, Boolean&,
			 Double&, Boolean&);

static bool CreateValue(const char*, const IdlValueInfo&, Boolean, Boolean);

static bool CreateValueFwd(const char*, Boolean);

//
// The current line number and the filename stack
//
long IdlLineno;
IdlStringSeq IdlFileStack;

//
// The current documentation comment
//
String_var IdlDocComment;

//
// The ORB
//
ORB_ptr IdlORB;

//
// The repository that is to be feeded
//
Repository_ptr IdlRepository;

//
// All repository ids of all definitions in the top-level IDL file
//
RepositoryIdSeq* IdlTopFileContent;

//
// All IDL files that are directly included from the top-level IDL file
//
IdlStringSeq* IdlIncludes;

//
// All comments for all repository ids
//
IdlCommentSeq* IdlComments;

//
// The stack for the containers
//
IdlContainerSeq IdlContStack;

//
// A stack of stacks for the prefixes
// 
IdlStringSeqSeq IdlPrefStackStack;

//
// This is needed for unions
//
IDLType_ptr IdlUnionSwitchType;

//
// Sequence with interfaces that have been forward declared
//
IdlStringSeq IdlFwdInterfaces;

//
// Sequence with valuetypes that have been forward declared
//
IdlStringSeq IdlFwdValuetypes;

%}

//
// Expect shift/reduce conflicts
//
%expect 4

/*** Return type ***/

%union
{
    //
    // General types
    //
    char* string;
    IdlStringSeq* stringSeq;

    CORBA::Long longVal;
    CORBA::ULong uLongVal;
    CORBA::Char charVal;
    CORBA::Double doubleVal;
    CORBA::Boolean booleanVal;

    CORBA::Any* any;
    IdlAnySeq* anySeq;

    //
    // Special types
    //
    IdlDeclarator* declarator;
    IdlDeclaratorSeq* declaratorSeq;
    IdlValueInfo* valueInfo;

    //
    // Types from the IR
    //
    CORBA::IDLType_ptr idlType;
    CORBA::StructDef_ptr structDef;
    CORBA::UnionDef_ptr unionDef;
    CORBA::ExceptionDef_ptr exceptionDef;
    CORBA::AttributeMode attributeMode;
    CORBA::StructMember* structMember;
    CORBA::StructMemberSeq* structMemberSeq;
    CORBA::UnionMemberSeq* unionMemberSeq;
    CORBA::UnionMember* unionMember;
    CORBA::InterfaceDef_ptr interfaceDef;
    CORBA::InterfaceDefSeq* interfaceDefSeq;
    CORBA::OperationMode operationMode;
    CORBA::ParameterMode parameterMode;
    CORBA::ParameterDescription* parameterDescription;
    CORBA::ParDescriptionSeq* parDescriptionSeq;
    CORBA::ExceptionDefSeq* exceptionDefSeq;
    CORBA::Visibility memberVisibility;
    CORBA::Initializer* initializer;
};

/*** Token definitions ***/

%token				IDL_ABSTRACT
%token				IDL_ANY
%token				IDL_ATTRIBUTE
%token				IDL_BOOLEAN
%token				IDL_CASE
%token				IDL_CHAR
%token				IDL_CONST
%token				IDL_CONTEXT
%token				IDL_CUSTOM
%token				IDL_DEFAULT
%token				IDL_DOUBLE
%token				IDL_ENUM
%token				IDL_EXCEPTION
%token				IDL_FIXED
%token				IDL_FLOAT
%token				IDL_IN
%token				IDL_INIT
%token				IDL_INOUT
%token				IDL_INTERFACE
%token				IDL_LONG
%token				IDL_MODULE
%token				IDL_NATIVE
%token				IDL_OBJECT
%token				IDL_OCTET
%token				IDL_ONEWAY
%token				IDL_OUT
%token				IDL_PRIVATE
%token				IDL_PUBLIC
%token				IDL_RAISES
%token				IDL_READONLY
%token				IDL_SEQUENCE
%token				IDL_SHORT
%token				IDL_STRING
%token				IDL_STRUCT
%token				IDL_SUPPORTS
%token				IDL_SWITCH
%token				IDL_TRUNCATABLE
%token				IDL_TYPEDEF
%token				IDL_UNSIGNED
%token				IDL_UNION
%token				IDL_VALUEBASE
%token				IDL_VALUETYPE
%token				IDL_VOID
%token				IDL_WCHAR
%token				IDL_WSTRING
%token				IDL_LEFT_SHIFT
%token				IDL_RIGHT_SHIFT
%token				IDL_SCOPE_DELIMITOR

%token <string>			IDL_IDENTIFIER
%token <uLongVal>		IDL_INTEGER_LITERAL
%token <string>			IDL_STRING_LITERAL
%token <wstring>		IDL_WIDE_STRING_LITERAL
%token <charVal>		IDL_CHARACTER_LITERAL
%token <charVal>		IDL_WIDE_CHARACTER_LITERAL
%token <doubleVal>		IDL_FLOATING_PT_LITERAL
%token		                IDL_FIXED_PT_LITERAL
%token <booleanVal>		IDL_FALSE
%token <booleanVal>		IDL_TRUE

%token <string>			IDL_PRAGMA

/*** Types definitions ***/

%type <any>			literal
%type <booleanVal>		boolean_literal
%type <stringSeq>		COMMA_string_literal_STAR
%type <string>			interface_name
%type <stringSeq>		COMMA_interface_name_STAR
%type <string>			scoped_name
%type <stringSeq>		COMMA_scoped_name_STAR
%type <string>			enumerator
%type <stringSeq>		COMMA_enumerator_STAR
%type <uLongVal>		positive_int_const
%type <uLongVal>		fixed_array_size
%type <interfaceDefSeq>		interface_inheritance_spec
%type <interfaceDefSeq>		interface_inheritance_spec_OPT

%type <string>			concrete_value_name
%type <string>			abstract_value_name
%type <stringSeq>		COMMA_abstract_value_name_STAR
%type <valueInfo>		value_inheritance_spec_OPT
%type <valueInfo>		value_inheritance_spec
%type <valueInfo>		value_inheritance_spec_1
%type <interfaceDef>		value_inheritance_spec_2
%type <structMemberSeq>		init_param_decls_OPT
%type <structMemberSeq>		init_param_decls
%type <structMemberSeq>		COMMA_init_param_decl_STAR
%type <structMember>		init_param_decl

%type <idlType>			type_spec
%type <idlType>			const_type
%type <idlType>			simple_type_spec
%type <idlType>			param_type_spec
%type <idlType>			base_type_spec
%type <idlType>			template_type_spec
%type <idlType>			constr_type_spec
%type <idlType>			floating_pt_type
%type <idlType>			integer_type
%type <idlType>			signed_int
%type <idlType>			signed_long_int
%type <idlType>			signed_longlong_int
%type <idlType>			signed_short_int
%type <idlType>			unsigned_int
%type <idlType>			unsigned_long_int
%type <idlType>			unsigned_longlong_int
%type <idlType>			unsigned_short_int
%type <idlType>			char_type
%type <idlType>			wide_char_type
%type <idlType>			boolean_type
%type <idlType>			octet_type
%type <idlType>			any_type
%type <idlType>			object_type
%type <idlType>			sequence_type
%type <idlType>			string_type
%type <idlType>			wide_string_type
%type <idlType>			fixed_pt_type
%type <idlType>			fixed_pt_const_type
%type <idlType>			value_base_type
%type <structDef>		struct_type
%type <unionDef>		union_type
%type <idlType>			enum_type
%type <idlType>			switch_type_spec
%type <idlType>			op_type_spec

%type <declaratorSeq>		declarators
%type <declaratorSeq>		COMMA_declarator_STAR
%type <declarator>		declarator
%type <declarator>		simple_declarator
%type <declaratorSeq>		COMMA_simple_declarator_STAR
%type <declarator>		complex_declarator
%type <declarator>		array_declarator
%type <declarator>		fixed_array_size_PLUS

%type <structMemberSeq>		member_list
%type <structMemberSeq>		member_PLUS
%type <structMemberSeq>		member_STAR
%type <structMemberSeq>		member
%type <structMemberSeq>		member_NO_SEM

%type <unionMemberSeq>		switch_body
%type <unionMemberSeq>		case_PLUS
%type <unionMemberSeq>		case
%type <unionMemberSeq>		case_NO_SEM
%type <anySeq>			case_label_PLUS
%type <any>			case_label
%type <unionMember>		element_spec

%type <operationMode>		op_attribute
%type <operationMode>		op_attribute_OPT
%type <parDescriptionSeq>	COMMA_param_dcl_STAR
%type <parDescriptionSeq>	parameter_dcls
%type <parameterDescription>	param_dcl
%type <parameterMode>		param_attribute

%type <exceptionDefSeq>		raises_expr
%type <exceptionDefSeq>		raises_expr_OPT

%type <stringSeq>		context_expr
%type <stringSeq>		context_expr_OPT

%type <any>			const_exp
%type <any>			or_expr
%type <any>			xor_expr
%type <any>			and_expr
%type <any>			shift_expr
%type <any>			add_expr
%type <any>			mult_expr
%type <any>			unary_expr
%type <charVal>			unary_operator
%type <any>			primary_expr

%type <attributeMode>		readonly_OPT
%type <booleanVal>		abstract_OPT
%type <booleanVal>		truncatable_OPT
%type <memberVisibility>	public_OR_private

%%

start		: specification
		{
		}
		| /* empty */
		{
		}
		;

/* (1) */
specification	:
                {
		    //
		    // Push repository on the container stack
		    //
		    IdlContStack.append(Repository::_duplicate(IdlRepository));

		    //
		    // Push new empty prefix on the prefix stacks
		    //
		    for(ULong i = 0 ; i < IdlPrefStackStack.length() ; i++)
			IdlPrefStackStack[i].append(string_dup(""));
                }
                definition_PLUS
		{
		    //
		    // Pop container from the stack
		    //
		    assert(IdlContStack.length() > 0);
		    IdlContStack.length(IdlContStack.length() - 1);
		    assert(yynerrs > 0 || IdlContStack.length() == 0);

		    //
		    // Pop prefix from the prefix stacks
		    //
		    for(ULong i = 0 ; i < IdlPrefStackStack.length() ; i++)
		    {
			assert(IdlPrefStackStack[i].length() > 0);
			IdlPrefStackStack[i].length(
			    IdlPrefStackStack[i].length() - 1);
			assert(yynerrs > 0 ||
			       IdlPrefStackStack[i].length() == 0);
		    }
		}
		;

/* (2) */
definition_PLUS	: definition
		{
		}
		| definition definition_PLUS
		{
		}
	        | error ';'
		{
		    yyerrok;
		}
		;

/* (2) */
definition	: definition_NO_SEM ';'
		{
		}
		| definition_NO_SEM
		{
		    IdlError("`;' missing after definition");
		    yynerrs++;
		}
		| IDL_PRAGMA
		{
		    String_var s = $1;
		    if(IdlParsePragma(s) == false)
		    {
			yynerrs++;
		    }
		}
		;

/* (2) */
definition_NO_SEM: type_dcl
		{
		}
		| const_dcl
		{
		}
		| except_dcl
		{
		}
		| interface
		{
		}
		| module
		{
		}
		| value
		{
		}
		;

/* (3) */
module		: IDL_MODULE IDL_IDENTIFIER
		{
		    //
		    // Get container from container stack
		    //
		    assert(IdlContStack.length() > 0);
		    Container_var container =
			IdlContStack[IdlContStack.length() - 1];

		    //
		    // Get name, id and version
		    //
		    Identifier_var name = $2;
		    RepositoryId_var id = IdlGetRepositoryId(name);
		    VersionSpec_var version = string_dup("1.0");

		    //
		    // Check if name already exists
		    //
		    Contained_var c;
		    try
		    {
			c = container -> lookup(name);
		    }
		    catch(const SystemException& ex)
		    {
			IdlPrintException(ex, id, name);
			yynerrs++;
			YYERROR;
		    }

		    ModuleDef_var module;
		    if(!is_nil(c))
		    {
			//
			// Name exists, check if it is an module declaration
			//
			module = ModuleDef::_narrow(c);

			if(is_nil(module))
			{	
			    String_var err = string_dup("`");
			    err += name;
			    err += "' is not a module";
			    IdlError(err);
			    yynerrs++;
			    YYERROR;
			}
		    }
		    else
		    {
			//
			// Create new module
			//
			try
			{
			    module =
				container -> create_module(id, name, version);
			}
			catch(const SystemException& ex)
			{
			    IdlPrintException(ex, id, name);
			    yynerrs++;
			    YYERROR;
			}
		    }

		    assert(!is_nil(module));

		    //
		    // Push new container on the stack
		    //
		    IdlContStack.append(ModuleDef::_duplicate(module));

		    //
		    // Push new prefix on the prefix stacks
		    //
		    for(ULong i = 0 ; i < IdlPrefStackStack.length() ; i++)
		    {
			String_var prefix =
			    IdlPrefStackStack[i][IdlPrefStackStack[i].length()
						- 1];
			prefix += name;
			prefix += "/";
			IdlPrefStackStack[i].append(prefix);
		    }

		    //
		    // Add ID to IdlTopFileContent and IdlComments
		    //
		    IdlAddID(id);
		}
		'{' definition_PLUS '}'
		{
		    //
		    // Pop container from the stack
		    //
		    assert(IdlContStack.length() > 0);
		    IdlContStack.length(IdlContStack.length() - 1);

		    //
		    // Pop prefix from the prefix stacks
		    //
		    for(ULong i = 0 ; i < IdlPrefStackStack.length() ; i++)
		    {
			assert(IdlPrefStackStack[i].length() > 0);
			IdlPrefStackStack[i].length(
			    IdlPrefStackStack[i].length() - 1);
		    }
		}
		;

/* (4) */
interface	: interface_dcl
		{
		}
		| forward_dcl
		{
		}
		;

/* (5) */
interface_dcl	: interface_header '{' interface_body '}'
		{
		    //
		    // Pop container from the stack
		    //
		    assert(IdlContStack.length() > 0);
		    IdlContStack.length(IdlContStack.length() - 1);

		    //
		    // Pop prefix from the prefix stacks
		    //
		    for(ULong i = 0 ; i < IdlPrefStackStack.length() ; i++)
		    {
			assert(IdlPrefStackStack[i].length() > 0);
			IdlPrefStackStack[i].length(
			    IdlPrefStackStack[i].length() - 1);
		    }
		}
		;

/* (6) */
forward_dcl	: abstract_OPT IDL_INTERFACE IDL_IDENTIFIER
		{
		    Boolean isAbstract = $1;

		    //
		    // Get container from container stack
		    //
		    assert(IdlContStack.length() > 0);
		    Container_var container =
			IdlContStack[IdlContStack.length() - 1];

		    //
		    // Get name, id and version
		    //
		    Identifier_var name = $3;
		    RepositoryId_var id = IdlGetRepositoryId(name);
		    VersionSpec_var version = string_dup("1.0");

		    //
		    // Check if name already exists
		    //
		    Contained_var c;
		    try
		    {
			c = container -> lookup(name);
		    }
		    catch(const SystemException& ex)
		    {
			IdlPrintException(ex, id, name);
			yynerrs++;
			YYERROR;
		    }

		    if(!is_nil(c))
		    {
			//
			// Check for ID equivalence
			//
// 			RepositoryId_var oldId = c -> id();
// 			if(strcmp(oldId, id) != 0)
// 			{
// 			    String_var err =
// 				string_dup("conflicting repository ids "
// 						 "for `interface ");
// 			    err += name;
// 			    err += "'";
// 			    IdlError(err);
// 			    yynerrs++;
// 			    YYERROR;
// 			}

			//
			// Name exists, check if it is an interface
			//
			InterfaceDef_var i = InterfaceDef::_narrow(c);

			if(is_nil(i))
			{
			    String_var err = string_dup("conflicting types "
						        "for `interface ");
			    err += name;
			    err += "'";
			    IdlError(err);
			    yynerrs++;
			    YYERROR;
			}

			//
			// Make sure forward decl matches
			//
			if(isAbstract != i -> is_abstract())
			{
			    String_var err;
			    err = string_dup("interface `");
			    err += name;
			    err += "' does not match forward declaration";
			    IdlError(err);
			    yynerrs++;
			    YYERROR;
			}
		    }
		    else
		    {
			//
			// Create "empty" interface as forward
			// declaration, do not put it on the container
			// stack, but put it in the list of forward
			// declared interfaces
			//
			InterfaceDefSeq seq;

			try
			{
			    InterfaceDef_var interf =
				container -> create_interface(id, name,
							      version, seq,
							      isAbstract);
			    IdlFwdInterfaces.append(interf -> absolute_name());
		        }
			catch(const SystemException& ex)
			{
			    IdlPrintException(ex, id, name);
			    yynerrs++;
			    YYERROR;
			}
		    }
		}
		;

/* (7) */
interface_header: abstract_OPT IDL_INTERFACE IDL_IDENTIFIER
                  interface_inheritance_spec_OPT
		{
		    Boolean isAbstract = $1;

		    //
		    // Get container from container stack
		    //
		    assert(IdlContStack.length() > 0);
		    Container_var container =
			IdlContStack[IdlContStack.length() - 1];

		    //
		    // Get name, id and version
		    //
		    Identifier_var name = $3;
		    RepositoryId_var id = IdlGetRepositoryId(name);
		    VersionSpec_var version = string_dup("1.0");

		    //
		    // Get interface inheritance specification sequence
		    //
		    InterfaceDefSeq_var seq = $4;
		    
		    //
		    // Check if name already exists
		    //
		    Contained_var c;
		    try
		    {
			c = container -> lookup(name);
		    }
		    catch(const SystemException& ex)
		    {
			IdlPrintException(ex, id, name);
			yynerrs++;
			YYERROR;
		    }

		    InterfaceDef_var interf;
		    if(!is_nil(c))
		    {
			//
			// Check for ID equivalence
			//
// 			RepositoryId_var oldId = c -> id();
// 			if(strcmp(oldId, id) != 0)
// 			{
// 			    String_var err =
// 				string_dup("conflicting repository ids "
// 						 "for `interface ");
// 			    err += name;
// 			    err += "'";
// 			    IdlError(err);
// 			    yynerrs++;
// 			    YYERROR;
// 			}

			//
			// Name exists, check if it is an interface
			// forward declaration
			//
			interf = InterfaceDef::_narrow(c);

			if(is_nil(interf))
			{	
			    String_var err = string_dup("conflicting types "
						        "for `interface ");
			    err += name;
			    err += "'";
			    IdlError(err);
			    yynerrs++;
			    YYERROR;
			}

			//
			// Don't use ID for comparison, because it's
			// possible that the ID was changed with
			// #pragma
			//
			ScopedName_var abs = interf -> absolute_name();
			ULong i;
			for(i = 0 ; i < IdlFwdInterfaces.length() ; i++)
			{
			    if(strcmp(abs, IdlFwdInterfaces[i]) == 0)
				break;
			}

			if(i >= IdlFwdInterfaces.length())
			{
			    String_var err =
				string_dup("redefinition of `interface ");
			    err += name;
			    err += "'";
			    IdlError(err);
			    yynerrs++;
			    YYERROR;
			}

			//
			// Make sure forward decl matches
			//
			if(isAbstract != interf -> is_abstract())
			{
			    String_var err;
			    err = string_dup("interface `");
			    err += name;
			    err += "' does not match forward declaration";
			    IdlError(err);
			    yynerrs++;
			    YYERROR;
			}

			IdlFwdInterfaces.remove(i);

			try
			{
			    //
			    // Fill interface
			    //
			    interf -> id(id);
			    // Not necessary, move() sets new name and
			    // version:
			    //interf -> name(name);
			    //interf -> version(version);
			    interf -> base_interfaces(seq);
			    
			    //
			    // Move interface (this will make this interface
			    // the last entry in the container)
			    //
			    interf -> move(container, name, version);
			}
			catch(const SystemException& ex)
			{
			    IdlPrintException(ex, id, name);
			    yynerrs++;
			    YYERROR;
			}
		    }
		    else
		    {
			//
			// Create new interface
			//
			try
			{
			    interf = container ->
				create_interface(id, name, version, seq,
						 isAbstract);
			}
			catch(const SystemException& ex)
			{
			    IdlPrintException(ex, id, name);
			    yynerrs++;
			    YYERROR;
			}
		    }
		    
		    assert(!is_nil(interf));

		    //
		    // Push new container on the stack
		    //
		    IdlContStack.append(InterfaceDef::_duplicate(interf));
		    
		    //
		    // Push new prefix on the prefix stacks
		    //
		    for(ULong i = 0 ; i < IdlPrefStackStack.length() ; i++)
		    {
			String_var prefix =
			    IdlPrefStackStack[i][IdlPrefStackStack[i].length()
						- 1];
			prefix += name;
			prefix += "/";
			IdlPrefStackStack[i].append(prefix);
		    }
		    
		    //
		    // Add ID to IdlTopFileContent and IdlComments
		    //
		    IdlAddID(id);
		}
		;
/* (8) */
interface_body	: export_STAR
		{
		}
		;

/* (9) */
export_STAR	: export export_STAR
		{
		}
		| /* empty */
		{
		}
		| error ';'
		{
		    yyerrok;
		}
		;

/* (9) */
export		: export_NO_SEM ';'
		{
		}
		| export_NO_SEM
		{
		    IdlError("`;' missing after definition");
		    yynerrs++;
		}
		| IDL_PRAGMA
		{
		    String_var s = $1;
		    if(!IdlParsePragma(s))
		    {
			yynerrs++;
		    }
		}
		;

/* (9) */
export_NO_SEM	: type_dcl
		{
		}
		| const_dcl
		{
		}
		| except_dcl
		{
		}
		| attr_dcl
		{
		}
		| op_dcl
		{
		}
		;

/* (10) */
interface_inheritance_spec_OPT: interface_inheritance_spec
		{
		    $$ = $1;
		}
		| /* empty */
		{
		    $$ = new InterfaceDefSeq;
		}
		;

/* (10) */
interface_inheritance_spec: ':' interface_name COMMA_interface_name_STAR
		{
		    IdlStringSeq_var scopedNameSeq = $3;
		    scopedNameSeq -> insert($2);

		    //
		    // Get container from container stack
		    //
		    assert(IdlContStack.length() > 0);
		    Container_var container =
			IdlContStack[IdlContStack.length() - 1];

		    //
		    // Create interface inheritance specification sequence
		    //
		    InterfaceDefSeq_var iDefSeq =
			new InterfaceDefSeq(scopedNameSeq -> length());

		    for(ULong i = 0 ; i < scopedNameSeq -> length() ; i++)
		    {
			//
			// Find scoped name
			//
			const char* scoped = scopedNameSeq[i];
			IRObject_var irObj =
			    IdlFindScopedName(container, scoped);

			if(is_nil(irObj))
			{
			    yynerrs++;
			    YYERROR;
			}

			while(true)
			{
			    AliasDef_var aliasDef = AliasDef::_narrow(irObj);
			    
			    if(is_nil(aliasDef))
				break;
			    
			    irObj = aliasDef -> original_type_def();
			}

			InterfaceDef_var interfaceDef =
			    InterfaceDef::_narrow(irObj);

			if(is_nil(interfaceDef))
			{	
			    String_var err = string_dup("`");
			    err += scoped;
			    err += "' is not an interface type";
			    IdlError(err);
			    yynerrs++;
			    YYERROR;
			}

			ScopedName_var abs = interfaceDef -> absolute_name();
			for(ULong j = 0 ; j < IdlFwdInterfaces.length() ; j++)
			{
			    if(strcmp(abs, IdlFwdInterfaces[j]) == 0)
			    {
				String_var err =
				    string_dup("can't inherit from "
					       "undefined interface `");
				err += scoped;
				err += "'";
				IdlError(err);
				yynerrs++;
				YYERROR;
			    }
			}

			iDefSeq -> append(interfaceDef);
		    }

		    $$ = iDefSeq._retn();
		}
		;

/* (11) */
COMMA_interface_name_STAR: ',' interface_name COMMA_interface_name_STAR
		{
		    $3 -> insert($2);
		    $$ = $3;
		}
		| /* empty */
		{
		    $$ = new IdlStringSeq;
		}
		;

/* (11) */
interface_name  : scoped_name
                {
		    $$ = $1;
		}
                ;

/* (12) */
COMMA_scoped_name_STAR: ',' scoped_name COMMA_scoped_name_STAR
		{
		    $3 -> insert($2);
		    $$ = $3;
		}
		| /* empty */
		{
		    $$ = new IdlStringSeq;
		}
		;

/* (12) */
scoped_name	: IDL_IDENTIFIER
		{
		    $$ = $1;
		}
		| IDL_SCOPE_DELIMITOR IDL_IDENTIFIER
		{
		    Identifier_var ident = $2;

		    $$ = string_alloc(2 + strlen(ident));
		    strcpy($$, "::");
		    strcat($$, ident);
		}
		| scoped_name IDL_SCOPE_DELIMITOR IDL_IDENTIFIER
		{
		    Identifier_var scoped = $1;
		    Identifier_var ident = $3;

		    $$ = string_alloc(strlen(scoped) + 2 + strlen(ident));
		    strcpy($$, scoped);
		    strcat($$, "::");
		    strcat($$, ident);
		}
		;

/* (13) */
/*
value		: value_dcl
		{
		}
		| value_abs_dcl
		{
		}
		| value_box_dcl
		{
		}
		| value_forward_dcl
		{
		}
		;
*/
value		: CUSTOM_value_dcl
		{
		}
		| IDL_VALUETYPE value_dcl
		{
		}
		| value_abs_dcl
		{
		}
		| IDL_VALUETYPE value_box_dcl
		{
		}
		| ABSTRACT_value_forward_dcl
		{
		}
		| IDL_VALUETYPE value_forward_dcl
		{
		}
		;

/* (14) */
/*
value_forward_dcl: abstract_OPT IDL_VALUETYPE IDL_IDENTIFIER
		{
		}
		;
*/

/* (14) */
ABSTRACT_value_forward_dcl: IDL_ABSTRACT IDL_VALUETYPE IDL_IDENTIFIER
		{
		    Identifier_var name = $3;
                    if(! CreateValueFwd(name, TRUE))
                    {
                        YYERROR;
                    }
		}
		;

/* (14) */
value_forward_dcl: IDL_IDENTIFIER
		{
		    Identifier_var name = $1;
                    if(! CreateValueFwd(name, FALSE))
                    {
                        YYERROR;
                    }
		}
		;

/* (15) */
value_box_dcl	: IDL_IDENTIFIER type_spec
		{
                    //
                    // Get container from container stack
                    //
                    assert(IdlContStack.length() > 0);
                    Container_var container =
                        IdlContStack[IdlContStack.length() - 1];

                    //
                    // Get name, id and version
                    //
                    Identifier_var name = $1;
                    RepositoryId_var id = IdlGetRepositoryId(name);
                    VersionSpec_var version = string_dup("1.0");

                    IDLType_var idlType = $2;

                    //
                    // Check if name already exists
                    //
                    Contained_var c;
                    try
                    {
                        c = container -> lookup(name);
                    }
                    catch(const SystemException& ex)
                    {
                        IdlPrintException(ex, id, name);
                        yynerrs++;
                        YYERROR;
                    }

                    if(!is_nil(c))
                    {
                        String_var err = string_dup("conflicting types "
                                                    "for `valuetype ");
                        err += name;
                        err += "'";
                        IdlError(err);
                        yynerrs++;
                        YYERROR;
                    }
                    else
                    {
                        ValueBoxDef_var box;

                        //
                        // Create new valuebox
                        //
                        try
                        {
                            box = container -> create_value_box(id, name,
                                                                version,
                                                                idlType);
                            assert(!is_nil(box));

                            //
                            // Add ID to IdlTopFileContent and IdlComments
                            //
                            IdlAddID(id);
                        }
                        catch(const SystemException& ex)
                        {
                            IdlPrintException(ex, id, name);
                            yynerrs++;
                            YYERROR;
                        }
                    }
		}
		;

/* (16) */
value_abs_dcl	: IDL_ABSTRACT IDL_VALUETYPE IDL_IDENTIFIER
		  value_inheritance_spec_OPT
		{
		    Identifier_var name = $3;
                    IdlValueInfo_var info = $4;
                    if(! CreateValue(name, info, false, true))
                    {
                        YYERROR;
                    }
		}
		'{' export_STAR '}'
		{
                    //
                    // Pop container from the stack
                    // Pop prefixes from the stack
                    //
                    assert(IdlContStack.length() > 0);
                    IdlContStack.length(IdlContStack.length() - 1);

                    //
                    // Pop prefix from the prefix stacks
                    //
                    for(ULong i = 0 ; i < IdlPrefStackStack.length() ; i++)
                    {
                        assert(IdlPrefStackStack[i].length() > 0);
                        IdlPrefStackStack[i].length(
                            IdlPrefStackStack[i].length() - 1);
                    }
		}
		;

/* (17) */
/*
value_dcl	: value_header '{' value_element_STAR '}'
		{
		}
		;
*/

/* (17) */
CUSTOM_value_dcl: CUSTOM_value_header '{' value_element_STAR '}'
		{
                    //
                    // Pop container from the stack
                    // Pop prefixes from the stack
                    //
                    assert(IdlContStack.length() > 0);
                    IdlContStack.length(IdlContStack.length() - 1);

                    //
                    // Pop prefix from the prefix stacks
                    //
                    for(ULong i = 0 ; i < IdlPrefStackStack.length() ; i++)
                    {
                        assert(IdlPrefStackStack[i].length() > 0);
                        IdlPrefStackStack[i].length(
                            IdlPrefStackStack[i].length() - 1);
                    }
		}
		;

/* (17) */
value_dcl	: value_header '{' value_element_STAR '}'
		{
                    //
                    // Pop container from the stack
                    // Pop prefixes from the stack
                    //
                    assert(IdlContStack.length() > 0);
                    IdlContStack.length(IdlContStack.length() - 1);

                    //
                    // Pop prefix from the prefix stacks
                    //
                    for(ULong i = 0 ; i < IdlPrefStackStack.length() ; i++)
                    {
                        assert(IdlPrefStackStack[i].length() > 0);
                        IdlPrefStackStack[i].length(
                            IdlPrefStackStack[i].length() - 1);
                    }
		}
		;

/* (18) */
/*
value_header	: custom_OPT IDL_VALUETYPE IDL_IDENTIFIER
		  value_inheritance_spec_OPT
		{
		}
		;
*/

/* (18) */
CUSTOM_value_header: IDL_CUSTOM IDL_VALUETYPE IDL_IDENTIFIER
		  value_inheritance_spec_OPT
		{
		    Identifier_var name = $3;
                    IdlValueInfo_var info = $4;
                    if(! CreateValue(name, info, true, false))
                    {
                        YYERROR;
                    }
		}
		;

/* (18) */
value_header	: IDL_IDENTIFIER value_inheritance_spec_OPT
		{
		    Identifier_var name = $1;
                    IdlValueInfo_var info = $2;
                    if(! CreateValue(name, info, false, false))
                    {
                        YYERROR;
                    }
		}
		;

/* (19) */
value_inheritance_spec_OPT: value_inheritance_spec
		{
		    $$ = $1;
		}
		| /* empty */
		{
		    $$ = new IdlValueInfo;
		    $$ -> is_truncatable = FALSE;
		}
		;

/* (19) */
/*
This causes a reduce/reduce conflict:

value_inheritance_spec: value_inheritance_spec_1_OPT
		        value_inheritance_spec_2_OPT
*/
value_inheritance_spec: value_inheritance_spec_1
		{
		    $$ = $1;
		}
		| value_inheritance_spec_2
		{
		    IdlValueInfo_var info = new IdlValueInfo;
		    info -> supported_interface = $1;
		    info -> is_truncatable = FALSE;
		    $$ = info._retn();
		}
		| value_inheritance_spec_1 value_inheritance_spec_2
		{
		    $$ = $1;
                    $$ -> supported_interface = $2;
		}
		;

/* (19) */
value_inheritance_spec_1: ':' truncatable_OPT concrete_value_name
		  COMMA_abstract_value_name_STAR
		{
		    IdlValueInfo_var info = new IdlValueInfo;
		    info -> is_truncatable = $2;

                    String_var baseName = $3;
                    IdlStringSeq_var names = $4;

                    //
                    // It's legal to specify "ValueBase" as a base type,
                    // but it's not necessary.  If it's provided, just
                    // ignore it.
                    //
                    // XXX Currently there is discussion in orb_revision
                    // about making ValueBase required if no other concrete
                    // base value is used.
                    //
                    if(strcmp(baseName, "ValueBase") != 0)
                        names -> insert(baseName._retn());

                    //
                    // Get container from container stack
                    //
                    assert(IdlContStack.length() > 0);
                    Container_var container =
                        IdlContStack[IdlContStack.length() - 1];

                    ULong i, j;
                    for(i = 0 ; i < names -> length() ; i++)
                    {
                        //
                        // Find scoped name
                        //
                        const char* scoped = names[i];
                        IRObject_var irObj =
                            IdlFindScopedName(container, scoped);

                        if(is_nil(irObj))
                        {
                            yynerrs++;
                            YYERROR;
                        }

                        while(true)
                        {
                            AliasDef_var aliasDef = AliasDef::_narrow(irObj);

                            if(is_nil(aliasDef))
                                break;

                            irObj = aliasDef -> original_type_def();
                        }

                        ValueDef_var valueDef = ValueDef::_narrow(irObj);

                        if(is_nil(valueDef))
                        {
                            String_var err = string_dup("`");
                            err += scoped;
                            err += "' is not a valuetype";
                            IdlError(err);
                            yynerrs++;
                            YYERROR;
                        }

                        ScopedName_var abs = valueDef -> absolute_name();
                        for(j = 0 ; j < IdlFwdValuetypes.length() ; j++)
                        {
                            if(strcmp(abs, IdlFwdValuetypes[j]) == 0)
                            {
                                String_var err =
                                    string_dup("can't inherit from "
                                               "undefined valuetype `");
                                err += scoped;
                                err += "'";
                                IdlError(err);
                                yynerrs++;
                                YYERROR;
                            }
                        }

                        info -> base_values.append(valueDef);
                    }

		    $$ = info._retn();
		}
		;

/* (19) */
value_inheritance_spec_2: IDL_SUPPORTS interface_name
		{
                    //
                    // Get container from container stack
                    //
                    assert(IdlContStack.length() > 0);
                    Container_var container =
                        IdlContStack[IdlContStack.length() - 1];

                    //
                    // Find scoped name
                    //
		    ScopedName_var scoped = $2;
                    IRObject_var irObj = IdlFindScopedName(container, scoped);

                    if(is_nil(irObj))
                    {
                        yynerrs++;
                        YYERROR;
                    }

                    while(true)
                    {
                        AliasDef_var aliasDef = AliasDef::_narrow(irObj);

                        if(is_nil(aliasDef))
                            break;

                        irObj = aliasDef -> original_type_def();
                    }

                    InterfaceDef_var interfaceDef =
                        InterfaceDef::_narrow(irObj);

                    if(is_nil(interfaceDef))
                    {
                        String_var err = string_dup("`");
                        err += scoped;
                        err += "' is not an interface type";
                        IdlError(err);
                        yynerrs++;
                        YYERROR;
                    }

                    ScopedName_var abs = interfaceDef -> absolute_name();
                    for(ULong j = 0 ; j < IdlFwdInterfaces.length() ; j++)
                    {
                        if(strcmp(abs, IdlFwdInterfaces[j]) == 0)
                        {
                            String_var err =
                                string_dup("value can't support an "
                                           "undefined interface `");
                            err += scoped;
                            err += "'";
                            IdlError(err);
                            yynerrs++;
                            YYERROR;
                        }
                    }

                    $$ = interfaceDef._retn();
		}
		;

/* (20) */
concrete_value_name: scoped_name
		{
		    $$ = $1;
		}
		| value_base_type
		{
		    IDLType_var idlType = $1;
                    $$ = string_dup("ValueBase");
		}
		;

/* (21) */
COMMA_abstract_value_name_STAR: ',' abstract_value_name COMMA_abstract_value_name_STAR
		{
                    $3 -> insert($2);
                    $$ = $3;
		}
		| /* empty */
		{
		    $$ = new IdlStringSeq;
		}
		;
/* (21) */
abstract_value_name: scoped_name
		{
                    $$ = $1;
		}
		;

/* (22) */
value_element_STAR: value_element value_element_STAR
		{
		}
		| /* empty */
		{
		}
		| error ';'
		{
		    yyerrok;
		}
		;
/* (22) */
value_element	: export
		{
		}
		| state_member
		{
		}
		| init_dcl
		{
		}
		;

/* (23) */
state_member	: state_member_NO_SEM ';'
		{
		}
		| state_member_NO_SEM
		{
		    IdlError("`;' missing after state member");
		    yynerrs++;
		}
		;

/* (23) */
state_member_NO_SEM: public_OR_private type_spec declarators
                {
                    //
                    // Get container from container stack
                    //
                    assert(IdlContStack.length() > 0);
                    Container_var container =
                        IdlContStack[IdlContStack.length() - 1];

                    ValueDef_var valueDef = ValueDef::_narrow(container);
                    assert(! is_nil(valueDef));

                    Visibility access = $1;
                    IDLType_var simpleIDLType = $2;
                    IdlDeclaratorSeq_var declarators = $3;

                    ULong i;
                    Long j;

                    for(i = 0 ; i < declarators -> length() ; i++)
                    {
                        const IdlDeclarator& declarator = declarators[i];

                        //
                        // Get name, id and version
                        //
                        Identifier_var name = declarator.identifier;
                        RepositoryId_var id = IdlGetRepositoryId(name);
                        VersionSpec_var version = string_dup("1.0");

                        //
                        // Handle arrays
                        //
                        IDLType_var idlType = simpleIDLType;

                        for(j = declarator.sizes.length() - 1 ; j >= 0 ; j--)
                        {
                            ArrayDef_var arrayDef = IdlRepository ->
                                create_array(declarator.sizes[j], idlType);

                            idlType = IDLType::_duplicate(arrayDef);
                        }

                        //
                        // Create member
                        //
                        try
                        {
                            release(valueDef ->
                                create_value_member(id, name, version, idlType,
                                                    access));
                        }
                        catch(const SystemException& ex)
                        {
                            IdlPrintException(ex, id, name);
                            yynerrs++;
                            YYERROR;
                        }

                        //
                        // Add ID to IdlTopFileContent and IdlComments
                        //
                        IdlAddID(id);
                    }
		}
		;

/* (24) */
init_dcl	: init_dcl_NO_SEM ';'
		{
		}
		| init_dcl_NO_SEM
		{
		    IdlError("`;' missing after struct init declaration");
		    yynerrs++;
		}
		;

/* (24) */
init_dcl_NO_SEM	: IDL_INIT '(' init_param_decls_OPT ')'
		{
                    Initializer_var init = new Initializer;
                    StructMemberSeq_var members = $3;
                    init -> members = members;

                    //
                    // Get container from container stack
                    //
                    assert(IdlContStack.length() > 0);
                    Container_var container =
                        IdlContStack[IdlContStack.length() - 1];

                    ValueDef_var valueDef = ValueDef::_narrow(container);
                    assert(! is_nil(valueDef));

                    //
                    // Add the new initializer
                    //
                    InitializerSeq_var seq = valueDef -> initializers();
                    seq -> append(init);
                    try
                    {
                        valueDef -> initializers(seq);
                    }
                    catch(const SystemException& ex)
                    {
                        RepositoryId_var id = valueDef -> id();
                        Identifier_var name = valueDef -> name();
                        IdlPrintException(ex, id, name);
                        yynerrs++;
                        YYERROR;
                    }
		}
		;

/* (25) */
init_param_decls_OPT: init_param_decls
		{
		    $$ = $1;
		}
		| /* empty */
		{
		    $$ = new StructMemberSeq;
		}
		;
/* (25) */
init_param_decls: init_param_decl COMMA_init_param_decl_STAR
		{
                    StructMember_var member = $1;
                    $2 -> insert(member);
                    $$ = $2;
		}
		;

/* (26) */
COMMA_init_param_decl_STAR: ',' init_param_decl COMMA_init_param_decl_STAR
		{
                    StructMember_var member = $2;
                    $3 -> insert(member);
                    $$ = $3;
		}
		| /* empty */
		{
                    $$ = new StructMemberSeq;
		}
		;
/* (26) */
init_param_decl: init_param_attribute param_type_spec simple_declarator
		{
                    StructMember_var member = new StructMember;
                    IdlDeclarator_var declarator = $3;
                    assert(declarator -> sizes.length() == 0);
                    member -> name = declarator -> identifier;
                    member -> type_def = $2;
                    member -> type = $2 -> type();
                    $$ = member._retn();
		}
		;

/* (27) */
init_param_attribute: IDL_IN
		{
		}
		;

/* (28) */
const_dcl	: IDL_CONST const_type IDL_IDENTIFIER '=' const_exp
		{
		    //
		    // Get type and value
		    //
		    IDLType_var idlType = $2;
		    Any_var any = $5;

		    //
		    // Get name, id and version
		    //
		    Identifier_var name = $3;
		    RepositoryId_var id = IdlGetRepositoryId(name);
		    VersionSpec_var version = string_dup("1.0");

		    //
		    // Check if conversion is necessary
		    //
		    TypeCode_var tc1 = idlType -> type();
		    tc1 = GetOrigType(tc1);
		    TypeCode_var tc2 = any -> type();

		    if(!tc1 -> equivalent(tc2))
		    {
			ULong ul;
			Boolean ulb;
			Long l;
			Boolean lb;
			Double d;
			Boolean db;
			TranslateAny(any, ul, ulb, l, lb, d, db);

			const char* err =
			    "wrong type for constant initialization";
		    
			//
			// Convert value type
			//
			switch(tc1 -> kind())
			{
			case tk_short:
			    
			    if(ulb)
				any <<= (Short)ul;
			    else if(lb)
				any <<= (Short)l;
			    else if(db)
			    {
				IdlWarning("converting floating point to "
					   "short");
				any <<= (Short)d;
			    }
			    else
			    {
				IdlError(err);
				yynerrs++;
				YYERROR;
			    }
			    
			    break;
			    
			case tk_long:
			    
			    if(ulb)
				any <<= (Long)ul;
			    else if(lb)
				any <<= (Long)l;
			    else if(db)
			    {
				IdlWarning("converting floating point to "
					   "long");
				any <<= (Long)d;
			    }
			    else
			    {
				IdlError(err);
				yynerrs++;
				YYERROR;
			    }
			    
			    break;
			    
			case tk_ushort:
			    
			    if(ulb)
				any <<= (UShort)ul;
			    else if(lb)
			    {
				if(l < 0)
				    IdlWarning("converting negative value to "
					       "unsigned short");
				any <<= (UShort)l;
			    }
			    else if(db)
			    {
				IdlWarning("converting floating point to "
					   "unsigned short");
				any <<= (UShort)d;
			    }
			    else
			    {
				IdlError(err);
				yynerrs++;
				YYERROR;
			    }
			    
			    break;
			    
			case tk_ulong:
			    
			    if(ulb)
				any <<= (ULong)ul;
			    else if(lb)
			    {
				if(l < 0)
				    IdlWarning("converting negative value to "
					       "unsigned short");
				any <<= (ULong)l;
			    }
			    else if(db)
			    {
				IdlWarning("converting floating point to "
					   "unsigned long");
				any <<= (ULong)d;
			    }
			    else
			    {
				IdlError(err);
				yynerrs++;
				YYERROR;
			    }
			    
			    break;
			    
			case tk_float:
			    
			    if(ulb)
				any <<= (Float)ul;
			    else if(lb)
				any <<= (Float)l;
			    else if(db)
				any <<= (Float)d;
			    else
			    {
				IdlError(err);
				yynerrs++;
				YYERROR;
			    }
			    
			    break;
			    
			case tk_double:
			    
			    if(ulb)
				any <<= (Double)ul;
			    else if(lb)
				any <<= (Double)l;
			    else if(db)
				any <<= (Double)d;
			    else
			    {
				IdlError(err);
				yynerrs++;
				YYERROR;
			    }
			    
			    break;

			case tk_string:
			{
			    char* s;
			    if(any >>= s)
			    {
				ULong len = tc1 -> length();
				Any::from_string from(s, len);
				any <<= from;
			    }
			    break;
			}

			default:
			    IdlError(err);
			    yynerrs++;
			    YYERROR;
			    break;
			}
		    }

		    //
		    // Get container from container stack
		    //
		    assert(IdlContStack.length() > 0);
		    Container_var container =
			IdlContStack[IdlContStack.length() - 1];

		    //
		    // Create constant definition
		    //
		    try
		    {
			release(container ->
				      create_constant(id, name, version,
						      idlType, any));
		    }
		    catch(const SystemException& ex)
		    {
			IdlPrintException(ex, id, name);
			yynerrs++;
			YYERROR;
		    }

		    //
		    // Add ID to IdlTopFileContent and IdlComments
		    //
		    IdlAddID(id);
		}
		;

/* (29) */
const_type	: integer_type
		{
		    $$ = $1;
		}
		| char_type
		{
		    $$ = $1;
		}
		| wide_char_type
		{
		    $$ = $1;
		}
		| boolean_type
		{
		    $$ = $1;
		}
		| floating_pt_type
		{
		    $$ = $1;
		}
		| string_type
		{
		    $$ = $1;
		}
		| wide_string_type
		{
		    IDLType_var idlType = $1;

		    TypeCode_var tc = idlType -> type();
		    assert(tc -> kind() == tk_wstring);

		    if(tc -> length() != 0)
		    {
			IdlError("bounded wide strings are no "
				 "valid constant types");
			yynerrs++;
			YYERROR;
		    }

		    $$ = idlType._retn();
		}
		| fixed_pt_const_type
		{
		    IDLType_var idlType = $1;

		    TypeCode_var tc = idlType -> type();
		    assert(tc -> kind() == tk_fixed);

		    assert_nca(false, NCANotImpl);

		    $$ = idlType._retn();
		}
		| scoped_name
		{
		    ScopedName_var scoped = $1;

		    //
		    // Get container from container stack
		    //
		    assert(IdlContStack.length() > 0);
		    Container_var container =
			IdlContStack[IdlContStack.length() - 1];

		    //
		    // Find scoped name
		    //
		    IRObject_var irObj =
			IdlFindScopedName(container, scoped);

		    if(is_nil(irObj))
		    {
			yynerrs++;
			YYERROR;
		    }

		    bool valid = false;

		    IDLType_var idlType = IDLType::_narrow(irObj);
		    if(!is_nil(idlType))
		    {
			TypeCode_var tc = idlType -> type();
			tc = GetOrigType(tc);

			TCKind k = tc -> kind();

			if(k == tk_short || k == tk_ushort ||
			   k == tk_long || k == tk_ulong ||
			   k == tk_float || k == tk_double ||
			   k == tk_char || k == tk_boolean ||
			   (k == tk_string && tc -> length() == 0))
			    valid = true;
		    }

		    if(!valid)
		    {
			String_var err = string_dup("`");
			err += scoped;
			err += "' is not a valid constant type";
			IdlError(err);
			yynerrs++;
			YYERROR;
		    }

		    $$ = idlType._retn();
		}
		| octet_type
		{
		    $$ = $1;
		}
		;

/* (30) */
const_exp	: or_expr
		{
		    $$ = $1;
		}
		;

/* (31) */
or_expr		: xor_expr
		{
		    $$ = $1;
		}
		| or_expr '|' xor_expr
		{
		    Any_var left = $1;
		    Any_var right = $3;
		    Any_var result = new Any;

		    ULong uLongL, uLongR;
		    Boolean uLongLBool, uLongRBool;
		    Long longL, longR;
		    Boolean longLBool, longRBool;
		    Double doubleL, doubleR;
		    Boolean doubleLBool, doubleRBool;

		    TranslateAny(left,
				 uLongL, uLongLBool,
				 longL, longLBool,
				 doubleL, doubleLBool);

		    TranslateAny(right,
				 uLongR, uLongRBool,
				 longR, longRBool,
				 doubleR, doubleRBool);

		    if((uLongLBool || longLBool) &&
		       (uLongRBool || longRBool))
		    {
			if(longLBool || longRBool)
			    result <<= (longL | longR);
			else
			    result <<= (uLongL | uLongR);
		    }
		    else
		    {
			IdlError("invalid operand for binary operator |");
			yynerrs++;
			YYERROR;
		    }

		    $$ = result._retn();
		}
		;

/* (32) */
xor_expr	: and_expr
		{
		    $$ = $1;
		}
		| xor_expr '^' and_expr
		{
		    Any_var left = $1;
		    Any_var right = $3;
		    Any_var result = new Any;

		    ULong uLongL, uLongR;
		    Boolean uLongLBool, uLongRBool;
		    Long longL, longR;
		    Boolean longLBool, longRBool;
		    Double doubleL, doubleR;
		    Boolean doubleLBool, doubleRBool;

		    TranslateAny(left,
				 uLongL, uLongLBool,
				 longL, longLBool,
				 doubleL, doubleLBool);

		    TranslateAny(right,
				 uLongR, uLongRBool,
				 longR, longRBool,
				 doubleR, doubleRBool);

		    if((uLongLBool || longLBool) &&
		       (uLongRBool || longRBool))
		    {
			if(longLBool || longRBool)
			    result <<= (longL ^ longR);
			else
			    result <<= (uLongL ^ uLongR);
		    }
		    else
		    {
			IdlError("invalid operand for binary operator ^");
			yynerrs++;
			YYERROR;
		    }

		    $$ = result._retn();
		}
		;

/* (33) */
and_expr	: shift_expr
		{
		    $$ = $1;
		}
		| and_expr '&' shift_expr
		{
		    Any_var left = $1;
		    Any_var right = $3;
		    Any_var result = new Any;

		    ULong uLongL, uLongR;
		    Boolean uLongLBool, uLongRBool;
		    Long longL, longR;
		    Boolean longLBool, longRBool;
		    Double doubleL, doubleR;
		    Boolean doubleLBool, doubleRBool;

		    TranslateAny(left,
				 uLongL, uLongLBool,
				 longL, longLBool,
				 doubleL, doubleLBool);

		    TranslateAny(right,
				 uLongR, uLongRBool,
				 longR, longRBool,
				 doubleR, doubleRBool);

		    if((uLongLBool || longLBool) &&
		       (uLongRBool || longRBool))
		    {
			if(longLBool || longRBool)
			    result <<= (longL & longR);
			else
			    result <<= (uLongL & uLongR);
		    }
		    else
		    {
			IdlError("invalid operand for binary operator &");
			yynerrs++;
			YYERROR;
		    }

		    $$ = result._retn();
		}
		;

/* (34) */
shift_expr	: add_expr
		{
		    $$ = $1;
		}
		| shift_expr IDL_RIGHT_SHIFT add_expr
		{
		    Any_var left = $1;
		    Any_var right = $3;
		    Any_var result = new Any;

		    ULong uLongL, uLongR;
		    Boolean uLongLBool, uLongRBool;
		    Long longL, longR;
		    Boolean longLBool, longRBool;
		    Double doubleL, doubleR;
		    Boolean doubleLBool, doubleRBool;

		    TranslateAny(left,
				 uLongL, uLongLBool,
				 longL, longLBool,
				 doubleL, doubleLBool);

		    TranslateAny(right,
				 uLongR, uLongRBool,
				 longR, longRBool,
				 doubleR, doubleRBool);

		    if((uLongLBool || longLBool) &&
		       (uLongRBool || longRBool))
		    {
			if(!(0 <= longR && longR < 32))
			{
			    IdlError("right operand of binary operator >> "
				     "must be in the range 0 <= right "
				     "operand < 32");
			    yynerrs++;
			    YYERROR;
			}

			if(longLBool)
			    result <<= (longL >> longR);
			else
			    result <<= (uLongL >> longR);
		    }
		    else
		    {
			IdlError("invalid operand for binary operator >>");
			yynerrs++;
			YYERROR;
		    }

		    $$ = result._retn();
		}
		| shift_expr IDL_LEFT_SHIFT add_expr
		{
		    Any_var left = $1;
		    Any_var right = $3;
		    Any_var result = new Any;

		    ULong uLongL, uLongR;
		    Boolean uLongLBool, uLongRBool;
		    Long longL, longR;
		    Boolean longLBool, longRBool;
		    Double doubleL, doubleR;
		    Boolean doubleLBool, doubleRBool;

		    TranslateAny(left,
				 uLongL, uLongLBool,
				 longL, longLBool,
				 doubleL, doubleLBool);

		    TranslateAny(right,
				 uLongR, uLongRBool,
				 longR, longRBool,
				 doubleR, doubleRBool);

		    if((uLongLBool || longLBool) &&
		       (uLongRBool || longRBool))
		    {
			if(!(0 <= longR && longR < 32))
			{
			    IdlError("right operand of binary operator << "
				     "must be in the range 0 <= right "
				     "operand < 32");
			    yynerrs++;
			    YYERROR;
			}

			if(longLBool)
			    result <<= (longL << longR);
			else
			    result <<= (uLongL << longR);
		    }
		    else
		    {
			IdlError("invalid operand for binary operator <<");
			yynerrs++;
			YYERROR;
		    }

		    $$ = result._retn();
		}
		;

/* (35) */
add_expr	: mult_expr
		{
		    $$ = $1;
		}
		| add_expr '+' mult_expr
		{
		    Any_var left = $1;
		    Any_var right = $3;
		    Any_var result = new Any;

		    ULong uLongL, uLongR;
		    Boolean uLongLBool, uLongRBool;
		    Long longL, longR;
		    Boolean longLBool, longRBool;
		    Double doubleL, doubleR;
		    Boolean doubleLBool, doubleRBool;

		    TranslateAny(left,
				 uLongL, uLongLBool,
				 longL, longLBool,
				 doubleL, doubleLBool);

		    TranslateAny(right,
				 uLongR, uLongRBool,
				 longR, longRBool,
				 doubleR, doubleRBool);

		    if((uLongLBool || longLBool || doubleLBool) &&
		       (uLongRBool || longRBool || doubleRBool))
		    {
			if(doubleLBool || doubleRBool)
			    result <<= (doubleL + doubleR);
			else if(longLBool || longRBool)
			    result <<= (longL + longR);
			else
			    result <<= (uLongL + uLongR);
		    }
		    else
		    {
			IdlError("invalid operand for binary operator +");
			yynerrs++;
			YYERROR;
		    }

		    $$ = result._retn();
		}
		| add_expr '-' mult_expr
		{
		    Any_var left = $1;
		    Any_var right = $3;
		    Any_var result = new Any;

		    ULong uLongL, uLongR;
		    Boolean uLongLBool, uLongRBool;
		    Long longL, longR;
		    Boolean longLBool, longRBool;
		    Double doubleL, doubleR;
		    Boolean doubleLBool, doubleRBool;

		    TranslateAny(left,
				 uLongL, uLongLBool,
				 longL, longLBool,
				 doubleL, doubleLBool);

		    TranslateAny(right,
				 uLongR, uLongRBool,
				 longR, longRBool,
				 doubleR, doubleRBool);

		    if((uLongLBool || longLBool || doubleLBool) &&
		       (uLongRBool || longRBool || doubleRBool))
		    {
			if(doubleLBool || doubleRBool)
			    result <<= (doubleL - doubleR);
			else if(longLBool || longRBool)
			    result <<= (longL - longR);
			else
			{
			    if(uLongR > uLongL)
				result <<= ((Long)uLongL -
					    (Long)uLongR);
			    else
				result <<= (uLongL - uLongR);
			}
		    }
		    else
		    {
			IdlError("invalid operand for binary operator -");
			yynerrs++;
			YYERROR;
		    }

		    $$ = result._retn();
		}
		;

/* (36) */
mult_expr	: unary_expr
		{
		    $$ = $1;
		}
		| mult_expr '*' unary_expr
		{
		    Any_var left = $1;
		    Any_var right = $3;
		    Any_var result = new Any;

		    ULong uLongL, uLongR;
		    Boolean uLongLBool, uLongRBool;
		    Long longL, longR;
		    Boolean longLBool, longRBool;
		    Double doubleL, doubleR;
		    Boolean doubleLBool, doubleRBool;

		    TranslateAny(left,
				 uLongL, uLongLBool,
				 longL, longLBool,
				 doubleL, doubleLBool);

		    TranslateAny(right,
				 uLongR, uLongRBool,
				 longR, longRBool,
				 doubleR, doubleRBool);

		    if((uLongLBool || longLBool || doubleLBool) &&
		       (uLongRBool || longRBool || doubleRBool))
		    {
			if(doubleLBool || doubleRBool)
			    result <<= (doubleL * doubleR);
			else if(longLBool || longRBool)
			    result <<= (longL * longR);
			else
			    result <<= (uLongL * uLongR);
		    }
		    else
		    {
			IdlError("invalid operand for binary operator *");
			yynerrs++;
			YYERROR;
		    }

		    $$ = result._retn();
		}
		| mult_expr '/' unary_expr
		{
		    Any_var left = $1;
		    Any_var right = $3;
		    Any_var result = new Any;

		    ULong uLongL, uLongR;
		    Boolean uLongLBool, uLongRBool;
		    Long longL, longR;
		    Boolean longLBool, longRBool;
		    Double doubleL, doubleR;
		    Boolean doubleLBool, doubleRBool;

		    TranslateAny(left,
				 uLongL, uLongLBool,
				 longL, longLBool,
				 doubleL, doubleLBool);

		    TranslateAny(right,
				 uLongR, uLongRBool,
				 longR, longRBool,
				 doubleR, doubleRBool);

		    if((uLongLBool || longLBool || doubleLBool) &&
		       (uLongRBool || longRBool || doubleRBool))
		    {
			if(doubleLBool || doubleRBool)
			    result <<= (doubleL / doubleR);
			else if(longLBool || longRBool)
			    result <<= (longL / longR);
			else
			    result <<= (uLongL / uLongR);
		    }
		    else
		    {
			IdlError("invalid operand for binary operator /");
			yynerrs++;
			YYERROR;
		    }

		    $$ = result._retn();
		}
		| mult_expr '%' unary_expr
		{
		    Any_var left = $1;
		    Any_var right = $3;
		    Any_var result = new Any;

		    ULong uLongL, uLongR;
		    Boolean uLongLBool, uLongRBool;
		    Long longL, longR;
		    Boolean longLBool, longRBool;
		    Double doubleL, doubleR;
		    Boolean doubleLBool, doubleRBool;

		    TranslateAny(left,
				 uLongL, uLongLBool,
				 longL, longLBool,
				 doubleL, doubleLBool);

		    TranslateAny(right,
				 uLongR, uLongRBool,
				 longR, longRBool,
				 doubleR, doubleRBool);

		    if((uLongLBool || longLBool) &&
		       (uLongRBool || longRBool))
		    {
			if(longLBool || longRBool)
			    result <<= (longL % longR);
			else
			    result <<= (uLongL % uLongR);
		    }
		    else
		    {
			IdlError("invalid operand for binary operator %");
			yynerrs++;
			YYERROR;
		    }

		    $$ = result._retn();
		}
		;

/* (37) */
unary_expr	: unary_operator primary_expr
		{
		    Any_var result = $2;
		    
		    const char* err = 0;
		    
		    if($1 == '~')
			err = "wrong argument type for bit-complement";
		    else if($1 == '+')
			err = "wrong argument type for unary plus";
		    else if($1 == '-')
			err = "wrong argument type for unary minus";
		    else
			assert(false);

		    ULong ul;
		    Boolean ulb;
		    Long l;
		    Boolean lb;
		    Double d;
		    Boolean db;
		    TranslateAny(result, ul, ulb, l, lb, d, db);

		    if(ulb)
		    {
			if($1 == '-')
			{
			    l = -(Long)ul;
			    result <<= l;
			}
			else if($1 == '~')
			{
			    ul = ~ul;
			    result <<= ul;
			}
		    }
		    else if(lb)
		    {
			if($1 == '-')
			{
			    l = -l;
			    result <<= l;
			}
			else if($1 == '~')
			{
			    l = ~l;
			    result <<= l;
			}
		    }
		    else if(db)
		    {
			if($1 == '-')
			{
			    d = -d;
			    result <<= d;
			}
			else if($1 == '~')
			{
			    IdlError(err);
			    yynerrs++;
			    YYERROR;
			}
		    }
		    else
		    {
			IdlError(err);
			yynerrs++;
			YYERROR;
		    }

		    $$ = result._retn();
		}
		| primary_expr
		{
		    $$ = $1;
		}
		;

/* (38) */
unary_operator	: '-'
		{
		    $$ = '-';
		}
		| '+'
		{
		    $$ = '+';
		}
		| '~'
		{
		    $$ = '~';
		}
		;

/* (39) */
primary_expr	: scoped_name
		{
		    ScopedName_var scoped = $1;
		    $$ = 0;
		    
		    //
		    // Get container from container stack
		    //
		    assert(IdlContStack.length() > 0);
		    Container_var container =
			IdlContStack[IdlContStack.length() - 1];
		    
		    if(!is_nil(IdlUnionSwitchType))
		    {
			//
			// Special treatment for unions with enums
			//
			IDLType_var switchType =
			    IDLType::_duplicate(IdlUnionSwitchType);
			
			while(true)
			{
			    AliasDef_var aliasDef =
				AliasDef::_narrow(switchType);
			    
			    if(is_nil(aliasDef))
				break;
			    
			    switchType = aliasDef -> original_type_def();
			}
			
			EnumDef_var enumDef = EnumDef::_narrow(switchType);
			
			if(!is_nil(enumDef))
			{
			    Long memberNum =
				IdlFindEnumMember(container, enumDef, scoped);
			    
			    if(memberNum == -1)
			    {
				yynerrs++;
				YYERROR;
			    }
			    
			    TypeCode_var tc = enumDef -> type();
			    ULong* p = new ULong;
			    *p = (ULong)memberNum;
			    $$ = new Any;
			    $$ -> replace(tc, p, true);
			}
		    }

		    if($$ == 0)
		    {
			//
			// Find scoped name
			//
			IRObject_var irObj =
			    IdlFindScopedName(container, scoped);
			
			if(is_nil(irObj))
			{
			    yynerrs++;
			    YYERROR;
			}
			
			ConstantDef_var co = ConstantDef::_narrow(irObj);
			if(is_nil(co))
			{
			    String_var err = string_dup("`");
			    err += scoped;
			    err += "' is not a constant";
			    IdlError(err);
			    yynerrs++;
			    YYERROR;
			}
			
			$$ = co -> value();
		    }
		}
		| literal
		{
		    $$ = $1;
		}
		| '(' const_exp ')'
		{
		    $$ = $2;
		}
		;

/* (40) */
literal		: IDL_INTEGER_LITERAL
		{
		    Any* any = new Any;
		    *any <<= $1;
		    $$ = any;
		}
		| IDL_STRING_LITERAL
		{
		    String_var origString = string_dup($1);

		    if(!IdlRemoveEscapes($1, true))
		    {
			String_var err = string_dup("`");
			err += origString;
			err += "' contains \\0 character";
			IdlError(err);
			yynerrs++;
			YYERROR;
		    }

		    Any* any = new Any;
		    *any <<= $1;
		    $$ = any;
		}
		| IDL_WIDE_STRING_LITERAL
		{
		    assert_nca(false, NCANotImpl);
		}
		| IDL_CHARACTER_LITERAL
		{
		    Any* any = new Any;
		    *any <<= Any::from_char($1);
		    $$ = any;
		}
		| IDL_WIDE_CHARACTER_LITERAL
		{
		    assert_nca(false, NCANotImpl);
		}
		| IDL_FIXED_PT_LITERAL
		{
		    assert_nca(false, NCANotImpl);
		}
		| IDL_FLOATING_PT_LITERAL
		{
		    Any* any = new Any;
		    *any <<= $1;
		    $$ = any;
		}
		| boolean_literal
		{
		    Any* any = new Any;
		    *any <<= Any::from_boolean($1);
		    $$ = any;
		}
		;

/* (41) */
boolean_literal	: IDL_TRUE
		{
		    assert($1 == TRUE);
		    $$ = true;
		}
		| IDL_FALSE
		{
		    assert($1 == FALSE);
		    $$ = false;
		}
		;

/* (42) */
positive_int_const: const_exp
		{
		    Any_var any = $1;
		    TypeCode_var tc = any -> type();
		    tc = GetOrigType(tc);
		    
		    switch(tc -> kind())
		    {
		    case tk_short:
		    {
			Short v;
			any >>= v;
			if(v <= 0)
			{
			    IdlError("integer constant must be > 0");
			    yynerrs++;
			    YYERROR;
			}
			$$ = (ULong)v;
			break;
		    }

		    case tk_long:
		    {
			Long v;
			any >>= v;
			if(v <= 0)
			{
			    IdlError("integer constant must be > 0");
			    yynerrs++;
			    YYERROR;
			}
			$$ = (ULong)v;
			break;
		    }
		    
		    case tk_ushort:
		    {
			UShort v;
			any >>= v;
			$$ = (ULong)v;
			break;
		    }

		    case tk_ulong:
		    {
			ULong v;
			any >>= v;
			$$ = (ULong)v;
			break;
		    }
		    
		    default:
		    {
			IdlError("not an integer constant");
			yynerrs++;
			YYERROR;
			break;
		    }
		    }
		}
		;

/* (43) */
type_dcl	: IDL_TYPEDEF type_declarator
		{
		}
		| struct_type
		{
		    release($1);
		}
		| union_type
		{
		    release($1);
		}
		| enum_type
		{
		    release($1);
		}
		| IDL_NATIVE simple_declarator
		{
		    //
		    // Get container from container stack
		    //
		    assert(IdlContStack.length() > 0);
		    Container_var container =
			IdlContStack[IdlContStack.length() - 1];

		    IdlDeclarator_var declarator = $2;
		    assert(declarator -> sizes.length() == 0);

		    //
		    // Get name, id and version
		    //
		    Identifier_var name = declarator -> identifier;
		    RepositoryId_var id = IdlGetRepositoryId(name);
		    VersionSpec_var version = string_dup("1.0");

		    //
		    // Create native type
		    //
		    try
		    {
			release(container ->
				      create_native(id, name, version));
		    }
		    catch(const SystemException& ex)
		    {
			IdlPrintException(ex, id, name);
			yynerrs++;
			YYERROR;
		    }
		    
		    //
		    // Add ID to IdlTopFileContent and IdlComments
		    //
		    IdlAddID(id);
		}
		;

/* (44) */
type_declarator	: type_spec declarators
		{
		    //
		    // Get container from container stack
		    //
		    assert(IdlContStack.length() > 0);
		    Container_var container =
			IdlContStack[IdlContStack.length() - 1];

		    IDLType_var simpleIDLType = $1;
		    IdlDeclaratorSeq_var declarators = $2;

		    for(ULong i = 0 ; i < declarators -> length() ; i++)
		    {
			IdlDeclarator declarator = declarators[i];

			//
			// Get name, id and version
			//
			Identifier_var name = declarator.identifier;
			RepositoryId_var id = IdlGetRepositoryId(name);
			VersionSpec_var version =
			    string_dup("1.0");

			//	
			// Handle arrays
			//
			IDLType_var idlType = simpleIDLType;

			for(Long j = declarator.sizes.length() - 1 ;
			    j >= 0 ; j--)
			{
			    ArrayDef_var arrayDef = IdlRepository ->
				create_array(declarator.sizes[j], idlType);

			    idlType = IDLType::_duplicate(arrayDef);
			}

			//
			// Create alias
			//
			try
			{
			    release(container ->
					  create_alias(id, name, version,
						       idlType));
			}
			catch(const SystemException& ex)
			{
			    IdlPrintException(ex, id, name);
			    yynerrs++;
			    YYERROR;
			}

			//
			// Add ID to IdlTopFileContent and IdlComments
			//
			IdlAddID(id);
		    }
		}
		;

/* (45) */
type_spec	: simple_type_spec
		{
		    $$ = $1;
		}
		| constr_type_spec
		{
		    $$ = $1;
		}
		;

/* (46) */
simple_type_spec: base_type_spec
		{
		    $$ = $1;
		}
		| template_type_spec
		{
		    $$ = $1;
		}
		| scoped_name
		{
		    ScopedName_var scoped = $1;

		    //
		    // Get container from container stack
		    //
		    assert(IdlContStack.length() > 0);
		    Container_var container =
			IdlContStack[IdlContStack.length() - 1];

		    //
		    // Find scoped name
		    //
		    IRObject_var irObj =
			IdlFindScopedName(container, scoped);

		    if(is_nil(irObj))
		    {
			yynerrs++;
			YYERROR;
		    }

		    IDLType_var idlType = IDLType::_narrow(irObj);
		    if(is_nil(idlType))
		    {
			String_var err = string_dup("`");
			err += scoped;
			err += "' is not a type specification";
			IdlError(err);
			yynerrs++;
			YYERROR;
		    }

		    $$ = idlType._retn();
		}
		;

/* (47) */
base_type_spec	: floating_pt_type
		{
		    $$ = $1;
		}
		| integer_type
		{
		    $$ = $1;
		}
		| char_type
		{
		    $$ = $1;
		}
		| wide_char_type
		{
		    $$ = $1;
		}
		| boolean_type
		{
		    $$ = $1;
		}
		| octet_type
		{
		    $$ = $1;
		}
		| any_type
		{
		    $$ = $1;
		}
		| object_type
		{
		    $$ = $1;
		}
		| value_base_type
		{
		    $$ = $1;
		}
		;

/* (48) */
template_type_spec: sequence_type
		{
		    $$ = $1;
		}
		| string_type
		{
		    $$ = $1;
		}
		| wide_string_type
		{
		    $$ = $1;
		}
		| fixed_pt_type
		{
		    $$ = $1;
		}
		;

/* (49) */
constr_type_spec: struct_type
		{
		    $$ = $1;
		}
		| union_type
		{
		    $$ = $1;
		}
		| enum_type
		{
		    $$ = $1;
		}
		;

/* (50) */
declarators	: declarator COMMA_declarator_STAR
		{
		    IdlDeclarator_var declarator = $1;
		    $2 -> insert(declarator);
		    $$ = $2;
		}
		;

/* (51) */
COMMA_declarator_STAR: ',' declarator COMMA_declarator_STAR
		{
		    IdlDeclarator_var declarator = $2;
		    $3 -> insert(declarator);
		    $$ = $3;
		}
		| /* empty */
		{
		    $$ = new IdlDeclaratorSeq;
		}
		;

/* (51) */
declarator	: simple_declarator
		{
		    $$ = $1;
		}
		| complex_declarator
		{
		    $$ = $1;
		}
		;

/* (52) */
simple_declarator: IDL_IDENTIFIER
		{
		    $$ = new IdlDeclarator;
		    $$ -> identifier = $1;
		}
		;

/* (52) */
COMMA_simple_declarator_STAR: ',' simple_declarator
			      COMMA_simple_declarator_STAR
		{
		    IdlDeclarator_var declarator = $2;
		    $3 -> insert(declarator);
		    $$ = $3;
		}
		| /* empty */
		{
		    $$ = new IdlDeclaratorSeq;
		}
		;

/* (53) */
complex_declarator: array_declarator
		{
		    $$ = $1;
		}
		;

/* (54) */
floating_pt_type: IDL_FLOAT
		{
		    $$ = IdlRepository -> get_primitive(pk_float);
		}
		| IDL_DOUBLE
		{
		    $$ = IdlRepository -> get_primitive(pk_double);
		}
		| IDL_LONG IDL_DOUBLE
		{
		    $$ = IdlRepository -> get_primitive(pk_longdouble);
		}
		;

/* (55) */
integer_type	: signed_int
		{
		    $$ = $1;
		}
		| unsigned_int
		{
		    $$ = $1;
		}
		;

/* (56) */
signed_int	: signed_short_int
		{
		    $$ = $1;
		}
                | signed_long_int
		{
		    $$ = $1;
		}
		| signed_longlong_int
		{
		    $$ = $1;
		}
		;

/* (57) */
signed_short_int: IDL_SHORT
		{
		    $$ = IdlRepository -> get_primitive(pk_short);
		}
		;

/* (58) */
signed_long_int	: IDL_LONG
		{
		    $$ = IdlRepository -> get_primitive(pk_long);
		}
		;
/* (59) */
signed_longlong_int: IDL_LONG IDL_LONG
		{
		    $$ = IdlRepository -> get_primitive(pk_longlong);
		}
		;

/* (60) */
unsigned_int	: unsigned_short_int
		{
		    $$ = $1;
		}
                | unsigned_long_int
		{
		    $$ = $1;
		}
		| unsigned_longlong_int
		{
		    $$ = $1;
		}
		;

/* (61) */
unsigned_short_int: IDL_UNSIGNED IDL_SHORT
		{
		    $$ = IdlRepository -> get_primitive(pk_ushort);
		}
		;

/* (62) */
unsigned_long_int: IDL_UNSIGNED IDL_LONG
		{
		    $$ = IdlRepository -> get_primitive(pk_ulong);
		}
		;

/* (63) */
unsigned_longlong_int: IDL_UNSIGNED IDL_LONG IDL_LONG
		{
		    $$ = IdlRepository -> get_primitive(pk_ulonglong);
		}
		;

/* (64) */
char_type	: IDL_CHAR
		{
		    $$ = IdlRepository -> get_primitive(pk_char);
		}
		;

/* (65) */
wide_char_type	: IDL_WCHAR
		{
		    $$ = IdlRepository -> get_primitive(pk_wchar);
		}
		;

/* (66) */
boolean_type	: IDL_BOOLEAN
		{
		    $$ = IdlRepository -> get_primitive(pk_boolean);
		}
		;

/* (67) */
octet_type	: IDL_OCTET
		{
		    $$ = IdlRepository -> get_primitive(pk_octet);
		}
		;

/* (68) */
any_type	: IDL_ANY
		{
		    $$ = IdlRepository -> get_primitive(pk_any);
		}
		;

/* (69) */
object_type	: IDL_OBJECT
		{
		    $$ = IdlRepository -> get_primitive(pk_objref);
		}
		;

/* (70) */
struct_type	: IDL_STRUCT IDL_IDENTIFIER
                {
		    //
		    // Get container from container stack
		    //
		    assert(IdlContStack.length() > 0);
		    Container_var container =
			IdlContStack[IdlContStack.length() - 1];

		    //
		    // Get name, id and version
		    //
		    const char* name = $2; // No _var here!
		    RepositoryId_var id = IdlGetRepositoryId(name);
		    VersionSpec_var version = string_dup("1.0");

		    //
		    // Create empty member sequence
		    //
		    StructMemberSeq_var members = new StructMemberSeq;

		    //
		    // Create new struct
		    //
		    try
                    {
			$<structDef>$ = container ->
			    create_struct(id, name, version, members);
		    }
		    catch(const SystemException& ex)
		    {
			IdlPrintException(ex, id, name);
			yynerrs++;
			YYERROR;
		    }
		    
		    assert(!is_nil($<structDef>$));

		    //
		    // Push new container on the stack
		    //
		    IdlContStack.append(StructDef::
					_duplicate($<structDef>$));
		    
		    //
		    // Push new prefix on the prefix stacks
		    //
		    for(ULong i = 0 ; i < IdlPrefStackStack.length() ; i++)
		    {
			String_var prefix =
			    IdlPrefStackStack[i][IdlPrefStackStack[i].length()
						- 1];
			prefix += name;
			prefix += "/";
			IdlPrefStackStack[i].append(prefix);
		    }
		    
		    //
		    // Add ID to IdlTopFileContent and IdlComments
		    //
		    IdlAddID(id);
                }
                '{' member_list '}'
		{
		    $$ = $<structDef>3;

		    //
		    // Get name and id
		    //
		    Identifier_var name = $2;
		    RepositoryId_var id = IdlGetRepositoryId(name);

		    //
		    // Get member sequence
		    //
		    StructMemberSeq_var members = $5;

		    //
		    // Set the member sequence
		    //
		    try
                    {
			$$ -> members(members);
		    }
		    catch(const SystemException& ex)
		    {
			IdlPrintException(ex, id, $2);
			yynerrs++;
			YYERROR;
		    }

		    //
		    // Pop container from the stack
		    //
		    assert(IdlContStack.length() > 0);
		    IdlContStack.length(IdlContStack.length() - 1);

		    //
		    // Pop prefix from the prefix stacks
		    //
		    for(ULong i = 0 ; i < IdlPrefStackStack.length() ; i++)
		    {
			assert(IdlPrefStackStack[i].length() > 0);
			IdlPrefStackStack[i].length(
			    IdlPrefStackStack[i].length() - 1);
		    }
                }
		;

/* (71) */
member_list	: member_PLUS
		{
		    $$ = $1;
		}
		;

/* (72) */
member_PLUS	: member
		{
		    $$ = $1;
		}
		| member member_PLUS
		{
		    StructMemberSeq_var members1 = $1;
		    StructMemberSeq_var members2 = $2;

		    for(ULong i = 0 ; i < members2 -> length() ; i++)
		    {
			//
			// Check for duplicate struct members
			//
			for(ULong j = 0 ; j < members1 -> length() ; j++)
			{
			    if(strcmp(members2[i].name, members1[j].name) == 0)
			    {
				String_var err =
				    string_dup("duplicate struct "
						     "member `");
				err += members2[i].name;
				err += "'";
				IdlError(err);
				yynerrs++;
				YYERROR;
			    }
			}

			members1 -> append(members2[i]);
		    }

		    $$ = members1._retn();
		}
		;

/* (72) */
member_STAR	: member member_STAR
		{
		    StructMemberSeq_var members1 = $1;
		    StructMemberSeq_var members2 = $2;

		    for(ULong i = 0 ; i < members2 -> length() ; i++)
		    {
			//
			// Check for duplicate exception members
			//
			for(ULong j = 0 ; j < members1 -> length() ; j++)
			{
			    if(strcmp(members2[i].name, members1[j].name) == 0)
			    {
				String_var err =
				    string_dup("duplicate exception "
						     "member `");
				err += members2[i].name;
				err += "'";
				IdlError(err);
				yynerrs++;
				YYERROR;
			    }
			}

			members1 -> append(members2[i]);
		    }

		    $$ = members1._retn();
		}
		| /* empty */
		{
		    $$ = new StructMemberSeq;
		}
		;

/* (72) */
member		: member_NO_SEM ';'
		{
		    $$ = $1;
		}
		| member_NO_SEM
		{
		    $$ = $1;
		    IdlError("`;' missing after struct member");
		    yynerrs++;
		}
		;

/* (72) */
member_NO_SEM	: type_spec declarators
		{
		    IDLType_var typeSpec = $1;
		    IdlDeclaratorSeq_var declarators = $2;

		    StructMemberSeq_var memberSeq =
			new StructMemberSeq(declarators -> length());

		    for(ULong i = 0 ; i < declarators -> length() ; i++)
		    {
			IdlDeclarator declarator = declarators[i];

			//	
			// Handle arrays
			//
			IDLType_var idlType = typeSpec;

			for(Long j = declarator.sizes.length() - 1 ;
			    j >= 0 ; j--)
			{
			    ArrayDef_var arrayDef = IdlRepository ->
				create_array(declarator.sizes[j], idlType);

			    idlType = IDLType::_duplicate(arrayDef);
			}

			//
			// Create member
			//
			StructMember member;
			member.name = declarator.identifier;
			member.type = idlType -> type();
			member.type_def = idlType;

			memberSeq -> append(member);
		    }

		    $$ = memberSeq._retn();
		}
		;

/* (73) */
union_type	: IDL_UNION IDL_IDENTIFIER IDL_SWITCH '(' switch_type_spec ')'
		{
		    //
		    // Get container from container stack
		    //
		    assert(IdlContStack.length() > 0);
		    Container_var container =
			IdlContStack[IdlContStack.length() - 1];

		    //
		    // Get name, id and version
		    //
		    const char* name = $2; // No _var here!
		    RepositoryId_var id = IdlGetRepositoryId(name);
		    VersionSpec_var version = string_dup("1.0");

		    //
		    // Get union switch type
		    //
		    IDLType_ptr disc = $5; // No _var here!

		    //
		    // Create empty member sequence
		    //
		    UnionMemberSeq_var members =
		        new UnionMemberSeq;

		    //
		    // Create new union
		    //
		    try
                    {
			$<unionDef>$ = container ->
			    create_union(id, name, version, disc, members);
		    }
		    catch(const SystemException& ex)
		    {
			IdlPrintException(ex, id, name);
			yynerrs++;
			YYERROR;
		    }
		    
		    assert(!is_nil($<unionDef>$));

		    //
		    // Push new container on the stack
		    //
		    IdlContStack.append(UnionDef::
					_duplicate($<unionDef>$));
		    
		    //
		    // Push new prefix on the prefix stacks
		    //
		    for(ULong i = 0 ; i < IdlPrefStackStack.length() ; i++)
		    {
			String_var prefix =
			    IdlPrefStackStack[i][IdlPrefStackStack[i].length()
						- 1];
			prefix += name;
			prefix += "/";
			IdlPrefStackStack[i].append(prefix);
		    }
		    
		    // Add ID to IdlTopFileContent and IdlComments
		    //
		    IdlAddID(id);

		    //
		    // Set IdlUnionSwitchType
		    //
		    // assert(is_nil(IdlUnionSwitchType));
		    // ^^^ Don't do this. In case of an syntax error
		    // IdlUnionSwitchType will not be set to nil.
		    //
		    IdlUnionSwitchType = IDLType::_duplicate(disc);
		}
		'{' switch_body '}'
		{
		    $$ = $<unionDef>7;

		    //
		    // Reset IdlUnionSwitchType
		    //
		    release(IdlUnionSwitchType);
		    IdlUnionSwitchType = IDLType::_nil();
		    
		    //
		    // Get name and id
		    //
		    Identifier_var name = $2;
		    RepositoryId_var id = IdlGetRepositoryId(name);

		    //
		    // Get union switch type
		    //
		    IDLType_var disc = $5;

		    //
		    // Get member sequence
		    //
		    UnionMemberSeq_var members = $9;

		    bool haveDefault = false;
		    ULong i;
		    for(i = 0 ; i < members -> length() ; i++)
		    {
			//
			// Check for multiple default labels
			//
			Octet oct;
			Boolean b =
			    members[i].label >>= Any::to_octet(oct);

			if(b)
			{
			    assert(oct == 0);

			    if(haveDefault)
			    {
				IdlError("multiple default labels");
				yynerrs++;
				YYERROR;
			    }

			    haveDefault = true;
			}
			else
			{
			    //
			    // Check for duplicate labels
			    //
			    for(ULong j = 0 ;
				j < members -> length() ; j++)
			    {
				if(i == j)
				    continue;

				TypeCode_var tc1 =
				    members[i].label.type();
				TypeCode_var tc2 =
				    members[j].label.type();

				tc1 = GetOrigType(tc1);
				tc2 = GetOrigType(tc2);

				if(tc2 -> kind() == tk_octet)
				    continue;

				assert(tc1 -> equivalent(tc2)); 

				const void* v1 = members[i].label.value();
				const void* v2 = members[j].label.value();

				switch(tc1 -> kind())
				{
				case tk_short:
				    if(*(Short*)v1 ==
				       *(Short*)v2)
				    {
					String_var err =
					    string_dup("duplicate label "
							     "`");
					err += *(Short*)v1;
					err += "'";
					IdlError(err);
					yynerrs++;
					YYERROR;
				    }
				    break;

				case tk_ushort:
				    if(*(UShort*)v1 ==
				       *(UShort*)v2)
				    {
					String_var err =
					    string_dup("duplicate label "
							     "`");
					err += *(UShort*)v1;
					err += "'";
					IdlError(err);
					yynerrs++;
					YYERROR;
				    }
				    break;

				case tk_long:
				    if(*(Long*)v1 ==
				       *(Long*)v2)
				    {
					String_var err =
					    string_dup("duplicate label "
							     "`");
					err += *(Long*)v1;
					err += "'";
					IdlError(err);
					yynerrs++;
					YYERROR;
				    }
				    break;

				case tk_ulong:
				    if(*(ULong*)v1 ==
				       *(ULong*)v2)
				    {
					String_var err =
					    string_dup("duplicate label "
							     "`");
					err += *(ULong*)v1;
					err += "'";
					IdlError(err);
					yynerrs++;
					YYERROR;
				    }
				    break;

				case tk_enum:
				    if(*(ULong*)v1 ==
				       *(ULong*)v2)
				    {
					String_var err =
					    string_dup("duplicate label "
							     "`");
					String_var s =
					    tc1 ->
					    member_name(*(ULong*)v1);
					err += "'";
					IdlError(err);
					yynerrs++;
					YYERROR;
				    }
				    break;
				    
				case tk_char:
				    if(*(Char*)v1 ==
				       *(Char*)v2)
				    {
					String_var err =
					    string_dup("duplicate label "
							     "`");
					err += (char)*(Char*)v1;
					err += "'";
					IdlError(err);
					yynerrs++;
					YYERROR;
				    }
				    break;

				case tk_boolean:
				    if(*(Boolean*)v1 ==
				       *(Boolean*)v2)
				    {
					String_var err =
					    string_dup("duplicate label "
							     "`");
					err += *(Boolean*)v1 ?
					    "TRUE" : "FALSE";
					err += "'";
					IdlError(err);
					yynerrs++;
					YYERROR;
				    }
				    break;
				    
				default:
				    assert(false);
				    break;
				}
			    }
			}
		    }

		    //
		    // Check for illegal default label
		    //
		    if(haveDefault)
		    {
			TypeCode_var tc = disc -> type();
			tc = GetOrigType(tc);

			if((tc -> kind() == tk_char &&
			    members -> length() > 256) ||
			   (tc -> kind() == tk_boolean &&
			    members -> length() > 2) ||
			   (tc -> kind() == tk_enum &&
			    members -> length() > tc -> member_count()))
			{
			    IdlError("No value available for an explicit "
				     "default label");
			    yynerrs++;
			    YYERROR;
			}
		    }

		    //
		    // Set the member sequence
		    //
		    try
                    {
			$$ -> members(members);
		    }
		    catch(const SystemException& ex)
		    {
			IdlPrintException(ex, id, $2);
			yynerrs++;
			YYERROR;
		    }

		    //
		    // Pop container from the stack
		    //
		    assert(IdlContStack.length() > 0);
		    IdlContStack.length(IdlContStack.length() - 1);

		    //
		    // Pop prefix from the prefix stacks
		    //
		    for(i = 0 ; i < IdlPrefStackStack.length() ; i++)
		    {
			assert(IdlPrefStackStack[i].length() > 0);
			IdlPrefStackStack[i].length(
			    IdlPrefStackStack[i].length() - 1);
		    }
		}
		;

/* (74) */
switch_type_spec: integer_type
		{
		    $$ = $1;
		}
		| char_type
		{
		    $$ = $1;
		}
		| boolean_type
		{
		    $$ = $1;
		}
		| enum_type
		{
		    $$ = $1;
		}
		| scoped_name
		{
		    ScopedName_var scoped = $1;

		    //
		    // Get container from container stack
		    //
		    assert(IdlContStack.length() > 0);
		    Container_var container =
			IdlContStack[IdlContStack.length() - 1];

		    //
		    // Find scoped name
		    //
		    IRObject_var irObj =
			IdlFindScopedName(container, scoped);

		    if(is_nil(irObj))
		    {
			yynerrs++;
			YYERROR;
		    }

		    bool valid = false;

		    IDLType_var idlType = IDLType::_narrow(irObj);
		    if(!is_nil(idlType))
		    {
			TypeCode_var tc = idlType -> type();
			tc = GetOrigType(tc);

			TCKind k = tc -> kind();

			if(k == tk_short || k == tk_ushort ||
			   k == tk_long || k == tk_ulong ||
			   k == tk_char || k == tk_boolean ||
			   k == tk_enum)
			    valid = true;
		    }

		    if(!valid)
		    {
			String_var err = string_dup("`");
			err += scoped;
			err += "' is not a valid union discriminator type";
			IdlError(err);
			yynerrs++;
			YYERROR;
		    }

		    $$ = idlType._retn();
		}
		;

/* (75) */
switch_body	: case_PLUS
		{
		    $$ = $1;
		}
		;

/* (76) */
case_PLUS	: case
		{
		    $$ = $1;
		}
		| case case_PLUS
		{
		    UnionMemberSeq_var members1 = $1;
		    UnionMemberSeq_var members2 = $2;

		    //
		    // Check for duplicate union members
		    //
		    for(ULong i = 0 ; i < members2 -> length() ; i++)
		    {
			//
			// It's only necessary to compare against the
			// first element of members1, since all
			// elements in members1 are equal.
			//
			if(strcmp(members2[i].name, members1[0].name) == 0)
			{
			    String_var err =
				string_dup("duplicate union member `");
			    err += members2[i].name;
			    err += "'";
			    IdlError(err);
			    yynerrs++;
			    YYERROR;
			}

			members1 -> append(members2[i]);
		    }

		    $$ = members1._retn();
		}
		;

/* (76) */
case		: case_NO_SEM ';'
		{
		    $$ = $1;
		}
		| case_NO_SEM
		{
		    $$ = $1;
		    IdlError("`;' missing after union member");
		    yynerrs++;
		}
		;

/* (76) */
case_NO_SEM	: case_label_PLUS element_spec
		{
		    IdlAnySeq_var anySeq = $1;
		    UnionMember_var member = $2;

		    UnionMemberSeq_var memberSeq =
			new UnionMemberSeq(anySeq -> length());
		    memberSeq -> length(anySeq -> length());

		    for(ULong i = 0 ; i < anySeq -> length() ; i++)
		    {
			memberSeq[i] = member;
			memberSeq[i].label = anySeq[i];
		    }

		    $$ = memberSeq._retn();
		}
		;

/* (77) */
case_label_PLUS	: case_label
		{
		    Any_var any = $1;
		    $$ = new IdlAnySeq;
		    $$ -> insert(any);
		}
		| case_label case_label_PLUS
		{
		    Any_var any = $1;
		    $2 -> insert(any);
		    $$ = $2;
		}
		;

/* (77) */
case_label	: IDL_CASE const_exp ':'
		{
		    Any_var result = $2;

		    //
		    // Check if discriminator type is matched
		    //
		    assert(!is_nil(IdlUnionSwitchType));
		    TypeCode_var tc1 = IdlUnionSwitchType -> type();
		    tc1 = GetOrigType(tc1);
		    TypeCode_var tc2 = result -> type();
		    tc2 = GetOrigType(tc2);

		    if(!tc1 -> equivalent(tc2))
		    {
			//
			// Check if conversion is possible
			//
			ULong ul;
			Boolean ulb;
			Long l;
			Boolean lb;
			Double d;
			Boolean db;
			TranslateAny(result, ul, ulb, l, lb, d, db);

			const char* err =
			    "case label type doesn't "
			    "match the discriminator type";

			//
			// Convert value type
			//
			switch(tc1 -> kind())
			{
			case tk_short:
			    
			    if(ulb)
				result <<= (Short)ul;
			    else if(lb)
				result <<= (Short)l;
			    else
			    {
				IdlError(err);
				yynerrs++;
				YYERROR;
			    }
			    
			    break;
			    
			case tk_long:
			    
			    if(ulb)
				result <<= (Long)ul;
			    else if(lb)
				result <<= (Long)l;
			    else
			    {
				IdlError(err);
				yynerrs++;
				YYERROR;
			    }
			    
			    break;
			    
			case tk_ushort:
			    
			    if(ulb)
				result <<= (UShort)ul;
			    else
			    {
				IdlError(err);
				yynerrs++;
				YYERROR;
			    }
			    
			    break;
			    
			case tk_ulong:
			    
			    if(ulb)
				result <<= (ULong)ul;
			    else
			    {
				IdlError(err);
				yynerrs++;
				YYERROR;
			    }
			    
			    break;
			    
			default:
			    IdlError(err);
			    yynerrs++;
			    YYERROR;
			    break;
			}
		    }

		    $$ = result._retn();
		}
		| IDL_DEFAULT ':'
		{
		    $$ = new Any;
		    *$$ <<= Any::from_octet(0);
		}
		;

/* (78) */
element_spec	: type_spec declarator
		{
		    IDLType_var typeSpec = $1;
		    IdlDeclarator_var declarator = $2;

		    //	
		    // Handle arrays
		    //
		    IDLType_var idlType = typeSpec;
		    
		    for(Long j = declarator -> sizes.length() - 1 ;
			j >= 0 ; j--)
		    {
			ArrayDef_var arrayDef = IdlRepository ->
			    create_array(declarator -> sizes[j], idlType);

			idlType = IDLType::_duplicate(arrayDef);
		    }
		    
		    //
		    // Create member
		    //
		    $$ = new UnionMember;
		    $$ -> name = declarator -> identifier;
		    $$ -> type = idlType -> type();
		    $$ -> type_def = idlType;
		}
		;

/* (79) */
enum_type	: IDL_ENUM IDL_IDENTIFIER '{' enumerator
		  COMMA_enumerator_STAR '}'
		{
		    //
		    // Get container from container stack
		    //
		    assert(IdlContStack.length() > 0);
		    Container_var container =
			IdlContStack[IdlContStack.length() - 1];

		    //
		    // Get member sequence
		    //
                    //XXX
		    //EnumMemberSeq_var members = $5;
		    //members -> insert($4);
		    EnumMemberSeq_var members =
			new EnumMemberSeq($5 -> length());
                    members -> length($5 -> length());
		    for(ULong i = 0; i < $5 -> length(); ++i)
                    {
			members[i] = string_dup((*$5)[i]);
                    }
                    members -> insert($4);

		    //
		    // Get name, id and version
		    //
		    Identifier_var name = $2;
		    RepositoryId_var id = IdlGetRepositoryId(name);
		    VersionSpec_var version = string_dup("1.0");

		    //
		    // Create enumerator
		    //
		    try
		    {
			$$ = container -> create_enum(id, name, version,
						      members);
		    }
		    catch(const SystemException& ex)
		    {
			IdlPrintException(ex, id, name);
			yynerrs++;
			YYERROR;
		    }

		    //
		    // Add ID to IdlTopFileContent and IdlComments
		    //
		    IdlAddID(id);
		}
		;

/* (80) */
COMMA_enumerator_STAR: ',' enumerator COMMA_enumerator_STAR
		{
		    Identifier_var member = $2;
		    IdlStringSeq* members = $3;
		    
		    //
		    // Check for duplicate enum members
		    //
		    for(ULong i = 0 ; i < members -> length() ; i++)
		    {
			if(strcmp((*members)[i], member) == 0)
			{
			    String_var err =
				string_dup("duplicate enum member `");
			    err += member;
			    err += "'";
			    IdlError(err);
			    yynerrs++;
			    YYERROR;
			}
		    }

		    members -> insert(member);
		    $$ = members;//XXX ._retn();
		}
		| /* empty */
		{
		    //$$ = new EnumMemberSeq;
		    $$ = new IdlStringSeq;
		}
		| ','
		{
		    IdlError("unexpected `,'");
		    yynerrs++;
		    YYERROR;
		}
		;

/* (80) */
enumerator	: IDL_IDENTIFIER
		{
		    $$ = $1;
		}
		;

/* (81) */
sequence_type	: IDL_SEQUENCE '<' simple_type_spec ',' positive_int_const '>'
		{
		    IDLType_var idlType = $3;
		    ULong bound = $5;

		    //
		    // Create sequence
		    //
		    $$ = IdlRepository -> create_sequence(bound, idlType);
		}
		| IDL_SEQUENCE '<' simple_type_spec '>'
		{
		    IDLType_var idlType = $3;

		    //
		    // Create sequence
		    //
		    $$ = IdlRepository -> create_sequence(0, idlType);
		}
		;

/* (82) */
string_type	: IDL_STRING '<' positive_int_const '>'
		{
		    ULong bound = $3;

		    //
		    // Create string
		    //
		    $$ = IdlRepository -> create_string(bound);
		}
		| IDL_STRING
		{
		    //
		    // Get string primitive
		    //
		    $$ = IdlRepository -> get_primitive(pk_string);
		}
		;

/* (83) */
wide_string_type: IDL_WSTRING '<' positive_int_const '>'
		{
		    ULong bound = $3;

		    //
		    // Create string
		    //
		    $$ = IdlRepository -> create_wstring(bound);
		}
		| IDL_WSTRING
		{
		    //
		    // Get string primitive
		    //
		    $$ = IdlRepository -> get_primitive(pk_wstring);
		}
		;

/* (84) */
array_declarator: IDL_IDENTIFIER fixed_array_size_PLUS
		{
		    $$ = $2;
		    $$ -> identifier = $1;
		}
		;

/* (85) */
fixed_array_size_PLUS: fixed_array_size
		{
		    $$ = new IdlDeclarator;
		    $$ -> sizes.insert($1);
		}
		| fixed_array_size fixed_array_size_PLUS
		{
		    $$ = $2;
		    $$ -> sizes.insert($1);
		}
		;

/* (85) */
fixed_array_size: '[' positive_int_const ']'
		{
		    $$ = $2;
		}
		;

/* (86) */
attr_dcl	: readonly_OPT IDL_ATTRIBUTE param_type_spec
		  simple_declarator COMMA_simple_declarator_STAR
		{
		    //
		    // Get container from container stack
		    //
		    assert(IdlContStack.length() > 0);
		    Container_var container =
			IdlContStack[IdlContStack.length() - 1];

		    InterfaceDef_var interf =
			InterfaceDef::_narrow(container);
		    ValueDef_var value =
			ValueDef::_narrow(container);
		    assert(!is_nil(interf) || !is_nil(value));

		    //
		    // Get attribute mode
		    //
		    AttributeMode mode = $1;

		    //
		    // Get parameter/return type
		    //
		    IDLType_var idlType = $3;

		    //
		    // Get declarators
		    //
		    IdlDeclarator_var declarator = $4;
		    IdlDeclaratorSeq_var declarators = $5;
		    declarators -> insert(declarator);

		    for(ULong i = 0 ; i < declarators -> length() ; i++)
		    {
			assert(declarators[i].sizes.length() == 0);

			//
			// Get name, id and version
			//
			Identifier_var name = declarators[i].identifier;
			RepositoryId_var id = IdlGetRepositoryId(name);
			VersionSpec_var version =
			    string_dup("1.0");

			//
			// Create attribute
			//
			try
			{
			    if(! is_nil(interf))
			        release(interf ->
					  create_attribute(id, name, version,
							   idlType, mode));
			    else
			        release(value ->
					  create_attribute(id, name, version,
							   idlType, mode));
			}
			catch(const SystemException& ex)
			{
			    IdlPrintException(ex, id, name);
			    yynerrs++;
			    YYERROR;
			}

			//
			// Add ID to IdlTopFileContent and IdlComments
			//
			IdlAddID(id);
		    }
		}
		;

/* (87) */
except_dcl	: IDL_EXCEPTION IDL_IDENTIFIER
                {
		    //
		    // Get container from container stack
		    //
		    assert(IdlContStack.length() > 0);
		    Container_var container =
			IdlContStack[IdlContStack.length() - 1];

		    //
		    // Get name, id and version
		    //
		    const char* name = $2; // No _var here!
		    RepositoryId_var id = IdlGetRepositoryId(name);
		    VersionSpec_var version = string_dup("1.0");

		    //
		    // Create empty member sequence
		    //
		    StructMemberSeq_var members =
		        new StructMemberSeq;

		    //
		    // Create new exception
		    //
		    try
                    {
			$<exceptionDef>$ = container ->
			    create_exception(id, name, version, members);
		    }
		    catch(const SystemException& ex)
		    {
			IdlPrintException(ex, id, name);
			yynerrs++;
			YYERROR;
		    }
		    
		    assert(!is_nil($<exceptionDef>$));

		    //
		    // Push new container on the stack
		    //
		    IdlContStack.append(ExceptionDef::
					_duplicate($<exceptionDef>$));
		    
		    //
		    // Push new prefix on the prefix stacks
		    //
		    for(ULong i = 0 ; i < IdlPrefStackStack.length() ; i++)
		    {
			String_var prefix =
			    IdlPrefStackStack[i][IdlPrefStackStack[i].length()
						- 1];
			prefix += name;
			prefix += "/";
			IdlPrefStackStack[i].append(prefix);
		    }
		    
		    //
		    // Add ID to IdlTopFileContent and IdlComments
		    //
		    IdlAddID(id);
                }
                '{' member_STAR '}'
		{
		    ExceptionDef_var ex = $<exceptionDef>3;

		    //
		    // Get name and id
		    //
		    Identifier_var name = $2;
		    RepositoryId_var id = IdlGetRepositoryId(name);

		    //
		    // Get member sequence
		    //
		    StructMemberSeq_var members = $5;

		    //
		    // Set the member sequence
		    //
		    try
                    {
			ex -> members(members);
		    }
		    catch(const SystemException& ex)
		    {
			IdlPrintException(ex, id, $2);
			yynerrs++;
			YYERROR;
		    }

		    //
		    // Pop container from the stack
		    //
		    assert(IdlContStack.length() > 0);
		    IdlContStack.length(IdlContStack.length() - 1);

		    //
		    // Pop prefix from the prefix stacks
		    //
		    for(ULong i = 0 ; i < IdlPrefStackStack.length() ; i++)
		    {
			assert(IdlPrefStackStack[i].length() > 0);
			IdlPrefStackStack[i].length(
			    IdlPrefStackStack[i].length() - 1);
		    }
                }
		;

/* (88) */
op_dcl		: op_attribute_OPT op_type_spec IDL_IDENTIFIER
		  parameter_dcls raises_expr_OPT context_expr_OPT
		{
		    //
		    // Get container from container stack
		    //
		    assert(IdlContStack.length() > 0);
		    Container_var container =
			IdlContStack[IdlContStack.length() - 1];

		    InterfaceDef_var interf =
			InterfaceDef::_narrow(container);
		    ValueDef_var value =
			ValueDef::_narrow(container);

		    assert(!is_nil(interf) || !is_nil(value));

		    //
		    // Get operation mode
		    //
		    OperationMode mode = $1;

		    //
		    // Get return type
		    //
		    IDLType_var result = $2;

		    //
		    // Get member sequence
		    //
		    ParDescriptionSeq_var params = $4;

		    //
		    // Get exceptions definition sequence
		    //
		    ExceptionDefSeq_var exceptions = $5;

		    //
		    // Get context ID sequence
		    //
                    //XXX
		    //ContextIdSeq_var contexts = $6;
		    ContextIdSeq_var contexts =
			new ContextIdSeq($6 -> length());
                    contexts -> length($6 -> length());
		    for(ULong i = 0; i < $6 -> length(); ++i)
			contexts[i] = string_dup((*$6)[i]);

		    //
		    // Get name, id and version
		    //
		    Identifier_var name = $3;
		    RepositoryId_var id = IdlGetRepositoryId(name);
		    VersionSpec_var version = string_dup("1.0");

		    //
		    // Check for oneway constraints
		    //
		    if(mode == OP_ONEWAY)
		    {
			TypeCode_var resultTC = result -> type();
			if(resultTC -> kind() != tk_void)
			{
			    IdlError("oneway operations cannot have "
				     "return value");
			    yynerrs++;
			    YYERROR;
			}

			for(ULong i = 0 ; i < params -> length() ; i++)
			    if(params[i].mode != PARAM_IN)
			    {
				IdlError("oneway operations can only have "
					 "`in' parameters");
				yynerrs++;
				YYERROR;
			    }

			if(exceptions -> length() > 0)
			{
			    IdlError("oneway operations cannot have "
				     "user defined exceptions");
			    yynerrs++;
			    YYERROR;
			}
		    }

		    //
		    // Create operation
		    //
		    try
		    {
			if(! is_nil(interf))
			    release(interf ->
				      create_operation(id, name, version,
						       result, mode, params,
						       exceptions, contexts));
			else
			    release(value ->
				      create_operation(id, name, version,
						       result, mode, params,
						       exceptions, contexts));
		    }
		    catch(const SystemException& ex)
		    {
			IdlPrintException(ex, id, name);
			yynerrs++;
			YYERROR;
		    }

		    //
		    // Add ID to IdlTopFileContent and IdlComments
		    //
		    IdlAddID(id);
		}
		;

/* (89) */
op_attribute_OPT: op_attribute
		{
		     $$ = $1;
		}
		| /* empty */
		{
		     $$ = OP_NORMAL;
		}
		;

/* (89) */
op_attribute	: IDL_ONEWAY
		{
		    $$ = OP_ONEWAY;
		}
		;

/* (90) */
op_type_spec	: param_type_spec
		{
		    $$ = $1;
		}
		| IDL_VOID
		{
		    $$ = IdlRepository -> get_primitive(pk_void);
		}
		;

/* (91) */
parameter_dcls	: '(' param_dcl COMMA_param_dcl_STAR ')'
		{
		    ParameterDescription_var param = $2;
		    $3 -> insert(param);
		    $$ = $3;
		}
		| '(' ')'
		{
		    $$ = new ParDescriptionSeq;
		}
		;

/* (92) */
COMMA_param_dcl_STAR: ',' param_dcl COMMA_param_dcl_STAR
		{
		    ParameterDescription_var param = $2;
		    $3 -> insert(param);
		    $$ = $3;
		}
		| /* empty */
		{
		    $$ = new ParDescriptionSeq;
		}
		;

/* (92) */
param_dcl	: param_attribute param_type_spec simple_declarator
		{
		    ParameterMode mode = $1;
		    IDLType_var typeSpec = $2;
		    IdlDeclarator_var declarator = $3;
		    assert(declarator -> sizes.length() == 0);

		    ParameterDescription_var param =
			new ParameterDescription;

		    param -> name = declarator -> identifier;
		    param -> type = typeSpec -> type();
		    param -> type_def = typeSpec;
		    param -> mode = mode;

		    $$ = param._retn();
		}
		| param_type_spec
		{
		    IDLType_var typeSpec = $1;

		    IdlError("`in', `inout' or `out' missing");
		    IdlError("parameter name missing");
		    yynerrs++;
		    YYERROR;
		}
		| param_type_spec simple_declarator
		{
		    IDLType_var typeSpec = $1;
		    IdlDeclarator_var declarator = $2;

		    IdlError("`in', `inout' or `out' missing");
		    yynerrs++;
		    YYERROR;
		}
		| param_attribute param_type_spec
		{
		    IDLType_var typeSpec = $2;

		    IdlError("parameter name missing");
		    yynerrs++;
		    YYERROR;
		}
		;

/* (93) */
param_attribute	: IDL_IN
		{
		    $$ = PARAM_IN;
		}
		| IDL_OUT
		{
		    $$ = PARAM_OUT;
		}
		| IDL_INOUT
		{
		    $$ = PARAM_INOUT;
		}
		;

/* (94) */
raises_expr_OPT	: raises_expr
		{
		    $$ = $1;
		}
		| /* empty */
		{
		    $$ = new ExceptionDefSeq;
		}
		;

/* (94) */
raises_expr	: IDL_RAISES '(' scoped_name COMMA_scoped_name_STAR ')'
		{
		    IdlStringSeq_var scopedNameSeq = $4;
		    scopedNameSeq -> insert($3);

		    //
		    // Get container from container stack
		    //
		    assert(IdlContStack.length() > 0);
		    Container_var container =
			IdlContStack[IdlContStack.length() - 1];

		    //
		    // Create exception definition sequence
		    //
		    ExceptionDefSeq_var exDefSeq =
			new ExceptionDefSeq(scopedNameSeq -> length());

		    for(ULong i = 0 ; i < scopedNameSeq -> length() ; i++)
		    {
			//
			// Find scoped name
			//
			const char* scoped = scopedNameSeq[i];
			IRObject_var irObj =
			    IdlFindScopedName(container, scoped);

			if(is_nil(irObj))
			{
			    yynerrs++;
			    YYERROR;
			}

			ExceptionDef_var exceptionDef =
			    ExceptionDef::_narrow(irObj);

			if(is_nil(exceptionDef))
			{	
			    String_var err = string_dup("`");
			    err += scoped;
			    err += "' is not an exception type";
			    IdlError(err);
			    yynerrs++;
			    YYERROR;
			}

			exDefSeq -> append(exceptionDef);
		    }

		    $$ = exDefSeq._retn();
		}
		;

/* (95) */
context_expr_OPT: context_expr
		{
		    $$ = $1;
		}
		| /* empty */
		{
		    //$$ = new ContextIdSeq;
		    $$ = new IdlStringSeq;
		}
		;

/* (95) */
context_expr	: IDL_CONTEXT '(' IDL_STRING_LITERAL COMMA_string_literal_STAR ')'
		{
		    String_var string = $3;
		    String_var origString = string_dup(string);
		    IdlStringSeq_var result = $4;

		    if(!IdlRemoveEscapes(string.inout(), true))
		    {
			String_var err = string_dup("`");
			err += origString;
			err += "' contains \\0 character";
			IdlError(err);
			yynerrs++;
			YYERROR;
		    }

		    String_var err2 = string_dup("`");
		    err2 += origString;
		    err2 += "' is not a valid context name";

		    if(strlen(string) == 0)
		    {
			IdlError(err2);
			yynerrs++;
			YYERROR;
		    }

		    for(ULong i = 0 ; i < strlen(string) ; i++)
		    {
			char c = string[i];

			if(isalpha(c))
			    continue;

			if(i != 0 && (isdigit(c) || c == '_' || c == '.'))
			    continue;

			if(i == strlen(string) - 1 && c == '*')
			    continue;

			IdlError(err2);
			yynerrs++;
			YYERROR;
		    }

		    result -> insert(string);
		    $$ = result._retn();
		}
		;

/* (96) */
param_type_spec	: base_type_spec
		{
		    $$ = $1;
		}
		| sequence_type
		{
		    IDLType_var typeSpec = $1;

		    IdlError("typedefs must be used for sequence type "
			     "parameters");
		    yynerrs++;
		    YYERROR;
		}
		| string_type
		{
		    $$ = $1;
		}
		| wide_string_type
		{
		    $$ = $1;
		}
		| scoped_name
		{
		    ScopedName_var scoped = $1;

		    //
		    // Get container from container stack
		    //
		    assert(IdlContStack.length() > 0);
		    Container_var container =
			IdlContStack[IdlContStack.length() - 1];

		    //
		    // Find scoped name
		    //
		    IRObject_var irObj =
			IdlFindScopedName(container, scoped);

		    if(is_nil(irObj))
		    {
			yynerrs++;
			YYERROR;
		    }

		    IDLType_var idlType = IDLType::_narrow(irObj);
		    if(is_nil(idlType))
		    {
			String_var err = string_dup("`");
			err += scoped;
			err += "' is not a parameter type specification";
			IdlError(err);
			yynerrs++;
			YYERROR;
		    }

		    $$ = idlType._retn();
		}
		;

/* (97) */
fixed_pt_type   : IDL_FIXED '<' positive_int_const ',' IDL_INTEGER_LITERAL '>'
		{
		    UShort digits = (UShort)$3;
		    Short scale = (Short)$5;

		    //
		    // Create fixed type
		    //
		    $$ = IdlRepository -> create_fixed(digits, scale);
		}
                ;

/* (98) */
fixed_pt_const_type: IDL_FIXED
		{
		    assert_nca(false, NCANotImpl);
		}
                ;

/* (99) */
value_base_type	: IDL_VALUEBASE
		{
		    $$ = IdlRepository -> get_primitive(pk_ValueBase);
		}
		;

/* other */
readonly_OPT	: IDL_READONLY
		{
		    $$ = ATTR_READONLY;
		}
		| /* empty */
		{
		    $$ = ATTR_NORMAL;
		}
		;

/* other */
COMMA_string_literal_STAR: ',' IDL_STRING_LITERAL COMMA_string_literal_STAR
		{
		    String_var string = $2;
		    String_var origString = string_dup(string);
		    IdlStringSeq_var result = $3;

		    if(!IdlRemoveEscapes(string.inout(), true))
		    {
			String_var err = string_dup("`");
			err += origString;
			err += "' contains \\0 character";
			IdlError(err);
			yynerrs++;
			YYERROR;
		    }

		    result -> insert(string);
		    $$ = result._retn();
		}
		| /* empty */
		{
		    $$ = new IdlStringSeq;
		}
		;

/* other */
abstract_OPT	: IDL_ABSTRACT
		{
		    $$ = TRUE;
		}
		| /* empty */
		{
		    $$ = FALSE;
		}
		;
 
/* other */
truncatable_OPT	: IDL_TRUNCATABLE
		{
		    $$ = TRUE;
		}
		| /* empty */
		{
		    $$ = FALSE;
		}
		;

/* other */
public_OR_private: IDL_PUBLIC
		{
		    $$ = PUBLIC;
		}
		| IDL_PRIVATE
		{
		    $$ = PRIVATE;
		}
		;

%%

// ----------------------------------------------------------------------
// Do parsing
// ----------------------------------------------------------------------

int
IdlParse(ORB_ptr orb,
	 Repository_ptr repository,
	 bool ignoreUndefined,
	 RepositoryIdSeq*& topFileContent,
	 IdlStringSeq*& includes,
	 IdlCommentSeq*& comments)
{
    //
    // Initialize global variables
    //   
    IdlLineno = 1;
    IdlDocComment = string_dup("");
    IdlORB = orb;
    IdlRepository = repository;
    IdlTopFileContent = new RepositoryIdSeq;
    IdlIncludes = new IdlStringSeq;
    IdlComments = new IdlCommentSeq;
    IdlUnionSwitchType = IDLType::_nil();
    IdlFwdInterfaces.length(0);    

    //
    // Set out parameters
    //
    topFileContent = IdlTopFileContent;
    includes = IdlIncludes;
    comments = IdlComments;

    //
    // Add empty file name to file name stack
    // Add empty prefix stack to the prefix stack stack
    //
    IdlFileStack.append(string_dup(""));
    IdlPrefStackStack.length(1);

    //
    // Parse
    //
    yynerrs = 0;
    int yyparse();
    int status = yyparse();
    assert(yynerrs > 0 || status == 0);

    //
    // Do some cleanup
    //
    IdlFileStack.length(0);
    IdlPrefStackStack.length(0);

    //
    // Check whether there are some interface that have been forward
    // declared but never defined
    //
    if(!ignoreUndefined)
    {
	for(ULong i = 0 ; i < IdlFwdInterfaces.length() ; i++)
	{
	    //
	    // Don't use IdlError here, because filename and linenumber
	    // are not defined
	    //
	    fflush(stderr);
	    fprintf(stderr,
		    "interface `%s' has been declared but not defined\n",
		    IdlFwdInterfaces[i].in());
	    
	    yynerrs++;
	}
    }

    //
    // Important note:
    //
    // If IdlParse returns with a status != 0, it is most certain that
    // there are memory leaks. The reason for that is that in case of
    // an error, it is not possible to free any values of the yyval
    // union that have been created with "new".
    //

    //
    // Return the status
    //
    return yynerrs;
}

// ----------------------------------------------------------------------
// Translate an Any into ULong, Long and Double
// ----------------------------------------------------------------------

static void TranslateAny(const Any& any,
			 ULong& ul, Boolean& ulBool,
			 Long& l, Boolean& lBool,
			 Double& d, Boolean& dBool)
{
    ulBool = any >>= ul;
    lBool = any >>= l;
    dBool = any >>= d;

    //
    // If it's not a Long, or ULong, or a Double check if it's a short
    // or an unsigned short. If it's either of these then assign the
    // value to the long.
    //
    if(!ulBool && !lBool && !dBool)
    {
	Short u;
	UShort us;
	if(any >>= u)
	{
	    lBool = true;
	    l = u;
	}
	else if(any >>= us)
	{
	    lBool = true;
	    l = us;
	}
    }

    if(!lBool)
	if(ulBool)
	    l = (Long)ul;

    if(!dBool)
    {
	if(ulBool)
	    d = (Double)ul;
	else if(lBool)
	    d = (Double)l;
    }
}

// ----------------------------------------------------------------------
// Create valuetype
// ----------------------------------------------------------------------

bool
CreateValue(const char* name, const IdlValueInfo& info, Boolean isCustom,
            Boolean isAbstract)
{
    //
    // Get container from container stack
    //
    assert(IdlContStack.length() > 0);
    Container_var container =
        IdlContStack[IdlContStack.length() - 1];

    //
    // Get id and version
    //
    RepositoryId_var id = IdlGetRepositoryId(name);
    VersionSpec_var version = string_dup("1.0");

    //
    // Check if name already exists
    //
    Contained_var c;
    try
    {
        c = container -> lookup(name);
    }
    catch(const SystemException& ex)
    {
        IdlPrintException(ex, id, name);
        yynerrs++;
        return false;
    }

    ValueDef_var value;
    if(!is_nil(c))
    {
        //
        // Name exists, check if it is a valuetype forward declaration
        //
        value = ValueDef::_narrow(c);

        if(is_nil(value))
        {        
            String_var err = string_dup("conflicting types for `valuetype ");
            err += name;
            err += "'";
            IdlError(err);
            yynerrs++;
            return false;
        }

        //
        // Don't use ID for comparison, because it's
        // possible that the ID was changed with
        // #pragma
        //
        ScopedName_var abs = value -> absolute_name();
        ULong i;
        for(i = 0 ; i < IdlFwdValuetypes.length() ; i++)
        {
            if(strcmp(abs, IdlFwdValuetypes[i]) == 0)
                break;
        }

        if(i >= IdlFwdValuetypes.length())
        {
            String_var err = string_dup("redefinition of `valuetype ");
            err += name;
            err += "'";
            IdlError(err);
            yynerrs++;
            return false;
        }

        //
        // Make sure forward decl matches
        //
        if(isAbstract != value -> is_abstract())
        {
            String_var err = string_dup("valuetype `");
            err += name;
            err += "' does not match forward declaration";
            IdlError(err);
            yynerrs++;
            return false;
        }

        IdlFwdValuetypes.remove(i);

        try
        {
            //
            // Fill valuetype
            //
            value -> id(id);
	    // Not necessary, move() sets new name and version:
	    //value -> name(name);
	    //value -> version(version);
            value -> supported_interface(info.supported_interface);
            value -> is_truncatable(info.is_truncatable);
            if(isAbstract)
                value -> abstract_base_values(info.base_values);
            else if(info.base_values.length() > 0)
            {
                value -> base_value(info.base_values[0]);
                ValueDefSeq seq = info.base_values;
                seq.remove(0);
                value -> abstract_base_values(seq);
            }

            //
            // Move interface (this will make this valuetype
            // the last entry in the container)
            //
            value -> move(container, name, version);
        }
        catch(const SystemException& ex)
        {
            IdlPrintException(ex, id, name);
            yynerrs++;
            return false;
        }
    }
    else
    {
        //
        // Create new value
        //
        try
        {
            InitializerSeq initializers;
            ValueDef_var baseValue;
            ValueDefSeq abstractBaseValues;

            //
            // Determine base value and abstract base values.
            //
            if(isAbstract)
                abstractBaseValues = info.base_values;
            else if(info.base_values.length() > 0)
            {
                if(info.base_values[0] -> is_abstract())
                    abstractBaseValues = info.base_values;
                else
                {
                    baseValue = ValueDef::_duplicate(info.base_values[0]);
                    abstractBaseValues = info.base_values;
                    abstractBaseValues.remove(0);
                }
            }

            value = container ->
                    create_value(id, name, version,
                                 isCustom, isAbstract,
                                 baseValue,
                                 info.is_truncatable,
                                 abstractBaseValues,
                                 info.supported_interface,
                                 initializers);
        }
        catch(const SystemException& ex)
        {
            IdlPrintException(ex, id, name);
            yynerrs++;
            return false;
        }
    }

    //
    // Push filled valuetype on the stack
    //
    IdlContStack.append(ValueDef::_duplicate(value));

    //
    // Push new prefix on the prefix stacks
    //
    for(ULong i = 0 ; i < IdlPrefStackStack.length() ; i++)
    {
        String_var prefix =
            IdlPrefStackStack[i][IdlPrefStackStack[i].length() - 1];
        prefix += name;
        prefix += "/";
        IdlPrefStackStack[i].append(prefix);
    }

    //
    // Add ID to IdlTopFileContent and IdlComments
    //
    IdlAddID(id);

    return true;
}

// ----------------------------------------------------------------------
// Create valuetype forward declaration
// ----------------------------------------------------------------------

bool
CreateValueFwd(const char* name, Boolean isAbstract)
{
    //
    // Get container from container stack
    //
    assert(IdlContStack.length() > 0);
    Container_var container =
        IdlContStack[IdlContStack.length() - 1];

    //
    // Get id and version
    //
    RepositoryId_var id = IdlGetRepositoryId(name);
    VersionSpec_var version = string_dup("1.0");

    //
    // Check if name already exists
    //
    Contained_var c;
    try
    {
        c = container -> lookup(name);
    }
    catch(const SystemException& ex)
    {
        IdlPrintException(ex, id, name);
        yynerrs++;
        return false;
    }

    if(!is_nil(c))
    {
        //
        // Name exists, check if it is a valuetype
        //
        ValueDef_var v = ValueDef::_narrow(c);

        if(is_nil(v))
        {
            String_var err = string_dup("conflicting types for `valuetype ");
            err += name;
            err += "'";
            IdlError(err);
            yynerrs++;
            return false;
        }

        //
        // Make sure forward decl matches
        //
        if(isAbstract != v -> is_abstract())
        {
            String_var err = string_dup("valuetype `");
            err += name;
            err += "' does not match forward declaration";
            IdlError(err);
            yynerrs++;
            return false;
        }
    }
    else
    {
        //
        // Create "empty" valuetype as forward
        // declaration, do not put it on the container
        // stack, but put it in the list of forward
        // declared valuetypes
        //
        ValueDefSeq baseValues;
        InitializerSeq initializers;

        try
        {
            ValueDef_var value =
                container -> create_value(id, name, version, false, isAbstract,
                                          ValueDef::_nil(), false, baseValues,
                                          InterfaceDef::_nil(), initializers);
            IdlFwdValuetypes.append(value -> absolute_name());
        }
        catch(const SystemException& ex)
        {
            IdlPrintException(ex, id, name);
            yynerrs++;
            return false;
        }
    }

    return true;
}

// ----------------------------------------------------------------------
// Error output
// ----------------------------------------------------------------------

void
yyerror(const char* s)
{
    IdlError(s);
}
