// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef IFR2_IMPL_H
#define IFR2_IMPL_H

#include <IFR1_impl.h>
#include <OB/IFR2_skel.h>

namespace CORBA
{

//
// Declare Container_impl
//
class Container_impl;

//
// The Contained_impl class
//
class Contained_impl : public virtual POA_CORBA::Contained,
		       public virtual IRObject_impl
{
protected:

    Container_var definedIn_;
    RepositoryId_var id_;
    Identifier_var name_;
    VersionSpec_var version_;

public:

    Contained_impl(ORB_ptr, Container_ptr);
    virtual ~Contained_impl();

    virtual void releaseInternal();

    virtual void destroy()
        throw(SystemException);

    virtual char* id()
        throw(SystemException);
    virtual void id(const char*)
        throw(SystemException);

    virtual char* name()
        throw(SystemException);
    virtual void name(const char*)
        throw(SystemException);

    virtual char* version()
        throw(SystemException);
    virtual void version(const char*)
        throw(SystemException);

    virtual Container_ptr defined_in()
        throw(SystemException);
    virtual char* absolute_name()
        throw(SystemException);
    virtual Repository_ptr containing_repository()
        throw(SystemException);

    virtual CORBA::Contained::Description* describe()
        throw(SystemException)= 0;

    virtual void move(Container_ptr new_container,
		      const char* identifier,
		      const char* new_version)
        throw(SystemException);
};

//
// The Container_impl class
//
class Container_impl : public virtual POA_CORBA::Container,
		       public virtual IRObject_impl
{
protected:

    ContainedSeq contents_;

    friend class Contained_impl;
    void add(Contained_ptr);
    void remove(Contained_ptr);
    static void merge(ContainedSeq*, const ContainedSeq&);

public:

    Container_impl(ORB_ptr);
    virtual ~Container_impl();

    virtual void releaseInternal();

    virtual Contained_ptr lookup(const char* search_name)
        throw(SystemException);
    
    virtual ContainedSeq* contents(DefinitionKind limit_type,
				   Boolean exclude_inherited)
        throw(SystemException);
    
    virtual ContainedSeq* lookup_name(const char* search_name,
				      Long levels_to_search,
				      DefinitionKind limit_type,
				      Boolean exclude_inherited)
        throw(SystemException);
    
    virtual CORBA::Container::DescriptionSeq*
    describe_contents(DefinitionKind limit_type,
		      Boolean exclude_inherited,
		      Long max_returned_objs)
        throw(SystemException);
    
    virtual ModuleDef_ptr
    create_module(const char* id,
		  const char* name,
		  const char* version)
        throw(SystemException);
    
    virtual ConstantDef_ptr
    create_constant(const char* id,
		    const char* name,
		    const char* version,
		    IDLType_ptr type,
		    const Any& value)
        throw(SystemException);
    
    virtual StructDef_ptr
    create_struct(const char* id,
		  const char* name,
		  const char* version,
		  const StructMemberSeq& members)
        throw(SystemException);
    
    virtual UnionDef_ptr
    create_union(const char* id,
		 const char* name,
		 const char* version,
		 IDLType_ptr discriminator_type,
		 const UnionMemberSeq& members)
        throw(SystemException);
    
    virtual EnumDef_ptr
    create_enum(const char* id,
		const char* name,
		const char* version,
		const EnumMemberSeq& members)
        throw(SystemException);
    
    virtual AliasDef_ptr
    create_alias(const char* id,
		 const char* name,
		 const char* version,
		 IDLType_ptr original_type)
        throw(SystemException);
    
    virtual InterfaceDef_ptr
    create_interface(const char* id,
		     const char* name,
		     const char* version,
		     const InterfaceDefSeq& base_interfaces,
		     Boolean is_abstract)
        throw(SystemException);
    
    virtual ExceptionDef_ptr
    create_exception(const char* id,
		     const char* name,
		     const char* version,
		     const StructMemberSeq& members)
        throw(SystemException);
    
    virtual ValueDef_ptr
    create_value(const char* id,
		 const char* name,
		 const char* version,
		 Boolean is_custom,
		 Boolean is_abstract,
		 ValueDef_ptr base_value,
		 Boolean is_truncatable,
		 const ValueDefSeq& abstract_base_values,
		 InterfaceDef_ptr supported_interface,
		 const InitializerSeq& initializers)
        throw(SystemException);

    virtual ValueBoxDef_ptr
    create_value_box(const char* id,
		     const char* name,
		     const char* version,
		     IDLType_ptr original_type_def)
        throw(SystemException);

    virtual NativeDef_ptr
    create_native(const char* id,
		  const char* name,
		  const char* version)
        throw(SystemException);
};

//
// The ExceptionDef_impl class
//
class ExceptionDef_impl : public virtual POA_CORBA::ExceptionDef,
			  public virtual Contained_impl,
			  public virtual Container_impl
{
    StructMemberSeq members_;

public:

    ExceptionDef_impl(ORB_ptr, Container_ptr);
    virtual ~ExceptionDef_impl();

    virtual void releaseInternal();

    virtual DefinitionKind def_kind()
        throw(SystemException)
    {
        return dk_Exception;
    }

    virtual CORBA::Contained::Description* describe()
        throw(SystemException);

    virtual TypeCode_ptr type()
        throw(SystemException);
    virtual TypeCode_ptr typeRecursive(ContainedSeq&)
        throw(SystemException);

    virtual StructMemberSeq* members()
        throw(SystemException);
    virtual void members(const StructMemberSeq&)
        throw(SystemException);
};

//
// The AttributeDef_impl class
//
class AttributeDef_impl : public virtual POA_CORBA::AttributeDef,
			  public virtual Contained_impl
{
    IDLType_var idlType_;
    AttributeMode attributeMode_;

public:

    AttributeDef_impl(ORB_ptr, Container_ptr);
    virtual ~AttributeDef_impl();

    virtual void releaseInternal();

    virtual DefinitionKind def_kind()
        throw(SystemException)
    {
        return dk_Attribute;
    }

    virtual char* name()
        throw(SystemException)
    {
        return Contained_impl::name();
    }
    
    virtual void name(const char*)
        throw(SystemException);

    virtual CORBA::Contained::Description* describe()
        throw(SystemException);

    virtual TypeCode_ptr type()
        throw(SystemException);

    virtual IDLType_ptr type_def()
        throw(SystemException);
    virtual void type_def(IDLType_ptr)
        throw(SystemException);

    virtual AttributeMode mode()
        throw(SystemException);
    virtual void mode(AttributeMode)
        throw(SystemException);
};

//
// The OperationDef_impl class
//
class OperationDef_impl : public virtual POA_CORBA::OperationDef,
			  public virtual Contained_impl
{
    IDLType_var result_;
    OperationMode mode_;
    ParDescriptionSeq params_;
    ExceptionDefSeq exceptions_;
    ContextIdSeq contexts_;

public:

    OperationDef_impl(ORB_ptr, Container_ptr);
    virtual ~OperationDef_impl();

    virtual void releaseInternal();

    virtual DefinitionKind def_kind()
        throw(SystemException)
    {
        return dk_Operation;
    }

    virtual char* name()
        throw(SystemException)
    {
        return Contained_impl::name();
    }
    virtual void name(const char*)
        throw(SystemException);

    virtual CORBA::Contained::Description* describe()
        throw(SystemException);

    virtual TypeCode_ptr result()
        throw(SystemException);

    virtual IDLType_ptr result_def()
        throw(SystemException);
    virtual void result_def(IDLType_ptr)
        throw(SystemException);

    virtual ParDescriptionSeq* params()
            throw(SystemException);
    virtual void params(const ParDescriptionSeq&)
            throw(SystemException);

    virtual OperationMode mode()
        throw(SystemException);
    virtual void mode(OperationMode)
        throw(SystemException);

    virtual ContextIdSeq* contexts()
        throw(SystemException);
    virtual void contexts(const ContextIdSeq&)
        throw(SystemException);

    virtual ExceptionDefSeq* exceptions()
        throw(SystemException);
    virtual void exceptions(const ExceptionDefSeq&)
        throw(SystemException);
};

//
// The InterfaceDef_impl class
//
class InterfaceDef_impl : public virtual POA_CORBA::InterfaceDef,
			  public virtual Container_impl,
			  public virtual Contained_impl,
			  public virtual IDLType_impl
{
    InterfaceDefSeq baseInterfaces_;
    Boolean isAbstract_;

public:

    InterfaceDef_impl(ORB_ptr, Container_ptr);
    virtual ~InterfaceDef_impl();

    virtual void releaseInternal();

    virtual DefinitionKind def_kind()
        throw(SystemException)
    {
        return dk_Interface;
    }

    virtual TypeCode_ptr type()
        throw(SystemException);
    virtual TypeCode_ptr typeRecursive(ContainedSeq&)
        throw(SystemException)
    {
	return type();
    }

    virtual CORBA::Contained::Description* describe()
        throw(SystemException);

    virtual Contained_ptr lookup(const char* search_name)
        throw(SystemException);

    virtual ContainedSeq* contents(DefinitionKind limit_type,
				   Boolean exclude_inherited)
        throw(SystemException);

    virtual ContainedSeq* lookup_name(const char* search_name,
				      Long levels_to_search,
				      DefinitionKind limit_type,
				      Boolean exclude_inherited)
        throw(SystemException);

    virtual InterfaceDefSeq* base_interfaces()
        throw(SystemException);
    virtual void base_interfaces(const InterfaceDefSeq&)
        throw(SystemException);

    virtual Boolean is_abstract()
        throw(SystemException);
    virtual void is_abstract(Boolean)
        throw(SystemException);

    virtual Boolean is_a(const char* interface_id)
        throw(SystemException);

    virtual CORBA::InterfaceDef::FullInterfaceDescription* describe_interface()
        throw(SystemException);

    virtual AttributeDef_ptr
    create_attribute(const char* id,
		     const char* name,
		     const char* version,
		     IDLType_ptr type,
		     AttributeMode mode)
        throw(SystemException);

    virtual OperationDef_ptr
    create_operation(const char* id,
		     const char* name,
		     const char* version,
		     IDLType_ptr result,
		     OperationMode mode,
		     const ParDescriptionSeq& params,
		     const ExceptionDefSeq& exceptions,
		     const ContextIdSeq& contexts)
        throw(SystemException);
};

} // End of namespace CORBA

#endif
