// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <Util.h>
#include <Types.h>
#include <IFR3_impl.h>

using namespace CORBA;
using namespace OB;
using namespace PortableServer;

// ----------------------------------------------------------------------
// Repository_impl member implementations
// ----------------------------------------------------------------------

Contained_ptr
Repository_impl::lookup_id_in_seq(const ContainedSeq& seq,
				  const char* search_id)
{
    for(ULong i = 0 ; i < seq.length() ; i++)
    {
	RepositoryId_var id = seq[i] -> id();
	if(strcmp(id, search_id) == 0)
	    return CORBA::Contained::_duplicate(seq[i]);

	Container_var container = CORBA::Container::_narrow(seq[i]);
	if(!is_nil(container))
	{
	    ContainedSeq_var seq = container -> contents(dk_all, false);
	    Contained_var contained = lookup_id_in_seq(seq, search_id);
	    if(!is_nil(contained))
		return CORBA::Contained::_duplicate(contained);
	}
    }
    
    return CORBA::Contained::_nil();
}

Repository_impl::Repository_impl(ORB_ptr orb)
    : IRObject_impl(orb),
      Container_impl(orb)
{
}

Repository_impl::~Repository_impl()
{
}

void
Repository_impl::releaseInternal()
{
    ULong i;

    for(i = 0 ; i < strings_.length() ; i++)
    {
	POA_var poa = _default_POA();
	ServantBase_var servant = poa -> reference_to_servant(strings_[i]);
	StringDef_impl* p = dynamic_cast<StringDef_impl*>(servant.in());
	if(p)
	    p -> releaseInternal();
    }
    
    for(i = 0 ; i < wstrings_.length() ; i++)
    {
	POA_var poa = _default_POA();
	ServantBase_var servant = poa -> reference_to_servant(wstrings_[i]);
	WstringDef_impl* p = dynamic_cast<WstringDef_impl*>(servant.in());
	if(p)
	    p -> releaseInternal();
    }
    
    for(i = 0 ; i < fixeds_.length() ; i++)
    {
	POA_var poa = _default_POA();
	ServantBase_var servant = poa -> reference_to_servant(fixeds_[i]);
	WstringDef_impl* p = dynamic_cast<WstringDef_impl*>(servant.in());
	if(p)
	    p -> releaseInternal();
    }
    
    for(i = 0 ; i < sequences_.length() ; i++)
    {
	POA_var poa = _default_POA();
	ServantBase_var servant = poa -> reference_to_servant(sequences_[i]);
	SequenceDef_impl* p = dynamic_cast<SequenceDef_impl*>(servant.in());
	if(p)
	    p -> releaseInternal();
    }
    
    for(i = 0 ; i < arrays_.length() ; i++)
    {
	POA_var poa = _default_POA();
	ServantBase_var servant = poa -> reference_to_servant(arrays_[i]);
	ArrayDef_impl* p = dynamic_cast<ArrayDef_impl*>(servant.in());
	if(p)
	    p -> releaseInternal();
    }
    
    primDefNull_ = CORBA::PrimitiveDef::_nil();
    primDefVoid_ = CORBA::PrimitiveDef::_nil();
    primDefShort_ = CORBA::PrimitiveDef::_nil();
    primDefLong_ = CORBA::PrimitiveDef::_nil();
    primDefUShort_ = CORBA::PrimitiveDef::_nil();
    primDefULong_ = CORBA::PrimitiveDef::_nil();
    primDefFloat_ = CORBA::PrimitiveDef::_nil();
    primDefDouble_ = CORBA::PrimitiveDef::_nil();
    primDefBoolean_ = CORBA::PrimitiveDef::_nil();
    primDefChar_ = CORBA::PrimitiveDef::_nil();
    primDefOctet_ = CORBA::PrimitiveDef::_nil();
    primDefAny_ = CORBA::PrimitiveDef::_nil();
    primDefTypeCode_ = CORBA::PrimitiveDef::_nil();
    primDefPrincipal_ = CORBA::PrimitiveDef::_nil();
    primDefString_ = CORBA::PrimitiveDef::_nil();
    primDefObjref_ = CORBA::PrimitiveDef::_nil();
    primDefLongLong_ = CORBA::PrimitiveDef::_nil();
    primDefULongLong_ = CORBA::PrimitiveDef::_nil();
    primDefLongDouble_ = CORBA::PrimitiveDef::_nil();
    primDefWChar_ = CORBA::PrimitiveDef::_nil();
    primDefWString_ = CORBA::PrimitiveDef::_nil();
    primDefValueBase_ = CORBA::PrimitiveDef::_nil();

    strings_.length(0);
    wstrings_.length(0);
    fixeds_.length(0);
    sequences_.length(0);
    arrays_.length(0);

    Container_impl::releaseInternal();
}

void
Repository_impl::destroy()
    throw(SystemException)
{
    //
    // Invoking destroy on a Repository is an error
    //
    throw BAD_INV_ORDER(2, COMPLETED_NO);
}

Contained_ptr
Repository_impl::lookup(const char* search_name)
    throw(SystemException)
{
    //
    // Make name relative (everything is relative to the repository)
    //
    while(strncmp(search_name, "::", 2) == 0)
	search_name += 2;

    return Container_impl::lookup(search_name);
}

Contained_ptr
Repository_impl::lookup_id(const char* search_id)
    throw(SystemException)
{
    ContainedSeq_var seq = contents(dk_all, false);
    return lookup_id_in_seq(seq, search_id);
}

TypeCode_ptr
Repository_impl::get_canonical_typecode(TypeCode_ptr tc)
    throw(SystemException)
{
    // ORBacus TypeCodes are always canonical
    return TypeCode::_duplicate(tc);
}

PrimitiveDef_ptr
Repository_impl::get_primitive(PrimitiveKind kind)
    throw(SystemException)
{
    switch(kind)
    {
    case pk_null:
	if(is_nil(primDefNull_))
	{
	    primDefNull_ = (new PrimitiveDef_impl(orb_, pk_null)) -> _this();
	}
	return CORBA::PrimitiveDef::_duplicate(primDefNull_);
	
    case pk_void:
	if(is_nil(primDefVoid_))
	{
	    primDefVoid_ = (new PrimitiveDef_impl(orb_, pk_void)) -> _this();
	}
	return CORBA::PrimitiveDef::_duplicate(primDefVoid_);
	
    case pk_short:
	if(is_nil(primDefShort_))
	{
	    primDefShort_ = (new PrimitiveDef_impl(orb_, pk_short)) -> _this();
	}
	return CORBA::PrimitiveDef::_duplicate(primDefShort_);
	
    case pk_long:
	if(is_nil(primDefLong_))
	{
	    primDefLong_ = (new PrimitiveDef_impl(orb_, pk_long)) -> _this();
	}
	return CORBA::PrimitiveDef::_duplicate(primDefLong_);
	
    case pk_ushort:
	if(is_nil(primDefUShort_))
	{
	    primDefUShort_ =
		(new PrimitiveDef_impl(orb_, pk_ushort)) -> _this();
	}
	return CORBA::PrimitiveDef::_duplicate(primDefUShort_);
	
    case pk_ulong:
	if(is_nil(primDefULong_))
	{
	    primDefULong_ = (new PrimitiveDef_impl(orb_, pk_ulong)) -> _this();
	}
	return CORBA::PrimitiveDef::_duplicate(primDefULong_);
	
    case pk_float:
	if(is_nil(primDefFloat_))
	{
	    primDefFloat_ = (new PrimitiveDef_impl(orb_, pk_float)) -> _this();
	}
	return CORBA::PrimitiveDef::_duplicate(primDefFloat_);
	
    case pk_double:
	if(is_nil(primDefDouble_))
	{
	    primDefDouble_ =
		(new PrimitiveDef_impl(orb_, pk_double)) -> _this();
	}
	return CORBA::PrimitiveDef::_duplicate(primDefDouble_);
	
    case pk_boolean:
	if(is_nil(primDefBoolean_))
	{
	    primDefBoolean_ =
		(new PrimitiveDef_impl(orb_, pk_boolean)) -> _this();
	}
	return CORBA::PrimitiveDef::_duplicate(primDefBoolean_);
	
    case pk_char:
	if(is_nil(primDefChar_))
	{
	    primDefChar_ = (new PrimitiveDef_impl(orb_, pk_char)) -> _this();
	}
	return CORBA::PrimitiveDef::_duplicate(primDefChar_);
	
    case pk_octet:
	if(is_nil(primDefOctet_))
	{
	    primDefOctet_ = (new PrimitiveDef_impl(orb_, pk_octet)) -> _this();
	}
	return CORBA::PrimitiveDef::_duplicate(primDefOctet_);
	
    case pk_any:
	if(is_nil(primDefAny_))
	{
	    primDefAny_ = (new PrimitiveDef_impl(orb_, pk_any)) -> _this();
	}
	return CORBA::PrimitiveDef::_duplicate(primDefAny_);
	
    case pk_TypeCode:
	if(is_nil(primDefTypeCode_))
	{
	    primDefTypeCode_ =
		(new PrimitiveDef_impl(orb_, pk_TypeCode)) -> _this();
	}
	return CORBA::PrimitiveDef::_duplicate(primDefTypeCode_);
	
    case pk_Principal:
	if(is_nil(primDefPrincipal_))
	{
	    primDefPrincipal_ =
		(new PrimitiveDef_impl(orb_, pk_Principal)) -> _this();
	}
	return CORBA::PrimitiveDef::_duplicate(primDefPrincipal_);
	
    case pk_string:
	if(is_nil(primDefString_))
	{
	    primDefString_ =
		(new PrimitiveDef_impl(orb_, pk_string)) -> _this();
	}
	return CORBA::PrimitiveDef::_duplicate(primDefString_);
	
    case pk_objref:
	if(is_nil(primDefObjref_))
	{
	    primDefObjref_ =
		(new PrimitiveDef_impl(orb_, pk_objref)) -> _this();
	}
	return CORBA::PrimitiveDef::_duplicate(primDefObjref_);

    case pk_longlong:
	if(is_nil(primDefLongLong_))
	{
	    primDefLongLong_ =
		(new PrimitiveDef_impl(orb_, pk_longlong)) -> _this();
	}
	return CORBA::PrimitiveDef::_duplicate(primDefLongLong_);
	
    case pk_ulonglong:
	if(is_nil(primDefULongLong_))
	{
	    primDefULongLong_ =
		(new PrimitiveDef_impl(orb_, pk_ulonglong)) -> _this();
	}
	return CORBA::PrimitiveDef::_duplicate(primDefULongLong_);
	
    case pk_longdouble:
	if(is_nil(primDefLongDouble_))
	{
	    primDefLongDouble_ =
		(new PrimitiveDef_impl(orb_, pk_longdouble)) -> _this();
	}
	return CORBA::PrimitiveDef::_duplicate(primDefLongDouble_);

    case pk_wchar:
	if(is_nil(primDefWChar_))
	{
	    primDefWChar_ = (new PrimitiveDef_impl(orb_, pk_wchar)) -> _this();
	}
	return CORBA::PrimitiveDef::_duplicate(primDefWChar_);
	
    case pk_wstring:
	if(is_nil(primDefWString_))
	{
	    primDefWString_ =
		(new PrimitiveDef_impl(orb_, pk_wstring)) -> _this();
	}
	return CORBA::PrimitiveDef::_duplicate(primDefWString_);

    case pk_ValueBase:
	if(is_nil(primDefValueBase_))
	{
	    primDefValueBase_ =
		(new PrimitiveDef_impl(orb_, pk_ValueBase)) -> _this();
	}
	return CORBA::PrimitiveDef::_duplicate(primDefValueBase_);

    }

    return CORBA::PrimitiveDef::_nil();
}

StringDef_ptr
Repository_impl::create_string(ULong bound)
    throw(SystemException)
{
    TypeCode_var t1 = orb_ -> create_string_tc(bound);
    
    for(ULong i = 0 ; i < strings_.length() ; i++)
    {
	StringDef_ptr p2 = strings_[i];
	TypeCode_var t2 = p2 -> type();

	if(t1 -> equal(t2))
	    return CORBA::StringDef::_duplicate(strings_[i]);
    }

    Repository_var self = _this();
    StringDef_impl* p = new StringDef_impl(orb_, self);
    try
    {
        p -> bound(bound);
    }
    catch(...)
    {
        p -> _remove_ref();
        throw;
    }
    strings_.append(p -> _this());
    return p -> _this();
}

WstringDef_ptr
Repository_impl::create_wstring(ULong bound)
    throw(SystemException)
{
    TypeCode_var t1 = orb_ -> create_wstring_tc(bound);
    
    for(ULong i = 0 ; i < wstrings_.length() ; i++)
    {
	WstringDef_ptr p2 = wstrings_[i];
	TypeCode_var t2 = p2 -> type();

	if(t1 -> equal(t2))
	    return CORBA::WstringDef::_duplicate(wstrings_[i]);
    }

    Repository_var self = _this();
    WstringDef_impl* p = new WstringDef_impl(orb_, self);

    try
    {
        p -> bound(bound);
    }
    catch(...)
    {
        p -> _remove_ref();
        throw;
    }
    wstrings_.append(p -> _this());
    return p -> _this();
}

SequenceDef_ptr
Repository_impl::create_sequence(ULong bound, IDLType_ptr element_type)
    throw(SystemException)
{
    TypeCode_var elem = element_type -> type();
    TypeCode_var t1 = orb_ -> create_sequence_tc(bound, elem);
    
    for(ULong i = 0 ; i < sequences_.length() ; i++)
    {
	SequenceDef_ptr p2 = sequences_[i];
	TypeCode_var t2 = p2 -> type();

	if(t1 -> equal(t2))
	    return CORBA::SequenceDef::_duplicate(sequences_[i]);
    }

    Repository_var self = _this();
    SequenceDef_impl* p = new SequenceDef_impl(orb_, self);

    try
    {
        p -> bound(bound);
        p -> element_type_def(element_type);
    }
    catch(...)
    {
        p -> _remove_ref();
        throw;
    }
    sequences_.append(p -> _this());
    return p -> _this();
}

ArrayDef_ptr
Repository_impl::create_array(ULong length, IDLType_ptr element_type)
    throw(SystemException)
{
    TypeCode_var elem = element_type -> type();
    TypeCode_var t1 = orb_ -> create_array_tc(length, elem);
    
    for(ULong i = 0 ; i < arrays_.length() ; i++)
    {
	ArrayDef_ptr p2 = arrays_[i];
	TypeCode_var t2 = p2 -> type();

	if(t1 -> equal(t2))
	    return CORBA::ArrayDef::_duplicate(arrays_[i]);
    }

    Repository_var self = _this();
    ArrayDef_impl* p = new ArrayDef_impl(orb_, self);
    try
    {
        p -> length(length);
        p -> element_type_def(element_type);
    }
    catch(...)
    {
        p -> _remove_ref();
        throw;
    }
    arrays_.append(p -> _this());
    return p -> _this();
}

FixedDef_ptr
Repository_impl::create_fixed(UShort digits, Short scale)
    throw(SystemException)
{
    TypeCode_var t1 = orb_ -> create_fixed_tc(digits, scale);
    
    for(ULong i = 0 ; i < fixeds_.length() ; i++)
    {
	FixedDef_ptr p2 = fixeds_[i];
	TypeCode_var t2 = p2 -> type();

	if(t1 -> equal(t2))
	    return CORBA::FixedDef::_duplicate(fixeds_[i]);
    }

    Repository_var self = _this();
    FixedDef_impl* p = new FixedDef_impl(orb_, self);
    try
    {
        p -> digits(digits);
        p -> scale(scale);
    }
    catch(...)
    {
        p -> _remove_ref();
        throw;
    }
    fixeds_.append(p -> _this());
    return p -> _this();
}

// ----------------------------------------------------------------------
// ModuleDef_impl member implementations
// ----------------------------------------------------------------------

ModuleDef_impl::ModuleDef_impl(ORB_ptr orb, Container_ptr p)
    : IRObject_impl(orb),
      Container_impl(orb),
      Contained_impl(orb, p)
{
}

ModuleDef_impl::~ModuleDef_impl()
{
}

void
ModuleDef_impl::releaseInternal()
{
    Container_impl::releaseInternal();
    Contained_impl::releaseInternal();
}

Contained::Description*
ModuleDef_impl::describe()
    throw(SystemException)
{
    ModuleDescription* desc = new ModuleDescription;

    desc -> name = name();
    desc -> id = id();

    Contained_var contained = CORBA::Contained::_narrow(definedIn_);
    if(!is_nil(contained))
	desc -> defined_in = contained -> id();
    else
	desc -> defined_in = string_dup("");

    desc -> version = version();

    CORBA::Contained::Description* description =
	new CORBA::Contained::Description;
    description -> kind = def_kind();
    description -> value <<= desc;

    return description;
}

// ----------------------------------------------------------------------
// ConstantDef_impl member implementations
// ----------------------------------------------------------------------

ConstantDef_impl::ConstantDef_impl(ORB_ptr orb, Container_ptr p)
    : IRObject_impl(orb),
      Contained_impl(orb, p)
{
}

ConstantDef_impl::~ConstantDef_impl()
{
}

void
ConstantDef_impl::releaseInternal()
{
    idlType_ = CORBA::IDLType::_nil();
    Contained_impl::releaseInternal();
}

Contained::Description*
ConstantDef_impl::describe()
    throw(SystemException)
{
    ConstantDescription* desc = new ConstantDescription;

    desc -> name = name();
    desc -> id = id();

    Contained_var contained = CORBA::Contained::_narrow(definedIn_);
    if(!is_nil(contained))
	desc -> defined_in = contained -> id();
    else
	desc -> defined_in = string_dup("");

    desc -> version = version();
    desc -> type = type();
    desc -> value = value_;

    CORBA::Contained::Description* description =
        new CORBA::Contained::Description;
    description -> kind = def_kind();
    description -> value <<= desc;

    return description;
}

TypeCode_ptr
ConstantDef_impl::type()
    throw(SystemException)
{
    return idlType_ -> type();
}

IDLType_ptr
ConstantDef_impl::type_def()
    throw(SystemException)
{
    return CORBA::IDLType::_duplicate(idlType_);
}

void
ConstantDef_impl::type_def(IDLType_ptr t)
    throw(SystemException)
{
    idlType_ = CORBA::IDLType::_duplicate(t);
}

Any*
ConstantDef_impl::value()
    throw(SystemException)
{
    return new Any(value_);
}

void
ConstantDef_impl::value(const Any& v)
    throw(SystemException)
{
    value_ = v;
}

// ----------------------------------------------------------------------
// TypedefDef_impl member implementations
// ----------------------------------------------------------------------

TypedefDef_impl::TypedefDef_impl(ORB_ptr orb, Container_ptr p)
    : IRObject_impl(orb),
      Contained_impl(orb, p),
      IDLType_impl(orb)
{
}

TypedefDef_impl::~TypedefDef_impl()
{
}

Contained::Description*
TypedefDef_impl::describe()
    throw(SystemException)
{
    TypeDescription* desc = new TypeDescription;

    desc -> name = name();
    desc -> id = id();

    Contained_var contained = CORBA::Contained::_narrow(definedIn_);
    if(!is_nil(contained))
	desc -> defined_in = contained -> id();
    else
	desc -> defined_in = string_dup("");

    desc -> version = version();
    desc -> type = type();

    CORBA::Contained::Description* description =
	new CORBA::Contained::Description;
    description -> kind = def_kind();
    description -> value <<= desc;

    return description;
}

// ----------------------------------------------------------------------
// StructDef_impl member implementations
// ----------------------------------------------------------------------

StructDef_impl::StructDef_impl(ORB_ptr orb, Container_ptr p)
    : IRObject_impl(orb),
      Container_impl(orb),
      Contained_impl(orb, p),
      IDLType_impl(orb),
      TypedefDef_impl(orb, p)
{
}

StructDef_impl::~StructDef_impl()
{
}

void
StructDef_impl::releaseInternal()
{
    members_.length(0);
    TypedefDef_impl::releaseInternal();
}

TypeCode_ptr
StructDef_impl::type()
    throw(SystemException)
{
    ContainedSeq seq;
    TypeCode_var tc = typeRecursive(seq);
    assert(seq.length() == 0);
    return tc._retn();
}

TypeCode_ptr
StructDef_impl::typeRecursive(ContainedSeq& seq)
    throw(SystemException)
{
    seq.length(seq.length() + 1);
    seq[seq.length() - 1] = _this();
    
    POA_var poa = _default_POA();
    for(ULong i = 0 ; i < members_.length() ; i++)
    {
	ServantBase_var servant =
	    poa -> reference_to_servant(members_[i].type_def);
	IDLType_impl* idlTypeImpl = dynamic_cast<IDLType_impl*>(servant.in());
	assert(idlTypeImpl);
	members_[i].type = idlTypeImpl -> typeRecursive(seq);
    }

    seq.length(seq.length() - 1);
    return orb_ -> create_struct_tc(id_, name_, members_);
}

StructMemberSeq*
StructDef_impl::members()
    throw(SystemException)
{
    //
    // Recalculate member TypeCodes, as ids may have changed, and also
    // in order to calculate recursive TypeCodes correctly
    //
    POA_var poa = _default_POA();
    for(ULong i = 0 ; i < members_.length() ; i++)
    {
	ServantBase_var servant =
	    poa -> reference_to_servant(members_[i].type_def);
	IDLType_impl* idlTypeImpl = dynamic_cast<IDLType_impl*>(servant.in());
	assert(idlTypeImpl);
	ContainedSeq seq;
	members_[i].type = idlTypeImpl -> typeRecursive(seq);
	assert(seq.length() == 0);
    }

    return new StructMemberSeq(members_);
}

void
StructDef_impl::members(const StructMemberSeq& members)
    throw(SystemException)
{
    members_ = members;
}

// ----------------------------------------------------------------------
// UnionDef_impl member implementations
// ----------------------------------------------------------------------

UnionDef_impl::UnionDef_impl(ORB_ptr orb, Container_ptr p)
    : IRObject_impl(orb),
      Container_impl(orb),
      Contained_impl(orb, p),
      IDLType_impl(orb),
      TypedefDef_impl(orb, p)
{
}

UnionDef_impl::~UnionDef_impl()
{
}

void
UnionDef_impl::releaseInternal()
{
    discType_ = CORBA::IDLType::_nil();
    members_.length(0);
    TypedefDef_impl::releaseInternal();
}

TypeCode_ptr
UnionDef_impl::type()
    throw(SystemException)
{
    ContainedSeq seq;
    TypeCode_var tc = typeRecursive(seq);
    assert(seq.length() == 0);
    return tc._retn();
}

TypeCode_ptr
UnionDef_impl::typeRecursive(ContainedSeq& seq)
    throw(SystemException)
{
    seq.length(seq.length() + 1);
    seq[seq.length() - 1] = _this();
    
    POA_var poa = _default_POA();
    for(ULong i = 0 ; i < members_.length() ; i++)
    {
	ServantBase_var servant =
	    poa -> reference_to_servant(members_[i].type_def);
	IDLType_impl* idlTypeImpl = dynamic_cast<IDLType_impl*>(servant.in());
	assert(idlTypeImpl);
	members_[i].type = idlTypeImpl -> typeRecursive(seq);
    }

    seq.length(seq.length() - 1);
    TypeCode_var discTC = discriminator_type();
    return orb_ -> create_union_tc(id_, name_, discTC, members_);
}

TypeCode_ptr
UnionDef_impl::discriminator_type()
    throw(SystemException)
{
    return discType_ -> type();
}

IDLType_ptr
UnionDef_impl::discriminator_type_def()
    throw(SystemException)
{
    return CORBA::IDLType::_duplicate(discType_);
}

void
UnionDef_impl::discriminator_type_def(IDLType_ptr d)
    throw(SystemException)
{
    discType_ = CORBA::IDLType::_duplicate(d);
}

UnionMemberSeq*
UnionDef_impl::members()
    throw(SystemException)
{
    //
    // Recalculate member TypeCodes, as ids may have changed, and also
    // in order to calculate recursive TypeCodes correctly
    //
    POA_var poa = _default_POA();
    for(ULong i = 0 ; i < members_.length() ; i++)
    {
	ServantBase_var servant =
	    poa -> reference_to_servant(members_[i].type_def);
	IDLType_impl* idlTypeImpl = dynamic_cast<IDLType_impl*>(servant.in());
	assert(idlTypeImpl);
	ContainedSeq seq;
	members_[i].type = idlTypeImpl -> typeRecursive(seq);
	assert(seq.length() == 0);
    }

    return new UnionMemberSeq(members_);
}

void
UnionDef_impl::members(const UnionMemberSeq& members)
    throw(SystemException)
{
    members_ = members;
}

// ----------------------------------------------------------------------
// EnumDef_impl member implementations
// ----------------------------------------------------------------------

EnumDef_impl::EnumDef_impl(ORB_ptr orb, Container_ptr p)
    : IRObject_impl(orb),
      Contained_impl(orb, p),
      IDLType_impl(orb),
      TypedefDef_impl(orb, p)
{
}

EnumDef_impl::~EnumDef_impl()
{
}

TypeCode_ptr
EnumDef_impl::type()
    throw(SystemException)
{
    return orb_ -> create_enum_tc(id_, name_, members_);
}

EnumMemberSeq*
EnumDef_impl::members()
    throw(SystemException)
{
    return new EnumMemberSeq(members_);
}

void
EnumDef_impl::members(const EnumMemberSeq& m)
    throw(SystemException)
{
    members_ = m;
}

// ----------------------------------------------------------------------
// AliasDef_impl member implementations
// ----------------------------------------------------------------------

AliasDef_impl::AliasDef_impl(ORB_ptr orb, Container_ptr p)
    : IRObject_impl(orb),
      Contained_impl(orb, p),
      IDLType_impl(orb),
      TypedefDef_impl(orb, p)
{
}

AliasDef_impl::~AliasDef_impl()
{
}

void
AliasDef_impl::releaseInternal()
{
    idlType_ = CORBA::IDLType::_nil();
    TypedefDef_impl::releaseInternal();
}

TypeCode_ptr
AliasDef_impl::type()
    throw(SystemException)
{
    ContainedSeq seq;
    TypeCode_var tc = typeRecursive(seq);
    assert(seq.length() == 0);
    return tc._retn();
}

TypeCode_ptr
AliasDef_impl::typeRecursive(ContainedSeq& seq)
    throw(SystemException)
{
    seq.length(seq.length() + 1);
    seq[seq.length() - 1] = _this();
    
    CORBA::Contained_var contained = CORBA::Contained::_narrow(idlType_);
    if(!is_nil(contained))
    {
	RepositoryId_var id = contained -> id();
	for(ULong i = 0 ; i < seq.length() ; i++)
	{
	    RepositoryId_var id2 = seq[i] -> id();
	    if(strcmp(id, id2) == 0)
	    {
		throw INTF_REPOS(MinorIllegalRecursion, COMPLETED_NO);
	    }
	}
    }

    POA_var poa = _default_POA();
    ServantBase_var servant = poa -> reference_to_servant(idlType_);
    IDLType_impl* idlTypeImpl = dynamic_cast<IDLType_impl*>(servant.in());
    assert(idlTypeImpl);
    TypeCode_var tc = idlTypeImpl -> typeRecursive(seq);
    
    seq.length(seq.length() - 1);
    return orb_ -> create_alias_tc(id_, name_, tc);
}

IDLType_ptr
AliasDef_impl::original_type_def()
    throw(SystemException)
{
    return CORBA::IDLType::_duplicate(idlType_);
}

void
AliasDef_impl::original_type_def(IDLType_ptr t)
    throw(SystemException)
{
    idlType_ = CORBA::IDLType::_duplicate(t);
}

// ----------------------------------------------------------------------
// NativeDef_impl member implementations
// ----------------------------------------------------------------------

NativeDef_impl::NativeDef_impl(ORB_ptr orb, Container_ptr p)
    : IRObject_impl(orb),
      Contained_impl(orb, p),
      IDLType_impl(orb),
      TypedefDef_impl(orb, p)
{
}

NativeDef_impl::~NativeDef_impl()
{
}

TypeCode_ptr
NativeDef_impl::type()
    throw(SystemException)
{
    return orb_ -> create_native_tc(id_, name_);
}

// ----------------------------------------------------------------------
// PrimitiveDef_impl member implementations
// ----------------------------------------------------------------------

PrimitiveDef_impl::PrimitiveDef_impl(ORB_ptr orb, PrimitiveKind pk)
    : IRObject_impl(orb),
      IDLType_impl(orb),
      primitiveKind(pk)
{
}

PrimitiveDef_impl::~PrimitiveDef_impl()
{
}

void
PrimitiveDef_impl::destroy()
    throw(SystemException)
{
    //
    // Invoking destroy on a PrimitiveDef is an error
    //
    throw BAD_INV_ORDER(2, COMPLETED_NO);
}

TypeCode_ptr
PrimitiveDef_impl::type()
    throw(SystemException)
{
    TypeCode_ptr p = TypeCode::_nil();

    switch(primitiveKind)
    {
    case pk_null:
	p = _tc_null;
	break;

    case pk_void:
	p = _tc_void;
	break;

    case pk_short:
	p = _tc_short;
	break;

    case pk_long:
	p = _tc_long;
	break;

    case pk_ushort:
	p = _tc_ushort;
	break;

    case pk_ulong:
	p = _tc_ulong;
	break;

    case pk_float:
	p = _tc_float;
	break;

    case pk_double:
	p = _tc_double;
	break;

    case pk_boolean:
	p = _tc_boolean;
	break;

    case pk_char:
	p = _tc_char;
	break;

    case pk_octet:
	p = _tc_octet;
	break;

    case pk_any:
	p = _tc_any;
	break;

    case pk_TypeCode:
	p = _tc_TypeCode;
	break;

    case pk_Principal:
	p = _tc_Principal;
	break;

    case pk_string:
	p = _tc_string;
	break;

    case pk_objref:
	p = _tc_Object;
	break;

    case pk_longlong:
	p = _tc_longlong;
	break;

    case pk_ulonglong:
	p = _tc_ulonglong;
	break;

    case pk_longdouble:
	p = _tc_longdouble;
	break;

    case pk_wchar:
	p = _tc_wchar;
	break;

    case pk_wstring:
	p = _tc_wstring;
	break;

    case pk_ValueBase:
    {
        ValueMemberSeq members;
        return orb_ -> create_value_tc("IDL:omg.org/CORBA/ValueBase:1.0",
                                       "ValueBase", VTM_ABSTRACT,
                                       TypeCode::_nil(), members);
    }
    }
    
    return TypeCode::_duplicate(p);
}

PrimitiveKind
PrimitiveDef_impl::kind()
    throw(SystemException)
{
    return primitiveKind;
}

// ----------------------------------------------------------------------
// StringDef_impl member implementations
// ----------------------------------------------------------------------

StringDef_impl::StringDef_impl(ORB_ptr orb, Repository_ptr p)
    : IRObject_impl(orb),
      IDLType_impl(orb),
      repository_(CORBA::Repository::_duplicate(p))
{
}

StringDef_impl::~StringDef_impl()
{
}

void
StringDef_impl::releaseInternal()
{
    repository_ = CORBA::Repository::_nil();
    IDLType_impl::releaseInternal();
}

TypeCode_ptr
StringDef_impl::type()
    throw(SystemException)
{
    return orb_ -> create_string_tc(bound_);
}

ULong
StringDef_impl::bound()
    throw(SystemException)
{
    return bound_;
}

void
StringDef_impl::bound(ULong b)
    throw(SystemException)
{
    bound_ = b;
}

// ----------------------------------------------------------------------
// WstringDef_impl member implementations
// ----------------------------------------------------------------------

WstringDef_impl::WstringDef_impl(ORB_ptr orb, Repository_ptr p)
    : IRObject_impl(orb),
      IDLType_impl(orb),
      repository_(CORBA::Repository::_duplicate(p))
{
}

WstringDef_impl::~WstringDef_impl()
{
}

void
WstringDef_impl::releaseInternal()
{
    repository_ = CORBA::Repository::_nil();
    IDLType_impl::releaseInternal();
}

TypeCode_ptr
WstringDef_impl::type()
    throw(SystemException)
{
    return orb_ -> create_wstring_tc(bound_);
}

ULong
WstringDef_impl::bound()
    throw(SystemException)
{
    return bound_;
}

void
WstringDef_impl::bound(ULong b)
    throw(SystemException)
{
    bound_ = b;
}

// ----------------------------------------------------------------------
// FixedDef_impl member implementations
// ----------------------------------------------------------------------

FixedDef_impl::FixedDef_impl(ORB_ptr orb, Repository_ptr p)
    : IRObject_impl(orb),
      IDLType_impl(orb),
      repository_(CORBA::Repository::_duplicate(p))
{
}

FixedDef_impl::~FixedDef_impl()
{
}

void
FixedDef_impl::releaseInternal()
{
    repository_ = CORBA::Repository::_nil();
    IDLType_impl::releaseInternal();
}

TypeCode_ptr
FixedDef_impl::type()
    throw(SystemException)
{
    return orb_ -> create_fixed_tc(digits_, scale_);
}

UShort
FixedDef_impl::digits()
    throw(SystemException)
{
    return digits_;
}

void
FixedDef_impl::digits(UShort digits)
    throw(SystemException)
{
    digits_ = digits;
}

Short
FixedDef_impl::scale()
    throw(SystemException)
{
    return scale_;
}

void
FixedDef_impl::scale(Short scale)
    throw(SystemException)
{
    scale_ = scale;
}

// ----------------------------------------------------------------------
// SequenceDef_impl member implementations
// ----------------------------------------------------------------------

SequenceDef_impl::SequenceDef_impl(ORB_ptr orb, Repository_ptr p)
    : IRObject_impl(orb),
      IDLType_impl(orb),
      repository_(CORBA::Repository::_duplicate(p))
{
}

SequenceDef_impl::~SequenceDef_impl()
{
}

void
SequenceDef_impl::releaseInternal()
{
    repository_ = CORBA::Repository::_nil();
    elementType_ = CORBA::IDLType::_nil();
    IDLType_impl::releaseInternal();
}

TypeCode_ptr
SequenceDef_impl::type()
    throw(SystemException)
{
    ContainedSeq seq;
    TypeCode_var tc = typeRecursive(seq);
    assert(seq.length() == 0);
    return tc._retn();
}

TypeCode_ptr
SequenceDef_impl::typeRecursive(ContainedSeq& seq)
    throw(SystemException)
{
    TypeCode_var tc;
    CORBA::Contained_var contained = CORBA::Contained::_narrow(elementType_);
    if(!is_nil(contained))
    {
	RepositoryId_var id = contained -> id();
	for(ULong i = 0 ; i < seq.length() ; i++)
	{
	    RepositoryId_var id2 = seq[i] -> id();
	    if(strcmp(id, id2) == 0)
	    {
		DefinitionKind elemDefKind = contained -> def_kind();
		if(elemDefKind != dk_Struct &&
		   elemDefKind != dk_Exception &&
		   elemDefKind != dk_Union)
		{
		    throw INTF_REPOS(MinorIllegalRecursion, COMPLETED_NO);
		}
		
		tc = orb_ -> create_recursive_tc(id);
	    }
	}
    }

    if(is_nil(tc))
    {
	POA_var poa = _default_POA();
	ServantBase_var servant = poa -> reference_to_servant(elementType_);
	IDLType_impl* elementTypeImpl =
	    dynamic_cast<IDLType_impl*>(servant.in());
	assert(elementTypeImpl);
	tc = elementTypeImpl -> typeRecursive(seq);
    }
    
    return orb_ -> create_sequence_tc(bound_, tc);
}

ULong
SequenceDef_impl::bound()
    throw(SystemException)
{
    return bound_;
}

void
SequenceDef_impl::bound(ULong b)
    throw(SystemException)
{
    bound_ = b;
}

TypeCode_ptr
SequenceDef_impl::element_type()
    throw(SystemException)
{
    return elementType_ -> type();
}

IDLType_ptr
SequenceDef_impl::element_type_def()
    throw(SystemException)
{
    return CORBA::IDLType::_duplicate(elementType_);
}

void
SequenceDef_impl::element_type_def(IDLType_ptr p)
    throw(SystemException)
{
    elementType_ = CORBA::IDLType::_duplicate(p);
}

// ----------------------------------------------------------------------
// ArrayDef_impl member implementations
// ----------------------------------------------------------------------

ArrayDef_impl::ArrayDef_impl(ORB_ptr orb, Repository_ptr p)
    : IRObject_impl(orb),
      IDLType_impl(orb),
      repository_(CORBA::Repository::_duplicate(p))
{
}

ArrayDef_impl::~ArrayDef_impl()
{
}

void
ArrayDef_impl::releaseInternal()
{
    repository_ = CORBA::Repository::_nil();
    elementType_ = CORBA::IDLType::_nil();
    IDLType_impl::releaseInternal();
}

TypeCode_ptr
ArrayDef_impl::type()
    throw(SystemException)
{
    ContainedSeq seq;
    TypeCode_var tc = typeRecursive(seq);
    assert(seq.length() == 0);
    return tc._retn();
}

TypeCode_ptr
ArrayDef_impl::typeRecursive(ContainedSeq& seq)
    throw(SystemException)
{
    Contained_var contained = Contained::_narrow(elementType_);
    if(!is_nil(contained))
    {
	RepositoryId_var id = contained -> id();
	for(ULong i = 0 ; i < seq.length() ; i++)
	{
	    RepositoryId_var id2 = seq[i] -> id();
	    if(strcmp(id, id2) == 0)
	    {
		throw INTF_REPOS(MinorIllegalRecursion, COMPLETED_NO);
	    }
	}
    }

    POA_var poa = _default_POA();
    ServantBase_var servant = poa -> reference_to_servant(elementType_);
    IDLType_impl* elementTypeImpl = dynamic_cast<IDLType_impl*>(servant.in());
    assert(elementTypeImpl);
    TypeCode_var tc = elementTypeImpl -> typeRecursive(seq);
    
    return orb_ -> create_array_tc(length_, tc);
}

ULong
ArrayDef_impl::length()
    throw(SystemException)
{
    return length_;
}

void
ArrayDef_impl::length(ULong b)
    throw(SystemException)
{
    length_ = b;
}

TypeCode_ptr
ArrayDef_impl::element_type()
    throw(SystemException)
{
    return elementType_ -> type();
}

IDLType_ptr
ArrayDef_impl::element_type_def()
    throw(SystemException)
{
    return CORBA::IDLType::_duplicate(elementType_);
}

void
ArrayDef_impl::element_type_def(IDLType_ptr p)
    throw(SystemException)
{
    elementType_ = CORBA::IDLType::_duplicate(p);
}
