// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef IFR3_IMPL_H
#define IFR3_IMPL_H

#include <IFR2_impl.h>
#include <OB/IFR3_skel.h>

namespace CORBA
{

//
// The Repository_impl class
//
class Repository_impl : public virtual POA_CORBA::Repository,
			public virtual Container_impl
{
    Contained_ptr lookup_id_in_seq(const ContainedSeq&, const char*);

protected:

    PrimitiveDef_var primDefNull_;
    PrimitiveDef_var primDefVoid_;
    PrimitiveDef_var primDefShort_;
    PrimitiveDef_var primDefLong_;
    PrimitiveDef_var primDefUShort_;
    PrimitiveDef_var primDefULong_;
    PrimitiveDef_var primDefFloat_;
    PrimitiveDef_var primDefDouble_;
    PrimitiveDef_var primDefBoolean_;
    PrimitiveDef_var primDefChar_;
    PrimitiveDef_var primDefOctet_;
    PrimitiveDef_var primDefAny_;
    PrimitiveDef_var primDefTypeCode_;
    PrimitiveDef_var primDefPrincipal_;
    PrimitiveDef_var primDefString_;
    PrimitiveDef_var primDefObjref_;
    PrimitiveDef_var primDefLongLong_;
    PrimitiveDef_var primDefULongLong_;
    PrimitiveDef_var primDefLongDouble_;
    PrimitiveDef_var primDefWChar_;
    PrimitiveDef_var primDefWString_;
    PrimitiveDef_var primDefValueBase_;

    IdlStringDefSeq strings_;
    IdlWstringDefSeq wstrings_;
    IdlFixedDefSeq fixeds_;
    IdlSequenceDefSeq sequences_;
    IdlArrayDefSeq arrays_;

public:

    Repository_impl(ORB_ptr);
    virtual ~Repository_impl();

    virtual void releaseInternal();

    virtual DefinitionKind def_kind()
        throw(SystemException)
    {
        return dk_Repository;
    }

    virtual void destroy()
        throw(SystemException);

    virtual Contained_ptr lookup(const char* search_name)
        throw(SystemException);

    virtual Contained_ptr lookup_id(const char* search_id)
        throw(SystemException);

    TypeCode_ptr get_canonical_typecode(TypeCode_ptr tc)
        throw(SystemException);

    PrimitiveDef_ptr get_primitive(PrimitiveKind kind)
        throw(SystemException);

    StringDef_ptr
    create_string(ULong bound)
        throw(SystemException);
    
    WstringDef_ptr
    create_wstring(ULong bound)
        throw(SystemException);

    SequenceDef_ptr
    create_sequence(ULong bound,
		    IDLType_ptr element_type)
        throw(SystemException);
    
    ArrayDef_ptr
    create_array(ULong length,
		 IDLType_ptr element_type)
        throw(SystemException);
    
    FixedDef_ptr
    create_fixed(UShort digits,
		 Short scale)
        throw(SystemException);
};

//
// The ModuleDef_impl class
//
class ModuleDef_impl : public virtual POA_CORBA::ModuleDef,
		       public virtual Container_impl,
		       public virtual Contained_impl
{
public:
    
    ModuleDef_impl(ORB_ptr, Container_ptr);
    virtual ~ModuleDef_impl();
    
    virtual void releaseInternal();
    
    virtual DefinitionKind def_kind()
        throw(SystemException)
    {
        return dk_Module;
    }

    virtual CORBA::Contained::Description* describe()
        throw(SystemException);
};

//
// The ConstantDef_impl class
//
class ConstantDef_impl : public virtual POA_CORBA::ConstantDef,
			 public virtual Contained_impl
{
    IDLType_var idlType_;
    Any value_;

public:

    ConstantDef_impl(ORB_ptr, Container_ptr);
    virtual ~ConstantDef_impl();
    
    virtual void releaseInternal();
    
    virtual DefinitionKind def_kind()
        throw(SystemException)
    {
        return dk_Constant;
    }
    
    virtual CORBA::Contained::Description* describe()
        throw(SystemException);
    
    virtual TypeCode_ptr type()
        throw(SystemException);
    
    virtual IDLType_ptr type_def()
        throw(SystemException);
    virtual void type_def(IDLType_ptr)
        throw(SystemException);

    virtual Any* value()
        throw(SystemException);
    virtual void value(const Any&)
        throw(SystemException);
};

//
// The TypedefDef_impl class
//
class TypedefDef_impl : public virtual POA_CORBA::TypedefDef,
			public virtual Contained_impl,
			public virtual IDLType_impl
{
public:

    TypedefDef_impl(ORB_ptr, Container_ptr);
    virtual ~TypedefDef_impl();

    virtual DefinitionKind def_kind()
        throw(SystemException)
    {
        return dk_Typedef;
    }

    virtual CORBA::Contained::Description* describe()
        throw(SystemException);
};

//
// The StructDef_impl class
//
class StructDef_impl : public virtual POA_CORBA::StructDef,
		       public virtual TypedefDef_impl,
		       public virtual Container_impl
{
    StructMemberSeq members_;
    
public:
    
    StructDef_impl(ORB_ptr, Container_ptr);
    virtual ~StructDef_impl();
    
    virtual void releaseInternal();
    
    virtual DefinitionKind def_kind()
        throw(SystemException)
    {
        return dk_Struct;
    }
    
    virtual TypeCode_ptr type()
        throw(SystemException);
    virtual TypeCode_ptr typeRecursive(ContainedSeq&)
        throw(SystemException);

    virtual StructMemberSeq* members()
        throw(SystemException);
    virtual void members(const StructMemberSeq&)
        throw(SystemException);
};

//
// The UnionDef_impl class
//
class UnionDef_impl : public virtual POA_CORBA::UnionDef,
		      public virtual TypedefDef_impl,
		      public virtual Container_impl
{
    IDLType_var discType_;
    UnionMemberSeq members_;
    
public:
    
    UnionDef_impl(ORB_ptr, Container_ptr);
    virtual ~UnionDef_impl();
    
    virtual void releaseInternal();
    
    virtual DefinitionKind def_kind()
        throw(SystemException)
    {
        return dk_Union;
    }

    virtual TypeCode_ptr type()
        throw(SystemException);
    virtual TypeCode_ptr typeRecursive(ContainedSeq&)
        throw(SystemException);

    virtual TypeCode_ptr discriminator_type()
        throw(SystemException);

    virtual IDLType_ptr discriminator_type_def()
        throw(SystemException);
    virtual void discriminator_type_def(IDLType_ptr)
        throw(SystemException);

    virtual UnionMemberSeq* members()
        throw(SystemException);
    virtual void members(const UnionMemberSeq&)
        throw(SystemException);
};

//
// The EnumDef_impl class
//
class EnumDef_impl : public virtual POA_CORBA::EnumDef,
		     public virtual TypedefDef_impl
{
    EnumMemberSeq members_;

public:

    EnumDef_impl(ORB_ptr, Container_ptr);
    virtual ~EnumDef_impl();

    virtual DefinitionKind def_kind()
        throw(SystemException)
    {
        return dk_Enum;
    }

    virtual TypeCode_ptr type()
        throw(SystemException);
    virtual TypeCode_ptr typeRecursive(ContainedSeq&)
        throw(SystemException)
    {
	return type();
    }

    virtual EnumMemberSeq* members()
        throw(SystemException);
    virtual void members(const EnumMemberSeq&)
        throw(SystemException);
};

//
// The AliasDef_impl class
//
class AliasDef_impl : public virtual POA_CORBA::AliasDef,
		      public virtual TypedefDef_impl
{
    IDLType_var idlType_;

public:

    AliasDef_impl(ORB_ptr, Container_ptr);
    virtual ~AliasDef_impl();

    virtual void releaseInternal();

    virtual DefinitionKind def_kind()
        throw(SystemException)
    {
        return dk_Alias;
    }

    virtual TypeCode_ptr type()
        throw(SystemException);
    virtual TypeCode_ptr typeRecursive(ContainedSeq&)
        throw(SystemException);

    virtual IDLType_ptr original_type_def()
        throw(SystemException);
    virtual void original_type_def(IDLType_ptr)
        throw(SystemException);
};

//
// The NativeDef_impl class
//
class NativeDef_impl : public virtual POA_CORBA::NativeDef,
		       public virtual TypedefDef_impl
{
public:

    NativeDef_impl(ORB_ptr, Container_ptr);
    virtual ~NativeDef_impl();

    virtual DefinitionKind def_kind()
        throw(SystemException)
    {
        return dk_Native;
    }

    virtual TypeCode_ptr type()
        throw(SystemException);
    virtual TypeCode_ptr typeRecursive(ContainedSeq&)
        throw(SystemException)
    {
	return type();
    }
};

//
// The PrimitiveDef_impl class
//
class PrimitiveDef_impl : public virtual POA_CORBA::PrimitiveDef,
			  public virtual IDLType_impl
{
    PrimitiveKind primitiveKind;

public:

    PrimitiveDef_impl(ORB_ptr, PrimitiveKind);
    virtual ~PrimitiveDef_impl();

    virtual DefinitionKind def_kind()
	throw(SystemException)
    {
	return dk_Primitive;
    }

    virtual void destroy()
        throw(SystemException);

    virtual TypeCode_ptr type()
        throw(SystemException);
    virtual TypeCode_ptr typeRecursive(ContainedSeq&)
        throw(SystemException)
    {
	return type();
    }

    virtual PrimitiveKind kind()
        throw(SystemException);
};

//
// The StringDef_impl class
//
class StringDef_impl : public virtual POA_CORBA::StringDef,
		       public virtual IDLType_impl
{
    Repository_var repository_;

    ULong bound_;

public:

    StringDef_impl(ORB_ptr, Repository_ptr);
    virtual ~StringDef_impl();

    virtual void releaseInternal();

    virtual DefinitionKind def_kind()
        throw(SystemException)
    {
        return dk_String;
    }

    virtual TypeCode_ptr type()
        throw(SystemException);
    virtual TypeCode_ptr typeRecursive(ContainedSeq&)
        throw(SystemException)
    {
	return type();
    }

    virtual ULong bound()
        throw(SystemException);
    virtual void bound(ULong)
        throw(SystemException);
};

//
// The WstringDef_impl class
//
class WstringDef_impl : public virtual POA_CORBA::WstringDef,
			public virtual IDLType_impl
{
    Repository_var repository_;

    ULong bound_;

public:

    WstringDef_impl(ORB_ptr, Repository_ptr);
    virtual ~WstringDef_impl();

    virtual void releaseInternal();

    virtual DefinitionKind def_kind()
        throw(SystemException)
    {
        return dk_Wstring;
    }

    virtual TypeCode_ptr type()
        throw(SystemException);
    virtual TypeCode_ptr typeRecursive(ContainedSeq&)
        throw(SystemException)
    {
	return type();
    }

    virtual ULong bound()
        throw(SystemException);
    virtual void bound(ULong)
        throw(SystemException);
};

//
// The FixedDef_impl class
//
class FixedDef_impl : public virtual POA_CORBA::FixedDef,
		      public virtual IDLType_impl
{
    Repository_var repository_;

    UShort digits_;
    Short scale_;
    
public:

    FixedDef_impl(ORB_ptr, Repository_ptr);
    virtual ~FixedDef_impl();

    virtual void releaseInternal();

    virtual DefinitionKind def_kind()
	throw(SystemException)
    {
	return dk_Fixed;
    }

    virtual TypeCode_ptr type()
	throw(SystemException);
    virtual TypeCode_ptr typeRecursive(ContainedSeq&)
        throw(SystemException)
    {
	return type();
    }

    virtual UShort digits()
	throw(SystemException);
    virtual void digits(UShort)
	throw(SystemException);

    virtual Short scale()
	throw(SystemException);
    virtual void scale(Short)
	throw(SystemException);
};

//
// The SequenceDef_impl class
//
class SequenceDef_impl : public virtual POA_CORBA::SequenceDef,
			 public virtual IDLType_impl
{
    Repository_var repository_;

    ULong bound_;
    IDLType_var elementType_;

public:

    SequenceDef_impl(ORB_ptr, Repository_ptr);
    virtual ~SequenceDef_impl();

    virtual void releaseInternal();

    virtual DefinitionKind def_kind()
	throw(SystemException)
    {
	return dk_Sequence;
    }

    virtual TypeCode_ptr type()
	throw(SystemException);
    virtual TypeCode_ptr typeRecursive(ContainedSeq&)
        throw(SystemException);

    virtual ULong bound()
	throw(SystemException);
    virtual void bound(ULong)
	throw(SystemException);

    virtual TypeCode_ptr element_type()
	throw(SystemException);

    virtual IDLType_ptr element_type_def()
	throw(SystemException);
    virtual void element_type_def(IDLType_ptr)
	throw(SystemException);
};

//
// The ArrayDef_impl class
//
class ArrayDef_impl : public virtual POA_CORBA::ArrayDef,
		      public virtual IDLType_impl
{
    Repository_var repository_;

    ULong length_;
    IDLType_var elementType_;

public:

    ArrayDef_impl(ORB_ptr, Repository_ptr);
    virtual ~ArrayDef_impl();

    virtual void releaseInternal();

    virtual DefinitionKind def_kind()
	throw(SystemException)
    {
    	return dk_Array;
    }

    virtual TypeCode_ptr type()
	throw(SystemException);
    virtual TypeCode_ptr typeRecursive(ContainedSeq&)
        throw(SystemException);

    virtual ULong length()
	throw(SystemException);
    virtual void length(ULong)
	throw(SystemException);

    virtual TypeCode_ptr element_type()
	throw(SystemException);

    virtual IDLType_ptr element_type_def()
	throw(SystemException);
    virtual void element_type_def(IDLType_ptr)
	throw(SystemException);
};

} // End of namespace CORBA

#endif
