// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <Types.h>
#include <IFR4_impl.h>

using namespace CORBA;
using namespace OB;
using namespace PortableServer;

// ----------------------------------------------------------------------
// ValueMemberDef_impl member implementations
// ----------------------------------------------------------------------

ValueMemberDef_impl::ValueMemberDef_impl(ORB_ptr orb, Container_ptr p)
    : IRObject_impl(orb),
      Contained_impl(orb, p)
{
}

ValueMemberDef_impl::~ValueMemberDef_impl()
{
}

void
ValueMemberDef_impl::releaseInternal()
{
    typeDef_ = CORBA::IDLType::_nil();
    Contained_impl::releaseInternal();
}

Contained::Description*
ValueMemberDef_impl::describe()
    throw(SystemException)
{
    ValueMember* desc = new ValueMember;

    desc -> name = name();
    desc -> id = id();

    Contained_var contained = CORBA::Contained::_narrow(definedIn_);
    if(!is_nil(contained))
        desc -> defined_in = contained -> id();
    else
        desc -> defined_in = string_dup("");

    desc -> version = version();
    desc -> type = type();
    desc -> type_def = type_def();
    desc -> access = access();

    CORBA::Contained::Description* description =
        new CORBA::Contained::Description;
    description -> kind = def_kind();
    description -> value <<= desc;

    return description;
}

TypeCode_ptr
ValueMemberDef_impl::type()
    throw(SystemException)
{
    // TODO: Call IdlCalcRecTypeCode(orb, seq, self)

    return typeDef_ -> type();
}

IDLType_ptr
ValueMemberDef_impl::type_def()
    throw(SystemException)
{
    return CORBA::IDLType::_duplicate(typeDef_);
}

void
ValueMemberDef_impl::type_def(IDLType_ptr t)
    throw(SystemException)
{
    typeDef_ = CORBA::IDLType::_duplicate(t);
}

Visibility
ValueMemberDef_impl::access()
    throw(SystemException)
{
    return access_;
}

void
ValueMemberDef_impl::access(Visibility v)
    throw(SystemException)
{
    access_ = v;
}

// ----------------------------------------------------------------------
// ValueDef_impl member implementations
// ----------------------------------------------------------------------

ValueDef_impl::ValueDef_impl(ORB_ptr orb, Container_ptr p)
    : IRObject_impl(orb),
      Container_impl(orb),
      Contained_impl(orb, p),
      IDLType_impl(orb),
      isAbstract_(FALSE),
      isCustom_(FALSE),
      isTruncatable_(FALSE)
{
}

ValueDef_impl::~ValueDef_impl()
{
}

void
ValueDef_impl::releaseInternal()
{
    supportedInterface_ = CORBA::InterfaceDef::_nil();
    initializers_.length(0);
    baseValue_ = CORBA::ValueDef::_nil();
    abstractBaseValues_.length(0);

    Container_impl::releaseInternal();
    Contained_impl::releaseInternal();
    IDLType_impl::releaseInternal();
}

Contained::Description*
ValueDef_impl::describe()
    throw(SystemException)
{
    ValueDescription* desc = new ValueDescription;
    ULong i;

    desc -> name = name();
    desc -> id = id();

    desc -> is_abstract = is_abstract();
    desc -> is_custom = is_custom();

    Contained_var contained = CORBA::Contained::_narrow(definedIn_);
    if(!is_nil(contained))
        desc -> defined_in = contained -> id();
    else
        desc -> defined_in = string_dup("");

    desc -> version = version();

    if(!is_nil(supportedInterface_))
        desc -> supported_interface = supportedInterface_ -> id();
    else
        desc -> supported_interface = string_dup("");

    desc -> abstract_base_values.length(abstractBaseValues_.length());
    for(i = 0 ; i < abstractBaseValues_.length() ; i++)
        desc -> abstract_base_values[i] = abstractBaseValues_[i] -> id();

    desc -> is_truncatable = is_truncatable();

    if(!is_nil(baseValue_))
        desc -> base_value = baseValue_ -> id();
    else
        desc -> base_value = string_dup("");

    CORBA::Contained::Description* description =
        new CORBA::Contained::Description;
    description -> kind = def_kind();
    description -> value <<= desc;

    return description;
}

TypeCode_ptr
ValueDef_impl::type()
    throw(SystemException)
{
    // TODO: Call IdlCalcRecTypeCode(orb, seq, self)

    TypeCode_var concreteBase;
    ValueModifier typeModifier = VTM_NONE;
    ValueMemberSeq members;

    if(is_abstract())
        typeModifier = VTM_ABSTRACT;
    else if(is_custom())
        typeModifier = VTM_CUSTOM;
    else if(is_truncatable())
        typeModifier = VTM_TRUNCATABLE;

    if(!is_nil(baseValue_))
        if(!baseValue_ -> is_abstract())
            concreteBase = baseValue_ -> type();

    for(ULong i = 0 ; i < contents_.length() ; i++)
    {
        if(contents_[i] -> def_kind() == dk_ValueMember)
        {
            CORBA::Contained::Description_var d = contents_[i] -> describe();
            ValueMember* memberDesc;
            Boolean b = d -> value >>= memberDesc;
            assert(b);
            members.append(*memberDesc);
        }
    }

    return orb_ -> create_value_tc(id_, name_, typeModifier, concreteBase,
                                   members);
}

Contained_ptr
ValueDef_impl::lookup(const char* search_name)
    throw(SystemException)
{
    Contained_var contained = Container_impl::lookup(search_name);

    if(is_nil(contained))
    {
        if(!is_nil(supportedInterface_))
            contained = supportedInterface_ -> lookup(search_name);
    }

    if(is_nil(contained))
    {
        if(!is_nil(baseValue_))
            contained = baseValue_ -> lookup(search_name);
    }

    if(is_nil(contained))
    {
        for(ULong i = 0 ; i < abstractBaseValues_.length() ; i++)
        {
            Contained_var contained2 =
                abstractBaseValues_[i] -> lookup(search_name);

            if(!is_nil(contained2))
            {
                if(!is_nil(contained))
                {
		    RepositoryId_var id = contained -> id();
		    RepositoryId_var id2 = contained2 -> id();
		    if(strcmp(id, id2) != 0)
                    {
                        //
                        // Name is ambiguous
                        //
                        throw INTF_REPOS(MinorLookupAmbiguous, COMPLETED_NO);
                    }
                }

                contained = contained2;
            }
        }
    }

    return contained._retn();
}

ContainedSeq*
ValueDef_impl::contents(DefinitionKind limit_type, Boolean exclude_inherited)
    throw(SystemException)
{
    ContainedSeq* s =
        Container_impl::contents(limit_type, exclude_inherited);

    if(!exclude_inherited)
    {
        if(!is_nil(supportedInterface_))
        {
            ContainedSeq_var s2 =
                supportedInterface_ -> contents(limit_type, false);
            merge(s, s2);
        }

        if(!is_nil(baseValue_))
        {
            ContainedSeq_var s2 = baseValue_ -> contents(limit_type, false);
            merge(s, s2);
        }

        for(ULong i = 0 ; i < abstractBaseValues_.length() ; i++)
        {
            ContainedSeq_var s2 =
                abstractBaseValues_[i] -> contents(limit_type, false);
            merge(s, s2);
        }
    }

    return s;
}

ContainedSeq*
ValueDef_impl::lookup_name(const char* search_name,
                           Long levels_to_search,
                           DefinitionKind limit_type,
                           Boolean exclude_inherited)
    throw(SystemException)
{
    ContainedSeq* s =
        Container_impl::lookup_name(search_name,
                                    levels_to_search,
                                    limit_type,
                                    exclude_inherited);

    if(!exclude_inherited)
    {
        if(!is_nil(supportedInterface_))
        {
            ContainedSeq_var s2 =
                supportedInterface_ -> lookup_name(search_name,
                                                   levels_to_search,
                                                   limit_type,
                                                   false);
            merge(s, s2);
        }

        if(!is_nil(baseValue_))
        {
            ContainedSeq_var s2 =
                baseValue_ -> lookup_name(search_name,
                                          levels_to_search,
                                          limit_type,
                                          false);
            merge(s, s2);
        }

        for(ULong i = 0 ; i < abstractBaseValues_.length() ; i++)
        {
            ContainedSeq_var s2 =
                abstractBaseValues_[i] -> lookup_name(search_name,
                                                      levels_to_search,
                                                      limit_type,
                                                      false);
            merge(s, s2);
        }
    }

    return s;
}

InterfaceDef_ptr
ValueDef_impl::supported_interface()
    throw(SystemException)
{
    return CORBA::InterfaceDef::_duplicate(supportedInterface_);
}

void
ValueDef_impl::supported_interface(InterfaceDef_ptr p)
    throw(SystemException)
{
    if(!is_nil(p))
    {
        ContainedSeq_var cont = contents(dk_all, false);
        ContainedSeq_var seq;
        ULong i, j;

        //
        // Check for conflicts in operations
        //
        seq = p -> contents(dk_Operation, false);
        for(i = 0 ; i < seq -> length() ; i++)
        {
            Identifier_var name1 = seq[i] -> name();
            for(j = 0 ; j < cont -> length() ; j++)
            {
                Identifier_var name2 = cont[j] -> name();
                if(strcmp(name1, name2) == 0)
                    throw BAD_PARAM(5, COMPLETED_NO);
            }
        }

        //
        // Check for conflicts in attributes
        //
        seq = p -> contents(dk_Attribute, false);
        for(i = 0 ; i < seq -> length() ; i++)
        {
            Identifier_var name1 = seq[i] -> name();
            for(j = 0 ; j < cont -> length() ; j++)
            {
                Identifier_var name2 = cont[j] -> name();
                if(strcmp(name1, name2) == 0)
                    throw BAD_PARAM(5, COMPLETED_NO);
            }
        }
    }

    supportedInterface_ = CORBA::InterfaceDef::_duplicate(p);
}

InitializerSeq*
ValueDef_impl::initializers()
    throw(SystemException)
{
    return new InitializerSeq(initializers_);
}

void
ValueDef_impl::initializers(const InitializerSeq& seq)
    throw(SystemException)
{
    ULong i, j, k;

    //
    // Cannot have initializers if abstract
    //
    if(is_abstract() && seq.length() > 0)
        throw BAD_PARAM(MinorAbstractValueInit, COMPLETED_NO);

    //
    // Ensure uniqueness of the signatures of all initializers -
    // we do NOT need to check the initializers of less-derived types
    //
    for(i = 0 ; i < seq.length() ; i++)
    {
        for(j = 0 ; j < seq[i].members.length() ; j++)
        {
            for(k = j + 1 ; k < seq[i].members.length() ; k++)
            {
                if(strcmp(seq[i].members[j].name, seq[i].members[k].name) == 0)
                    throw BAD_PARAM(MinorDuplicateFormalParam, COMPLETED_NO);
            }
        }

        for(j = i + 1 ; j < seq.length() ; j++)
        {
            if(seq[i].members.length() == seq[j].members.length())
            {
                for(k = 0 ; k < seq[i].members.length() ; k++)
                {
                    if(!seq[i].members[k].type ->
                       equal(seq[j].members[k].type))
                        break;
                }

                if(k >= seq[i].members.length())
                    throw BAD_PARAM(MinorDuplicateValueInit, COMPLETED_NO);
            }
        }
    }

    initializers_ = seq;
}

ValueDef_ptr
ValueDef_impl::base_value()
    throw(SystemException)
{
    return CORBA::ValueDef::_duplicate(baseValue_);
}

void
ValueDef_impl::base_value(ValueDef_ptr p)
    throw(SystemException)
{
    if(!is_nil(p))
    {
        //
        // Cannot have base value if abstract
        //
        if(is_abstract())
            throw BAD_PARAM(MinorInvalidValueInheritance, COMPLETED_NO);

        //
        // Base value cannot be abstract
        //
        if(p -> is_abstract())
            throw BAD_PARAM(MinorInvalidValueInheritance, COMPLETED_NO);

        //
        // A non-custom value cannot inherit from a custom value
        //
        if(!is_custom() && p -> is_custom())
            throw BAD_PARAM(MinorInvalidValueInheritance, COMPLETED_NO);

        //
        // If this value is truncatable, then its base value must
        // be truncatable (if the base value has a concrete base type)
        //
        if(is_truncatable())
        {
            ValueDef_var base = p -> base_value();
            if(!is_nil(base) && !p -> is_truncatable())
                throw BAD_PARAM(MinorInvalidValueInheritance, COMPLETED_NO);
        }

        ContainedSeq_var cont = contents(dk_all, false);
        ContainedSeq_var seq = p -> contents(dk_all, false);
        ULong i, j;

        //
        // Check for conflicts
        //
        for(i = 0 ; i < seq -> length() ; i++)
        {
            Identifier_var name1 = seq[i] -> name();
            for(j = 0 ; j < cont -> length() ; j++)
            {
                Identifier_var name2 = cont[j] -> name();
                if(strcmp(name1, name2) == 0)
                    throw BAD_PARAM(5, COMPLETED_NO);
            }
        }
    }

    baseValue_ = CORBA::ValueDef::_duplicate(p);
}

ValueDefSeq*
ValueDef_impl::abstract_base_values()
    throw(SystemException)
{
    return new ValueDefSeq(abstractBaseValues_);
}

void
ValueDef_impl::abstract_base_values(const ValueDefSeq& seq)
    throw(SystemException)
{
    ULong i, j, k;

    //
    // Check that all valuetypes are abstract, and that there are
    // no duplicates
    //
    for(i = 0 ; i < seq.length() ; i++)
    {
        if(!seq[i] -> is_abstract())
            throw BAD_PARAM(MinorInvalidValueInheritance, COMPLETED_NO);

        String_var id1 = seq[i] -> id();
        for(j = i + 1 ; j < seq.length() ; j++)
        {
            String_var id2 = seq[j] -> id();
            if(strcmp(id1, id2) == 0)
                throw BAD_PARAM(MinorDuplicateBaseType, COMPLETED_NO);
        }
    }

    //
    // Check for conflicts
    //
    ContainedSeq_var cont = contents(dk_all, false);
    for(i = 0 ; i < seq.length() ; i++)
    {
        ContainedSeq_var s = seq[i] -> contents(dk_all, false);

        for(j = 0 ; j < s -> length() ; j++)
        {
            Identifier_var name1 = s[j] -> name();
            for(k = 0 ; k < cont -> length() ; k++)
            {
                Identifier_var name2 = cont[k] -> name();
                if(strcmp(name1, name2) == 0)
                    throw BAD_PARAM(5, COMPLETED_NO);
            }
        }
    }

    abstractBaseValues_ = seq;
}

Boolean
ValueDef_impl::is_abstract()
    throw(SystemException)
{
    return isAbstract_;
}

void
ValueDef_impl::is_abstract(Boolean b)
    throw(SystemException)
{
    if(b)
    {
        //
        // Cannot be abstract and custom
        //
        if(is_custom())
            throw BAD_PARAM(MinorInvalidValueModifier, COMPLETED_NO);

        //
        // Cannot be abstract and truncatable
        //
        if(is_truncatable())
            throw BAD_PARAM(MinorInvalidValueModifier, COMPLETED_NO);

        //
        // Cannot have concrete base value
        //
        if(!is_nil(baseValue_))
            throw BAD_PARAM(MinorInvalidValueInheritance, COMPLETED_NO);
    }

    isAbstract_ = b;
}

Boolean
ValueDef_impl::is_custom()
    throw(SystemException)
{
    return isCustom_;
}

void
ValueDef_impl::is_custom(Boolean b)
    throw(SystemException)
{
    if(b)
    {
        //
        // Cannot be custom and truncatable
        //
        if(is_truncatable())
            throw BAD_PARAM(MinorInvalidValueModifier, COMPLETED_NO);

        //
        // Cannot be abstract and custom
        //
        if(is_abstract())
            throw BAD_PARAM(MinorInvalidValueModifier, COMPLETED_NO);
    }

    isCustom_ = b;
}

Boolean
ValueDef_impl::is_truncatable()
    throw(SystemException)
{
    return isTruncatable_;
}

void
ValueDef_impl::is_truncatable(Boolean b)
    throw(SystemException)
{
    if(b)
    {
        //
        // Cannot be custom and truncatable
        //
        if(is_custom())
            throw BAD_PARAM(MinorInvalidValueModifier, COMPLETED_NO);

        //
        // Cannot be abstract and truncatable
        //
        if(is_abstract())
            throw BAD_PARAM(MinorInvalidValueModifier, COMPLETED_NO);

        //
        // If this value is truncatable, then its base value must
        // be truncatable (if the base value has a concrete base type)
        //
        if(!is_nil(baseValue_))
        {
            ValueDef_var base = baseValue_ -> base_value();
            if(!is_nil(base) && !baseValue_ -> is_truncatable())
                throw BAD_PARAM(MinorInvalidValueInheritance, COMPLETED_NO);
        }
    }

    isTruncatable_ = b;
}

Boolean
ValueDef_impl::is_a(const char* value_id)
    throw(SystemException)
{
    if(strcmp(value_id, id_) == 0)
        return TRUE;

    if(!is_nil(supportedInterface_))
        if(supportedInterface_ -> is_a(value_id))
            return TRUE;

    if(!is_nil(baseValue_))
        if(baseValue_ -> is_a(value_id))
            return TRUE;

    for(ULong i = 0 ; i < abstractBaseValues_.length() ; i++)
        if(abstractBaseValues_[i] -> is_a(value_id))
            return TRUE;

    return FALSE;
}

CORBA::ValueDef::FullValueDescription*
ValueDef_impl::describe_value()
    throw(SystemException)
{
    CORBA::ValueDef::FullValueDescription* desc =
        new CORBA::ValueDef::FullValueDescription;
    ULong i;

    desc -> name = name();
    desc -> id = id();
    desc -> is_abstract = is_abstract();
    desc -> is_custom = is_custom();

    Contained_var contained = CORBA::Contained::_narrow(definedIn_);
    if(!is_nil(contained))
        desc -> defined_in = contained -> id();
    else
        desc -> defined_in = string_dup("");

    desc -> version = version();

    //
    // Collect operations, attributes and members
    //
    for(i = 0 ; i < contents_.length() ; i++)
    {
        Contained_ptr contained = contents_[i];

        OperationDef_var operation = CORBA::OperationDef::_narrow(contained);
        AttributeDef_var attribute = CORBA::AttributeDef::_narrow(contained);
        ValueMemberDef_var member = CORBA::ValueMemberDef::_narrow(contained);

        if(!is_nil(operation))
        {
            CORBA::Contained::Description_var d = operation -> describe();
            OperationDescription* operationDesc;
            Boolean b = d -> value >>= operationDesc;
            assert(b);

            desc -> operations.append(*operationDesc);
        }
        else if(!is_nil(attribute))
        {
            CORBA::Contained::Description_var d = attribute -> describe();
            AttributeDescription* attributeDesc;
            Boolean b = d -> value >>= attributeDesc;
            assert(b);

            desc -> attributes.append(*attributeDesc);
        }
        else if(!is_nil(member))
        {
            CORBA::Contained::Description_var d = member -> describe();
            ValueMember* memberDesc;
            Boolean b = d -> value >>= memberDesc;
            assert(b);

            desc -> members.append(*memberDesc);
        }
    }

    desc -> initializers = initializers_;

    //
    // Collect supported interfaces
    //

    if(!is_nil(supportedInterface_))
    {
        desc -> supported_interfaces.append(supportedInterface_ -> id());
    }

    if(!is_nil(baseValue_))
    {
        InterfaceDef_var intf = baseValue_ -> supported_interface();
        if(!is_nil(intf))
            desc -> supported_interfaces.append(intf -> id());
    }

    for(i = 0 ; i < abstractBaseValues_.length() ; i++)
    {
        InterfaceDef_var intf =
            abstractBaseValues_[i] -> supported_interface();
        if(!is_nil(intf))
            desc -> supported_interfaces.append(intf -> id());
    }

    //
    // Copy abstract value IDs
    //
    desc -> abstract_base_values.length(abstractBaseValues_.length());
    for(i = 0 ; i < abstractBaseValues_.length() ; i++)
        desc -> abstract_base_values[i] = abstractBaseValues_[i] -> id();

    desc -> is_truncatable = is_truncatable();

    if(!is_nil(baseValue_))
        desc -> base_value = baseValue_ -> id();
    else
        desc -> base_value = string_dup("");

    desc -> type = type();

    return desc;
}

ValueMemberDef_ptr
ValueDef_impl::create_value_member(
    const char* id,
    const char* name,
    const char* version,
    IDLType_ptr type,
    Visibility access)
    throw(SystemException)
{
    ValueDef_var self = _this();
    ValueMemberDef_impl* p = new ValueMemberDef_impl(orb_, self);
    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
        p -> type_def(type);
        p -> access(access);
    }
    catch(...)
    {
        p -> _remove_ref();
        throw;
    }
          
    contents_.append(p -> _this());
    return p -> _this();
}

AttributeDef_ptr
ValueDef_impl::create_attribute(
    const char* id,
    const char* name,
    const char* version,
    IDLType_ptr type,
    AttributeMode mode)
    throw(SystemException)
{
    ValueDef_var self = _this();
    AttributeDef_impl* p = new AttributeDef_impl(orb_, self);

    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
        p -> type_def(type);
        p -> mode(mode);
    }
    catch(...)
    {
        p -> _remove_ref();
        throw;
    }
    contents_.append(p -> _this());
    return p -> _this();
}

OperationDef_ptr
ValueDef_impl::create_operation(
    const char* id,
    const char* name,
    const char* version,
    IDLType_ptr result,
    OperationMode mode,
    const ParDescriptionSeq& params,
    const ExceptionDefSeq& exceptions,
    const ContextIdSeq& contexts)
    throw(SystemException)
{
    ValueDef_var self = _this();
    OperationDef_impl* p = new OperationDef_impl(orb_, self);

    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
        p -> result_def(result);
        p -> mode(mode);
        p -> params(params);
        p -> exceptions(exceptions);
        p -> contexts(contexts);
    }
    catch(...)
    {
        p -> _remove_ref();
        throw;
    }
    contents_.append(p -> _this());
    return p -> _this();
}

// ----------------------------------------------------------------------
// ValueBoxDef_impl member implementations
// ----------------------------------------------------------------------

ValueBoxDef_impl::ValueBoxDef_impl(ORB_ptr orb, Container_ptr p)
    : IRObject_impl(orb),
      Contained_impl(orb, p),
      IDLType_impl(orb),
      TypedefDef_impl(orb, p)
{
}

ValueBoxDef_impl::~ValueBoxDef_impl()
{
}

void
ValueBoxDef_impl::releaseInternal()
{
    originalTypeDef_ = CORBA::IDLType::_nil();
    TypedefDef_impl::releaseInternal();
}

TypeCode_ptr
ValueBoxDef_impl::type()
    throw(SystemException)
{
    ContainedSeq seq;
    TypeCode_var tc = typeRecursive(seq);
    assert(seq.length() == 0);
    return tc._retn();
}

TypeCode_ptr
ValueBoxDef_impl::typeRecursive(ContainedSeq& seq)
    throw(SystemException)
{
    if(is_nil(originalTypeDef_))
        throw BAD_INV_ORDER();

    seq.length(seq.length() + 1);
    seq[seq.length() - 1] = _this();
    
    CORBA::Contained_var contained =
        CORBA::Contained::_narrow(originalTypeDef_);
    if(!is_nil(contained))
    {
	RepositoryId_var id = contained -> id();
	for(ULong i = 0 ; i < seq.length() ; i++)
	{
	    RepositoryId_var id2 = seq[i] -> id();
	    if(strcmp(id, id2) == 0)
	    {
		throw INTF_REPOS(MinorIllegalRecursion, COMPLETED_NO);
	    }
	}
    }

    POA_var poa = _default_POA();
    ServantBase_var servant = poa -> reference_to_servant(originalTypeDef_);
    IDLType_impl* originalTypeDefImpl =
	dynamic_cast<IDLType_impl*>(servant.in());
    assert(originalTypeDefImpl);
    TypeCode_var tc = originalTypeDefImpl -> typeRecursive(seq);

    seq.length(seq.length() - 1);
    return orb_ -> create_boxed_value_tc(id_, name_, tc);
}

IDLType_ptr
ValueBoxDef_impl::original_type_def()
    throw(SystemException)
{
    return CORBA::IDLType::_duplicate(originalTypeDef_);
}

void
ValueBoxDef_impl::original_type_def(IDLType_ptr p)
    throw(SystemException)
{
    originalTypeDef_ = CORBA::IDLType::_duplicate(p);
}
