// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef IFR4_IMPL_H
#define IFR4_IMPL_H

#include <IFR3_impl.h>
#include <OB/IFR4_skel.h>

namespace CORBA
{

//
// The ValueMemberDef_impl class
//
class ValueMemberDef_impl : public virtual POA_CORBA::ValueMemberDef,
                            public virtual Contained_impl
{
    IDLType_var typeDef_;
    Visibility access_;

public:
    ValueMemberDef_impl(ORB_ptr, Container_ptr);
    virtual ~ValueMemberDef_impl();

    virtual DefinitionKind def_kind()
	throw(SystemException)
    {
	return dk_ValueMember;
    }

    virtual void releaseInternal();

    virtual CORBA::Contained::Description* describe()
	throw(SystemException);

    virtual TypeCode_ptr type()
	throw(SystemException);
    virtual TypeCode_ptr typeRecursive(ContainedSeq&)
        throw(SystemException)
    {
	// TODO
	return type();
    }

    virtual IDLType_ptr type_def()
	throw(SystemException);
    virtual void type_def(IDLType_ptr)
	throw(SystemException);

    virtual Visibility access()
	throw(SystemException);
    virtual void access(Visibility)
	throw(SystemException);
};

//
// The ValueDef_impl class
//
class ValueDef_impl : public virtual POA_CORBA::ValueDef,
		      public virtual Container_impl,
		      public virtual Contained_impl,
		      public virtual IDLType_impl
{
    InterfaceDef_var supportedInterface_;
    InitializerSeq initializers_;
    ValueDef_var baseValue_;
    ValueDefSeq abstractBaseValues_;
    Boolean isAbstract_;
    Boolean isCustom_;
    Boolean isTruncatable_;

public:
    ValueDef_impl(ORB_ptr, Container_ptr);
    virtual ~ValueDef_impl();

    virtual DefinitionKind def_kind()
	throw(SystemException)
    {
	return dk_Value;
    }

    virtual void releaseInternal();

    virtual CORBA::Contained::Description* describe()
	throw(SystemException);

    virtual TypeCode_ptr type()
	throw(SystemException);
    virtual TypeCode_ptr typeRecursive(ContainedSeq&)
        throw(SystemException)
    {
	// TODO
	return type();
    }

    virtual Contained_ptr lookup(const char* search_name)
	throw(SystemException);

    virtual ContainedSeq* contents(DefinitionKind limit_type,
                                   Boolean exclude_inherited)
	throw(SystemException);

    virtual ContainedSeq* lookup_name(const char* search_name,
                                      Long levels_to_search,
                                      DefinitionKind limit_type,
                                      Boolean exclude_inherited)
	throw(SystemException);

    virtual InterfaceDef_ptr supported_interface()
	throw(SystemException);
    virtual void supported_interface(InterfaceDef_ptr)
	throw(SystemException);

    virtual InitializerSeq* initializers()
	throw(SystemException);
    virtual void initializers(const InitializerSeq&)
	throw(SystemException);

    virtual ValueDef_ptr base_value()
	throw(SystemException);
    virtual void base_value(ValueDef_ptr)
	throw(SystemException);

    virtual ValueDefSeq* abstract_base_values()
	throw(SystemException);
    virtual void abstract_base_values(const ValueDefSeq&)
	throw(SystemException);

    virtual Boolean is_abstract()
	throw(SystemException);
    virtual void is_abstract(Boolean)
	throw(SystemException);

    virtual Boolean is_custom()
	throw(SystemException);
    virtual void is_custom(Boolean)
	throw(SystemException);

    virtual Boolean is_truncatable()
	throw(SystemException);
    virtual void is_truncatable(Boolean)
	throw(SystemException);

    virtual Boolean is_a(const char* value_id)
	throw(SystemException);

    virtual CORBA::ValueDef::FullValueDescription* describe_value()
	throw(SystemException);

    virtual ValueMemberDef_ptr create_value_member(
        const char* id,
        const char* name,
        const char* version,
        IDLType_ptr type,
        Visibility access)
	throw(SystemException);

    virtual AttributeDef_ptr create_attribute(
        const char* id,
        const char* name,
        const char* version,
        IDLType_ptr type,
        AttributeMode mode)
	throw(SystemException);

    virtual OperationDef_ptr create_operation(
        const char* id,
        const char* name,
        const char* version,
        IDLType_ptr result,
        OperationMode mode,
        const ParDescriptionSeq& params,
        const ExceptionDefSeq& exceptions,
        const ContextIdSeq& contexts)
	throw(SystemException);
};

//
// The ValueBoxDef_impl class
//
class ValueBoxDef_impl : public virtual POA_CORBA::ValueBoxDef,
                         public virtual TypedefDef_impl
{
    IDLType_var originalTypeDef_;

public:
    ValueBoxDef_impl(ORB_ptr, Container_ptr);
    virtual ~ValueBoxDef_impl();

    virtual DefinitionKind def_kind()
	throw(SystemException)
    {
	return dk_ValueBox;
    }

    virtual void releaseInternal();

    virtual TypeCode_ptr type()
	throw(SystemException);
    virtual TypeCode_ptr typeRecursive(ContainedSeq&)
        throw(SystemException);

    virtual IDLType_ptr original_type_def()
	throw(SystemException);
    virtual void original_type_def(IDLType_ptr)
	throw(SystemException);
};

} // End of namespace CORBA

#endif
