// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Convert strings with escape sequences
// ----------------------------------------------------------------------

void
IdlAddEscapes(char*& str, bool addQuotes)
{
    char* s = str;
    String_var result = string_dup("");

    if(addQuotes)
	result += '"';

    while(*s)
    {
	switch(*s)
	{
	case '\n':
	    result += "\\n";
	    break;

	case '\t':
	    result += "\\t";
	    break;

	case '\v':
	    result += "\\v";
	    break;

	case '\b':
	    result += "\\b";
	    break;

	case '\r':
	    result += "\\r";
	    break;

	case '\f':
	    result += "\\f";
	    break;

	case '\a':
	    result += "\\a";
	    break;

	case '\\':
	    result += "\\\\";
	    break;
	    
 	case '\?':
 	    result += "\\?";
 	    break;
	    
	case '\'':
	    result += "\\'";
	    break;

	case '"':
	    result += "\\\"";
	    break;

	default:
	    if(isprint(*s))
		result += *s;
	    else
	    {
		Char c = *s;
		char str[5];
		sprintf(str, "\\%.3o", (unsigned char)c);
		result += str;
	    }
	    break;
	}
	
	s++;
    }

    if(addQuotes)
	result += '"';

    string_free(str);
    str = result._retn();
}

void
IdlAddEscapes(wchar_t*& str, bool addQuotes)
{
    wchar_t* s = str;
    WString_var result = wstring_dup(L"");;

    if(addQuotes)
	result += L'"';

    while(*s)
    {
	switch(*s)
	{
	case L'\n':
	    result += L"\\n";
	    break;

	case L'\t':
	    result += L"\\t";
	    break;

	case L'\v':
	    result += L"\\v";
	    break;

	case L'\b':
	    result += L"\\b";
	    break;

	case L'\r':
	    result += L"\\r";
	    break;

	case L'\f':
	    result += L"\\f";
	    break;

	case L'\a':
	    result += L"\\a";
	    break;

	case L'\\':
	    result += L"\\\\";
	    break;
	    
 	case L'\?':
 	    result += L"\\?";
 	    break;
	    
	case L'\'':
	    result += L"\\'";
	    break;

	case L'"':
	    result += L"\\\"";
	    break;

	default:
// XXX Alternatives wanted
	    if(*s < 256)
	    {
		if(isprint((char)*s))
		    result += *s;
		else
		{
		    result += *s;
		}
	    }
	    else
		result += *s;
	    break;
	}
	
	s++;
    }

    if(addQuotes)
	result += L'"';

    wstring_free(str);
    str = result._retn();
}

bool
IdlRemoveEscapes(char*& str, bool removeQuotes)
{
    char* s = str;
    String_var result = string_dup("");
    bool b = true;

    if(removeQuotes && *s == '"')
	s++;

    while(*s)
    {
	if(removeQuotes && *(s + 1) == 0 && *s == '"')
	    break;

	if(*s != '\\')
	{
	    result += *s;
	    s++;
	}
	else
	{
	    switch(*(s + 1))
	    {
	    case 'n':
		result += '\n';
		s += 2;
		break;

	    case 't':
		result += '\t';
		s += 2;
		break;

	    case 'v':
		result += '\v';
		s += 2;
		break;

	    case 'b':
		result += '\b';
		s += 2;
		break;

	    case 'r':
		result += '\r';
		s += 2;
		break;

	    case 'f':
		result += '\f';
		s += 2;
		break;

	    case 'a':
		result += '\a';
		s += 2;
		break;

	    case '\\':
		result += '\\';
		s += 2;
		break;

	    case '?':
		result += '\?';
		s += 2;
		break;

	    case '\'':
		result += '\'';
		s += 2;
		break;

	    case '"':
		result += '\"';
		s += 2;
		break;

	    case 'x':
	    {
		if((*(s + 2) >= '0' && *(s + 2) <= '9') ||
		   (*(s + 2) >= 'a' && *(s + 2) <= 'f') ||
		   (*(s + 2) >= 'A' && *(s + 2) <= 'F'))
		{
		    char* s2;
		    Char c = (Char)strtol(s + 2, &s2, 16);

 		    if(c == 0)
			b = false;

		    result += c;

		    s = s2;
		}
		else
		{
		    result += '\\';
		    s++;
		}

		break;
	    }
	    
	    case '0':
	    case '1':
	    case '2':
	    case '3':
	    case '4':
	    case '5':
	    case '6':
	    case '7':
	    {
		char* s2;
		Char c = (Char)strtol(s + 1, &s2, 8);
		
 		if(c == 0)
		    b = false;
			
		result += c;

		s = s2;

		break;
	    }

	    default:
		result += '\\';
		s++;
		break;
	    }
	}
    }

    string_free(str);
    str = result._retn();

    return b;
}

// ----------------------------------------------------------------------
// Calculate a recursive TypeCode for an IDLType
// ----------------------------------------------------------------------

/*
char*
IdlCalcRecTypeCodeHelper(const ContainedSeq& seq,
			 IRObject_ptr irObject,
			 DefinitionKind& defKind)
{
    Contained_var contained = Contained::_narrow(irObject);
    if(!is_nil(contained))
    {
	RepositoryId_var id = contained -> id();
	for(ULong i = 0 ; i < seq.length() ; i++)
	{
	    RepositoryId_var id2 = seq[i] -> id();
	    if(strcmp(id, id2) == 0)
	    {
		defKind = contained -> def_kind();
		return id._retn();
	    }
	}
    }

    defKind = dk_none;
    return string_dup("");
}

TypeCode_ptr
IdlCalcRecTypeCode(ORB_ptr orb, ContainedSeq& seq, IRObject_ptr irObject)
{
    TypeCode_var result;

    Contained_var contained = Contained::_narrow(irObject);
    if(!is_nil(contained))
    {
	seq.length(seq.length() + 1);
	seq[seq.length() - 1] = contained;
    }

    DefinitionKind defKind = irObject -> def_kind();
    switch(defKind)
    {
    case dk_none:
    case dk_all:
    case dk_Constant:
    case dk_Attribute:
    case dk_Operation:
    case dk_Module:
    case dk_Repository:
	assert(false);
	break;

    case dk_Typedef:
	assert(false); // What is dk_Typedef for anyway?
	break;

    case dk_Primitive:
    case dk_String:
    case dk_Wstring:
    case dk_Enum:
    case dk_Fixed:
    case dk_Native:
    case dk_Interface:
    {
	IDLType_var idlType = IDLType::_narrow(irObject);
	assert(!is_nil(idlType));
	result = idlType -> type();
	break;
    }

    case dk_Alias:
    {
	AliasDef_var aliasDef = AliasDef::_narrow(irObject);
	assert(!is_nil(aliasDef));
	IDLType_var origIdlType = aliasDef -> original_type_def();

	DefinitionKind origDefKind;
	RepositoryId_var origId =
	    IdlCalcRecTypeCodeHelper(seq, origIdlType, origDefKind);
	if(origDefKind != dk_none)
	{
	    throw INTF_REPOS(MinorIllegalRecursion, COMPLETED_NO);
	}
	
	TypeCode_var origType =
	    IdlCalcRecTypeCode(orb, seq, origIdlType);
	RepositoryId_var id = aliasDef -> id();
	Identifier_var name = aliasDef -> name();
	result = orb -> create_alias_tc(id, name, origType);
	break;
    }

    case dk_Struct:
    {
	StructDef_var structDef = StructDef::_narrow(irObject);
	assert(!is_nil(structDef));
	StructMemberSeq_var members = structDef -> members();
	for(ULong i = 0 ; i < members -> length() ; i++)
	    members[i].type =
		IdlCalcRecTypeCode(orb, seq, members[i].type_def);

	RepositoryId_var id = structDef -> id();
	Identifier_var name = structDef -> name();
	result = orb -> create_struct_tc(id, name, members);
	break;
    }

    case dk_Exception:
    {
	ExceptionDef_var excDef = ExceptionDef::_narrow(irObject);
	assert(!is_nil(excDef));
	StructMemberSeq_var members = excDef -> members();
	for(ULong i = 0 ; i < members -> length() ; i++)
	    members[i].type =
		IdlCalcRecTypeCode(orb, seq, members[i].type_def);

	RepositoryId_var id = excDef -> id();
	Identifier_var name = excDef -> name();
	result = orb -> create_exception_tc(id, name, members);
	break;
    }

    case dk_Union:
    {
	UnionDef_var unionDef = UnionDef::_narrow(irObject);
	assert(!is_nil(unionDef));
	UnionMemberSeq_var members = unionDef -> members();
	for(ULong i = 0 ; i < members -> length() ; i++)
	    members[i].type =
		IdlCalcRecTypeCode(orb, seq, members[i].type_def);

	RepositoryId_var id = unionDef -> id();
	Identifier_var name = unionDef -> name();
	TypeCode_var discType = unionDef -> discriminator_type();
	result = orb -> create_union_tc(id, name, discType, members);
	break;
    }

    case dk_Sequence:
    {
	SequenceDef_var seqDef = SequenceDef::_narrow(irObject);
	assert(!is_nil(seqDef));
	IDLType_var elemIdlType = seqDef -> element_type_def();

	TypeCode_var elemType;
	DefinitionKind elemDefKind;
	RepositoryId_var elemId =
	    IdlCalcRecTypeCodeHelper(seq, elemIdlType, elemDefKind);
	if(elemDefKind != dk_none)
	{
	    if(elemDefKind != dk_Struct &&
	       elemDefKind != dk_Exception &&
	       elemDefKind != dk_Union)
	    {
		throw INTF_REPOS(MinorIllegalRecursion, COMPLETED_NO);
	    }
	    
	    elemType = orb -> create_recursive_tc(elemId);
	}
	
	if(is_nil(elemType))
	    elemType = IdlCalcRecTypeCode(orb, seq, elemIdlType);
	result = orb -> create_sequence_tc(seqDef -> bound(), elemType);
	break;
    }

    case dk_Array:
    {
	ArrayDef_var arrDef = ArrayDef::_narrow(irObject);
	assert(!is_nil(arrDef));
	IDLType_var elemIdlType = arrDef -> element_type_def();

	DefinitionKind elemDefKind;
	RepositoryId_var elemId =
	    IdlCalcRecTypeCodeHelper(seq, elemIdlType, elemDefKind);
	if(elemDefKind != dk_none)
	{
	    throw INTF_REPOS(MinorIllegalRecursion, COMPLETED_NO);
	}
	
	TypeCode_var elemType =
	    IdlCalcRecTypeCode(orb, seq, elemIdlType);
	result = orb -> create_array_tc(arrDef -> length(), elemType);
	break;
    }

    case dk_ValueBox:
    {
	ValueBoxDef_var valueBoxDef =
	    ValueBoxDef::_narrow(irObject);
	assert(!is_nil(valueBoxDef));
	IDLType_var origIdlType = valueBoxDef -> original_type_def();
	
	DefinitionKind origDefKind;
	RepositoryId_var origId =
	    IdlCalcRecTypeCodeHelper(seq, origIdlType, origDefKind);
	if(origDefKind != dk_none)
	{
	    throw INTF_REPOS(MinorIllegalRecursion, COMPLETED_NO);
	}
	
	TypeCode_var origType =
	    IdlCalcRecTypeCode(orb, seq, origIdlType);
	RepositoryId_var id = valueBoxDef -> id();
	Identifier_var name = valueBoxDef -> name();
	result = orb -> create_boxed_value_tc(id, name, origType);
	break;
    }

    case dk_Value:
	// TODO
	break;

    case dk_ValueMember:
	// TODO
	break;
    }

    if(!is_nil(contained))
	seq.length(seq.length() - 1);

    assert(!is_nil(result));
    return result._retn();
}
*/
