// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>

#include <Test1.h>

#include <TestMacro.h>

#include <stdlib.h>
#include <errno.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

int
main(int argc, char* argv[], char*[])
{
    try
    {
	//
	// Create ORB
	//
	CORBA::ORB_var orb = CORBA::ORB_init(argc, argv);
	
	//
	// Get "Test1" object
	//
	const char* refFile = "Test1.ref";
	ifstream in; // Must use open(name), not ifstream in(name) (VC++ bug)
	in.open(refFile);
	if(in.fail())
	{
	    cerr << argv[0] << ": can't open `" << refFile << "': "
		 << strerror(errno) << endl;
	    return 1;
	}
	
	char s[1000];
	in >> s;
	
	CORBA::Object_var obj = orb -> string_to_object(s);
	assert(!CORBA::is_nil(obj));
	
	Test1_var t = Test1::_narrow(obj);
	assert(!CORBA::is_nil(t));

	cout << "Testing LongSeq... " << flush;
	{
	    Test1::LongSeq l;
	    l.length(1);
	    l[0] = 10;

	    TEST(t -> opLongSeq(l));
	}
	cout << "OK!" << endl;

	cout << "Testing LongSeq2... " << flush;
	{
	    Test1::LongSeq2 l;
	    l.length(1);
	    l[0] = 10;

	    TEST(t -> opLongSeq2(l));
	}
	cout << "OK!" << endl;

	cout << "Testing Any... " << flush;
	{
	    Test1::LongSeq l;
	    l.length(1);
	    l[0] = 10;

	    CORBA::Any a;
	    a <<= l;

	    TEST(t -> opLongSeqAny(a));
	    TEST(!t -> opLongSeq2Any(a));
	}

	{
	    Test1::LongSeq2 l;
	    l.length(1);
	    l[0] = 10;

	    CORBA::Any a;
	    a <<= l;

	    TEST(!t -> opLongSeqAny(a));
	    TEST(t -> opLongSeq2Any(a));
	}
	cout << "OK!" << endl;

	cout << "Testing LongSeqSeq... " << flush;
	{
	    Test1::LongSeqSeq l;
	    l.length(1);
	    l.length(1);
	    l[0].length(1);
	    l[0][0] = 10;

	    TEST(t -> opLongSeqSeq(l));
	}
	cout << "OK!" << endl;

	cout << "Testing LongSeqSeq2... " << flush;
	{
	    Test1::LongSeqSeq2 l;
	    l.length(1);
	    l.length(1);
	    l[0].length(1);
	    l[0][0] = 10;

	    TEST(t -> opLongSeqSeq2(l));
	}
	cout << "OK!" << endl;

	cout << "Testing Any... " << flush;
	{
	    Test1::LongSeqSeq l;
	    l.length(1);
	    l.length(1);
	    l[0].length(1);
	    l[0][0] = 10;

	    CORBA::Any a;
	    a <<= l;

	    TEST(t -> opLongSeqSeqAny(a));
	    TEST(!t -> opLongSeqSeq2Any(a));
	}

	{
	    Test1::LongSeqSeq2 l;
	    l.length(1);
	    l.length(1);
	    l[0].length(1);
	    l[0][0] = 10;

	    CORBA::Any a;
	    a <<= l;

	    TEST(!t -> opLongSeqSeqAny(a));
	    TEST(t -> opLongSeqSeq2Any(a));
	}
	cout << "OK!" << endl;
    }
    catch(CORBA::SystemException& ex)
    {
	OB::PrintException(ex);
	return 1;
    }

    return 0;
}

