// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>

#include <Test4.h>

#include <TestMacro.h>

#include <stdlib.h>
#include <errno.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

int
main(int argc, char* argv[], char*[])
{
    try
    {
	//
	// Create ORB
	//
	CORBA::ORB_var orb = CORBA::ORB_init(argc, argv);
	
	//
	// Get "Test4" object
	//
	const char* refFile = "Test4.ref";
	ifstream in; // Must use open(name), not ifstream in(name) (VC++ bug)
	in.open(refFile);
	if(in.fail())
	{
	    cerr << argv[0] << ": can't open `" << refFile << "': "
		 << strerror(errno) << endl;
	    return 1;
	}
	
	char s[1000];
	in >> s;
	
	CORBA::Object_var obj = orb -> string_to_object(s);
	assert(!CORBA::is_nil(obj));
	
	Test4_var t = Test4::_narrow(obj);
	assert(!CORBA::is_nil(t));

	cout << "Testing FixedArray... " << flush;
	{
	    Test4::FixedArray a;
	    for(int i = 0; i < 10; ++i)
		a[i] = i;

	    TEST(t -> opFixedArray(a));
	}
	cout << "OK!" << endl;

	cout << "Testing FixedArraySequence... " << flush;
	{
	    Test4::FixedArraySequence s;
	    s.length(1);
	    for(int i = 0; i < 10; ++i)
		s[0][i] = i;

	    TEST(t -> opFixedArraySequence(s));
	}
	cout << "OK!" << endl;

	cout << "Testing FixedArray2... " << flush;
	{
	    Test4::FixedArray2 a;

	    for(int i = 0; i < 2; ++i)
		for(int j = 0; j < 3; ++j)
		    for(int k = 0; k < 4; ++k)
			a[i][j][k] = i+j+k;
	    TEST(t -> opFixedArray2(a));
	}
	cout << "OK!" << endl;

	cout << "Testing FixedArraySequence2... " << flush;
	{
	    Test4::FixedArraySequence2 s;
	    s.length(1);
	    for(int i = 0; i < 2; ++i)
		for(int j = 0; j < 3; ++j)
		    for(int k = 0; k < 4; ++k)
			s[0][i][j][k] = i+j+k;

	    TEST(t -> opFixedArraySequence2(s));
	}
	cout << "OK!" << endl;

	cout << "Testing Any... " << flush;
	{
	    Test4::FixedArray a;
	    for(int i = 0; i < 10; ++i)
		a[i] = i;

	    CORBA::Any any;
	    any <<= a;
	    TEST(t -> opFixedArrayAny(any));
	    TEST(!t -> opFixedArraySequenceAny(any));
	}

	{
	    Test4::FixedArraySequence s;
	    s.length(1);
	    for(int i = 0; i < 10; ++i)
		s[0][i] = i;

	    CORBA::Any any;
	    any <<= s;

	    TEST(!t -> opFixedArrayAny(any));
	    TEST(t -> opFixedArraySequenceAny(any));
	}

	{
	    Test4::FixedArray2 a;

	    for(int i = 0; i < 2; ++i)
		for(int j = 0; j < 3; ++j)
		    for(int k = 0; k < 4; ++k)
			a[i][j][k] = i+j+k;

	    CORBA::Any any;
	    any <<= a;
	    TEST(!t -> opFixedArraySequence2Any(any));
	    TEST(t -> opFixedArray2Any(any));
	}
	cout << "OK!" << endl;

	cout << "Testing FixedArraySequence2... " << flush;
	{
	    Test4::FixedArraySequence2 s;
	    s.length(1);
	    for(int i = 0; i < 2; ++i)
		for(int j = 0; j < 3; ++j)
		    for(int k = 0; k < 4; ++k)
			s[0][i][j][k] = i+j+k;

	    CORBA::Any any;
	    any <<= s;
	    TEST(!t -> opFixedArray2Any(any));
	    TEST(t -> opFixedArraySequence2Any(any));
	}
	cout << "OK!" << endl;
    }
    catch(CORBA::SystemException& ex)
    {
	OB::PrintException(ex);
	return 1;
    }

    return 0;
}

