// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>

#include <Test_impl.h>
#include <TestDSI_impl.h>
#include <TestInterfaceTimeout_skel_tie.h>

#include <errno.h>

#ifdef HAVE_UNISTD_H
#   include <unistd.h>
#endif

#ifdef HAVE_IO_H
#   include <io.h>
#endif

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#ifndef HAVE_NO_EXPLICIT_TEMPLATES
template class POA_TestInterfaceTimeout_tie< TestInterfaceTimeout_impl >;
#else
#ifdef HAVE_PRAGMA_DEFINE
#pragma define(POA_TestInterfaceTimeout_tie< TestInterfaceTimeout_impl >)
#endif
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;

int
ServerRun(ORB_ptr orb, PortableServer::POA_ptr poa, int argc, char* argv[])
{
    //
    // Create implementation objects
    //
    TestInterfaceTimeout_impl* i = new TestInterfaceTimeout_impl(orb);
    TestInterfaceTimeout_var p = i -> _this();
    String_var impl = orb -> object_to_string(p);

    TestInterfaceExDSI_impl* iDSI = new TestInterfaceExDSI_impl(p, orb);
    PortableServer::ObjectId_var id = poa -> activate_object(iDSI);
    Object_var pDSI =
	poa -> create_reference_with_id(id, "IDL:TestInterfaceEx:1.0");
    String_var implDSI = orb -> object_to_string(pDSI);

    POA_TestInterfaceTimeout_tie< TestInterfaceTimeout_impl >* iTie = 
	new POA_TestInterfaceTimeout_tie< TestInterfaceTimeout_impl >
	(i, poa, FALSE);
    Object_var pTie = iTie -> _this();
    String_var implTie = orb -> object_to_string(pTie);

    const char* refFile = "TestInterface.ref";
    ofstream out(refFile);
    if(out.fail())
    {
	cerr << argv[0] << ": can't open `" << refFile << "': "
	     << strerror(errno) << endl;
	return 1;
    }
    
    out << impl << endl;
    out << implDSI << endl;
    out << implTie << endl;
    out.close();

    //
    // Change the POA manager state to active
    //
    PortableServer::POAManager_var manager = poa -> the_POAManager();
    manager -> activate();

    //
    // Run the ORB
    //
    orb -> run();

    //
    // Deactivate each of the servants
    //
    id = poa -> servant_to_id(i);
    poa -> deactivate_object(id);
    delete i;

    id = poa -> servant_to_id(iDSI);
    poa -> deactivate_object(id);
    delete iDSI;

    id = poa -> servant_to_id(iTie);
    poa -> deactivate_object(id);
    delete iTie;

#ifdef HAVE__UNLINK
    _unlink(refFile);
#else
    unlink(refFile);
#endif

    return 0;
}
