// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>

#include <stdlib.h>

#ifdef TEST_X11_REACTOR
#   include <X11/Intrinsic.h>
#   include <OB/X11.h>
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;

int
main(int argc, char* argv[], char*[])
{
    int status;

    try
    {
#ifdef TEST_X11_REACTOR
	XtAppContext appContext;
	Widget topLevel = XtAppInitialize(&appContext, "Test", 0, 0,
					  &argc, argv, 0, 0, 0);
	OBX11Init(appContext);
#endif

        //
        // Create ORB and POA
        //
        ORB_var orb = ORB_init(argc, argv);
	Object_var obj;
	try
	{
	    obj = orb -> resolve_initial_references("RootPOA");
	}
	catch(const ORB::InvalidName&)
	{
	    cerr << "Error: can't resolve `RootPOA'" << endl;
	    exit(1);
	}
	if(is_nil(obj))
	{
	    cerr << "Error: `RootPOA' is a nil object reference" << endl;
	    exit(1);
	}

	PortableServer::POA_var root = PortableServer::POA::_narrow(obj);

	if(is_nil(root))
	{
	    cerr << "Error: `RootPOA' is not a POA object reference" << endl;
	    exit(1);
	}

	//
	// Run tests
	//
	int ServerRun(ORB_ptr, PortableServer::POA_ptr, int, char*[]);
	status = ServerRun(orb, root, argc, argv);
    }
    catch(SystemException& ex)
    {
	OB::PrintException(ex);
	status = 1;
    }

    return status;
}
