// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <TestInterface.h>
#include <TestMacro.h>

using namespace CORBA;

void
TestDII(Object_ptr ti, ORB_ptr orb)
{
    ULong i, j;

    {
	Request_var request;

	request = ti -> _request("opVoid");
	request -> invoke();
	if(request -> env() -> exception() != 0)
	    request -> env() -> exception() -> _raise();
    }

    {
	Request_var request;

	Short ret;
	Short inOut;
	Short out;

	request = ti -> _request("_set_attrShort");
	request -> add_in_arg() <<= (Short)-32768;
	request -> invoke();
	if(request -> env() -> exception() != 0)
	    request -> env() -> exception() -> _raise();
	request = ti -> _request("_get_attrShort");
	request -> set_return_type(_tc_short);
	request -> invoke();
	if(request -> env() -> exception() != 0)
	    request -> env() -> exception() -> _raise();
	TEST(request -> return_value() >>= ret);
	TEST(ret == -32768);

	request = ti -> _request("_set_attrShort");
	request -> add_in_arg() <<= (Short)32767;
	request -> invoke();
	if(request -> env() -> exception() != 0)
	    request -> env() -> exception() -> _raise();
	request = ti -> _request("_get_attrShort");
	request -> set_return_type(_tc_short);
	request -> invoke();
	if(request -> env() -> exception() != 0)
	    request -> env() -> exception() -> _raise();
	TEST(request -> return_value() >>= ret);
	TEST(ret == 32767);

	request = ti -> _request("opShort");
	request -> add_in_arg() <<= (Short)10;
	Any& inOutAny = request -> add_inout_arg();
	inOutAny <<= (Short)20;
	Any& outAny = request -> add_out_arg();
	outAny.replace(_tc_short, 0);
	request -> set_return_type(_tc_short);
	request -> invoke();
	if(request -> env() -> exception() != 0)
	    request -> env() -> exception() -> _raise();
	TEST(inOutAny >>= inOut);
	TEST(outAny >>= out);
	TEST(request -> return_value() >>= ret);
	TEST(ret == 30);
        TEST(inOut == 30);
        TEST(out == 30);
    }

    {
	Request_var request;

	Double ret;
	Double inOut;
	Double out;
	    
	request = ti -> _request("_set_attrDouble");
	request -> add_in_arg() <<= 1.7976931348623157E+308;
	request -> invoke();
	if(request -> env() -> exception() != 0)
	    request -> env() -> exception() -> _raise();
	request = ti -> _request("_get_attrDouble");
	request -> set_return_type(_tc_double);
	request -> invoke();
	if(request -> env() -> exception() != 0)
	    request -> env() -> exception() -> _raise();
	TEST(request -> return_value() >>= ret);
	TEST(ret == 1.7976931348623157E+308);

	request = ti -> _request("_set_attrDouble");
	request -> add_in_arg() <<= 2.2250738585072014E-308;
	request -> invoke();
	if(request -> env() -> exception() != 0)
	    request -> env() -> exception() -> _raise();
	request = ti -> _request("_get_attrDouble");
	request -> set_return_type(_tc_double);
	request -> invoke();
	if(request -> env() -> exception() != 0)
	    request -> env() -> exception() -> _raise();
	TEST(request -> return_value() >>= ret);
#if !( defined(__alpha) && defined(__linux) )
	TEST(ret == 2.2250738585072014E-308);
#endif

	request = ti -> _request("opDouble");
	request -> add_in_arg() <<= 10.0;
	Any& inOutAny = request -> add_inout_arg();
	inOutAny <<= 20.0;
	Any& outAny = request -> add_out_arg();
	outAny.replace(_tc_double, 0);
	request -> set_return_type(_tc_double);
	request -> invoke();
	if(request -> env() -> exception() != 0)
	    request -> env() -> exception() -> _raise();
	TEST(inOutAny >>= inOut);
	TEST(outAny >>= out);
	TEST(request -> return_value() >>= ret);
	TEST(ret == 30);
        TEST(inOut == 30);
        TEST(out == 30);
    }

#ifndef WIN32
#if SIZEOF_LONG_DOUBLE != 0
    {
	try
	{
	    Request_var request;

	    LongDouble ret;
	    LongDouble inOut;
	    LongDouble out;

	    request = ti -> _request("_set_attrLongDouble");
	    request -> add_in_arg() <<= (LongDouble)1.7976931348623157E+308;
	    request -> invoke();
            if(request -> env() -> exception() != 0)
                request -> env() -> exception() -> _raise();
	    request = ti -> _request("_get_attrLongDouble");
	    request -> set_return_type(_tc_longdouble);
	    request -> invoke();
            if(request -> env() -> exception() != 0)
                request -> env() -> exception() -> _raise();
	    TEST(request -> return_value() >>= ret);
	    TEST(ret == (LongDouble)1.7976931348623157E+308);

	    request = ti -> _request("_set_attrLongDouble");
	    request -> add_in_arg() <<= (LongDouble)2.2250738585072014E-308;
	    request -> invoke();
            if(request -> env() -> exception() != 0)
                request -> env() -> exception() -> _raise();
	    request = ti -> _request("_get_attrLongDouble");
	    request -> set_return_type(_tc_longdouble);
	    request -> invoke();
            if(request -> env() -> exception() != 0)
                request -> env() -> exception() -> _raise();
	    TEST(request -> return_value() >>= ret);
#if !( defined(__alpha) && defined(__linux) )
	    TEST(ret == (LongDouble)2.2250738585072014E-308);
#endif

	    request = ti -> _request("opLongDouble");
	    request -> add_in_arg() <<= (LongDouble)10.0;
	    Any& inOutAny = request -> add_inout_arg();
	    inOutAny <<= (LongDouble)20.0;
	    Any& outAny = request -> add_out_arg();
	    outAny.replace(_tc_longdouble, 0);
	    request -> set_return_type(_tc_longdouble);
	    request -> invoke();
            if(request -> env() -> exception() != 0)
                request -> env() -> exception() -> _raise();
	    TEST(inOutAny >>= inOut);
	    TEST(outAny >>= out);
	    TEST(request -> return_value() >>= ret);
	    TEST(ret == 30);
	    TEST(inOut == 30);
	    TEST(out == 30);
	}
	catch(const MARSHAL& ex)
	{
	    //
	    // Ignore exceptions caused by unsupported data types
	    //
	    if(ex.minor() != OB::MinorLongDoubleNotSupported)
		throw;
	}
    }
#endif
#endif

/*    {
	Request_var request;

	TestInterface::TestFixed ret;
	TestInterface::TestFixed inOut;
	TestInterface::TestFixed out;
	    
	request = ti -> _request("_set_attrFixed");
	request -> add_in_arg() <<= Any::from_fixed(
	    TestInterface::TestFixed("1234567890.12345670"), 24, 8);
	request -> invoke();
	if(request -> env() -> exception() != 0)
	    request -> env() -> exception() -> _raise();
	request = ti -> _request("_get_attrFixed");
	request -> set_return_type(TestInterface::_tc_TestFixed);
	request -> invoke();
	if(request -> env() -> exception() != 0)
	    request -> env() -> exception() -> _raise();
	TEST(request -> return_value() >>= Any::to_fixed(ret, 24, 8));
	TEST(ret == TestInterface::TestFixed("1234567890.12345670"));

	request = ti -> _request("opFixed");
	request -> add_in_arg() <<= Any::from_fixed(
	    TestInterface::TestFixed("10.00000000"), 24, 8);
	Any& inOutAny = request -> add_inout_arg();
	inOutAny <<= Any::from_fixed(
	    TestInterface::TestFixed("20.00000000"), 24, 8);
	Any& outAny = request -> add_out_arg();
	outAny.replace(TestInterface::_tc_TestFixed, 0);
	request -> set_return_type(TestInterface::_tc_TestFixed);
	request -> invoke();
	if(request -> env() -> exception() != 0)
	    request -> env() -> exception() -> _raise();
	TEST(inOutAny >>= Any::to_fixed(inOut, 24, 8));
	TEST(outAny >>= Any::to_fixed(out, 24, 8));
	TEST(request -> return_value() >>= Any::to_fixed(ret, 24, 8));
	TEST(ret == TestInterface::TestFixed("30.00000000"));
        TEST(inOut == TestInterface::TestFixed("30.00000000"));
        TEST(out == TestInterface::TestFixed("30.00000000"));
	}*/

    {
	Request_var request;

	char* ret;
	char* inOut;
	char* out;

	request = ti -> _request("_set_attrString");
	request -> add_in_arg() <<= (const char*)"Hello";
	request -> invoke();
	if(request -> env() -> exception() != 0)
	    request -> env() -> exception() -> _raise();
	request = ti -> _request("_get_attrString");
	request -> set_return_type(_tc_string);
	request -> invoke();
	if(request -> env() -> exception() != 0)
	    request -> env() -> exception() -> _raise();
	TEST(request -> return_value() >>= ret);
	TEST(strcmp(ret, "Hello") == 0);

	request = ti -> _request("opString");
	request -> add_in_arg() <<= (const char*)"Hello, ";
	Any& inOutAny = request -> add_inout_arg();
	inOutAny <<= (const char*)"world!";
	Any& outAny = request -> add_out_arg();
	outAny.replace(_tc_string, 0);
	request -> set_return_type(_tc_string);
	request -> invoke();
	if(request -> env() -> exception() != 0)
	    request -> env() -> exception() -> _raise();
	TEST(inOutAny >>= inOut);
	TEST(outAny >>= out);
	TEST(request -> return_value() >>= ret);
	TEST(strcmp(ret, "Hello, world!") == 0);
	TEST(strcmp(out, "Hello, world!") == 0);
    }

    {
	Request_var request;

	TestInterface::StringSequence seq;
	seq.length(3);
	seq[0] = string_dup("!!!");
	seq[1] = string_dup("@@@");
	seq[2] = string_dup("###");

	TestInterface::StringSequence* ret;
	TestInterface::StringSequence* inOut;
	TestInterface::StringSequence* out;

	request = ti -> _request("_set_attrStringSequence");
	request -> add_in_arg() <<= seq;
	request -> invoke();
	if(request -> env() -> exception() != 0)
	    request -> env() -> exception() -> _raise();
	request = ti -> _request("_get_attrStringSequence");
	request -> set_return_type(TestInterface::_tc_StringSequence);
	request -> invoke();
	if(request -> env() -> exception() != 0)
	    request -> env() -> exception() -> _raise();
	TEST(request -> return_value() >>= ret);
	TEST(ret -> length() == 3);
	TEST(strcmp((*ret)[0], "!!!") == 0);
	TEST(strcmp((*ret)[1], "@@@") == 0);
	TEST(strcmp((*ret)[2], "###") == 0);

	inOut = new TestInterface::StringSequence;
	inOut -> length(2);
	(*inOut)[0] = string_dup("%");
	(*inOut)[1] = string_dup("^^");
	request = ti -> _request("opStringSequence");
	request -> add_in_arg() <<= seq;
	Any& inOutAny = request -> add_inout_arg();
	inOutAny <<= inOut;
	Any& outAny = request -> add_out_arg();
	outAny.replace(TestInterface::_tc_StringSequence, 0);
	request -> set_return_type(TestInterface::_tc_StringSequence);
	request -> invoke();
	if(request -> env() -> exception() != 0)
	    request -> env() -> exception() -> _raise();
	TEST(request -> return_value() >>= ret);
	TEST(inOutAny >>= inOut);
	TEST(outAny >>= out);
	TEST(ret -> length() == 5);
	TEST(strcmp((*ret)[0], "!!!") == 0);
	TEST(strcmp((*ret)[1], "@@@") == 0);
	TEST(strcmp((*ret)[2], "###") == 0);
	TEST(strcmp((*ret)[3], "%") == 0);
	TEST(strcmp((*ret)[4], "^^") == 0);
	TEST(inOut -> length() == 5);
	TEST(strcmp((*inOut)[0], "!!!") == 0);
	TEST(strcmp((*inOut)[1], "@@@") == 0);
	TEST(strcmp((*inOut)[2], "###") == 0);
	TEST(strcmp((*inOut)[3], "%") == 0);
	TEST(strcmp((*inOut)[4], "^^") == 0);
	TEST(out -> length() == 5);
	TEST(strcmp((*out)[0], "!!!") == 0);
	TEST(strcmp((*out)[1], "@@@") == 0);
	TEST(strcmp((*out)[2], "###") == 0);
	TEST(strcmp((*out)[3], "%") == 0);
	TEST(strcmp((*out)[4], "^^") == 0);
    }

    {
	Request_var request;

	Char ret;
	Char inOut;
	Char out;
	    
	request = ti -> _request("_set_attrChar");
	request -> add_in_arg() <<= Any::from_char('a');
	request -> invoke();
	if(request -> env() -> exception() != 0)
	    request -> env() -> exception() -> _raise();
	request = ti -> _request("_get_attrChar");
	request -> set_return_type(_tc_char);
	request -> send_deferred();
	request -> get_response();
	TEST(request -> return_value() >>= Any::to_char(ret));
	TEST(ret == 'a');

	request = ti -> _request("opChar");
	request -> add_in_arg() <<= Any::from_char('a');
	Any& inOutAny = request -> add_inout_arg();
	inOutAny <<= Any::from_char(1);
	Any& outAny = request -> add_out_arg();
	outAny.replace(_tc_char, 0);
	request -> set_return_type(_tc_char);
	request -> send_deferred();
	request -> get_response();
	TEST(inOutAny >>= Any::to_char(inOut));
	TEST(outAny >>= Any::to_char(out));
	TEST(request -> return_value() >>= Any::to_char(ret));
	TEST(ret == 'b');
        TEST(inOut == 'b');
        TEST(out == 'b');
    }

    {
	Request_var request;

	TestInterface::VariableArray ar =
	{ 
	    {
		string_dup("aa"),
		string_dup("bb"),
		string_dup("cc")
	    },
	    {
		string_dup("AA"),
		string_dup("BB"),
		string_dup("CC")
	    }
	};
	
	TestInterface::VariableArray_forany ret;
	TestInterface::VariableArray_forany inOut;
	TestInterface::VariableArray_forany out;

	request = ti -> _request("_set_attrVariableArray");
	request -> add_in_arg() <<= TestInterface::VariableArray_forany(ar);
	request -> invoke();
	if(request -> env() -> exception() != 0)
	    request -> env() -> exception() -> _raise();
	request = ti -> _request("_get_attrVariableArray");
	request -> set_return_type(TestInterface::_tc_VariableArray);
	request -> send_deferred();
	while(!request -> poll_response())
	    ;
	TEST(request -> return_value() >>= ret);
	for(i = 0 ; i < 2 ; i++)
	    for(j = 0 ; j < 3 ; j++)
	    {
		TEST(strcmp(ar[i][j], ret[i][j]) == 0);
	    }

	TestInterface::VariableArray_slice* ar2 =
	    TestInterface::VariableArray_alloc();
	for(i = 0 ; i < 2 ; i++)
	    for(j = 0 ; j < 3 ; j++)
		ar2[i][j] = string_dup("abc");
	request = ti -> _request("opVariableArray");
	request -> add_in_arg() <<= TestInterface::VariableArray_forany(ar);
	Any& inOutAny = request -> add_inout_arg();
	inOutAny <<= TestInterface::VariableArray_forany(ar2, true);
	Any& outAny = request -> add_out_arg();
	outAny.replace(TestInterface::_tc_VariableArray, 0);
	request -> set_return_type(TestInterface::_tc_VariableArray);
	request -> send_deferred();
	while(!request -> poll_response())
	    ;
	TEST(request -> return_value() >>= ret);
	TEST(inOutAny >>= inOut);
	TEST(outAny >>= out);
	for(i = 0 ; i < 2 ; i++)
	    for(j = 0 ; j < 3 ; j++)
	    {
		TEST(strcmp(ar[i][j], ret[i][j]) == 0);
		TEST(strcmp(ar[i][j], inOut[i][j]) == 0);
		TEST(strcmp(ar[i][j], out[i][j]) == 0);
	    }
    }

    {
	Request_var request;

	request = ti -> _request("_set_attrUShort");
	request -> add_in_arg() <<= (UShort)1234;
	request -> invoke();
	if(request -> env() -> exception() != 0)
	    request -> env() -> exception() -> _raise();
	
	ORB::RequestSeq requests;
	requests.length(5);
	
	for(i = 0 ; i < requests.length() ; i++)
	{
	    requests[i] = ti -> _request("_get_attrUShort");
	    requests[i] -> set_return_type(_tc_ushort);
	}
	
	orb -> send_multiple_requests_deferred(requests);
	
	for(i = 0 ; i < requests.length() ; i++)
	{
	    while(!orb -> poll_next_response())
		;
	    
	    orb -> get_next_response(request.out());

	    UShort ret;
	    TEST(request -> return_value() >>= ret);
	    TEST(ret == 1234);
	}
    }

    {
	Float ret;
	Float inOut;
	Float out;

	Request_var request1;
	request1 = ti -> _request("_set_attrFloat");
	request1 -> add_in_arg() <<= (Float)1;
	request1 -> send_deferred();

	Request_var request2;
	request2 = ti -> _request("_get_attrFloat");
	request2 -> set_return_type(_tc_float);
	request2 -> send_deferred();

	Request_var request3;
	request3 = ti -> _request("_set_attrFloat");
	request3 -> add_in_arg() <<= (Float)-1;
	request3 -> send_deferred();

	Request_var request4;
	request4 = ti -> _request("_get_attrFloat");
	request4 -> set_return_type(_tc_float);
	request4 -> send_deferred();

	Request_var request5;
	request5 = ti -> _request("opFloat");
	request5 -> add_in_arg() <<= (Float)10;
	Any& inOutAny = request5 -> add_inout_arg();
	inOutAny <<= (Float)20;
	Any& outAny = request5 -> add_out_arg();
	outAny.replace(_tc_float, 0);
	request5 -> set_return_type(_tc_float);
	request5 -> send_deferred();

	request5 -> get_response();
	TEST(inOutAny >>= inOut);
	TEST(outAny >>= out);
	TEST(request5 -> return_value() >>= ret);
	TEST(ret == 30);
        TEST(inOut == 30);
        TEST(out == 30);

	request2 -> get_response();
	TEST(request2 -> return_value() >>= ret);
	TEST(ret == 1);

	request4 -> get_response();
	TEST(request4 -> return_value() >>= ret);
	TEST(ret == -1);

	request1 -> get_response();
	request3 -> get_response();
    }
}
