// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <TestInterfaceEx.h>
#include <TestMacro.h>

using namespace CORBA;

void
TestDIIException(Object_ptr ti)
{
    Request_var request;

    {
	request = ti -> _request("opVoidEx");
	request -> invoke();
	Exception* ex = request -> env() -> exception();
	TEST(UNKNOWN::_downcast(ex));

	request = ti -> _request("opVoidEx");
	request -> exceptions() -> add(TestInterfaceEx::_tc_ExVoid);
	request -> invoke();
	ex = request -> env() -> exception();
	UnknownUserException* uex;
	TEST(uex = UnknownUserException::_downcast(ex));
	TestInterfaceEx::ExVoid* iex;
	TEST(uex -> exception() >>= iex);
    }

    {
	request = ti -> _request("opShortEx");
	request -> exceptions() -> add(TestInterfaceEx::_tc_ExShort);
	request -> add_in_arg() <<= (Short)10;
	Any& inOutAny = request -> add_inout_arg();
	inOutAny <<= (Short)20;
	Any& outAny = request -> add_out_arg();
	outAny.replace(_tc_short, 0);
	request -> set_return_type(_tc_short);
	request -> invoke();
	Exception* ex = request -> env() -> exception();
	UnknownUserException* uex;
	TEST(uex = UnknownUserException::_downcast(ex));
	TestInterfaceEx::ExShort* iex;
	TEST(uex -> exception() >>= iex);
	TEST(iex -> value == 30);
    }

    {
	request = ti -> _request("opDoubleEx");
	request -> exceptions() -> add(TestInterfaceEx::_tc_ExDouble);
	request -> add_in_arg() <<= 10.0;
	Any& inOutAny = request -> add_inout_arg();
	inOutAny <<= 20.0;
	Any& outAny = request -> add_out_arg();
	outAny.replace(_tc_double, 0);
	request -> set_return_type(_tc_double);
	request -> invoke();
	Exception* ex = request -> env() -> exception();
	UnknownUserException* uex;
	TEST(uex = UnknownUserException::_downcast(ex));
	TestInterfaceEx::ExDouble* iex;
	TEST(uex -> exception() >>= iex);
	TEST(iex -> value == 30);
    }

    {
	TestInterfaceEx::StringSequence in;
	in.length(3);
	in[0] = string_dup("!!!");
	in[1] = string_dup("@@@");
	in[2] = string_dup("###");
	TestInterfaceEx::StringSequence inOut;
	inOut.length(2);
	inOut[0] = string_dup("%");
	inOut[1] = string_dup("^^");
	request = ti -> _request("opStringSequenceEx");
	request -> exceptions() -> add(TestInterfaceEx::_tc_ExStringSequence);
	request -> add_in_arg() <<= in;
	Any& inOutAny = request -> add_inout_arg();
	inOutAny <<= inOut;
	Any& outAny = request -> add_out_arg();
	outAny.replace(TestInterfaceEx::_tc_StringSequence, 0);
	request -> set_return_type(TestInterfaceEx::_tc_StringSequence);
	request -> invoke();
	Exception* ex = request -> env() -> exception();
	UnknownUserException* uex;
	TEST(uex = UnknownUserException::_downcast(ex));
	TestInterfaceEx::ExStringSequence* iex;
	TEST(uex -> exception() >>= iex);
	TEST(iex -> value.length() == 5);
	TEST(strcmp(iex -> value[0], "!!!") == 0);
	TEST(strcmp(iex -> value[1], "@@@") == 0);
	TEST(strcmp(iex -> value[2], "###") == 0);
	TEST(strcmp(iex -> value[3], "%") == 0);
	TEST(strcmp(iex -> value[4], "^^") == 0);
    }
}
