// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <TestDSI_impl.h>

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;

TestInterfaceExDSI_impl::TestInterfaceExDSI_impl(TestInterfaceEx_ptr ti,
						 ORB_ptr orb)
    : ti_(TestInterfaceEx::_duplicate(ti)),
      orb_(ORB::_duplicate(orb))
{
}

RepositoryId
TestInterfaceExDSI_impl::_primary_interface(const PortableServer::ObjectId& oid,
					    PortableServer::POA_ptr poa)
{
    return string_dup("IDL:TestInterfaceEx:1.0");
}

Boolean
TestInterfaceExDSI_impl::_is_a(const char* name)
    throw(SystemException)
{
    if(strcmp(name, "IDL:TestInterface:1.0") == 0 ||
       strcmp(name, "IDL:TestInterfaceEx:1.0") == 0)
    {
	return true;
    }
    else
    {
#ifdef HAVE_VCPLUSPLUS_BUGS
	return ServantBase::_is_a(name);
#else
	return PortableServer::ServantBase::_is_a(name);
#endif
    }
}

void
TestInterfaceExDSI_impl::invoke(ServerRequest_ptr request) throw()
{
    String_var name = request -> op_name();

    bool ex;
    if(strlen(name) > 2 &&
       strcmp((const char*)name + strlen(name) - 2, "Ex") == 0)
    {
	name[strlen(name) - 2] = '\0';
	ex = true;
    }
    else
	ex = false;

    if(strcmp(name, "opVoid") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	request -> params(list);

	if(ex)
	{
	    any = new Any;
	    *any <<= new TestInterfaceEx::ExVoid();
	    request -> exception(any);
	}

	return;
    }

    if(strcmp(name, "opContext") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(_tc_string, 0);
	request -> params(list);
	Context_ptr ctx = request -> ctx();

	any = list -> item(0) -> value();
	char* pattern;
	*any >>= pattern;

	TestInterface::StringSequence* ret = ti_ -> opContext(pattern, ctx);

	any = new Any;
	*any <<= ret;
	request -> result(any);

	return;
    }

    if(strcmp(name, "_get_attrShort") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	request -> params(list);

	Short ret = ti_ -> attrShort();

	any = new Any;
	*any <<= ret;
	request -> result(any);

	return;
    }

    if(strcmp(name, "_set_attrShort") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value() -> replace(_tc_short, 0);
	request -> params(list);

	any = list -> item(0) -> value();
	Short arg;
	*any >>= arg;

	ti_ -> attrShort(arg);

	return;
    }

    if(strcmp(name, "opShort") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(_tc_short, 0);
	list -> add(ARG_INOUT) -> value()
	    -> replace(_tc_short, 0);
	list -> add(ARG_OUT);
	request -> params(list);

	any = list -> item(0) -> value();
	Short arg0;
	*any >>= arg0;

	any = list -> item(1) -> value();
	Short arg1;
	*any >>= arg1;

	Short arg2;

	Short ret = ti_ -> opShort(arg0, arg1, arg2);

	any = list -> item(1) -> value();
	*any <<= arg1;
	
	any = list -> item(2) -> value();
	*any <<= arg2;
	
	if(ex)
	{
	    any = new Any;
	    *any <<= new TestInterfaceEx::ExShort(ret);
	    request -> exception(any);
	}
	else
	{
	    any = new Any;
	    *any <<= ret;
	    request -> result(any);
	}

	return;
    }

    if(strcmp(name, "_get_attrUShort") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	request -> params(list);

	UShort ret = ti_ -> attrUShort();

	any = new Any;
	*any <<= ret;
	request -> result(any);

	return;
    }

    if(strcmp(name, "_set_attrUShort") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(_tc_ushort, 0);
	request -> params(list);

	any = list -> item(0) -> value();
	UShort arg;
	*any >>= arg;

	ti_ -> attrUShort(arg);

	return;
    }

    if(strcmp(name, "opUShort") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(_tc_ushort, 0);
	list -> add(ARG_INOUT) -> value()
	    -> replace(_tc_ushort, 0);
	list -> add(ARG_OUT);
	request -> params(list);

	any = list -> item(0) -> value();
	UShort arg0;
	*any >>= arg0;

	any = list -> item(1) -> value();
	UShort arg1;
	*any >>= arg1;

	UShort arg2;

	UShort ret = ti_ -> opUShort(arg0, arg1, arg2);

	any = list -> item(1) -> value();
	*any <<= arg1;

	any = list -> item(2) -> value();
	*any <<= arg2;

	if(ex)
	{
	    any = new Any;
	    *any <<= new TestInterfaceEx::ExUShort(ret);
	    request -> exception(any);
	}
	else
	{
	    any = new Any;
	    *any <<= ret;
	    request -> result(any);
	}

	return;
    }

    if(strcmp(name, "_get_attrLong") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	request -> params(list);

	Long ret = ti_ -> attrLong();

	any = new Any;
	*any <<= ret;
	request -> result(any);

	return;
    }

    if(strcmp(name, "_set_attrLong") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(_tc_long, 0);
	request -> params(list);

	any = list -> item(0) -> value();
	Long arg;
	*any >>= arg;

	ti_ -> attrLong(arg);

	return;
    }

    if(strcmp(name, "opLong") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(_tc_long, 0);
	list -> add(ARG_INOUT) -> value()
	    -> replace(_tc_long, 0);
	list -> add(ARG_OUT);
	request -> params(list);

	any = list -> item(0) -> value();
	Long arg0;
	*any >>= arg0;

	any = list -> item(1) -> value();
	Long arg1;
	*any >>= arg1;

	Long arg2;

	Long ret = ti_ -> opLong(arg0, arg1, arg2);

	any = list -> item(1) -> value();
	*any <<= arg1;

	any = list -> item(2) -> value();
	*any <<= arg2;

	if(ex)
	{
	    any = new Any;
	    *any <<= new TestInterfaceEx::ExLong(ret);
	    request -> exception(any);
	}
	else
	{
	    any = new Any;
	    *any <<= ret;
	    request -> result(any);
	}

	return;
    }

    if(strcmp(name, "_get_attrULong") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	request -> params(list);

	ULong ret = ti_ -> attrULong();

	any = new Any;
	*any <<= ret;
	request -> result(any);

	return;
    }

    if(strcmp(name, "_set_attrULong") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(_tc_ulong, 0);
	request -> params(list);

	any = list -> item(0) -> value();
	ULong arg;
	*any >>= arg;

	ti_ -> attrULong(arg);

	return;
    }

    if(strcmp(name, "opULong") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(_tc_ulong, 0);
	list -> add(ARG_INOUT) -> value()
	    -> replace(_tc_ulong, 0);
	list -> add(ARG_OUT);
	request -> params(list);

	any = list -> item(0) -> value();
	ULong arg0;
	*any >>= arg0;

	any = list -> item(1) -> value();
	ULong arg1;
	*any >>= arg1;

	ULong arg2;

	ULong ret = ti_ -> opULong(arg0, arg1, arg2);

	any = list -> item(1) -> value();
	*any <<= arg1;

	any = list -> item(2) -> value();
	*any <<= arg2;

	if(ex)
	{
	    any = new Any;
	    *any <<= new TestInterfaceEx::ExULong(ret);
	    request -> exception(any);
	}
	else
	{
	    any = new Any;
	    *any <<= ret;
	    request -> result(any);
	}

	return;
    }

    if(strcmp(name, "_get_attrLongLong") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	request -> params(list);

	LongLong ret = ti_ -> attrLongLong();

	any = new Any;
	*any <<= ret;
	request -> result(any);

	return;
    }

    if(strcmp(name, "_set_attrLongLong") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(_tc_longlong, 0);
	request -> params(list);

	any = list -> item(0) -> value();
	LongLong arg;
	*any >>= arg;

	ti_ -> attrLongLong(arg);

	return;
    }

    if(strcmp(name, "opLongLong") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(_tc_longlong, 0);
	list -> add(ARG_INOUT) -> value()
	    -> replace(_tc_longlong, 0);
	list -> add(ARG_OUT);
	request -> params(list);

	any = list -> item(0) -> value();
	LongLong arg0;
	*any >>= arg0;

	any = list -> item(1) -> value();
	LongLong arg1;
	*any >>= arg1;

	LongLong arg2;

	LongLong ret = ti_ -> opLongLong(arg0, arg1, arg2);

	any = list -> item(1) -> value();
	*any <<= arg1;

	any = list -> item(2) -> value();
	*any <<= arg2;

	if(ex)
	{
	    any = new Any;
	    *any <<= new TestInterfaceEx::ExLongLong(ret);
	    request -> exception(any);
	}
	else
	{
	    any = new Any;
	    *any <<= ret;
	    request -> result(any);
	}

	return;
    }

    if(strcmp(name, "_get_attrULongLong") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	request -> params(list);

	ULongLong ret = ti_ -> attrULongLong();

	any = new Any;
	*any <<= ret;
	request -> result(any);

	return;
    }

    if(strcmp(name, "_set_attrULongLong") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(_tc_ulonglong, 0);
	request -> params(list);

	any = list -> item(0) -> value();
	ULongLong arg;
	*any >>= arg;

	ti_ -> attrULongLong(arg);

	return;
    }

    if(strcmp(name, "opULongLong") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(_tc_ulonglong, 0);
	list -> add(ARG_INOUT) -> value()
	    -> replace(_tc_ulonglong, 0);
	list -> add(ARG_OUT);
	request -> params(list);

	any = list -> item(0) -> value();
	ULongLong arg0;
	*any >>= arg0;

	any = list -> item(1) -> value();
	ULongLong arg1;
	*any >>= arg1;

	ULongLong arg2;

	ULongLong ret = ti_ -> opULongLong(arg0, arg1, arg2);

	any = list -> item(1) -> value();
	*any <<= arg1;

	any = list -> item(2) -> value();
	*any <<= arg2;

	if(ex)
	{
	    any = new Any;
	    *any <<= new TestInterfaceEx::ExULongLong(ret);
	    request -> exception(any);
	}
	else
	{
	    any = new Any;
	    *any <<= ret;
	    request -> result(any);
	}

	return;
    }

    if(strcmp(name, "_get_attrFloat") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	request -> params(list);

	Float ret = ti_ -> attrFloat();

	any = new Any;
	*any <<= ret;
	request -> result(any);

	return;
    }

    if(strcmp(name, "_set_attrFloat") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(_tc_float, 0);
	request -> params(list);

	any = list -> item(0) -> value();
	Float arg;
	*any >>= arg;

	ti_ -> attrFloat(arg);

	return;
    }

    if(strcmp(name, "opFloat") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(_tc_float, 0);
	list -> add(ARG_INOUT) -> value()
	    -> replace(_tc_float, 0);
	list -> add(ARG_OUT);
	request -> params(list);

	any = list -> item(0) -> value();
	Float arg0;
	*any >>= arg0;

	any = list -> item(1) -> value();
	Float arg1;
	*any >>= arg1;

	Float arg2;

	Float ret = ti_ -> opFloat(arg0, arg1, arg2);

	any = list -> item(1) -> value();
	*any <<= arg1;

	any = list -> item(2) -> value();
	*any <<= arg2;

	if(ex)
	{
	    any = new Any;
	    *any <<= new TestInterfaceEx::ExFloat(ret);
	    request -> exception(any);
	}
	else
	{
	    any = new Any;
	    *any <<= ret;
	    request -> result(any);
	}

	return;
    }

    if(strcmp(name, "_get_attrDouble") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	request -> params(list);

	Double ret = ti_ -> attrDouble();

	any = new Any;
	*any <<= ret;
	request -> result(any);

	return;
    }

    if(strcmp(name, "_set_attrDouble") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(_tc_double, 0);
	request -> params(list);

	any = list -> item(0) -> value();
	Double arg;
	*any >>= arg;

	ti_ -> attrDouble(arg);

	return;
    }

    if(strcmp(name, "opDouble") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(_tc_double, 0);
	list -> add(ARG_INOUT) -> value()
	    -> replace(_tc_double, 0);
	list -> add(ARG_OUT);
	request -> params(list);

	any = list -> item(0) -> value();
	Double arg0;
	*any >>= arg0;

	any = list -> item(1) -> value();
	Double arg1;
	*any >>= arg1;

	Double arg2;

	Double ret = ti_ -> opDouble(arg0, arg1, arg2);

	any = list -> item(1) -> value();
	*any <<= arg1;

	any = list -> item(2) -> value();
	*any <<= arg2;

	if(ex)
	{
	    any = new Any;
	    *any <<= new TestInterfaceEx::ExDouble(ret);
	    request -> exception(any);
	}
	else
	{
	    any = new Any;
	    *any <<= ret;
	    request -> result(any);
	}

	return;
    }

    if(strcmp(name, "_get_attrLongDouble") == 0)
    {
#if SIZEOF_LONG_DOUBLE == 0
	throw MARSHAL(OB::MinorLongDoubleNotSupported, COMPLETED_NO);
#endif

	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	request -> params(list);

	LongDouble ret = ti_ -> attrLongDouble();

	any = new Any;
	*any <<= ret;
	request -> result(any);

	return;
    }

    if(strcmp(name, "_set_attrLongDouble") == 0)
    {
#if SIZEOF_LONG_DOUBLE == 0
	throw MARSHAL(OB::MinorLongDoubleNotSupported, COMPLETED_NO);
#endif
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(_tc_longdouble, 0);
	request -> params(list);

	any = list -> item(0) -> value();
	LongDouble arg;
	*any >>= arg;

	ti_ -> attrLongDouble(arg);

	return;
    }

    if(strcmp(name, "opLongDouble") == 0)
    {
#if SIZEOF_LONG_DOUBLE == 0
	throw MARSHAL(OB::MinorLongDoubleNotSupported, COMPLETED_NO);
#endif

	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(_tc_longdouble, 0);
	list -> add(ARG_INOUT) -> value()
	    -> replace(_tc_longdouble, 0);
	list -> add(ARG_OUT);
	request -> params(list);

	any = list -> item(0) -> value();
	LongDouble arg0;
	*any >>= arg0;

	any = list -> item(1) -> value();
	LongDouble arg1;
	*any >>= arg1;

	LongDouble arg2;

	LongDouble ret = ti_ -> opLongDouble(arg0, arg1, arg2);

	any = list -> item(1) -> value();
	*any <<= arg1;

	any = list -> item(2) -> value();
	*any <<= arg2;

	if(ex)
	{
	    any = new Any;
	    *any <<= new TestInterfaceEx::ExLongDouble(ret);
	    request -> exception(any);
	}
	else
	{
	    any = new Any;
	    *any <<= ret;
	    request -> result(any);
	}

	return;
    }

    if(strcmp(name, "_get_attrFixed") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	request -> params(list);

	TestInterface::TestFixed ret = ti_ -> attrFixed();

	any = new Any;
	*any <<= Any::from_fixed(ret, 24, 8);
	request -> result(any);

	return;
    }

    if(strcmp(name, "_set_attrFixed") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(_tc_fixed.fixed(24, 8), 0);
	request -> params(list);

	any = list -> item(0) -> value();
	TestInterface::TestFixed arg;
	*any >>= Any::to_fixed(arg, 24, 8);
	ti_ -> attrFixed(arg);

	return;
    }

    if(strcmp(name, "opFixed") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value() -> replace(_tc_fixed.fixed(24, 8), 0);
	list -> add(ARG_INOUT) -> value()
	    -> replace(_tc_fixed.fixed(24, 8), 0);
	list -> add(ARG_OUT);
	request -> params(list);

	any = list -> item(0) -> value();
	TestInterface::TestFixed arg0;
	*any >>= Any::to_fixed(arg0, 24, 8);

	any = list -> item(1) -> value();
	TestInterface::TestFixed arg1;
	*any >>= Any::to_fixed(arg1, 24, 8);

	TestInterface::TestFixed arg2;

	TestInterface::TestFixed ret = ti_ -> opFixed(arg0, arg1, arg2);

	any = list -> item(1) -> value();
	*any <<= Any::from_fixed(arg1, 24, 8);

	any = list -> item(2) -> value();
	*any <<= Any::from_fixed(arg2, 24, 8);

	if(ex)
	{
	    any = new Any;
	    *any <<= new TestInterfaceEx::ExFixed(ret);
	    request -> exception(any);
	}
	else
	{
	    any = new Any;
	    *any <<= Any::from_fixed(ret, 24, 8);
	    request -> result(any);
	}

	return;
    }

    if(strcmp(name, "_get_attrBoolean") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	request -> params(list);

	Boolean ret = ti_ -> attrBoolean();

	any = new Any;
	*any <<= Any::from_boolean(ret);
	request -> result(any);

	return;
    }

    if(strcmp(name, "_set_attrBoolean") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value() -> replace(_tc_boolean, 0);
	request -> params(list);

	any = list -> item(0) -> value();
	Boolean arg;
	*any >>= Any::to_boolean(arg);

	ti_ -> attrBoolean(arg);

	return;
    }

    if(strcmp(name, "opBoolean") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(_tc_boolean, 0);
	list -> add(ARG_INOUT) -> value()
	    -> replace(_tc_boolean, 0);
	list -> add(ARG_OUT);
	request -> params(list);

	any = list -> item(0) -> value();
	Boolean arg0;
	*any >>= Any::to_boolean(arg0);

	any = list -> item(1) -> value();
	Boolean arg1;
	*any >>= Any::to_boolean(arg1);

	Boolean arg2;

	Boolean ret = ti_ -> opBoolean(arg0, arg1, arg2);

	any = list -> item(1) -> value();
	*any <<= Any::from_boolean(arg1);

	any = list -> item(2) -> value();
	*any <<= Any::from_boolean(arg2);

	if(ex)
	{
	    any = new Any;
	    *any <<= new TestInterfaceEx::ExBoolean(ret);
	    request -> exception(any);
	}
	else
	{
	    any = new Any;
	    *any <<= Any::from_boolean(ret);
	    request -> result(any);
	}

	return;
    }

    if(strcmp(name, "_get_attrChar") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	request -> params(list);

	Char ret = ti_ -> attrChar();

	any = new Any;
	*any <<= Any::from_char(ret);
	request -> result(any);

	return;
    }

    if(strcmp(name, "_set_attrChar") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value() -> replace(_tc_char, 0);
	request -> params(list);

	any = list -> item(0) -> value();
	Char arg;
	*any >>= Any::to_char(arg);

	ti_ -> attrChar(arg);

	return;
    }

    if(strcmp(name, "opChar") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(_tc_char, 0);
	list -> add(ARG_INOUT) -> value()
	    -> replace(_tc_char, 0);
	list -> add(ARG_OUT);
	request -> params(list);

	any = list -> item(0) -> value();
	Char arg0;
	*any >>= Any::to_char(arg0);

	any = list -> item(1) -> value();
	Char arg1;
	*any >>= Any::to_char(arg1);

	Char arg2;

	Char ret = ti_ -> opChar(arg0, arg1, arg2);

	any = list -> item(1) -> value();
	*any <<= Any::from_char(arg1);

	any = list -> item(2) -> value();
	*any <<= Any::from_char(arg2);

	if(ex)
	{
	    any = new Any;
	    *any <<= new TestInterfaceEx::ExChar(ret);
	    request -> exception(any);
	}
	else
	{
	    any = new Any;
	    *any <<= Any::from_char(ret);
	    request -> result(any);
	}

	return;
    }

    if(strcmp(name, "_get_attrWChar") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	request -> params(list);

	WChar ret = ti_ -> attrWChar();

	any = new Any;
	*any <<= Any::from_wchar(ret);
	request -> result(any);

	return;
    }

    if(strcmp(name, "_set_attrWChar") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value() -> replace(_tc_wchar, 0);
	request -> params(list);

	any = list -> item(0) -> value();
	WChar arg;
	*any >>= Any::to_wchar(arg);

	ti_ -> attrWChar(arg);

	return;
    }

    if(strcmp(name, "opWChar") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(_tc_wchar, 0);
	list -> add(ARG_INOUT) -> value()
	    -> replace(_tc_wchar, 0);
	list -> add(ARG_OUT);
	request -> params(list);

	any = list -> item(0) -> value();
	WChar arg0;
	*any >>= Any::to_wchar(arg0);

	any = list -> item(1) -> value();
	WChar arg1;
	*any >>= Any::to_wchar(arg1);

	WChar arg2;

	WChar ret = ti_ -> opWChar(arg0, arg1, arg2);

	any = list -> item(1) -> value();
	*any <<= Any::from_wchar(arg1);

	any = list -> item(2) -> value();
	*any <<= Any::from_wchar(arg2);

	if(ex)
	{
	    any = new Any;
	    *any <<= new TestInterfaceEx::ExWChar(ret);
	    request -> exception(any);
	}
	else
	{
	    any = new Any;
	    *any <<= Any::from_wchar(ret);
	    request -> result(any);
	}

	return;
    }

    if(strcmp(name, "_get_attrOctet") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	request -> params(list);

	Octet ret = ti_ -> attrOctet();

	any = new Any;
	*any <<= Any::from_octet(ret);
	request -> result(any);

	return;
    }

    if(strcmp(name, "_set_attrOctet") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value() -> replace(_tc_octet, 0);
	request -> params(list);

	any = list -> item(0) -> value();
	Octet arg;
	*any >>= Any::to_octet(arg);

	ti_ -> attrOctet(arg);

	return;
    }

    if(strcmp(name, "opOctet") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(_tc_octet, 0);
	list -> add(ARG_INOUT) -> value()
	    -> replace(_tc_octet, 0);
	list -> add(ARG_OUT);
	request -> params(list);

	any = list -> item(0) -> value();
	Octet arg0;
	*any >>= Any::to_octet(arg0);

	any = list -> item(1) -> value();
	Octet arg1;
	*any >>= Any::to_octet(arg1);

	Octet arg2;

	Octet ret = ti_ -> opOctet(arg0, arg1, arg2);

	any = list -> item(1) -> value();
	*any <<= Any::from_octet(arg1);

	any = list -> item(2) -> value();
	*any <<= Any::from_octet(arg2);

	if(ex)
	{
	    any = new Any;
	    *any <<= new TestInterfaceEx::ExOctet(ret);
	    request -> exception(any);
	}
	else
	{
	    any = new Any;
	    *any <<= Any::from_octet(ret);
	    request -> result(any);
	}

	return;
    }

    if(strcmp(name, "_get_attrString") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	request -> params(list);

	String_var ret = ti_ -> attrString();

	any = new Any;
	*any <<= ret;
	request -> result(any);

	return;
    }

    if(strcmp(name, "_set_attrString") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value() -> replace(_tc_string, 0);
	request -> params(list);

	any = list -> item(0) -> value();
	char* arg;
	*any >>= arg;

	ti_ -> attrString(arg);

	return;
    }

    if(strcmp(name, "opString") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(_tc_string, 0);
	list -> add(ARG_INOUT) -> value()
	    -> replace(_tc_string, 0);
	list -> add(ARG_OUT);
	request -> params(list);

	any = list -> item(0) -> value();
	char* arg0;
	*any >>= arg0;

	any = list -> item(1) -> value();
	char* tmpArg1;
	*any >>= tmpArg1;

	String_var arg1 = string_dup(tmpArg1);
	String_var arg2;

	String_var ret = ti_ -> opString(arg0, arg1.inout(), arg2.out());

	any = list -> item(1) -> value();
	*any <<= arg1;

	any = list -> item(2) -> value();
	*any <<= arg2;

	if(ex)
	{
	    any = new Any;
	    *any <<= new TestInterfaceEx::ExString(ret);
	    request -> exception(any);
	}
	else
	{
	    any = new Any;
	    *any <<= ret;
	    request -> result(any);
	}

	return;
    }

    if(strcmp(name, "_get_attrWString") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	request -> params(list);

	WString_var ret = ti_ -> attrWString();

	any = new Any;
	*any <<= ret;
	request -> result(any);

	return;
    }

    if(strcmp(name, "_set_attrWString") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value() -> replace(_tc_wstring, 0);
	request -> params(list);

	any = list -> item(0) -> value();
	wchar_t* arg;
	*any >>= arg;

	ti_ -> attrWString(arg);

	return;
    }

    if(strcmp(name, "opWString") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(_tc_wstring, 0);
	list -> add(ARG_INOUT) -> value()
	    -> replace(_tc_wstring, 0);
	list -> add(ARG_OUT);
	request -> params(list);

	any = list -> item(0) -> value();
	wchar_t* arg0;
	*any >>= arg0;

	any = list -> item(1) -> value();
	wchar_t* tmpArg1;
	*any >>= tmpArg1;

	WString_var arg1 = wstring_dup(tmpArg1);
	WString_var arg2;

	WString_var ret = ti_ -> opWString(arg0, arg1.inout(),
						 arg2.out());

	any = list -> item(1) -> value();
	*any <<= arg1;

	any = list -> item(2) -> value();
	*any <<= arg2;

	if(ex)
	{
	    any = new Any;
	    *any <<= new TestInterfaceEx::ExWString(ret);
	    request -> exception(any);
	}
	else
	{
	    any = new Any;
	    *any <<= ret;
	    request -> result(any);
	}

	return;
    }

    if(strcmp(name, "_get_attrAny") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	request -> params(list);

	Any* ret = ti_ -> attrAny();

	any = new Any;
	*any <<= ret;
	request -> result(any);

	return;
    }

    if(strcmp(name, "_set_attrAny") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value() -> replace(_tc_any, 0);
	request -> params(list);

	any = list -> item(0) -> value();
	Any* arg;
	*any >>= arg;

	ti_ -> attrAny(*arg);

	return;
    }

    if(strcmp(name, "opAny") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(_tc_any, 0);
	list -> add(ARG_INOUT) -> value()
	    -> replace(_tc_any, 0);
	list -> add(ARG_OUT);
	request -> params(list);

	any = list -> item(0) -> value();
	Any* arg0;
	*any >>= arg0;

	any = list -> item(1) -> value();
	Any* arg1;
	*any >>= arg1;
	
	Any* arg2;

	Any* ret = ti_ -> opAny(*arg0, *arg1, arg2);

	any = list -> item(2) -> value();
	*any <<= arg2;

	if(ex)
	{
	    any = new Any;
	    *any <<= new TestInterfaceEx::ExAny(*ret);
	    delete ret;
	    request -> exception(any);
	}
	else
	{
	    any = new Any;
	    *any <<= ret;
	    request -> result(any);
	}

	return;
    }

    if(strcmp(name, "_get_attrTestEnum") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	request -> params(list);

	TestInterface::TestEnum ret = ti_ -> attrTestEnum();

	any = new Any;
	*any <<= ret;
	request -> result(any);

	return;
    }

    if(strcmp(name, "_set_attrTestEnum") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(TestInterface::_tc_TestEnum, 0);
	request -> params(list);

	any = list -> item(0) -> value();
	TestInterface::TestEnum arg;
	*any >>= arg;

	ti_ -> attrTestEnum(arg);

	return;
    }

    if(strcmp(name, "opTestEnum") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(TestInterface::_tc_TestEnum, 0);
	list -> add(ARG_INOUT) -> value()
	    -> replace(TestInterface::_tc_TestEnum, 0);
	list -> add(ARG_OUT);
	request -> params(list);

	any = list -> item(0) -> value();
	TestInterface::TestEnum arg0;
	*any >>= arg0;

	any = list -> item(1) -> value();
	TestInterface::TestEnum arg1;
	*any >>= arg1;

	TestInterface::TestEnum arg2;

	TestInterface::TestEnum ret = ti_ -> opTestEnum(arg0, arg1, arg2);

	any = list -> item(1) -> value();
	*any <<= arg1;

	any = list -> item(2) -> value();
	*any <<= arg2;

	if(ex)
	{
	    any = new Any;
	    *any <<= new TestInterfaceEx::ExTestEnum(ret);
	    request -> exception(any);
	}
	else
	{
	    any = new Any;
	    *any <<= ret;
	    request -> result(any);
	}

	return;
    }

    if(strcmp(name, "_get_attrTestInterface") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any = aTestInterface;
	request -> result(any);

	return;
    }

    if(strcmp(name, "_set_attrTestInterface") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(_tc_TestInterface, 0);
	request -> params(list);

	any = list -> item(0) -> value();
	aTestInterface = *any;

	return;
    }

    if(strcmp(name, "opTestInterface") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(_tc_TestInterface, 0);
	list -> add(ARG_INOUT) -> value()
	    -> replace(_tc_TestInterface, 0);
	list -> add(ARG_OUT);
	request -> params(list);

	any = list -> item(0) -> value();
	aTestInterface = *any;

	any = list -> item(1) -> value();
	*any = aTestInterface;

	any = list -> item(2) -> value();
	*any = aTestInterface;

	any = new Any;
	*any = aTestInterface;
	request -> result(any);

	return;
    }

    if(strcmp(name, "_get_attrFixedStruct") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	request -> params(list);

	TestInterface::FixedStruct ret = ti_ -> attrFixedStruct();

	any = new Any;
	*any <<= ret;
	request -> result(any);

	return;
    }

    if(strcmp(name, "_set_attrFixedStruct") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(TestInterface::_tc_FixedStruct, 0);
	request -> params(list);

	any = list -> item(0) -> value();
	TestInterface::FixedStruct* arg;
	*any >>= arg;

	ti_ -> attrFixedStruct(*arg);

	return;
    }

    if(strcmp(name, "opFixedStruct") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(TestInterface::_tc_FixedStruct, 0);
	list -> add(ARG_INOUT) -> value()
	    -> replace(TestInterface::_tc_FixedStruct, 0);
	list -> add(ARG_OUT);
	request -> params(list);

	any = list -> item(0) -> value();
	TestInterface::FixedStruct* arg0;
	*any >>= arg0;

	any = list -> item(1) -> value();
	TestInterface::FixedStruct* arg1;
	*any >>= arg1;

	TestInterface::FixedStruct arg2;

	TestInterface::FixedStruct ret =
	    ti_ -> opFixedStruct(*arg0, *arg1, arg2);


	any = list -> item(2) -> value();
	*any <<= arg2;

	if(ex)
	{
	    any = new Any;
	    *any <<= new TestInterfaceEx::ExFixedStruct(ret);
	    request -> exception(any);
	}
	else
	{
	    any = new Any;
	    *any <<= ret;
	    request -> result(any);
	}

	return;
    }

    if(strcmp(name, "_get_attrVariableStruct") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	request -> params(list);

	TestInterface::VariableStruct* ret = ti_ -> attrVariableStruct();

	any = new Any;
	*any <<= ret;
	request -> result(any);

	return;
    }

    if(strcmp(name, "_set_attrVariableStruct") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(TestInterface::_tc_VariableStruct, 0);
	request -> params(list);

	any = list -> item(0) -> value();
	TestInterface::VariableStruct* arg;
	*any >>= arg;

	ti_ -> attrVariableStruct(*arg);

	return;
    }

    if(strcmp(name, "opVariableStruct") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(TestInterface::_tc_VariableStruct, 0);
	list -> add(ARG_INOUT) -> value()
	    -> replace(TestInterface::_tc_VariableStruct, 0);
	list -> add(ARG_OUT);
	request -> params(list);

	any = list -> item(0) -> value();
	TestInterface::VariableStruct* arg0;
	*any >>= arg0;

	any = list -> item(1) -> value();
	TestInterface::VariableStruct* arg1;
	*any >>= arg1;

	TestInterface::VariableStruct* arg2;

	TestInterface::VariableStruct* ret =
	    ti_ -> opVariableStruct(*arg0, *arg1, arg2);


	any = list -> item(2) -> value();
	*any <<= arg2;

	if(ex)
	{
	    any = new Any;
	    *any <<= new TestInterfaceEx::ExVariableStruct(*ret);
	    delete ret;
	    request -> exception(any);
	}
	else
	{
	    any = new Any;
	    *any <<= ret;
	    request -> result(any);
	}

	return;
    }

    if(strcmp(name, "_get_attrFixedUnion") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	request -> params(list);

	TestInterface::FixedUnion ret = ti_ -> attrFixedUnion();

	any = new Any;
	*any <<= ret;
	request -> result(any);

	return;
    }

    if(strcmp(name, "_set_attrFixedUnion") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(TestInterface::_tc_FixedUnion, 0);
	request -> params(list);

	any = list -> item(0) -> value();
	TestInterface::FixedUnion* arg;
	*any >>= arg;

	ti_ -> attrFixedUnion(*arg);

	return;
    }

    if(strcmp(name, "opFixedUnion") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(TestInterface::_tc_FixedUnion, 0);
	list -> add(ARG_INOUT) -> value()
	    -> replace(TestInterface::_tc_FixedUnion, 0);
	list -> add(ARG_OUT);
	request -> params(list);

	any = list -> item(0) -> value();
	TestInterface::FixedUnion* arg0;
	*any >>= arg0;

	any = list -> item(1) -> value();
	TestInterface::FixedUnion* arg1;
	*any >>= arg1;

	TestInterface::FixedUnion arg2;

	TestInterface::FixedUnion ret =
	    ti_ -> opFixedUnion(*arg0, *arg1, arg2);

	any = list -> item(2) -> value();
	*any <<= arg2;

	if(ex)
	{
	    any = new Any;
	    *any <<= new TestInterfaceEx::ExFixedUnion(ret);
	    request -> exception(any);
	}
	else
	{
	    any = new Any;
	    *any <<= ret;
	    request -> result(any);
	}

	return;
    }

    if(strcmp(name, "_get_attrVariableUnion") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any = aVariableUnion;
	request -> result(any);

	return;
    }

    if(strcmp(name, "_set_attrVariableUnion") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(TestInterface::_tc_VariableUnion, 0);
	request -> params(list);

	any = list -> item(0) -> value();
	aVariableUnion = *any;

	return;
    }

    if(strcmp(name, "opVariableUnion") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(TestInterface::_tc_VariableUnion, 0);
	list -> add(ARG_INOUT) -> value()
	    -> replace(TestInterface::_tc_VariableUnion, 0);
	list -> add(ARG_OUT);
	request -> params(list);

	any = list -> item(0) -> value();
	aVariableUnion = *any;

	any = list -> item(1) -> value();
	*any = aVariableUnion;

	any = list -> item(2) -> value();
	*any = aVariableUnion;

	any = new Any;
	*any = aVariableUnion;
	request -> result(any);

	return;
    }

    if(strcmp(name, "_get_attrStringSequence") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	request -> params(list);

	TestInterface::StringSequence* ret = ti_ -> attrStringSequence();

	any = new Any;
	*any <<= ret;
	request -> result(any);

	return;
    }

    if(strcmp(name, "_set_attrStringSequence") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(TestInterface::_tc_StringSequence, 0);
	request -> params(list);

	any = list -> item(0) -> value();
	TestInterface::StringSequence* arg;
	*any >>= arg;

	ti_ -> attrStringSequence(*arg);

	return;
    }

    if(strcmp(name, "opStringSequence") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(TestInterface::_tc_StringSequence, 0);
	list -> add(ARG_INOUT) -> value()
	    -> replace(TestInterface::_tc_StringSequence, 0);
	list -> add(ARG_OUT);
	request -> params(list);

	any = list -> item(0) -> value();
	TestInterface::StringSequence* arg0;
	*any >>= arg0;

	any = list -> item(1) -> value();
	TestInterface::StringSequence* arg1;
	*any >>= arg1;

	TestInterface::StringSequence* arg2;

	TestInterface::StringSequence* ret =
	    ti_ -> opStringSequence(*arg0, *arg1, arg2);

	any = list -> item(2) -> value();
	*any <<= arg2;

	if(ex)
	{
	    any = new Any;
	    *any <<= new TestInterfaceEx::ExStringSequence(*ret);
	    delete ret;
	    request -> exception(any);
	}
	else
	{
	    any = new Any;
	    *any <<= ret;
	    request -> result(any);
	}

	return;
    }

    if(strcmp(name, "_get_attrFixedArray") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	request -> params(list);

	TestInterface::FixedArray_slice* ret = ti_ -> attrFixedArray();

	any = new Any;
	*any <<= TestInterface::FixedArray_forany(ret, true);
	request -> result(any);

	return;
    }

    if(strcmp(name, "_set_attrFixedArray") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(TestInterface::_tc_FixedArray, 0);
	request -> params(list);

	any = list -> item(0) -> value();
	TestInterface::FixedArray_forany arg;
	*any >>= arg;

	ti_ -> attrFixedArray(arg);

	return;
    }

    if(strcmp(name, "opFixedArray") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(TestInterface::_tc_FixedArray, 0);
	list -> add(ARG_INOUT) -> value()
	    -> replace(TestInterface::_tc_FixedArray, 0);
	list -> add(ARG_OUT);
	request -> params(list);

	any = list -> item(0) -> value();
	TestInterface::FixedArray_forany arg0;
	*any >>= arg0;

	any = list -> item(1) -> value();
	TestInterface::FixedArray_forany arg1;
	*any >>= arg1;

	TestInterface::FixedArray arg2;

	TestInterface::FixedArray_slice* ret =
	    ti_ -> opFixedArray(arg0, arg1.inout(), arg2);
	
	any = list -> item(2) -> value();
	*any <<= TestInterface::FixedArray_forany(arg2, false);

	if(ex)
	{
	    any = new Any;
#ifdef HAVE_VCPLUSPLUS_BUGS
	    *any <<= new TestInterfaceEx::ExFixedArray(
		(const TestInterface::FixedArray_slice*)ret);
#else
	    *any <<= new TestInterfaceEx::ExFixedArray(ret);
#endif
	    TestInterface::FixedArray_free(ret);
	    request -> exception(any);
	}
	else
	{
	    any = new Any;
	    *any <<= TestInterface::FixedArray_forany(ret, true);
	    request -> result(any);
	}

	return;
    }

    if(strcmp(name, "_get_attrVariableArray") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	request -> params(list);

	TestInterface::VariableArray_slice* ret = ti_ -> attrVariableArray();

	any = new Any;
	*any <<= TestInterface::VariableArray_forany(ret, true);
	request -> result(any);

	return;
    }

    if(strcmp(name, "_set_attrVariableArray") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(TestInterface::_tc_VariableArray, 0);
	request -> params(list);

	any = list -> item(0) -> value();
	TestInterface::VariableArray_forany arg;
	*any >>= arg;

	ti_ -> attrVariableArray(arg);

	return;
    }

    if(strcmp(name, "opVariableArray") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(TestInterface::_tc_VariableArray, 0);
	list -> add(ARG_INOUT) -> value()
	    -> replace(TestInterface::_tc_VariableArray, 0);
	list -> add(ARG_OUT);
	request -> params(list);

	any = list -> item(0) -> value();
	TestInterface::VariableArray_forany arg0;
	*any >>= arg0;

	any = list -> item(1) -> value();
	TestInterface::VariableArray_forany arg1;
	*any >>= arg1;

	TestInterface::VariableArray_slice* arg2;

	TestInterface::VariableArray_slice* ret =
	    ti_ -> opVariableArray(arg0, arg1.inout(), arg2);
	
	any = list -> item(2) -> value();
	*any <<= TestInterface::VariableArray_forany(arg2, true);

	if(ex)
	{
	    any = new Any;
#ifdef HAVE_VCPLUSPLUS_BUGS
	    *any <<= new TestInterfaceEx::ExVariableArray(
		(const TestInterface::VariableArray_slice*)ret);
#else
	    *any <<= new TestInterfaceEx::ExVariableArray(ret);
#endif
	    TestInterface::VariableArray_free(ret);
	    request -> exception(any);
	}
	else
	{
	    any = new Any;
	    *any <<= TestInterface::VariableArray_forany(ret, true);
	    request -> result(any);
	}

	return;
    }

    if(strcmp(name, "_get_attrFixedArraySequence") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	request -> params(list);

	TestInterface::FixedArraySequence* ret =
	    ti_ -> attrFixedArraySequence();

	any = new Any;
	*any <<= ret;
	request -> result(any);

	return;
    }

    if(strcmp(name, "_set_attrFixedArraySequence") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(TestInterface::_tc_FixedArraySequence, 0);
	request -> params(list);

	any = list -> item(0) -> value();
	TestInterface::FixedArraySequence* arg;
	*any >>= arg;

	ti_ -> attrFixedArraySequence(*arg);

	return;
    }

    if(strcmp(name, "opFixedArraySequence") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(TestInterface::_tc_FixedArraySequence, 0);
	list -> add(ARG_INOUT) -> value()
	    -> replace(TestInterface::_tc_FixedArraySequence, 0);
	list -> add(ARG_OUT);
	request -> params(list);

	any = list -> item(0) -> value();
	TestInterface::FixedArraySequence* arg0;
	*any >>= arg0;

	any = list -> item(1) -> value();
	TestInterface::FixedArraySequence* arg1;
	*any >>= arg1;

	TestInterface::FixedArraySequence* arg2;

	TestInterface::FixedArraySequence* ret =
	    ti_ -> opFixedArraySequence(*arg0, *arg1, arg2);

	any = list -> item(2) -> value();
	*any <<= arg2;

	if(ex)
	{
	    any = new Any;
	    *any <<= new TestInterfaceEx::ExFixedArraySequence(*ret);
	    delete ret;
	    request -> exception(any);
	}
	else
	{
	    any = new Any;
	    *any <<= ret;
	    request -> result(any);
	}

	return;
    }

    if(strcmp(name, "_get_attrVariableArraySequence") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	request -> params(list);

	TestInterface::VariableArraySequence* ret =
	    ti_ -> attrVariableArraySequence();

	any = new Any;
	*any <<= ret;
	request -> result(any);

	return;
    }

    if(strcmp(name, "_set_attrVariableArraySequence") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(TestInterface::_tc_VariableArraySequence, 0);
	request -> params(list);

	any = list -> item(0) -> value();
	TestInterface::VariableArraySequence* arg;
	*any >>= arg;

	ti_ -> attrVariableArraySequence(*arg);

	return;
    }

    if(strcmp(name, "opVariableArraySequence") == 0)
    {
	NVList_ptr list;
	Any* any;

	orb_ -> create_list(0, list);
	list -> add(ARG_IN) -> value()
	    -> replace(TestInterface::_tc_VariableArraySequence, 0);
	list -> add(ARG_INOUT) -> value()
	    -> replace(TestInterface::_tc_VariableArraySequence, 0);
	list -> add(ARG_OUT);
	request -> params(list);

	any = list -> item(0) -> value();
	TestInterface::VariableArraySequence* arg0;
	*any >>= arg0;

	any = list -> item(1) -> value();
	TestInterface::VariableArraySequence* arg1;
	*any >>= arg1;

	TestInterface::VariableArraySequence* arg2;

	TestInterface::VariableArraySequence* ret =
	    ti_ -> opVariableArraySequence(*arg0, *arg1, arg2);

	any = list -> item(2) -> value();
	*any <<= arg2;

	if(ex)
	{
	    any = new Any;
	    *any <<= new TestInterfaceEx::ExVariableArraySequence(*ret);
	    delete ret;
	    request -> exception(any);
	}
	else
	{
	    any = new Any;
	    *any <<= ret;
	    request -> result(any);
	}

	return;
    }

    if(strcmp(name, "op_UNKNOWN_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new UNKNOWN(1, COMPLETED_YES);
	request -> exception(any);

	return;
    }

    if(strcmp(name, "op_BAD_PARAM_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new BAD_PARAM(2, COMPLETED_NO);
	request -> exception(any);

	return;
    }

    if(strcmp(name, "op_NO_MEMORY_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new NO_MEMORY(3, COMPLETED_MAYBE);
	request -> exception(any);

	return;
    }

    if(strcmp(name, "op_IMP_LIMIT_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new IMP_LIMIT(4, COMPLETED_YES);
	request -> exception(any);

	return;
    }

    if(strcmp(name, "op_COMM_FAILURE_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new COMM_FAILURE(5, COMPLETED_NO);
	request -> exception(any);

	return;
    }

    if(strcmp(name, "op_INV_OBJREF_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new INV_OBJREF(6, COMPLETED_MAYBE);
	request -> exception(any);

	return;
    }

    if(strcmp(name, "op_NO_PERMISSION_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new NO_PERMISSION(7, COMPLETED_YES);
	request -> exception(any);

	return;
    }

    if(strcmp(name, "op_INTERNAL_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new INTERNAL(8, COMPLETED_NO);
	request -> exception(any);

	return;
    }

    if(strcmp(name, "op_MARSHAL_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new MARSHAL(9, COMPLETED_MAYBE);
	request -> exception(any);

	return;
    }

    if(strcmp(name, "op_INITIALIZE_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new INITIALIZE(10, COMPLETED_YES);
	request -> exception(any);

	return;
    }

    if(strcmp(name, "op_NO_IMPLEMENT_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new NO_IMPLEMENT(11, COMPLETED_NO);
	request -> exception(any);

	return;
    }

    if(strcmp(name, "op_BAD_TYPECODE_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new BAD_TYPECODE(12, COMPLETED_MAYBE);
	request -> exception(any);

	return;
    }

    if(strcmp(name, "op_BAD_OPERATION_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new BAD_OPERATION(13, COMPLETED_YES);
	request -> exception(any);

	return;
    }

    if(strcmp(name, "op_NO_RESOURCES_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new NO_RESOURCES(14, COMPLETED_NO);
	request -> exception(any);

	return;
    }

    if(strcmp(name, "op_NO_RESPONSE_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new NO_RESPONSE(15, COMPLETED_MAYBE);
	request -> exception(any);

	return;
    }

    if(strcmp(name, "op_PERSIST_STORE_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new PERSIST_STORE(16, COMPLETED_YES);
	request -> exception(any);

	return;
    }

    if(strcmp(name, "op_BAD_INV_ORDER_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new BAD_INV_ORDER(17, COMPLETED_NO);
	request -> exception(any);

	return;
    }

    if(strcmp(name, "op_TRANSIENT_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new TRANSIENT(18, COMPLETED_MAYBE);
	request -> exception(any);

	return;
    }

    if(strcmp(name, "op_FREE_MEM_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new FREE_MEM(19, COMPLETED_YES);
	request -> exception(any);

	return;
    }

    if(strcmp(name, "op_INV_IDENT_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new INV_IDENT(20, COMPLETED_NO);
	request -> exception(any);

	return;
    }

    if(strcmp(name, "op_INV_FLAG_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new INV_FLAG(21, COMPLETED_MAYBE);
	request -> exception(any);

	return;
    }

    if(strcmp(name, "op_INTF_REPOS_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new INTF_REPOS(22, COMPLETED_YES);
	request -> exception(any);

	return;
    }

    if(strcmp(name, "op_BAD_CONTEXT_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new BAD_CONTEXT(23, COMPLETED_NO);
	request -> exception(any);

	return;
    }

    if(strcmp(name, "op_OBJ_ADAPTER_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new OBJ_ADAPTER(24, COMPLETED_MAYBE);
	request -> exception(any);

	return;
    }

    if(strcmp(name, "op_DATA_CONVERSION_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new DATA_CONVERSION(25, COMPLETED_YES);
	request -> exception(any);

	return;
    }

    if(strcmp(name, "op_OBJECT_NOT_EXIST_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new OBJECT_NOT_EXIST(26, COMPLETED_NO);
	request -> exception(any);

	return;
    }

    if(strcmp(name, "op_TRANSACTION_REQUIRED_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new TRANSACTION_REQUIRED(27, COMPLETED_NO);
	request -> exception(any);

	return;
    }

    if(strcmp(name, "op_TRANSACTION_ROLLEDBACK_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new TRANSACTION_ROLLEDBACK(28, COMPLETED_NO);
	request -> exception(any);

	return;
    }

    if(strcmp(name, "op_INVALID_TRANSACTION_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new INVALID_TRANSACTION(29, COMPLETED_NO);
	request -> exception(any);

	return;
    }

    if(strcmp(name, "op_INV_POLICY_") == 0)
    {
	assert(ex);

	NVList_ptr list;
	Any* any;
	
	orb_ -> create_list(0, list);
	request -> params(list);

	any = new Any;
	*any <<= new INV_POLICY(30, COMPLETED_NO);
	request -> exception(any);

	return;
    }

    cerr << "DSI implementation: unknown operation: " << name << endl;

    NVList_ptr list;
    Any* any;
    
    orb_ -> create_list(0, list);
    request -> params(list);
    
    any = new Any;
    *any <<= new BAD_OPERATION();
    request -> exception(any);
}

void
TestInterfaceExDSI_impl::releaseInternal()
{
    //
    // Avoid cyclic object dependencies
    //
    Any dummy;
    aTestInterface = dummy;
    aVariableUnion = dummy;
}
