// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <Test_impl.h>

#include <limits.h>

#ifdef HAVE_UNISTD_H
#   include <unistd.h>
#endif

using namespace CORBA;

TestInterface_impl::TestInterface_impl(ORB_ptr orb)
    : orb_(ORB::_duplicate(orb))
{
}

void
TestInterface_impl::opVoid()
    throw(SystemException)
{
}

TestInterface::StringSequence*
TestInterface_impl::opContext(const char* pattern, Context_ptr ctx)
    throw(SystemException)
{
    NVList_var values;
    ctx -> get_values("", 0, pattern, values.out());

    ULong len = values -> count();
    TestInterface::StringSequence* result =
        new TestInterface::StringSequence(len * 2);
    result -> length(len * 2);
    for(ULong i = 0 ; i < len ; i++)
    {
	char* s;
	*(values -> item(i) -> value()) >>= s;
	(*result)[i * 2] = values -> item(i) -> name();
	(*result)[i * 2 + 1] = string_dup(s);
    }

    return result;
}

TestInterfaceEx_impl::TestInterfaceEx_impl(ORB_ptr orb)
    : TestInterface_impl(orb)
{
}

void
TestInterfaceEx_impl::opVoidEx()
    throw(TestInterfaceEx::ExVoid,
	  SystemException)
{
    throw TestInterfaceEx::ExVoid();
}

Short
TestInterface_impl::attrShort()
    throw(SystemException)
{
    return aShort;
}

void
TestInterface_impl::attrShort(Short val)
    throw(SystemException)
{
    aShort = val;
}

Short
TestInterface_impl::opShort(Short a0,
			    Short& a1,
			    Short& a2)
    throw(SystemException)
{
    aShort = a0 + a1;
    a1 = a2 = aShort;
    return aShort;
}

Short
TestInterfaceEx_impl::opShortEx(Short a0,
				Short& a1,
				Short&)
    throw(TestInterfaceEx::ExShort,
	  SystemException)
{
    aShort = a0 + a1;
    throw TestInterfaceEx::ExShort(aShort);
    return 0; // Some compilers need this
}

Long
TestInterface_impl::attrLong()
    throw(SystemException)
{
    return aLong;
}

void
TestInterface_impl::attrLong(Long val)
    throw(SystemException)
{
    aLong = val;
}

Long
TestInterface_impl::opLong(Long a0,
			   Long& a1,
			   Long& a2)
    throw(SystemException)
{
    aLong = a0 + a1;
    a1 = a2 = aLong;
    return aLong;
}

Long
TestInterfaceEx_impl::opLongEx(Long a0,
			       Long& a1,
			       Long&)
    throw(TestInterfaceEx::ExLong,
	  SystemException)
{
    aLong = a0 + a1;
    throw TestInterfaceEx::ExLong(aLong);
    return 0; // Some compilers need this
}

LongLong
TestInterface_impl::attrLongLong()
    throw(SystemException)
{
    return aLongLong;
}

void
TestInterface_impl::attrLongLong(LongLong val)
    throw(SystemException)
{
    aLongLong = val;
}

LongLong
TestInterface_impl::opLongLong(LongLong a0,
			       LongLong& a1,
			       LongLong& a2)
    throw(SystemException)
{
    aLongLong = a0 + a1;
    a1 = a2 = aLongLong;
    return aLongLong;
}

LongLong
TestInterfaceEx_impl::opLongLongEx(LongLong a0,
				   LongLong& a1,
				   LongLong&)
    throw(TestInterfaceEx::ExLongLong,
	  SystemException)
{
    aLongLong = a0 + a1;
    throw TestInterfaceEx::ExLongLong(aLongLong);
    return 0; // Some compilers need this
}

UShort
TestInterface_impl::attrUShort()
    throw(SystemException)
{
    return aUShort;
}

void
TestInterface_impl::attrUShort(UShort val)
    throw(SystemException)
{
    aUShort = val;
}

UShort
TestInterface_impl::opUShort(UShort a0,
			     UShort& a1,
			     UShort& a2)
    throw(SystemException)
{
    aUShort = a0 + a1;
    a1 = a2 = aUShort;
    return aUShort;
}

UShort
TestInterfaceEx_impl::opUShortEx(UShort a0,
				 UShort& a1,
				 UShort&)
    throw(TestInterfaceEx::ExUShort,
	  SystemException)
{
    aUShort = a0 + a1;
    throw TestInterfaceEx::ExUShort(aUShort);
    return 0; // Some compilers need this
}

ULong
TestInterface_impl::attrULong()
    throw(SystemException)
{
    return aULong;
}

void
TestInterface_impl::attrULong(ULong val)
    throw(SystemException)
{
    aULong = val;
}

ULong
TestInterface_impl::opULong(ULong a0,
			    ULong& a1,
			    ULong& a2)
    throw(SystemException)
{
    aULong = a0 + a1;
    a1 = a2 = aULong;
    return aULong;
}

ULong
TestInterfaceEx_impl::opULongEx(ULong a0,
				ULong& a1,
				ULong&)
    throw(TestInterfaceEx::ExULong,
	  SystemException)
{
    aULong = a0 + a1;
    throw TestInterfaceEx::ExULong(aULong);
    return 0; // Some compilers need this
}

ULongLong
TestInterface_impl::attrULongLong()
    throw(SystemException)
{
    return aULongLong;
}

void
TestInterface_impl::attrULongLong(ULongLong val)
    throw(SystemException)
{
    aULongLong = val;
}

ULongLong
TestInterface_impl::opULongLong(ULongLong a0,
				ULongLong& a1,
				ULongLong& a2)
    throw(SystemException)
{
    aULongLong = a0 + a1;
    a1 = a2 = aULongLong;
    return aULongLong;
}

ULongLong
TestInterfaceEx_impl::opULongLongEx(ULongLong a0,
				    ULongLong& a1,
				    ULongLong&)
    throw(TestInterfaceEx::ExULongLong,
	  SystemException)
{
    aULongLong = a0 + a1;
    throw TestInterfaceEx::ExULongLong(aULongLong);
    return 0; // Some compilers need this
}

Float
TestInterface_impl::attrFloat()
    throw(SystemException)
{
    return aFloat;
}

void
TestInterface_impl::attrFloat(Float val)
    throw(SystemException)
{
    aFloat = val;
}

Float
TestInterface_impl::opFloat(Float a0,
			    Float& a1,
			    Float& a2)
    throw(SystemException)
{
    aFloat = a0 + a1;
    a1 = a2 = aFloat;
    return aFloat;
}

Float
TestInterfaceEx_impl::opFloatEx(Float a0,
				Float& a1,
				Float&)
    throw(TestInterfaceEx::ExFloat,
	  SystemException)
{
    aFloat = a0 + a1;
    throw TestInterfaceEx::ExFloat(aFloat);
    return 0; // Some compilers need this
}

Double
TestInterface_impl::attrDouble()
    throw(SystemException)
{
    return aDouble;
}

void
TestInterface_impl::attrDouble(Double val)
    throw(SystemException)
{
    aDouble = val;
}

Double
TestInterface_impl::opDouble(Double a0,
			     Double& a1,
			     Double& a2)
    throw(SystemException)
{
    aDouble = a0 + a1;
    a1 = a2 = aDouble;
    return aDouble;
}

Double
TestInterfaceEx_impl::opDoubleEx(Double a0,
				 Double& a1,
				 Double&)
    throw(TestInterfaceEx::ExDouble,
	  SystemException)
{
    aDouble = a0 + a1;
    throw TestInterfaceEx::ExDouble(aDouble);
    return 0; // Some compilers need this
}

LongDouble
TestInterface_impl::attrLongDouble()
    throw(SystemException)
{
    return aLongDouble;
}

void
TestInterface_impl::attrLongDouble(LongDouble val)
    throw(SystemException)
{
    aLongDouble = val;
}

LongDouble
TestInterface_impl::opLongDouble(LongDouble a0,
				 LongDouble& a1,
				 LongDouble& a2)
    throw(SystemException)
{
#if SIZEOF_LONG_DOUBLE != 0
    aLongDouble = a0 + a1;
    a1 = a2 = aLongDouble;
#endif
    return aLongDouble;
}

LongDouble
TestInterfaceEx_impl::opLongDoubleEx(LongDouble a0,
				     LongDouble& a1,
				     LongDouble&)
    throw(TestInterfaceEx::ExLongDouble,
	  SystemException)
{
#if SIZEOF_LONG_DOUBLE != 0
    aLongDouble = a0 + a1;
    throw TestInterfaceEx::ExLongDouble(aLongDouble);
    return 0; // Some compilers need this
#else
    return aLongDouble; // Some compilers need this
#endif
}

TestInterface::TestFixed
TestInterface_impl::attrFixed()
    throw(SystemException)
{
    return aFixed;
}

void
TestInterface_impl::attrFixed(TestInterface::TestFixed val)
    throw(SystemException)
{
    aFixed = val;
}

TestInterface::TestFixed
TestInterface_impl::opFixed(TestInterface::TestFixed a0,
			    TestInterface::TestFixed& a1,
			    TestInterface::TestFixed_out a2)
    throw(SystemException)
{
    aFixed = a0 + a1;
    a1 = a2 = aFixed;
    return aFixed;
}

TestInterface::TestFixed
TestInterfaceEx_impl::opFixedEx(TestInterface::TestFixed a0,
				TestInterface::TestFixed& a1,
				TestInterface::TestFixed_out)
    throw(TestInterfaceEx::ExFixed,
	  SystemException)
{
    aFixed = a0 + a1;
    throw TestInterfaceEx::ExFixed(aDouble);
    return 0L; // Some compilers need this
}

Boolean
TestInterface_impl::attrBoolean()
    throw(SystemException)
{
    return aBoolean;
}

void
TestInterface_impl::attrBoolean(Boolean val)
    throw(SystemException)
{
    aBoolean = val;
}

Boolean
TestInterface_impl::opBoolean(Boolean a0,
			      Boolean& a1,
			      Boolean& a2)
    throw(SystemException)
{
    aBoolean = a0 && a1;
    a1 = a2 = aBoolean;
    return aBoolean;
}

Boolean
TestInterfaceEx_impl::opBooleanEx(Boolean a0,
				  Boolean& a1,
				  Boolean&)
    throw(TestInterfaceEx::ExBoolean,
	  SystemException)
{
    aBoolean = a0 && a1;
    throw TestInterfaceEx::ExBoolean(aBoolean);
    return 0; // Some compilers need this
}

Char
TestInterface_impl::attrChar()
    throw(SystemException)
{
    return aChar;
}

void
TestInterface_impl::attrChar(Char val)
    throw(SystemException)
{
    aChar = val;
}

Char
TestInterface_impl::opChar(Char a0,
			   Char& a1,
			   Char& a2)
    throw(SystemException)
{
    aChar = a0 + a1;
    a1 = a2 = aChar;
    return aChar;
}

Char
TestInterfaceEx_impl::opCharEx(Char a0,
			       Char& a1,
			       Char&)
    throw(TestInterfaceEx::ExChar,
	  SystemException)
{
    aChar = a0 + a1;
    throw TestInterfaceEx::ExChar(aChar);
    return 0; // Some compilers need this
}

WChar
TestInterface_impl::attrWChar()
    throw(SystemException)
{
    return aWChar;
}

void
TestInterface_impl::attrWChar(WChar val)
    throw(SystemException)
{
    aWChar = val;
}

WChar
TestInterface_impl::opWChar(WChar a0,
			    WChar& a1,
			    WChar& a2)
    throw(SystemException)
{
    aWChar = a0 + a1;
    a1 = a2 = aWChar;
    return aWChar;
}

WChar
TestInterfaceEx_impl::opWCharEx(WChar a0,
				WChar& a1,
				WChar&)
    throw(TestInterfaceEx::ExWChar,
	  SystemException)
{
    aWChar = a0 + a1;
    throw TestInterfaceEx::ExWChar(aWChar);
    return 0; // Some compilers need this
}

Octet
TestInterface_impl::attrOctet()
    throw(SystemException)
{
    return aOctet;
}

void
TestInterface_impl::attrOctet(Octet val)
    throw(SystemException)
{
    aOctet = val;
}

Octet
TestInterface_impl::opOctet(Octet a0,
			    Octet& a1,
			    Octet& a2)
    throw(SystemException)
{
    aOctet = a0 + a1;
    a1 = a2 = aOctet;
    return aOctet;
}

Octet
TestInterfaceEx_impl::opOctetEx(Octet a0,
				Octet& a1,
				Octet&)
    throw(TestInterfaceEx::ExOctet,
	  SystemException)
{
    aOctet = a0 + a1;
    throw TestInterfaceEx::ExOctet(aOctet);
    return 0; // Some compilers need this
}

char*
TestInterface_impl::attrString()
    throw(SystemException)
{
    return string_dup(aString);
}

void
TestInterface_impl::attrString(const char* val)
    throw(SystemException)
{
    aString = val;
}

char*
TestInterface_impl::opString(const char* a0,
			     char*& a1,
			     String_out a2)
    throw(SystemException)
{
    aString = string_alloc(strlen(a0) + strlen(a1));
    strcpy(aString.inout(), a0);
    strcat(aString.inout(), a1);
    string_free(a1);
    a1 = string_dup(aString);
    a2 = string_dup(aString);
    return string_dup(aString);
}

char*
TestInterfaceEx_impl::opStringEx(const char* a0,
				 char*& a1,
				 String_out)
    throw(TestInterfaceEx::ExString,
	  SystemException)
{
    aString = string_alloc(strlen(a0) + strlen(a1));
    strcpy(aString.inout(), a0);
    strcat(aString.inout(), a1);
    throw TestInterfaceEx::ExString(aString);
    return 0; // Some compilers need this
}

wchar_t*
TestInterface_impl::attrWString()
    throw(SystemException)
{
    return wstring_dup(aWString);
}

void
TestInterface_impl::attrWString(const wchar_t* val)
    throw(SystemException)
{
    aWString = val;
}

wchar_t*
TestInterface_impl::opWString(const wchar_t* a0,
			      wchar_t*& a1,
			      WString_out a2)
    throw(SystemException)
{
    aWString = wstring_alloc(OB::wcslen(a0) + OB::wcslen(a1));
    OB::wcscpy(aWString.inout(), a0);
    OB::wcscat(aWString.inout(), a1);
    wstring_free(a1);
    a1 = wstring_dup(aWString);
    a2 = wstring_dup(aWString);
    return wstring_dup(aWString);
}

wchar_t*
TestInterfaceEx_impl::opWStringEx(const wchar_t* a0,
				  wchar_t*& a1,
				  WString_out)
    throw(TestInterfaceEx::ExWString,
	  SystemException)
{
    aWString = wstring_alloc(OB::wcslen(a0) + OB::wcslen(a1));
    OB::wcscpy(aWString.inout(), a0);
    OB::wcscat(aWString.inout(), a1);
    throw TestInterfaceEx::ExWString(aWString);
    return 0; // Some compilers need this
}

Any*
TestInterface_impl::attrAny()
    throw(SystemException)
{
    return new Any(aAny);
}

void
TestInterface_impl::attrAny(const Any& val)
    throw(SystemException)
{
    aAny = val;
}

Any*
TestInterface_impl::opAny(const Any& a0,
			  Any& a1,
			  Any_out a2)
    throw(SystemException)
{
    aAny = a0;
    a1 = aAny;
    a2 = new Any(aAny);
    return new Any(aAny);
}

Any*
TestInterfaceEx_impl::opAnyEx(const Any& a0,
			      Any&,
			      Any_out)
    throw(TestInterfaceEx::ExAny,
	  SystemException)
{
    aAny = a0;
    throw TestInterfaceEx::ExAny(aAny);
    return 0; // Some compilers need this
}

TestInterface::TestEnum
TestInterface_impl::attrTestEnum()
    throw(SystemException)
{
    return aTestEnum;
}

void
TestInterface_impl::attrTestEnum(TestInterface::TestEnum val)
    throw(SystemException)
{
    aTestEnum = val;
}

TestInterface::TestEnum
TestInterface_impl::opTestEnum(TestInterface::TestEnum a0,
			       TestInterface::TestEnum& a1,
			       TestInterface::TestEnum& a2)
    throw(SystemException)
{
    aTestEnum = a0;
    a1 = a2 = aTestEnum;
    return aTestEnum;
}

TestInterface::TestEnum
TestInterfaceEx_impl::opTestEnumEx(TestInterface::TestEnum a0,
				   TestInterface::TestEnum&,
				   TestInterface::TestEnum&)
    throw(TestInterfaceEx::ExTestEnum,
	  SystemException)
{
    aTestEnum = a0;
    throw TestInterfaceEx::ExTestEnum(aTestEnum);
    return TestInterface::TestEnum1; // Some compilers need this
}

TestInterface_ptr
TestInterface_impl::attrTestInterface()
    throw(SystemException)
{
    return TestInterface::_duplicate(aTestInterface);
}

void
TestInterface_impl::attrTestInterface(TestInterface_ptr val)
    throw(SystemException)
{
    aTestInterface = TestInterface::_duplicate(val);
}

TestInterface_ptr
TestInterface_impl::opTestInterface(TestInterface_ptr a0,
				    TestInterface_ptr& a1,
				    TestInterface_out a2)
    throw(SystemException)
{
    aTestInterface = TestInterface::_duplicate(a0);
    release(a1);
    a1 = TestInterface::_duplicate(aTestInterface);
    a2 = TestInterface::_duplicate(aTestInterface);
    return TestInterface::_duplicate(aTestInterface);
}

TestInterface_ptr
TestInterfaceEx_impl::opTestInterfaceEx(TestInterface_ptr a0,
					TestInterface_ptr&,
					TestInterface_out)
    throw(TestInterfaceEx::ExTestInterface,
	  SystemException)
{
    aTestInterface = TestInterface::_duplicate(a0);
    throw TestInterfaceEx::ExTestInterface(aTestInterface);
    return TestInterface::_nil(); // Some compilers need this
}

TestInterface::FixedStruct
TestInterface_impl::attrFixedStruct()
    throw(SystemException)
{
    return aFixedStruct;
}

void
TestInterface_impl::attrFixedStruct(const TestInterface::FixedStruct& val)
    throw(SystemException)
{
    aFixedStruct = val;
}

TestInterface::FixedStruct
TestInterface_impl::opFixedStruct(const TestInterface::FixedStruct& a0,
				  TestInterface::FixedStruct& a1,
				  TestInterface::FixedStruct& a2)
    throw(SystemException)
{
    aFixedStruct = a0;
    a1 = a2 = aFixedStruct;
    return aFixedStruct;
}

TestInterface::FixedStruct
TestInterfaceEx_impl::opFixedStructEx(const TestInterface::FixedStruct& a0,
				      TestInterface::FixedStruct&,
				      TestInterface::FixedStruct&)
    throw(TestInterfaceEx::ExFixedStruct,
	  SystemException)
{
    aFixedStruct = a0;
    throw TestInterfaceEx::ExFixedStruct(aFixedStruct);
    return TestInterface::FixedStruct(); // Some compilers need this
}

TestInterface::VariableStruct*
TestInterface_impl::attrVariableStruct()
    throw(SystemException)
{
    return new TestInterface::VariableStruct(aVariableStruct);
}

void
TestInterface_impl::attrVariableStruct(const TestInterface::VariableStruct&
				       val)
    throw(SystemException)
{
    aVariableStruct = val;
}

TestInterface::VariableStruct*
TestInterface_impl::opVariableStruct(const TestInterface::VariableStruct& a0,
				     TestInterface::VariableStruct& a1,
				     TestInterface::VariableStruct_out a2)
    throw(SystemException)
{
    aVariableStruct = a0;
    a1 = aVariableStruct;
    a2 = new TestInterface::VariableStruct(aVariableStruct);
    return new TestInterface::VariableStruct(aVariableStruct);
}

TestInterface::VariableStruct*
TestInterfaceEx_impl::opVariableStructEx(const TestInterface::VariableStruct&
					 a0,
					 TestInterface::VariableStruct&,
					 TestInterface::VariableStruct_out)
    throw(TestInterfaceEx::ExVariableStruct,
	  SystemException)
{
    aVariableStruct = a0;
    throw TestInterfaceEx::ExVariableStruct(aVariableStruct);
    return new TestInterface::VariableStruct; // Some compilers need this
}

TestInterface::FixedUnion
TestInterface_impl::attrFixedUnion()
    throw(SystemException)
{
    return aFixedUnion;
}

void
TestInterface_impl::attrFixedUnion(const TestInterface::FixedUnion& val)
    throw(SystemException)
{
    aFixedUnion = val;
}

TestInterface::FixedUnion
TestInterface_impl::opFixedUnion(const TestInterface::FixedUnion& a0,
				 TestInterface::FixedUnion& a1,
				 TestInterface::FixedUnion& a2)
    throw(SystemException)
{
    aFixedUnion = a0;
    a1 = a2 = aFixedUnion;
    return aFixedUnion;
}

TestInterface::FixedUnion
TestInterfaceEx_impl::opFixedUnionEx(const TestInterface::FixedUnion& a0,
				     TestInterface::FixedUnion&,
				     TestInterface::FixedUnion&)
    throw(TestInterfaceEx::ExFixedUnion,
	  SystemException)
{
    aFixedUnion = a0;
    throw TestInterfaceEx::ExFixedUnion(aFixedUnion);
    return TestInterface::FixedUnion(); // Some compilers need this
}

TestInterface::VariableUnion*
TestInterface_impl::attrVariableUnion()
    throw(SystemException)
{
    return new TestInterface::VariableUnion(aVariableUnion);
}

void
TestInterface_impl::attrVariableUnion(const TestInterface::VariableUnion& val)
    throw(SystemException)
{
    aVariableUnion = val;
}

TestInterface::VariableUnion*
TestInterface_impl::opVariableUnion(const TestInterface::VariableUnion& a0,
				    TestInterface::VariableUnion& a1,
				    TestInterface::VariableUnion_out a2)
    throw(SystemException)
{
    aVariableUnion = a0;
    a1 = aVariableUnion;
    a2 = new TestInterface::VariableUnion(aVariableUnion);
    return new TestInterface::VariableUnion(aVariableUnion);
}

TestInterface::VariableUnion*
TestInterfaceEx_impl::opVariableUnionEx(const TestInterface::VariableUnion& a0,
					TestInterface::VariableUnion&,
					TestInterface::VariableUnion_out)
    throw(TestInterfaceEx::ExVariableUnion,
	  SystemException)
{
    aVariableUnion = a0;
    throw TestInterfaceEx::ExVariableUnion(aVariableUnion);
    return new TestInterface::VariableUnion; // Some compilers need this
}

TestInterface::StringSequence*
TestInterface_impl::attrStringSequence()
    throw(SystemException)
{
    return new TestInterface::StringSequence(aStringSequence);
}

void
TestInterface_impl::attrStringSequence(const TestInterface::StringSequence&
				       val)
    throw(SystemException)
{
    aStringSequence = val;
}

TestInterface::StringSequence*
TestInterface_impl::opStringSequence(const TestInterface::StringSequence& a0,
				     TestInterface::StringSequence& a1,
				     TestInterface::StringSequence_out a2)
    throw(SystemException)
{
    aStringSequence = a0;
    for(ULong i = 0 ; i < a1.length() ; i++)
	aStringSequence.append(a1[i]);

    a1 = aStringSequence;
    a2 = new TestInterface::StringSequence(aStringSequence);
    return new TestInterface::StringSequence(aStringSequence);
}

TestInterface::StringSequence*
TestInterfaceEx_impl::
opStringSequenceEx(const TestInterface::StringSequence& a0,
		   TestInterface::StringSequence& a1,
		   TestInterface::StringSequence_out)
    throw(TestInterfaceEx::ExStringSequence,
	  SystemException)
{
    aStringSequence = a0;
    for(ULong i = 0 ; i < a1.length() ; i++)
	aStringSequence.append(a1[i]);

    throw TestInterfaceEx::ExStringSequence(aStringSequence);
    return new TestInterface::StringSequence; // Some compilers need this
}

TestInterface::FixedArray_slice*
TestInterface_impl::attrFixedArray()
    throw(SystemException)
{
    return TestInterface::FixedArray_dup(aFixedArray);
}

void
TestInterface_impl::attrFixedArray(const TestInterface::FixedArray val)
    throw(SystemException)
{
    TestInterface::FixedArray_copy(val, aFixedArray);
}

TestInterface::FixedArray_slice*
TestInterface_impl::opFixedArray(const TestInterface::FixedArray a0,
				 TestInterface::FixedArray a1,
				 TestInterface::FixedArray a2)
    throw(SystemException)
{
    TestInterface::FixedArray_copy(a0, aFixedArray);
    TestInterface::FixedArray_copy(aFixedArray, a1);
    TestInterface::FixedArray_copy(aFixedArray, a2);
    return TestInterface::FixedArray_dup(aFixedArray);
}

TestInterface::FixedArray_slice*
TestInterfaceEx_impl::opFixedArrayEx(const TestInterface::FixedArray a0,
				     TestInterface::FixedArray,
				     TestInterface::FixedArray)
    throw(TestInterfaceEx::ExFixedArray,
	  SystemException)
{
    TestInterface::FixedArray_copy(a0, aFixedArray);
#ifdef HAVE_VCPLUSPLUS_BUGS
    throw TestInterfaceEx::ExFixedArray(
	(const TestInterface::FixedArray_slice*)aFixedArray);
#else
    throw TestInterfaceEx::ExFixedArray(aFixedArray);
#endif
    return TestInterface::FixedArray_alloc(); // Some compilers need this
}

TestInterface::VariableArray_slice*
TestInterface_impl::attrVariableArray()
    throw(SystemException)
{
    return TestInterface::VariableArray_dup(aVariableArray);
}

void
TestInterface_impl::attrVariableArray(const TestInterface::VariableArray val)
    throw(SystemException)
{
    TestInterface::VariableArray_copy(val, aVariableArray);
}

TestInterface::VariableArray_slice*
TestInterface_impl::opVariableArray(const TestInterface::VariableArray a0,
				    TestInterface::VariableArray a1,
				    TestInterface::VariableArray_out a2)
    throw(SystemException)
{
    TestInterface::VariableArray_copy(a0, aVariableArray);
    TestInterface::VariableArray_copy(aVariableArray, a1);
    a2 = TestInterface::VariableArray_dup(aVariableArray);
    return TestInterface::VariableArray_dup(aVariableArray);
}

TestInterface::VariableArray_slice*
TestInterfaceEx_impl::opVariableArrayEx(const TestInterface::VariableArray a0,
					TestInterface::VariableArray,
					TestInterface::VariableArray_out)
    throw(TestInterfaceEx::ExVariableArray,
	  SystemException)
{
    TestInterface::VariableArray_copy(a0, aVariableArray);
#ifdef HAVE_VCPLUSPLUS_BUGS
    throw TestInterfaceEx::ExVariableArray(
	(const TestInterface::VariableArray_slice*)aVariableArray);
#else
    throw TestInterfaceEx::ExVariableArray(aVariableArray);
#endif
    return TestInterface::VariableArray_alloc(); // Some compilers need this
}

TestInterface::FixedArraySequence*
TestInterface_impl::attrFixedArraySequence()
    throw(SystemException)
{
    return new TestInterface::FixedArraySequence(aFixedArraySequence);
}

void
TestInterface_impl::
attrFixedArraySequence(const TestInterface::FixedArraySequence& val)
    throw(SystemException)
{
    aFixedArraySequence = val;
}

TestInterface::FixedArraySequence*
TestInterface_impl::
opFixedArraySequence(const TestInterface::FixedArraySequence& a0,
		     TestInterface::FixedArraySequence& a1,
		     TestInterface::FixedArraySequence_out a2)
    throw(SystemException)
{
    aFixedArraySequence.length(a0.length() + a1.length());
    
    ULong i;
    ULong n = 0;

    for(i = 0 ; i < a0.length() ; i++)
	TestInterface::FixedArray_copy(a0[i], aFixedArraySequence[n++]);
	
    for(i = 0 ; i < a1.length() ; i++)
	TestInterface::FixedArray_copy(a1[i], aFixedArraySequence[n++]);

    a1 = aFixedArraySequence;
    a2 = new TestInterface::FixedArraySequence(aFixedArraySequence);
    return new TestInterface::FixedArraySequence(aFixedArraySequence);
}

TestInterface::FixedArraySequence*
TestInterfaceEx_impl::
opFixedArraySequenceEx(const TestInterface::FixedArraySequence& a0,
		       TestInterface::FixedArraySequence& a1,
		       TestInterface::FixedArraySequence_out)
    throw(TestInterfaceEx::ExFixedArraySequence,
	  SystemException)
{
    aFixedArraySequence.length(a0.length() + a1.length());
    
    ULong i;
    ULong n = 0;

    for(i = 0 ; i < a0.length() ; i++)
	TestInterface::FixedArray_copy(a0[i], aFixedArraySequence[n++]);
	
    for(i = 0 ; i < a1.length() ; i++)
	TestInterface::FixedArray_copy(a1[i], aFixedArraySequence[n++]);

    throw TestInterfaceEx::ExFixedArraySequence(aFixedArraySequence);

    // Some compilers need this
    return new TestInterface::FixedArraySequence(aFixedArraySequence);
}

TestInterface::VariableArraySequence*
TestInterface_impl::attrVariableArraySequence()
    throw(SystemException)
{
    return new TestInterface::VariableArraySequence(aVariableArraySequence);
}

void
TestInterface_impl::
attrVariableArraySequence(const TestInterface::VariableArraySequence& val)
    throw(SystemException)
{
    aVariableArraySequence = val;
}

TestInterface::VariableArraySequence*
TestInterface_impl::
opVariableArraySequence(const TestInterface::VariableArraySequence& a0,
			TestInterface::VariableArraySequence& a1,
			TestInterface::VariableArraySequence_out a2)
    throw(SystemException)
{
    aVariableArraySequence.length(a0.length() + a1.length());
    
    ULong i;
    ULong n = 0;

    for(i = 0 ; i < a0.length() ; i++)
	TestInterface::VariableArray_copy(a0[i], aVariableArraySequence[n++]);
	
    for(i = 0 ; i < a1.length() ; i++)
	TestInterface::VariableArray_copy(a1[i], aVariableArraySequence[n++]);

    a1 = aVariableArraySequence;
    a2 = new TestInterface::VariableArraySequence(aVariableArraySequence);
    return new TestInterface::VariableArraySequence(aVariableArraySequence);
}

TestInterface::VariableArraySequence*
TestInterfaceEx_impl::
opVariableArraySequenceEx(const TestInterface::VariableArraySequence& a0,
			  TestInterface::VariableArraySequence& a1,
			  TestInterface::VariableArraySequence_out)
    throw(TestInterfaceEx::ExVariableArraySequence,
	  SystemException)
{
    aVariableArraySequence.length(a0.length() + a1.length());
    
    ULong i;
    ULong n = 0;

    for(i = 0 ; i < a0.length() ; i++)
	TestInterface::VariableArray_copy(a0[i], aVariableArraySequence[n++]);
	
    for(i = 0 ; i < a1.length() ; i++)
	TestInterface::VariableArray_copy(a1[i], aVariableArraySequence[n++]);

    throw TestInterfaceEx::ExVariableArraySequence(aVariableArraySequence);

    // Some compilers need this
    return new TestInterface::VariableArraySequence(aVariableArraySequence);
}

void
TestInterfaceEx_impl::op_UNKNOWN_Ex()
    throw(SystemException)
{
    throw UNKNOWN(1, COMPLETED_YES);
}

void
TestInterfaceEx_impl::op_BAD_PARAM_Ex()
    throw(SystemException)
{
    throw BAD_PARAM(2, COMPLETED_NO);
}

void
TestInterfaceEx_impl::op_NO_MEMORY_Ex()
    throw(SystemException)
{
    throw NO_MEMORY(3, COMPLETED_MAYBE);
}

void
TestInterfaceEx_impl::op_IMP_LIMIT_Ex()
    throw(SystemException)
{
    throw IMP_LIMIT(4, COMPLETED_YES);
}

void
TestInterfaceEx_impl::op_COMM_FAILURE_Ex()
    throw(SystemException)
{
    throw COMM_FAILURE(5, COMPLETED_NO);
}

void
TestInterfaceEx_impl::op_INV_OBJREF_Ex()
    throw(SystemException)
{
    throw INV_OBJREF(6, COMPLETED_MAYBE);
}

void
TestInterfaceEx_impl::op_NO_PERMISSION_Ex()
    throw(SystemException)
{
    throw NO_PERMISSION(7, COMPLETED_YES);
}

void
TestInterfaceEx_impl::op_INTERNAL_Ex()
    throw(SystemException)
{
    throw INTERNAL(8, COMPLETED_NO);
}

void
TestInterfaceEx_impl::op_MARSHAL_Ex()
    throw(SystemException)
{
    throw MARSHAL(9, COMPLETED_MAYBE);
}

void
TestInterfaceEx_impl::op_INITIALIZE_Ex()
    throw(SystemException)
{
    throw INITIALIZE(10, COMPLETED_YES);
}

void
TestInterfaceEx_impl::op_NO_IMPLEMENT_Ex()
    throw(SystemException)
{
    throw NO_IMPLEMENT(11, COMPLETED_NO);
}

void
TestInterfaceEx_impl::op_BAD_TYPECODE_Ex()
    throw(SystemException)
{
    throw BAD_TYPECODE(12, COMPLETED_MAYBE);
}

void
TestInterfaceEx_impl::op_BAD_OPERATION_Ex()
    throw(SystemException)
{
    throw BAD_OPERATION(13, COMPLETED_YES);
}

void
TestInterfaceEx_impl::op_NO_RESOURCES_Ex()
    throw(SystemException)
{
    throw NO_RESOURCES(14, COMPLETED_NO);
}

void
TestInterfaceEx_impl::op_NO_RESPONSE_Ex()
    throw(SystemException)
{
    throw NO_RESPONSE(15, COMPLETED_MAYBE);
}

void
TestInterfaceEx_impl::op_PERSIST_STORE_Ex()
    throw(SystemException)
{
    throw PERSIST_STORE(16, COMPLETED_YES);
}

void
TestInterfaceEx_impl::op_BAD_INV_ORDER_Ex()
    throw(SystemException)
{
    throw BAD_INV_ORDER(17, COMPLETED_NO);
}

void
TestInterfaceEx_impl::op_TRANSIENT_Ex()
    throw(SystemException)
{
    throw TRANSIENT(18, COMPLETED_MAYBE);
}

void
TestInterfaceEx_impl::op_FREE_MEM_Ex()
    throw(SystemException)
{
    throw FREE_MEM(19, COMPLETED_YES);
}

void
TestInterfaceEx_impl::op_INV_IDENT_Ex()
    throw(SystemException)
{
    throw INV_IDENT(20, COMPLETED_NO);
}

void
TestInterfaceEx_impl::op_INV_FLAG_Ex()
    throw(SystemException)
{
    throw INV_FLAG(21, COMPLETED_MAYBE);
}

void
TestInterfaceEx_impl::op_INTF_REPOS_Ex()
    throw(SystemException)
{
    throw INTF_REPOS(22, COMPLETED_YES);
}

void
TestInterfaceEx_impl::op_BAD_CONTEXT_Ex()
    throw(SystemException)
{
    throw BAD_CONTEXT(23, COMPLETED_NO);
}

void
TestInterfaceEx_impl::op_OBJ_ADAPTER_Ex()
    throw(SystemException)
{
    throw OBJ_ADAPTER(24, COMPLETED_MAYBE);
}

void
TestInterfaceEx_impl::op_DATA_CONVERSION_Ex()
    throw(SystemException)
{
    throw DATA_CONVERSION(25, COMPLETED_YES);
}

void
TestInterfaceEx_impl::op_OBJECT_NOT_EXIST_Ex()
    throw(SystemException)
{
    throw OBJECT_NOT_EXIST(26, COMPLETED_NO);
}

void
TestInterfaceEx_impl::op_TRANSACTION_REQUIRED_Ex()
    throw(SystemException)
{
    throw TRANSACTION_REQUIRED(27, COMPLETED_NO);
}

void
TestInterfaceEx_impl::op_TRANSACTION_ROLLEDBACK_Ex()
    throw(SystemException)
{
    throw TRANSACTION_ROLLEDBACK(28, COMPLETED_NO);
}

void
TestInterfaceEx_impl::op_INVALID_TRANSACTION_Ex()
    throw(SystemException)
{
    throw INVALID_TRANSACTION(29, COMPLETED_NO);
}

void
TestInterfaceEx_impl::op_INV_POLICY_Ex()
    throw(SystemException)
{
    throw INV_POLICY(30, COMPLETED_NO);
}

void
TestInterface_impl::releaseInternal()
{
    //
    // Avoid cyclic object dependencies
    //
    aTestInterface = TestInterface::_nil();
    aVariableUnion = TestInterface::VariableUnion();
}

void
TestInterface_impl::deactivate()
    throw(SystemException)
{
    orb_ -> shutdown(false);
}

TestInterfaceTimeout_impl::TestInterfaceTimeout_impl(ORB_ptr orb)

    : TestInterfaceEx_impl(orb), TestInterface_impl(orb)
{
}

void
TestInterfaceTimeout_impl::sleepOneway(ULong sec)
        throw(SystemException)
{
#ifdef WIN32
    Sleep(sec * 1000);
#else
    sleep(sec);
#endif
};

void
TestInterfaceTimeout_impl::sleepTwoway(ULong sec)
        throw(SystemException)
{
#ifdef WIN32
    Sleep(sec * 1000);
#else
    sleep(sec);
#endif
};

