// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>

#include <TestOBV.h>
#include <Factories.h>
#include <TestMacro.h>

#include <TestOBV_skel.h>
#include <TestOBVColo_impl.h>

#include <stdlib.h>
#include <string.h>
#include <errno.h>

#ifdef HAVE_UNISTD_H
#   include <unistd.h>
#endif

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;
using namespace PortableServer;

static void
testValue(ORB_ptr orb, TestOBV_ptr to)
{
    ValueBase_var vb;
    TestValue_var value, v1, v2;
    TestValueSub_var valueSub;
    TestValueSub* pValueSub;
    TestValueSub* s1;
    TestValueSub* s2;
    TestAbsValue1_var a1;
    TestAbsValue2_var a2;

    vb = to -> get_null_valuebase();
    TEST(vb.in() == 0);
    to -> set_null_valuebase(0);

    valueSub = to -> get_null_valuesub();
    TEST(valueSub.in() == 0);
    to -> set_null_valuesub(0);

    a1 = to -> get_abs_value1();
    TEST(a1.in() != 0);
    a1 -> ping1();
    to -> set_abs_value1(a1);

    a2 = to -> get_abs_value2();
    TEST(a2.in() != 0);
    a2 -> ping2();
    to -> set_abs_value2(a2);

    value = to -> get_value();
    TEST(value.in() != 0);
    value -> ping1();
    TEST(value -> count() == 500);
    to -> set_value(value);

    valueSub = to -> get_valuesub();
    TEST(valueSub.in() != 0);
    valueSub -> ping1();
    valueSub -> ping2();
    TEST(valueSub -> count() == 501);
    TEST(strcmp(valueSub -> name(), "ValueSub") == 0);
    to -> set_valuesub(valueSub);

    value = to -> get_valuesub_as_value();
    TEST(value.in() != 0);
    value -> ping1();
    TEST(value -> count() == 501);
    pValueSub = TestValueSub::_downcast(value);
    TEST(pValueSub != 0);
    pValueSub -> ping2();
    TEST(strcmp(pValueSub -> name(), "ValueSub") == 0);
    to -> set_valuesub_as_value(value);

    to -> get_two_values(v1.out(), v2.out());
    TEST(v1.in() != 0);
    TEST(v2.in() != 0);
    TEST(v1.in() == v2.in());
    v1 -> ping1();
    TEST(v1 -> count() == 500);
    to -> set_two_values(v1, v2);

    to -> get_two_valuesubs_as_values(v1.out(), v2.out());
    TEST(v1.in() != 0);
    TEST(v2.in() != 0);
    v1 -> ping1();
    v2 -> ping1();
    TEST(v1 -> count() == v2 -> count());
    s1 = TestValueSub::_downcast(v1);
    TEST(s1 != 0);
    s2 = TestValueSub::_downcast(v2);
    TEST(s2 != 0);
    s1 -> ping2();
    s2 -> ping2();
    TEST(strcmp(s1 -> name(), s2 -> name()) == 0);
    to -> set_two_valuesubs_as_values(v1, v2);
}

#if 0 // XXX enable when custom marshalling is supported
static void
testCustom(ORB_ptr orb, TestOBV_ptr to)
{
    TestCustom_var cust;
    TestCustom* pCust;
    TestAbsValue1_var a1;

    cust = to -> get_custom();
    TEST(cust.in() != 0);
    cust -> ping1();
    TEST(cust -> shortVal() == -99);
    TEST(cust -> longVal() == -123456);
    TEST(strcmp(cust -> stringVal(), "CustomVal") == 0);
    TEST(cust -> doubleVal() == 100.997);
    to -> set_custom(cust);

    a1 = to -> get_abs_custom();
    TEST(a1.in() != 0);
    a1 -> ping1();
    pCust = TestCustom::_downcast(a1);
    TEST(pCust != 0);
    TEST(pCust -> shortVal() == -99);
    TEST(pCust -> longVal() == -123456);
    TEST(strcmp(pCust -> stringVal(), "CustomVal") == 0);
    TEST(pCust -> doubleVal() == 100.997);
    to -> set_abs_custom(a1);
}
#endif

static void
testValueBox(ORB_ptr orb, TestOBV_ptr to)
{
    ULong i;

    TestStringBox_var sb;
    sb = to -> get_string_box("hi there");
    TEST(sb.in() != 0);
    TEST(strcmp(sb -> _value(), "hi there") == 0);
    sb = new TestStringBox(string_dup("bye now"));
    to -> set_string_box(sb, "bye now");

    TestULongBox_var ub;
    ub = to -> get_ulong_box(999);
    TEST(ub.in() != 0);
    TEST(ub -> _value() == 999);
    ub = new TestULongBox(77777);
    to -> set_ulong_box(ub, 77777);

    TestFixStructBox_var fsb;
    TestFixStruct fs;
    fs.x = 111;
    fs.y = 222;
    fs.radius = 3.33;
    fsb = to -> get_fix_struct_box(fs);
    TEST(fsb.in() != 0);
    TEST(fsb -> x() == fs.x);
    TEST(fsb -> y() == fs.y);
    TEST(fsb -> radius() == fs.radius);
    fsb = new TestFixStructBox(fs);
    to -> set_fix_struct_box(fsb, fs);

    TestVarStructBox_var vsb;
    TestVarStruct vs;
    vs.name = string_dup("Joe Bob Briggs");
    vs.email = string_dup("jbb@cheese.com");
    vsb = to -> get_var_struct_box(vs);
    TEST(vsb.in() != 0);
    TEST(strcmp(vsb -> name(), vs.name) == 0);
    TEST(strcmp(vsb -> email(), vs.email) == 0);
    vsb = new TestVarStructBox(vs);
    to -> set_var_struct_box(vsb, vs);

    TestFixUnionBox_var fub;
    TestFixUnion fu;
    fu.o((Octet)55);
    fub = to -> get_fix_union_box(fu);
    TEST(fub.in() != 0);
    TEST(fub -> o() == (Octet)55);
    fu.d(99.88);
    fub = new TestFixUnionBox(fu);
    to -> set_fix_union_box(fub, fu);

    TestVarUnionBox_var vub;
    TestVarUnion vu;
    vu.s(string_dup("howdy"));
    vub = to -> get_var_union_box(vu);
    TEST(vub.in() != 0);
    TEST(strcmp(vub -> s(), "howdy") == 0);
    vu.fs(fs);
    vub = new TestVarUnionBox(vu);
    to -> set_var_union_box(vub, vu);

    TestAnonSeqBox_var asb;
    asb = to -> get_anon_seq_box(10);
    TEST(asb.in() != 0);
    TEST(asb -> length() == 10);
    for(i = 0 ; i < asb -> length() ; i++)
        TEST((*asb)[i] == (Short)i);
    to -> set_anon_seq_box(asb, 10);

    TestStringSeqBox_var ssb;
    TestStringSeq ss;
    ss.length(5);
    for(i = 0 ; i < 5 ; i++)
    {
        String_var str = string_dup("s");
        str += i;
        ss[i] = str;
    }
    ssb = to -> get_string_seq_box(ss);
    TEST(ssb.in() != 0);
    TEST(ssb -> length() == ss.length());
    for(i = 0 ; i < ssb -> length() ; i++)
        TEST(strcmp((*ssb)[i], ss[i]) == 0);
    to -> set_string_seq_box(ssb, ss);
}

static void
testCollocated(ORB_ptr orb, TestOBV_ptr to)
{
    Object_var obj = orb -> resolve_initial_references("RootPOA");
    POA_var poa = POA::_narrow(obj);
    assert(!is_nil(poa));

    TestValue_var value;
    TestOBVColo_impl coloImpl;
    TestOBVColo_var colo = coloImpl._this();

    value = to -> get_value();
    TEST(value.in() != 0);
    value -> count(111);
    colo -> set_expected_count(111);

    colo -> test_attribute(0);
    colo -> test_op(0);

    colo -> test_attribute(value);
    TEST(value -> count() == 111);

    value -> count(222);
    colo -> set_expected_count(222);

    colo -> test_op(value);
    TEST(value -> count() == 222);

    ObjectId_var id = poa -> servant_to_id(&coloImpl);
    poa -> deactivate_object(id);
}

static void
testAbstract(ORB_ptr orb, TestOBV_ptr to)
{
    TestAbstract_var ai;
    ai = to -> get_ai_interface();
    assert(!is_nil(ai));
    ai -> abstract_op();
    to -> set_ai_interface(ai);

    Any_var any;
    any = to -> get_ai_interface_any();
    Object_ptr pobj;
    Any::to_object to_obj(pobj);
    assert(any >>= to_obj);
    assert(!is_nil(pobj));
    TestAbstractSub_var sub = TestAbstractSub::_narrow(pobj);
    assert(!is_nil(sub));
    sub -> abstract_op();
    sub -> sub_op();
    to -> set_ai_interface_any(any);
}

static void
testTruncated(ORB_ptr orb, TestOBV_ptr to)
{
    TestTruncBase_var truncBase;
    TestTrunc1* trunc1;
    TestTrunc2* trunc2;
    TestAbsValue1* a;
    TestValue* v;

    //
    // With factory installed, we should be able to downcast to TestTrunc1
    //
    TestTrunc1Factory_impl::install(orb);
    truncBase = to -> get_trunc1();
    trunc1 = TestTrunc1::_downcast(truncBase);
    assert(trunc1 != 0);
    assert(trunc1 -> cost() > 1.99 && trunc1 -> cost() < 2.0);
    assert(trunc1 -> boolVal() == TRUE);
    a = trunc1 -> v();
    assert(a != 0);
    v = TestValue::_downcast(a);
    assert(v != 0);
    assert(v -> count() == 999);
    assert(trunc1 -> shortVal() == 12667);

    //
    // With factory removed, we should not be able to downcast to TestTrunc1
    //
    orb -> unregister_value_factory("IDL:TestTrunc1:1.0");
    truncBase = to -> get_trunc1();
    trunc1 = TestTrunc1::_downcast(truncBase);
    assert(trunc1 == 0);
    assert(truncBase -> cost() > 1.99 && truncBase -> cost() < 2.0);

    //
    // With factories installed, we should be able to downcast to TestTrunc2
    //
    TestTrunc1Factory_impl::install(orb);
    TestTrunc2Factory_impl::install(orb);
    truncBase = to -> get_trunc2();
    trunc2 = TestTrunc2::_downcast(truncBase);
    assert(trunc2 != 0);
    assert(trunc2 -> cost() > 5.99 && trunc2 -> cost() < 6.0);
    trunc1 = TestTrunc1::_downcast(trunc2 -> t());
    assert(trunc1 != 0);
    assert(trunc1 -> cost() > 1.99 && trunc1 -> cost() < 2.0);
    assert(trunc1 -> boolVal() == TRUE);
    a = trunc1 -> v();
    assert(a != 0);
    v = TestValue::_downcast(a);
    assert(v != 0);
    assert(v -> count() == 999);
    assert(trunc1 -> shortVal() == 12667);
    a = trunc2 -> a();
    assert(a != 0);
    v = TestValue::_downcast(a);
    assert(v != 0);
    assert(v -> count() == 9999);
    assert(trunc2 -> b1() == trunc2); // recursion
    assert(trunc2 -> v() == trunc1 -> v()); // indirection
    assert(trunc2 -> b2() == trunc1); // indirection

    //
    // Without a factory for TestTrunc1, some nested values of TestTrunc2
    // will be truncated
    //
    orb -> unregister_value_factory("IDL:TestTrunc1:1.0");
    truncBase = to -> get_trunc2();
    trunc2 = TestTrunc2::_downcast(truncBase);
    assert(trunc2 != 0);
    assert(trunc2 -> cost() > 5.99 && trunc2 -> cost() < 6.0);
    assert(trunc2 -> t() != 0);
    trunc1 = TestTrunc1::_downcast(trunc2 -> t());
    assert(trunc1 == 0);
    assert(trunc2 -> t() -> cost() > 1.99 && trunc2 -> t() -> cost() < 2.0);
    a = trunc2 -> a();
    assert(a != 0);
    v = TestValue::_downcast(a);
    assert(v != 0);
    assert(v -> count() == 9999);
    assert(trunc2 -> b1() == trunc2); // recursion
    assert(trunc2 -> v() != 0); // indirection
    assert(trunc2 -> v() -> count() == 999); // indirection
    assert(trunc2 -> b2() == trunc2 -> t()); // indirection

    //
    // With factory removed, we should not be able to downcast to TestTrunc2
    //
    TestTrunc1Factory_impl::install(orb);
    orb -> unregister_value_factory("IDL:TestTrunc2:1.0");
    truncBase = to -> get_trunc2();
    trunc2 = TestTrunc2::_downcast(truncBase);
    assert(trunc2 == 0);
    assert(truncBase -> cost() > 5.99 && truncBase -> cost() < 6.0);
}

int
ClientRun(ORB_ptr orb, int argc, char* argv[])
{
    //
    // Get TestOBV
    //
    const char* refFile = "TestOBV.ref";
    ifstream in; // Must use open(name), not ifstream in(name) (VC++ bug)
    in.open(refFile);
    if(in.fail())
    {
        cerr << argv[0] << ": can't open `" << refFile << "': "
             << strerror(errno) << endl;
        return 1;
    }
    
    char impl[1000];
    in >> impl;

    in.close(); // The explicit close is necessary to satisfy the Windows NT
                // file locking

    Object_var obj = orb -> string_to_object(impl);
    TEST(!is_nil(obj));
    TestOBV_var to = TestOBV::_narrow(obj);
    TEST(!is_nil(to));

    //
    // Install valuetype factories
    //
    ValueFactoryBase_var valueFactory =
        TestValueFactory_impl::install(orb);
    ValueFactoryBase_var valueSubFactory =
        TestValueSubFactory_impl::install(orb);
    TestTruncBaseFactory_impl::install(orb);
    ValueFactoryBase_var customFactory =
        TestCustomFactory_impl::install(orb);
    TestStringBoxFactory_impl::install(orb);
    TestULongBoxFactory_impl::install(orb);
    TestFixStructBoxFactory_impl::install(orb);
    TestVarStructBoxFactory_impl::install(orb);
    TestFixUnionBoxFactory_impl::install(orb);
    TestVarUnionBoxFactory_impl::install(orb);
    TestAnonSeqBoxFactory_impl::install(orb);
    TestStringSeqBoxFactory_impl::install(orb);

    //
    // Run tests
    //

    cout << "Testing valuetypes... " << flush;
    testValue(orb, to);
    cout << "OK!" << endl;

#if 0 // XXX enable when custom marshalling is supported
    cout << "Testing custom marshalling... " << flush;
    testCustom(orb, to);
    cout << "OK!" << endl;
#endif

    cout << "Testing value boxes... " << flush;
    testValueBox(orb, to);
    cout << "OK!" << endl;

    cout << "Testing collocated valuetypes... " << flush;
    testCollocated(orb, to);
    cout << "OK!" << endl;

    cout << "Testing abstract interfaces... " << flush;
    testAbstract(orb, to);
    cout << "OK!" << endl;

    cout << "Testing truncatable valuetypes... " << flush;
    testTruncated(orb, to);
    cout << "OK!" << endl;


    try
    {
        to -> deactivate();
    }
    catch(const COMM_FAILURE&)
    {
        // This call will deactivate the server and thus throw an
        // exception
    }

    return(0);
}
