// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>

#include <TestOBV.h>
#include <Factories.h>
#include <TestOBV_skel.h>
#include <TestOBV_impl.h>
#include <TestAbstractSub_impl.h>

#include <errno.h>

#ifdef HAVE_UNISTD_H
#   include <unistd.h>
#endif

#ifdef HAVE_IO_H
#   include <io.h>
#endif

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;
using namespace PortableServer;

int
ServerRun(ORB_ptr orb, POA_ptr poa, int argc, char* argv[])
{
    //
    // Install value factories
    //
    TestValue_init* valueFactory = TestValueFactory_impl::install(orb);
    TestValueSub_init* valueSubFactory =
        TestValueSubFactory_impl::install(orb);
    TestTruncBaseFactory_impl::install(orb);
    TestCustom_init* customFactory = TestCustomFactory_impl::install(orb);
    TestStringBoxFactory_impl::install(orb);
    TestULongBoxFactory_impl::install(orb);
    TestFixStructBoxFactory_impl::install(orb);
    TestVarStructBoxFactory_impl::install(orb);
    TestFixUnionBoxFactory_impl::install(orb);
    TestVarUnionBoxFactory_impl::install(orb);
    TestAnonSeqBoxFactory_impl::install(orb);
    TestStringSeqBoxFactory_impl::install(orb);

    //
    // Create implementation objects
    //
    TestAbstractSub_impl* absSubImpl = new TestAbstractSub_impl();
    TestAbstract_var absInterface = absSubImpl -> _this();

    TestOBV_impl* i = new TestOBV_impl(orb, valueFactory, valueSubFactory,
                                       customFactory, absInterface);
    TestOBV_var p = i -> _this();
    String_var ref = orb -> object_to_string(p);

    const char* refFile = "TestOBV.ref";
    ofstream out(refFile);
    if(out.fail())
    {
        cerr << argv[0] << ": can't open `" << refFile << "': "
             << strerror(errno) << endl;
        return 1;
    }

    out << ref << endl;
    out.close();

    POAManager_var manager = poa -> the_POAManager();
    manager -> activate();

    orb -> run();

    //
    // To avoid memory leaks, we need to deactivate and delete our servants
    //
    ObjectId_var id;

    id = poa -> servant_to_id(i);
    poa -> deactivate_object(id);
    delete i;

    id = poa -> servant_to_id(absSubImpl);
    poa -> deactivate_object(id);
    delete absSubImpl;

#ifdef HAVE__UNLINK
    _unlink(refFile);
#else
    unlink(refFile);
#endif

    return 0;
}
