// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <TestMacro.h>
#include <TestOBV_skel.h>
#include <TestOBV_impl.h>
#include <TestTrunc_impl.h>

using namespace CORBA;

//
// IDL:TestOBV:1.0
//
TestOBV_impl::TestOBV_impl(ORB_ptr orb, TestValue_init* valueFactory,
                           TestValueSub_init* valueSubFactory,
                           TestCustom_init* customFactory,
                           TestAbstract_ptr absInterface)
    : orb_(ORB::_duplicate(orb))
{
    value_ = valueFactory -> create(500);
    valueSub_ = valueSubFactory -> create(501, "ValueSub");
    custom_ = customFactory -> create(-99, -123456, "CustomVal", 100.997);
    absInterface_ = TestAbstract::_duplicate(absInterface);
}

TestOBV_impl::~TestOBV_impl()
{
}

//
// IDL:TestOBV/get_null_valuebase:1.0
//
ValueBase*
TestOBV_impl::get_null_valuebase()
    throw(SystemException)
{
    return (ValueBase*)0;
}

//
// IDL:TestOBV/set_null_valuebase:1.0
//
void
TestOBV_impl::set_null_valuebase(ValueBase* v)
    throw(SystemException)
{
    TEST(v == 0);
}

//
// IDL:TestOBV/get_null_valuesub:1.0
//
TestValueSub*
TestOBV_impl::get_null_valuesub()
    throw(SystemException)
{
    return (TestValueSub*)0;
}

//
// IDL:TestOBV/set_null_valuesub:1.0
//
void
TestOBV_impl::set_null_valuesub(TestValueSub* v)
    throw(SystemException)
{
    TEST(v == 0);
}

//
// IDL:TestOBV/get_abs_value1:1.0
//
TestAbsValue1*
TestOBV_impl::get_abs_value1()
    throw(SystemException)
{
    TEST(value_ -> _refcount_value() == 1);
    add_ref(value_);
    return value_;
}

//
// IDL:TestOBV/set_abs_value1:1.0
//
void
TestOBV_impl::set_abs_value1(TestAbsValue1* v)
    throw(SystemException)
{
    TEST(v != 0);
    TEST(v -> _refcount_value() == 1);
    TestValue* value = TestValue::_downcast(v);
    TEST(value != 0);
    value -> ping1();
    TEST(value -> count() == value_ -> count());
}

//
// IDL:TestOBV/get_abs_value2:1.0
//
TestAbsValue2*
TestOBV_impl::get_abs_value2()
    throw(SystemException)
{
    TEST(valueSub_ -> _refcount_value() == 1);
    add_ref(valueSub_);
    return valueSub_;
}

//
// IDL:TestOBV/set_abs_value2:1.0
//
void
TestOBV_impl::set_abs_value2(TestAbsValue2* v)
    throw(SystemException)
{
    TEST(v != 0);
    TEST(v -> _refcount_value() == 1);
    TestValueSub* value = TestValueSub::_downcast(v);
    TEST(value != 0);
    value -> ping1();
    value -> ping2();
    TEST(value -> count() == valueSub_ -> count());
    TEST(strcmp(value -> name(), valueSub_ -> name()) == 0);
}

//
// IDL:TestOBV/get_value:1.0
//
TestValue*
TestOBV_impl::get_value()
    throw(SystemException)
{
    TEST(value_ -> _refcount_value() == 1);
    add_ref(value_);
    return value_;
}

//
// IDL:TestOBV/set_value:1.0
//
void
TestOBV_impl::set_value(TestValue* v)
    throw(SystemException)
{
    TEST(v != 0);
    TEST(v -> _refcount_value() == 1);
    v -> ping1();
    TEST(v -> count() == value_ -> count());
}

//
// IDL:TestOBV/get_valuesub:1.0
//
TestValueSub*
TestOBV_impl::get_valuesub()
    throw(SystemException)
{
    TEST(valueSub_ -> _refcount_value() == 1);
    add_ref(valueSub_);
    return valueSub_;
}

//
// IDL:TestOBV/set_valuesub:1.0
//
void
TestOBV_impl::set_valuesub(TestValueSub* v)
    throw(SystemException)
{
    TEST(v != 0);
    TEST(v -> _refcount_value() == 1);
    v -> ping1();
    v -> ping2();
    TEST(v -> count() == valueSub_ -> count());
    TEST(strcmp(v -> name(), valueSub_ -> name()) == 0);
}

//
// IDL:TestOBV/get_valuesub_as_value:1.0
//
TestValue*
TestOBV_impl::get_valuesub_as_value()
    throw(SystemException)
{
    TEST(valueSub_ -> _refcount_value() == 1);
    add_ref(valueSub_);
    return valueSub_;
}

//
// IDL:TestOBV/set_valuesub_as_value:1.0
//
void
TestOBV_impl::set_valuesub_as_value(TestValue* v)
    throw(SystemException)
{
    TEST(v != 0);
    TEST(v -> _refcount_value() == 1);
    TestValueSub* value = TestValueSub::_downcast(v);
    TEST(value != 0);
    value -> ping1();
    value -> ping2();
    TEST(value -> count() == valueSub_ -> count());
    TEST(strcmp(value -> name(), valueSub_ -> name()) == 0);
}

//
// IDL:TestOBV/get_two_values:1.0
//
void
TestOBV_impl::get_two_values(TestValue*& v1, TestValue*& v2)
    throw(SystemException)
{
    TEST(value_ -> _refcount_value() == 1);
    add_ref(value_);
    add_ref(value_);
    v1 = value_.in();
    v2 = value_.in();
}

//
// IDL:TestOBV/set_two_values:1.0
//
void
TestOBV_impl::set_two_values(TestValue* v1, TestValue* v2)
    throw(SystemException)
{
    TEST(v1 != 0);
    TEST(v2 != 0);
    TEST(v1 == v2);
    TEST(v1 -> _refcount_value() == 2);
    v1 -> ping1();
    v2 -> ping1();
    TEST(v1 -> count() == value_ -> count());
}

//
// IDL:TestOBV/get_two_valuesubs_as_values:1.0
//
void
TestOBV_impl::get_two_valuesubs_as_values(TestValue*& v1, TestValue*& v2)
    throw(SystemException)
{
    //
    // Lookup the factory and create two distinct TestValueSub instances
    //

    ValueFactoryBase_var factory;
    factory = orb_ -> lookup_value_factory("IDL:TestValueSub:1.0");
    TEST(factory.in() != 0);
    TestValueSub_init* subFactory = TestValueSub_init::_downcast(factory);
    TEST(subFactory != 0);

    v1 = subFactory -> create(999, "ValueSub");
    v2 = subFactory -> create(999, "ValueSub");
}

//
// IDL:TestOBV/set_two_valuesubs_as_values:1.0
//
void
TestOBV_impl::set_two_valuesubs_as_values(TestValue* v1, TestValue* v2)
    throw(SystemException)
{
    TEST(v1 != 0);
    TEST(v2 != 0);
    TEST(v1 -> _refcount_value() == 1);
    TEST(v2 -> _refcount_value() == 1);
    v1 -> ping1();
    v2 -> ping1();
    TEST(v1 -> count() == v2 -> count());

    TestValueSub* s1 = TestValueSub::_downcast(v1);
    TEST(s1 != 0);
    s1 -> ping2();
    TestValueSub* s2 = TestValueSub::_downcast(v2);
    TEST(s2 != 0);
    s2 -> ping2();
    TEST(strcmp(s1 -> name(), s2 -> name()) == 0);
}

//
// IDL:TestOBV/get_custom:1.0
//
TestCustom*
TestOBV_impl::get_custom()
    throw(SystemException)
{
    TEST(custom_ -> _refcount_value() == 1);
    add_ref(custom_);
    return custom_;
}

//
// IDL:TestOBV/set_custom:1.0
//
void
TestOBV_impl::set_custom(TestCustom* v)
    throw(SystemException)
{
    TEST(v != 0);
    TEST(v -> _refcount_value() == 1);
    v -> ping1();
    TEST(v -> shortVal() == custom_ -> shortVal());
    TEST(v -> longVal() == custom_ -> longVal());
    TEST(v -> doubleVal() == custom_ -> doubleVal());
    TEST(strcmp(v -> stringVal(), custom_ -> stringVal()) == 0);
}

//
// IDL:TestOBV/get_abs_custom:1.0
//
TestAbsValue1*
TestOBV_impl::get_abs_custom()
    throw(SystemException)
{
    TEST(custom_ -> _refcount_value() == 1);
    add_ref(custom_);
    return custom_;
}

//
// IDL:TestOBV/set_abs_custom:1.0
//
void
TestOBV_impl::set_abs_custom(TestAbsValue1* v)
    throw(SystemException)
{
    TEST(v != 0);
    TEST(v -> _refcount_value() == 1);
    v -> ping1();
    TestCustom* c = TestCustom::_downcast(v);
    TEST(c != 0);
    TEST(c -> shortVal() == custom_ -> shortVal());
    TEST(c -> longVal() == custom_ -> longVal());
    TEST(c -> doubleVal() == custom_ -> doubleVal());
    TEST(strcmp(c -> stringVal(), custom_ -> stringVal()) == 0);
}

//
// IDL:TestOBV/get_string_box:1.0
//
TestStringBox*
TestOBV_impl::get_string_box(const char* value)
    throw(SystemException)
{
    return new TestStringBox(value);
}

//
// IDL:TestOBV/set_string_box:1.0
//
void
TestOBV_impl::set_string_box(TestStringBox* b, const char* value)
    throw(SystemException)
{
    TEST(b != 0);
    TEST(b -> _refcount_value() == 1);
    TEST(strcmp(b -> _value(), value) == 0);
}

//
// IDL:TestOBV/get_ulong_box:1.0
//
TestULongBox*
TestOBV_impl::get_ulong_box(ULong value)
    throw(SystemException)
{
    return new TestULongBox(value);
}

//
// IDL:TestOBV/set_ulong_box:1.0
//
void
TestOBV_impl::set_ulong_box(TestULongBox* b,
                            ULong value)
    throw(SystemException)
{
    TEST(b != 0);
    TEST(b -> _refcount_value() == 1);
    TEST(b -> _value() == value);
}

//
// IDL:TestOBV/get_fix_struct_box:1.0
//
TestFixStructBox*
TestOBV_impl::get_fix_struct_box(const TestFixStruct& value)
    throw(SystemException)
{
    return new TestFixStructBox(value);
}

//
// IDL:TestOBV/set_fix_struct_box:1.0
//
void
TestOBV_impl::set_fix_struct_box(TestFixStructBox* b,
                                 const TestFixStruct& value)
    throw(SystemException)
{
    TEST(b != 0);
    TEST(b -> _refcount_value() == 1);
    TEST(b -> x() == value.x);
    TEST(b -> y() == value.y);
    TEST(b -> radius() == value.radius);
}

//
// IDL:TestOBV/get_var_struct_box:1.0
//
TestVarStructBox*
TestOBV_impl::get_var_struct_box(const TestVarStruct& value)
    throw(SystemException)
{
    return new TestVarStructBox(value);
}

//
// IDL:TestOBV/set_var_struct_box:1.0
//
void
TestOBV_impl::set_var_struct_box(TestVarStructBox* b,
                                 const TestVarStruct& value)
    throw(SystemException)
{
    TEST(b != 0);
    TEST(b -> _refcount_value() == 1);
    TEST(strcmp(b -> name(), value.name) == 0);
    TEST(strcmp(b -> email(), value.email) == 0);
}

//
// IDL:TestOBV/get_fix_union_box:1.0
//
TestFixUnionBox*
TestOBV_impl::get_fix_union_box(const TestFixUnion& value)
    throw(SystemException)
{
    return new TestFixUnionBox(value);
}

//
// IDL:TestOBV/set_fix_union_box:1.0
//
void
TestOBV_impl::set_fix_union_box(TestFixUnionBox* b, const TestFixUnion& value)
    throw(SystemException)
{
    TEST(b != 0);
    TEST(b -> _refcount_value() == 1);

    switch(value._d())
    {
    case TRUE:
        TEST(b -> o() == value.o());
        break;

    case FALSE:
        TEST(b -> d() == value.d());
        break;
    }
}

//
// IDL:TestOBV/get_var_union_box:1.0
//
TestVarUnionBox*
TestOBV_impl::get_var_union_box(const TestVarUnion& value)
    throw(SystemException)
{
    return new TestVarUnionBox(value);
}

//
// IDL:TestOBV/set_var_union_box:1.0
//
void
TestOBV_impl::set_var_union_box(TestVarUnionBox* b, const TestVarUnion& value)
    throw(SystemException)
{
    TEST(b != 0);
    TEST(b -> _refcount_value() == 1);

    switch(value._d())
    {
    case 0:
        TEST(strcmp(b -> s(), value.s()) == 0);
        break;

    case 9:
    {
        const TestFixStruct& fs1 = b -> fs();
        const TestFixStruct& fs2 = value.fs();
        TEST(fs1.x == fs2.x);
        TEST(fs1.y == fs2.y);
        TEST(fs1.radius == fs2.radius);
        break;
    }
    }
}

//
// IDL:TestOBV/get_anon_seq_box:1.0
//
TestAnonSeqBox*
TestOBV_impl::get_anon_seq_box(ULong length)
    throw(SystemException)
{
    TestAnonSeqBox* result = new TestAnonSeqBox;
    result -> length(length);
    for(ULong i = 0 ; i < length ; i++)
        (*result)[i] = (Short)i;
    return result;
}

//
// IDL:TestOBV/set_anon_seq_box:1.0
//
void
TestOBV_impl::set_anon_seq_box(TestAnonSeqBox* b, ULong length)
    throw(SystemException)
{
    TEST(b != 0);
    TEST(b -> _refcount_value() == 1);
    TEST(b -> length() == length);
    for(ULong i = 0 ; i < length ; i++)
        TEST((*b)[i] == (Short)i);
}

//
// IDL:TestOBV/get_string_seq_box:1.0
//
TestStringSeqBox*
TestOBV_impl::get_string_seq_box(const TestStringSeq& value)
    throw(SystemException)
{
    return new TestStringSeqBox(value);
}

//
// IDL:TestOBV/set_string_seq_box:1.0
//
void
TestOBV_impl::set_string_seq_box(TestStringSeqBox* b,
                                 const TestStringSeq& value)
    throw(SystemException)
{
    TEST(b != 0);
    TEST(b -> _refcount_value() == 1);
    TEST(b -> length() == value.length());
    for(ULong i = 0 ; i < b -> length() ; i++)
        TEST(strcmp((*b)[i], value[i]) == 0);
}

//
// IDL:TestOBV/get_ai_interface:1.0
//
TestAbstract_ptr
TestOBV_impl::get_ai_interface()
    throw(SystemException)
{
    return TestAbstract::_duplicate(absInterface_);
}

// 
// IDL:TestOBV/set_ai_interface:1.0
//
void 
TestOBV_impl::set_ai_interface(TestAbstract_ptr a)
    throw(SystemException)
{
    if(!is_nil(a))
    {
        a -> abstract_op();
    }
}

//
// IDL:TestOBV/get_ai_interface_any:1.0
//
Any*
TestOBV_impl::get_ai_interface_any()
    throw(SystemException)
{
    Any* result = new Any;
    *result <<= absInterface_;

    //
    // Test local any extraction
    //
    AbstractBase_ptr base;
    Any::to_abstract_base to_base(base);
    assert(*result >>= to_base);
    assert(!is_nil(base));
    TestAbstract_var ab = TestAbstract::_narrow(base);
    assert(!is_nil(ab));
    ab -> abstract_op();

    return result;
}

//
// IDL:TestOBV/set_ai_interface_any:1.0
//
void 
TestOBV_impl::set_ai_interface_any(const Any& a)
    throw(SystemException)
{
    //
    // Test remote any extraction
    //
    Object_ptr obj;
    Any::to_object to_obj(obj);
    assert(a >>= to_obj);
    assert(!is_nil(obj));
    TestAbstractSub_var sub = TestAbstractSub::_narrow(obj);
    assert(!is_nil(sub));
    sub -> abstract_op();
    sub -> sub_op();
}

//
// IDL:TestOBV/get_trunc1:1.0
//
TestTruncBase*
TestOBV_impl::get_trunc1()
    throw(SystemException)
{
    //
    // This test addresses several issues:
    //
    // 1) truncation
    // 2) skipping chunks and nested values during truncation
    // 3) nested value with repository ID information
    //

    TestTrunc1_var t1 = new TestTrunc1_impl;
    t1 -> cost((Float)1.993);
    t1 -> boolVal(TRUE);
    ValueBase_var vb = value_ -> _copy_value();
    TestValue* v = TestValue::_downcast(vb);
    v -> count(999);
    t1 -> v(v);
    t1 -> shortVal(12667);
    return t1._retn();
}

//
// IDL:TestOBV/get_trunc2:1.0
//
TestTruncBase*
TestOBV_impl::get_trunc2()
    throw(SystemException)
{
    //
    // This test addresses several issues:
    //
    // 1) truncation - skipping chunks, nested values
    // 2) nested values with repository ID information
    // 3) cotermination of outer and nested values
    // 4) nested truncatable values
    // 5) indirection of repository ID information
    // 6) recursion
    // 7) value indirection into truncated portion of nested value
    //

    TestTrunc2_var t2 = new TestTrunc2_impl;
    t2 -> cost((Float)5.993);

    TestTrunc1_var t1 = new TestTrunc1_impl;
    t1 -> cost((Float)1.993);
    t1 -> boolVal(TRUE);
    ValueBase_var vb = value_ -> _copy_value();
    TestValue* v = TestValue::_downcast(vb);
    v -> count(999);
    t1 -> v(v);
    t1 -> shortVal(12667);
    t2 -> t(t1); // issues 2, 4

    vb = value_ -> _copy_value();
    TestValue* v2 = TestValue::_downcast(vb);
    v2 -> count(9999);
    t2 -> a(v2); // issues 2, 5

    t2 -> b1(t2); // issues 2, 4, 5, 6

    t2 -> v(v); // issue 7

    t2 -> b2(t1);

    return t2._retn();
}

//
// IDL:TestOBV/deactivate:1.0
//
void
TestOBV_impl::deactivate()
    throw(SystemException)
{
    orb_ -> shutdown(false);
}
