// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>

#include <Test.h>

#include <stdlib.h>
#include <errno.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;

int
main(int argc, char* argv[], char*[])
{
    try
    {
        //
        // Create ORB
        //
        ORB_var orb = ORB_init(argc, argv);
        
        //
        // Read all object references from file
        //
        const char* refFile = "Test.ref";
        ifstream in; // Must use open(name), not ifstream in(name) (VC++ bug)
        in.open(refFile);
        if(in.fail())
        {
            cerr << argv[0] << ": can't open `" << refFile << "': "
                 << strerror(errno) << endl;
            return 1;
        }

        char s[1000];
        in >> s;
	in.close();

        Object_var obj = orb -> string_to_object(s);
        assert(!is_nil(obj));
        TestServer_var server = TestServer::_narrow(obj);
        assert(!is_nil(server));

        TestInfoSeq_var info = server -> get_info();

        for(ULong i = 0 ; i < info -> length() ; i++)
        {
            try
            {
                info[i].obj -> aMethod();
                assert(strlen(info[i].except_id) == 0);
            }
            catch(const SystemException& ex)
            {
                assert(strcmp(info[i].except_id, ex._OB_typeId()) == 0);
            }
        }

        try
        {
            server -> deactivate();
        }
        catch(const COMM_FAILURE&)
        {
            // ignore
        }
    }
    catch(const SystemException& ex)
    {
        OB::PrintException(ex);
        return 1;
    }

    return 0;
}
