// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>

#include <TestUtil.h>

#include <assert.h>
#include <string.h>

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;
using namespace PortableServer;

void
TestCreate(ORB_ptr orb, POA_ptr root)
{
    Object_var obj;
    PolicyList policies;
    POA_var poa, parent, poa2, poa3;
    POAManager_var mgr;
    String_var str;

    POAManager_var rootMgr = root -> the_POAManager();
    assert(!is_nil(rootMgr));

    //
    // Test: POAManager should be in HOLDING state
    //
    assert(rootMgr -> get_state() == PortableServer::POAManager::HOLDING);

    //
    // Create child POA
    //
    poa = root -> create_POA("poa1", PortableServer::POAManager::_nil(),
                             policies);

    //
    // Test: POAManager should NOT be the same as the root's manager
    //
    mgr = poa -> the_POAManager();
    assert(!mgr -> _is_equivalent(rootMgr));

    //
    // Test: POAManager should be in HOLDING state
    //
    assert(mgr -> get_state() == PortableServer::POAManager::HOLDING);

    //
    // Test: Confirm name
    //
    str = poa -> the_name();
    assert(strcmp("poa1", str) == 0);

    //
    // Test: Confirm parent
    //
    parent = poa -> the_parent();
    assert(parent -> _is_equivalent(root));

    //
    // Test: AdapterAlreadyExists exception
    //
    try
    {
        poa2 = root -> create_POA("poa1", PortableServer::POAManager::_nil(),
                                  policies);
        assert("create_POA should not have succeeded" == 0);
    }
    catch(const POA::AdapterAlreadyExists&)
    {
        // expected
    }

    //
    // Test: XXX InvalidPolicy exception
    //

    //
    // Create another child of root POA
    //
    poa2 = root -> create_POA("poa2", rootMgr, policies);

    //
    // Test: POAManager should be the same as the root's manager
    //
    mgr = poa2 -> the_POAManager();
    assert(mgr -> _is_equivalent(rootMgr));

    //
    // Create child of child POA
    //
    poa3 = poa2 -> create_POA("child", rootMgr, policies);

    //
    // Test: Confirm parent
    //
    parent = poa3 -> the_parent();
    assert(parent -> _is_equivalent(poa2));

    poa -> destroy(TRUE, TRUE);
    poa2 -> destroy(TRUE, TRUE);
}

int
main(int argc, char* argv[], char*[])
{
    int status;

    try
    {
        //
        // Create ORB
        //
        ORB_var orb = ORB_init(argc, argv);

        POA_var root = TestUtil_GetRootPOA(orb);

        cout << "Testing POA::create_POA()... " << flush;
        TestCreate(orb, root);
        cout << "OK!" << endl;
    }
    catch(const UserException&)
    {
        cerr << argv[0] << ": Unexpected user exception was thrown" << endl;
        status = 1;
    }
    catch(const SystemException& ex)
    {
        OB::PrintException(ex);
        status = 1;
    }

    return status;
}
