// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <TestDSI_impl.h>

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;
using namespace PortableServer;

TestDSI_impl::TestDSI_impl(ORB_ptr orb, const char* name, Boolean compare)
    : orb_(ORB::_duplicate(orb)), name_(string_dup(name)), compare_(compare),
      defaultServant_(false)
{
}

TestDSI_impl::~TestDSI_impl()
{
    //cout << "TestDSI_impl destroyed!" << endl;
}

void
TestDSI_impl::invoke(ServerRequest_ptr request)
{
    String_var name = request -> op_name();

    if(strcmp(name, "aMethod") != 0)
    {
        NVList_ptr list;
        Any* any;
    
        orb_ -> create_list(0, list);
        request -> params(list);
    
        any = new Any;
        *any <<= new BAD_OPERATION();
        request -> exception(any);
        return;
    }
    
    Object_var currentObj;

    try
    {
        currentObj = orb_ -> resolve_initial_references("POACurrent");
    }
    catch(const UserException&)
    {
    }

    PortableServer::Current_var current;
    if(!is_nil(currentObj))
        current = PortableServer::Current::_narrow(currentObj);
    assert(!is_nil(current));

    ObjectId_var oid = current -> get_object_id();
    String_var oidString = ObjectId_to_string(oid);

    if(compare_)
        assert(strcmp(oidString, name_) == 0);
    if(defaultServant_)
    {
	PortableServer::POA_var poa = current -> get_POA();
	ObjectId_var servantId = poa -> servant_to_id(this);
	assert(servantId -> length() == oid -> length());
	assert(memcmp(servantId -> get_buffer(), oid -> get_buffer(),
		      oid -> length()) == 0);
    }
    
    //cout << "DSI: aMethod" << endl;
}
    
RepositoryId
TestDSI_impl::_primary_interface(const ObjectId& id, POA_ptr poa)
{
    return string_dup("IDL:Test:1.0");
}
