// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>

#include <Test_skel.h>
#include <Test_impl.h>
#include <TestDSI_impl.h>
#include <TestServer_impl.h>
#include <TestUtil.h>

#include <stdlib.h>
#include <errno.h>

#ifdef HAVE_UNISTD_H
#   include <unistd.h>
#endif

#ifdef HAVE_IO_H
#   include <io.h>
#endif

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;
using namespace PortableServer;

int
main(int argc, char* argv[], char*[])
{
    try
    {
        //
        // Create ORB
        //
        ORB_var orb = ORB_init(argc, argv);

        POA_var rootPoa = TestUtil_GetRootPOA(orb);
        POAManager_var manager = rootPoa -> the_POAManager();

        CORBA::PolicyList policies(6);
        policies.length(6);
        policies[0] = rootPoa ->
            create_lifespan_policy(PortableServer::PERSISTENT);
        policies[1] = rootPoa ->
            create_id_assignment_policy(PortableServer::USER_ID);
        policies[2] = rootPoa ->
         create_request_processing_policy(PortableServer::USE_DEFAULT_SERVANT);
        policies[3] = rootPoa ->
            create_servant_retention_policy(PortableServer::RETAIN);
        policies[4] = rootPoa -> create_implicit_activation_policy(
            PortableServer::NO_IMPLICIT_ACTIVATION);
        policies[5] = rootPoa ->
            create_id_uniqueness_policy(PortableServer::MULTIPLE_ID);

        POA_var persistentPOA =
            rootPoa -> create_POA("persistent", manager, policies);

        //
        // Use a DSI servant as the default
        //
        TestDSI_impl* defaultServant = new TestDSI_impl(orb);
	defaultServant -> setDefaultServant(true);

        persistentPOA -> set_servant(defaultServant);

        //
        // Create some servants
        //
        Test_impl* test1 = new Test_impl(orb, "test1");
        Test_impl* test2 = new Test_impl(orb, "test2");

        //
        // Create ObjectIds
        //
        ObjectId_var oid1 = string_to_ObjectId("test1");
        ObjectId_var oid2 = string_to_ObjectId("test2");
        ObjectId_var oid3 = string_to_ObjectId("testDefault");

        //
        // Activate servants
        //
        persistentPOA -> activate_object_with_id(oid1, test1);
        persistentPOA -> activate_object_with_id(oid2, test2);

        //
        // Create object references
        //
        Object_var reference1 =
            persistentPOA -> create_reference_with_id(oid1, "IDL:Test:1.0");
        Object_var reference2 =
            persistentPOA -> create_reference_with_id(oid2, "IDL:Test:1.0");
        Object_var reference3 =
            persistentPOA -> create_reference_with_id(oid3, "IDL:Test:1.0");

        //
        // Create server
        //
        TestInfoSeq info;
        info.length(3);
        info[0].obj = Test::_narrow(reference1);
        info[0].except_id = string_dup("");
        info[1].obj = Test::_narrow(reference2);
        info[1].except_id = string_dup("");
        info[2].obj = Test::_narrow(reference3);
        info[2].except_id = string_dup("");
        TestServer_impl* serverImpl = new TestServer_impl(orb, info);
        TestServer_var server = serverImpl -> _this();

        //
        // Save reference
        //
        const char* refFile = "Test.ref";
        ofstream out(refFile);
        if(out.fail())
        {
            cerr << argv[0] << ": can't open `" << refFile << "': "
                 << strerror(errno) << endl;
            return 1;
        }

        String_var ref = orb -> object_to_string(server);
        out << ref << endl;
        out.close();

        //
        // Run implementation
        //
        manager -> activate();
        orb -> run();

        //
        // Clean up
        //
        persistentPOA -> destroy(TRUE, TRUE);
        defaultServant -> _remove_ref();
	serverImpl -> _remove_ref();
        delete test1;
        delete test2;

#ifdef HAVE__UNLINK
        _unlink(refFile);
#else
        unlink(refFile);
#endif
    }
    catch(const UserException&)
    {
        abort();
        return 1;
    }
    catch(const SystemException& ex)
    {
        OB::PrintException(ex);
        return 1;
    }

    return 0;
}
