// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>

#include <TestUtil.h>

#include <assert.h>
#include <string.h>

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;
using namespace PortableServer;

void
TestDestroy(ORB_ptr orb, POA_ptr root)
{
    Object_var obj;
    PolicyList policies;
    POA_var poa, parent, poa2, poa3;
    POAManager_var mgr;
    String_var str;

    POAManager_var rootMgr = root -> the_POAManager();
    assert(!is_nil(rootMgr));

    //
    // Create child POA
    //
    poa = root -> create_POA("poa1", rootMgr, policies);

    //
    // Test: destroy
    //
    poa -> destroy(TRUE, TRUE);

    //
    // Ensure parent no longer knows about child
    //
    try
    {
        root -> find_POA("poa1", FALSE);
        assert("find_POA should not have succeeded" == 0);
    }
    catch(const POA::AdapterNonExistent&)
    {
        // expected
    }

    //
    // Create child POA
    //
    poa = root -> create_POA("poa1", rootMgr, policies);

    //
    // Create child of child POA
    //
    poa2 = poa -> create_POA("child1", rootMgr, policies);

    //
    // Test: destroy - should destroy poa1 and poa1/child1
    //
    poa -> destroy(TRUE, TRUE);

    //
    // Ensure parent no longer knows about child
    //
    try
    {
        root -> find_POA("poa1", FALSE);
        assert("find_POA should not have succeeded" == 0);
    }
    catch(const POA::AdapterNonExistent&)
    {
        // expected
    }

    //
    // XXX Test: etherialize w/ servant manager
    //
}

int
main(int argc, char* argv[], char*[])
{
    int status;

    try
    {
        //
        // Create ORB
        //
        ORB_var orb = ORB_init(argc, argv);

        POA_var root = TestUtil_GetRootPOA(orb);

        cout << "Testing POA::destroy()..." << flush;
        TestDestroy(orb, root);
        cout << "OK!" << endl;
    }
    catch(const UserException&)
    {
        cerr << argv[0] << ": Unexpected user exception was thrown" << endl;
        status = 1;
    }
    catch(const SystemException& ex)
    {
        OB::PrintException(ex);
        status = 1;
    }

    return status;
}
