// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>

#include <TestUtil.h>

#include <assert.h>
#include <string.h>

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;
using namespace PortableServer;

void
TestFind(ORB_ptr orb, POA_ptr root)
{
    Object_var obj;
    PolicyList policies;
    POA_var poa, parent, poa2, poa3;
    POAManager_var mgr;
    String_var str;

    POAManager_var rootMgr = root -> the_POAManager();
    assert(!is_nil(rootMgr));

    //
    // Create child POA
    //
    poa = root -> create_POA("poa1", rootMgr, policies);

    //
    // Test: find_POA
    //
    poa2 = root -> find_POA("poa1", FALSE);
    assert(!is_nil(poa2));
    assert(poa2 -> _is_equivalent(poa));

    //
    // Test: AdapterNonExistent exception
    //
    try
    {
        poa2 = root -> find_POA("poaX", FALSE);
        assert("find_POA should not have succeeded" == 0);
    }
    catch(const POA::AdapterNonExistent&)
    {
        // expected
    }

    //
    // Create child POA
    //
    poa2 = root -> create_POA("poa2", rootMgr, policies);

    //
    // Test: Confirm parent knows about child
    //
    poa3 = root -> find_POA("poa2", FALSE);
    assert(!is_nil(poa3));
    assert(poa3 -> _is_equivalent(poa2));
}

int
main(int argc, char* argv[], char*[])
{
    int status;

    try
    {
        //
        // Create ORB
        //
        ORB_var orb = ORB_init(argc, argv);

        POA_var root = TestUtil_GetRootPOA(orb);

        cout << "Testing POA::find_POA()... " << flush;
        TestFind(orb, root);
        cout << "OK!" << endl;
    }
    catch(const UserException&)
    {
        cerr << argv[0] << ": Unexpected user exception was thrown" << endl;
        status = 1;
    }
    catch(const SystemException& ex)
    {
        OB::PrintException(ex);
        status = 1;
    }

    return status;
}
