// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>

#include <TestUtil.h>

#include <TestLocationForwardActivator_skel.h>
#include <TestLocationForwardActivator_impl.h>
#include <Test_impl.h>
#include <TestLocationForward_impl.h>

#include <stdlib.h>
#include <errno.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;
using namespace PortableServer;

int
main(int argc, char* argv[], char*[])
{
    int status;

    try
    {
        //
        // Create ORB
        //
        ORB_var orb = ORB_init(argc, argv);

        POA_var root = TestUtil_GetRootPOA(orb);

        POA_var poa;
        PolicyList policies;

        POAManager_var manager = root -> the_POAManager();

        //
        // Create POA
        //
        policies.length(4);
        policies[0] =
            root -> create_id_assignment_policy(PortableServer::USER_ID);
        policies[1] =
            root -> create_id_uniqueness_policy(PortableServer::UNIQUE_ID);
        policies[2] =
            root -> create_servant_retention_policy(PortableServer::RETAIN);
        policies[3] = root ->
         create_request_processing_policy(PortableServer::USE_SERVANT_MANAGER);
        poa = root -> create_POA("poa", manager, policies);

        TestLocationForwardActivator_impl activatorImpl;
        TestLocationForwardActivator_var activator = activatorImpl._this();
        poa -> set_servant_manager(activator);

        ObjectId_var oid = string_to_ObjectId("test");
        Object_var reference =
            poa -> create_reference_with_id(oid, "IDL:Test:1.0");

        //
        // Read all object references from file
        //
        const char* refFile = "Test.ref";
        ifstream in; // Must use open(name), not ifstream in(name) (VC++ bug)
        in.open(refFile);
        if(in.fail())
        {
            cerr << argv[0] << ": can't open `" << refFile << "': "
                 << strerror(errno) << endl;
            return 1;
        }

        char s[1000];
        in >> s;
        in.close();

        Object_var obj = orb -> string_to_object(s);
        assert(!is_nil(obj));
        TestLocationForwardServer_var server =
            TestLocationForwardServer::_narrow(obj);
        assert(!is_nil(server));
        Object_var servant = server -> get_servant();

        activator -> setForwardRequest(servant);

        TestLocationForward_impl testImpl(orb);
        activatorImpl.setActivatedServant(&testImpl);

        manager -> activate();

        TestLocationForwardActivator_var remoteActivator =
            server -> get_activator();
        assert(!is_nil(remoteActivator));
        remoteActivator -> setForwardRequest(reference);

        //
        // Run some calls
        //
        TestLocationForward_var local =
            TestLocationForward::_narrow(reference);

        //
        // First should be local
        //
        local -> aMethod();
        local -> deactivate_servant();

        //
        // Second, should be remote
        //
        local -> aMethod();
        local -> deactivate_servant();

        //
        // Third should be local again
        //
        local -> aMethod();
        local -> deactivate_servant();

        //
        // Clean up
        //
        poa -> destroy(TRUE, TRUE);

        //
        // Since activatorImpl is a stack-based servant, we need to deactivate
        // it before it goes out of scope
        //
        ObjectId_var id = root -> servant_to_id(&activatorImpl);
        root -> deactivate_object(id);

        try
        {
            server -> deactivate();
        }
        catch(const COMM_FAILURE&)
        {
            // ignore
        }
    }
    catch(const SystemException& ex)
    {
        OB::PrintException(ex);
        status = 1;
    }

    return status;
}
