// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>

#include <TestUtil.h>

#include <TestLocationForwardActivator_skel.h>
#include <TestLocationForwardActivator_impl.h>
#include <Test_impl.h>
#include <TestLocationForward_impl.h>

#include <stdlib.h>
#include <errno.h>

#ifdef HAVE_UNISTD_H
#   include <unistd.h>
#endif

#ifdef HAVE_IO_H
#   include <io.h>
#endif

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;
using namespace PortableServer;

class Server_impl : virtual public POA_TestLocationForwardServer
{
    CORBA::ORB_var orb_;
    TestLocationForwardActivator_var activator_;
    Object_var servant_;

public:
    Server_impl(ORB_ptr orb, TestLocationForwardActivator_ptr activator,
                Object_ptr servant)
        : orb_(ORB::_duplicate(orb)),
          activator_(TestLocationForwardActivator::_duplicate(activator)),
          servant_(Object::_duplicate(servant))
    {
    }

    virtual TestLocationForwardActivator_ptr get_activator()
        throw(SystemException)
    {
        return TestLocationForwardActivator::_duplicate(activator_);
    }

    virtual Object_ptr get_servant()
        throw(SystemException)
    {
        return Object::_duplicate(servant_);
    }

    virtual void deactivate()
        throw(SystemException)
    {
        orb_ -> shutdown(false);
    }
};


int
main(int argc, char* argv[], char*[])
{
    int status;

    try
    {
        //
        // Create ORB
        //
        ORB_var orb = ORB_init(argc, argv);

        POA_var root = TestUtil_GetRootPOA(orb);

        POA_var poa;
        PolicyList policies;

        POAManager_var manager = root -> the_POAManager();

        //
        // Create POAs
        //
        policies.length(4);
        policies[0] =
            root -> create_id_assignment_policy(PortableServer::USER_ID);
        policies[1] =
            root -> create_id_uniqueness_policy(PortableServer::UNIQUE_ID);
        policies[2] =
            root -> create_servant_retention_policy(PortableServer::RETAIN);
        policies[3] = root ->
         create_request_processing_policy(PortableServer::USE_SERVANT_MANAGER);
        poa = root -> create_POA("poa", manager, policies);

        TestLocationForwardActivator_impl activatorImpl;
        TestLocationForwardActivator_var activator = activatorImpl._this();
        poa -> set_servant_manager(activator);

        ObjectId_var oid = string_to_ObjectId("test");
        Object_var obj = poa -> create_reference_with_id(oid, "IDL:Test:1.0");
        TestLocationForward_impl testImpl(orb);

        activatorImpl.setActivatedServant(&testImpl);

        Server_impl serverImpl(orb, activator, obj);
        TestLocationForwardServer_var server = serverImpl._this();

        //
        // Save reference
        //
        const char* refFile = "Test.ref";
        ofstream out(refFile);
        if(out.fail())
        {
            cerr << argv[0] << ": can't open `" << refFile << "': "
                 << strerror(errno) << endl;
            return 1;
        }

        String_var ref = orb -> object_to_string(server);
        out << ref << endl;
        out.close();

        //
        // Run implementation
        //
        manager -> activate();
        orb -> run();

        //
        // Clean up
        //
        poa -> destroy(TRUE, TRUE);

        //
        // Deactivate stack-based servants before they go out of scope
        //
        ObjectId_var id;
        id = root -> servant_to_id(&activatorImpl);
        root -> deactivate_object(id);
        id = root -> servant_to_id(&serverImpl);
        root -> deactivate_object(id);

#ifdef HAVE__UNLINK
        _unlink(refFile);
#else
        unlink(refFile);
#endif
    }
    catch(const SystemException& ex)
    {
        OB::PrintException(ex);
        status = 1;
    }

    return status;
}
